//
// Parse, store, and print parameters specific to the cluster environment.
//

using System;
using System.Collections;
using System.Diagnostics;

using SkipNet=Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Common;

using Microsoft.Research.Taro;
using Microsoft.Research.Herald.ClusterLib;

namespace Microsoft.Research.Herald.ClusterLib
{
    public class ClusterLibParameters
    {
        public static int nodes = 200;
        public static int nodeNumber;

        public static ClusterNode clusterNode;

        public static void Init(ClusterNode clusterNode)
        {
            ClusterLibParameters.clusterNode = clusterNode;
            ClusterNodeTable.ReadClusterNodeTable();
        }

        // Parse cluster-specific command line arguments from the specified index,
        // returning the index of the first unparsed parameter.
        // Returns -1 if Usage() needs to be called.
        public static int ParseArgs(string [] args, int starting_at)
        {
            int i;
            if ((i = clusterNode.ProcessParameters(args, starting_at)) < 0) // Parse node name, seed node, etc.
                return -1;

            // Parse additional cluster-specific parameters
            for (; i < args.Length; i++) 
            {
                if ( (args[i][0] == '-') || (args[i][0] == '/') ) 
                {
                    string opt = args[i].Substring(1);
                    if (opt.Length < 1) break;

                    if ("node".StartsWith(opt)) // -node 0 of 30
                    {
                        if (i + 3 >= args.Length) return -1;
                        nodeNumber = Convert.ToInt32(args[++i]);
                        if (! args[++i].Equals("of")) return -1;
                        nodes = Convert.ToInt32(args[++i]);
                    } 
                    else if ("xmlserializer".StartsWith(opt)) // -xmlserializer
                    {
                        clusterNode.useXmlSerializer = true;
                    }
                    else if ("binaryformatter".StartsWith(opt)) // -binaryformatter
                    {
                        clusterNode.useXmlSerializer = false;
                    }
                    else if ("compressed".StartsWith(opt)) // -compressed
                    {
                        Taro.TaroSocketConfig.USE_COMPRESSION = true;
                    }
                    else if ("uncompressed".StartsWith(opt)) // -uncompressed
                    {
                        Taro.TaroSocketConfig.USE_COMPRESSION = false;
                    }
                    else
                        break;
                }
                else
                    break;
            }

            return i;
        }
        
        // Print the common portion of the usage message
        public static void PrintUsage()
        {
            clusterNode.Usage();
            Console.WriteLine("  Cluster Configuration Parameters:");
            Console.WriteLine("    [-node this_node_index_starting_from_zero of total_nodes]");
            Console.WriteLine("    [-binaryformatter] -- uses binary formatter instead of the XML serializer");
            Console.WriteLine("    [-xmlserializer] -- uses XML serializer instead of the binary formatter");
            Console.WriteLine("    [-uncompressed] -- does not compress serialized messages");
            Console.WriteLine("    [-compressed] -- compresses serialized messages");
        }

        // Set the parameter values based on the arguments selected
        public static void SetParameters()
        {
            // Validate current node's parameters against its table entry
            if (nodes > ClusterNodeTable.clusterNodeTable.Length || nodeNumber >= nodes) 
            {
                Console.WriteLine("Insufficient ClusterNodeTable entries {0} for node count {1} on node {2}",
                    ClusterNodeTable.clusterNodeTable.Length, nodes, nodeNumber);
                Environment.Exit(-1);
            }
            ClusterNodeTableEntry entry = ClusterNodeTable.clusterNodeTable[nodeNumber];
            if (! Environment.MachineName.ToLower().Equals(entry.machineName.ToLower()) ||
                ! clusterNode.port.ToString().Equals(entry.portNumber) ||
                ! clusterNode.nodeNameSuffix.Equals("/" + entry.logicalNodeName + "/"))
            {
                Console.WriteLine("Mismatch between node parameters and ClusterNodeTable[{0}] entry:", nodeNumber);
                Console.WriteLine("  Node parameters: machine {0} port {1} suffix {2}",
                    Environment.MachineName, clusterNode.port, clusterNode.nodeNameSuffix);
                Console.WriteLine("  Table entry[{0}]: machine {1} port {2} suffix {3}",
                    nodeNumber, entry.machineName, entry.portNumber, "/" + entry.logicalNodeName + "/");
                Environment.Exit(-1);
            }
            if (clusterNode.socketMode)
            {
                string ipAddress = clusterNode.nodeAddr.Substring(0, clusterNode.nodeAddr.LastIndexOf(':'));
                if (! ipAddress.Equals(entry.ipAddress)) 
                {
                    Console.WriteLine("Mismatch between node address {0} and ClusterNodeTable[{1}] address {2}",
                        ipAddress, nodeNumber, entry.ipAddress);
                    Environment.Exit(-1);
                }
            }
        }

        // Print the arguments selected and information about them
        public static void PrintArgs()
        {
            clusterNode.PrintArgs();
            Console.WriteLine("  Cluster Configuration Parameters:");
            Console.WriteLine("    -node {0} of {1}", nodeNumber, nodes);
            if (clusterNode.useXmlSerializer)
                Console.WriteLine("    -xmlserializer");
            else
                Console.WriteLine("    -binaryformatter");
            if (TaroSocketConfig.USE_COMPRESSION)
                Console.WriteLine("    -compressed");
            else
                Console.WriteLine("    -uncompressed");
        }
    }
}
