//
// Create cluster node data structures
//

using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Xml.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

using MessageRouting;
using SkipNet=Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Common;

using Microsoft.Research.Taro;
using Microsoft.Research.Herald.ClusterLib;

namespace Microsoft.Research.Herald.ClusterLib
{
    public class ClusterNode
    {
        // Configuration
        public  bool                    socketMode=false;
        public  int                     port;

        public  string                  nodeNameSuffix;     // e.g. "/herald1/"     (User-supplied input)

        public  string                  nodeName;           // String name of the local node (used for SkipNet)
        public  string                  nodeAddr;           // Physical address where clients can access this service
        // e.g.,  "192.168.200.1:10001"     (IP mode)
        public  string                  virtualIp;           // modelnet style Physical address where clients can access this service
        // e.g. "10.0.0.1"

        public  string                  overlaySeed;
        public  bool localMode=false, ipMode=false;
        private bool                    useIcmpPing=false;

        private HeraldUI                ui;  // For I/O redirection

        // Overlay/Taro-related members
        public  SkipNet.Node            node;
        private NodePhysicalAddress     addr;
        private TaroTopology            topo;
        public  TaroSendMessage         sendMessage;
        public  TaroReceiveMessage      receiveMessage;
        public  TaroLocalEvent          localEvent;
        
        // Serialization members
        public  bool useXmlSerializer = true;
        private Type[]                  taroAndSkipNetTypes;

        // Counter callbacks
        private CounterCallback         sentMessageCallback;
        private CounterCallback         recvMessageCallback;

//============================================= Functions =============================================

        public ClusterNode(HeraldUI ui)
        {
            this.ui = ui;

            InitSerialization();
        }

        public string IpAddrToLogicalNodeName(String address) 
        {
            foreach (ClusterNodeTableEntry entry in ClusterNodeTable.clusterNodeTable)
            {
                if (address.Equals(entry.ipAddress))
                    return entry.logicalNodeName;
            }
            Console.WriteLine("No entry for IP address {0} in ClusterNodeTable", address);
            Debug.Assert(false);
            return null;
        }

        public void Usage()
        {
            ui.Error("  Node Configuration Parameters:");
            ui.Error("    [/local] [/socket] [/ip] [/ping] /port:portnum [/virtual:modelnet-IP-addr /modelnet] nodeNameSuffix [seed-node]");
            ui.Error("      e.g.: /socket /port:10000 /herald2/ 192.168.200.2:10000");
            ui.Error("      or:   /socket /port:10001 /virtual:10.0.0.1 /modelnet /herald1/");
            ui.Error("      or:   /port:10000 /herald2/ soap://mymachine2:10000/herald1/");
        }

        public void PrintArgs()
        {
            ui.Message("  Node Configuration Parameters:");
            ui.MessageNoNewLine("    ");
            if (localMode) ui.MessageNoNewLine("/local ");
            if (useIcmpPing) ui.MessageNoNewLine("/ping ");
            if (socketMode) ui.MessageNoNewLine("/socket ");
            if (ipMode) ui.MessageNoNewLine("/ip ");
            if (virtualIp != null) ui.MessageNoNewLine("/virtual:{0} ", virtualIp);
            if (TaroSocketConfig.USE_MODELNET) ui.MessageNoNewLine("/modelnet ");
            ui.MessageNoNewLine("/port:{0} ", port);
            ui.MessageNoNewLine("{0} ", nodeNameSuffix);
            if (overlaySeed != null) ui.Message("{0}", overlaySeed); else ui.Message("(no seed)");
        }

        // Parse cluster-specific command line arguments from the specified index,
        // returning the index of the first unparsed parameter.
        // Returns -1 if Usage() needs to be called.
        public int ProcessParameters(String[] args, int starting_at)
        {
            string machineName, machinePhysicalName;
            int i;

            // Ensure we have enough parameters
            if( args.Length - starting_at <= 0 ) return -1;

            // Process optional parameters
            i = starting_at;
            for(;;) 
            {
                if( args[i]=="/local" ) 
                {
                    localMode=true;
                } 
                else if( args[i]=="/ping" ) 
                {
                    Console.WriteLine("Using genuine ICMP pings");
                    useIcmpPing=true;
                } 
                else if( args[i]=="/socket" ) 
                {
                    socketMode=true;
                } 
                else if( args[i]=="/ip" ) 
                {
                    ipMode=true;
                } 
                else if( args[i].StartsWith("/port") ) 
                {
                    int separatorIndex = args[i].IndexOf(':');
                    if( separatorIndex==-1 ) return -1;
                    port = Int32.Parse( args[i].Substring(separatorIndex+1) );
                } 
                else if( args[i].StartsWith("/virtual") ) 
                {
                    int separatorIndex = args[i].IndexOf(':');
                    if( separatorIndex==-1 ) return -1;
                    virtualIp = args[i].Substring(separatorIndex+1);
                } 
                else if( args[i].StartsWith("/modelnet") ) 
                {
                    TaroSocketConfig.USE_MODELNET = true;
                } 
                else 
                {
                    break;
                }
                i++;
                // Ensure we have enough parameters left for the required parameters
                if( i>=args.Length ) return -1;
            }

            // User must give a port
            if( this.port==0 ) return -1;

            // Determine the 'hostname' part of our local URI
            if( localMode ) 
            {
                machineName = "localhost";
            } 
            else 
            {
                if (virtualIp!= null)
                {
                    machineName = IpAddrToLogicalNodeName(virtualIp);
                }
                else
                {
                    machineName = Environment.MachineName;
                }
            }
//              Console.WriteLine("machineName {0}", machineName);

            // Determine the machine's "physical" name: either its DNS name or IP address
            if( ipMode ) 
            {
                if (virtualIp!= null)
                {
                    machinePhysicalName = virtualIp;
	
                }
                else
                {
                    machinePhysicalName = Dns.Resolve(Environment.MachineName).AddressList[0].ToString();	
                }
            } 
            else 
            {
                machinePhysicalName = machineName;
            }

            // The first parameter is the relative part of our local URI
            nodeNameSuffix = args[i++];
            if( nodeNameSuffix[0]!='/' ) 
            {
                // Local service name must start with a '/'
                nodeNameSuffix = '/' + nodeNameSuffix;
            }
            if( nodeNameSuffix[ nodeNameSuffix.Length-1 ]!='/' ) 
            {
                // Local service name must end in a '/'
                nodeNameSuffix = nodeNameSuffix + '/';
            }

            nodeName = "skip://" + machineName + nodeNameSuffix;

            // If IP-mode is enabled, use the IP address in the node address
            if( socketMode ) 
            {
                if (virtualIp!= null)
                {
                    nodeAddr = virtualIp + ":" + port.ToString();
                }
                else
                {
                    nodeAddr = Dns.Resolve(Environment.MachineName).AddressList[0].ToString()
                        + ":" + port.ToString();	
                }
            } 
            else 
            {
                nodeAddr = "soap://" + machinePhysicalName + ":" + port + nodeNameSuffix;
            }

            // The second parameter is our Overlay seed URI
            if (args.Length-i>=1 && (args[i][0] != '/' && args[i][0] != '-') )
            {
                overlaySeed = args[i++];

                if( ! socketMode ) 
                {
                    // If not in IP mode, seed must be a URI
                    if( overlaySeed.IndexOf("soap://")!=0 ) return -1;

                    if( overlaySeed[ overlaySeed.Length-1 ]!='/' ) 
                    {
                        // Overlay seed URI must end in a '/'
                        overlaySeed = overlaySeed + '/';
                    }
                }
            }

            return i;
        }

        
        // Initialize the array of Types that are needed when serializing objects
        private void InitSerialization() 
        {
            taroAndSkipNetTypes = new Type[] {
                                        typeof(MessageRouting.RouteMsg),

                                        typeof(TaroAddress),
                                        typeof(TaroPing),

                                        // SkipNet objects
                                        typeof(SkipNet.NodeId),
                                        typeof(SkipNet.NumericId),
                                        typeof(SkipNet.NodeIdAddressBind),
                                        typeof(SkipNet.CLBId),

                                        typeof(SkipNet.RouteMsg),
                                        typeof(SkipNet.RouteDirectMsg),
                                        typeof(SkipNet.RouteByNameMsg),
                                        typeof(SkipNet.RouteByCLBMsg),
                                        typeof(SkipNet.RoutePayloadControlMsg),

                                        typeof(SkipNet.JoinRequest),
                                        typeof(SkipNet.JoinReply),
                                        typeof(SkipNet.LeafSetFixup),
                                        typeof(SkipNet.PiggybackWrapper),
                                        typeof(SkipNet.FixupHigher),
                                        typeof(SkipNet.LeafSetSimplePing),
                                        typeof(SkipNet.LeafSetSimplePingReply),

                                        typeof(SkipNet.LeafSet),
                                        typeof(SkipNet.LeafSetHalf),

                                        // PTable classes
                                        typeof(SkipNet.ProximityTable),
                                        typeof(SkipNet.ProximityTableHalf),
                                        typeof(SkipNet.FindEntriesRequest),
                                        typeof(SkipNet.FindEntriesReply),
                                        typeof(SkipNet.PTableEntry),
                                        typeof(SkipNet.UpdatePTableMsg),

                                        // CTable classes
                                        typeof(SkipNet.CLBJoinRequest),
                                        typeof(SkipNet.CLBJoinReply),
                                        typeof(SkipNet.UpdateCTableMsg)
            };
        }

        // Set the callback to be called every time a message is sent
        // This must be called before Initialize().
        public CounterCallback SentMessageCounter 
        {
            set 
            {
                sentMessageCallback = value;
            }
        }

        // Set the callback to be called every time a message is received
        // This must be called before Initialize().
        public CounterCallback RecvMessageCounter 
        {
            set 
            {
                recvMessageCallback = value;
            }
        }

        // Set up our local Overlay node and the Taro objects that it depends on.
        // When this is complete, we start a thread which will process the Overlay messages.
        public void InitNode(Type [] appTypes) 
        {
            TaroAddress seedNode=null;

            ui.Message("Initializing Taro and SkipNet");

            if( overlaySeed!=null ) 
            {
                ui.Message("Seed node: {0}", overlaySeed);
            } 
            else 
            {
                ui.Message("No seed");
            }

            // Create a local SkipNet node
            node = new SkipNet.Node( nodeName );

            // Create a NodePhysicalAddress
            addr = new TaroAddress( nodeAddr );
            
            // Attach the Overlay node to its Physical Address
            //addr.AttachLogicalNode( node );

            // Create LocalEvent object
            localEvent = new TaroLocalEvent( node );

            Type [] extraTypes;
            if (appTypes == null || appTypes.Length == 0)
                extraTypes = taroAndSkipNetTypes;
            else
            { // Concatenate Taro and SkipNet serialization types with application serialization types
                extraTypes = new Type [taroAndSkipNetTypes.Length + appTypes.Length];
                Array.Copy(taroAndSkipNetTypes, 0, extraTypes, 0, taroAndSkipNetTypes.Length);
                Array.Copy(appTypes, 0, extraTypes, taroAndSkipNetTypes.Length, appTypes.Length);
            }
            // Set up the XmlSerializer, along with the types of derived classes
            object serializer = useXmlSerializer ?
                (object) new XmlSerializer(typeof(object), extraTypes) :
                (object) new BinaryFormatter();

            // Create communication objects
            if( socketMode ) 
            {
                // Create Socket objects
                sendMessage = new TaroSocketSendMessage(localEvent, serializer, sentMessageCallback );
                topo = new TaroTopology( sendMessage, node, useIcmpPing );
            
                // Next find a seed node to connect to
                if( overlaySeed!=null ) 
                {
                    seedNode = new TaroAddress( overlaySeed );
                }

                // Next call InitNode on the Overlay node.
                // This must be called before TaroSocketReceiveMessage since it needs the PhysAddr to be assigned,
                // which happens in InitNode().  InitNode() initiates sends for the node join, but since they
                // are non-blocking we quickly get to the receive creation code below, in time to listen.
                node.InitNode(localEvent, sendMessage, topo, addr, seedNode);

                receiveMessage = new TaroSocketReceiveMessage( port, localEvent, node, serializer, recvMessageCallback );
            }
            else 
            {
                Debug.Assert(false, "Non-socket mode communications not supported in this code release");
            }

            node.ActivateNode(true);
        }
    }
  }
