//
// Provide for repeatable control flow of experiments on both the cluster and the simulator
//

using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using SkipNet = Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Common;

namespace Microsoft.Research.Herald.Common.ControlFlow
{

//========================= Server message base classes =========================//

    /// <summary>
    /// All ControlFlow request messages derive from ControlFlowRequestMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("ControlFlowRequestMsg",Namespace=ControlFlowApplication.CONTROLFLOW_NAMESPACE,IsNullable=true)]
    abstract public class ControlFlowRequestMsg: OverlayRequestMsg
    {
        public ControlFlowRequestMsg() {} // For XML Serialization only

        public ControlFlowRequestMsg(Node localNode, NodeIdAddressBind source, OverlayReplyCallback cb)
            : base(localNode, cb, 1 * 60 * 1000 /* TimeoutMs */, false)
        {
            this.MaxRetries = 0; // Don't retry program steps that fail to respond
            this.Source = source;
        }

        // Version allowing timeout to be specified
        public ControlFlowRequestMsg(Node localNode, NodeIdAddressBind source, OverlayReplyCallback cb, long timeoutMs)
            : base(localNode, cb, timeoutMs, false)
        {
            this.MaxRetries = 0; // Don't retry program steps that fail to respond
            this.Source = source;
        }
    }

    /// <summary>
    /// All ControlFlow reply messages derive from ControlFlowReplyMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("ControlFlowReplyMsg",Namespace=ControlFlowApplication.CONTROLFLOW_NAMESPACE,IsNullable=true)]
    public class ControlFlowReplyMsg: OverlayReplyMsg 
    {
        public ControlFlowReplyMsg() {} // For XML Serialization only

        public ControlFlowReplyMsg(MessageRouting.OverlayRequestMsg requestMsg)
            : base(requestMsg)
        {
        }
    }

    /// <summary>
    /// All ControlFlow asynchronous messages derive from ControlFlowAsyncMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("ControlFlowAsyncMsg",Namespace=ControlFlowApplication.CONTROLFLOW_NAMESPACE,IsNullable=true)]
    abstract public class ControlFlowAsyncMsg: OverlayMsg
    {
        public ControlFlowAsyncMsg() {} // For XML Serialization only
    }

//======================================== ControlFlowApplication class ========================================

    /// <summary>
    /// The main ControlFlow class.
    /// </summary>
    public class ControlFlowApplication: Application 
    {
        public const string CONTROLFLOW_NAMESPACE = "http://research.microsoft.com/sn/Herald/ControlFlow";
        public static Type [] serializationTypes = new Type[]
        {
            typeof(ControlFlowRequestMsg),
            typeof(ControlFlowReplyMsg),
            typeof(ControlFlowAsyncMsg),
            typeof(RendezvousWithNodeRequest),
            typeof(RendezvousWithNodeReply),
            typeof(OnNodeDoStepRequest),
            typeof(OnNodeDoStepReply),
            typeof(FailNodeMsg),
            typeof(ExitNodeMsg)
        };

        public SkipNet.Node myServer;
        public SendMessage sendMsg;
        public LocalEvent local;

        ControlFlowUpcalls upcalls;

        public static bool DebugOutput = false;

        public ControlFlowApplication(SkipNet.Node myServer, SendMessage s, LocalEvent l, ControlFlowUpcalls upcalls)
        {
            this.myServer = myServer;
            this.sendMsg = s;
            this.local = l;
            this.upcalls = upcalls;
        }

//========================= ControlFlowCallback =========================//

        public abstract class ControlFlowCallback : OverlayReplyCallback
        {
            public ControlFlowApplication app;

            public ControlFlowCallback(ControlFlowApplication app)
            {
                this.app = app;
            }
        }

//========================= RendezvousWithNode messages & functions  =========================//

        [Serializable, XmlRootAttribute("RendezvousWithNodeRequest", Namespace=CONTROLFLOW_NAMESPACE, IsNullable=true)]
            public class RendezvousWithNodeRequest : ControlFlowRequestMsg
        {
            public int nodeNumber;
            public int attempt;

            public RendezvousWithNodeRequest() {} // For XML Serialization only

            public RendezvousWithNodeRequest(Node localNode, NodeIdAddressBind source, OverlayReplyCallback cb, int nodeNumber, int attempt, long timeoutMs)
                : base(localNode, source, cb, timeoutMs)
            {
                this.nodeNumber = nodeNumber;
                this.attempt = attempt;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                ControlFlowApplication cfApp = (ControlFlowApplication) localNode.GetApp(typeof(ControlFlowApplication));
                cfApp.ServerRendezvousWithNode(this);
            }
        }

        [Serializable, XmlRootAttribute("RendezvousWithNodeReply", Namespace=CONTROLFLOW_NAMESPACE, IsNullable=true)]
            public class RendezvousWithNodeReply : ControlFlowReplyMsg
        {
            public int nodeNumber;
            public int attempt;

            public RendezvousWithNodeReply() {} // For XML Serialization only

            public RendezvousWithNodeReply(MessageRouting.OverlayRequestMsg requestMsg, int nodeNumber, int attempt)
                : base(requestMsg)
            {
                this.nodeNumber = nodeNumber;
                this.attempt = attempt;
            }
        }

        public void ClientRendezvousWithNode(NodeIdAddressBind dest, int nodeNumber, int attempt, long timeoutMs)
        {
            RendezvousWithNodeCallback cb = new RendezvousWithNodeCallback(this, nodeNumber, attempt);
            RendezvousWithNodeRequest request = new RendezvousWithNodeRequest(myServer, myServer.NodeIdAddressBind, cb, nodeNumber, attempt, timeoutMs);
            myServer.SendMessageDirect(request, dest);
        }

        void ServerRendezvousWithNode(RendezvousWithNodeRequest msg)
        {
            upcalls.RendezvousWithNodeRequestUpcall(myServer, msg.nodeNumber, msg.attempt);
            RendezvousWithNodeReply replyMsg = new RendezvousWithNodeReply(msg, msg.nodeNumber, msg.attempt);
            myServer.SendReply(replyMsg, msg.Source);
        }

        class RendezvousWithNodeCallback: ControlFlowCallback 
        {
            int nodeNumber;
            int attempt;

            public RendezvousWithNodeCallback(ControlFlowApplication cfApp, int nodeNumber, int attempt)
                : base (cfApp)
            {
                this.nodeNumber = nodeNumber;
                this.attempt = attempt;
            }

            public override void Execute(Node localNode) 
            {
                if (DebugOutput) 
                {
                    if (Reply != null && (
                        nodeNumber != ((RendezvousWithNodeReply) Reply).nodeNumber ||
                        attempt != ((RendezvousWithNodeReply) Reply).attempt))
                    {
                        Console.WriteLine("RendezvousWithNodeCallback match error at time {0:T}:", DateTime.Now);
                        Console.WriteLine("    nodeNumber {0} attempt {1} reply nodeNumber {2} reply attempt {3}",
                            nodeNumber, attempt,
                            ((RendezvousWithNodeReply) Reply).nodeNumber, ((RendezvousWithNodeReply) Reply).attempt);
                    }
                }

                app.upcalls.RendezvousWithNodeReplyUpcall(app, Reply == null, 
                    (Reply == null) ? nodeNumber : ((RendezvousWithNodeReply) Reply).nodeNumber,
                    (Reply == null) ? attempt : ((RendezvousWithNodeReply) Reply).attempt);
            }
        }


        //========================= OnNodeDoStep messages & functions  =========================//

        [Serializable, XmlRootAttribute("OnNodeDoStepRequest", Namespace=CONTROLFLOW_NAMESPACE, IsNullable=true)]
            public class OnNodeDoStepRequest : ControlFlowRequestMsg
        {
            public int nodeNumber;
            public int programStep;
            public int loopIndex;

            public OnNodeDoStepRequest() {} // For XML Serialization only

            public OnNodeDoStepRequest(Node localNode, NodeIdAddressBind source, OverlayReplyCallback cb,
                int nodeNumber, int programStep, int loopIndex)
                : base(localNode, source, cb)
            {
                this.nodeNumber = nodeNumber;
                this.programStep = programStep;
                this.loopIndex = loopIndex;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                ControlFlowApplication cfApp = (ControlFlowApplication) localNode.GetApp(typeof(ControlFlowApplication));
                cfApp.ServerOnNodeDoStep(this);
            }
        }

        [Serializable, XmlRootAttribute("OnNodeDoStepReply", Namespace=CONTROLFLOW_NAMESPACE, IsNullable=true)]
            public class OnNodeDoStepReply : ControlFlowReplyMsg
        {
            public OnNodeDoStepReply() {} // For XML Serialization only

            public OnNodeDoStepReply(MessageRouting.OverlayRequestMsg requestMsg)
                : base(requestMsg)
            {
            }
        }

        public void ClientOnNodeDoStep(NodeIdAddressBind dest, int nodeNumber, int programStep, int loopIndex)
        {
            if (DebugOutput)
                Console.WriteLine("OnNodeDoStep on node {0}: {1} {2} {3}", myServer.GetNodeIdAddressBind.GetAddress,
                    nodeNumber, programStep, loopIndex);

            OnNodeDoStepCallback cb = new OnNodeDoStepCallback(this);
            OnNodeDoStepRequest request = new OnNodeDoStepRequest(myServer, myServer.NodeIdAddressBind, cb,
                nodeNumber, programStep, loopIndex);
            myServer.SendMessageDirect(request, dest);
        }

        void ServerOnNodeDoStep(OnNodeDoStepRequest msg) 
        {
            int nodeNumber = msg.nodeNumber;
            int programStep = msg.programStep;
            int loopIndex = msg.loopIndex;
            if (DebugOutput)
                Console.WriteLine("Received OnNodeDoStep on node {0}: {1} {2} {3}", myServer.GetNodeIdAddressBind.GetAddress,
                    nodeNumber, programStep, loopIndex);
            upcalls.OnNodeDoStepRequestUpcall(myServer, msg, nodeNumber, programStep, loopIndex);
        }

        public void SendOnNodeDoStepReply(object replyToken)
        {
            OnNodeDoStepRequest requestMsg = (OnNodeDoStepRequest) replyToken;
            OnNodeDoStepReply replyMsg = new OnNodeDoStepReply(requestMsg);
            if (DebugOutput)
                Console.WriteLine("Sending OnNodeDoStep reply on node {0}", myServer.GetNodeIdAddressBind.GetAddress);
            myServer.SendReply(replyMsg, requestMsg.Source);
        }

        class OnNodeDoStepCallback: ControlFlowCallback 
        {
            public OnNodeDoStepCallback(ControlFlowApplication cfApp)
                : base (cfApp)
            {
            }

            public override void Execute(Node localNode) 
            {
                if (DebugOutput) Console.WriteLine("OnNodeDoStepCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);
                if (Reply == null) 
                    if (DebugOutput)
                        Console.WriteLine("OnNodeDoStepCallback: request timed out");

                app.upcalls.OnNodeDoStepReplyUpcall(app, Reply == null);
            }
        }

//========================= FailNode message & functions  =========================//

        [Serializable, XmlRootAttribute("FailNodeMsg", Namespace=CONTROLFLOW_NAMESPACE, IsNullable=true)]
            public class FailNodeMsg : ControlFlowAsyncMsg
        {
            public FailNodeMsg() {} // For XML Serialization only

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                ControlFlowApplication cfApp = (ControlFlowApplication) localNode.GetApp(typeof(ControlFlowApplication));
                cfApp.ServerFailNode(this);
            }
        }

        public void ClientFailNode(NodeIdAddressBind dest)
        {
            // If debugging high latency replication skip this message sometimes so cache is actually used
            FailNodeMsg msg = new FailNodeMsg();
            myServer.SendMessageDirect(msg, dest);
        }

        void ServerFailNode(FailNodeMsg msg) 
        {
            if (DebugOutput)
                Console.WriteLine("Received FailNode on node {0}", myServer.GetNodeIdAddressBind.GetAddress);
            upcalls.FailNodeUpcall(myServer);
        }

//========================= ExitNode message & functions  =========================//

        [Serializable, XmlRootAttribute("ExitNodeMsg", Namespace=CONTROLFLOW_NAMESPACE, IsNullable=true)]
            public class ExitNodeMsg : ControlFlowAsyncMsg
        {
            public ExitNodeMsg() {} // For XML Serialization only

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                ControlFlowApplication cfApp = (ControlFlowApplication) localNode.GetApp(typeof(ControlFlowApplication));
                cfApp.ServerExitNode(this);
            }
        }

        public void ClientExitNode(NodeIdAddressBind dest)
        {
            // If debugging high latency replication skip this message sometimes so cache is actually used
            ExitNodeMsg msg = new ExitNodeMsg();
            myServer.SendMessageDirect(msg, dest);
        }

        void ServerExitNode(ExitNodeMsg msg) 
        {
            if (DebugOutput)
                Console.WriteLine("Received ExitNode on node {0}", myServer.GetNodeIdAddressBind.GetAddress);
            upcalls.ExitNodeUpcall(myServer);
        }

//================================ Required Application class housekeeping methods ================================

        public override void NewNeighbour(MessageRouting.NodeIdAddressBind id) 
        {
            // Nothing to do for the ControlFlow application
        }

        public override void NeighbourRemoved(MessageRouting.NodeIdAddressBind id, bool failure) 
        {
            // Nothing to do for the ControlFlow application
        }
    }

//================================ Functions to be provided by clients of this application ================================

    /// <summary>
    /// Interface to be supplied by code using the ControlFlowApplication allowing them to receive upcalls on message reciept.
    /// </summary>
    public interface ControlFlowUpcalls
    {
        void RendezvousWithNodeRequestUpcall(SkipNet.Node node, int nodeNumber, int attempt);
        void RendezvousWithNodeReplyUpcall(ControlFlowApplication cfApp, bool timedOut, int nodeNumber, int attempt);
        void OnNodeDoStepRequestUpcall(SkipNet.Node node, object replyToken, int nodeNumber, int programStep, int loopIndex);
        void OnNodeDoStepReplyUpcall(ControlFlowApplication cfApp, bool timedOut);
        void FailNodeUpcall(SkipNet.Node node);
        void ExitNodeUpcall(SkipNet.Node node);
    }



    //================================ Simple control flow classes ====================

    public class SimpleControlFlow {

        public static ExecutionEnvironment exEnv = null;
        public static bool quitImmediately = true;

        // Variables for sending control RPCs.
        public SkipNet.Node myServer = null;
        public bool AppRPCReplyPending = false;
        public bool AppRPCTimeout = false;
        public OverlayReplyMsg AppRPCReplyMsg = null;

        public SimpleControlFlow(SkipNet.Node myServer, ExecutionEnvironment exEnvParam, bool quitImmediatelyParam) {
            this.myServer = myServer;
            exEnv = exEnvParam;
            quitImmediately = quitImmediatelyParam;
        }

       
        /// <summary>
        /// Tell a specified node to fail itself.
        /// </summary>
        /// <param name="nodeNumber"></param>
        public void FailNode(int nodeNumber) {
            ExecuteBasicNodeOpn(nodeNumber, BasicNodeOpn.FailNode);
        }

        /// <summary>
        /// Execute a BasicNodeOpn on a specified node.
        /// 
        /// *** NOTE *** Only one of these may be executing at a time.
        /// </summary>
        /// <param name="nodeNumber"></param>
        /// <param name="opn"></param>
        public bool ExecuteBasicNodeOpn(int nodeNumber, BasicNodeOpn opn, long timeoutMs) {
            BasicNodeOpnMsg msg = new BasicNodeOpnMsg(myServer, this, opn, timeoutMs);
            BasicNodeOpnReplyMsg replyMsg = (BasicNodeOpnReplyMsg) SendRequestMsg(exEnv.NodeIdAddrBind(nodeNumber), msg);
            if (replyMsg == null) {
                // Rendezvous timed out.  Abort.
                if (ControlFlowApplication.DebugOutput) Console.WriteLine("BasicNodeOpnMsg to {0} timed out.", exEnv.NodeIdAddrBind(nodeNumber).GetNodeId.ToString());
                return false;
            }
            else {
                return true;
            }
        }
        
        /// <summary>
        /// Execute a BasicNodeOpn on a specified node.
        /// 
        /// *** NOTE *** Only one of these may be executing at a time.
        /// </summary>
        /// <param name="nodeNumber"></param>
        /// <param name="opn"></param>
        public bool ExecuteBasicNodeOpn(int nodeNumber, BasicNodeOpn opn) 
        {
            return ExecuteBasicNodeOpn(nodeNumber, opn, Parameters.BaseNodeOpnTimeout);
        }

        // JD 8-28-03 overloaded this in order to scoop up the SkipNet naming information
        public bool ExecuteBasicNodeOpn(int nodeNumber, BasicNodeOpn opn, Hashtable nodeNumber_to_NodeIDAddressBind, long timeoutMs) 
        {
            BasicNodeOpnMsg msg = new BasicNodeOpnMsg(myServer, this, opn, timeoutMs);
            BasicNodeOpnReplyMsg replyMsg = (BasicNodeOpnReplyMsg) SendRequestMsg(exEnv.NodeIdAddrBind(nodeNumber), msg);
            if (replyMsg == null) 
            {
                // Rendezvous timed out.  Abort.
                if (ControlFlowApplication.DebugOutput) Console.WriteLine("BasicNodeOpnMsg to {0} timed out.", exEnv.NodeIdAddrBind(nodeNumber).GetNodeId.ToString());
                return false;
            }
            else 
            {
                nodeNumber_to_NodeIDAddressBind[nodeNumber] = replyMsg.originatorOfReply;
                return true;
            }
        }

        // JD 8-28-03 overloaded this in order to scoop up the SkipNet naming information
        public bool ExecuteBasicNodeOpn(int nodeNumber, BasicNodeOpn opn, Hashtable nodeNumber_to_NodeIDAddressBind) 
        {
            return ExecuteBasicNodeOpn(nodeNumber, opn, nodeNumber_to_NodeIDAddressBind, Parameters.BaseNodeOpnTimeout);
        }

        /// <summary>
        /// RPC routine: send a 'msg' that is a subtype of 'RPCRequestMsg' and wait for a reply message.
        /// 
        /// *** NOTE *** Only one of these may be executing at a time.
        /// </summary>
        /// <param name="dest"></param>
        /// <param name="msg"></param>
        /// <returns></returns>
        public OverlayReplyMsg SendRequestMsg(NodeIdAddressBind dest, RPCRequestMsg msg) {
            Debug.Assert(AppRPCReplyPending == false);
            AppRPCReplyPending = true;
            AppRPCTimeout = false;
            lock (exEnv.OverlayLock) {
                myServer.SendMessageDirect(msg, dest);
            }
            while (AppRPCReplyPending) {
                //if (exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment) {
                //    System.Threading.Thread.Sleep(5); // XXX Replace with something better
                //}
                //else {
                    exEnv.GoFor(5);
                //}
            }
            if (AppRPCTimeout) return null;
            else return AppRPCReplyMsg;
        }
    }

    /// <summary>
    /// Operations supported by ExecuteBasicNodeOpn() on a node.
    /// </summary>
    public enum BasicNodeOpn {Ping, ExitProgram, FailNode};

    /// <summary>
    /// Needed for XML serialization of messages.
    /// </summary>
    public class SimpleControlFlow_SerializationConfig {
        public const string RPCCONTROL_NAMESPACE = "http://research.microsoft.com/sn/Herald/RPCCONTROL";
        public static Type [] serializationTypes = new Type[] {
                                                                  typeof(RPCRequestMsg),
                                                                  typeof(BasicNodeOpnMsg),
                                                                  typeof(BasicNodeOpnReplyMsg)
                                                              };
    }

    /// <summary>
    /// RPC control parameters.
    /// </summary>
    public class Parameters {
        public const int RPCTimeout = 60 * 1000;  // 1 minute
        public const int BaseNodeOpnTimeout = 1 * 60 * 1000; // 1 minute
    }

    /// <summary>
    /// Internal callback class used with RPCRequestMsg messages to do
    /// an RPC-style rendezvous at the caller.
    /// </summary>
    public class SendReplyCallback: OverlayReplyCallback {
        SimpleControlFlow cf;

        public SendReplyCallback(SimpleControlFlow cf) {
            this.cf = cf;
        }

        public override void Execute( MessageRouting.Node localNode ) {
            Debug.Assert(cf.AppRPCReplyPending == true);
            cf.AppRPCReplyMsg = this.Reply;
            cf.AppRPCReplyPending = false;
            if (this.Reply == null) cf.AppRPCTimeout = true;
            else cf.AppRPCTimeout = false;
        }
    }

    /// <summary>
    /// Abstract message class that is the supertype for all messages
    /// that will be sent using SendRequestMsg().
    /// </summary>
    [Serializable,
    XmlRootAttribute("RPCRequestMsg",Namespace=SimpleControlFlow_SerializationConfig.RPCCONTROL_NAMESPACE,IsNullable=true)]
    public abstract class RPCRequestMsg: OverlayRequestMsg {
         
        public NodeIdAddressBind Requestor;

        public RPCRequestMsg() {}  // For XML serialization

        public RPCRequestMsg(MessageRouting.Node localNode, SimpleControlFlow cf):
            this(localNode, cf, Parameters.RPCTimeout) {
        }

        public RPCRequestMsg(MessageRouting.Node localNode, SimpleControlFlow cf, long timeout):
            base(localNode, new SendReplyCallback(cf), timeout, false) {
            this.Requestor = localNode.GetNodeIdAddressBind;
            this.MaxRetries = 0;  // Give up after the first transmission attempt (we're on TCP).
        }
    }

    /// <summary>
    /// Message used to rendevous with a specified node.
    /// </summary>
    [Serializable,
    XmlRootAttribute("BasicNodeOpnMsg",Namespace=SimpleControlFlow_SerializationConfig.RPCCONTROL_NAMESPACE,IsNullable=true)]
    public class BasicNodeOpnMsg : RPCRequestMsg {
        
        public BasicNodeOpn opn;

        public BasicNodeOpnMsg() {}

        public BasicNodeOpnMsg( MessageRouting.Node localNode, SimpleControlFlow cf, BasicNodeOpn opn): 
            base(localNode, cf, Parameters.BaseNodeOpnTimeout) {
            this.opn = opn;
        }
 
        // Version allowing timeout to be specified
        public BasicNodeOpnMsg( MessageRouting.Node localNode, SimpleControlFlow cf, BasicNodeOpn opn, long timeoutMs): 
            base(localNode, cf, timeoutMs) 
        {
            this.opn = opn;
        }

        public override void ArrivedAtDestination( MessageRouting.Node overlayNode ) {

            if (ControlFlowApplication.DebugOutput) Console.WriteLine("BasicNodeOpnMsg {0} (opn: {1})", overlayNode.GetNodeIdAddressBind.GetNodeId.ToString(), this.opn);

            SkipNet.Node skipNode = (SkipNet.Node) overlayNode;
            SkipNet.NodeIdAddressBind here = (SkipNet.NodeIdAddressBind) skipNode.GetNodeIdAddressBind;
            OverlayReplyMsg replyMsg = new BasicNodeOpnReplyMsg(this, here);
            overlayNode.SendReply(replyMsg, this.Requestor);

            bool quitFlag = false;
            if (this.opn == BasicNodeOpn.ExitProgram) {
                if (SimpleControlFlow.exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment) 
                    quitFlag = true;
            }
            else if (this.opn == BasicNodeOpn.FailNode) {
                if (SimpleControlFlow.exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment) 
                    quitFlag = true;
                else 
                    overlayNode.Failed = true;
            }
            else if (this.opn == BasicNodeOpn.Ping) 
            {
                if (SimpleControlFlow.exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment) 
                    Console.WriteLine("Ping received at time {0:T}", DateTime.Now);
            }
            if (quitFlag) {
                if (!SimpleControlFlow.quitImmediately) {
                    Console.WriteLine();
                    Console.WriteLine("Hit Enter to exit program.");
                    Console.ReadLine();
                }
                if (SimpleControlFlow.exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment) 
                {
                    System.Threading.Thread.Sleep(10 * 1000); // Give reply time to go out
                    Console.WriteLine("Exiting at {0:T}", DateTime.Now);
                }
                Environment.Exit(0);
            }
        }
    }

    /// <summary>
    /// Reply to a BasicNodeOpnMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("BasicNodeOpnReplyMsg",Namespace=SimpleControlFlow_SerializationConfig.RPCCONTROL_NAMESPACE,IsNullable=true)]
    public class BasicNodeOpnReplyMsg : OverlayReplyMsg {

        public SkipNet.NodeIdAddressBind originatorOfReply;

        public BasicNodeOpnReplyMsg() {}

        public BasicNodeOpnReplyMsg( OverlayRequestMsg requestMsg, SkipNet.NodeIdAddressBind originatorOfReply): base(requestMsg) {
            this.originatorOfReply = originatorOfReply;
        }
    }
}
