using System;
using System.Collections;

namespace Microsoft.Research.Herald.Common
{
    public class DoubleHistogram 
    {
	private int numBuckets;
	private double[] bucketBounds;
	private int[] bucketValues;
	private double overflowSum;

	// Create a histogram object with the specified buckets.
	// Each entry in 'bucketBounds' specifies an upper bound for a
	// bucket.  The histogram object will maintain one additional bucket
	// for values larger than the largest specified bucket value.
	public DoubleHistogram(double[] bucketBounds) {
	    this.numBuckets = bucketBounds.Length;
	    this.bucketBounds = new double[this.numBuckets];
	    this.bucketValues = new int[this.numBuckets + 1];
	    for (int i = 0; i < this.numBuckets; i++) {
		this.bucketBounds[i] = bucketBounds[i];
	    }
            this.overflowSum = 0.0;
	}

	// Add a new value to the histogram.
	public void Update(double val) {
	    bool foundIt = false;
	    for (int i = 0; i < this.numBuckets; i++) {
		if (val <= this.bucketBounds[i]) {
		    this.bucketValues[i] += 1;
		    foundIt = true;
		    break;
		}
	    }
	    if (!foundIt) {
		this.bucketValues[this.numBuckets] += 1;
		this.overflowSum += val;
	    }
	}

	// Reset the histogram to empty.
	public void Clear()
	{
	    for (int i = 0; i <= this.numBuckets; i++)
	    {
		this.bucketValues[i] = 0;
	    }
            this.overflowSum = 0.0;
	}

	// Print the histogram, prefixed by the contents of 'label'.
	public void Print(string label) {
	    // First, figure out the total number of bucket values, so that we can
	    // compute percentages.
	    int total = 0;
	    for (int i = 0; i <= this.numBuckets; i++) {
		total += this.bucketValues[i];
	    }
	    if (total == 0) total = 1; // Avoid arithmetic faults.

	    // Print out the histogram.
	    Console.WriteLine(label);
	    double percentage = (double)(this.bucketValues[0] * 100) / (double)total;
	    Console.WriteLine("     -  , {0,5}]: {1}  ({2:f}%)",
			      this.bucketBounds[0],
			      this.bucketValues[0], percentage);
	    for (int i = 1; i < this.numBuckets; i++) {
		percentage = (double)(this.bucketValues[i] * 100) / (double)total;
		Console.WriteLine("  ({0,5}, {1,5}]: {2}  ({3:f}%)",
				  this.bucketBounds[i-1], this.bucketBounds[i],
				  this.bucketValues[i], percentage);
	    }
	    percentage = (double)(this.bucketValues[this.numBuckets] * 100) / (double)total;
	    Console.WriteLine("  ({0,5},   -   : {1}  ({2:f}%, overflow sum={3})",
		this.bucketBounds[this.numBuckets - 1],
		this.bucketValues[this.numBuckets], percentage, this.overflowSum);
	}

        public void PrintCDF(string label) 
        {
	    // First, figure out the total number of bucket values, so that we can
	    // compute percentages.
	    int total = 0;
	    for (int i = 0; i <= this.numBuckets; i++) {
		total += this.bucketValues[i];
	    }
	    if (total == 0) total = 1; // Avoid arithmetic faults.

	    // Print out the histogram.
	    Console.WriteLine(label);

            int currentSum = this.bucketValues[0];
	    double percentage = (double)(currentSum * 100) / (double)total;
	    Console.WriteLine("     -  , {0,5}]: {1:f}%",
			      this.bucketBounds[0], percentage);
            bool done = false;
	    for (int i = 1; i < this.numBuckets; i++) {
                // special support to avoid printing lots of 0% values at the beginning
                // and lots of 100% values at the end
                if (currentSum == 0) {
                    if (this.bucketValues[i] == 0)
                        continue;
                    // if i == 1, then the previous entry was already printed
                    if (this.bucketValues[i] > 0 && (i >=2))
                        Console.WriteLine("  ({0,5}, {1,5}]: {2:f}%",
                                          this.bucketBounds[i-2], this.bucketBounds[i-1], 0);
                }
                if (currentSum == total) {
                    // print out the first 100% entry, but none after that
                    if (done == true)
                        continue;
                    done = true;
                }
                currentSum += this.bucketValues[i];
		percentage = (double)(currentSum * 100) / (double)total;
		Console.WriteLine("  ({0,5}, {1,5}]: {2:f}%",
				  this.bucketBounds[i-1], this.bucketBounds[i],
				  percentage);
	    }
            currentSum += this.bucketValues[this.numBuckets];
	    percentage = (double)(currentSum * 100) / (double)total;
	    Console.WriteLine("  ({0,5},   -   : {1:f}%  (overflow sum={2})",
                              this.bucketBounds[this.numBuckets - 1],
                              percentage, this.overflowSum);
        }
        
	// Histogram bucket bounds.
	public double[] Bounds {
	    get {return this.bucketBounds;}
	}

	// Histogram bucket values.
	public int[] Values {
	    get {return this.bucketValues;}
	}
    
	// Histogram bucket values.
	public double OverflowSum 
	{
	    get {return this.overflowSum;}
	}
    }
    
    public class Histogram {

	private int numBuckets;
	private int[] bucketBounds;
	private int[] bucketValues;
	private long overflowSum;

	// Create a histogram object with the specified buckets.
	// Each entry in 'bucketBounds' specifies an upper bound for a
	// bucket.  The histogram object will maintain one additional bucket
	// for values larger than the largest specified bucket value.
	public Histogram(int[] bucketBounds) {
	    this.numBuckets = bucketBounds.Length;
	    this.bucketBounds = new int[this.numBuckets];
	    this.bucketValues = new int[this.numBuckets + 1];
	    for (int i = 0; i < this.numBuckets; i++) {
		this.bucketBounds[i] = bucketBounds[i];
	    }
            this.overflowSum = 0;
	}

	// Add a new value to the histogram.
	public void Update(int val) {
	    bool foundIt = false;
	    for (int i = 0; i < this.numBuckets; i++) {
		if (val <= this.bucketBounds[i]) {
		    this.bucketValues[i] += 1;
		    foundIt = true;
		    break;
		}
	    }
	    if (!foundIt) {
		this.bucketValues[this.numBuckets] += 1;
		this.overflowSum += val;
	    }
	}

	// Reset the histogram to empty.
	public void Clear()
	{
	    for (int i = 0; i <= this.numBuckets; i++)
	    {
		this.bucketValues[i] = 0;
	    }
            this.overflowSum = 0;
	}

	// Print the histogram, prefixed by the contents of 'label'.
	public void Print(string label) {
	    // First, figure out the total number of bucket values, so that we can
	    // compute percentages.
	    int total = 0;
	    for (int i = 0; i <= this.numBuckets; i++) {
		total += this.bucketValues[i];
	    }
	    if (total == 0) total = 1; // Avoid arithmetic faults.

	    // Print out the histogram.
	    Console.WriteLine(label);
	    double percentage = (double)(this.bucketValues[0] * 100) / (double)total;
	    Console.WriteLine("     -  , {0,5}]: {1}  ({2:f}%)",
			      this.bucketBounds[0],
			      this.bucketValues[0], percentage);
	    for (int i = 1; i < this.numBuckets; i++) {
		percentage = (double)(this.bucketValues[i] * 100) / (double)total;
		Console.WriteLine("  ({0,5}, {1,5}]: {2}  ({3:f}%)",
				  this.bucketBounds[i-1], this.bucketBounds[i],
				  this.bucketValues[i], percentage);
	    }
	    percentage = (double)(this.bucketValues[this.numBuckets] * 100) / (double)total;
	    Console.WriteLine("  ({0,5},   -   : {1}  ({2:f}%,  overflow sum={3})",
		this.bucketBounds[this.numBuckets - 1],
		this.bucketValues[this.numBuckets], percentage, this.overflowSum);
	}

        public void PrintCDF(string label) 
        {
	    // First, figure out the total number of bucket values, so that we can
	    // compute percentages.
	    int total = 0;
	    for (int i = 0; i <= this.numBuckets; i++) {
		total += this.bucketValues[i];
	    }
	    if (total == 0) total = 1; // Avoid arithmetic faults.

	    // Print out the histogram.
	    Console.WriteLine(label);

            int currentSum = this.bucketValues[0];
	    double percentage = (double)(currentSum * 100) / (double)total;
	    Console.WriteLine("     -  , {0,5}]: {1:f}%",
			      this.bucketBounds[0], percentage);
            bool done = false;
	    for (int i = 1; i < this.numBuckets; i++) {
                // special support to avoid printing lots of 0% values at the beginning
                // and lots of 100% values at the end
                if (currentSum == 0) {
                    if (this.bucketValues[i] == 0)
                        continue;
                    // if i == 1, then the previous entry was already printed
                    if (this.bucketValues[i] > 0 && (i >=2))
                        Console.WriteLine("  ({0,5}, {1,5}]: {2:f}%",
                                          this.bucketBounds[i-2], this.bucketBounds[i-1], 0);
                }
                if (currentSum == total) {
                    // print out the first 100% entry, but none after that
                    if (done == true)
                        continue;
                    done = true;
                }
                currentSum += this.bucketValues[i];
		percentage = (double)(currentSum * 100) / (double)total;
		Console.WriteLine("  ({0,5}, {1,5}]: {2:f}%",
				  this.bucketBounds[i-1], this.bucketBounds[i],
				  percentage);
	    }
            currentSum += this.bucketValues[this.numBuckets];
	    percentage = (double)(currentSum * 100) / (double)total;
	    Console.WriteLine("  ({0,5},   -   : {1:f}%  (overflow sum={2})",
                              this.bucketBounds[this.numBuckets - 1],
                              percentage, this.overflowSum);
        }
        
	// Histogram bucket bounds.
	public int[] Bounds {
	    get {return this.bucketBounds;}
	}

	// Histogram bucket values.
	public int[] Values {
	    get {return this.bucketValues;}
	}
    
	// Histogram bucket values.
	public long OverflowSum 
	{
	    get {return this.overflowSum;}
	}
    }
}
