using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;
using System.Security.Cryptography;

using MessageRouting;
using SkipNet = Microsoft.Research.Herald.SkipNet;
using NodeIdAddressBind = MessageRouting.NodeIdAddressBind;
using Microsoft.Research.Herald.Logger;

namespace  Microsoft.Research.Herald.DLC
{
    /// <summary>
    /// DLCid is a globally unique identifier.
    /// </summary>
    public class DLCid
    {
        public DLCid(long timePortion, int randomPortion)
        {
            this.timePortion = timePortion;
            this.randomPortion = randomPortion;
        }
        public override string ToString ()
        {
            return timePortion + ":" + randomPortion;
        }

        protected int randomPortion;
        protected long timePortion;
    }

    /// <summary>
    /// DLC methods can return instances of this exception.
    /// </summary>
    public class DLCException : Exception
    {
        public DLCException() {}
        public DLCException(string errMsg) 
        {
            this.errMsg = errMsg;
        }
        public string errMsg;
    }
    
    /// <summary>
    /// This is used by clients of the DLC package to specify their continuations
    /// for each public method.
    /// </summary>
    public abstract class DLCCContinuation : Continuation 
    {
        public string ErrMsg;  // null => success
        public string DLCCId;  // Id of the delegated liveness checking chain
    }

    public class NullCallback : OverlayReplyCallback 
    {
        public NullCallback() : base() {}
        public override void Execute(Node localNode) {}
    }

    /// <summary>
    /// This defines a function that can be registered using RegisterFailureNotifier() to receive
    /// notifications when there are failures or exceptions on a DLCC.
    /// Note that DLCCs automatically deallocate themselves when they signal a failure.
    /// </summary>
    public delegate void DLCCFailureNotification(Node localNode, string dlccId, string errMsg);

    /// <summary>
    /// If this timer fires, we're at the root, create/recreate has failed, and we should act accordingly. 
    /// </summary>
    public class CreateRecreateTimer : Timer 
    {
        private string dlcId;
        private bool immediateRecreate;
        public CreateRecreateTimer(long delayMs, string dlcId) : base(delayMs) 
        {
            this.dlcId = dlcId;
            immediateRecreate = false;
        }
        public bool isImmediateRecreate() { return immediateRecreate; }
        public void doImmediateRecreate() { immediateRecreate = true; }
        public override void Execute( MessageRouting.Node node )		 
        {
            DLCC dlcApp = (DLCC) node.GetApp(typeof(DLCC));
            Debug.Assert(dlcApp.dlcEntryTable[dlcId] != null);
            DlcEntry entry = dlcApp.dlcEntryTable[dlcId];
            Debug.Assert(entry.rootState != null);
            Debug.Assert(entry.rootState.rootStateCondition != DlcRootState.RootStateCondition.Live);
            entry.rootState.createRecreateTimer = null; // remove this reference to self so that
            // this timer is not redundantly cancelled (it's firing right now)
            dlcApp.HardNotificationUpcall(dlcId, null); // This will do all the appropriate cleanup.
        }
    }

    /// <summary>
    /// If this timer fires, we're at a member, recreate has failed, and we should act accordingly.
    /// </summary>
    public class RecreateTimer : Timer 
    {
        private string dlcId;
        public RecreateTimer(long delayMs, string dlcId) : base(delayMs) 
        {
            this.dlcId = dlcId;
        }
        public override void Execute( MessageRouting.Node node )		 
        {
            DLCC dlcApp = (DLCC) node.GetApp(typeof(DLCC));
            Debug.Assert(dlcApp.dlcEntryTable[dlcId] != null);
            DlcEntry entry = dlcApp.dlcEntryTable[dlcId];
            Debug.Assert(entry.memberState != null);
            Debug.Assert(entry.memberState.memberStateCondition != DlcMemberState.MemberStateCondition.Live);
            entry.memberState.recreateTimer = null; // remove this reference to self so that
            // this timer is not redundantly cancelled (it's firing right now)
            dlcApp.HardNotificationUpcall(dlcId, null); // This will do all the appropriate cleanup.
        }
    }

    /// <summary>
    /// If this timer fires, we're at the root, the InstallChecking messages never finished arriving, and we
    /// need to recreate.
    /// </summary>
    public class InstallCheckingTimer : Timer 
    {
        private DlcIdPlusSeqnum dlcIdPlusSeqnum;
        public InstallCheckingTimer( long delayMs, DlcIdPlusSeqnum dlcIdPlusSeqnum) : base(delayMs) 
        {
            this.dlcIdPlusSeqnum = dlcIdPlusSeqnum;
        }
        public override void Execute( MessageRouting.Node node )		 
        {
            string dlcId = dlcIdPlusSeqnum.dlcId;
            DLCC dlcApp = (DLCC) node.GetApp(typeof(DLCC));
            Debug.Assert(dlcApp.dlcEntryTable[dlcId] != null);
            DlcEntry entry = dlcApp.dlcEntryTable[dlcId];
            Debug.Assert(entry.rootState != null);
            entry.rootState.installCheckingTimer = null; // remove this reference to self so that
            // this timer is not redundantly cancelled (it's firing right now)
            Log.Write("dbg", LogLvl.Hi, "SoftNotification due to InstallChecking timer firing on dlcId {0}", dlcId);
            dlcApp.SoftNotificationUpcall(dlcIdPlusSeqnum, null); // This will do all the appropriate cleanup.
        }    
    }
    /// <summary>
    /// If this timer fires, we have not received liveness confirming messages sufficiently quickly, and we
    /// need to do a SoftNotificationUpcall.
    /// </summary>
    public class DlcPathRefreshTimer : Timer 
    {
        private DlcIdPlusSeqnum dlcIdPlusSeqnum;
        private DlcPath path;

        public DlcPathRefreshTimer( long delayMs, DlcIdPlusSeqnum dlcIdPlusSeqnum, 
            MessageRouting.Node localNode, DlcPath path) : base(delayMs) 
        {
            this.dlcIdPlusSeqnum = dlcIdPlusSeqnum;
            this.path = path;
            Debug.Assert(path.dlcPathRefreshTimer == null);
            path.dlcPathRefreshTimer = this;
            localNode.SetTimer(this);
            Log.Write("dbg", LogLvl.Med, "DlcPathRefreshTimer created and set for dlc {0}", dlcIdPlusSeqnum);
        }
        public void Reset(MessageRouting.Node localNode) 
        {
            path.dlcPathRefreshTimer = null;
            DlcPathRefreshTimer newTimer = new DlcPathRefreshTimer(80*1000, dlcIdPlusSeqnum,
                localNode, path);
            localNode.CancelTimer(this);
            Log.Write("dbg", LogLvl.Med, "DlcPathRefreshTimer reset for dlc {0}", dlcIdPlusSeqnum);
        }
        public void Remove(MessageRouting.Node localNode) 
        {
            localNode.CancelTimer(this);
            path.dlcPathRefreshTimer = null;
            Log.Write("dbg", LogLvl.Med, "DlcPathRefreshTimer removed for dlc {0}", dlcIdPlusSeqnum);
        }
        public override void Execute( MessageRouting.Node node )		 
        {
            path.dlcPathRefreshTimer = null;
            DLCC dlcApp = (DLCC) node.GetApp(typeof(DLCC));
            DlcIdPlusSeqnum dlcIdPlusSeqnumForNotification = new DlcIdPlusSeqnum( dlcIdPlusSeqnum.dlcId,
                path.containingEntry.seqnum);
            Log.Write("dbg", LogLvl.Med, "SoftNotification due to DlcPathRefresh timer firing on dlcId {0}", dlcIdPlusSeqnum.dlcId);
            dlcApp.SoftNotificationUpcall(dlcIdPlusSeqnumForNotification, null); // This will do all the appropriate cleanup.
        }    
    }

    //========================= DLCC class =========================//

    /// <summary>
    /// The DLC Application that must run on top of the overlay.
    /// </summary>
    public class DLCC : Application 
    {
        // ================ Member Data ================ //
        
        // We avoid spurious teardowns during setup by ignoring soft
        // notifications until after the group has been in place for at least this much time.
        public const int SETUP_BUFFER_MSECS = 60 * 1000;

        public const int DLC_INFINITY = 10000; //HACK
        public const int DLC_AUTO_REPAIR_TIMEOUT = 5 * 60 * 1000;

        public static byte [] NULL_DLC_PING = {0};

        protected Node myServer;
        public LocalEvent localEvent;
        public DLCEntryTable dlcEntryTable;
     
        public Hashtable dlccids_to_nodeIDAddressBindsOfGroupMembersPlusDelegates;

        protected int dlcGroupCounter = 0; // used in generating globally unique fuse group IDs
        protected int dlcSeqnumCounter = 1; // used in generating monotonically increasing seqnums
        public static bool DebugOutput = false;

        protected DLCCFailureNotification failureHandler;
        protected Hashtable  failureHandlerTable;

        public DLCC(LocalEvent local) 
        { 
            failureHandlerTable = new Hashtable();
            dlcEntryTable = new DLCEntryTable(); 
            dlccids_to_nodeIDAddressBindsOfGroupMembersPlusDelegates = new Hashtable();
            // This is the current API we export. 
            // I'll change the API to roll init in to the constructor
            // (and to pass in fewer parameters) once we can pass the bvt's again.
            dlcidToRecreateGroupBackoffInfo = new Hashtable();
        }
        public void DLCInit(Node myServer, SendMessage s, LocalEvent l) 
        {
            this.myServer = myServer;
            this.localEvent = l;
            SkipNet.PingSendUpcall sendUpcall = new SkipNet.PingSendUpcall( PingSend );
            SkipNet.PingReceiveUpcall receiveUpcall = new SkipNet.PingReceiveUpcall( PingReceive);
            ((SkipNet.Node)myServer).RegisterPingNotificationUpcall(sendUpcall, receiveUpcall, 13); //HACK
        }

        /// <summary>
        /// Called to appropriately clear state for churn experiments.
        /// Call after resurrecting the node at other levels.
        /// </summary>
        public void Resurrect() 
        {
            failureHandlerTable = new Hashtable();
            dlcEntryTable = new DLCEntryTable(); 
            dlccids_to_nodeIDAddressBindsOfGroupMembersPlusDelegates = new Hashtable();
        }


        Object PingSend( SkipNet.Node localNode, SkipNet.NodeIdAddressBind n ) 
        {
            Log.Write("dbg", LogLvl.Lo, "node {0} is sending a ping to node {1}", 
                localNode.GetNodeIdAddressBind.GetAddress, n.GetAddress);
            if (dlcEntryTable.pingChecksums[ n.GetAddress] != null) 
                Log.Write("dbg", LogLvl.Lo, "HAVE A PING TO SEND");
            Object o = dlcEntryTable.pingChecksums[ n.GetAddress];
            if (o == null) o = NULL_DLC_PING;
            return o;
        }
        void PingReceive ( SkipNet.Node localNode, SkipNet.NodeIdAddressBind source, Object piggybackObj ) 
        {
            CompareChecksum(source,(byte[]) piggybackObj);
        }

        
        //========================= DLCC externally visible methods  =========================//

        /// <summary>
        /// All apps must provide this interface. DLCC does not act on calls to this function.
        /// </summary>
        /// <param name="id"></param>
        public override void NewNeighbour(MessageRouting.NodeIdAddressBind id) 
        {
            Log.Write("dlc", LogLvl.Hi, "NewNeighbour called in node {0} adding node {1}", 
                myServer.GetNodeIdAddressBind.GetAddress, id.GetAddress);
        }

        /// <summary>
        /// All apps must provide this interface. DLCC acts on calls to removed node by
        /// generating a soft notification on every relevant group. 
        /// This is true even if the node was seen to fail and it was a member -- this allows us to use one 
        /// measure of reliability for our overlay-routing table pings without forcing DLC to accept that 
        /// measure of reliability in its decisions.
        /// </summary>
        /// <param name="id">The id of the node being removed from the routing table</param>
        /// <param name="failure">true if the node was observed to have failed</param>
        public override void NeighbourRemoved(MessageRouting.NodeIdAddressBind nodeBeingRemoved, 
            bool nodeObservedToFail)  
        {
            Log.Write("dlc", LogLvl.Hi, "DLCC.NeighbourRemoved called in node {0} for node {1}, failure {2}",
                myServer.GetNodeIdAddressBind.GetAddress, nodeBeingRemoved.GetAddress, nodeObservedToFail);

            ArrayList relevantDlcIds = dlcEntryTable.DlcEntriesWithNeighbor(nodeBeingRemoved);
            foreach (string dlcId in relevantDlcIds) 
            {
                Log.Write("dbg", LogLvl.Hi, "SoftNotification due to NeighbourRemoved on dlcId {0}", dlcId);
                SoftNotificationUpcall(new DlcIdPlusSeqnum(dlcId, dlcEntryTable[dlcId].seqnum), null);
            }
        }

        /// <summary>
        /// Triggers a hard notification on this dlcId. 
        /// Sends out the appropriate hard notification messages77
        /// (to the root if we're a non-root member, to all members if we're the root),
        /// then soft notification messages (to proactively clean up state on other nodes),
        /// then notifies the failure handler on this node, and 
        /// then cleans up state associated with the dlc group on this node.
        /// Returns false if the dlcId is not installed on this node, true if it is.
        /// </summary>
        /// <param name="dlcId">The group to notify.</param>
        /// <returns></returns>
        public bool RaiseException(string dlcId) 
        {
            if (dlcEntryTable[dlcId] != null) 
            {
                HardNotificationUpcall(dlcId, null); // this function does all the work
                return true;
            } 
            else return false;
        }

        /// <summary>
        /// Assign a global failure handler for all dlcIds.
        /// Overrides previously registered failure handlers.
        /// </summary>
        /// <param name="failureHandler">The callback that handles failures.</param>
        public void RegisterFailureNotifier(DLCCFailureNotification failureHandler)
        {
            this.failureHandler = failureHandler;
        }

        /// <summary>
        /// Add a failure handler for a particular id. Overrides previously registered failure handlers for this id.
        /// This callback will be called instead of the global failure handler.
        /// </summary>
        public void RegisterFailureIDNotifier(DLCCFailureNotification particularFailureHandler, string dlcId)
        {
            Debug.Assert(dlcId != null);
            if (dlcEntryTable[dlcId] == null) 
            {
                particularFailureHandler(myServer, dlcId, "dlcId did not exist when handler was registered"); 
            } 
            else 
            {
                failureHandlerTable[dlcId] = particularFailureHandler;
            }
        }

        /// <summary>
        /// Legacy API.
        /// Create a DLC group with two members, root and member. Must be called on root.
        /// The creationCallback will return the dlcId if the group is successfully created, null otherwise.
        /// </summary>
        /// <param name="root"></param>
        /// <param name="member"></param>
        /// <param name="creationCallback"></param>
        public void CreatePath(NodeIdAddressBind root, NodeIdAddressBind member, DLCCContinuation creationCallback) 
        {
            Log.Write("dlc", LogLvl.Hi, "DLCC.Createpath called in node {0} between {1} and {2}",
                myServer.GetNodeIdAddressBind.GetAddress, root.GetAddress, member.GetAddress);
            Debug.Assert(root == myServer.GetNodeIdAddressBind);
            NodeIdAddressBind[] members = new NodeIdAddressBind[2];
            members[0] = root;
            members[1] = member;
            CreateGroupInternal(members, creationCallback);
        }
 
        /// <summary>
        /// Legacy API.
        /// Create a DLC group with the given membership. The root must be the first member of the group,
        /// and the group must be of size 2 or more.
        /// The creationCallback will return the dlcId if the group is successfully created, null otherwise.
        /// </summary>
        /// <param name="members"></param>
        /// <param name="interestedPathElems">This is ignored</param>
        /// <param name="creationCallback"></param>
        public void CreatePath(NodeIdAddressBind[] members, bool[] interestedPathElems, 
            DLCCContinuation creationCallback) 
        {
            Log.Write("dlc", LogLvl.Hi, "DLCC.Createpath called in node {0}", 
                myServer.GetNodeIdAddressBind.GetAddress);
            Debug.Assert(members[0] == myServer.GetNodeIdAddressBind);
            Debug.Assert(members.Length >= 2);
            CreateGroupInternal(members, creationCallback);
        }

        /// <summary>
        /// Preferred API.
        /// Create a DLC Group with the given membership.
        /// </summary>
        /// <param name="members">members[0] is considered root and must be current node</param>
        /// <param name="creationCallback">Callback to execute when create fails/succeeds.</param>
        public void CreateGroup (NodeIdAddressBind[] members, DLCCContinuation creationCallback) 
        {
            Log.Write("dlc", LogLvl.Hi, "DLCC.CreateGroup called in node {0}", 
                myServer.GetNodeIdAddressBind.GetAddress );
            CreateGroupInternal(members, creationCallback);
        }


        /// <summary>
        /// Returns true if the specified DLC group is currently installed on this node 
        /// and includes this node as a member.
        /// </summary>
        /// <param name="dlcId"></param>
        /// <returns></returns>
        public bool Exists(string dlcId) 
        {
            DlcEntry entry = dlcEntryTable[dlcId];
            if (entry == null) return false;
            if (entry.memberState != null || entry.rootState != null) return true;
            else return false;
        }

        //========================= DLCC internal methods  =========================//

        /// <summary>
        /// Generates a globally unique identifier for a new DLC group.
        /// </summary>
        protected DLCid GenerateUniqueIdentifier()
        {
            return new DLCid (myServer.GetNodeIdAddressBind.GetAddress.GetHashCode(), ++dlcGroupCounter);
            // if we wanted to avoid name collisions even when there were SkipNet name collisions, we
            // could use the commented out implementation below
            // return new DLCid (System.DateTime.Today.Ticks, random.Next());
        }

       
        /// <summary>
        /// Delivers the notification to the application using the appropriate handler.
        /// The notification is dropped if no handler exists.
        /// If a failure handler specific to this dlcId is registered, it is called, and then
        /// removed. If no specific failure handler exists, but a 
        /// generic failure handler exists, the failure is delivered to it.
        /// </summary>
        /// <param name="dlcId">The dlc group id on which to signal a notification</param>
        protected void DeliverNotificationToApplication (string dlcId, string message) 
        {
            if (BUFFER_NOTIFICATIONS_FOR_DEBUGGING) // added during debugging to flush out timing bugs
            {
                BufferNotificationAndThenDeliver(dlcId, message);
                return;
            } 

            DLCCFailureNotification particularFailureHandler = (DLCCFailureNotification) failureHandlerTable[dlcId];
            if (particularFailureHandler != null) 
            {
                particularFailureHandler(myServer, dlcId, message);
                failureHandlerTable.Remove(dlcId);
            } 
            else if (failureHandler != null) 
            {
                failureHandler(myServer, dlcId, message);
            }
        }
        
        public bool BUFFER_NOTIFICATIONS_FOR_DEBUGGING = false;

        /// <summary>
        /// I hypothesize that the current SV BVT assert I am tracking down is really a timing bug in SV. 
        /// This will help me test that hypothesis by adding in a short delay to DLC notification.
        /// </summary>
        /// <param name="dlcId"></param>
        /// <param name="message"></param>
        protected void BufferNotificationAndThenDeliver(string dlcId, string message) 
        {
            DLCCFailureNotification particularFailureHandler = (DLCCFailureNotification) failureHandlerTable[dlcId];
            if (particularFailureHandler != null) 
            {
                failureHandlerTable.Remove(dlcId);
                BufferedNotificationTimer bnt = new BufferedNotificationTimer(particularFailureHandler,
                    dlcId, message);
                myServer.SetTimer(bnt);
            } 
            else if (failureHandler != null) 
            {
                BufferedNotificationTimer bnt = new BufferedNotificationTimer(failureHandler,
                    dlcId, message);
                myServer.SetTimer(bnt);
            }
        }

        public class BufferedNotificationTimer : Timer 
        {
            DLCCFailureNotification failureHandler;
            string dlcId;
            string message;
            public BufferedNotificationTimer(DLCCFailureNotification failureHandler, string dlcId, 
                string message) : base(60*1000) //15 secs
            {
                this.failureHandler = failureHandler;
                this.dlcId = dlcId;
                this.message = message;
            }
                
            public override void Execute(MessageRouting.Node node) 
            {
                    failureHandler(node, dlcId, message);
            }
        }

        




        /// <summary>
        /// This function processes a hard notification on this dlcId.
        /// It is the internal version of RaiseException, and it can be called both 
        /// from a message passed in to this node, and from an application above it.
        /// 
        /// This function sends out the appropriate hard notification messages
        /// (to the root if we're a non-root member, to all members if we're the root),
        /// then soft notification messages (to proactively clean up state on other nodes),
        /// then cleans up state associated with the dlc group on this node, and
        /// then notifies the failure handler on this node.
        /// </summary>
        /// <param name="dlcId">The group to notify.</param>
        /// <param name="source">The node from which the HardNotificationMsg was heard. 
        /// We don't send the HardNotificationMsg back to the node that sent it to us.
        /// Set to null if notification is generated internally. </param>
        public void HardNotificationUpcall(string dlcId, NodeIdAddressBind source) 
        {
            DlcEntry entry = dlcEntryTable[dlcId];
            if (entry == null) return;

            // 1. Generate list of nodes to send notifications.
            Hashtable nodesToSendHardNotification = new Hashtable();
            Hashtable nodesToSendSoftNotification = new Hashtable();

            if (entry.rootState != null) 
            {
                foreach (NodeIdAddressBind n in entry.rootState.conditionOfMembers.Keys)
                    if (n != source) nodesToSendHardNotification[n] = null;
            }
            else 
            {
                if (entry.memberState != null)
                    if (entry.memberState.root != source && entry.memberState.root != null) // BAD DEFN: null also used to denote that root is self
                        nodesToSendHardNotification[entry.memberState.root] = null;
            }
            
            if (entry.delegateState != null)
                foreach(DlcPath path in entry.delegateState.dlcPaths) 
                {
                    if (path.neighborTowardsMember != null &&
                        !nodesToSendHardNotification.ContainsKey(path.neighborTowardsMember) &&
                        path.neighborTowardsMember != source)
                        nodesToSendSoftNotification[path.neighborTowardsMember] = null;
                    if (path.neighborTowardsRoot != null &&
                        !nodesToSendHardNotification.ContainsKey(path.neighborTowardsRoot) &&
                        path.neighborTowardsRoot != source)
                        nodesToSendSoftNotification[path.neighborTowardsRoot] = null;
                }

            // 2. Send out the notifications.
            foreach(NodeIdAddressBind n in nodesToSendHardNotification.Keys) 
            {
                DLCHardNotificationMsg msg = new DLCHardNotificationMsg(dlcId, myServer.GetNodeIdAddressBind);
                myServer.SendMessageDirect(msg, n);
            }
            foreach(NodeIdAddressBind n in nodesToSendSoftNotification.Keys) 
            {   // seqnum set so no one ignores this soft notification
                DLCSoftNotificationMsg msg = new DLCSoftNotificationMsg(new 
                    DlcIdPlusSeqnum(dlcId, DLC_INFINITY), myServer.GetNodeIdAddressBind);
                myServer.SendMessageDirect(msg, n);
            }

            // 3. Clean up internal state associated with this group.
            if (entry.memberState != null && entry.memberState.recreateTimer != null)
                myServer.CancelTimer(entry.memberState.recreateTimer);
            if (entry.rootState != null && entry.rootState.createRecreateTimer != null)
                myServer.CancelTimer(entry.rootState.createRecreateTimer);
            if (entry.rootState != null && entry.rootState.installCheckingTimer != null)
                myServer.CancelTimer(entry.rootState.installCheckingTimer);
            if (entry.delegateState != null)
                foreach (DlcPath path in entry.delegateState.dlcPaths) 
                    path.dlcPathRefreshTimer.Remove(myServer);
            
            // Before removing entry, save away information that might be 
            // neccesary for notifying the application
            DLCCContinuation appCreationCallback = null;
            if (entry.rootState != null &&
                entry.rootState.rootStateCondition == DlcRootState.RootStateCondition.Creating)
            {
                Debug.Assert(entry.rootState.appCreationCallback != null);
                appCreationCallback = entry.rootState.appCreationCallback;
            }

            dlcEntryTable.Remove(dlcId);
            dlcEntryTable.Refresh(); // must call after changing the dlcEntryTable state to update ping calculations

            // 4. If this group is the root and we're not done creating, signal creation failure. 
            // If we are done creating, call the appropriate failure handler.
            if (appCreationCallback != null) 
            {
                appCreationCallback.DLCCId = null;
                appCreationCallback.ErrMsg = "Failed create";
                appCreationCallback.Execute(myServer);
            }
            else 
                DeliverNotificationToApplication(dlcId, "you have been notified");
        }
        

        /// <summary>
        /// This function processes a soft notification on this dlcIdPlusSeqnum. If the seqnum being passed 
        /// in is LESS THAN the seqnum of record, this is a no-op. 
        /// 
        /// Otherwise, what we do depends on the the current state of this node. If we're the root, 
        /// and in a creating phase, we need to ensure that, if create finishes,
        /// a re-create occurs immediately thereafter. We do this by modifying the createRecreateTimer.
        /// If we're the root and in a re-creating phase, we immediately move to recreate on the 
        /// next higher sequence number. If we're the root and in a live phase, we just go to re-create.
        /// If we're a member, we go to the recreating state.
        /// For all node roles, we wipe out the delegate state.
        /// 
        /// WARNING: It is not the responsibility of this function to avoid the 
        /// group-creation/ping-reconciliation race condition.
        /// </summary>
        /// <param name="dlcIdPlusSeqnum">The group and sequence number to notify.</param>
        /// <param name="source">The node from which we received the soft notification.
        /// We avoid sending a soft notification back to the person we heard this from.
        /// Set to null if this node is the source of the soft notification.</param>
        public void SoftNotificationUpcall(DlcIdPlusSeqnum dlcIdPlusSeqnum, NodeIdAddressBind source) 
        {
            Log.Write("dbg", LogLvl.Hi, "SoftNotification entered on dlcIdPlusSeqnum {0}", 
                dlcIdPlusSeqnum.ToString());
            string dlcId = dlcIdPlusSeqnum.dlcId;
            DlcEntry entry = dlcEntryTable[dlcId];
            if (entry == null) return;
            if (entry.seqnum > dlcIdPlusSeqnum.seqnum) return;
            
            if (entry.rootState != null && 
                entry.rootState.rootStateCondition != DlcRootState.RootStateCondition.Live) 
            {
                Log.Write("dbg", LogLvl.Hi, "SoftNotification set up for immediate recreate on dlcIdPlusSeqnum {0}", 
                    dlcIdPlusSeqnum.ToString());
                // We shouldn't need to check this null condition, but since we added in the hard
                // notification buffering for debugging, we've expanded the state space (sadly).
                // The hard notification buffering should definitely be removed soon.
                // JD 10-27-03
                if (entry.rootState.createRecreateTimer != null)
                    entry.rootState.createRecreateTimer.doImmediateRecreate();
                return;
            }

            entry.seqnum = dlcIdPlusSeqnum.seqnum;

            // Create soft notifications and clean up delegate state.
            Hashtable nodesToSendSoftNotification = new Hashtable();
            if (entry.delegateState != null) 
            {
                foreach(DlcPath path in entry.delegateState.dlcPaths) 
                {
                    if (path.neighborTowardsMember != null &&
                        source != path.neighborTowardsMember)
                        nodesToSendSoftNotification[path.neighborTowardsMember] = null;
                    if (path.neighborTowardsRoot != null &&
                        source != path.neighborTowardsRoot)
                        nodesToSendSoftNotification[path.neighborTowardsRoot] = null;
                    if (path.dlcPathRefreshTimer != null) 
                        path.dlcPathRefreshTimer.Remove(myServer);
                }
                entry.delegateState = null;
            }


            if (entry.memberState != null) 
            {   // 1. Set the state here to be recreating, 2. set a timer for recreation to succeed, 
                // 3. send a "need recreate" message to the root.
                entry.memberState.memberStateCondition = DlcMemberState.MemberStateCondition.Recreating;
                if (entry.memberState.recreateTimer == null) // don't override a previously existing timer
                {
                    entry.memberState.recreateTimer = new RecreateTimer(60*1000, dlcId);
                    myServer.SetTimer(entry.memberState.recreateTimer);
                }
                DLCNeedRecreateMsg msg = new DLCNeedRecreateMsg(myServer.GetNodeIdAddressBind, 
                    entry.memberState.root, dlcIdPlusSeqnum);
                myServer.SendMessageDirect(msg, entry.memberState.root);
            }

            if (entry.rootState != null) 
            {   // We already know that the rootState is not "creating," so we know to 
                // do re-creation. Re-creation has been separated out in to another function.
                Log.Write("dbg", LogLvl.Hi, "SoftNotification do recreate on dlcIdPlusSeqnum {0}", 
                    dlcIdPlusSeqnum.ToString());
                if (entry.rootState.rootStateCondition == DlcRootState.RootStateCondition.Live) 
                {
                    entry.rootState.rootStateCondition = DlcRootState.RootStateCondition.Recreating;
                    RecreateGroupInternal(dlcId);
                } 
                /* else if (entry.rootState.rootStateCondition == DlcRootState.RootStateCondition.Recreating) 
                {   
                    CreateRecreateTimer timer = entry.rootState.createRecreateTimer;
                    timer.doImmediateRecreate();
                } */

            } 

            // Send the soft notifications.
            foreach(NodeIdAddressBind n in nodesToSendSoftNotification.Keys) 
            {   
                Log.Write("dbg", LogLvl.Hi, "SoftNotificationMsg create on dlcId {0} in soft notificationupcall",
                    dlcIdPlusSeqnum.ToString());
            
                DLCSoftNotificationMsg msg = new DLCSoftNotificationMsg(dlcIdPlusSeqnum, 
                    myServer.GetNodeIdAddressBind);
                myServer.SendMessageDirect(msg, n);
            }
            
            if (entry.memberState == null && entry.rootState == null) dlcEntryTable.Remove(dlcId);
            // Refresh the ping checksums to be safe regardless of whether any changes occurred.
            dlcEntryTable.Refresh(); 
            Log.Write("dbg", LogLvl.Hi, "SoftNotification done on dlcIdPlusSeqnum {0}", 
                dlcIdPlusSeqnum.ToString());
        }

        /// <summary>
        /// When a hash piggybacked on a ping arrives, it calls this function to check the payload against 
        /// this nodes' set of live dlc groups.
        /// If the pings don't match, a DLCIdListRequestMsg is sent back to neighbor.
        /// </summary>
        /// <param name="neighbor">The source of the checksum.</param>
        /// <param name="checksum">The checksum.</param>
        protected void CompareChecksum(NodeIdAddressBind neighbor, byte[] checksum) 
        {
            Log.Write("dbg", LogLvl.Med, "CompareChecksum entry point");
            if (checksum[0] == 0 && checksum.Length == 1) // DLCC.NULL_DLC_PING 
            {
                Log.Write("dbg", LogLvl.Med, "null ping from neighbor {0}", neighbor.GetAddress);
                return;
            }
            byte[] myChecksum = (byte[]) dlcEntryTable.pingChecksums[neighbor.GetAddress];
            bool differenceDetected = false;
            if (myChecksum == null || checksum.Length != myChecksum.Length)
                differenceDetected = true;
            else 
            {
                for (int i = 0; i < checksum.Length; i++) 
                {
                    if (checksum[i] != myChecksum[i])
                        differenceDetected = true;
                }
            }
            if (differenceDetected) 
            {
                Log.Write("dbg", LogLvl.Med, "ping upcall from neighbor {0} DID NOT MATCH", neighbor.GetAddress);
                ArrayList dlcIdsWhereNeighbor = dlcEntryTable.DlcEntriesWithNeighbor(neighbor);
                ArrayList dlcIdsPlusSeqnums = new ArrayList();
                foreach(string dlcId in dlcIdsWhereNeighbor)
                    dlcIdsPlusSeqnums.Add(new DlcIdPlusSeqnum(dlcId, dlcEntryTable[dlcId].seqnum));
                DLCIdListRequestMsg msg = new DLCIdListRequestMsg(dlcIdsPlusSeqnums, 
                    myServer.GetNodeIdAddressBind, neighbor);
                myServer.SendMessageDirect(msg, neighbor);
            } 
            else 
            {
                Log.Write("dbg", LogLvl.Med, "ping upcall from neighbor {0} MATCHED", neighbor.GetAddress);
                ArrayList dlcIdsWhereNeighbor = dlcEntryTable.DlcEntriesWithNeighbor(neighbor);
                Log.Write("dbg", LogLvl.Med, "ping upcall from neighbor {0} MATCHED --> {1} relevant dlcIds", neighbor.GetAddress, dlcIdsWhereNeighbor.Count);
                foreach(string dlcId in dlcIdsWhereNeighbor) 
                {
                    Log.Write("dbg", LogLvl.Med, "ping upcall from neighbor {0} MATCHED --> reset on dlcid {1}", neighbor.GetAddress, dlcId);
                    DlcEntry entry = dlcEntryTable[dlcId];
                    foreach(DlcPath path in entry.delegateState.dlcPaths) 
                    {
                        path.dlcPathRefreshTimer.Reset(myServer);
                    }
                }
            }
        }
        
             
        /// <summary>
        /// Internal function to create a group. Called only when initially creating a group, 
        /// not for recreates. Sends messages to every member.
        /// </summary>
        /// <param name="group">Group members - must contain more than 2 members, 
        /// and 1st member must be this node. Members cannot be repeated.</param>
        /// <param name="creationCallback">Application level callback for when creation fails/succeeds. 
        /// Cannot be null.</param>        
        protected void CreateGroupInternal (NodeIdAddressBind[] group, DLCCContinuation appCreationCallback) 
        {
            Debug.Assert(group.Length > 1);
            Debug.Assert(group[0] == myServer.GetNodeIdAddressBind);
            for (int i = 0; i < group.Length; i++)
                for (int j = 0; j < group.Length; j++)
                    Debug.Assert(i == j || group[i] != group[j]);
            Debug.Assert(appCreationCallback != null);

            string dlcId = GenerateUniqueIdentifier().ToString();

            // Create a DlcEntry with rootState and timers.
            Debug.Assert(dlcEntryTable[dlcId] == null);
            dlcEntryTable[dlcId] = new DlcEntry(dlcId);
            ArrayList nonRootMembers = new ArrayList();
            for (int i = 1; i < group.Length; i++) nonRootMembers.Add(group[i]);
            dlcEntryTable[dlcId].rootState = new DlcRootState(nonRootMembers);
            dlcEntryTable[dlcId].rootState.appCreationCallback = appCreationCallback;
            CreateRecreateTimer createRecreateTimer = new CreateRecreateTimer(2*60 * 1000, dlcId);
            myServer.SetTimer(createRecreateTimer);
            dlcEntryTable[dlcId].rootState.createRecreateTimer = createRecreateTimer;
            InstallCheckingTimer installCheckingTimer = new InstallCheckingTimer(60 * 1000, 
                new DlcIdPlusSeqnum(dlcId, 0));
            myServer.SetTimer(installCheckingTimer);
            dlcEntryTable[dlcId].rootState.installCheckingTimer = installCheckingTimer;

            // Send out a DLCCreateGroupRequestMsg to all the non-root members.
            for (int i = 1; i < group.Length; i++) 
            {
                NodeIdAddressBind member = group[i];
                CreateGroupInternalCallback internalCreationCallback = new 
                    CreateGroupInternalCallback(dlcId, member, this);
                DLCCreateGroupRequestMsg msg = new DLCCreateGroupRequestMsg(myServer, 
                    internalCreationCallback, myServer.GetNodeIdAddressBind, member, dlcId);
                myServer.SendMessageDirect(msg, member);
            }
        }

        /// <summary>
        /// The callback executed by a DlcGroupCreateReplyMsg. 
        /// Handles the second half of CreateGroupInternal's job.
        /// </summary>
        protected class CreateGroupInternalCallback : OverlayReplyCallback 
        {
            string dlcId;
            NodeIdAddressBind member;
            DLCC dlcApp;
            
            public CreateGroupInternalCallback (string dlcId, 
                NodeIdAddressBind member, DLCC dlcApp) 
            {
                this.dlcId = dlcId;
                this.dlcApp = dlcApp;
                this.member = member;
            }

            public override void Execute(MessageRouting.Node localNode)
            {
                // 4 cases: 
                // 1. this callback is irrelevant b/c this create has already been declared a failure
                // 2. this callback is the one where create fails
                // 3. this callback is the one where create succeeds
                // 4. the verdict is still open - update state, but otherwise 
                // do nothing and wait for another callback
                // ----------------------
                // If we look in the list of entries being created in the dlcEntryTable and this one isn't there, 
                // we're in case 1. We exit. Otherwise, we look if this callback was a timeout. If it was,
                // we're in case 2. Otherwise, we look to see if this callback corresponds to filling in 
                // the last missing entry in dlcEntry. If so, we're in case 3. Otherwise, we're in case 4.
                Log.Write("dbg", LogLvl.Hi, "CreateGroupCallback on dlcid {0}", dlcId);

                if (dlcApp.dlcEntryTable[dlcId] == null) return; // case 1.
                Log.Write("dbg", LogLvl.Hi, "CreateGroupCallback continuing on dlcid {0}", dlcId);

                DlcEntry entry = dlcApp.dlcEntryTable[dlcId];
                if (this.Reply == null) 
                {   // case 2.
                    Log.Write("dbg", LogLvl.Hi, "CreateGroupCallback failing on dlcid {0}", dlcId);
                    dlcApp.HardNotificationUpcall(dlcId, null);
                } 
                else 
                {
                    Debug.Assert(((DLCCreateGroupReplyMsg) this.Reply).Member == member);
                    Debug.Assert(((DLCCreateGroupReplyMsg) this.Reply).Root == dlcApp.myServer.GetNodeIdAddressBind);
                    Debug.Assert(entry.rootState.rootStateCondition == DlcRootState.RootStateCondition.Creating);
                    DlcRootState.ConditionOfMember condition = (DlcRootState.ConditionOfMember) entry.rootState.conditionOfMembers[member];
                    condition.haveReceivedCreateRecreateReply = true;
                    if (entry.rootState.isCreateRecreateComplete()) 
                    {   // case 3.
                        Log.Write("dbg", LogLvl.Hi, "CreateGroupCallback succeeding on dlcid {0}", dlcId);
                        CreateRecreateTimer timer = entry.rootState.createRecreateTimer; // create local ref
                        entry.rootState.createRecreateTimer = null; // remove non-local ref
                        dlcApp.myServer.CancelTimer(timer); // remove scheduled firing
                        entry.rootState.rootStateCondition = DlcRootState.RootStateCondition.Live;
                        // return success to application
                        entry.rootState.appCreationCallback.ErrMsg = null;
                        entry.rootState.appCreationCallback.DLCCId = dlcId;
                        entry.rootState.appCreationCallback.Execute(dlcApp.myServer);
                        // if we're in an immediate recreate situation, do so now
                        if(timer.isImmediateRecreate()) 
                        {
                            Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c immediate recreate was set on dlcId {0}", dlcId);
                            dlcApp.SoftNotificationUpcall(new DlcIdPlusSeqnum(dlcId, 0), null);
                        }
                        

                    } 
                    else // otherwise we're in case 4. There's nothing to do.
                    {
                        Log.Write("dbg", LogLvl.Hi, "CreateGroupCallback not yet conclusive on dlcid {0}", dlcId);
                    }
                }
            }
        }

        /// <summary>
        /// Implements backoff for group recreate by interposing between this and
        /// RecreateGroupInternalImmmediate.
        /// </summary>
        /// <param name="dlcId"></param>
        protected void RecreateGroupInternal(string dlcId) 
        {
            if (dlcidToRecreateGroupBackoffInfo[dlcId] == null) 
            {
                dlcidToRecreateGroupBackoffInfo[dlcId] = new RecreateGroupBackoffInfo(localEvent.GetTime()); 
                RecreateGroupInternalImmediate(dlcId);
            } 
            else 
            {
                RecreateGroupBackoffInfo info = (RecreateGroupBackoffInfo) dlcidToRecreateGroupBackoffInfo[dlcId];
                long elapsed = localEvent.GetTime() - info.previousRecreateAttemptTime;
                Debug.Assert(elapsed >= 0);
                // schedule the next attempt for either 5,10,20, or 40 secs since the most recent one
                long offsetFromPreviousToForce;
                if (info.previousRecreateAttemptIncurredOffset < 5*1000) 
                {
                    offsetFromPreviousToForce = 5*1000;
                } 
                else if (info.previousRecreateAttemptIncurredOffset < 10*1000) 
                {
                    offsetFromPreviousToForce = 10*1000;
                } 
                else if (info.previousRecreateAttemptIncurredOffset < 20*1000) 
                {
                    offsetFromPreviousToForce = 20*1000;
                } 
                else 
                {
                    offsetFromPreviousToForce = 40*1000;
                } 
                if (elapsed < offsetFromPreviousToForce) 
                {
                    long offsetFromCurrent = offsetFromPreviousToForce - elapsed;
                    info.previousRecreateAttemptTime = localEvent.GetTime() + offsetFromCurrent;
                    info.previousRecreateAttemptIncurredOffset = offsetFromCurrent;
                    RecreateGroupBuffered bufferedRecreate = new RecreateGroupBuffered(dlcId, 5*1000);
                    myServer.SetTimer(bufferedRecreate); 
                } 
                else 
                {
                    info.previousRecreateAttemptTime = localEvent.GetTime();
                    info.previousRecreateAttemptIncurredOffset = 0;
                    RecreateGroupInternalImmediate(dlcId);
                }
            }
        }
   
        protected Hashtable dlcidToRecreateGroupBackoffInfo;
        // we should put this info in dlcEntry, but I'm in a rush - 
        // it's the Saturday 10 days before the usenix deadline - JD

        protected class RecreateGroupBackoffInfo 
        {
            public long previousRecreateAttemptIncurredOffset;
            public long previousRecreateAttemptTime;
            public RecreateGroupBackoffInfo(long nowMsecs) 
            {
                previousRecreateAttemptIncurredOffset = 0;
                previousRecreateAttemptTime = nowMsecs;
            }

        }

        protected class RecreateGroupBuffered : Timer {
            string dlcId;
            public RecreateGroupBuffered(string dlcId, long msecDelay) : base (msecDelay) {this.dlcId = dlcId;}
            public override void Execute(MessageRouting.Node node) {
                DLCC dlcc = (DLCC) node.GetApp(typeof(DLCC));
                dlcc.RecreateGroupInternalImmediate(dlcId);
            }
        }

        /// <summary>
        /// Internal function to recreate a group. Called only when recreating a group, 
        /// not for initial creates. Sends messages to every member.
        /// Had to make it public in order to interact nicely with our callbacks.
        /// </summary>
        /// <param name="group">Group members - must contain more than 2 members, 
        /// and 1st member must be this node. Members cannot be repeated.</param>
        /// <param name="creationCallback">Application level callback for when creation fails/succeeds.</param>        
        public void RecreateGroupInternalImmediate (string dlcId) 
        {
            Debug.Assert(dlcEntryTable[dlcId] != null);
            Debug.Assert(dlcEntryTable[dlcId].rootState != null);
            Debug.Assert(dlcEntryTable[dlcId].rootState.rootStateCondition == DlcRootState.RootStateCondition.Recreating);
            DlcEntry entry = dlcEntryTable[dlcId];
            DlcRootState rootState = entry.rootState;
                        
            // The current implementation of DLC only calls this function from within SoftNotificationUpcall,
            // but we may or may not have finished a previous re-creation.
            foreach (NodeIdAddressBind member in rootState.conditionOfMembers.Keys) 
            {
                ((DlcRootState.ConditionOfMember) rootState.conditionOfMembers[member]).haveReceivedCreateRecreateReply = false;
                ((DlcRootState.ConditionOfMember) rootState.conditionOfMembers[member]).haveReceivedInstallChecking = false;
            }
            entry.seqnum++;
            DlcIdPlusSeqnum dlcIdPlusSeqnum = new DlcIdPlusSeqnum(dlcId, entry.seqnum);

            // Create timers if they don't already exist.
            if (rootState.createRecreateTimer == null) 
            {
                rootState.createRecreateTimer = new CreateRecreateTimer(2*60 * 1000, dlcId);
                myServer.SetTimer(rootState.createRecreateTimer);
            }
            if (rootState.installCheckingTimer == null) 
            {
                rootState.installCheckingTimer = new InstallCheckingTimer(60 * 1000, dlcIdPlusSeqnum);
                myServer.SetTimer(rootState.installCheckingTimer);
            }

            // Send out a DLCRecreateGroupRequestMsg to all the non-root members.
            foreach (NodeIdAddressBind member in rootState.conditionOfMembers.Keys)
            {
                RecreateGroupInternalCallback internalRecreationCallback = new 
                    RecreateGroupInternalCallback(dlcIdPlusSeqnum, member, this);
                DLCRecreateGroupRequestMsg msg = new DLCRecreateGroupRequestMsg(myServer, 
                    internalRecreationCallback, myServer.GetNodeIdAddressBind, member, dlcIdPlusSeqnum);
                myServer.SendMessageDirect(msg, member);
            }
        }

        /// <summary>
        /// The callback executed by a DLCRecreateGroupReplyMsg. 
        /// Handles the second half of RecreateGroupInternal's job.
        /// </summary>
        protected class RecreateGroupInternalCallback : OverlayReplyCallback 
        {
            DlcIdPlusSeqnum dlcIdPlusSeqnum;
            NodeIdAddressBind member;
            DLCC dlcApp;
            
            public RecreateGroupInternalCallback (DlcIdPlusSeqnum dlcIdPlusSeqnum,  
                NodeIdAddressBind member, DLCC dlcApp) 
            {
                this.dlcIdPlusSeqnum = dlcIdPlusSeqnum;
                this.dlcApp = dlcApp;
                this.member = member;
            }

            public override void Execute(MessageRouting.Node localNode)
            {
                // 4 cases: 
                // 1. this callback is irrelevant b/c this recreate has already been declared a failure
                //    or this callback is occurring with a recreate msg that's out-of-date (old seqnum)
                // 2. this callback is the one where recreate fails
                // 3. this callback is the one where recreate succeeds
                // 4. the verdict is still open - update state, but otherwise 
                // do nothing and wait for another callback
                // ----------------------
                // If we look in the list of entries in the dlcEntryTable and this one isn't there, 
                // we're in case 1. We exit. Otherwise, we look if this callback was a timeout. If it was,
                // we're in case 2. Otherwise, we look to see if this callback corresponds to filling in 
                // the last missing entry in dlcEntry. If so, we're in case 3. Otherwise, we're in case 4.
                string dlcId = dlcIdPlusSeqnum.dlcId;
                Log.Write("dbg", LogLvl.Hi, "RecreateGroupCallback initiated on dlcid {0}", dlcId);
                if (dlcApp.dlcEntryTable[dlcId] == null ||
                    dlcApp.dlcEntryTable[dlcId].rootState == null ||
                    dlcApp.dlcEntryTable[dlcId].seqnum > dlcIdPlusSeqnum.seqnum) return; // case 1.

                Log.Write("dbg", LogLvl.Hi, "RecreateGroupCallback continuing on dlcid {0}", dlcId);

                DlcEntry entry = dlcApp.dlcEntryTable[dlcId];
                // Since we suppress duplicate messages, if the DlcEntry exists on this seqnum, 
                // we should still be re-creating it. If we've already moved to a higher seqnum, 
                // we won't reach this point in the code.
                Debug.Assert(entry.rootState.rootStateCondition == DlcRootState.RootStateCondition.Recreating);
                if (this.Reply == null) 
                {   // case 2.
                    Log.Write("dbg", LogLvl.Hi, "RecreateGroupCallback failing on dlcid {0}", dlcId);
                    dlcApp.HardNotificationUpcall(dlcId, null);
                } 
                else 
                {
                    Debug.Assert(((DLCRecreateGroupReplyMsg) this.Reply).Member == member);
                    DlcRootState.ConditionOfMember condition = (DlcRootState.ConditionOfMember) entry.rootState.conditionOfMembers[member];
                    Debug.Assert(!condition.haveReceivedCreateRecreateReply);
                    condition.haveReceivedCreateRecreateReply = true;
                    if (entry.rootState.isCreateRecreateComplete()) 
                    {   // case 3.
                        Log.Write("dbg", LogLvl.Hi, "RecreateGroupCallback succeeding on dlcid {0}", dlcId);
                        CreateRecreateTimer timer = entry.rootState.createRecreateTimer; // create local ref
                        entry.rootState.createRecreateTimer = null; // remove non-local ref
                        dlcApp.myServer.CancelTimer(timer); // remove scheduled firing
                        entry.rootState.rootStateCondition = DlcRootState.RootStateCondition.Live;
                        // if we're in an immediate recreate situation, do so now
                        if(timer.isImmediateRecreate()) 
                        {
                            Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c immediate recreate was set on dlcId {0}", dlcId);
                            dlcApp.SoftNotificationUpcall(dlcIdPlusSeqnum, null);
                        }
                        

                    } 
                    else // otherwise we're in case 4. There's nothing to do.
                    {
                        Log.Write("dbg", LogLvl.Hi, "RecreateGroupCallback not yet conclusive on dlcid {0}", dlcId);
                    }
                }
            }
        }


    }
}
    

