using System;
using System.Collections;
using System.Diagnostics;
using NodeIdAddressBind = MessageRouting.NodeIdAddressBind;
using System.Security.Cryptography;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.DLC
{
    /// <summary>
    /// The DlcEntryTable contains one DlcEntry per DlcId.
    /// 
    /// Each DlcEntry can contain member state, root state, and delegate state. 
    /// The root state contains whether there is an outstanding create/re-create.
    /// The delegate state can contain several DlcPaths. A DlcPath is a chain between a root node and a non-root 
    /// group member node.
    /// 
    /// The DlcEntryTable is not just a structured hashtable, it has additional features. Some examples are:
    /// 1. It allows one to ask for all the DlcsEntries on which a given node plays a role.
    /// 2. It maintains the information we piggyback on pings. This information is:
    ///     for a given neighbor, a simple hash of all the DlcIdPlusSeqnum's of the 
    ///     DlcPaths passing through that neighbor.
    /// The current implementation does not carefully keep track of the relationships between information that 
    /// impacts the ping checksums. Instead, it requires that a client of the class call Refresh() after 
    /// making a change to any DLCEntry in order to make the class aware of the change; otherwise feature 
    /// 2. above returns stale information.
    /// The interface to this class has been designed so that Refresh() can be made a noop without
    /// changing the interface -- this would require making all accesses occur through properties,
    /// which would update the appropriate additional information themselves. This would increase 
    /// the efficiency of the implementation at the cost of complexity, and so it is currently postponed.
    /// </summary>
    public class DLCEntryTable
    {
        // member data
        protected   Hashtable dlcEntries;
    
        // TODO: we would like the outside world to treat pingChecksums as read-only
        // this can be done by subclassing the Hashtable and overriding the get call.
        public      Hashtable pingChecksums; 

        // This SHA1 implementation is for generating hashes of the list of DlcsIdPlusSeqnums we share
        // with another given node.
        private static SHA1 md = new SHA1Managed();

        /// <summary>
        /// Create this on DLCC initialization
        /// </summary>
        public DLCEntryTable() 
        {
            dlcEntries = new Hashtable();
            pingChecksums = new Hashtable();
        }

        /// <summary>
        /// DlcEntryTable is a hashtable, plus additional features.
        /// On get, the Hashtable semantics apply -- accesses based on a 
        /// key that is not in the Hashtable return a null value.
        /// On set, the Hashtable semantics similarly apply.
        /// </summary>
        /// <param name="dlcId">The key to the Hashtable.</param>
        /// <returns></returns>      
        public DlcEntry this[string dlcId] 
        {
            get { return (DlcEntry) dlcEntries[dlcId]; }
            set { dlcEntries[dlcId]= value; }
        }

        /// <summary>
        /// Removes the item from the hashtable.
        /// </summary>
        /// <param name="dlcId">Key to object being removed.</param>
        public void Remove(string dlcId) 
        {
            Debug.Assert(dlcEntries.ContainsKey(dlcId));
            dlcEntries.Remove(dlcId);
        }
         


        /// <summary>
        /// Retrieve an Arraylist of the dlcIds (strings) on which this node plays a role.
        /// Returns an empty ArrayList if there are no such entries.
        /// If the node is not a neighbor, then there should be no such entries
        /// (assuming that the RemoveNeighbor upcall has already occurred).
        /// </summary>
        /// <param name="node">the neighbor in question</param>
        /// <returns></returns>
        public ArrayList DlcEntriesWithNeighbor (NodeIdAddressBind node) 
        {
            Log.Write("dbg", LogLvl.Med, "DlcEntriesWithNeighbor for node {0}", node.GetAddress);
            // In the future, we could cache this information or precompute it.
            // but we're not bothering to do so in this implementation.
            ArrayList dlcIdsOfRelevantEntries = new ArrayList();
            foreach (string dlcId in dlcEntries.Keys) 
            {
                bool found = false;
                DlcEntry entry = (DlcEntry) dlcEntries[dlcId];
                if (entry.delegateState != null) 
                    foreach (DlcPath path in entry.delegateState.dlcPaths) 
                    {
                        //if (path.neighborTowardsMember == node || path.neighborTowardsRoot == node
                        //    || path.member == node || path.root == node) found = true;
                        if ((path.neighborTowardsMember != null && 
                             path.neighborTowardsMember.GetAddress.ToString() == node.GetAddress.ToString()) 
                            || path.neighborTowardsRoot != null &&  
                           path.neighborTowardsRoot.GetAddress.ToString() == node.GetAddress.ToString())
                            found = true;

                        // It's unfortunate that we earlier defined our invariants so that the root being a neighbor means
                        // we don't have to specify neighborTowardsRoot == root.
                        // This should be changed in the postUSENIX timeframe.

                        // DOH. The problem was not over-riding equality; the invariant was fine.

                        if (node.GetAddress.ToString() == "10.0.2.130:10086") 
                        { // for debug
                            Log.Write("dbg", LogLvl.Med, "path info: path.neighborTowardsMember = {0}" +
                                " path.neighborTowardsRoot = {1}" +
                                " path.member = {2}" +
                                " path.root ={3}, found?={4}, dlcId={5}", 
                                path.neighborTowardsMember == null ? "NULL" : path.neighborTowardsMember.GetAddress.ToString(), 
                                path.neighborTowardsRoot == null ? "NULL" : path.neighborTowardsRoot.GetAddress.ToString(),
                                path.member == null ? "NULL" : path.member.GetAddress.ToString(), 
                                path.root == null ? "NULL" : path.root.GetAddress.ToString(),
                                found, dlcId);
                        }
                    }
                if (found) dlcIdsOfRelevantEntries.Add(dlcId);
            }
            // need to unique-ify //HACK
            Hashtable h = new Hashtable();
            foreach (string dlcId in dlcIdsOfRelevantEntries) h[dlcId] = "";
            dlcIdsOfRelevantEntries.Clear();
            foreach (string dlcId in h.Keys) dlcIdsOfRelevantEntries.Add(dlcId);
            return dlcIdsOfRelevantEntries;
        }
       
        /// <summary>
        /// Call this after making any changes in order to refresh the piggybacked ping information.
        /// </summary>
        public void Refresh() 
        {
            // Look through all the dlc information. For every node that's a neighbor,
            // generate a sorted list of dlcIdPlusSeqnums passing through that node.
            Hashtable nodeToListOfDlcIds = new Hashtable();
            foreach (string dlcId in dlcEntries.Keys) 
            {                
                DlcEntry entry = (DlcEntry) dlcEntries[dlcId];
                if (entry.delegateState != null)
                    foreach (DlcPath path in entry.delegateState.dlcPaths) 
                    {
                        string s_dlcIdPlusSeqnum = dlcId + "." + entry.seqnum.ToString();
                        NodeIdAddressBind neighbor;

                        if (path.neighborTowardsMember != null) 
                        {
                            neighbor = path.neighborTowardsMember;
                            SortedList listOfDlcIds = (SortedList) nodeToListOfDlcIds[neighbor];
                            if (listOfDlcIds == null) listOfDlcIds = new SortedList();
                            if (!listOfDlcIds.Contains(s_dlcIdPlusSeqnum)) listOfDlcIds.Add(s_dlcIdPlusSeqnum, null);
                            nodeToListOfDlcIds[neighbor] = listOfDlcIds;
                        }

                        if (path.neighborTowardsRoot != null) 
                        {
                            neighbor = path.neighborTowardsRoot;
                            SortedList listOfDlcIds = (SortedList) nodeToListOfDlcIds[neighbor];
                            if (listOfDlcIds == null) listOfDlcIds = new SortedList();
                            if (!listOfDlcIds.Contains(s_dlcIdPlusSeqnum)) listOfDlcIds.Add(s_dlcIdPlusSeqnum, null);
                            nodeToListOfDlcIds[neighbor] = listOfDlcIds;
                        }
                    }
            }

            pingChecksums.Clear(); //necessary to flush information for nodes that are no longer neighbors
            
            Log.Write("dbg", LogLvl.Lo, "clearing all checksums");

            foreach (NodeIdAddressBind node in nodeToListOfDlcIds.Keys) 
            {
                // generate string, hash it, put it in the pingChecksum
                string valueToHash = "";
                SortedList dlcIdsPlusSeqnums = (SortedList) nodeToListOfDlcIds[node];      
                if (dlcIdsPlusSeqnums != null)
                    foreach (string s in dlcIdsPlusSeqnums.Keys)
                        valueToHash += "/" + s;
             
                
                    Log.Write("dbg", LogLvl.Lo, "for neighbor node {0}, value being used to " + 
                        "compute checksum is: {1}", node.GetAddress, valueToHash);
                

                byte[] data = new Byte[valueToHash.Length];
                char[] chars = valueToHash.ToCharArray(); 
                for (int i = 0; i < chars.Length; i++)
                    data[i] = (byte) chars[i];
                byte[] hash = md.ComputeHash(data);

                pingChecksums[node.GetAddress] = hash;
            }
        }
        
    }

    /// <summary>
    /// A DlcEntry can contain member state, root state, and delegate state. 
    /// The root state contains whether there is an outstanding create/re-create.
    /// The delegate state can contain several DlcPaths. A DlcPath is a chain between a root node and a non-root 
    /// group member node.
    /// There is only one sequence number -- we always proactively clean up things with out-of-date 
    /// sequence numbers.
    /// </summary>
    public class DlcEntry
    {
        // member data
        public DlcRootState        rootState;
        public DlcMemberState      memberState;
        public DlcDelegateState    delegateState;
        public int seqnum;
        public string  dlcId; // same as the key in the dlcEntries hashtable

        public DlcEntry (string dlcId) 
        {
            this.dlcId = dlcId;
            seqnum = 0;
        }
    }

    public class DlcRootState 
    {
        public enum RootStateCondition 
        {
            Creating, Live, Recreating
        };

        public class ConditionOfMember 
        {
            public bool haveReceivedCreateRecreateReply;
            public bool haveReceivedInstallChecking;
            public ConditionOfMember () 
            {
                haveReceivedCreateRecreateReply = false;
                haveReceivedInstallChecking = false;
            }
        };

        // member data
        public Hashtable conditionOfMembers;
        public RootStateCondition rootStateCondition;
        public DLCCContinuation appCreationCallback;

        // these next two must be set manually
        public CreateRecreateTimer createRecreateTimer;
        public InstallCheckingTimer installCheckingTimer;

        public DlcRootState(ArrayList nonRootMembers) 
        {
            Debug.Assert(nonRootMembers.Count >= 1);
            conditionOfMembers = new Hashtable();
            foreach(NodeIdAddressBind n in nonRootMembers)
                conditionOfMembers[n] = new ConditionOfMember();
            rootStateCondition = RootStateCondition.Creating;
        }
        public bool isCreateRecreateComplete() 
        {
            foreach(ConditionOfMember condition in conditionOfMembers.Values) 
            {
                if (condition.haveReceivedCreateRecreateReply == false) return false;
            }
            return true;
        }
    }

    public class DlcMemberState 
    {
        public enum MemberStateCondition 
        {
            Live, Recreating
        };
        public MemberStateCondition memberStateCondition;
        public NodeIdAddressBind root;
        public RecreateTimer recreateTimer;

        public DlcMemberState(NodeIdAddressBind root) 
        {
            memberStateCondition = MemberStateCondition.Live;
            this.root = root;
        }
    }

    public class DlcDelegateState 
    {
        public ArrayList dlcPaths;
        public DlcDelegateState() 
        {
            dlcPaths = new ArrayList();
        }
        /// <summary>
        /// Return the path containing the given neighbor.
        /// </summary>
        /// <param name="neighbor"></param>
        /// <returns></returns>
        public DlcPath PathForNeighbor (NodeIdAddressBind neighbor) 
        {
            foreach (DlcPath path in dlcPaths) 
            {
                if ((path.neighborTowardsMember != null &&
                     path.neighborTowardsMember.GetAddress.ToString() == neighbor.GetAddress.ToString()) ||
                     (path.neighborTowardsRoot != null &&
                     path.neighborTowardsRoot.GetAddress.ToString() == neighbor.GetAddress.ToString())) 
                {
                    return path;
                }
            }
            Debug.Assert(false, "neighbor was not found");
            return null; 
        }
    }




    /// <summary>
    /// A DLCPath is a DLC chain between two group members, one of whom is the root. 
    /// In addition to these two nodes, we keep track of our two 
    /// neighbors (only one if we're a member).
    /// </summary>
    /// 

    public class DlcPath
    {
        public NodeIdAddressBind root; 
        public NodeIdAddressBind member;
        public NodeIdAddressBind neighborTowardsRoot; 
        public NodeIdAddressBind neighborTowardsMember;
        public long creationTime;
        public DlcPathRefreshTimer dlcPathRefreshTimer;
        public DlcEntry containingEntry;

        public DlcPath (NodeIdAddressBind root, NodeIdAddressBind neighborTowardsRoot, 
            long creationTime, NodeIdAddressBind member, 
            NodeIdAddressBind neighborTowardsMember,
            DlcEntry containingEntry) 
        {
            this.root = root;
            this.member = member; 
            this.neighborTowardsRoot = neighborTowardsRoot;
            this.neighborTowardsMember = neighborTowardsMember;
            this.creationTime = creationTime;
            this.containingEntry = containingEntry;
        }
    }
    
    /// <summary>
    /// The DlcIdPlusSeqnum is a convenient bundle for passing around these two pieces of data.
    /// It doesn't require any xml goop to be serializable.
    /// </summary>
    [Serializable]
    public class DlcIdPlusSeqnum 
    {
        public DlcIdPlusSeqnum() {}
        public DlcIdPlusSeqnum(string dlcId, int seqnum) 
        {
            this.dlcId = dlcId;
            this.seqnum = seqnum;
        }
        public string dlcId;
        public int seqnum;
        public override string ToString() 
        {
            return (dlcId + "." + seqnum);
        }
    }

}
