using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using Microsoft.Research.Herald.Logger;
using MessageRouting;
using SkipNet = Microsoft.Research.Herald.SkipNet;
using NodeIdAddressBind = MessageRouting.NodeIdAddressBind;

namespace  Microsoft.Research.Herald.DLC 
{
    public class SerializationConfig
    { 
        public const string DLC_NAMESPACE = "http://research.microsoft.com/sn/Herald/DLC";
        public static Type [] serializationTypes = new Type[] {
                                                                  typeof(DLCCreateGroupRequestMsg),
                                                                  typeof(DLCCreateGroupReplyMsg),
                                                                  typeof(DLCInstallCheckingMsg),
                                                                  typeof(DLCNeedRecreateMsg),
                                                                  typeof(DLCRecreateGroupRequestMsg),
                                                                  typeof(DLCRecreateGroupReplyMsg),
                                                                  typeof(DLCSoftNotificationMsg),
                                                                  typeof(DLCHardNotificationMsg),
                                                                  typeof(DLCIdListRequestMsg),
                                                                  typeof(DLCIdListReplyMsg),
                                                                  typeof(DlcIdPlusSeqnum),
        }; 
    }

    //======================= Message classes for DLC ========================/

    /// <summary>
    /// This message is sent from the root directly to a member as part of DLC group creation.
    /// When it reaches the member, it generates a DLCCreateGroupReplyMsg, which returns directly, and a
    /// DLCInstallChecking message, which installs the checking path.
    /// </summary>
    [Serializable, XmlRootAttribute("DLCCreateGroupRequestMsg",Namespace=SerializationConfig.DLC_NAMESPACE,
                       IsNullable=true)]
    public class DLCCreateGroupRequestMsg : OverlayRequestMsg 
    {
        private NodeIdAddressBind root;
        private NodeIdAddressBind member; // only carrying this information for debugging purposes
        private string dlcId;
 
        public DLCCreateGroupRequestMsg(Node localNode, OverlayReplyCallback cb, 
            NodeIdAddressBind root, NodeIdAddressBind member,
            string dlcId) : base(localNode, cb, 6 * 60 * 1000 /* delayMs */, false) 
        {
            Debug.Assert(localNode.GetNodeIdAddressBind == root);
            this.root = root;
            this.member = member;
            this.dlcId = dlcId;
            Log.Write("dbg", LogLvl.Hi, "Group create on id {0} headed from {1} to {2}", 
                dlcId, root.GetAddress.ToString(), member.GetAddress.ToString());
        }

        public override void ArrivedAtDestination(MessageRouting.Node localNode)
        {
            // 1. Install member state.
            // 2. Send a DLCCreateGroupReplyMsg back to the root directly.
            // 3. Route a DLCInstallCheckingMsg back to the root through the overlay.
            Log.Write("dbg", LogLvl.Hi, "Group create on id {0} has arrived", dlcId);

            Debug.Assert(localNode.GetNodeIdAddressBind == member);
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            if (dlcApp.dlcEntryTable[dlcId] == null) dlcApp.dlcEntryTable[dlcId] = new DlcEntry(dlcId);
            DlcEntry entry = dlcApp.dlcEntryTable[dlcId];
            Debug.Assert(entry.memberState == null,
                "received create for group that already exists as member");
            Debug.Assert(entry.rootState == null,
                "received create for group that exists as root");
            entry.memberState = new DlcMemberState(root); 
        
            DLCCreateGroupReplyMsg creationReply = new DLCCreateGroupReplyMsg(this);
            localNode.SendMessageDirect(creationReply, root);
 
            // The DLCInstallCheckingMsg sets the appropriate timers for liveness checking.
            DLCInstallCheckingMsg installChecking = new DLCInstallCheckingMsg(this);
            localNode.RouteMessage(installChecking, root.GetNodeId); 
        }

        // xml serialization goop
        public DLCCreateGroupRequestMsg() {} 

        [XmlElementAttribute("Root",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Root 
        {
            set { root = value; }
            get { return root; }
        }
        [XmlElementAttribute("Member",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Member 
        {
            set { member = value; }
            get { return member; }
        }
        public string DlcId
        {
            set { dlcId = value; }
            get { return dlcId; }
        }
    }

    /// <summary>
    /// This message is sent directly from a member to the root of the group, confirming the group creation
    /// along this branch of the liveness checking tree.
    /// </summary>
    [Serializable, XmlRootAttribute("DLCCreateGroupReplyMsg",Namespace=SerializationConfig.DLC_NAMESPACE,
                       IsNullable=true)]
    public class DLCCreateGroupReplyMsg : OverlayReplyMsg 
    {
        // only carrying the root's name for debugging purposes
        private NodeIdAddressBind root; // destination of this message 
        private NodeIdAddressBind member; // originator of this message
        private string dlcId;

        public DLCCreateGroupReplyMsg(DLCCreateGroupRequestMsg request) : base(request)
        {
            // copy in state from the request message we are responding to
            root = request.Root;
            member = request.Member;
            dlcId = request.DlcId;
            Log.Write("dbg", LogLvl.Hi, "Group createReply on id {0} headed from {1} to {2}", 
                dlcId, member.GetAddress.ToString(), root.GetAddress.ToString());

        }
 
        // We would allow the reply callback to occur immediately, except we're adding some debugging information.
        public override void ArrivedAtDestination(MessageRouting.Node localNode) 
        {
            Log.Write("dbg", LogLvl.Hi, "Group createReply on id {0} has arrived from {1}", 
                dlcId, member.GetAddress.ToString());
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            ArrayList bagOfIds = (ArrayList) dlcApp.dlccids_to_nodeIDAddressBindsOfGroupMembersPlusDelegates[dlcId];
            if (bagOfIds == null) bagOfIds = new ArrayList();
            bagOfIds.Add(member);
            dlcApp.dlccids_to_nodeIDAddressBindsOfGroupMembersPlusDelegates[dlcId] = bagOfIds; // done
            base.ArrivedAtDestination(localNode);
        }

        // xml serialization goop

        public DLCCreateGroupReplyMsg() {} 

        [XmlElementAttribute("Root",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Root 
        {
            set { root = value; }
            get { return root; }
        }
        [XmlElementAttribute("Member",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Member 
        {
            set { member = value; }
            get { return member; }
        }
        public string DlcId
        {
            set { dlcId = value; }
            get { return dlcId; }
        }
    }

    /// <summary>
    /// This message is routed through the overlay from the member to the root. On every node it encounters
    /// (including the member and root), it installs checking information in both directions.
    /// If the message at any point finds that a higher sequence number exists, it
    /// sends a SoftNotificationMsg back the way it came, and then dies.
    /// If the message finds a lower sequence number, it initiates pro-active cleanup 
    /// (except along the path that it expects to coninue travelling).
    /// </summary>
    [Serializable, XmlRootAttribute("DLCInstallCheckingMsg",Namespace=SerializationConfig.DLC_NAMESPACE,
                       IsNullable=true)]
    public class DLCInstallCheckingMsg : OverlayMsg {
    
        private NodeIdAddressBind root; // destination of this message
        private NodeIdAddressBind member; // originator of this message
        private NodeIdAddressBind previousHop; 
        private DlcIdPlusSeqnum dlcIdPlusSeqnum;

        // Record this informaiton so we can easily do statistics at 
        // the group creator on how many delegates were involved.
        private ArrayList nodesVisitedByThisMessage;

        public DLCInstallCheckingMsg(DLCCreateGroupRequestMsg request)
        {
            // copy in state from the request message we are responding to
            root = request.Root;
            member = request.Member;
            previousHop = null;
            dlcIdPlusSeqnum = new DlcIdPlusSeqnum(request.DlcId, 0);
            nodesVisitedByThisMessage = new ArrayList();
        }
 
        public DLCInstallCheckingMsg(DLCRecreateGroupRequestMsg request)
        {
            // copy in state from the request message we are responding to
            root = request.Root;
            member = request.Member;
            previousHop = null;
            dlcIdPlusSeqnum = request.DlcIdPlusSeqnum;
            nodesVisitedByThisMessage = new ArrayList();
        }

        /// <summary>
        /// We rely on the SkipNet layer exporting the GetNextHop() function in order to 
        /// know the next node we will visit. It is acceptable for this information to sometimes be wrong --
        /// it will be caught by the ping checks, and trigger a soft notification.
        /// We considered implementing pro-active checking of the previousHop/nextHop 
        /// relationship, but decided against it because (1) it adds additional complexity
        /// and (2) the benefit is not clear.
        /// </summary>
        /// <param name="localNode"></param>
        /// <returns></returns>
        public override bool CheckContinueRouting(MessageRouting.Node localNode) 
        {
            nodesVisitedByThisMessage.Add(localNode.GetNodeIdAddressBind);
            string dlcId = dlcIdPlusSeqnum.dlcId;
            Log.Write("dbg", LogLvl.Hi, "InstallChecking on dlcId {0} with seqnum {1}", dlcId, dlcIdPlusSeqnum.seqnum);
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            if (dlcApp.dlcEntryTable[dlcId] == null) dlcApp.dlcEntryTable[dlcId] = new DlcEntry(dlcId);
            DlcEntry entry = dlcApp.dlcEntryTable[dlcId];
            if (entry.seqnum > this.dlcIdPlusSeqnum.seqnum) 
            {   // If a higher seqnum already exists on this node, (1) send a soft notification 
                // back the way we came and (2) drop this message.
                Log.Write("dbg", LogLvl.Hi, "SoftNotificationMsg create on dlcId {0} b/c install checking seqnum too low",
                    dlcIdPlusSeqnum.ToString());
            
                DLCSoftNotificationMsg msg = new DLCSoftNotificationMsg(dlcIdPlusSeqnum, 
                    localNode.GetNodeIdAddressBind);
                if(previousHop != null) localNode.SendMessageDirect(msg, previousHop);
                return false;
            } 
            
            NodeIdAddressBind nextHop = null;
            if (localNode.GetNodeIdAddressBind.GetAddress.ToString() != root.GetAddress.ToString()) 
            {   // Fill in nextHop -- HACK -- code taken from SkipNet routing implementation
                MessageRouting.NodeId dest = this.root.GetNodeId;
                SkipNet.CLBId destId;
                if( dest is SkipNet.CLBId ) 
                {
                    destId = (SkipNet.CLBId) dest;
                } 
                else 
                {
                    destId = new SkipNet.CLBId( dest.ToString() + "!nodeid" );
                }
                SkipNet.RouteByCLBMsg snMsg = new SkipNet.RouteByCLBMsg((SkipNet.NodeIdAddressBind)
                    localNode.GetNodeIdAddressBind, destId, this);
                snMsg.forceRTable = true;
                nextHop = snMsg.GetNextHop((SkipNet.Node) localNode);
            }

            // If either the previous hop or the next hop points at the current node, 
            // (1) send a soft notification back the way we came and (2) drop this message.
            // Note that this is allowed to happen because the routing layer is allowed to be
            // 'out-of-whack' while routing updates occur.

            if (nextHop != null) 
                Log.Write("dbg", LogLvl.Hi, "InstallChecking on dlcId {0} has nextHop {1}",
                dlcIdPlusSeqnum.ToString(), nextHop.GetAddress.ToString());
            if (previousHop != null) 
                Log.Write("dbg", LogLvl.Hi, "InstallChecking on dlcId {0} has prevHop {1}",
                    dlcIdPlusSeqnum.ToString(), previousHop.GetAddress.ToString());

            if ((nextHop != null && nextHop == localNode.GetNodeIdAddressBind) ||
                (previousHop != null && previousHop == localNode.GetNodeIdAddressBind)) 
            {
                Log.Write("dbg", LogLvl.Hi, "SoftNotificationMsg create on dlcId {0} b/c of GetNextHop failure",
                    dlcIdPlusSeqnum.ToString());
            
                DLCSoftNotificationMsg msg = new DLCSoftNotificationMsg(dlcIdPlusSeqnum, 
                    localNode.GetNodeIdAddressBind);
                if(previousHop != null) localNode.SendMessageDirect(msg, previousHop);
                return false;
            }


            if (entry.delegateState == null) entry.delegateState = new DlcDelegateState();
            DlcDelegateState delegateState = entry.delegateState;



            // If some dlcPaths with a lower sequence number exist, clean them up
            // (unless a path corresponds to the node we're headed towards).
            if (entry.seqnum < dlcIdPlusSeqnum.seqnum) 
            {   
                Hashtable livenessCheckingTreeNeighbors = new Hashtable();
                foreach (DlcPath path in delegateState.dlcPaths) 
                {
                    if (path.neighborTowardsMember != null) 
                        livenessCheckingTreeNeighbors[path.neighborTowardsMember] = null;
                    if (path.neighborTowardsRoot != null) 
                        livenessCheckingTreeNeighbors[path.neighborTowardsRoot] = null;
                    if (path.dlcPathRefreshTimer != null) 
                        path.dlcPathRefreshTimer.Remove(localNode);
                    
                }
                delegateState.dlcPaths.Clear();
                foreach (NodeIdAddressBind neighbor in livenessCheckingTreeNeighbors.Keys) 
                {
                    if (neighbor != nextHop) 
                    {
                        Log.Write("dbg", LogLvl.Hi, "SoftNotificationMsg create on dlcId {0} destined to {1} b/c GetNextHop does not match actual enxt hop",
                            dlcIdPlusSeqnum.ToString(), neighbor.GetAddress.ToString());
            
                        DLCSoftNotificationMsg msg = new DLCSoftNotificationMsg(new 
                            DlcIdPlusSeqnum(dlcId, dlcIdPlusSeqnum.seqnum -1), localNode.GetNodeIdAddressBind);
                        localNode.SendMessageDirect(msg, neighbor);
                    }
                }
                entry.seqnum = dlcIdPlusSeqnum.seqnum;
            }
            // Because we are not doing this through the SoftNotificationUpcall, we may have allowed 
            // some member state to have become associated with a higher sequence number than when 
            // it was installed. This is okay, because the chain we are installing will notify this member if
            // there is a loss of connectivity to the root. The logic for DLC group re-create messages,
            // should take in to account that this situation can occur.

            // Create and install the path that we're going to add, then update the ping checksums.
            DlcPath newPath = new DlcPath(root, nextHop, dlcApp.localEvent.GetTime(), 
                                       member, previousHop, entry);
            DlcPathRefreshTimer t = new DlcPathRefreshTimer(80*1000, new DlcIdPlusSeqnum(entry.dlcId, entry.seqnum), localNode, newPath);
            newPath.dlcPathRefreshTimer = t; // This should be redundant to the work 
            // done by the constructor, but I'm having trouble figuring out where the bug is
            delegateState.dlcPaths.Add(newPath);
            dlcApp.dlcEntryTable.Refresh();
            Log.Write("dbg", LogLvl.Med, "Install checking: timer exists {0}, number of distinct paths = {1}", 
                newPath.dlcPathRefreshTimer == null ? false : true, delegateState.dlcPaths.Count);
            previousHop = localNode.GetNodeIdAddressBind;
            return true;
        }
 
        public override void ArrivedAtDestination(MessageRouting.Node localNode)
        {   
            // If we haven't made it back to root, this install checking message has failed; 
            // trigger a soft notification upcall for this dlcId+seqnum.
            // If we have made it back to the root, we need to additionally update the 
            // root state to record that another timer (installed by this message's 
            // CheckContinueRouting function) is now monitoring this path.
            // We also need to cancel that other timer.
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            if (localNode.GetNodeIdAddressBind.GetAddress.ToString() != root.GetAddress.ToString()) 
            {
                Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c of root mismatch on dlcId {0}. Local is {1}, msg requests {2}", 
                    dlcIdPlusSeqnum.dlcId, localNode.GetNodeIdAddressBind.GetAddress, root.GetAddress);
                Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c of root mismatch on dlcId {0}. Local is {1}, msg requests {2}", 
                    dlcIdPlusSeqnum.dlcId, localNode.GetNodeIdAddressBind.GetNodeId, root.GetNodeId);
                dlcApp.SoftNotificationUpcall(dlcIdPlusSeqnum, null);
                return;
            } 
            else 
            {
                // The information about all the delegates we've seen is recorded so that we can do 
                // logging (for example, knowing the number of delegates in a given group).
                ArrayList bagOfIds = (ArrayList) dlcApp.dlccids_to_nodeIDAddressBindsOfGroupMembersPlusDelegates[dlcIdPlusSeqnum.dlcId];
                if (bagOfIds == null) bagOfIds = new ArrayList();
                bagOfIds.AddRange(nodesVisitedByThisMessage);
                dlcApp.dlccids_to_nodeIDAddressBindsOfGroupMembersPlusDelegates[dlcIdPlusSeqnum.dlcId] = bagOfIds; // done

                DlcEntry entry = dlcApp.dlcEntryTable[dlcIdPlusSeqnum.dlcId];
                Debug.Assert(entry != null); // Can't be null b/c we just installed a path in CheckContinueRouting.
                if (entry.rootState == null) 
                {   // this means the group has been notified -- bail out
                    Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c of earliier hard notification on dlcId {0}",
                        dlcIdPlusSeqnum.dlcId);
                    dlcApp.SoftNotificationUpcall(dlcIdPlusSeqnum, null);
                    return;
                }
                DlcRootState.ConditionOfMember condition = (DlcRootState.ConditionOfMember) 
                    entry.rootState.conditionOfMembers[member];
                condition.haveReceivedInstallChecking = true;
                
                bool installCheckingCompleted = true; // check if all install checking msg's arrived
                foreach (DlcRootState.ConditionOfMember cond in 
                    entry.rootState.conditionOfMembers.Values) 
                    if (!cond.haveReceivedInstallChecking) installCheckingCompleted = false;
                if (installCheckingCompleted && entry.rootState.installCheckingTimer != null) 
                {
                    localNode.CancelTimer(entry.rootState.installCheckingTimer);
                    entry.rootState.installCheckingTimer = null;
                }
            }
        }            
        // xml serialization goop

        public DLCInstallCheckingMsg() {} 

        [XmlElementAttribute("Root",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Root 
        {
            set { root = value; }
            get { return root; }
        }
        [XmlElementAttribute("Member",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Member 
        {
            set { member = value; }
            get { return member; }
        }
        [XmlElementAttribute("PreviousHop",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind PreviousHop 
        {
            set { previousHop = value; }
            get { return previousHop; }
        }
        [XmlElementAttribute("DlcIdPlusSeqnum",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public DlcIdPlusSeqnum DlcIdPlusSeqnum 
        {
            set { dlcIdPlusSeqnum = value; }
            get { return dlcIdPlusSeqnum; }
        }
        [XmlArray( ElementName="NodesVisitedByThisMessage",
             Namespace=SerializationConfig.DLC_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="NodeIdAddressBind", Type=typeof(NodeIdAddressBind), 
             IsNullable=true, Namespace=SerializationConfig.DLC_NAMESPACE)]
        public ArrayList NodesVisitedByThisMessage         
        {
            set { nodesVisitedByThisMessage = value; }
            get { return nodesVisitedByThisMessage; } 
        }   
    }

    /// <summary>
    /// Much like a SoftNotification, except that it's sent directly from a member to the root.
    /// </summary>
    [Serializable, XmlRootAttribute("DLCNeedRecreateMsg",Namespace=SerializationConfig.DLC_NAMESPACE,
        IsNullable=true)]
    public class DLCNeedRecreateMsg : OverlayMsg {

        // the root is only recorded for debugging purposes
        private NodeIdAddressBind root; // destination of this message
        private NodeIdAddressBind member; // originator of this message
        private DlcIdPlusSeqnum dlcIdPlusSeqnum;

        public DLCNeedRecreateMsg(NodeIdAddressBind member, NodeIdAddressBind root, 
            DlcIdPlusSeqnum dlcIdPlusSeqnum) 
        {
            this.member = member;
            this.root = root;
            this.dlcIdPlusSeqnum = dlcIdPlusSeqnum;
            Log.Write("dbg", LogLvl.Hi, "NeedRecreateMessage on id {0} headed from member {1} to root {2}", 
                dlcIdPlusSeqnum.dlcId, member.GetAddress.ToString(), root.GetAddress.ToString());

        }

        // If the seqnum we have is less than the one currently on the node, toss this message.
        // Otherwise signal that we need to re-create this group, starting with a soft notification.
        public override void ArrivedAtDestination(MessageRouting.Node localNode) 
        {   Debug.Assert(root == localNode.GetNodeIdAddressBind);
            Log.Write("dbg", LogLvl.Hi, "NeedRecreateMessage on id {0} has arrived from member {1}", 
                dlcIdPlusSeqnum.dlcId, member.GetAddress.ToString());
            string dlcId = dlcIdPlusSeqnum.dlcId;
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            DlcEntry entry = dlcApp.dlcEntryTable[dlcId];
            
            if (entry == null) return; // If entry is null, the state was already garbage-collected, 
            // and a hard notification was already sent to the member.
            // We choose not to re-send it (choosing bandwidth conservation over redundant messages; 
            // the effect of the hard notification will be felt through a timer even if the earlier 
            // explicit notification message doesn't get through).

            Debug.Assert(entry.rootState != null); // If the entry is here, it better be the root.

            if (entry.seqnum > dlcIdPlusSeqnum.seqnum) return; // Return if the recreate message is out of date.  
            Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c of needRecreateMessage on dlcId {0}",
                dlcIdPlusSeqnum.dlcId);
            dlcApp.SoftNotificationUpcall(dlcIdPlusSeqnum, member); // This starts the re-create process.
        }
        public DLCNeedRecreateMsg() {} 

        [XmlElementAttribute("Root",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Root 
        {
            set { root = value; }
            get { return root; }
        }
        [XmlElementAttribute("Member",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Member 
        {
            set { member = value; }
            get { return member; }
        }
        [XmlElementAttribute("DlcIdPlusSeqnum",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public DlcIdPlusSeqnum DlcIdPlusSeqnum 
        {
            set { dlcIdPlusSeqnum = value; }
            get { return dlcIdPlusSeqnum; }
        }

    }

    /// <summary>
    /// This message is sent from the root directly to a member during DLC group re-creation.
    /// When it reaches the member, it generates a DLCRecreateGroupReplyMsg, which returns directly, and a
    /// DLCInstallChecking message, which installs the checking path.
    /// </summary>
    [Serializable, XmlRootAttribute("DLCRecreateGroupRequestMsg",Namespace=SerializationConfig.DLC_NAMESPACE,
                       IsNullable=true)]
    public class DLCRecreateGroupRequestMsg : OverlayRequestMsg 
    {
        private NodeIdAddressBind root;
        private NodeIdAddressBind member; // recorded for debugging purposes
        private DlcIdPlusSeqnum dlcIdPlusSeqnum;
        
        public DLCRecreateGroupRequestMsg(Node localNode, OverlayReplyCallback cb, 
            NodeIdAddressBind root, NodeIdAddressBind member,
            DlcIdPlusSeqnum dlcIdPlusSeqnum) : base(localNode, cb, 6 * 60 * 1000 /* delayMs */, false) 
        {
            Debug.Assert(localNode.GetNodeIdAddressBind == root);
            this.root = root;
            this.member = member;
            this.dlcIdPlusSeqnum = dlcIdPlusSeqnum;
            Log.Write("dbg", LogLvl.Hi, "RecreateRequestMessage on id {0} headed from root {1} to member {2}", 
                dlcIdPlusSeqnum.dlcId, root.GetAddress.ToString(), member.GetAddress.ToString());

        }

        public override void ArrivedAtDestination(MessageRouting.Node localNode)
        {
            Log.Write("dbg", LogLvl.Hi, "RecreateRequestMessage on id {0} arrived", 
                dlcIdPlusSeqnum.dlcId);

            // 1. Update member state.
            // 2. Send a DLCRecreateGroupReplyMsg back to the root directly.
            // 3. Route a DLCInstallCheckingMsg back to the root through the overlay.

            Debug.Assert(localNode.GetNodeIdAddressBind == member);
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            DlcEntry entry = dlcApp.dlcEntryTable[dlcIdPlusSeqnum.dlcId];

            if (entry == null || entry.memberState == null) 
            {   // The DLC client app was already signalled here; send a HardNotification to the root.
                DLCHardNotificationMsg msg = new DLCHardNotificationMsg(dlcIdPlusSeqnum.dlcId, member);
                localNode.SendMessageDirect(msg, root);
                // Do a SoftNotificationUpcall in case there's delegate state to clean up.
                Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c of recreateGroupRequestMessage bailing out on dlcId {0}",
                    dlcIdPlusSeqnum.dlcId);
                if (entry != null) dlcApp.SoftNotificationUpcall(new 
                                          DlcIdPlusSeqnum(dlcIdPlusSeqnum.dlcId, DLCC.DLC_INFINITY), member); 
            } 
            else 
            {   
                // Check that the sequence number implies that we should act on this request.
                if (dlcIdPlusSeqnum.seqnum >= entry.seqnum) 
                {
                    entry.seqnum = dlcIdPlusSeqnum.seqnum;

                    // Update the member state appropriately.
                    DlcMemberState memberState = entry.memberState;
                    if (memberState.memberStateCondition == DlcMemberState.MemberStateCondition.Recreating) 
                    {
                        memberState.memberStateCondition = DlcMemberState.MemberStateCondition.Live; 
                        localNode.CancelTimer(memberState.recreateTimer); // take recreate timer off local event queue
                        memberState.recreateTimer = null; // remove all references to the timer
                    }
                    // Send out the messages.
                    DLCRecreateGroupReplyMsg creationReply = new DLCRecreateGroupReplyMsg(this);
                    localNode.SendMessageDirect(creationReply, root);
                    // The InstallChecking message will clean up any chains it is not replacing.
                    DLCInstallCheckingMsg installChecking = new DLCInstallCheckingMsg(this);
                    localNode.RouteMessage(installChecking, root.GetNodeId);
                }
            }
        }

        // xml serialization goop
        public DLCRecreateGroupRequestMsg() {} 

        [XmlElementAttribute("Root",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Root 
        {
            set { root = value; }
            get { return root; }
        }
        [XmlElementAttribute("Member",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Member 
        {
            set { member = value; }
            get { return member; }
        }
        [XmlElementAttribute("DlcIdPlusSeqnum",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public DlcIdPlusSeqnum DlcIdPlusSeqnum 
        {
            set { dlcIdPlusSeqnum = value; }
            get { return dlcIdPlusSeqnum; }
        }
    }


    [Serializable, XmlRootAttribute("DLCRecreateGroupReplyMsg",Namespace=SerializationConfig.DLC_NAMESPACE,
                       IsNullable=true)]
    public class DLCRecreateGroupReplyMsg : OverlayReplyMsg 
    {
        // root is only here for debugging purposes
        private NodeIdAddressBind root; // destination of this message
        private NodeIdAddressBind member; // originator of this message
        private DlcIdPlusSeqnum dlcIdPlusSeqnum;

        public DLCRecreateGroupReplyMsg(DLCRecreateGroupRequestMsg request) : base(request)
        {
            // copy in state from the request message we are responding to
            root = request.Root;
            member = request.Member;
            dlcIdPlusSeqnum = request.DlcIdPlusSeqnum;
            Log.Write("dbg", LogLvl.Hi, "RecreateReplyMessage on id {0} headed from member {1} to root {2}", 
                dlcIdPlusSeqnum.dlcId, member.GetAddress.ToString(), root.GetAddress.ToString());


        }

        public override void ArrivedAtDestination(MessageRouting.Node localNode) 
        {
            Log.Write("dbg", LogLvl.Hi, "RecreateReplyMessage on id {0} arrived from member {1}", 
                dlcIdPlusSeqnum.dlcId, member.GetAddress.ToString());
            base.ArrivedAtDestination(localNode);
        }



        // Request-Reply matching supersedes the ArrivedAtDestination call.

        // xml serialization goop

        public DLCRecreateGroupReplyMsg() {} 

        [XmlElementAttribute("Root",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Root 
        {
            set { root = value; }
            get { return root; }
        }
        [XmlElementAttribute("Member",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Member 
        {
            set { member = value; }
            get { return member; }
        }
        [XmlElementAttribute("DlcIdPlusSeqnum",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public DlcIdPlusSeqnum DlcIdPlusSeqnum 
        {
            set { dlcIdPlusSeqnum = value; }
            get { return dlcIdPlusSeqnum; }
        }
    }

    /// <summary>
    /// This message is sent in response to a hash discrepancy; the goal is to agree on which DlcPaths
    /// are live by removing all the ones they disagree about. The hash was piggybacked on a ping.
    /// </summary>
    [Serializable, XmlRootAttribute("DLCIdListRequestMsg",Namespace=SerializationConfig.DLC_NAMESPACE,
                       IsNullable=true)]
    public class DLCIdListRequestMsg  : OverlayMsg 
    {
        private ArrayList dlcIdsPlusSeqnums;
        private NodeIdAddressBind requestor;
        private NodeIdAddressBind neighbor;
                    
        /// <summary>
        /// The source of this request should put in all the dlcId's that the 
        /// source believes it has in common with the neighbor node.
        /// </summary>
        /// <param name="localNode"></param>
        /// <param name="cb"></param>
        /// <param name="dlcIdsPlusSeqnums"></param>
        /// <param name="source"></param>
        /// <param name="neighbor"></param>
        public DLCIdListRequestMsg(ArrayList dlcIdsPlusSeqnums, 
            NodeIdAddressBind requestor, NodeIdAddressBind neighbor)
        {
            Log.Write("dbg", LogLvl.Hi, "IdListRequest to {0}", 
                neighbor.GetAddress);
            foreach (DlcIdPlusSeqnum d in dlcIdsPlusSeqnums) 
            {
                Log.Write("dbg", LogLvl.Hi, "IdListRequest to {0} contains {1}", neighbor.GetAddress, d.ToString());
            }
            this.dlcIdsPlusSeqnums = dlcIdsPlusSeqnums;
            this.requestor = requestor;
            this.neighbor = neighbor;
        }
        
        /// <summary>
        /// 1. For groups that exist on this node but not the other node, we do a soft notification upcall, 
        /// being careful not to send unneccessary messages to the node that just told this node that it 
        /// does not know about the group.
        /// 2. For groups that exist on both nodes, we update their liveness timers just as a ping would.
        /// 3. We reply with a list of all live groups that this node believes the two nodes have in common. 
        /// NOTE:
        /// We skip doing notifications on groups that have just recently been established -- this prevents 
        /// a race condition involving pings and group create.
        /// </summary>
        /// <param name="localNode"></param>
        public override void ArrivedAtDestination(MessageRouting.Node localNode) 
        {
            Log.Write("dbg", LogLvl.Hi, "IdListRequest from {0}", 
                requestor.GetAddress);
            foreach (DlcIdPlusSeqnum d in dlcIdsPlusSeqnums) 
            {
                Log.Write("dbg", LogLvl.Hi, "IdListRequest from {0} contains {1}", requestor.GetAddress, d.ToString());
            }
            Debug.Assert(neighbor == localNode.GetNodeIdAddressBind);
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            Hashtable foreignDlcIdsToSeqnums = new Hashtable();

            foreach(DlcIdPlusSeqnum foreignDlcIdPlusSeqnum in dlcIdsPlusSeqnums) 
            {
                foreignDlcIdsToSeqnums[foreignDlcIdPlusSeqnum.dlcId] = foreignDlcIdPlusSeqnum.seqnum;
            }
            
            ArrayList commonDlcIdsPlusSeqnums = new ArrayList();
            long currentTime = dlcApp.localEvent.GetTime();
            foreach (string dlcId in dlcApp.dlcEntryTable.DlcEntriesWithNeighbor(requestor)) 
            {
                long offset = currentTime - dlcApp.dlcEntryTable[dlcId].delegateState.PathForNeighbor(requestor).creationTime;
                if (offset > DLCC.SETUP_BUFFER_MSECS) 
                {   // avoiding aforementioned race condition
                    if (foreignDlcIdsToSeqnums[dlcId] == null ||
                        ((int) foreignDlcIdsToSeqnums[dlcId]) > dlcApp.dlcEntryTable[dlcId].seqnum) // Case 1.
                    {
                        Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c of id list mismatch on dlcId {0}",
                            dlcId);
                        dlcApp.SoftNotificationUpcall(new DlcIdPlusSeqnum(dlcId, dlcApp.dlcEntryTable[dlcId].seqnum),
                            requestor); 
                    } 
                    else 
                    {
                        if (((int) foreignDlcIdsToSeqnums[dlcId]) == dlcApp.dlcEntryTable[dlcId].seqnum) 
                        {   // Cases 2. and 3.
                            DlcPath path = dlcApp.dlcEntryTable[dlcId].delegateState.PathForNeighbor(requestor);
                            Log.Write("dbg", LogLvl.Hi, "IdListRequestMessage offset {0}", offset);
                            Log.Write("dbg", LogLvl.Hi, "IdListRequestMessage resetting dlcId {0}, timer exists = {1}", dlcId,
                                path.dlcPathRefreshTimer == null ? false : true);
                            path.dlcPathRefreshTimer.Reset(localNode);
                            commonDlcIdsPlusSeqnums.Add(new DlcIdPlusSeqnum(dlcId, 
                                dlcApp.dlcEntryTable[dlcId].seqnum));
                        } 
                    }
                }
            }
            DLCIdListReplyMsg msg = new DLCIdListReplyMsg(this, commonDlcIdsPlusSeqnums);
        }
        
        // xml serialization goop

        public DLCIdListRequestMsg() {} 

        [XmlArray( ElementName="DlcIdsPlusSeqnums",
             Namespace=SerializationConfig.DLC_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="DlcIdPlusSeqnum", Type=typeof(DlcIdPlusSeqnum), 
             IsNullable=true, Namespace=SerializationConfig.DLC_NAMESPACE)]
        public ArrayList DlcIdsPlusSeqnums         
        {
            set { dlcIdsPlusSeqnums = value; }
            get { return dlcIdsPlusSeqnums; } 
        }        
        [XmlElementAttribute("Requestor",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Requestor 
        {
            set { requestor = value; }
            get { return requestor; }
        }
        [XmlElementAttribute("Neighbor",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Neighbor 
        {
            set { neighbor = value; }
            get { return neighbor; }
        }
    }

    /// <summary>
    /// This message returns with a list of id's that the other node believes the two nodes have in common.
    /// </summary>
    public class DLCIdListReplyMsg : OverlayMsg {
        private ArrayList dlcIdsPlusSeqnums;
        private NodeIdAddressBind replier;
    
        /// <summary>
        /// The source of the request should put in all the dlcId's it 
        /// believes it has in common with the foreign node.
        /// </summary>
        /// <param name="request"></param>
        public DLCIdListReplyMsg(DLCIdListRequestMsg request, ArrayList dlcIdsPlusSeqnums)
        {
            this.dlcIdsPlusSeqnums = dlcIdsPlusSeqnums;
            replier = request.Neighbor;
        }
        
        /// <summary>
        /// See comments for DLCIdListRequestMsg -- actions are nearly identical, except that we don't send
        /// a list of groups that we have in common a second time.
        /// </summary>
        /// <param name="localNode"></param>
        public override void ArrivedAtDestination(MessageRouting.Node localNode) 
        {
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            Hashtable foreignDlcIdsToSeqnums = new Hashtable();

            foreach(DlcIdPlusSeqnum foreignDlcIdPlusSeqnum in dlcIdsPlusSeqnums) 
            {
                foreignDlcIdsToSeqnums[foreignDlcIdPlusSeqnum.dlcId] = foreignDlcIdPlusSeqnum.seqnum;
            }
            
            long currentTime = dlcApp.localEvent.GetTime();
            foreach (string dlcId in dlcApp.dlcEntryTable.DlcEntriesWithNeighbor(replier)) 
            {
                if ((currentTime - dlcApp.dlcEntryTable[dlcId].delegateState.PathForNeighbor(replier).creationTime) 
                    > DLCC.SETUP_BUFFER_MSECS) 
                {   // avoiding aforementioned race condition
                    if (foreignDlcIdsToSeqnums[dlcId] == null ||
                        ((int) foreignDlcIdsToSeqnums[dlcId]) > dlcApp.dlcEntryTable[dlcId].seqnum) // Case 1.
                    {
                        Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c of id list mismatch (reply) on dlcId {0}",
                            dlcId);
                        dlcApp.SoftNotificationUpcall(new DlcIdPlusSeqnum(dlcId, (int) foreignDlcIdsToSeqnums[dlcId]),
                            replier); 
                    } 
                    else 
                    {
                        if (((int) foreignDlcIdsToSeqnums[dlcId]) == dlcApp.dlcEntryTable[dlcId].seqnum) 
                        {   // Cases 2. and 3.
                            DlcPath path = dlcApp.dlcEntryTable[dlcId].delegateState.PathForNeighbor(replier);
                            path.dlcPathRefreshTimer.Reset(localNode);
                        } 
                    }
                }
            }
        }

        // xml serialization goop

        public DLCIdListReplyMsg() {} 

        [XmlArray( ElementName="DlcIdsPlusSeqnums",
             Namespace=SerializationConfig.DLC_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="DlcIdPlusSeqnum", Type=typeof(DlcIdPlusSeqnum), 
             IsNullable=true, Namespace=SerializationConfig.DLC_NAMESPACE)]
        public ArrayList DlcIdsPlusSeqnums         
        {
            set { dlcIdsPlusSeqnums = value; }
            get { return dlcIdsPlusSeqnums; } 
        }    
        [XmlElementAttribute("Replier",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Replier 
        {
            set { replier = value; }
            get { return replier; }
        }
    }
  
    /// <summary>
    /// A soft notification message triggers a SoftNotificationUpcall.
    /// The subtlelies of SoftNotification logic is handled in the upcall.
    /// </summary>
    [Serializable, XmlRootAttribute("DLCSoftNotificationMsg",Namespace=SerializationConfig.DLC_NAMESPACE,
                       IsNullable=true)]
    public class DLCSoftNotificationMsg : OverlayMsg 
    {            
        private DlcIdPlusSeqnum dlcIdPlusSeqnum;
        private NodeIdAddressBind source;

        public DLCSoftNotificationMsg(DlcIdPlusSeqnum dlcIdPlusSeqnum, NodeIdAddressBind source) 
        {
            Log.Write("dbg", LogLvl.Hi, "SoftNotificationMsg create on dlcId {0} with source {1}",
                dlcIdPlusSeqnum.ToString(), source.GetAddress.ToString());
            this.dlcIdPlusSeqnum = dlcIdPlusSeqnum;
            this.source = source;
        }

        public override void ArrivedAtDestination(MessageRouting.Node localNode) 
        {
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            Log.Write("dbg", LogLvl.Hi, "SoftNotification b/c of SoftNotificationMsg on dlcId {0}",
                dlcIdPlusSeqnum.ToString());
            dlcApp.SoftNotificationUpcall(dlcIdPlusSeqnum, source);
        }

        // xml serialization goop

        public DLCSoftNotificationMsg() {} 

        [XmlElementAttribute("Source",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Source 
        {
            set { source = value; }
            get { return source; }
        }
        [XmlElementAttribute("DlcIdPlusSeqnum",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public DlcIdPlusSeqnum DlcIdPlusSeqnum 
        {
            set { dlcIdPlusSeqnum = value; }
            get { return dlcIdPlusSeqnum; }
        }
    }

    /// <summary>
    /// A hard notification is always directed towards a member (possibly the root).
    /// It does not contain a seqnum.
    /// As an optimization, we store the originator of the hard notification 
    /// (in order to avoid sending a redundant message back to the originator if it is also a member).
    /// </summary>
    [Serializable, XmlRootAttribute("DLCHardNotificationMsg",Namespace=SerializationConfig.DLC_NAMESPACE,
                       IsNullable=true)]
    public class DLCHardNotificationMsg : OverlayMsg 
    {
        private string dlcId;
        private NodeIdAddressBind source;
        
        public DLCHardNotificationMsg(string dlcId, NodeIdAddressBind source) 
        {
            this.dlcId = dlcId;
            this.source = source;
        }
        public override void ArrivedAtDestination(MessageRouting.Node localNode) 
        {
            Log.Write("dbg", LogLvl.Hi, "Hard notification on id {0} has arrived from {1}", dlcId, source.GetAddress.ToString());
            DLCC dlcApp = (DLCC) localNode.GetApp(typeof(DLCC));
            dlcApp.HardNotificationUpcall(dlcId, source);
        }
        // xml serialization goop

        public DLCHardNotificationMsg() {} 

        [XmlElementAttribute("DlcId",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public string DlcId 
        {
            set { dlcId = value; }
            get { return dlcId; }
        }    
        [XmlElementAttribute("Source",
             Namespace=SerializationConfig.DLC_NAMESPACE)]
        public NodeIdAddressBind Source 
        {
            set { source = value; }
            get { return source; }
        }
    }
}
