using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using LocalEvent = MessageRouting.LocalEvent;
using Node = MessageRouting.Node;

namespace Microsoft.Research.Herald.Logger {

    public enum LogLvl {
        Disabled, Hi, Med, Lo,
    };

    public class Log {

        //**********************************************************************************************
        // Static Members
        //**********************************************************************************************
        static string       identity;
        static Hashtable    enabledLevelTable;
        static LogLvl       defaultLevel;

        static bool         enableTimeStamping;
        static LocalEvent   localEvent; // We get the time from the localEvent.
        static PerformanceCounter CPUCycleCounter;
        static PerformanceCounter NonpagedCounter;
        static PerformanceCounter TCPConnectionCounter;

        // We log the statistics information under this category.
        public const string STATISTICS_CATEGORY = "stat";
 
        /// <summary>
        /// Static constructor
        /// </summary>
        static Log() {
            enabledLevelTable = new Hashtable();
            defaultLevel = LogLvl.Hi;
            enableTimeStamping = false;
            localEvent = null;
            CPUCycleCounter = null;
            NonpagedCounter = new PerformanceCounter("Memory", "Pool Nonpaged Bytes");
            TCPConnectionCounter = new PerformanceCounter("TCP", "Connections Established");
        }

        /// <summary>
        /// Need to pass in LocalEvent to allow access to timing information.
        /// </summary>
        public static void EnableTimeStamping(LocalEvent localEvent) 
        {
            enableTimeStamping = true;
            Log.localEvent = localEvent;
        }


        /// <summary>
        /// Creates and initializes a CPU performance counter.
        /// </summary>
        public static void EnableLoadMonitoring() 
        {
            try // This throws a "Category does not exist" exception only on herald004 -- go figure!!!
            {
                CPUCycleCounter = new PerformanceCounter("Processor", 
                                                         "% Processor Time", "_Total", true);   
            }
            catch (Exception e) 
            {
                Log.Write("dbg", LogLvl.Hi, "caught unexpected exception: {0}", e);
            }
        }

        /// <summary>
        /// Write a message to the log that is output at the given logging level.
        /// The 'category' for this message is as given.
        /// </summary>
        /// <param name="category">The category for this message</param>
        /// <param name="level">The severity level of this message</param>
        /// <param name="formatString">The string format of this message</param>
        /// <param name="args">The parameters for the string formatting</param>
        public static void Write( string category, LogLvl level, string formatString, params object[] args ) {
            Debug.Assert( level!=LogLvl.Disabled );

            // Check if this message should be filtered out
            if( ShouldFilter(category, level) ) {
                return;
            }
            
            // Output the message
            string s = String.Format(formatString, args);
            if( identity!=null ) 
            {
                s = String.Format("Node {0}: {1}", identity, s);
            }
            if (enableTimeStamping)
            {
                s = String.Format("{0}: {1}", Log.GetRealOrVirtualTime(), s);
            } 
            Console.WriteLine("{0}: {1}", category, s);
        }

        /// <summary>
        /// Constructs the time using the (previously assigned) localEvent
        /// and returns a string containing the time in a human-readable format.
        /// </summary>
        public static string GetRealOrVirtualTime() 
        {
            Debug.Assert(localEvent != null);
            // the constant 10000 converts the msecs returned by GetTime() to 
            // the 100-nanosec unit expected by
            // the DateTime constructor
            long time = localEvent.GetTime();
            DateTime dateTime = new DateTime(10000 * time);
            return String.Format("{0} ({1:T}.{2:000})", time, dateTime, dateTime.Millisecond);
        }

        /// <summary>
        /// Log message counts and (if appropriate) cpu load.
        /// </summary>
        public static void logStatistics(Node node)
        {
            logCounters(node);
            if (CPUCycleCounter != null) {
                logTransient();
            }
        }

        /// <summary>
        /// Log transient information (just CPU load right now).
        /// </summary>
        public static void logTransient() 
        {			
            Debug.Assert(CPUCycleCounter != null);
            Write(STATISTICS_CATEGORY, LogLvl.Hi, "cpuLoad={0}", 
                  CPUCycleCounter.NextValue().ToString());
        }

        /// <summary>
        /// Log counters (just message counts right now).
        /// </summary>
        public static void logCounters(Node node) 
        {			
            Debug.Assert(localEvent != null);
            Write(STATISTICS_CATEGORY, LogLvl.Hi, "msgsIn={0}: msgsOut={1}",
                  localEvent.GetNumMessagesReceivedFromWire(node),
                  localEvent.GetNumMessagesSentToWire(node));
            long npc = (long)NonpagedCounter.NextValue();
            Write(STATISTICS_CATEGORY, LogLvl.Hi, "nonpaged pool bytes = {0}",
                  npc);    
            long nconn = (long)TCPConnectionCounter.NextValue();
            Write(STATISTICS_CATEGORY, LogLvl.Hi, "TCP connections established = {0}",
                  nconn);
        }

        /// <summary>
        /// Set the identity of the current node. The simulator can use this to switch
        /// identities when it processes messages on different nodes.
        /// </summary>
        /// <param name="id">Current identity</param>
        public static void SetIdentity( string id ) {
            identity = id;
        }

        /// <summary>
        /// Return the identity of the current node.
        /// </summary>
        /// <returns></returns>
        public static string GetIdentity() {
            return identity;
        }

        /// <summary>
        /// Enable debug output to a file.
        /// </summary>
        /// <param name="filename">The name of the logging file</param>
        public static void SetOutputFile( string filename ) {}

        /// <summary>
        /// Set the default filter level. Messages at this level or above
        /// will be printed by default, unless overridden with SetCategoryFilterLevel().
        /// </summary>
        public static void SetDefaultFilterLevel( LogLvl level ) {
            defaultLevel = level;
        }

        /// <summary>
        /// Enable output for the given category, at the given level and above.
        /// </summary>
        public static void ClearCategoryFilters() {
            enabledLevelTable = new Hashtable();
        }

        /// <summary>
        /// Enable output for the given category, at the given level and above.
        /// </summary>
        public static void SetCategoryFilterLevel( string category, LogLvl level ) {
            enabledLevelTable[category] = level;
        }

        /// <summary>
        /// Determine if a message with the given category and level should be filtered.
        /// </summary>
        /// <returns>True if it should be filtered, False if it should be output</returns>
        private static bool ShouldFilter( string msgCategory, LogLvl msgLevel ) {
            int iMsgLevel = (int) msgLevel;
            int iFilterLevel;

            if( enabledLevelTable.ContainsKey(msgCategory) ) {
                LogLvl filterLevel = (LogLvl) enabledLevelTable[msgCategory];
                iFilterLevel = (int) filterLevel;
            } else {
                iFilterLevel = (int) defaultLevel;
            }

            return (iMsgLevel > iFilterLevel);
        }

        
        //**********************************************************************************************
        // Instance (non-static) Members
        //**********************************************************************************************
        
        private string     category; 

        /// <summary>
        /// Instance constructor
        /// </summary>
        public Log( string category ) {
            this.category = category;
        }

        /// <summary>
        /// Write a message to the log that is output at the given logging level.
        /// The 'category' for this message is determined by the instance.
        /// </summary>
        /// <param name="level">The severity level of this message</param>
        /// <param name="formatString">The string format of this message</param>
        /// <param name="args">The parameters for the string formatting</param>
        public virtual void Write( LogLvl level, string formatString, params object[] args ) {
            Write( category, level, formatString, args );
        }

    }

}
