using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Text;
using System.Xml.Serialization;

using MessageRouting;
using SkipNet=Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Logger;
using Microsoft.Research.Herald.Common;
using Microsoft.Research.Herald.TLog;
using ControlFlow=Microsoft.Research.Herald.Common.ControlFlow;
using DLC=Microsoft.Research.Herald.DLC;

namespace MeasureFUSE
{
    public class App 
    {
 
        // To be initialized by _sim or _cluster
        public static ExecutionEnvironment exEnv;
        public static ApplicationEnvironment appEnv;

        // Debugging controls
        public static bool DebugOutput = false;
        public static bool doLogAppFlow = false;
        public static bool debugDlc = false;

        // Other variables
        static bool quitImmediately = false;
        public static Random random = null;
        public static Hashtable index_to_NodeIdAddressBind = new Hashtable(); // implements a name sevrice on the control node, otherwise empty
        
        // RPC variables
        public static SkipNet.Node myServer = null;
        public static ControlFlow.SimpleControlFlow cf = null;

        // Experiment variables...
        // ...general to all the experiments
        public static int FUSEgroupSize = 2;
        public static bool doStatLogging = false;
        public static bool doNonLoadStatLoggingOnly = false;
        public static int periodOfPinging = 60 * 1000; // default is to ping every 60 secs
        public static int periodOfStatLogging = 5 * 1000; // default is to stat log every 5 secs
        
        //...specific to SteadyStateTest()
        public static bool doSteadyStateTest = false;
        public static int msecsToSpendInSteadyState = 10*60*1000; // 10 minutes
        
        //...specific to CreateOneAndFailHard()
        public static bool doCreateOneAndFailHardTest = false;
        // the intent of this offset is to give the person manually removing the ethernet cable some 
        // warning before the cable should actually be removed
        public static int msecsFromGlobalStartUntilHardFailureSignal = 1*10*1000; // 10 seconds

        // ...specific to SerialCreationAndNotificationTest()
        public static bool doSerialCreationAndNotificationTest = false;
        public static int totalNumSerialTests = 10;
        
        // ...specific to Test_1()
        public static bool doTest_1_Cluster = false;
        public static bool doTest_1_Sim = false;
        public static bool doTest_1_Debug = false;

        public static bool doSilentStatLogging = false;

        // ...specific to Test_2_Sim()
        public static bool doTest_2_Sim = false;
        public static bool doTest_2_Cluster = false;

        // ...specific to Test_3()
        public static bool doTest_3 = false;

        public static bool doImmediateLogging = false;
        public static bool doImmediateLoggingAndLogRTable = false;
        public static bool doLogRTable = false;

        // ...specific to ParallelCreationAndDestructionTest()
        public static bool doParallelCreationAndDestructionTest = false;
        public static int FUSEgroupsPerNode = 5;

        // ...specific to ParallelCreationAndDestructionTest()
        public static bool doTestAutoRepair = false;

        // This offset allows us to finish sending 'go' messages before starting the experiment.
        public static long msecsOffsetFromGlobalStartToGroupCreation = 10*60*1000; // 10 minutes
        // This offset is how long to spend waiting around after creating the groups
        public static long msecsToSpendQuiesced = 10*60*1000; // 10 minutes
        // This offset has a self-explanatory name
        public static long msecsToAllowForFailureNotificationsToPropagate = 2*60*1000; // 2 minutes



        //================================ Code for application program steps ================================

        //
        // Run the application.
        //
        public static void Application() 
        {
            Console.WriteLine("pingEnabled: {0}", SkipNet.Parameters.pingEnabled); // xxx mmt
            Log.Write("flow", LogLvl.Hi, "Pinging nodes for liveness and having joined overlay...");
            // Ping all nodes to confirm they are ready.
            PingNodesTillTheyReply();
            Log.Write("flow", LogLvl.Hi, ".. everyone has been pinged.");

            TLog.InitGlobalLoggingNode(App.myServer, App.exEnv);

            Log.Write("flow", LogLvl.Hi, "MeasureFUSEInitialization: Sending out list of node names and ids...");
            // MeasureFUSEInitialization: send out to every node the list of all other 
            // nodes in the experiment, and kicks off other events not specific to 
            // individual experiments.
            // Specifying doImmediateLogging causes the initialization to be done at boot.
            if (!doImmediateLogging) MeasureFUSEInitialization(); 
            Log.Write("flow", LogLvl.Hi, ".. finished sending out list to everyone.");

            // Experiments:
            if (doSteadyStateTest) { DoSteadyStateTest(); }
            if (doSerialCreationAndNotificationTest) { DoSerialCreationAndNotificationTest(); }
            if (doTest_1_Sim) { DoTest_1_Sim(); }
            if (doTest_1_Cluster) { DoTest_1_Cluster(); }
            if (doTest_1_Debug) { DoTest_1_Debug(); }
            if (doTest_2_Sim) { DoTest_2_Sim(30*60*1000, 5); }
            if (doTest_2_Cluster) { DoTest_2_Cluster(5); }
            if (doTest_3) { DoTest_3(); }
            if (doCreateOneAndFailHardTest) { DoCreateOneAndFailHardTest(); }
            if (doParallelCreationAndDestructionTest) { DoParallelCreationAndDestructionTest(); }
            if (doTestAutoRepair) {DoTestAutoRepair(); }

            // We're done.
            Log.Write("flow", LogLvl.Hi, "Starting 'TERMINATE' phase from master node");
            TerminateNodes();
            Log.Write("flow", LogLvl.Hi, "Finishing 'TERMINATE' phase from master node");

        }

        public static void DoSteadyStateTest()  
        {
            Log.Write("flow", LogLvl.Hi, "Beginning SteadyStateTest -- sleeping for {0} seconds",
                msecsToSpendInSteadyState / 1000);   
            exEnv.GoFor(msecsToSpendInSteadyState);
            Log.Write("flow", LogLvl.Hi, "Finished SteadyStateTest");        
        }

        public static void DoTestAutoRepair()  
        {
            Debug.Assert(4 <= App.exEnv.NodeCount, 
                "this experiment requires at least four nodes"); 
            Log.Write("flow", LogLvl.Hi, "Beginning to Test Auto Repair" +
                "from master node -- nothing to do");

            exEnv.GoFor(20*1000);

            NodeIdAddressBind [] FUSEgroupMembers = new NodeIdAddressBind [2];
            FUSEgroupMembers[0] = (SkipNet.NodeIdAddressBind) index_to_NodeIdAddressBind[0];
            FUSEgroupMembers[1] = (SkipNet.NodeIdAddressBind) index_to_NodeIdAddressBind[4];

            long cgStart = appEnv.envGetTime();
            MeasureGroupCreateControlMsg msg = new MeasureGroupCreateControlMsg(myServer, cf, FUSEgroupMembers);
            CreateGroup_ControlReplyMsg rMsg = (CreateGroup_ControlReplyMsg) 
                cf.SendRequestMsg(FUSEgroupMembers[0], msg);
            Log.Write("dbg", LogLvl.Hi, "FUSE group creation operation took {0} seconds",
                (appEnv.envGetTime() - cgStart) / 1000.0);
            if (rMsg == null) 
            {
                Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                    "establishment to address {0} never acked. ***",
                    FUSEgroupMembers[0].GetAddress.ToString());
            }
            else 
            {
                if (rMsg.dlccid == null) 
                {
                    Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                        "establishment to address {0} acked, but creation failed. ***",
                        FUSEgroupMembers[0].GetAddress.ToString());
                } 
                else 
                {
                    Log.Write("dbg", LogLvl.Hi, "Message requesting serial fuse group " +
                        "establishment to address {0} acked with dlccid {1}.",
                        FUSEgroupMembers[0].GetAddress.ToString(), rMsg.dlccid);
                }
            }

            Log.Write("flow", LogLvl.Hi, "pausing indefinitely now that we've constructed a dlc group");  
            int bigNumMsecs = 30*1000;
            while (true) 
            {
                if (App.DebugOutput) myServer.DumpRoutingTable();
                exEnv.GoFor(bigNumMsecs);
            }
        }

        public static void DoSerialCreationAndNotificationTest() 
        {
            for (int i = 0; i < totalNumSerialTests; i++) 
            {
                //generate the nodes to ask to participate in this fuse group
                NodeIdAddressBind [] FUSEgroupMembers = new NodeIdAddressBind [FUSEgroupSize];
                Debug.Assert(FUSEgroupSize <= App.exEnv.NodeCount, 
                    "too few nodes in overlay for FEASIBLY creating FUSE groups of this size"); 
                string listOfFUSEgroupMembersForDebugging = "";
                for (int j = 0; j < FUSEgroupSize; j++) 
                {
                    int nextIndex = App.random.Next(App.exEnv.NodeCount);
                    // check for uniqueness
                    NodeIdAddressBind nextAddr = (SkipNet.NodeIdAddressBind) index_to_NodeIdAddressBind[nextIndex];
                    bool unique = true;
                    for (int k = 0; k < j; k++) 
                    {
                        if (FUSEgroupMembers[k].GetAddress == nextAddr.GetAddress) unique = false;
                    }
                    if (unique) 
                    {
                        FUSEgroupMembers[j] = nextAddr;
                        listOfFUSEgroupMembersForDebugging += (" " + nextAddr.GetAddress.ToString());
                    } 
                    else 
                    {
                        j--;
                    }
                }
                Log.Write("dbg", LogLvl.Hi, "Creating FUSE group {0} with membership {1}", 
                    i, listOfFUSEgroupMembersForDebugging);
 
                // create the fuse group
                long cgStart = appEnv.envGetTime();
                MeasureGroupCreateControlMsg msg = new MeasureGroupCreateControlMsg(myServer, cf, FUSEgroupMembers);
                CreateGroup_ControlReplyMsg rMsg = (CreateGroup_ControlReplyMsg) 
                    cf.SendRequestMsg(FUSEgroupMembers[0], msg);
                Log.Write("dbg", LogLvl.Hi, "FUSE group creation control operation took {0} seconds",
                    (appEnv.envGetTime() - cgStart) / 1000.0);
                if (rMsg != null)
                    Log.Write("stat", LogLvl.Hi, "FUSE group localCreate operation took {0} msecs",
                        rMsg.createLatency);
                if (rMsg == null) 
                {
                    Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                        "establishment to address {0} never acked. ***",
                        FUSEgroupMembers[0].GetAddress.ToString());
                }
                else 
                {
                    if (rMsg.dlccid == null) 
                    {
                        Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                            "establishment to address {0} acked, but creation failed. ***",
                            FUSEgroupMembers[0].GetAddress.ToString());
                    } 
                    else 
                    {
                        Log.Write("dbg", LogLvl.Hi, "Message requesting serial fuse group " +
                            "establishment to address {0} acked with dlccid {1}.",
                            FUSEgroupMembers[0].GetAddress.ToString(), rMsg.dlccid);
                        // notify the fuse group -- allow a certain number of msecs for notifications to propagate
                        // so that we don't walk over ourselves during the experiment                    

                        // pick a node to be the one we notify, and notify that node
                        int indexToNotify = App.random.Next(FUSEgroupSize);
                        NodeIdAddressBind dest = FUSEgroupMembers[indexToNotify];
                        RaiseNotificationControlMsg nmsg = new RaiseNotificationControlMsg(myServer, cf, rMsg.dlccid);
                        ControlReplyMsg rnMsg = (ControlReplyMsg) cf.SendRequestMsg(dest, nmsg);
                        if (rnMsg == null) 
                        {
                            Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                                "notification to address {0} never acked. ***",
                                dest.GetAddress.ToString());
                        }

                        // We check with all nodes that notification has occurred;
                        // currently I'm just letting 5 secs elapse, but I could RPC this.
                        exEnv.GoFor(5*1000); 

                    }
                }
            }   
        }

        public static void DoTest_1_Sim() 
        {
            DoTest_1_Common();
        }

        public static void DoTest_1_Debug() 
        {
            // copied this code from DoTest_1_Cluster -- I apologize! JD   
            FUSEgroupSize = 5; totalNumSerialTests = 20;
            Log.Write("flow", LogLvl.Hi, "proceeding to create a small number ({0}) of size {1} FUSE groups\r\n" +
                "in order to measure overhead of quiesced system", totalNumSerialTests, FUSEgroupSize);

            for (int i = 0; i < totalNumSerialTests; i++) 
            {
                //generate the nodes to ask to participate in this fuse group
                NodeIdAddressBind [] FUSEgroupMembers = new NodeIdAddressBind [FUSEgroupSize];
                Debug.Assert(FUSEgroupSize <= App.exEnv.NodeCount, 
                    "too few nodes in overlay for FEASIBLY creating FUSE groups of this size"); 
                string listOfFUSEgroupMembersForDebugging = "";
                for (int j = 0; j < FUSEgroupSize; j++) 
                {
                    int nextIndex = App.random.Next(App.exEnv.NodeCount);
                    // check for uniqueness
                    NodeIdAddressBind nextAddr = (SkipNet.NodeIdAddressBind) index_to_NodeIdAddressBind[nextIndex];
                    bool unique = true;
                    for (int k = 0; k < j; k++) 
                    {
                        if (FUSEgroupMembers[k].GetAddress == nextAddr.GetAddress) unique = false;
                    }
                    if (unique) 
                    {
                        FUSEgroupMembers[j] = nextAddr;
                        listOfFUSEgroupMembersForDebugging += (" " + nextAddr.ToString());
                    } 
                    else 
                    {
                        j--;
                    }
                }
                Log.Write("dbg", LogLvl.Hi, "Creating FUSE group {0} with membership {1}", 
                    i, listOfFUSEgroupMembersForDebugging);
 
                // create the fuse group
                long cgStart = appEnv.envGetTime();
                MeasureGroupCreateControlMsg msg = new MeasureGroupCreateControlMsg(myServer, cf, FUSEgroupMembers);
                CreateGroup_ControlReplyMsg rMsg = (CreateGroup_ControlReplyMsg) 
                    cf.SendRequestMsg(FUSEgroupMembers[0], msg);
                Log.Write("dbg", LogLvl.Hi, "FUSE group creation operation took {0} seconds",
                    (appEnv.envGetTime() - cgStart) / 1000.0);
                if (rMsg == null) 
                {
                    Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                        "establishment to address {0} never acked. ***",
                        FUSEgroupMembers[0].GetAddress.ToString());
                }
                else 
                {
                    if (rMsg.dlccid == null) 
                    {
                        Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                            "establishment to address {0} acked, but creation failed. ***",
                            FUSEgroupMembers[0].GetAddress.ToString());
                    } 
                    else
                    {
                        Log.Write("dbg", LogLvl.Hi, "Message requesting serial fuse group " +
                            "establishment to address {0} acked with dlccid {1}.",
                            FUSEgroupMembers[0].GetAddress.ToString(), rMsg.dlccid);
                    }
                }
            }
            Log.Write("flow", LogLvl.Hi, "DONE creating all those FUSE groups");
            Log.Write("flow", LogLvl.Hi, "Now let's measure the quiescent overhead");

            msecsToSpendInSteadyState = 5*60*1000;
            Log.Write("flow", LogLvl.Hi, "spending {0} mins in steady state", 
                msecsToSpendInSteadyState/(60*1000));
            DoSteadyStateTest();
            Log.Write("flow", LogLvl.Hi, "Done measuring the quiescent overhead");
            Log.SetCategoryFilterLevel("SLOG", LogLvl.Disabled);
        }

        public static void DoTest_1_Cluster()
        {
          
            DoTest_1_Common();

            msecsToSpendInSteadyState = 5*60*1000;
            Log.Write("flow", LogLvl.Hi, "spending {0} mins in steady state", 
                msecsToSpendInSteadyState/(60*1000));
            DoSteadyStateTest();

            Log.SetCategoryFilterLevel("SLOG", LogLvl.Hi);
            FUSEgroupSize = 5; totalNumSerialTests = App.exEnv.NodeCount;
            Log.Write("flow", LogLvl.Hi, "proceeding to create a large number ({0}) of size {1} FUSE groups\r\n" +
                "in order to measure overhead of quiesced system", totalNumSerialTests, FUSEgroupSize);

            for (int i = 0; i < totalNumSerialTests; i++) 
            {
                //generate the nodes to ask to participate in this fuse group
                NodeIdAddressBind [] FUSEgroupMembers = new NodeIdAddressBind [FUSEgroupSize];
                Debug.Assert(FUSEgroupSize <= App.exEnv.NodeCount, 
                    "too few nodes in overlay for FEASIBLY creating FUSE groups of this size"); 
                string listOfFUSEgroupMembersForDebugging = "";
                for (int j = 0; j < FUSEgroupSize; j++) 
                {
                    int nextIndex = App.random.Next(App.exEnv.NodeCount);
                    // check for uniqueness
                    NodeIdAddressBind nextAddr = (SkipNet.NodeIdAddressBind) index_to_NodeIdAddressBind[nextIndex];
                    bool unique = true;
                    for (int k = 0; k < j; k++) 
                    {
                        if (FUSEgroupMembers[k].GetAddress == nextAddr.GetAddress) unique = false;
                    }
                    if (unique) 
                    {
                        FUSEgroupMembers[j] = nextAddr;
                        listOfFUSEgroupMembersForDebugging += (" " + nextAddr.ToString());
                    } 
                    else 
                    {
                        j--;
                    }
                }
                Log.Write("dbg", LogLvl.Hi, "Creating FUSE group {0} with membership {1}", 
                    i, listOfFUSEgroupMembersForDebugging);
 
                // create the fuse group
                long cgStart = appEnv.envGetTime();
                MeasureGroupCreateControlMsg msg = new MeasureGroupCreateControlMsg(myServer, cf, FUSEgroupMembers);
                CreateGroup_ControlReplyMsg rMsg = (CreateGroup_ControlReplyMsg) 
                    cf.SendRequestMsg(FUSEgroupMembers[0], msg);
                Log.Write("dbg", LogLvl.Hi, "FUSE group creation operation took {0} seconds",
                    (appEnv.envGetTime() - cgStart) / 1000.0);
                if (rMsg == null) 
                {
                    Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                        "establishment to address {0} never acked. ***",
                        FUSEgroupMembers[0].GetAddress.ToString());
                }
                else 
                {
                    if (rMsg.dlccid == null) 
                    {
                        Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                            "establishment to address {0} acked, but creation failed. ***",
                            FUSEgroupMembers[0].GetAddress.ToString());
                    } 
                    else
                    {
                        Log.Write("dbg", LogLvl.Hi, "Message requesting serial fuse group " +
                            "establishment to address {0} acked with dlccid {1}.",
                            FUSEgroupMembers[0].GetAddress.ToString(), rMsg.dlccid);
                    }
                }
            }
            Log.Write("flow", LogLvl.Hi, "DONE creating all those FUSE groups");
            Log.Write("flow", LogLvl.Hi, "Now let's measure the quiescent overhead");

            msecsToSpendInSteadyState = 5*60*1000;
            Log.Write("flow", LogLvl.Hi, "spending {0} mins in steady state", 
                msecsToSpendInSteadyState/(60*1000));
            DoSteadyStateTest();
            Log.Write("flow", LogLvl.Hi, "Done measuring the quiescent overhead");
            Log.SetCategoryFilterLevel("SLOG", LogLvl.Disabled);

        }
        public static void DoTest_1_Common()  
        {
            Log.SetCategoryFilterLevel("SLOG", LogLvl.Disabled);
            int[] sizes = new int[5];
            sizes[0] = 2; 
            sizes[1] = 4; 
            sizes[2] = 8; 
            sizes[3] = 16; 
            sizes[4] = 32; 
            totalNumSerialTests = 20;
            foreach (int size in sizes) 
            {
                if (size > exEnv.NodeCount) 
                {
                    Console.WriteLine("Insufficient nodes {0} for serial size {1} experiment -- skipping it",
                        exEnv.NodeCount, size);
                    continue;
                }
                Log.Write("flow", LogLvl.Hi, "START serial size {0}, proceeding to serially create and notify groups of given size", size);
                FUSEgroupSize = size;
                DoSerialCreationAndNotificationTest();
                Log.Write("flow", LogLvl.Hi, "END serial size {0}, done with serial create and notify", size);
            }
        }
      
        static void ActivateNodes() {
            for (int i = 0; i < exEnv.NodeCount; i++) {
                SkipNet.Node node = exEnv.GetNode(i);
                node.ActivateNode(false);
            }
        }

        /// <summary>
        /// Measure cost of auto-repair.
        ///
        /// *** Simulator-specific ***
        ///
        /// This experiment has several phases:
        /// - Disable half the nodes in the overlay and then let the system quiesce.
        /// - Measure steady-state message traffic to obtain msgs/node/sec overhead of having an overlay network.
        /// - Churn overlay membership and measure message traffic to obtain msgs/node/sec overhead of 
        ///   overlay network repair.
        /// - Stop churning the overlay membership and quiesce the system.
        /// - Create FUSE groups.
        /// - Churn overlay nodes that are not members of a FUSE group and measure message traffic to obtain
        ///   msgs/sec of doing auto-repair (and overlay network repair).
        ///   We define 2 notions of msgs/node/sec for this phase of the experiment:
        ///   - Divide msgs/sec by the number of nodes belonging to FUSE groups.  This gives us the system-wide
        ///     traffic overhead of supporting a given number of FUSE group member nodes.
        ///   - Divide msgs/sec by the number of nodes belonging to FUSE groups plus the number of delegate nodes
        ///     supporting the FUSE groups.  This gives us the per-node traffic overhead for FUSE group "participants"
        ///     (both member and delegate).
        /// </summary>
        public static void DoTest_2_Sim(int nodeHalfLife, int fuseGroupSize) 
        {
//            Log.SetCategoryFilterLevel(SkipNet.JoinRequest.CATEGORY, LogLvl.Hi);
//            Log.SetCategoryFilterLevel(SkipNet.Node.CATEGORY, LogLvl.Hi);

            //
            // Disable half the nodes in the overlay to arrive at our "base" configuration.
            //
            Log.Write("flow", LogLvl.Hi, "Disable half of all overlay nodes");
            // Enable pinging so that the overlay network can repair itself.
            SkipNet.Parameters.pingEnabled = true;
            ActivateNodes();
            // Fail nodes.
            int numberOfCurrentlyLiveNodes = exEnv.NodeCount;
            while(numberOfCurrentlyLiveNodes >  (exEnv.NodeCount/2))
            {
                for (int i = 0; i < (numberOfCurrentlyLiveNodes/10); i++) {
                    int indexOfNodeToFlip = random.Next(App.exEnv.NodeCount);
                    if (!(App.exEnv.GetNode(indexOfNodeToFlip).Failed) && (indexOfNodeToFlip != 0)) { 
                        // fail this node
                        numberOfCurrentlyLiveNodes--;
                        App.appEnv.envDeactivateNode(indexOfNodeToFlip);
                    }
                    if (numberOfCurrentlyLiveNodes <=  (exEnv.NodeCount/2)) break;
                }
                // Wait for repair to finish.
                QuiesceSystem();
            }

            //
            // Measure traffic overhead of simply having an overlay network.
            //
            int initialTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            long initialTime = App.appEnv.envGetTime();
            SteadyStatePhase(5 * 60 * 1000);
            int finalTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            long finalTime = App.appEnv.envGetTime();
            int deltaTime = (int) (finalTime - initialTime);
            int avgMsgPerSec = (finalTotalMsgs - initialTotalMsgs) * 1000 / deltaTime;
            Log.Write("flow", LogLvl.Hi, "    Number of live nodes in overlay: {0}", numberOfCurrentlyLiveNodes);
            Log.Write("flow", LogLvl.Hi, "      Total msgs sent: {0}", finalTotalMsgs - initialTotalMsgs);
            Log.Write("flow", LogLvl.Hi, "    Avg msgs/sec: {0}, avg msgs/node/sec: {1}", avgMsgPerSec, ((float)avgMsgPerSec) / ((float)numberOfCurrentlyLiveNodes));

            //
            // Churn overlay membership (and log message traffic rates).
            //
            Log.Write("flow", LogLvl.Hi, "Begin churn phase");

            // Define how many churn events we want to observe.
            int numberOfDistinctChurnEvents = 100;

            Log.Write("flow", LogLvl.Hi, "numer of distinct churn events = {0}", numberOfDistinctChurnEvents);
            int msecsToSpendChurning = numberOfDistinctChurnEvents * nodeHalfLife / exEnv.NodeCount;
            if (msecsToSpendChurning < 5 * SkipNet.Parameters.neighbourPingPeriod) {
                msecsToSpendChurning = 5 * SkipNet.Parameters.neighbourPingPeriod;
            }
            Log.Write("flow", LogLvl.Hi, "time to spend churning = {0} msecs", msecsToSpendChurning);
            int churnDelayInMsecs = (int) msecsToSpendChurning/numberOfDistinctChurnEvents;
            Log.Write("flow", LogLvl.Hi, "churn event delay = {0} msecs", churnDelayInMsecs);
            Debug.Assert(churnDelayInMsecs > 0);

            int referenceNumberOfLiveNodes = numberOfCurrentlyLiveNodes;

            initialTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            initialTime = App.appEnv.envGetTime();
            for (int i = 0; i < numberOfDistinctChurnEvents; i++) 
            {
                // we need to find a node to fail, and a node to resurrect
                while (numberOfCurrentlyLiveNodes == referenceNumberOfLiveNodes) 
                { // look for a node to fail
                    int index = random.Next(App.exEnv.NodeCount);
                    if (!(App.exEnv.GetNode(index).Failed) && (index != 0))   // node is alive, fail it
                    {                        
                        numberOfCurrentlyLiveNodes--;
                        Log.Write("dbg", LogLvl.Hi, "failing node {0} with bind {1}", index, exEnv.GetNode(index).GetNodeIdAddressBind.GetAddress.ToString());
                        App.appEnv.envDeactivateNode(index);
                    }
                }
                while (numberOfCurrentlyLiveNodes < referenceNumberOfLiveNodes) 
                { // look for a node to resurrect
                    int index = random.Next(App.exEnv.NodeCount);
                    if (App.exEnv.GetNode(index).Failed)   // node is dead, resurrect it
                    {
                        numberOfCurrentlyLiveNodes++;
                        Log.Write("dbg", LogLvl.Hi, "reviving node {0} with bind {1}", index, exEnv.GetNode(index).GetNodeIdAddressBind.GetAddress.ToString());
                        App.appEnv.envActivateNode(index);
                    }
                }
                        
                // run forward
                Log.Write("dbg", LogLvl.Hi, "number of live nodes - {0}", numberOfCurrentlyLiveNodes);
                //Log.Write("dbg", LogLvl.Hi, "routing table just after changing node membership");
                Log.Write("dbg", LogLvl.Hi, "ITERATION {0}", i);
                //myServer.DumpRoutingTable();
                exEnv.GoFor(churnDelayInMsecs);
                //Log.Write("dbg", LogLvl.Hi, "routing table just before changing node membership");
                //myServer.DumpRoutingTable();
            }
            Log.Write("flow", LogLvl.Hi, "End churn phase");
            finalTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            finalTime = App.appEnv.envGetTime();
            deltaTime = (int) (finalTime - initialTime);
            avgMsgPerSec = (finalTotalMsgs - initialTotalMsgs) * 1000 / deltaTime;
            Log.Write("flow", LogLvl.Hi, "      Total msgs sent: {0}", finalTotalMsgs - initialTotalMsgs);
            Log.Write("flow", LogLvl.Hi, "    Avg msgs/sec: {0}, avg msgs/node/sec: {1}", avgMsgPerSec, ((float)avgMsgPerSec) / ((float)numberOfCurrentlyLiveNodes));

            // Let the system quiet down. (Should already be quiesced; but just in case.)
            QuiesceSystem();

            //
            // Create enough FUSE groups so that half of all overlay nodes are group members.
            //
            int totalFuseDelegates = 0;
            Log.Write("flow", LogLvl.Hi, "Begin create FUSE group phase");
            // choose 2/3 so as to have alignment with the cluster verision of this experiment 
            // when we set initial node num to 600
            int numFuseGroups = numberOfCurrentlyLiveNodes * 2 / (3 * fuseGroupSize); 
            Log.Write("flow", LogLvl.Hi, "number of FUSE groups to create: {0}", numFuseGroups);
            bool[] isFuseGroupMember = new bool[exEnv.NodeCount];
            for (int i = 0; i < exEnv.NodeCount; i++) {
                isFuseGroupMember[i] = false;
            }
            NodeIdAddressBind[] fuseGroup = new NodeIdAddressBind[fuseGroupSize];
            initialTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            initialTime = App.appEnv.envGetTime();
            for (int i = 0; i < numFuseGroups; i++) {
                int creatorIndex = -1;
                Node creatorNode = null;
                for (int j = 0; j < fuseGroupSize; j++) {
                    int indx = random.Next(App.exEnv.NodeCount);
                    Node node = App.exEnv.GetNode(indx);
                    while (node.Failed || isFuseGroupMember[indx]) {
                        indx = random.Next(App.exEnv.NodeCount);
                        node = App.exEnv.GetNode(indx);
                    }
                    fuseGroup[j] = exEnv.GetNode(indx).GetNodeIdAddressBind;
                    isFuseGroupMember[indx] = true;
                    Log.Write("dbg", LogLvl.Hi, "assigning node {0} with bind {1} to a fuse group", indx, exEnv.GetNode(indx).GetNodeIdAddressBind.GetAddress.ToString());
                    if (j == 0) {
                        creatorIndex = indx;
                        creatorNode = node;
                    }
                }
                FuseGroupCreated = false;
                MessageRouting.Node creationNode = App.exEnv.GetNode(creatorIndex);
                DLC.DLCC dlc = (DLC.DLCC) creationNode.GetApp(typeof(DLC.DLCC));
                CreateGroupDoneContinuation cont = new CreateGroupDoneContinuation();
                dlc.CreateGroup(fuseGroup, cont);
                exEnv.Go();
                Debug.Assert(FuseGroupCreated);
                totalFuseDelegates += FuseGroupDelegateNumber;
            }
            Log.Write("flow", LogLvl.Hi, "End create FUSE group phase");
            finalTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            finalTime = App.appEnv.envGetTime();
            deltaTime = (int) (finalTime - initialTime);
            avgMsgPerSec = (finalTotalMsgs - initialTotalMsgs) * 1000 / deltaTime;
            Log.Write("flow", LogLvl.Hi, "      Total msgs sent: {0}", finalTotalMsgs - initialTotalMsgs);
            Log.Write("flow", LogLvl.Hi, "    Avg msgs/sec: {0}, avg msgs/(FUSE group member)/sec: {1}",
                       avgMsgPerSec, ((float)avgMsgPerSec) / ((float)(numFuseGroups * fuseGroupSize)));

            //
            // Measure traffic overhead of simply having an overlay network with FUSE gruops.
            //
            initialTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            initialTime = App.appEnv.envGetTime();
            SteadyStatePhase(5 * 60 * 1000);
            finalTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            finalTime = App.appEnv.envGetTime();
            deltaTime = (int) (finalTime - initialTime);
            avgMsgPerSec = (finalTotalMsgs - initialTotalMsgs) * 1000 / deltaTime;
            Log.Write("flow", LogLvl.Hi, "      Total msgs sent: {0}", finalTotalMsgs - initialTotalMsgs);
            Log.Write("flow", LogLvl.Hi, "    Avg msgs/sec: {0}, avg msgs/node/sec: {1}", avgMsgPerSec, ((float)avgMsgPerSec) / ((float)numberOfCurrentlyLiveNodes));

            //
            // Churn non-group-member overlay nodes (and log message traffic rates).
            //
            Log.Write("flow", LogLvl.Hi, "Begin FUSE group churn phase");
            initialTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            for (int i = 0; i < numberOfDistinctChurnEvents; i++) {
                // Find a node to fail.  Choose only from nodes that aren't FUSE group members.
                while (true) {
                    int indx = random.Next(App.exEnv.NodeCount);
                    if (!(App.exEnv.GetNode(indx).Failed) && (indx != 0) && !isFuseGroupMember[indx]) {
                        App.appEnv.envDeactivateNode(indx);
                        Log.Write("dbg", LogLvl.Hi, "killing node {0} with bind {1}", indx, exEnv.GetNode(indx).GetNodeIdAddressBind.GetAddress.ToString());
                        break;
                    }
                }

                // Find a node to resurrect.
                while (true) {
                    int indx = random.Next(App.exEnv.NodeCount);
                    if (App.exEnv.GetNode(indx).Failed) {
                        App.appEnv.envActivateNode(indx);
                    Log.Write("dbg", LogLvl.Hi, "resurrecting node {0} with bind {1} to a fuse group", indx, exEnv.GetNode(indx).GetNodeIdAddressBind.GetAddress.ToString());
                        break;
                    }
                }

                // Run forward.
                exEnv.GoFor(churnDelayInMsecs);
            }
            Log.Write("flow", LogLvl.Hi, "End FUSE group churn phase");
            finalTotalMsgs = App.appEnv.envGetTotalMsgsSent();
            avgMsgPerSec = (finalTotalMsgs - initialTotalMsgs) * 1000 / msecsToSpendChurning;
            Log.Write("flow", LogLvl.Hi, "      Total msgs sent: {0}", finalTotalMsgs - initialTotalMsgs);
            Log.Write("flow", LogLvl.Hi, "    Avg msgs/sec: {0}", avgMsgPerSec);
            int totalFuseMembers = numFuseGroups * fuseGroupSize;
            Log.Write("flow", LogLvl.Hi, "    Avg msgs/(FUSE group member)/sec: {0},  Avg msgs/(FUSE group member || delegate)/sec: {1}",
                       ((float)avgMsgPerSec) / ((float)totalFuseMembers), ((float)avgMsgPerSec) / ((float)(totalFuseMembers + totalFuseDelegates)));
  
            // Let the system quiet down. (Should already be quiesced; but just in case.)
            SteadyStatePhase(15 * 60 * 1000);
            Log.Write("flow", LogLvl.Hi, "Done!");               
        }

        public static void QuiesceSystem() {
            Log.Write("flow", LogLvl.Hi, "Beginning QuiesceSystem");   
             Log.Write("dbg", LogLvl.Hi, "num rebuilds: {0}", SkipNet.RoutingTable.NumNeighborRebuilds);
             int prevNumRebuilds = SkipNet.RoutingTable.NumNeighborRebuilds;
            exEnv.GoFor(2 * SkipNet.Parameters.neighbourPingPeriod); // Run at least for a couple of ping intervals.
             int numRebuilds = SkipNet.RoutingTable.NumNeighborRebuilds;
             Log.Write("dbg", LogLvl.Hi, "num rebuilds: {0}", SkipNet.RoutingTable.NumNeighborRebuilds);
            while (numRebuilds > prevNumRebuilds) {
                exEnv.GoFor(1 * 60 * 1000);
                prevNumRebuilds = numRebuilds;
                numRebuilds = SkipNet.RoutingTable.NumNeighborRebuilds;
                Log.Write("dbg", LogLvl.Hi, "num rebuilds: {0}", SkipNet.RoutingTable.NumNeighborRebuilds);
            }
            Log.Write("flow", LogLvl.Hi, "Finished QuiesceSystem");        
        }

        public static void SteadyStatePhase(int msecsToRun) {
            Log.Write("flow", LogLvl.Hi, "Beginning SteadyStatePhase");
            exEnv.GoFor(msecsToRun);
            Log.Write("flow", LogLvl.Hi, "Finished SteadyStatePhase");
       }

        static bool FuseGroupCreated;
        static int FuseGroupDelegateNumber;

        class CreateGroupDoneContinuation: DLC.DLCCContinuation {

            public CreateGroupDoneContinuation() {
            }

            public override void Execute( MessageRouting.Node localNode ) {
                FuseGroupCreated = true;
                DLC.DLCC dlc = (DLC.DLCC) localNode.GetApp(typeof(DLC.DLCC));
                ArrayList bagOfIDs = (ArrayList) dlc.dlccids_to_nodeIDAddressBindsOfGroupMembersPlusDelegates[this.DLCCId];
                Hashtable delegates = new Hashtable();
                foreach (NodeIdAddressBind n in bagOfIDs) {
                    if (!delegates.Contains(n.GetNodeId.ToString())) {
                        delegates[n.GetNodeId.ToString()] = 1;
                    }
                    else {
                        int cnt = (int) delegates[n.GetNodeId.ToString()];
                        delegates[n.GetNodeId.ToString()] = cnt + 1;
                    }
                }
                FuseGroupDelegateNumber = delegates.Count;
                Log.Write("flow", LogLvl.Hi, "FUSE group created: {0}", this.DLCCId);
            }
        }


        public static void DoTest_2_Cluster(int fuseGroupSize)
        {
            // start by waiting two minutes for any remaining node joins to complete
            exEnv.GoFor(2*60*1000);

            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "Test2: Start killing nodes: reduce overlay from 400 to 300 and quiesce (5min)");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");

            // allow 5 minutes for the kill script to run, reducing the system size from 400 to 300 nodes
            exEnv.GoFor(5*60*1000);

            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "Test2: End initial quiesce time");
            Log.Write("flow", LogLvl.Hi, "Test2: Start steady-state period #1 - overlay load with 300 nodes (5min)");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            // 5 minutes steady state
            exEnv.GoFor(5*60*1000);

            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "Test2: End steady-state period #1");
            Log.Write("flow", LogLvl.Hi, "Test2: Start steady-state period #2 - churn overhead without FUSE (10min)");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");

            // 10 minutes steady state
            exEnv.GoFor(10*60*1000);

            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "Test2: End steady-state period #2");
            Log.Write("flow", LogLvl.Hi, "Test2: Pause 5 before creating FUSE groups");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");

            // 5 minutes steady state
            exEnv.GoFor(5*60*1000);
            
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "Test2: Starting to create FUSE groups on the stable nodes");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");

            // (2) create a number of fuse groups on nodes that we know will not be killed
            // our convention for this experiment is that the first half of the cluster is immune from churn
            // First half of cluster is defined to be nodes herald001 through herald021.
            bool[] indicesEligibleToBeMembers = new bool[exEnv.NodeCount];
            bool[] indicesPickedToBeMembers = new bool[exEnv.NodeCount];
            for (int i=0; i < exEnv.NodeCount; i++) 
            {
                NodeIdAddressBind n = (NodeIdAddressBind) index_to_NodeIdAddressBind[i];
                NodePhysicalAddress addr = n.GetAddress;
                int port = addr.Port;

                // HACK: here we have hardcoded all sorts of detailed information about the
                // herald cluster and the contents of example mapping, as well as the convention 
                // that the herald virtual node index corresponds to the address port number
                // minus 10000.

                // We also assume that there are 46 physical nodes and 10 virtual per machine
                // in the mapping file
                int num = port - 10000;

                // heraldnum is the number of the physical node
                int heraldnum = (num % 46) + 1;
                // herald004 is not part of the cluster at the moment
                if (heraldnum > 3) 
                    heraldnum++;
                
                if (heraldnum <= 21) {
                    indicesEligibleToBeMembers[i] = true;
                } else {
                    indicesEligibleToBeMembers[i] = false;
                }
                indicesPickedToBeMembers[i] = false; //still haven't created any of the fuse groups yet
            }
            // now pick from amongst these nodes
            int totalNumMembers = 200;
            int numGroups = totalNumMembers/fuseGroupSize;
            for (int i = 0; i < numGroups; i++) 
            {
                for (int j = 0; j < exEnv.NodeCount; j++)
                {
                    indicesPickedToBeMembers[j] = false;
                }
                NodeIdAddressBind[] members = new NodeIdAddressBind[fuseGroupSize];
                for (int j = 0; j < fuseGroupSize; j++) 
                { 
                    int idx = App.random.Next(exEnv.NodeCount); 
                    while ((indicesEligibleToBeMembers[idx] == false) ||
                           (indicesPickedToBeMembers[idx] == true))
                    {
                        idx = App.random.Next(exEnv.NodeCount); 
                    }
                    members[j] = (NodeIdAddressBind) index_to_NodeIdAddressBind[idx];
                    indicesPickedToBeMembers[idx] = true;
                }
                long cgStart = appEnv.envGetTime();
                MeasureGroupCreateControlMsg msg = new MeasureGroupCreateControlMsg(myServer, cf, members);
                CreateGroup_ControlReplyMsg rMsg = (CreateGroup_ControlReplyMsg) 
                    cf.SendRequestMsg(members[0], msg);
                Log.Write("dbg", LogLvl.Hi, "FUSE group creation operation took {0} seconds",
                    (appEnv.envGetTime() - cgStart) / 1000.0);
                if (rMsg == null) 
                {
                    Log.Write("flow", LogLvl.Hi, "*** error: Message requesting fuse group " +
                        "establishment to address {0} never acked. ***",
                        members[0].GetAddress.ToString());
                } 
                else 
                {
                    if (rMsg.dlccid == null) 
                    {
                        Log.Write("dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                            "establishment to address {0} acked, but creation failed. ***",
                            members[0].GetAddress.ToString());
                    } 
                    else 
                    {
                        Log.Write("dbg", LogLvl.Hi, "Message requesting serial fuse group " +
                            "establishment to address {0} acked with dlccid {1}.",
                            members[0].GetAddress.ToString(), rMsg.dlccid);
                    }
                }
            }

            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "Test2: Finished creating fuse groups");
            Log.Write("flow", LogLvl.Hi, "Test2: Pausing 2 minutes before starting next churn");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");

            // 2 minutes steady state
            exEnv.GoFor(2*60*1000);

            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "Test2: Start steady-state period #3 - node churn with FUSE groups (10min)");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");

            // allow 10 minutes for there to be churn among the remaining nodes
            exEnv.GoFor(10*60*1000);

            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "Test2: End steady-state period #3");
            Log.Write("flow", LogLvl.Hi, "Test2: Pausing 2 minutes before starting terminate");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "*************************************************");
            Log.Write("flow", LogLvl.Hi, "Done!");               

            // 2 minutes steady state
            exEnv.GoFor(2*60*1000);

        }


        public static void DoTest_3() 
        {
            Log.SetCategoryFilterLevel("Test_3_dbg", LogLvl.Hi);

            // Enable pinging so that the overlay network can repair itself.
            SkipNet.Parameters.pingEnabled = true;
            ActivateNodes();
            
            // Create fuse groups of various sizes (varying from 2 to 32) until
            // we have a total number of members equal to half the overlay node number.
            int totalFuseGroupMembers = 0;
            while (true) {
                FUSEgroupSize = 2 + random.Next(30);

                //generate the nodes to ask to participate in this fuse group
                NodeIdAddressBind [] FUSEgroupMembers = new NodeIdAddressBind [FUSEgroupSize];
                Debug.Assert(FUSEgroupSize <= App.exEnv.NodeCount, 
                             "too few nodes in overlay for FEASIBLY creating FUSE groups of this size"); 
                string listOfFUSEgroupMembersForDebugging = "";
                for (int j = 0; j < FUSEgroupSize; j++) 
                {
                    int nextIndex = App.random.Next(App.exEnv.NodeCount);
                    // check for uniqueness
                    NodeIdAddressBind nextAddr = (SkipNet.NodeIdAddressBind) index_to_NodeIdAddressBind[nextIndex];
                    bool unique = true;
                    for (int k = 0; k < j; k++) 
                    {
                        if (FUSEgroupMembers[k].GetAddress == nextAddr.GetAddress) unique = false;
                    }
                    if (unique) 
                    {
                        FUSEgroupMembers[j] = nextAddr;
                        listOfFUSEgroupMembersForDebugging += (" " + nextIndex.ToString());
                    } 
                    else 
                    {
                        j--;
                    }
                }
                Log.Write("Test_3_dbg", LogLvl.Hi, "Creating FUSE group with membership {0}", 
                           listOfFUSEgroupMembersForDebugging);
 
                // create the fuse group
                long cgStart = appEnv.envGetTime();
                MeasureGroupCreateControlMsg msg = new MeasureGroupCreateControlMsg(myServer, cf, FUSEgroupMembers);
                CreateGroup_ControlReplyMsg rMsg = (CreateGroup_ControlReplyMsg) 
                    cf.SendRequestMsg(FUSEgroupMembers[0], msg);
                Log.Write("Test_3_dbg", LogLvl.Hi, "FUSE group creation operation took {0} seconds",
                    (appEnv.envGetTime() - cgStart) / 1000.0);
                if (rMsg == null) 
                {
                    Log.Write("Test_3_dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                               "establishment to address {0} never acked. ***",
                               FUSEgroupMembers[0].GetAddress.ToString());
                }
                else 
                {
                    if (rMsg.dlccid == null) 
                    {
                        Log.Write("Test_3_dbg", LogLvl.Hi, "*** error: Message requesting serial fuse group " +
                                   "establishment to address {0} acked, but creation failed. ***",
                                   FUSEgroupMembers[0].GetAddress.ToString());
                    } 
                    else 
                    {
                        Log.Write("Test_3_dbg", LogLvl.Hi, "Message requesting serial fuse group " +
                                   "establishment to address {0} acked with dlccid {1}.",
                                   FUSEgroupMembers[0].GetAddress.ToString(), rMsg.dlccid);
                    }
                }

                totalFuseGroupMembers += FUSEgroupSize;
                if (totalFuseGroupMembers > (App.exEnv.NodeCount / 2)) break;
            }

            // Let the system run for a while so that we can observe false positives
            // (which are logged using slog).
            App.appEnv.envEnableDroppedPackets(0.01);
            Log.Write("flow", LogLvl.Hi, "Wait to allow for false positives");               
            App.exEnv.GoFor(15 * 60 * 1000);  // Let experiment run for 15 minutes

            Log.Write("flow", LogLvl.Hi, "Experiment done!");               
        }

        public static void DoCreateOneAndFailHardTest()  
            {
            
            // 1. pick a node that we'll fail (via manually disconnecting in the cluster)
            // need to by hand pick a set of nodes that will be on different physical nodes, and which phyical node they're on
            // for the moment, we'll hard code this to be a set of nodes that probably won't work.
            NodeIdAddressBind [] FUSEgroupMembers = new NodeIdAddressBind[FUSEgroupSize];
            for (int i = 0; i < FUSEgroupSize; i++) 
            {
                FUSEgroupMembers[i] = (NodeIdAddressBind) index_to_NodeIdAddressBind[i];
            }
            
            // 2. signal on the control node which one it is
            Log.SetCategoryFilterLevel("hardfail", LogLvl.Hi);
            Log.Write("hardfail", LogLvl.Hi, "going to 1st set up a fuse group and then fail node {0}", 
                FUSEgroupMembers[FUSEgroupSize-1].ToString()); 

            // 3. set up a FUSE group through that node and to a few other nodes on distinct cluster nodes
            long cgStart = appEnv.envGetTime();
            MeasureGroupCreateControlMsg msg = new MeasureGroupCreateControlMsg(myServer, cf, FUSEgroupMembers);
            CreateGroup_ControlReplyMsg rMsg = (CreateGroup_ControlReplyMsg) 
                cf.SendRequestMsg(FUSEgroupMembers[0], msg);
            Log.Write("hardfail", LogLvl.Hi, "FUSE group creation operation took {0} seconds",
                (appEnv.envGetTime() - cgStart) / 1000.0);
            if (rMsg == null) 
            {
                Log.Write("hardfail", LogLvl.Hi, "*** error: Message requesting fuse group " +
                    "establishment to address {0} never acked. ***",
                    FUSEgroupMembers[0].GetAddress.ToString());
            }
            Log.Write("hardfail", LogLvl.Hi, "Created FUSE group, starting count down...");
            // 4. allow timer to elapse, and print on screen that it's time to manually disconnect the cable
            
            for (int i = 0; i < 10; i++) 
            {
                Log.Write("hardfail", LogLvl.Hi, "...{0} secs until you pull the cable", 
                    ((msecsFromGlobalStartUntilHardFailureSignal/1000)/10)*(10-i));
                exEnv.GoFor(msecsFromGlobalStartUntilHardFailureSignal/10);
            }
            Log.Write("hardfail", LogLvl.Hi, "PULL THE CABLE!");
            Log.Write("hardfail", LogLvl.Hi, "cable pulled (we hope)");

            // 5. gather up the TLog msgs corresponding to this event
            exEnv.GoFor((int) msecsToAllowForFailureNotificationsToPropagate);
            Log.Write("hardfail", LogLvl.Hi, "presuming that all failure notifications have been " +
                "gathered and that it's time to exit");

        }

        public static void DoParallelCreationAndDestructionTest() {
            Debug.Assert(false, "this method was broken by the removal of the global name service");
            Log.Write("flow", LogLvl.Hi, "Signalling the ParallelCreationAndDestructionTest " +
                "from master node");   

            // Send message to start a timer. Timer will trigger creating FUSE groups.
            long timeToStartCreatingGroups = appEnv.envGetTime() 
                + msecsOffsetFromGlobalStartToGroupCreation;
            for (int i = 0; i < exEnv.NodeCount; i++)  // i indexes which node to tell "go"
            {
                // Quick-and-easy rule for which nodes are going to fail early: 
                // the one's with higher nodeIndex than average.
                bool failEarly = (i >= exEnv.NodeCount/2); 
                StartParallelControlMsg msg = new StartParallelControlMsg(myServer, cf, 
                    FUSEgroupsPerNode, timeToStartCreatingGroups, msecsToSpendQuiesced, failEarly);
                ControlReplyMsg rMsg = (ControlReplyMsg) cf.SendRequestMsg(exEnv.NodeIdAddrBind(i), msg);
                if (rMsg == null) 
                {
                    Log.Write("dbg", LogLvl.Hi, "*** error: Message 'start' to node {0} address {1} never acked. ***",
                        i, exEnv.NodePhysAddr(i).ToString());
                }
                else 
                {
                    Log.Write("dbg", LogLvl.Hi, "Message 'start' to node {0} address {1} was acked.",
                        i, exEnv.NodePhysAddr(i).ToString());
                } 
            }
            Log.Write("flow", LogLvl.Hi, "Finished signaling the ParallelCreationAndDestructionTest " +
                "from  master node");

            long msecsUntilGroupCreation = (timeToStartCreatingGroups - appEnv.envGetTime());
            Debug.Assert(msecsUntilGroupCreation > 0, "BARF! - we're still sending out messages" +
                "suggesting that it's time to start creating groups,\r\n" +
                "when the time to start group creation has already passed!");

            Log.Write("flow", LogLvl.Hi, "going to wait {0} secs",
                 msecsUntilGroupCreation/1000);
            exEnv.GoFor((int) (msecsUntilGroupCreation + msecsToSpendQuiesced - 1));
            Log.Write("flow", LogLvl.Hi, "nodes are about to fail...");
            exEnv.GoFor((int) (msecsToAllowForFailureNotificationsToPropagate));


        }
    
        /// <summary>
        /// Broadcast node names, and start experiment-independent timers, like logging of
        /// cpu load and message counts.
        /// </summary>
        static void MeasureFUSEInitialization() 
        {
            for (int i = 0; i < exEnv.NodeCount; i++) 
            {
                InitializeMFAControlMsg msg = new InitializeMFAControlMsg(myServer, cf, FUSEgroupSize);
                ControlReplyMsg rMsg = (ControlReplyMsg) cf.SendRequestMsg(exEnv.NodeIdAddrBind(i), msg);
                if (rMsg == null) 
                {
                    Log.Write("dbg", LogLvl.Hi, "*** error: Message 'init' to node {0} address {1} never acked. ***",
                        i, exEnv.NodePhysAddr(i).ToString());
                }
                else 
                {
                    Log.Write("dbg", LogLvl.Hi, "Message 'init' to node {0} address {1} was acked.",
                        i, exEnv.NodePhysAddr(i).ToString());
                }
            }
        }

        /// <summary>
        /// Ping all nodes to confirm liveness.
        /// </summary>
        static void PingNodesTillTheyReply() 
        {
            index_to_NodeIdAddressBind[0] = myServer.GetNodeIdAddressBind;

            for (int i = 1; i < exEnv.NodeCount; i++) 
            {
                // Ping the node for liveness
                Log.Write("flow-action", LogLvl.Hi, "PingNodesTillTheyReply {0} at {1:T}", i, DateTime.Now);
                long timeoutMs = 4000; // Start with four seconds
                while (false == cf.ExecuteBasicNodeOpn(i, ControlFlow.BasicNodeOpn.Ping, index_to_NodeIdAddressBind, timeoutMs)) 
                {
                    Log.Write("flow-action", LogLvl.Hi, "Timeout while pinging node {0} at {1:T}.  Trying again ...", i, DateTime.Now);
                    timeoutMs *= 2;
                    if (timeoutMs > 1 * 60 * 1000) timeoutMs = 1 * 60 * 1000; // Cap at 1 minute
                }
            }
            Log.Write("flow-action", LogLvl.Hi, "PingNodesTillTheyReply finished at {0:T}", DateTime.Now);
        }

        /// <summary>
        /// Signal all nodes (including nodes that have already failed)
        /// to terminate themselves (i.e. exit their programs).
        /// This would be slow if we had strictly silent failures, but we don't.
        /// </summary>
        static void TerminateNodes() 
        {
            if (App.exEnv.Type() == ExecutionEnvironmentType.SimulatorEnvironment) ExitProg(0);
            for (int i = 1; i < exEnv.NodeCount; i++) 
            {
                Log.Write("flow-action", LogLvl.Hi, "Terminating node {0} with address {1} at {2:T}",
                    i, exEnv.NodePhysAddr(i), DateTime.Now);
                cf.ExecuteBasicNodeOpn(i, ControlFlow.BasicNodeOpn.ExitProgram);
            }
        }

        //================================ Handle program args and related state ================================

        /// <summary>
        /// Print the various command-line options to MeasureFUSE.
        /// </summary>
        public static void Usage() 
        {
            Console.WriteLine("MeasureFUSE");

            appEnv.envPrintUsage();
            Console.WriteLine("  Common Parameters:");
            CommonParameters.PrintUsage();

            Console.WriteLine("  Experiment Parameters:");
            Console.WriteLine("    [-F] -- Turns on application 'flow' output");
            Console.WriteLine("    [-L] -- Initialize message logging etc. at boot time instead of by explicit message");
            Console.WriteLine("    [-LR] -- Like -L, but also log RTable");
            Console.WriteLine("    [-LS] -- Silence logging");
            Console.WriteLine("    [-R] -- Log RTable, but central server based initialization");
            Console.WriteLine("    [-periodOfPinging] -- R-Table ping period in msecs");
            Console.WriteLine("    [-Measure] -- turns on instrumentation for experiments");
            Console.WriteLine("    [-MeasureOther] -- turns on non-load instrumentation only");
            Console.WriteLine("    [-FuseGroupSize] -- size of each Fuse group");
            Console.WriteLine("    TestAutoRepair parameters...");
            Console.WriteLine("    [-DoTestAutoRepair] -- run the test auto repair experiment");
            Console.WriteLine("    SteadyStateTest parameters...");
            Console.WriteLine("    [-DoSteadyState] -- run the steady state experiment");
            Console.WriteLine("    [-msecsSteadyState] -- msecs to spend in steady state");
            Console.WriteLine("    CreateOneAndFailHardTest parameters...");
            Console.WriteLine("    [-DoCreateOneAndFailHard] -- run the create one and fail hard experiment");
            Console.WriteLine("    [-msecsHardFailDelay] -- msecs to wait after pinging everyone before\r\n" +
                              "    suggesting to the operator that\r\n" +
                              "    it's time to remove the ethernet cable (i.e., a hard failure)");
            Console.WriteLine("    SerialCreationAndNotificationTest parameters...");
            Console.WriteLine("    [-DoSerialCreationAndNotification] -- run the serial create/notify experiment");
            Console.WriteLine("    [-SerialTestNum] -- how many distinct FUSE group tests to do in serial");
            Console.WriteLine("    Test_1 parameters...");
            Console.WriteLine("    [-DoTest_1_Debug] -- you are working on making the Test experiment #1 run");
            Console.WriteLine("    [-DoTest_1_Sim] -- run the Test experiment #1");
            Console.WriteLine("    [-DoTest_1_Cluster] -- run the Test experiment #1 (including the quiesced measurement)");
            Console.WriteLine("    Test_2 parameters...");
            Console.WriteLine("    [-DoTest_2_Sim] -- run the Test experiment #2 in the simulator");
            Console.WriteLine("    [-DoTest_2_Cluster] -- run the Test experiment #2 on the cluster");
            Console.WriteLine("    Test_3 parameters...");
            Console.WriteLine("    [-DoTest_3] -- run the Test experiment #3 (false positives)");
            Console.WriteLine("    ParallelCreationAndDestructionTest parameters...");
            Console.WriteLine("    [-DoParallelCreationAndDestruction] -- run the parallel create/parallel fail experiment");
            Console.WriteLine("    [-FuseGroupNum] -- number of FUSE groups per node");
            Console.WriteLine("    [-msecsStartToCreation] -- msecs from global start to FUSE group creation");
            Console.WriteLine("    [-msecsQuiesced] -- msecs to spend quiesced after creation and before failures");
            Console.WriteLine("    [-msecsNotificationBuffer] -- msecs to allow for failure notification to be noticed and propagate");
            Console.WriteLine("                                  THIS PARAMETER IS ALSO USED BY CreateOneAndFailHardTest");
            Console.WriteLine("    [-quit] -- Causes program to exit without waiting for input");
            Console.WriteLine("    Next revision should command line control of group size and total number of groups being created.");
            Console.WriteLine("  Debugging Parameters:");
            Console.WriteLine("    [-D] -- Turns on application debugging output");
            Console.WriteLine("    [-DebugDlc] -- Turns on dlc debugging output");

            ExitProg(0);
        }
       
        /// <summary>
        /// Print the parameters that are being used for this experiment
        /// </summary>
        public static void PrintExperimentParameters() 
        {
            Console.WriteLine("Running MeasureFUSE with parameters");

            appEnv.envPrintArgs();
            Console.WriteLine("  Common Parameters:");
            CommonParameters.PrintArgs();

            Console.WriteLine();
            Console.WriteLine("  Experiment Parameters:");
            Console.WriteLine("    -F: log application flow output: {0}", doLogAppFlow);
            Console.WriteLine("    -periodOfPinging (in msecs): {0}", SkipNet.Parameters.neighbourPingPeriod);
            Console.WriteLine("    -Measure: {0}", doStatLogging);
            Console.WriteLine("    -MeasureOther: {0}", doNonLoadStatLoggingOnly);
            Console.WriteLine("    -FuseGroupSize: {0}", FUSEgroupSize);
            if (doTest_1_Sim || doTest_1_Cluster || doTest_1_Debug) 
            {
                Console.WriteLine("    Going to do the suite of Test_1 experiments ...");
            }
            if (doTest_2_Sim || doTest_2_Cluster) 
            {
                Console.WriteLine("    Going to do Test_2 experiment (churn) ...");
            }
            if (doTest_3) {
                Console.WriteLine("    Going to do Test_3 experiment (false positives) ...");
            }
            if (doTestAutoRepair) 
            {
                Console.WriteLine("    Going to Test Auto Repair ...");
            }
            if (doSteadyStateTest) 
            {
                Console.WriteLine("    SteadyStateTest parameters...");
                Console.WriteLine("    -msecsSteadyState: {0}", msecsToSpendInSteadyState);
            }
            if (doCreateOneAndFailHardTest)
            {
                Console.WriteLine("    CreateOneAndFailHardTest parameters...");
                Console.WriteLine("    -msecsHardFailDelay: {0}", msecsFromGlobalStartUntilHardFailureSignal);
            }
            if (doSerialCreationAndNotificationTest) 
            {
                Console.WriteLine("    SerialCreationAndNotificationTest parameters...");
                Console.WriteLine("    -SerialTestNum: {0}", totalNumSerialTests);
            }
            if (doParallelCreationAndDestructionTest) 
            {
                Console.WriteLine("    ParallelCreationAndDestructionTest parameters...");
                Console.WriteLine("    -FuseGroupNum: {0}", FUSEgroupsPerNode);
                Console.WriteLine("    -msecsStartToCreation: {0}", msecsOffsetFromGlobalStartToGroupCreation);
                Console.WriteLine("    -msecsQuiesced: {0}", msecsToSpendQuiesced);
                Console.WriteLine("    -msecsNotificationBuffer: {0}", msecsToAllowForFailureNotificationsToPropagate);
            }
                
            Console.WriteLine("  App Level Debugging Parameters");            
            Console.WriteLine("    -DebugDlc: {0}", debugDlc);
            Console.WriteLine("    -D: (app debugging): {0}", DebugOutput);

            if (quitImmediately)
                Console.WriteLine("    -quit -- Exit without waiting for input");
            Console.WriteLine();
        }

        /// <summary>
        /// Process the command-line options.
        /// </summary>
        public static void GetOptions(string[] args, int starting_at) 
        {
            for (int i = starting_at; i < args.Length; i++) 
            {
                if ( (args[i][0] == '-') || (args[i][0] == '/') ) 
                {
                    switch(args[i][1]) 
                    {
                        case 'q':
                            quitImmediately = true;
                            break;
                        case 'p':
                            if (args[i].Equals("-periodOfPinging"))
                                periodOfPinging = Convert.ToInt32(args[++i]);
                            break;
                        case 'D':
                            if (args[i].Equals("-D"))
                                DebugOutput = true;
                            if (args[i].Equals("-DebugDlc"))
                                debugDlc = true;
                            if (args[i].Equals("-DoTestAutoRepair"))
                                doTestAutoRepair = true;
                            if (args[i].Equals("-DoSteadyState"))
                                doSteadyStateTest = true;
                            if (args[i].Equals("-DoCreateOneAndFailHard"))
                                doCreateOneAndFailHardTest = true;
                            if (args[i].Equals("-DoSerialCreationAndNotification"))
                                doSerialCreationAndNotificationTest = true;
                            if (args[i].Equals("-DoTest_1_Sim"))
                                doTest_1_Sim = true;
                            if (args[i].Equals("-DoTest_1_Debug"))
                                doTest_1_Debug = true;
                            if (args[i].Equals("-DoTest_2_Sim"))
                                doTest_2_Sim = true;
                            if (args[i].Equals("-DoTest_2_Cluster"))
                                doTest_2_Cluster = true;
                            if (args[i].Equals("-DoTest_1_Cluster"))
                                doTest_1_Cluster = true;
                            if (args[i].Equals("-DoTest_3"))
                                doTest_3 = true;
                            if (args[i].Equals("-DoParallelCreationAndDestruction"))
                                doParallelCreationAndDestructionTest = true;
                            break;
                        case 'L':
                            if (args[i].Equals("-L"))
                                doImmediateLogging = true;
                            if (args[i].Equals("-LR")) 
                            {
                                doImmediateLogging = true;
                                doImmediateLoggingAndLogRTable = true;
                            }
                            if (args[i].Equals("-LS")) 
                            {
                                doImmediateLogging = true;
                                doSilentStatLogging = true;
                            }
                            break;
                        case 'R':
                            doLogRTable = true;
                            break;
                        case 'M':
                            if (args[i].Equals("-Measure"))
                                doStatLogging = true;
                            if (args[i].Equals("-MeasureOther"))
                                doNonLoadStatLoggingOnly = true;
                            break;
                        case 'S':
                            if (args[i].Equals("-SerialTestNum"))
                                totalNumSerialTests = Convert.ToInt32(args[++i]);
                            break;
                        case 'F':
                            if (args[i].Equals("-F"))
                                doLogAppFlow = true;
                            if (args[i].Equals("-FuseGroupSize")) 
                                FUSEgroupSize = Convert.ToInt32(args[++i]);
                            if (args[i].Equals("-FuseGroupNum"))
                                FUSEgroupsPerNode = Convert.ToInt32(args[++i]);
                            break;
                        case 'm':
                            if (args[i].Equals("-msecsStartToCreation"))
                                msecsOffsetFromGlobalStartToGroupCreation = Convert.ToInt32(args[++i]);
                            if (args[i].Equals("-msecsQuiesced"))
                                msecsToSpendQuiesced = Convert.ToInt32(args[++i]);
                            if (args[i].Equals("-msecsNotificationBuffer"))
                                msecsToAllowForFailureNotificationsToPropagate = Convert.ToInt32(args[++i]);
                            if (args[i].Equals("-msecsSteadyState"))
                                msecsToSpendInSteadyState = Convert.ToInt32(args[++i]);
                            if (args[i].Equals("-msecsHardFailDelay"))
                                msecsFromGlobalStartUntilHardFailureSignal = Convert.ToInt32(args[++i]);
                            break;
                        default:
                            Usage();
                            break;
                    }
                }
                else
                    Usage();
            }

            //
            // Fixup & verify parameters
            //

            // Deal with environment-specific parameters.
            appEnv.envParameterFixup();
        }


        //======================== Functions to be called from environemnt-specific code ========================

        public static void SetParameterDefaults() 
        {
            Log.SetDefaultFilterLevel(LogLvl.Disabled);
            Log.SetCategoryFilterLevel("test", LogLvl.Hi);
            Log.SetCategoryFilterLevel("tlog", LogLvl.Hi);
            Log.SetCategoryFilterLevel("flow-action", LogLvl.Hi);
            Log.SetCategoryFilterLevel("dbg", LogLvl.Hi);
            Log.SetCategoryFilterLevel("TLog", LogLvl.Hi);


            SkipNet.Parameters.b = 8;
            SkipNet.Parameters.denseRTable = true;
            SkipNet.Parameters.proximityTableEnabled = false;
            SkipNet.Parameters.cTableEnabled = false;
            SkipNet.Parameters.clbTableEnabled = false;
            //SkipNet.Parameters.aggressiveFixups = true;

            //Log.SetCategoryFilterLevel(SkipNet.Node.CATEGORY, LogLvl.Med);
            //Log.SetCategoryFilterLevel(SkipNet.JoinRequest.CATEGORY, LogLvl.Disabled);
            //Log.SetCategoryFilterLevel(SkipNet.CLBTable.CATEGORY, LogLvl.Lo);
            appEnv.envSetParameterDefaults();

        }

        public static void SetRandomSeed(int seed) 
        {
            random = new Random(seed);
        }

        public static void PostProcessParameters() 
        {

            Debug.Assert(!(doParallelCreationAndDestructionTest && doCreateOneAndFailHardTest), 
                    "You can't do both the ParallelCreationAndDestructionTest and the CreateOneAndFailHardTest;\r\n" +
                    "neither one resets state from within the experiment.");
                    
            SkipNet.Parameters.neighbourPingPeriod  = periodOfPinging;

            if(doTest_1_Cluster) 
            {
                Log.SetCategoryFilterLevel("SLOG", LogLvl.Hi);
                Log.SetCategoryFilterLevel("stat", LogLvl.Hi);
            }
            if (doTest_1_Sim) 
            {
                Log.SetCategoryFilterLevel("stat", LogLvl.Hi);
            }
            if (doTest_2_Sim) {
//                Log.SetCategoryFilterLevel("tlog", LogLvl.Disabled);
//                Log.SetCategoryFilterLevel("Tlog", LogLvl.Disabled);
                Log.SetCategoryFilterLevel("flow-action", LogLvl.Disabled);
            }
            if (doTest_3) {
                Log.SetCategoryFilterLevel("flow-action", LogLvl.Disabled);
            }

            if (doTestAutoRepair) 
            {
                SkipNet.Parameters.leafSetHalfSize = 2;
                DLC.DLCC.DebugOutput = true;
            }
            if (CommonParameters.seed == -1) 
            {
                CommonParameters.seed = 19600322;
            }
            if (DebugOutput) 
            {
                Log.SetCategoryFilterLevel("dbg", LogLvl.Hi);
                Log.SetCategoryFilterLevel(SkipNet.RoutingTable.CATEGORY, LogLvl.Disabled);
            } else {
                Log.SetCategoryFilterLevel("dbg", LogLvl.Disabled);
                Log.SetCategoryFilterLevel(SkipNet.RoutingTable.CATEGORY, LogLvl.Disabled);
            } 
            if (doLogAppFlow) 
            {
                Log.SetCategoryFilterLevel("flow", LogLvl.Lo);
            }
            if (doStatLogging) 
            {
                Log.SetCategoryFilterLevel("SLOG", LogLvl.Lo);
                Log.SetCategoryFilterLevel("stat", LogLvl.Lo);
            } 
            if (doNonLoadStatLoggingOnly) 
            {
                Log.SetCategoryFilterLevel("stat", LogLvl.Lo);
            } 
        }

        /// <summary>
        /// Each overlay has various configuration parameters. We set them here.
        /// </summary>
        public static void SetOverlayParameters(bool doRandom) 
        {
            SkipNet.Parameters.recordMessagePath = false;
            if (doRandom) SkipNet.Parameters.SetSeed(CommonParameters.seed);
        }

        public class DummyApplication: Application 
        {

            // Need to keep a copy of this here so that the garbage collector
            // doesn't get rid of it.
            ControlFlow.SimpleControlFlow cf;

            public DummyApplication(ControlFlow.SimpleControlFlow cf) 
            {
                this.cf = cf;
            }

            public override void NewNeighbour(NodeIdAddressBind id) {}
            public override void NeighbourRemoved(NodeIdAddressBind id, bool failure) {}
        }

        public static void InitNode(SkipNet.Node node, int nodeIndex, SendMessage s, LocalEvent l) 
        {
            // Need to create an instance of SimpleControlFlow on each node so that
            // exit and failNode will work.
            ControlFlow.SimpleControlFlow myCf = new ControlFlow.SimpleControlFlow(node, exEnv, quitImmediately);
            // Store myCf in the application object so that it won't get GC'ed
            DummyApplication da = new DummyApplication(myCf);
            node.RegisterApp(da, typeof(DummyApplication));

            // If this is the master control node then record some useful 
            // information for App.
            if (nodeIndex == 0) 
            {
                myServer = node;
                cf = myCf;
            }

            MeasureFUSEApplication fuseApp = new MeasureFUSEApplication(node, nodeIndex, s, l);
            node.RegisterApp(fuseApp, typeof(MeasureFUSEApplication));
            DLC.DLCC dlc = new DLC.DLCC(l);
            dlc.DLCInit(node, s, l);
            if (!debugDlc) 
            {
                DLC.DLCC.DebugOutput = false; // the default was true for some reason
            } else {
                DLC.DLCC.DebugOutput = true; // the default was true for some reason
                // Log.SetCategoryFilterLevel("dbg", LogLvl.Lo); // need this to debug the ping hashes
                // Log.SetCategoryFilterLevel(SkipNet.RoutingTable.CATEGORY, LogLvl.Lo);
            }
            node.RegisterApp(dlc , typeof( DLC.DLCC));
        }
        public static void UseNode(SkipNet.Node node, int nodeIndex) 
        {
            MeasureFUSEApplication fuseApp = (MeasureFUSEApplication) 
                node.GetApp(typeof(MeasureFUSEApplication));
            if (doImmediateLogging && !doSilentStatLogging) fuseApp.init(FUSEgroupSize);
            if (doImmediateLogging && doSilentStatLogging) fuseApp.initSilent(FUSEgroupSize);
        }


        public static void PrintStats() 
        {
            appEnv.envPrintStats();
        }

        public static void ExitProg(int exitCode) 
        {
            if (!quitImmediately) 
            {
                Console.WriteLine();
                Console.ReadLine();
            }
            Environment.Exit(exitCode);
        }
    }


    /// <summary>
    /// Interface to be supplied by both _sim and _cluster.
    /// </summary>
    public interface ApplicationEnvironment 
    {
        void envDeactivateNode(int i);  // interface supplied by both, but these two calls are noops on cluster
        void envActivateNode(int i);
        void envPrintUsage();                   // Print environment-specific usage information
        void envPrintArgs();                    // Print environment-specific argument information
        void envPrintStats();    // Print environment-specific stats
        bool envLogDetails {get; set;}          // Turn on/off logging environment-specific details
        void envSetParameterDefaults();         // Set environment-specific parameter defaults
        void envParameterFixup();               // Reconcile environment-specific parameter settings
        //bool envRendezvousWithNode();           // Called to sequentially confirm a node's creation
        void envReset();                        // Reset environment-specific state, such as counters, etc.
        long envGetTime();                      // Return local time.
        void envExit(SkipNet.Node node);        // Exit on a particular (virtual) node.
        int envGetTotalMsgsSent();              // Return the total number of messages sent so far.
        void envEnableDroppedPackets(double DropFraction);  // Enable packet dropping with the specified drop percentage.
    }
    
        
    //================================ Messages and classes for the application ================================

    /// <summary>
    /// Needed for XML serialization of messages.
    /// </summary>
    public class SerializationConfig 
    {
        public const string MEASUREFUSE_NAMESPACE = "http://research.microsoft.com/sn/Herald/MEASUREFUSE";
        public static Type [] serializationTypes = new Type[] {
                                                                  typeof(ControlReplyMsg),
                                                                  typeof(StartParallelControlMsg),
                                                                  typeof(InitializeMFAControlMsg),
                                                                  typeof(RaiseNotificationControlMsg),
                                                                  typeof(CreateGroup_ControlReplyMsg),
                                                                  typeof(MeasureGroupCreateControlMsg)
                                                              };
    }

    /// <summary>
    /// Reply to a StartParallelControlMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("ControlReplyMsg",Namespace=SerializationConfig.MEASUREFUSE_NAMESPACE,IsNullable=true)]
    public class ControlReplyMsg : OverlayReplyMsg 
    {

        public NodeIdAddressBind Replier;  // Node sending the reply message.
        
        public ControlReplyMsg() 
        {
        }

        public ControlReplyMsg( NodeIdAddressBind replier, OverlayRequestMsg requestMsg ): base(requestMsg) 
        {
            this.Replier = replier;
        }
    }
    
    /// <summary>
    /// Message sent by the master node kicking off the parallel experiments.
    /// </summary>
    [Serializable,
    XmlRootAttribute("InitializeMFAControlMsg",Namespace=SerializationConfig.MEASUREFUSE_NAMESPACE,IsNullable=true)]
    public class InitializeMFAControlMsg : ControlFlow.RPCRequestMsg 
    {
        public int FUSEgroupSize; 
        public NodeIdAddressBind controlNode;
        //public NodeIdAddressBind [] index_to_NodeIdAddressBind;

        public InitializeMFAControlMsg() {} // just for serializer

        /// <summary>
        /// Constructor.
        /// </summary>
        public InitializeMFAControlMsg( MessageRouting.Node localNode, ControlFlow.SimpleControlFlow cf, 
            int FUSEgroupSize): base(localNode, cf) 
        {
            this.FUSEgroupSize = FUSEgroupSize;
            controlNode = localNode.GetNodeIdAddressBind;
            //this.index_to_NodeIdAddressBind = new NodeIdAddressBind [index_to_NodeIdAddressBind.Count];
            /*for (int i = 0; i < index_to_NodeIdAddressBind.Count; i++) 
            {
                this.index_to_NodeIdAddressBind[i] = (NodeIdAddressBind) index_to_NodeIdAddressBind[i];
            }*/
        }

        /// <summary>
        /// Initialize the MeasureFUSEApplication on this node, then 
        /// tell the master node that things are great. 
        /// </summary>
        /// <param name="overlayNode"></param>
        public override void ArrivedAtDestination( MessageRouting.Node overlayNode ) 
        {
            Log.Write("dbg", LogLvl.Hi, "InitializeMFAControlMsg received", overlayNode.GetNodeIdAddressBind.GetAddress.ToString());

            MeasureFUSEApplication fuseApp = (MeasureFUSEApplication) 
                overlayNode.GetApp(typeof(MeasureFUSEApplication));
            fuseApp.init(FUSEgroupSize);

            OverlayReplyMsg replyMsg = new ControlReplyMsg(overlayNode.GetNodeIdAddressBind, this);
            overlayNode.SendReply(replyMsg, controlNode);
        }
    }
    /// <summary>
    /// Raise notification on a particular node and wait for it to respond that things worked.
    /// </summary>
    [Serializable,
    XmlRootAttribute("RaiseNotificationControlMsg",Namespace=SerializationConfig.MEASUREFUSE_NAMESPACE,IsNullable=true)]
    public class RaiseNotificationControlMsg : ControlFlow.RPCRequestMsg 
    {
        public string dlccid;
        public RaiseNotificationControlMsg() {} // just for serializer
        public RaiseNotificationControlMsg( MessageRouting.Node localNode, ControlFlow.SimpleControlFlow cf, 
            string dlccid): base(localNode, cf) 
        {
            this.dlccid = dlccid;
        }
        public override void ArrivedAtDestination( MessageRouting.Node overlayNode ) 
        {
            Log.Write("dbg", LogLvl.Hi, "RaiseNotificationControlMsg received", overlayNode.GetNodeIdAddressBind.GetAddress.ToString());

            DLC.DLCC dlc = (DLC.DLCC) overlayNode.GetApp(typeof(DLC.DLCC));
            bool haveId = dlc.RaiseException(dlccid);
            Log.Write("dbg", LogLvl.Hi, "...but did I have the id? {0}", haveId);
        
            OverlayReplyMsg replyMsg = new ControlReplyMsg(overlayNode.GetNodeIdAddressBind, this);
            overlayNode.SendReply(replyMsg, this.Requestor);
        }
    }

    /// <summary>
    /// Message sent by the master node to create a single group and measure the latency.
    /// </summary>
    [Serializable,
    XmlRootAttribute("MeasureGroupCreateControlMsg",Namespace=SerializationConfig.MEASUREFUSE_NAMESPACE,
        IsNullable=true)]
    public class MeasureGroupCreateControlMsg : ControlFlow.RPCRequestMsg 
    {
        public NodeIdAddressBind [] members;    
        [XmlArray( ElementName="Members",
             Namespace=SerializationConfig.MEASUREFUSE_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="MemberElem", Type=typeof(NodeIdAddressBind),
             IsNullable=true, Namespace=SerializationConfig.MEASUREFUSE_NAMESPACE)]
        public NodeIdAddressBind[] Members 
        {
            get { return members; }
            set { members = value; }
        }
       
        public MeasureGroupCreateControlMsg() {} // xml serializers! gotta love em.
        public MeasureGroupCreateControlMsg(MessageRouting.Node localNode, ControlFlow.SimpleControlFlow cf, 
            NodeIdAddressBind [] members) : base(localNode, cf, (13 * 60 * 1000)) { // 2 * path timeout + 1 minute
            // convention is that 1st member is the node from whom we initiate FUSE group creation
            this.members = members;
        }

        public override void ArrivedAtDestination( MessageRouting.Node overlayNode )  
        {
            Log.Write("dbg", LogLvl.Hi, "MeasureGroupCreateControlMsg received", overlayNode.GetNodeIdAddressBind.GetAddress.ToString());
            Debug.Assert(overlayNode.GetNodeIdAddressBind.GetAddress.ToString() == 
                members[0].GetAddress.ToString(), "MeasureGroupCreateControlMsg arrived at wrong destination");

            MeasureFUSEApplication fuseApp = (MeasureFUSEApplication) 
                overlayNode.GetApp(typeof(MeasureFUSEApplication));
            // log latency using TLog, not by passing it back in the message
            // pass in "this" because CreateGroupAndReturn has to initiate the 
            // response saying that the group was created
            fuseApp.CreateGroupAndReturn(members, this); 
        }
    }
    
        
    /// <summary>
    /// Message sent by the master node kicking off the parallel experiments.
    /// </summary>
    [Serializable,
    XmlRootAttribute("StartParallelControlMsg",Namespace=SerializationConfig.MEASUREFUSE_NAMESPACE,IsNullable=true)]
    public class StartParallelControlMsg : ControlFlow.RPCRequestMsg 
    {
        public int FUSEgroupsPerNode;
        public long msecsToSpendQuiesced;
        public long timeToStartCreatingGroups;
        public bool failEarly;
        public NodeIdAddressBind controlNode;

        public StartParallelControlMsg() {}
        /// <summary>
        /// Constructor.
        /// </summary>
        public StartParallelControlMsg( MessageRouting.Node localNode, ControlFlow.SimpleControlFlow cf, 
            int FUSEgroupsPerNode, long timeToStartCreatingGroups, long msecsToSpendQuiesced, bool failEarly): base(localNode, cf) 
        {
            this.FUSEgroupsPerNode = FUSEgroupsPerNode;
            this.msecsToSpendQuiesced = msecsToSpendQuiesced;
            this.timeToStartCreatingGroups = timeToStartCreatingGroups;
            this.failEarly = failEarly;
            controlNode = localNode.GetNodeIdAddressBind;
        }

        /// <summary>
        /// Kick off parallel experiments on this node, then 
        /// tell the master node that things are great. 
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node overlayNode ) 
        {
            Log.Write("dbg", LogLvl.Hi, "StartParallelControlMsg received", overlayNode.GetNodeIdAddressBind.GetAddress.ToString());

            MeasureFUSEApplication fuseApp = (MeasureFUSEApplication) 
                overlayNode.GetApp(typeof(MeasureFUSEApplication));
            fuseApp.initParallel(FUSEgroupsPerNode, msecsToSpendQuiesced, 
                 timeToStartCreatingGroups, failEarly);

            OverlayReplyMsg replyMsg = new ControlReplyMsg(overlayNode.GetNodeIdAddressBind, this);
            overlayNode.SendReply(replyMsg, controlNode);
        }
    }

    [Serializable,
    XmlRootAttribute("CreateGroup_ControlReplyMsg",Namespace=SerializationConfig.MEASUREFUSE_NAMESPACE,IsNullable=true)]
    public class CreateGroup_ControlReplyMsg : ControlReplyMsg
    {
        public string dlccid;
        public long createLatency;
        public CreateGroup_ControlReplyMsg() {}
        public CreateGroup_ControlReplyMsg(NodeIdAddressBind replier, OverlayRequestMsg msg,  string dlccid, long createLatency)
            : base(replier, msg) 
        {
            this.dlccid = dlccid;
            this.createLatency = createLatency;
        }
        public override void ArrivedAtDestination( MessageRouting.Node overlayNode ) 
        {
            base.ArrivedAtDestination(overlayNode);
        }
    }

    

    public class MeasureFUSEApplication : Application 
    {
        SkipNet.Node node;
        public int nodeIndex;
        SendMessage s;
        LocalEvent l;
        bool failEarly;
        bool logFailuresAssumingTimeredGlobalFailure;
        int FUSEgroupSize, FUSEgroupsPerNode;
        long msecsToSpendQuiesced, timeToStartCreatingGroups;
        //NodeIdAddressBind [] index_to_NodeIdAddressBind; // a global name service!

        string [] FUSEgroupIDs;
        public int numFUSEgroupsEstablished;

        // these next two functions are not ones we want to use
        public override void NeighbourRemoved(MessageRouting.NodeIdAddressBind n, bool b) {}
        public override void NewNeighbour(MessageRouting.NodeIdAddressBind n) {}

        // stats for reporting latency in creation time
        public long [] dlcGroupCreateCallStartTime;
        public long [] dlcGroupCreateCallReturnTime;
        public bool [] dlcGroupCreateCallSuccess;
        long timeWhenGlobalFailuresOccur;

        public MeasureFUSEApplication(SkipNet.Node node, int nodeIndex, SendMessage s, 
            LocalEvent l)
        {
            this.node = node;
            this.nodeIndex = nodeIndex;
            this.s = s;
            this.l = l;
            numFUSEgroupsEstablished = 0;
            logFailuresAssumingTimeredGlobalFailure = false;
         }

        // Initialize MeasureFUSEApp with state that's useful for each experiment.
        public void init(int FUSEgroupSize) 
        {
            this.FUSEgroupSize = FUSEgroupSize;  
            //this.index_to_NodeIdAddressBind = index_to_NodeIdAddressBind;
            //periodically log starting now
            MessageRouting.Timer periodicMessageStatLogging = new periodicMessageStatLogging(10*1000, App.periodOfStatLogging, this);
            node.SetTimer(periodicMessageStatLogging);

            DLC.DLCC dlc = (DLC.DLCC) node.GetApp(typeof(DLC.DLCC));
            LoggingFailureHandler failureCallback = new LoggingFailureHandler(this);
            dlc.RegisterFailureNotifier(new DLC.DLCCFailureNotification (failureCallback.LivenessIDFailure));
        }

        public void initSilent(int FUSEgroupSize) // avoid extra messages during run
        {
            this.FUSEgroupSize = FUSEgroupSize;  

            DLC.DLCC dlc = (DLC.DLCC) node.GetApp(typeof(DLC.DLCC));
            LoggingFailureHandler failureCallback = new LoggingFailureHandler(this);
            dlc.RegisterFailureNotifier(new DLC.DLCCFailureNotification (failureCallback.LivenessIDFailure));
        }

        public void CreateGroupAndReturn(NodeIdAddressBind[] members, MeasureGroupCreateControlMsg msg) // members _should_ include this ndoe
        {
            Debug.Assert(members[0].GetAddress.ToString() == node.GetNodeIdAddressBind.GetAddress.ToString());
            
            DLC.DLCC dlc = (DLC.DLCC) node.GetApp(typeof(DLC.DLCC));
            long time = l.GetTime();
            CreateGroupAndReturnContinuation continuation = new CreateGroupAndReturnContinuation (dlc, this, msg, time);
            dlc.CreateGroup(members, continuation);
        }

        public class CreateGroupAndReturnContinuation : DLC.DLCCContinuation 
        {
            DLC.DLCC dlc;
            MeasureFUSEApplication mfa;
            MeasureGroupCreateControlMsg msg;
            long timeCreateWasCalled;

            public CreateGroupAndReturnContinuation(DLC.DLCC dlc, MeasureFUSEApplication mfa, 
                MeasureGroupCreateControlMsg msg, long timeCreateWasCalled)
            {
                this.dlc = dlc;
                this.mfa = mfa;
                this.msg = msg;
                this.timeCreateWasCalled = timeCreateWasCalled;
            }
		
            public override void Execute( MessageRouting.Node localNode )
            {
                if (DLC.DLCC.DebugOutput) 
                {
                    Console.WriteLine("LivenessApplication path establishment called in node {0} for {1} with message {2}",
                        localNode.GetNodeIdAddressBind.GetAddress,
                        this.DLCCId,
                        this.ErrMsg);
                }
                long currentTime = mfa.l.GetTime();
                int timeDiff = (int) (currentTime - timeCreateWasCalled);
                if (this.DLCCId != null) 
                {
                    // TLOG --> we succeeded with the given latency
                    Log.Write("stat", LogLvl.Hi, "DLC group serial creation succeeded for group {0} with latency {1}",
                        this.DLCCId, timeDiff);                  
                } 
                else 
                {
                    // TLOG --> we failed with said latency
                    Log.Write("stat", LogLvl.Hi, "DLC group serial creation FAILED latency {0}",
                        timeDiff);
                }
                // signal that this group create finished, can now try to signal it
                OverlayReplyMsg replyMsg = new CreateGroup_ControlReplyMsg(localNode.GetNodeIdAddressBind, 
                    msg, this.DLCCId, timeDiff); // send along dlccid in order for control node to name the dlc group
                localNode.SendReply(replyMsg, msg.Requestor);
            }
        }

        // Start the timered events that will 1. create FUSE groups, 2. quiesce, 3. fail FUSE groups
        public void initParallel(int FUSEgroupsPerNode, long msecsToSpendQuiesced, 
            long timeToStartCreatingGroups, bool failEarly)
        {
            this.FUSEgroupsPerNode = FUSEgroupsPerNode;
            this.msecsToSpendQuiesced = msecsToSpendQuiesced;
            this.timeToStartCreatingGroups = timeToStartCreatingGroups;
            this.failEarly = failEarly;
            logFailuresAssumingTimeredGlobalFailure = true;

            FUSEgroupIDs = new string[FUSEgroupsPerNode];
            dlcGroupCreateCallStartTime = new long[FUSEgroupsPerNode];
            dlcGroupCreateCallReturnTime = new long[FUSEgroupsPerNode];
            dlcGroupCreateCallSuccess = new bool[FUSEgroupsPerNode];
            timeWhenGlobalFailuresOccur = (timeToStartCreatingGroups + msecsToSpendQuiesced);

            // start timered events here
            int delay = (int) (timeToStartCreatingGroups - l.GetTime());
            Log.Write("dbg", LogLvl.Hi, "It's now {0}, and the time to start is {1}", 
                l.GetTime(), timeToStartCreatingGroups);
            Debug.Assert(delay > 0, "BARF! - I missed the 'start creating groups deadline'!");
            MessageRouting.Timer parallelGroupCreationTimer = new ParallelGroupCreationTimer(delay, this);            
            //if (nodeIndex == 0) // added during debugging -- remove later
            node.SetTimer(parallelGroupCreationTimer);
            int delay2 = (int) (timeToStartCreatingGroups - l.GetTime() + msecsToSpendQuiesced);
            Timer groupFailureTimer = new GroupFailureTimer(delay2, this);
            node.SetTimer(groupFailureTimer);
        }

 
        public class GroupFailureTimer : MessageRouting.Timer 
        {
            MeasureFUSEApplication mfa;
            public GroupFailureTimer(int delay, MeasureFUSEApplication mfa) : base(delay) 
            {
                this.mfa = mfa;
            }
            public override void Execute( MessageRouting.Node node ) 
            {
                // do dlc group destruction here;
                // might want to do gentle destruction (i.e., just raising an exception) in the next iteration
                if(mfa.failEarly) 
                {
                    Log.Write("dbg", LogLvl.Hi, "dying");
                    TLog.WriteGlobal(node, "TLOG-GLOBAL", LogLvl.Hi, "node {0} failing", 
                        node.GetPhysicalAddr.ToString());
                    App.appEnv.envExit(mfa.node);
                }
            }
  
        }

        // this is our failure handler - it just does logging
        public class LoggingFailureHandler 
        {
            MeasureFUSEApplication mfa;
            public LoggingFailureHandler(MeasureFUSEApplication mfa) 
            { 
                this.mfa = mfa;
            }

            public void LivenessIDFailure(Node localNode, string dlccId, string errMsg)
            {
                if (DLC.DLCC.DebugOutput) 
                {
                    Console.WriteLine("LivenessContinuation.LivenessIDFailure called in node {0} for {1} with message {2}",
                        localNode.GetNodeIdAddressBind.GetAddress,
                        dlccId,
                        errMsg
                        );
                }
                // do the logging here for when the failure was seen
                Log.Write("dbg", LogLvl.Hi, "saw notification for FUSE group {0}, errMsg {1}", 
                    dlccId, errMsg);
                if (mfa.logFailuresAssumingTimeredGlobalFailure) 
                {
                    Log.Write("stat", LogLvl.Hi, "parallel: notification for FUSE group {0} had latency {1} (using local clock)",
                        dlccId, App.appEnv.envGetTime() - mfa.timeWhenGlobalFailuresOccur);
                    TLog.WriteGlobal(localNode, "TLOG-GLOBAL", LogLvl.Hi, "parallel: notification for FUSE group {0} occurred", 
                       dlccId);
                } 
                else 
                {
                    Log.Write("stat", LogLvl.Hi, "serial: notification for FUSE group {0} received at time {1} (using local clock)",
                        dlccId, App.appEnv.envGetTime());
                    TLog.WriteGlobal(localNode, "TLOG-GLOBAL", LogLvl.Hi, "serial: notification for FUSE group {0} received", dlccId);
                }
            }
        } 
    
        public class ParallelGroupCreationTimer : MessageRouting.Timer 
        {
            MeasureFUSEApplication mfa;
            
            public ParallelGroupCreationTimer(int delay, MeasureFUSEApplication mfa) : base(delay) 
            {
                this.mfa = mfa;
            }
            public override void Execute( MessageRouting.Node node ) 
            {
                DLC.DLCC dlc = (DLC.DLCC) node.GetApp(typeof(DLC.DLCC));
                
                Log.Write("dbg", LogLvl.Hi, "My address is {0}", 
                    mfa.node.NodeIdAddressBind.GetAddress.ToString());

                for (int i = 0; i < mfa.FUSEgroupsPerNode; i++) 
                {
                    NodeIdAddressBind [] FUSEgroupMembers = new NodeIdAddressBind [mfa.FUSEgroupSize];
                    NodeIdAddressBind myAddr = node.GetNodeIdAddressBind;
                    FUSEgroupMembers[0] = myAddr;
                    //Debug.Assert((2*mfa.FUSEgroupSize) <= App.exEnv.NodeCount, 
                    //    "too few nodes in overlay for easily creating FUSE groups of this size"); 
                    Debug.Assert(mfa.FUSEgroupSize <= App.exEnv.NodeCount, 
                        "too few nodes in overlay for FEASIBLY creating FUSE groups of this size"); 
                    string listOfFUSEgroupMembersForDebugging = mfa.nodeIndex.ToString();
                    for (int j = 1; j < mfa.FUSEgroupSize; j++) 
                    {
                        int nextIndex = App.random.Next(App.exEnv.NodeCount);
                        // check for uniqueness
                        // this next call should blow up if called in a distributed setting any time
                        // before we implement a global name service from ClusterNodeTable.txt
                        NodeIdAddressBind nextAddr = App.exEnv.GetNode(nextIndex).GetNodeIdAddressBind; 
                        bool unique = true;
                        for (int k = 0; k < j; k++) 
                        {
                            if (FUSEgroupMembers[k].GetAddress == nextAddr.GetAddress) unique = false;
                        }
                        if (unique) 
                        {
                            FUSEgroupMembers[j] = nextAddr;
                            listOfFUSEgroupMembersForDebugging += (" " + nextIndex.ToString());
                        } 
                        else 
                        {
                            j--;
                        }
                    }
                    Log.Write("dbg", LogLvl.Hi, "Creating FUSE group {0} with membership {1}", 
                        i, listOfFUSEgroupMembersForDebugging);
                    ParallelGroupCreationContinuation continuation = new ParallelGroupCreationContinuation (dlc, mfa, i);
                    mfa.dlcGroupCreateCallStartTime[i] = App.appEnv.envGetTime();
                    dlc.CreateGroup(FUSEgroupMembers, continuation);
                }
            }
        }
  
        public class periodicMessageStatLogging : MessageRouting.Timer 
        {
            MeasureFUSEApplication mfa;
            int counter = 0; // making this static was causing some ndoes to be excluded in the simulator

            public periodicMessageStatLogging(int delay, int period, 
                                              MeasureFUSEApplication mfa) : base(delay, period) 
            {
                this.mfa = mfa;
            }

            public override void Execute( MessageRouting.Node node ) 
            {
                Log.logCounters(node);
                counter++;
                if (App.doImmediateLoggingAndLogRTable || App.doLogRTable) {
                    // rather than print the r-table every five seconds, print it
                    // every 30 seconds
                    if ((counter % 6) == 0) {
                        mfa.node.DumpRoutingTable();
                        Log.Write("dbg", LogLvl.Hi, "RTABLE TOTAL SIZE = {0}", mfa.node.GetAllNeighbours().Count);
                    }
                }
            }  
        }

        // this contains the callback that adds the DLCid to our list of DLCid's 
        public class ParallelGroupCreationContinuation : DLC.DLCCContinuation 
        {
            DLC.DLCC dlc;
            MeasureFUSEApplication mfa;
            int indexOfDLCcreatedHere;

            public ParallelGroupCreationContinuation(DLC.DLCC dlc, MeasureFUSEApplication mfa, int indexOfDLCcreatedHere)
            {
                this.dlc = dlc;
                this.mfa = mfa;
                this.indexOfDLCcreatedHere = indexOfDLCcreatedHere;
            }
		
            public override void Execute( MessageRouting.Node localNode )
            {
                if (DLC.DLCC.DebugOutput) 
                {
                    Console.WriteLine("LivenessApplication path establishment called in node {0} for {1} with message {2}",
                                                      localNode.GetNodeIdAddressBind.GetAddress,
                                                      this.DLCCId,
                                                      this.ErrMsg);
                }
                if (this.DLCCId != null) 
                {
                    // 8-27-03 changing the failure handler to be global to 
                    // all DLC's, not instantiated on a per-dlc basis
                    // dlc.RegisterFailureIDNotifier (new DLC.DLCCFailureNotification (LivenessIDFailure), this.DLCCId);
                    mfa.FUSEgroupIDs[mfa.numFUSEgroupsEstablished] = this.DLCCId;
                    mfa.numFUSEgroupsEstablished++;
                    mfa.dlcGroupCreateCallReturnTime[indexOfDLCcreatedHere] = App.appEnv.envGetTime();
                    mfa.dlcGroupCreateCallSuccess[indexOfDLCcreatedHere] = true;
                    Log.Write("stat", LogLvl.Hi, "DLC group create latency = {0}", 
                        (mfa.dlcGroupCreateCallReturnTime[indexOfDLCcreatedHere] - 
                         mfa.dlcGroupCreateCallStartTime[indexOfDLCcreatedHere]));
                    // ouput the total number of nodes in this FUSE group (Members + Delegates)
                    ArrayList bagOfIDs = (ArrayList) dlc.dlccids_to_nodeIDAddressBindsOfGroupMembersPlusDelegates[DLCCId];
                    // unique-ify by sticking in a hashtable
                    Hashtable h = new Hashtable();
                    foreach (NodeIdAddressBind n in bagOfIDs) 
                    {
                        h[n] = 1;
                    }
                    Log.Write("stat", LogLvl.Hi, "FUSE group {0} has {1} members+delegates",
                        this.DLCCId, h.Keys.Count);
                    string s = "";
                    foreach (NodeIdAddressBind b in h.Keys)
                        s += b.ToString() +  " ";
                    Log.Write("dbg", LogLvl.Hi, "The members are: {0}", s);
                } 
                else 
                {
                    mfa.dlcGroupCreateCallReturnTime[indexOfDLCcreatedHere] = App.appEnv.envGetTime();
                    mfa.dlcGroupCreateCallSuccess[indexOfDLCcreatedHere] = false;
                    Log.Write("stat", LogLvl.Hi, "DLC group FAILED create latency = {0}", 
                        (mfa.dlcGroupCreateCallReturnTime[indexOfDLCcreatedHere] - 
                         mfa.dlcGroupCreateCallStartTime[indexOfDLCcreatedHere]));
                }
            }
         }
    }
}
