//
// Measuyret_sim -- RPC test application, simulator version
//

using System;
using System.Collections;
using System.Diagnostics;
using Application = MessageRouting.Application;
using SendMessage = MessageRouting.SendMessage;
using LocalEvent = MessageRouting.LocalEvent;
using Topology = MessageRouting.Topology;
using NodePhysicalAddress = MessageRouting.NodePhysicalAddress;
using SkipNet=Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Common;
using Microsoft.Research.Herald.Logger;
using Microsoft.Research.Herald.TLog;
using Microsoft.Research.Herald.DLC;
using Simulator;
using Microsoft.Research.Herald.SimLib;

namespace MeasureFUSE
{
    class Test: CheckIgnore, SimNodesClient, ApplicationEnvironment 
    {
        static long start;

        static int transitTransitBand = 250000;
        static int transitStubBand = 5625;
        static int stubStubBand = 5625;

        static SimInterface s 
        {
            get { return SimNodes.s; } // Allow use of just "s" instead of "SimNodes.s"
        }

        //================================ CheckIgnore functions ================================

        public bool IgnoreThis(object msg) 
        {
            return ((MessageRouting.RouteMsg) msg).IsBackground();
        }

        //================================ SimNodesClient functions ================================

        public void UseOrgs(Organization[] sortedOrgsList) 
        {
            // No special processing after creating the orgs
        }

        static Hashtable indexToNodeMap = new Hashtable();

        public void InitNode(SkipNet.Node node, int nodeIndex) 
        {
            App.InitNode(node, nodeIndex, s, s);
            // JD 9/12/03 - for the purpose of later being able to kill/revive these nodes (i.e., churn)
            // keep a hashtable around of poitners to them
            indexToNodeMap[nodeIndex] = node;
        }

        public void UseNode(SkipNet.Node node, int nodeIndex) 
        {
            // Begin periodic tasks
            node.ActivateNode(true);
            s.ProcessUntilIgnoreEventsOnly();
            App.UseNode(node, nodeIndex);
        }

        public void Terminate(int exitCode) 
        {
            App.ExitProg(exitCode);
        }

        //================================ ApplicationEnvironment functions ================================

        /// <summary>
        /// Tell another node to disappear (noop if not in simulator)
        /// </summary>
        public void envDeactivateNode(int nodeIndex) 
        {
            Debug.Assert(nodeIndex != 0, "0 is the seed node - what were you thinking failing it?");
            SkipNet.Node node = (SkipNet.Node) indexToNodeMap[nodeIndex];
            node.Failed = true;
        }
        /// <summary>
        /// Tell another node to re-appear (noop if not in simulator)
        /// This resurrection is not semantically clean. For example, some state of the node's may
        /// be left around, whlie other state is not. Timers that were set to fire while the node was
        /// failed will have been garbage-collected, timers set to fire far enough in the future that
        /// the node has been resurrected will still be on the queue. Additionally, we have to clean up
        /// the application state one app at a time.
        /// </summary>
        public void envActivateNode(int nodeIndex) 
        {
            SkipNet.Node node = (SkipNet.Node) indexToNodeMap[nodeIndex];
            SkipNet.Node seedNode = (SkipNet.Node) indexToNodeMap[0];
            node.Failed = false;
            node.InitNode(s,s,s.GetTopology(),node.GetPhysicalAddr,seedNode.GetPhysicalAddr);
            node.ActivateNode(false);
            DLCC dlc = (DLCC) node.GetApp(typeof(DLCC));
            dlc.Resurrect();
        }


        /// <summary>
        /// Print environment-specific program usage information
        /// </summary>
        public void envPrintUsage() 
        {
            SimLibParameters.PrintUsage();
        }

        /// <summary>
        /// Print environment-specific argument information
        /// </summary>
        public void envPrintArgs() 
        {
            SimLibParameters.PrintArgs();
        }

        /// <summary>
        /// Turn on/off logging environment-specific details
        /// </summary>
        public bool envLogDetails 
        {
            get {return Organizations.DebugOutput;}
            set {Organizations.DebugOutput = value;}
        }

        /// <summary>
        /// Set environment-specific parameter defaults
        /// </summary>
        public void envSetParameterDefaults() 
        {
            SimLibParameters.numOrgs = 10;
        }

        /// <summary>
        /// Reconcile environment-specific parameter settings
        /// </summary>
        public void envParameterFixup() 
        {
            if (SimLibParameters.nodes < SimLibParameters.numOrgs) 
            {
                Console.WriteLine("WARNING: Setting numOrgs from {0} to {1}.", SimLibParameters.numOrgs, SimLibParameters.nodes);
                SimLibParameters.numOrgs = SimLibParameters.nodes;
            }
        }

        /// <summary>
        /// Print environment-specific stats
        /// </summary>
        public void envPrintStats() 
        {
            // Nothing to print.
        }

        /// <summary>
        /// Reset environment-specific state, such as counters, etc.
        /// </summary>
        public void envReset() 
        {
            s.GetNetTopology.ResetStats();
        }

        /// <summary>
        /// Return local time.
        /// </summary>
        public long envGetTime() 
        {
            return s.GetTime();
        }

        /// <summary>
        /// Exit simulated node
        /// </summary>
        public void envExit(SkipNet.Node node) 
        {
            node.Failed = true;
        }

        // Create a single node, returning true if done creating nodes
        public bool envRendezvousWithNode() 
        {
            bool done = SimNodes.CreateNode();
            if (done) 
            {
                long seconds = (DateTime.Now.Ticks - start) / 10000000;
                Console.WriteLine("Elapsed time: {0}:{1:D2}", seconds / 60, seconds % 60);
            }

            return done;
        }

        public int envGetTotalMsgsSent() {
            int totalMsgs = 0;
            for (int i = 0; i < App.exEnv.NodeCount; i++) {
                totalMsgs += (int) s.GetNumMessagesSentToWire(App.exEnv.GetNode(i));
            }
            return totalMsgs;
        }

        public void envEnableDroppedPackets(double dropFraction) {
            Simulator.SimParameters.DropMessages = true;
            Simulator.SimParameters.DropFraction = dropFraction;
        }

        //================================ Environment-specific support functions ================================

        //================================ Simulator environment main program ================================

        /// <summary>
        /// Main program for simulator version of RPCTest
        /// </summary>
        [STAThread]
        public static int Main(string[] args) 
        {
            start = DateTime.Now.Ticks;

            Test test = new Test();
            
            App.appEnv = test;
            App.exEnv = new SimExecutionEnvironment();

            App.SetParameterDefaults();

            // Process the command line options
            int resume_at = 0;
            resume_at = SimLibParameters.ParseArgs(args, resume_at);
            if (resume_at < 0) App.Usage();
            resume_at = CommonParameters.ParseArgs(args, resume_at);
            if (resume_at < 0) App.Usage();
            App.GetOptions(args, resume_at);

            App.PostProcessParameters();

            NetTopology.SetRandomSeed(CommonParameters.seed);
            ZipfSelect.SetRandomSeed(CommonParameters.seed);
            MeasureFUSE.App.SetRandomSeed(CommonParameters.seed);

            SimLibParameters.SetParameters();
            CommonParameters.SetParameters();

            App.PrintExperimentParameters();

            App.SetOverlayParameters(true);

            long seconds = (DateTime.Now.Ticks - start) / 10000000;
            Console.WriteLine("Elapsed time: {0}:{1:D2}", seconds / 60, seconds % 60);
            Organizations orgs = SimNodes.InitNodes(test, test, SimLibParameters.numOrgs, transitTransitBand, transitStubBand, stubStubBand);
            seconds = (DateTime.Now.Ticks - start) / 10000000;
            Console.WriteLine("Elapsed time: {0}:{1:D2}", seconds / 60, seconds % 60);

            Log.EnableTimeStamping(Test.s);
            TLog.Init(Test.s);

            if (App.doTest_2_Sim) {
                SimParameters.TrackMessageReceipts = true;
            }
            App.Application();
        
            App.PrintStats();

            Console.WriteLine("Time: " + ((DateTime.Now.Ticks - start)/10000000) + " seconds.");

            Console.WriteLine();
            Console.WriteLine("FUSE has been measured. :)");
            App.ExitProg(0);
            return 0;
        }
    }
}


