using System;
using System.Collections;
using System.Xml.Serialization;


/// <summary>
/// Defines the interface of applications that live on top of the overlay.
/// </summary>


namespace MessageRouting {

    /// <summary>
    /// This class defines an application that lives on top of an overlay.
    /// </summary>
    public abstract class Application {
  
        /// <summary>
        /// This method is called whenever a new neighbour arrives.
        /// </summary>
        public abstract void NewNeighbour(NodeIdAddressBind id);

        /// <summary>
        /// This method is called whenever a neighbour is removed from a routing table.
        /// </summary>
        /// <param name="id">The id of the neighbour that was removed</param>
        /// <param name="failure">
        /// If true, the neighbour was removed because it failed.
        /// If false, the neighbour was removed because it was replaced with a better entry
        /// and it has not necessarily failed.
        /// </param>
        public abstract void NeighbourRemoved(NodeIdAddressBind id, bool failure);

        // TODO: We will need additional functionality like this for CLB.

    }


    /// <summary>
    /// A message that an application can send through the overlay.
    /// </summary>
    [Serializable,
        XmlRootAttribute("OverlayMsg", Namespace=SerializationConfig.MESSAGEROUTING_NAMESPACE)]
    public abstract class OverlayMsg {

        /// <summary>
        /// Determine if this message needs to continue being routed.
        /// Derived classes should only override this if necessary. In most cases,
        /// it's safe just to return true.
        /// </summary>
        /// <returns>True=continue routing, False=drop the message</returns>
        public virtual bool CheckContinueRouting( Node localNode ) {
            return true;
        }

        /// <summary>
        /// Called when a message has arrived at its destination.
        /// Must be overriden in derived classes.
        /// </summary>
        public abstract void ArrivedAtDestination( Node localNode );

        public object Clone() {
            return this.MemberwiseClone();
        }

    }


    /// <summary>
    /// An overlay message for which the sender expects a reply message.
    /// </summary>
    [Serializable,
        XmlRootAttribute("OverlayRequestMsg", Namespace=SerializationConfig.MESSAGEROUTING_NAMESPACE)]
    public abstract class OverlayRequestMsg : OverlayMsg {
        
        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        const int       MAX_RETRIES = 3; // JD 9/2/03 --> from 3 back to 0
        // JD 9/2/03 use to set new global default which can be overridden upwards
        const int       SUGGESTED_RETRIES = 0; 
                                                


        //**********************************************************************************************
        // Serialized Members
        //**********************************************************************************************
        string                  msgId;
        NodeIdAddressBind       source;
        long                    timeoutMs;

        // If set to true, no source-state will be installed for request-reply matching
        bool                    suppressSourceState;
        
        // If set to true, no state will be installed to prevent multiple calls to CheckContinueRouting
        bool                    allowDupCCR;    
        
        // If set to true, no state will be installed to prevent multiple calls to ArrivedAtDestination
        bool                    allowDupAD;    
        

        //**********************************************************************************************
        // NonSerialized Members
        //**********************************************************************************************
        [NonSerialized]
        OverlayReplyCallback    callback;
        [NonSerialized]
        int                     maxRetries;


        //**********************************************************************************************
        // Properties
        //**********************************************************************************************

        [XmlElementAttribute("MsgId", Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public string MsgId {
            get { return msgId; }
            set { msgId = value; }
        }
        
        [XmlElementAttribute("Source", Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public NodeIdAddressBind Source {
            get { return source; }
            set { source = value; }
        }

        [XmlElementAttribute("TimeoutMs", Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public long TimeoutMs {
            get { return timeoutMs; }
            set { timeoutMs = value; }
        }

        [XmlElementAttribute("SuppressSourceState", Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public bool SuppressSourceState {
            get { return suppressSourceState; }
            set { suppressSourceState = value; }
        }
        
        [XmlElementAttribute("AllowDupCCR", Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public bool AllowDupCCR {
            get { return allowDupCCR; }
            set { allowDupCCR = value; }
        }
        
        [XmlElementAttribute("AllowDupAD", Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public bool AllowDupAD {
            get { return allowDupAD; }
            set { allowDupAD = value; }
        }
        
        [XmlIgnore]
        internal OverlayReplyCallback Callback {
            get { return callback; }
            set { callback = value; }
        }

        [XmlIgnore]
        public int MaxRetries {
            get { return maxRetries; }
            set { maxRetries = value; }
        }


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        /// <summary>
        /// Default constructor. Should be used only by the serializer.
        /// </summary>
        public OverlayRequestMsg() {}

        /// <summary>
        /// Base Constructor. Subclasses must call this constructor too.
        /// </summary>
        /// <param name="localNode">Node sending the request</param>
        /// <param name="cb">Callback to invoke when reply message arrives</param>
        /// <param name="delayMs">Timeout after which cb will be invoked if no reply message has arrived by then</param>
        /// <param name="allowDuplicateCalls">If true, allow duplicate calls to CheckContinueRouting and ArrivedAtDestination</param>
        public OverlayRequestMsg(Node localNode, OverlayReplyCallback cb, long delayMs, bool allowDuplicateCalls) {
            this.msgId = localNode.RequestReplyTable.GenerateRequestId();
            this.callback = cb;
            this.timeoutMs = delayMs;
            this.suppressSourceState = false;
            if( allowDuplicateCalls ) {
                this.allowDupAD = true;
                this.allowDupCCR = true;
            }
            this.maxRetries = SUGGESTED_RETRIES;
        }
    }
    
    
    /// <summary>
    /// A reply to an OverlayRequestMsg.
    /// </summary>
    [Serializable,
        XmlRootAttribute("OverlayReplyMsg", Namespace=SerializationConfig.MESSAGEROUTING_NAMESPACE)]
    public abstract class OverlayReplyMsg : OverlayMsg {
        
        // Serialized Members
        string                  requestId;
        
        // NonSerialized Members
        [NonSerialized]
        long                    requestTimeoutMs;
        
        [XmlElementAttribute("RequestId",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public string RequestId {
            get { return requestId; }
            set { requestId = value; }
        }

        [XmlIgnore]
        internal long RequestTimeoutMs {
            get { return requestTimeoutMs; }
        }

        /// <summary>
        /// Default Constructor. Should only be used by the serializer.
        /// </summary>
        public OverlayReplyMsg() {}
        
        /// <summary>
        /// Base Constructor. Subclasses must call this too.
        /// </summary>
        /// <param name="requestMsg">The request to which this message is a reply</param>
        public OverlayReplyMsg( OverlayRequestMsg requestMsg ) {
            requestId = requestMsg.MsgId;
            requestTimeoutMs = requestMsg.TimeoutMs;
        }
        
        /// <summary>
        /// The reply has arrived at its destination. Invoke the necessary
        /// request-reply code.
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node localNode ) {
            localNode.RequestReplyTable.ReceivedReply(this);
        }
    }
    
    
    /// <summary>
    /// Class containing a callback that is invoked when an OverlayReplyMsg arrives or times out.
    /// Users should override the continuation's Execute() method and access the 'reply' member
    /// to determine if the reply was received or not. If 'reply' is null, this indicates that
    /// a timeout occurred.
    /// </summary>
    public abstract class OverlayReplyCallback : Continuation {

        protected OverlayReplyMsg     reply;

        /// <summary>
        /// Reply is the reply message
        /// </summary>
        public OverlayReplyMsg Reply {
            get { return reply; }
            set { reply = value; }
        }
    }


    /// <summary>
    /// This class is an abstract continuation: a piece of code that is executed later by the
    /// event-loop. Users should create a subclass that overrides the Execute method.
    /// </summary>
    public abstract class Continuation : OverlayMsg {

        //**********************************************************************************************
        // Abstract Methods
        //**********************************************************************************************

        /// <summary>
        /// This method is called when the continuation fires.
        /// </summary>
        public abstract void Execute( Node localNode );

        /// <summary>
        /// An indication to the simulator whether this is a 'background' message or not.
        /// Background messages are ignored by the simulator's "ProcessUntilIgnoreEventsOnly()".
        /// </summary>
        public virtual bool IsBackground() {
            return true;
        }


        //**********************************************************************************************
        // Overridden Methods
        //**********************************************************************************************

        public override void ArrivedAtDestination( Node localNode ) {
            this.Execute( localNode );
        }
    }

    
    /// <summary>
    /// This class is an abstract timer: a piece of code that is executed in the future,
    /// possibly periodically, by the event-loop. Users should create a subclass that
    /// overrides the Execute method.
    /// </summary>
    public abstract class Timer : Continuation {

        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        long        initialDelayMs;        // Initial delay
        long        nextDelayMs;           // Delay for next execution of timer
        
        bool        periodic;
        long        periodMs;              // Steady-state period
        

        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        // Constructor for one-shot timers
        public Timer( long delayMs ) {
            this.initialDelayMs = delayMs;
            this.nextDelayMs = delayMs;
            this.periodic = false;
        }
    
        // Constructor for periodic timers
        public Timer( long initialDelayMs, long periodMs ) {
            this.initialDelayMs = initialDelayMs;
            this.nextDelayMs = initialDelayMs;
            this.periodMs = periodMs;
            this.periodic = true;
        }
    
    
        //**********************************************************************************************
        // Properties
        //**********************************************************************************************

        /// <summary>
        /// The length of time to wait before next executing the timer.
        /// </summary>
        public virtual long NextDelayMs {
            get { return nextDelayMs; }
        }

        /// <summary>
        /// To make a non-periodic timer into a periodic one, set Periodic to true (before or during
        /// the timer's execution).
        /// 
        /// To cancel a periodic or non-periodic timer before the timer's execution, use Node.CancelTimer().
        /// 
        /// To cancel a periodic timer during the timer's execution (i.e., to make it non-periodic)
        /// set Periodic to false.
        /// </summary>
        public bool Periodic {
            get { return periodic; }
            set { periodic = value; }
        }


        //**********************************************************************************************
        // Overridden Methods
        //**********************************************************************************************

        public override void ArrivedAtDestination( Node localNode ) {
            this.Execute( localNode );
            if( this.periodic ) {
                this.nextDelayMs = this.periodMs;
                localNode.SetTimer(this);
            }
        }
    }

}
