using System;
using System.Collections;
using System.Xml.Serialization;


/// <summary>
/// Defines the interface of the Message Delivery system that lies underneath the overlay.
/// </summary>


namespace MessageRouting {

    /// <summary>
    /// Defines an interface for an event queue.
    /// </summary>
    public interface LocalEvent {

        /// <summary>
        /// Insert the Message onto the event queue.
        /// </summary>
        /// <param name="ms">Delay for this many milliseconds before delivering the message</param>
        void RaiseEvent(long ms, NodeIdAddressBind me, RouteMsg Message);

        /// <summary>
        /// Remove an event from the event queue.
        /// </summary>
        /// <param name="Message">A Message acting as a "comparison key" that is equal to the
        /// message to be removed.</param>
        void RemoveEvent(RouteMsg Message);

        /// <summary>
        /// The current time, according to the event queue.
        /// </summary>
        long GetTime();

        long GetNumMessagesSentToWire(Node node); // JD 8-21-03 -- for stats
        long GetNumMessagesReceivedFromWire(Node node); // JD 8-21-03 -- for stats
    }


    /// <summary>
    /// The address of a node, as used by the message-delivery system.
    /// </summary>
    [Serializable,
        XmlRootAttribute("NodePhysicalAddress", Namespace=SerializationConfig.MESSAGEROUTING_NAMESPACE)]
    public abstract class NodePhysicalAddress {
        // Get a string representation of the address.
        public abstract string Address { get; set; }


        //**********************************************************************************************
        // For Simulation
        //**********************************************************************************************
        [XmlIgnore]
        public bool Failed {
            get {return failed;}
            set {failed = value;}
        }
        [NonSerialized]
        private bool failed = false;

        [NonSerialized]
        Node attachedNode;

        public virtual void AttachLogicalNode(Node node) {
            attachedNode = node;
        }

        public virtual Node GetLogicalNode(OverlayId oid) {
            return attachedNode;
        }

        public virtual int Port { get { return 0; } set {} }
    }


    /// <summary>
    /// Defines an interface for sending messages between nodes.
    /// </summary>
    public interface SendMessage {
        // Send a message to the destination node
        bool Send(NodeIdAddressBind source, NodeIdAddressBind destination, RouteMsg obj);

        // Send a message asynchronously to the destination node
        // The ReplyCallback is invoked when an ack is received from the destination node
        void BeginSend(NodeIdAddressBind source, NodeIdAddressBind destination, RouteMsg obj, ReplyCallback cb);
    }

    public delegate void ReplyCallback(NodeIdAddressBind destination, bool success, RouteMsg sentMsg);


    /// <summary>
    /// Defines an interface for measuring the distance between nodes.
    /// </summary>
    public interface Topology {
        // Measure the distance between a and b.
        // A distance of -1 means the node is dead. A distance > 0 means the node is alive.
        int Ping(NodePhysicalAddress a, NodePhysicalAddress b, bool disableCache);

        // Asynchronously measure the distance between a and b.
        // A distance of -1 means the node is dead. A distance > 0 means the node is alive.
        void BeginPing(NodePhysicalAddress a, NodePhysicalAddress b, bool disableCache, PingCallback cb);

        // This must return another seed node
        NodePhysicalAddress GetClosest(NodePhysicalAddress n);
    }

    public delegate void PingCallback(NodePhysicalAddress destination, int distance);



    //**********************************************************************************************
    // For Simulation
    //**********************************************************************************************
    
    public struct OverlayId {
        public int idGlobal, idLocal;

        public OverlayId(int idGlobal, int idLocal) {
            this.idGlobal = idGlobal;
            this.idLocal = idLocal;
        }
 
        public static OverlayId DefaultOverlay = new OverlayId(0,0);

        public override string ToString() {
            return "[" + this.idGlobal + "," + this.idLocal + "]";
        }
    }

    /// <summary>
    /// Abstracts the transport layer available underneath an overlay network
    /// (either a simulator or an actual transport stack).
    /// </summary>
    public interface TransportLayer {
        // Return a characterization of the underlying network topology.
        Topology GetTopology();

        // Create an overlay node on top of this physical node:
        // - Informs the transport layer that the logical node should be attached to the physical node.
        // - Calls the node's InitNode method.
        Node CreateNode(Node node, NodePhysicalAddress addr, Object nodeInitData);
    }
}