using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;


/// <summary>
/// Defines the classes that an overlay (such as Pastry or SkipNet) must implement.
/// Also provides some basic functionality that is common to all overlays.
/// </summary>


namespace MessageRouting {

    public class SerializationConfig {
        public const string MESSAGEROUTING_NAMESPACE = "http://research.microsoft.com/sn/Herald/MessageRouting";
    }

    /// <summary>
    /// Defines an overlay node.
    /// </summary>
    public abstract class Node {

        //**********************************************************************************************
        // Base Constructor
        //**********************************************************************************************

        /// <summary>
        /// Base constructor. Subclasses must call this.
        /// </summary>
        public Node() {
            this.requestReplyTable = new RequestReplyTable(this);
        }


        //**********************************************************************************************
        // Public Properties
        //**********************************************************************************************

        public abstract NodePhysicalAddress GetPhysicalAddr { get; }

        public abstract NodeIdAddressBind GetNodeIdAddressBind { get; } 


        //**********************************************************************************************
        // Initialization
        //**********************************************************************************************

        // Initialize a node.  
        public abstract void InitNode(LocalEvent local, SendMessage output, Topology top,
            NodePhysicalAddress myAddr, NodePhysicalAddress seedNodeAddr);


        //**********************************************************************************************
        // NodeId generation
        //**********************************************************************************************

        // Create a node ID based on a given string
        public abstract NodeId GenerateNodeId(string s);
    

        //**********************************************************************************************
        // Upcalls
        //**********************************************************************************************

        /// <summary>
        /// Upcall from the message-delivery system when a message arrives.
        /// </summary>
        public void RaiseEventUpcall(RouteMsg msg) {
            msgsReceived++;
            RaiseEvent(msg);
        }

        /// <summary>
        /// Upcall to the overlay implementation to handle a message from the message-delivery system.
        /// </summary>
        protected abstract void RaiseEvent(RouteMsg msg);


        //**********************************************************************************************
        // Message-routing functions
        //**********************************************************************************************

        /// <summary>
        /// Route an overlay message using the overlay routing system.
        /// The message is not routed securely.
        /// </summary>
        /// <param name="msg">The message to send</param>
        /// <param name="dest">A string representation of the destination</param>
        public void RouteMessage(OverlayMsg msg, string dest) {
            RouteMessage(msg, dest, false);
        }

        /// <summary>
        /// Route an overlay message using the overlay routing system.
        /// The message is not routed securely.
        /// </summary>
        /// <param name="msg">The message to send</param>
        /// <param name="dest">A node ID representing the destination</param>
        public void RouteMessage(OverlayMsg msg, NodeId dest) {
            RouteMessage(msg, dest, false);
        }

        /// <summary>
        /// Route an overlay message using the overlay routing system.
        /// </summary>
        /// <param name="msg">The message to send</param>
        /// <param name="dest">A string representation of the destination</param>
        /// <param name="secure">Specifies if this message should be routed securely</param>
        public abstract void RouteMessage(OverlayMsg msg, string dest, bool secure);

        /// <summary>
        /// Route an overlay message using the overlay routing system.
        /// </summary>
        /// <param name="msg">The message to send</param>
        /// <param name="dest">A node ID representing the destination</param>
        /// <param name="secure">Specifies if this message should be routed securely</param>
        public abstract void RouteMessage(OverlayMsg msg, NodeId dest, bool secure);

        /// <summary>
        /// Internal method to route a RouteMsg.
        /// Not intended to be called by applications.
        /// </summary>
        public abstract void RouteMessageInternal(RouteMsg routeMsg);

        /// <summary>
        /// Send an overlay message directly to the given destination node.
        /// </summary>
        /// <param name="msg">The message to send</param>
        /// <param name="dest">The destination node</param>
        public abstract void SendMessageDirect(OverlayMsg msg, NodeIdAddressBind dest);

        public abstract void SendBackgroundMessageDirect(OverlayMsg msg, NodeIdAddressBind dest);
        
        /// <summary>
        /// Reply to a OverlayReply message.
        /// </summary>
        public void SendReply(OverlayReplyMsg reply, NodeIdAddressBind dest) {
            requestReplyTable.AddReply(reply, dest);
            SendMessageDirect(reply, dest);
        }

        public void SendBackgroundReply(OverlayReplyMsg reply, NodeIdAddressBind dest) {
            requestReplyTable.AddReply(reply, dest);
            SendBackgroundMessageDirect(reply, dest);
        }

        /// <summary>
        /// Schedule a continuation to execute.
        /// </summary>
        public abstract void ScheduleContinuation(Continuation cont);

        /// <summary>
        /// Set a timer to execute.
        /// </summary>
        public abstract void SetTimer(Timer t);

        /// <summary>
        /// Cancel a timer that has already been set.
        /// </summary>
        public abstract void CancelTimer(Timer t);


        //**********************************************************************************************
        // Neighbour Information
        //**********************************************************************************************
        
        // Table of all neighbours. The overlay is responsible for keeping this table up-to-date.
        protected Hashtable   allNeighbours;

        /// <summary>
        /// Determine if a given node is a neighbour of this node in the overlay
        /// </summary>
        public virtual bool IsNeighbour( NodeIdAddressBind n ) {
            bool result = allNeighbours.ContainsKey(n);
            return result;
        }

        /// <summary>
        /// Get an array containing all neighbours of this node.
        /// </summary>
        /// <returns>
        /// An array of NodeIdAddressBind objects. The array contains no duplicates.
        /// The caller may modify the array.
        /// </returns>
        public virtual ArrayList GetAllNeighbours() {
            ArrayList nbrs = new ArrayList( allNeighbours.Keys );
            return nbrs;
        }


        //**********************************************************************************************
        // Request-Reply Matching
        //**********************************************************************************************

        RequestReplyTable       requestReplyTable;

        internal RequestReplyTable RequestReplyTable {
            get { return requestReplyTable; }
        }

        /// <summary>
        /// Hook that must be called by the overlay node whenever it is the source of a new message.
        /// </summary>
        /// <param name="msg">The overlay message. Must not be null.</param>
        protected void MessageSourceHook( OverlayMsg overlayMsg, RouteMsg routeMsg ) {
            // If the message is a request-reply message, let the table handle it
            Debug.Assert( overlayMsg!=null );
            if( overlayMsg is OverlayRequestMsg ) {
                requestReplyTable.HandleRequestAtSource( (OverlayRequestMsg) overlayMsg, routeMsg );
            }
        }

        /// <summary>
        /// Hook that must be called by the overlay node in order to call "CheckContinueRouting"
        /// for any OverlayMsg.
        /// </summary>
        /// <returns>True to continue routing, false to stop</returns>
        public bool MessageCCRHook( OverlayMsg overlayMsg ) {
            bool continueRouting;
            if( overlayMsg is OverlayRequestMsg ) {
                continueRouting = requestReplyTable.HandleCheckContinueRouting( (OverlayRequestMsg) overlayMsg );
            } else {
                continueRouting = overlayMsg.CheckContinueRouting(this);
            }
            return continueRouting;
        }

        /// <summary>
        /// Method that must be called by the overlay node whenever a message arrives at its destination.
        /// This method will take care of notifying the message that it has arrived.
        /// </summary>
        /// <param name="overlayMsg">The overlay message. Must not be null.</param>
        protected void MessageArrivedHook( OverlayMsg overlayMsg ) {
            Debug.Assert( overlayMsg!=null );

            // Check if the arriving message is a request message
            if( overlayMsg is OverlayRequestMsg ) {
                requestReplyTable.HandleArrivedAtDestination( (OverlayRequestMsg) overlayMsg );
            } else {
                overlayMsg.ArrivedAtDestination(this);
            }
        }
        
        
        //**********************************************************************************************
        // Applications (that live on top of the overlay)
        //**********************************************************************************************

        protected Hashtable     appTable = new Hashtable();

        /// <summary>
        /// Register an application that lives on top of the overlay
        /// </summary>
        public void RegisterApp( Application app, Object key ) {
            Debug.Assert( appTable!=null );
            appTable[ key ] = app;
        }
        
        /// <summary>
        /// Unregister an application that lives on top of the overlay
        /// </summary>
        public void UnregisterApp( Application snApp, Object key ) {
            Debug.Assert( appTable!=null );
            Debug.Assert( appTable.ContainsKey(key) );
            appTable.Remove(key);
        }

        /// <summary>
        /// Get the application with the given key
        /// </summary>
        public Application GetApp( Object key ) {
            Debug.Assert( appTable!=null );

            Application app = (Application) appTable[ key ];
            Debug.Assert( app!=null );

            return app;
        }


        //**********************************************************************************************
        // For Simulation
        //**********************************************************************************************

        // Counts the number of messages received at this node
        public int      msgsReceived=0;

        public abstract bool Failed { get; set; }

        bool active=true;
        public bool IsActive {
            get { return active; }
            set { active = value; }
        }

        // Which overlay network this node belongs to.
        public OverlayId OverlayId { 
            get {return overlayId;}
        }
        protected OverlayId overlayId = OverlayId.DefaultOverlay;
    }


    /// <summary>
    /// Callback for the Node.SendMessageDirect method
    /// </summary>
    public delegate void ReplyDirectCallback(Node localNode, NodeIdAddressBind dest, bool success, RouteMsg sentMsg);


    /// <summary>
    /// A NodeId is effectively a routing destination. Details are overlay-specific.
    /// </summary>
    [Serializable,
        XmlRootAttribute("NodeId", Namespace=SerializationConfig.MESSAGEROUTING_NAMESPACE)]
    public abstract class NodeId {
        // Empty
    }


    /// <summary>
    /// A NodeIdAddressBind ties a NodeId to a NodePhysicalAddress
    /// </summary>
    [Serializable,
        XmlRootAttribute("NodeIdAddressBind", Namespace=SerializationConfig.MESSAGEROUTING_NAMESPACE)]
    public abstract class NodeIdAddressBind {
        public abstract NodeId GetNodeId { get; set; }
        public abstract NodePhysicalAddress GetAddress { get; set; }
    }
    
    
    /// <summary>
    /// A routing message that is given by the overlay to the message-delivery system.
    /// </summary>
    [Serializable,
        XmlRootAttribute("RouteMsg", Namespace=SerializationConfig.MESSAGEROUTING_NAMESPACE)]
    public abstract class RouteMsg {
        /// <summary>
        /// The destination of the message
        /// </summary>
        public abstract NodeId Destination { get; set; }

        /// <summary>
        /// The source address of the message
        /// </summary>
        public abstract NodeIdAddressBind Source { get; set; }

        /// <summary>
        /// An indication to the simulator whether this is a 'background' message or not.
        /// Background messages are ignored by the simulator's "ProcessUntilIgnoreEventsOnly()".
        /// </summary>
        public abstract bool IsBackground();

        
        //**********************************************************************************************
        // For Simulation
        //**********************************************************************************************

        // For multiplexing multiple overlays on a single node
        [NonSerialized]
        private OverlayId oid = OverlayId.DefaultOverlay;
        [XmlIgnore]
        public OverlayId Overlay {
            get {return oid;}
            set {oid = value;}
        }

        // Size of the message on the wire
        public const int DefaultMsgSize = 1024;
        [NonSerialized]
        private int msgSize = DefaultMsgSize;
        [XmlIgnore]
        public int MessageSize {
            get {return msgSize;}
            set {msgSize = value;}
        }
    }

}
