using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;


/// <summary>
/// Defines classes for performing request/reply messages.
/// Basically this is a simple implementation of RPC,
/// with additional machinery for maintaining state at intermediate nodes.
/// </summary>


namespace MessageRouting {

    /// <summary>
    /// Class to implement the logic of request-reply matching.
    /// </summary>
    internal class RequestReplyTable {

        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        Node            localNode;
        int             msgCounter;     // Counter for generating "unique" ids
        Hashtable       table;          // Map from Msg ID strings to TableEntry objects

        class TableEntry {
            // The following fields are populated at all nodes on a request's routing path */
            public string              msgId;
            public bool                ccrCalled;      // Has CheckContinueRouting been called at this node?
            public bool                ccrResult;      // The result of a previous call to CCR (if any)
            public bool                adCalled;       // Has ArrivedAtDestination been called at this node?
            public OverlayReplyMsg     reply;          // A reply. Populated only at a node that replies.
            public NodeIdAddressBind   replyDest;      // The destination of the reply message.

            // The following fields are populated only at the source
            public OverlayRequestMsg   request;        // The original request message.
            public RouteMsg            routeMsg;       // The original RouteMsg.
            public int                 retryCount;     // How many times has the request been re-sent?
            public ReplyTimer          replyTimer;     // Timer for doing resends / timeouts.

            /// <summary>
            /// Constructor for use at the source node
            /// </summary>
            public TableEntry( string msgId, OverlayRequestMsg request, RouteMsg routeMsg ) {
                this.msgId = msgId;
                this.request = request;
                this.routeMsg = routeMsg;
                this.retryCount = 0;
            }
            
            /// <summary>
            /// Constructor for use at intermediate nodes
            /// </summary>
            public TableEntry( string msgId ) {
                this.msgId = msgId;
            }
        }


        //**********************************************************************************************
        // Constructor
        //**********************************************************************************************

        public RequestReplyTable( Node node ) {
            this.localNode = node;
            this.msgCounter = 0;
            this.table = new Hashtable();
        }


        //**********************************************************************************************
        // Requests
        //**********************************************************************************************

        /// <summary>
        /// Generate a new unique id for a request message.
        /// </summary>
        public string GenerateRequestId() {
            string id = String.Format( "{0}!{1}", 
                localNode.GetNodeIdAddressBind.GetNodeId, msgCounter ); 
            msgCounter++;
            return id;
        }

        /// <summary>
        /// Called by the overlay when a request message is being sent from the source.
        /// Sets up timeouts, etc.
        /// </summary>
        /// <param name="request">The request message</param>
        public void HandleRequestAtSource( OverlayRequestMsg request, RouteMsg routeMsg ) {
            Debug.Assert( request!=null );
            Debug.Assert( request.Callback!=null );
            Debug.Assert( request.MsgId!=null );

            string id = request.MsgId;

            // The 'SuppressSourceState' flag indicates not to set up source-state at this node
            if( request.SuppressSourceState ) {
                return;
            }

            // Get the current table entry (if any), or create a new one
            TableEntry entry;
            if( table.ContainsKey(id) ) {
                entry = (TableEntry) table[id];
                Debug.Assert( entry.retryCount>0 );
            } else {
                // Create a new entry and add it to the hash table
                entry = new TableEntry( id, request, routeMsg );
                table[id] = entry;
            }

            // Schedule the timer to handle retries and timeouts
            ReplyTimer timer = new ReplyTimer( request, request.TimeoutMs, routeMsg.IsBackground() );
            entry.replyTimer = timer;
            localNode.SetTimer(timer);
        }
        
        /// <summary>
        /// Called by a reply message when it arrives at the original source.
        /// </summary>
        /// <param name="reply"></param>
        public void ReceivedReply( OverlayReplyMsg reply ) {
            // Get the request's id
            string id = reply.RequestId;
            
            // Look up the request's id in the table.
            if( table.ContainsKey(id) ) {

                // Get the entry for this id
                TableEntry entry = (TableEntry) table[id];
                if( entry.request==null ) {
                    // If there is no request message, this node is not the source
                    return;
                }
                OverlayRequestMsg request = entry.request;

                // Cancel the pending timer.
                Debug.Assert( entry.replyTimer!=null );
                localNode.CancelTimer( entry.replyTimer );

                // Invoke callback with reply.
                request.Callback.Reply = reply;
                request.Callback.Execute( localNode );

                // Remove the entry from the table
                RemoveEntry( id );
            } else {
                // Request table does not contain the request id.
                // Just throw away the reply.
            }
        }

        /// <summary>
        /// This method is called by a ReplyTimer when it fires.
        /// </summary>
        /// <param name="id">The id of the message whose timer fired</param>
        internal void ReplyTimerFired( string id ) {
            // Get the table entry for this message id
            Debug.Assert( table.ContainsKey(id) );
            TableEntry entry = (TableEntry) table[id];
            Debug.Assert( entry!=null );
            OverlayRequestMsg request = entry.request;
            Debug.Assert( request!=null );

            if( entry.retryCount<request.MaxRetries ) {
                // We should retry sending this request
                entry.retryCount++;
                localNode.RouteMessageInternal( entry.routeMsg );
            } else {
                // We have already retried this request enough times.
                // This time we just fail the request.

                // Invoke the Request's callback with a null reply,
                // indicating no reply was received
                request.Callback.Reply = null;
                request.Callback.Execute( localNode );

                // Remove the request from the RequestReplyTable
                RemoveEntry( request.MsgId );   
            }
        }

        /// <summary>
        /// Called to unregister an entry from the table, either:
        ///  - when a reply is received
        ///  - when a timeout expires
        ///  - when state at intermediate nodes is garbage-collected
        /// </summary>
        /// <param name="request">The request message</param>
        internal void RemoveEntry( string id ) {
            Debug.Assert( table.ContainsKey(id) );
            table.Remove(id);
        }


        //**********************************************************************************************
        // CheckContinueRouting
        //**********************************************************************************************

        /// <summary>
        /// Called by the overlay when CheckContinueRouting is about to be called for a request message.
        /// Sets up state to avoid calling CCR multiple times.
        /// 
        /// Also, checks if this node has already replied to this message.
        /// If so, sends the reply right away.
        /// </summary>
        /// <param name="request">The request message</param>
        public bool HandleCheckContinueRouting( OverlayRequestMsg request ) {
            // If the 'AllowDupCCR' option is on, we don't need to worry about
            // duplicate calls to CheckContinueRouting().
            if( request.AllowDupCCR ) {
                return request.CheckContinueRouting( localNode );
            }
            
            // Get the request's id
            string id = request.MsgId;

            // Look up the request's id in the table.
            TableEntry entry;
            if( table.ContainsKey(id) ) {
                // Get the entry for this id
                entry = (TableEntry) table[id];
            } else {
                // Create a new entry for this id and add it to the table
                entry = new TableEntry( id );
                table[id] = entry;

                // Start a timer to garbage-collect this table entry
                EntryGCTimer timer = new EntryGCTimer( request );
                localNode.SetTimer( timer );
            }

            // Check if this node has already generated a reply to this message
            if( entry.reply!=null ) {
                // This node has already generated a reply.
                // (And, by implication, CCR has already been called.)
                // Just resend the reply.
                localNode.SendMessageDirect( entry.reply, entry.replyDest );
                
                // No need to continue routing since we already have a reply.
                return false;
            }

            // Check if CCR has already been called
            if( ! entry.ccrCalled ) {
                // If not, call it and record the result
                entry.ccrResult = request.CheckContinueRouting( localNode );
                entry.ccrCalled = true;

                if( entry.ccrResult==true ) {
                    // If a node generates a reply, it should not continue routing the request
                    Debug.Assert( entry.reply==null );
                }
            }
            
            // Return the recorded result
            return entry.ccrResult;
        }


        //**********************************************************************************************
        // ArrivedAtDestination
        //**********************************************************************************************
        
        /// <summary>
        /// Called by the overlay when ArrivedAtDestination is about to be called for a request message.
        /// Sets up state to avoid calling ArrivedAtDestination multiple times.
        /// </summary>
        public void HandleArrivedAtDestination( OverlayRequestMsg request ) {
            // If the 'AllowDupAD' option is on, we don't need to worry about
            // duplicate calls to ArrivedAtDestination().
            if( request.AllowDupAD ) {
                request.ArrivedAtDestination( localNode );
                return;
            }
            
            // Get the request's id
            string id = request.MsgId;

            // Look up the request's id in the table.

            // Note: Normally, an entry for this id will already exist.
            // But, sometimes during CLB routing, ArrivedAtDestination() is called
            // without ever calling CheckContinueRouting() at that node. 
            TableEntry entry;
            if( table.ContainsKey(id) ) {
                // Get the entry for this id
                entry = (TableEntry) table[id];
            } else {
                // Create a new entry for this id and add it to the table
                entry = new TableEntry( id );
                table[id] = entry;

                // Start a timer to garbage-collect this table entry
                EntryGCTimer timer = new EntryGCTimer( request );
                localNode.SetTimer( timer );
            }
            
            // If ArrivedAtDestination has not already been called at this node, call it
            if( ! entry.adCalled ) {
                request.ArrivedAtDestination( localNode );
                entry.adCalled = true;
            }
        }
        
        
        //**********************************************************************************************
        // Replies
        //**********************************************************************************************
           
        /// <summary>
        /// Add a reply to the table.
        /// </summary>
        public void AddReply( OverlayReplyMsg reply, NodeIdAddressBind dest ) {
            // Get the table entry for this request id (which should already exist)
            string id = reply.RequestId;
            if( table.ContainsKey(id) ) {
                // Stick this reply into the table entry
                TableEntry entry = (TableEntry) table[id];
                Debug.Assert( entry.reply==null );
                entry.reply = reply;
                entry.replyDest = dest;
            } else {
                // No entry in the table.
                // This can occur if AllowDupCCR / AllowDupAD are true
                // Don't keep track of the reply
            }

        }
    }


    /// <summary>
    /// Timer for implementing reply-timeouts.
    /// </summary>
    public class ReplyTimer : Timer {
        OverlayRequestMsg   request;
        private bool background = false;

        public ReplyTimer( OverlayRequestMsg request, long periodMs )
            : base(periodMs)
        {
            this.request = request;
        }
   
        public ReplyTimer( OverlayRequestMsg request, long periodMs, bool background )
            : base(periodMs) {
            this.request = request;
            this.background = background;
        }
     
        /// <summary>
        /// Invoked when this timer fires (i.e., a reply is not received in time).
        /// </summary>
        public override void Execute( Node localNode ) {
            localNode.RequestReplyTable.ReplyTimerFired( request.MsgId );
        }

        /// <summary>
        /// ReplyTimers are not background messages since the application code is expecting a reply
        /// </summary>
        public override bool IsBackground() {
            return this.background;
        }    
    }
 

    /// <summary>
    /// Timer for garbage-collecting entries from the reply table.
    /// </summary>
    internal class EntryGCTimer : Timer {

        const int TTL_FACTOR = 5;

        string   msgId;

        /// <summary>
        /// Constructor. Determine the GC interval and record the message id.
        /// </summary>
        public EntryGCTimer( OverlayRequestMsg request )
            : base( TTL_FACTOR*request.TimeoutMs )
        {
            this.msgId = request.MsgId;
        }
        
        /// <summary>
        /// Invoked when a table entry is to be garbage-collected.
        /// </summary>
        public override void Execute( Node localNode ) {
            localNode.RequestReplyTable.RemoveEntry( msgId );
        }
    }
}