using System;
using System.Collections;
using System.Diagnostics;
using System.Threading;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Taro {

    public class QueueEntry {
        public RouteMsg             Message;
        public DateTime             FireAt;
    }
    
    /// <summary>
    /// A synchronized queue of messages
    /// </summary>
    public class TaroMessageQueue {
        // A list of ArrayList objects
        SortedList      list = new SortedList();
        Hashtable       table = new Hashtable();
        bool            killed = false;

        // Add the message to the queue
        public void Add( QueueEntry p ) {
            lock(this) {
                if( killed ) {
                    return;
                }

                // Add message to the list, sorted by delivery time
                ArrayList listEntry = (ArrayList) list[p.FireAt];
                if( listEntry!=null ) {
                    listEntry.Add( p );
                } else {
                    listEntry = new ArrayList();
                    listEntry.Add( p );
                    list.Add( p.FireAt, listEntry );
                }

                // Add the message to the table accessed by key
                Debug.Assert( ! table.ContainsKey(p.Message) );
                table[ p.Message ] = p;

                Monitor.Pulse(this);
            }
        }

        // Wait for the next event on the queue to be ready
        // Returns null if the queue was killed
        public QueueEntry Remove() {
            TimeSpan    tsInfinite=new TimeSpan(0,0,0,0,Timeout.Infinite);     // Infinite timeout
            QueueEntry  p = null;
            TimeSpan    ts;
            
            // Extract the next Message to process
            lock(this) {
                for(;;) {
                    if( killed ) {
                        return null;
                    }

                    if( list.Count>0 ) {
                        ArrayList listEntry = (ArrayList) list.GetByIndex(0);
                        Debug.Assert( listEntry.Count>0 );
                        p = (QueueEntry) listEntry[0];

                        // We got a message. Find out when it should be processed
                        DateTime now = DateTime.Now;
                        ts = p.FireAt.Subtract( DateTime.Now );
                        Log.Write(TaroLocalEvent.CATEGORY, LogLvl.Lo, "Message fires in {0} ms",
                            ts.TotalMilliseconds);

                        // Could probably use DateTime.CompareTo() here
                        if( ts.Ticks<=0 ) {
                            // It is time to process the message

                            // Remove the entry from the list
                            listEntry.RemoveAt(0);
                            if( listEntry.Count==0 ) {
                                list.RemoveAt(0);
                            }

                            // Remove the entry from the table
                            Debug.Assert( table.ContainsKey(p.Message) );
                            table.Remove(p.Message);

                            return p;
                        }

                        // Message is not ready to be processed.
                        // Wait for amount of time specified by 'ts'
                    } else {
                        // No messages in the queue. Wait until we are signalled.
                        ts = tsInfinite;
                    }

                    // Wait until we are signalled or the timeout expires.
                    Monitor.Wait(this, ts);
                }                
            }
        }

        
        /// <summary>
        /// Remove an event from the event queue.
        /// </summary>
        /// <param name="Message">A Message acting as a "comparison key" that is equal to the
        /// message to be removed.</param>
        public QueueEntry Remove(MessageRouting.RouteMsg msg) {
            lock(this) {
                // Retrieve the event matching the given key
                Debug.Assert( table.ContainsKey(msg) );
                QueueEntry p = (QueueEntry) table[msg];
                Debug.Assert( msg.Equals(p.Message) );
            
                // Remove the event
                ArrayList listEntry = (ArrayList) list[p.FireAt];
                Debug.Assert( listEntry!=null );
                Debug.Assert( listEntry.Contains(p) );
                listEntry.Remove(p);
                if( listEntry.Count==0 ) {
                    list.Remove(p.FireAt);
                }
                table.Remove(msg);

                // Return the event
                return p;
            }
        }

        // Kill the queue immediately
        public void Shutdown() {
            lock( this ) {
                killed = true;
                Monitor.Pulse(this);
            }
        }

        // Get the number of messages in the queue that are currently ready to be processed
        public int GetQueueLength() {
            int length=0;

            // Extract the next Message to process
            lock(this) {
                for( int i=0; i<list.Count; i++ ) {
                    ArrayList listEntry = (ArrayList) list.GetByIndex(i);
                    Debug.Assert( listEntry.Count>0 );
                    QueueEntry p = (QueueEntry) listEntry[0];

                    // We got a message. Find out when it should be processed
                    DateTime now = DateTime.Now;
                    TimeSpan ts = p.FireAt.Subtract( DateTime.Now );
                    // Could probably use DateTime.CompareTo() here

                    if( ts.Ticks<0 ) {
                        // Message _is_ ready to be processed
                        length += listEntry.Count;
                    } else if( ts.Ticks>0 ) {
                        // Message is still _not_ ready to be processed
                        break;
                    }
                }
            }

            return length;
        }

    }

    // This class is used for scheduling events to the local application in the
    // future.
    public class TaroLocalEvent : MessageRouting.LocalEvent {

        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        public const string     CATEGORY = "TaroLocalEvent";

        
        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        private TaroMessageQueue mq = new TaroMessageQueue();
        private Node node;

        // Synchronization
        public static Object overlayLock;


        //**********************************************************************************************
        // Constructor
        //**********************************************************************************************
        public TaroLocalEvent( Node n ) {
            node = n;
            overlayLock = this;
            numMessagesReceivedFromWire = 0;
            numMessagesSentToWire = 0;
        }

        public long numMessagesReceivedFromWire;
        public long numMessagesSentToWire;
        public long GetNumMessagesReceivedFromWire(Node node) 
        {
            return numMessagesReceivedFromWire;
        }
        public long GetNumMessagesSentToWire(Node node) 
        {
            return numMessagesSentToWire;
        }

        /// <summary>
        /// Raises an event immediately on the current thread
        /// </summary>
        /// <param name="message">Message to raise</param>
        public void RaiseEventNow(Object message) {
            RouteMsg rm = (RouteMsg) message;
            node.RaiseEventUpcall( rm );
        }
        
        /// <summary>
        /// Queues an operation to be run on a thread-pool thread
        /// </summary>
        /// <param name="ms"></param>
        /// <param name="me"></param>
        /// <param name="Message"></param>
        public void RaiseEvent(long ms, NodeIdAddressBind me, MessageRouting.RouteMsg message) {
            RouteMsg rm = (RouteMsg) message;
            const int TicksPerMillisecond=10000;
            
            //Log.Write(CATEGORY, LogLvl.Lo, "Raising local event: delay={0}ms", ms);

            QueueEntry p = new QueueEntry();
            p.Message = rm;
            p.FireAt = DateTime.Now.Add( new TimeSpan(TicksPerMillisecond*ms) );

            //Log.Write(CATEGORY, LogLvl.Lo, "Event will fire at: {0}", p.FireAt.ToLongTimeString());

            mq.Add( p );
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="message"></param>
        public void RemoveEvent(MessageRouting.RouteMsg message) {
            mq.Remove(message);
        }

        // This is the main method of the Overlay thread. It reads messages from the
        // MessageQueue and processes them.
        public void EventLoop() {
            for(;;) {
                QueueEntry p = mq.Remove();
                if( p==null ) {
                    Console.WriteLine("MessageQueue was killed. Exiting.");
                    return;
                }

                // Old, single-threaded code
                lock( overlayLock ) {
                    node.RaiseEventUpcall( p.Message );
                }
            }
        }

        // Callback from the thread-pool
        private void RaiseEventUpcall( Object o ) {
            RouteMsg rm = (RouteMsg) o;
            node.RaiseEventUpcall( rm );
        }

        // This method is called to shutdown the MessageQueue and the Overlay thread.
        public void Shutdown() {
            mq.Shutdown();
        }

        public long GetTime() {
            return (DateTime.Now.Ticks/10000); // time in msecs
        }

        public int GetQueueLength() {
            return mq.GetQueueLength();
        }
    }

}
