/* Ping.cs 
 * MSDN Magazine by Lance Olson 
 */
using System;
using System.Net;
using System.Net.Sockets;
using System.Diagnostics;
using System.Threading;
using System.Runtime.InteropServices;

namespace Microsoft.Research.Taro {

	/// <summary>
	///		The Main Ping Class 
	/// </summary>
	internal class Ping {

		/// <summary>
		///	Class that holds the ICMP Packet information
		/// </summary>
		private class IcmpPacket { 
			public Byte  Type;    // type of message
			public Byte  SubCode;    // type of sub code
			public UInt16 CheckSum;   // ones complement checksum of struct
			public UInt16 Identifier;      // identifier
			public UInt16 SequenceNumber;     // sequence number  
			public Byte [] Data;
		} // class IcmpPacket

		// These are the return values
		private const int HOST_NOT_FOUND = -1;
		private const int PACKET_CREATION = -2;
		private const int SEND_PACKET = -3;
		private const int RECEIVE_PACKET = -4;

		//Declare some Constant Variables
		private const int SOCKET_ERROR = -1;        
		private const int ICMP_ECHO = 8;

		// External functions
		[DllImport("kernel32.dll")]
		static extern void QueryPerformanceCounter(out long c);

		// Private members
		private Socket socket = new Socket(AddressFamily.InterNetwork, SocketType.Raw, ProtocolType.Icmp);
		private byte[] nonce;
		private int ttl;
		private byte[] ReceiveBuffer = new Byte[256];
		private EndPoint EndPointFrom;
			
		// Constructor
		public Ping() {
			// Set up the nonce we send in the ping
			nonce = new byte[32];
			for( int i=0; i<nonce.Length; i++ ) {
				nonce[i] = (byte) i;
			}
			ttl = 30;

			// Set up the receiving endpoint
			IPHostEntry fromHE = Dns.GetHostByName(Dns.GetHostName());
			IPEndPoint ipEndPointFrom = new IPEndPoint(fromHE.AddressList[0], 0);        
			EndPointFrom = ipEndPointFrom;
		}

		/// <summary>
		///	This method takes the "hostname" of the server and returns the number of *milliseconds*
		///	
		///	Note: Does not work well if the Ping is lost or the destination doesn't reply.
		/// </summary>
		public ulong SendPingHost(string host, out int errorCode)
		{
			//Declare the IPHostEntry 
			IPHostEntry serverHE;
			int nBytes = 0;
			errorCode = 0;

			// Get the server endpoint
			try { serverHE = Dns.Resolve(host); }
			catch(Exception) { errorCode = HOST_NOT_FOUND; return 0; }

			// Convert the server IP_EndPoint to an EndPoint
			Console.WriteLine("Sending ICMP Ping to: {0}",serverHE.AddressList[0].ToString());
			IPEndPoint ipepServer = new IPEndPoint(serverHE.AddressList[0], 0);
			EndPoint epServer = (ipepServer);	

			int PacketSize = 0;
			IcmpPacket packet = new IcmpPacket();
			// Construct the packet to send
			packet.Type = ICMP_ECHO; //8
			packet.SubCode = 0;
			packet.CheckSum = 0; 
			packet.Identifier = 45; 
			packet.SequenceNumber = 0; 
			int PingData = 32; // sizeof(IcmpPacket) - 8;
			packet.Data = new Byte[PingData];
			//Initilize the Packet.Data
			for (int i = 0; i < nonce.Length; i++)
				packet.Data[i] = nonce[i];
			//Variable to hold the total Packet size
			PacketSize = PingData + 8;
			Byte [] icmp_pkt_buffer = new Byte[ PacketSize ]; 
			Int32 Index = 0;
			//Call a Method Serialize which counts
			//The total number of Bytes in the Packet
			Index = Serialize(packet,icmp_pkt_buffer,PacketSize,PingData);
			//Error in Packet Size
			if (Index == -1)
			{
				errorCode = PACKET_CREATION;
				return 0;
			}
          
			// now get this critter into a UInt16 array
	         
			//Get the Half size of the Packet
			Double double_length = Convert.ToDouble(Index);
			Double dtemp = Math.Ceiling( double_length / 2);
			int cksum_buffer_length = Convert.ToInt32(dtemp);
			//Create a Byte Array
			UInt16 [] cksum_buffer = new UInt16[cksum_buffer_length];
			//Code to initilize the Uint16 array 
			int icmp_header_buffer_index = 0;
			for( int i = 0; i < cksum_buffer_length; i++ ) 
			{
				cksum_buffer[i] = BitConverter.ToUInt16(icmp_pkt_buffer,icmp_header_buffer_index);
				icmp_header_buffer_index += 2;
			}
			//Call a method which will return a checksum             
			UInt16 u_cksum = checksum(cksum_buffer, cksum_buffer_length);
			//Save the checksum to the Packet
			packet.CheckSum  = u_cksum; 
	            
			// Now that we have the checksum, serialize the packet again
			Byte [] sendbuf = new Byte[ PacketSize ]; 
			//again check the packet size
			Index = Serialize(packet,sendbuf,PacketSize,PingData );
			//if there is a error report it
			if( Index == -1 )
			{
				errorCode = PACKET_CREATION;
				return 0;
			}

			socket.SetSocketOption(SocketOptionLevel.IP,SocketOptionName.IpTimeToLive,ttl);
			
			// Start timer now
			ulong start;
			try {
				long tmp;
				QueryPerformanceCounter(out tmp);
				start = (ulong) tmp;
			}
			catch (Exception) {
				errorCode = SEND_PACKET;
				return 0;
			}
			
			// Stuff the ICMP packet into the socket.
			if ((nBytes = socket.SendTo(sendbuf, PacketSize, 0, epServer)) == SOCKET_ERROR) 
			{
				errorCode = SEND_PACKET;
				return 0;
			}
			Debug.Assert( nBytes==PacketSize );

			// Receive the reply from the socket.
			nBytes = socket.ReceiveFrom(ReceiveBuffer, 256, SocketFlags.None, ref EndPointFrom);
			if (nBytes>0) {
				if (ReceiveBuffer[20] != 11) {
					for (int i = 0; i < 20; i++) {
						byte b1 = nonce[i];
						byte b2 = ReceiveBuffer[28+i];
						if( b1!=b2 ) {
							errorCode = RECEIVE_PACKET;
							return 0;
						}
					}
				} else {
					errorCode = RECEIVE_PACKET;
					return 0;
				}
			} else {
				errorCode = RECEIVE_PACKET;
				return 0;
			}

			// Finished the receive. End timing now.
			ulong end;
			try {
				long tmp;
				QueryPerformanceCounter(out tmp);
				end = (ulong) tmp;
			}
			catch (Exception) {
				errorCode = SEND_PACKET;
				return 0;
			}
			ulong duration = end-start;

			return (end-start);
		}

		/// <summary>
		///  This method get the Packet and calculates the total size 
		///  of the Pack by converting it to byte array
		/// </summary>
		private static Int32 Serialize(IcmpPacket packet, Byte [] Buffer, Int32 PacketSize, Int32 PingData )
		{
			Int32 cbReturn = 0;
			// serialize the struct into the array
			int Index=0;

			Byte [] b_type = new Byte[1];
			b_type[0] = (packet.Type);

			Byte [] b_code = new Byte[1];
			b_code[0] = (packet.SubCode);

			Byte [] b_cksum = BitConverter.GetBytes(packet.CheckSum);
			Byte [] b_id = BitConverter.GetBytes(packet.Identifier);
			Byte [] b_seq = BitConverter.GetBytes(packet.SequenceNumber);
	        
			// Console.WriteLine("Serialize type ");
			Array.Copy( b_type, 0, Buffer, Index, b_type.Length );
			Index += b_type.Length;
	        
			// Console.WriteLine("Serialize code ");
			Array.Copy( b_code, 0, Buffer, Index, b_code.Length );
			Index += b_code.Length;

			// Console.WriteLine("Serialize cksum ");
			Array.Copy( b_cksum, 0, Buffer, Index, b_cksum.Length );
			Index += b_cksum.Length;

			// Console.WriteLine("Serialize id ");
			Array.Copy( b_id, 0, Buffer, Index, b_id.Length );
			Index += b_id.Length;

			Array.Copy( b_seq, 0, Buffer, Index, b_seq.Length );
			Index += b_seq.Length;

			// copy the data	        
			Array.Copy( packet.Data, 0, Buffer, Index, PingData );
			Index += PingData;
			if( Index != PacketSize/* sizeof(IcmpPacket)  */) 
			{
				cbReturn = -1;
				return cbReturn;
			}

			cbReturn = Index;
			return cbReturn;
		}
		/// <summary>
		///		This Method has the algorithm to make a checksum 
		/// </summary>
		public static UInt16 checksum( UInt16[] buffer, int size )
		{
			Int32 cksum = 0;
			int counter;

			counter = 0;

			while ( size > 0 ) 
			{

				UInt16 val = buffer[counter];

				cksum += Convert.ToInt32( buffer[counter] );
				counter += 1;
				size -= 1;
			}

			cksum = (cksum >> 16) + (cksum & 0xffff);
			cksum += (cksum >> 16);
			return (UInt16)(~cksum);
		}
	} // class ping
}