
using System;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Xml.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
//using ZlibDotNetWrapper;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

// Taro socket receive layer

namespace Microsoft.Research.Taro {

    class SocketState {
        Socket sock;
        byte[] buffer;
        int buflen;
        bool headerComplete;    // flag indicating whether the message header has been processed
        int messageType;        // type of message, as specified in the packet header
        int messageLength;      // total length of message, as specified in the packet header
        int curMessageLength;   // current length of message (i.e. the amount that has been received)
        MemoryStream packetData;
        int offset;

        // useful information for debugging
        public DateTime receiveStarted;
        public string source;
        public string dest;
        
        public SocketState(Socket s)
        {
            buffer = new Byte[8192];
            buflen = 8192;
            headerComplete = false;
            messageType = TaroSocketConfig.TARO_MSG_UNKNOWN;
            messageLength = 0;
            curMessageLength = 0;
            packetData = new MemoryStream();
            offset = 0;
            this.sock = s;
        }
        
        public void Reset()
        {
            packetData.Close();
            packetData = new MemoryStream();
            messageType = TaroSocketConfig.TARO_MSG_UNKNOWN;
            messageLength = 0;
            curMessageLength = 0;
            headerComplete = false;
            offset = 0;
        }
        
        public void AppendDataToStream(bool skipHeader)
        {
            try {
                if (skipHeader) {
                    packetData.Write(buffer, 8, (offset-8));
                    curMessageLength += (offset - 8);
                } else {    
                    packetData.Write(buffer, 0, offset);
                    curMessageLength += offset;
                }
                offset = 0;
            }
            catch (Exception e) {
                Console.WriteLine("AppendDataToStream Error: {0}", e.ToString());
            }
        }
        
        public void SetHeader()
        {
            byte[] length;
            byte[] type;
            type = new byte[4];
            Array.Copy(buffer,type,4);
            length = new byte[4];
            Array.Copy(buffer,4,length,0,4);

            MemoryStream ms = new MemoryStream(type, 0, 4);
            BinaryReader br = new BinaryReader(ms);
            messageType = br.ReadInt32();
            br.Close();
            ms.Close();

            ms = new MemoryStream(length, 0, 4);
            br = new BinaryReader(ms);
            messageLength = br.ReadInt32();
            br.Close();
            ms.Close();

            headerComplete = true;
        }
        
        public bool HeaderComplete
        {
            get { return headerComplete; }
        }
        
        public int MessageType
        {
            get { return messageType; }
        }
        
        public MemoryStream PacketData
        {
            get { return packetData; }
            set { packetData = value; }
        }
            
        public int MessageLength
        {
            get { return messageLength; }
        }

        public int CurMessageLength
        {
            get { return curMessageLength; }
        }
        
        public int Buflen 
        {
            get { return buflen; }
        }
            
        public byte[] Buffer 
        {
            get { return buffer; }
        }
        
        public Socket Sock
        {
            get { return sock; }
        }
        
        public int Offset 
        {
            get { return offset; }
            set { offset = value; }
        }
    }
    
    // TaroSocketReceiveMessage is an implementation of TaroReceiveMessage, but specific to
    // Sockets. It receives Overlay messages on its sockets and passes them up to
    // the LocalEvent object. It also receives and replies to Taro pings.
    public class TaroSocketReceiveMessage : TaroReceiveMessage {
        private int                     localPort;  // e.g. "10001"

        private Socket                  serverSocket;
        private IPEndPoint              localEP;
        private Thread                  serverThread;
        private byte[]                  eom;

        private int                     msgCounter;
        private int                     pingCounter;
        private CounterCallback         callback;

        public const string CATEGORY = "TaroSocketReceive";
        static Log log = new Log(CATEGORY);

        // Constructor
        public TaroSocketReceiveMessage(int localPort, TaroLocalEvent localEvent,
                                        Node node, object serializer, CounterCallback callback)
            : base( localEvent, node, serializer )
        {
            this.localPort = localPort;
            this.msgCounter = 0;
            this.pingCounter = 0;
            this.callback = callback;

            string address = node.GetPhysicalAddr.Address;
            int colon = address.IndexOf(':');
            IPAddress localAddr = IPAddress.Parse(address.Substring(0, colon));

            this.localEP = new IPEndPoint(localAddr, localPort);
            this.serverSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp );
            this.eom = new byte[4];

            // Try to bind the socket to the endpoint
            try {
                serverSocket.Bind(localEP);
                serverSocket.Listen(80);
                Console.WriteLine("Node listening on {0}", localEP.ToString());
            } 
            catch (SocketException e) {
                Console.WriteLine("SocketException in TaroSocketReceiveMessage constructor - local endpoint = {0}, error message = {1}", localEP.ToString(), e.Message);
            }
            catch (Exception e) {
                Console.WriteLine("TaroSocketReceiveMessage: exception near bind/listen - {0}", e.ToString());
            }
            
            serverThread = new Thread( new ThreadStart(ServerMain) );
            serverThread.Start();
        }

        // This method is invoked when the user exits. We
        // remove our handlers and so upon returning are no longer accepting commands.
        public override void ShutDown() {
            try {
                serverSocket.Shutdown(SocketShutdown.Both);
                serverSocket.Close();
            }
            catch (SocketException e) {
                Console.WriteLine("Shutdown on listen socket {0} generated a SocketException: {1}",
                                  localEP, e.Message);
            }
            catch (Exception e) {
                Console.WriteLine("ShutDown Error: {0}", e.ToString());
            }
        }
        
        AsyncCallback recvCallback = null;
        
        // this method is the asynchronous I/O completion callback
        private void ReceiveCallback(IAsyncResult asyn)
        {
            IAsyncResult result;
            SocketState state = (SocketState) asyn.AsyncState;
            int bytesRecvd;

            try {
                bytesRecvd = state.Sock.EndReceive(asyn);
                if (bytesRecvd == 0) {
                    // the other end of the socket has closed gracefully
                    if (TaroSocketConfig.Debug) {
                        Console.WriteLine("EndReceive closing socket {0} -> {1}", state.source, state.dest);
                    }
                    state.Sock.Close();
                    return;
                }
            }
            catch (ObjectDisposedException) {
                if (TaroSocketConfig.Debug) {
                    // TODO: can we access state.source and state.dest if the object has been disposed???
                    Console.WriteLine("EndReceive closing disposed socket");
                }
                // the socket has been closed
                return;
            }
            catch (SocketException e) {
                Console.WriteLine("EndReceive socket {0} -> {1} generated a SocketException: {2}",
                                  state.source, state.dest, e.Message);
                return;
            }
            catch (Exception e) {
                Console.WriteLine("EndReceive error: {0}", e.ToString());
                return;
            }

            if (recvCallback == null)
                recvCallback = new AsyncCallback(this.ReceiveCallback);

            // overall strategy: don't do any processing until the first 8 bytes (the header)
            // have been received.  once the header is complete, then we just copy data from 
            // recv buffer into the memorystream until the message is done.

            int totalBytesRecvd = bytesRecvd + state.Offset;
            state.Offset = totalBytesRecvd;

            bool skipHeader = false;
            // step 1: receive the message type and the length
            if (state.HeaderComplete == false) {
                // check if this is the first time ReceiveCallback has been called for this message
                if (bytesRecvd == totalBytesRecvd) {
                    state.receiveStarted = DateTime.Now;
                    if (TaroSocketConfig.Debug) {
                        log.Write(LogLvl.Med,"ReceiveCallback: received first segment: {0} bytes", bytesRecvd);
                        log.Write(LogLvl.Med,"ReceiveMessage from {0} began at time {1:T}.{2:000}",
                                  state.Sock.RemoteEndPoint, state.receiveStarted, 
                                  state.receiveStarted.Millisecond);
                    }
                }
                if (totalBytesRecvd >= 8) {
                    state.SetHeader();
                    skipHeader = true;
                } else {
                    // call BeginReceive to listen for new messages on this socket
                    result = state.Sock.BeginReceive(state.Buffer, state.Offset, 
                                                     (state.Buflen - state.Offset), SocketFlags.None, 
                                                     recvCallback, state);
                    return;
                }
            }

            // copy message bytes to memory stream
            state.AppendDataToStream(skipHeader);
            state.Offset = 0;

            if (state.CurMessageLength == state.MessageLength) {
                try 
                {
                    // Send EOM signal indicating we have received the whole message
                    state.Sock.Send(eom);
                }
                catch (SocketException e) 
                {
                    Console.WriteLine("Send of EOM ACK: socket {0} -> {1} generated a SocketException: {2}",
                                      state.source, state.dest, e.Message);
                    return;
                }
                catch (Exception e) 
                {
                    Console.WriteLine("Send of EOM ACK error: {0}", e.ToString());
                    return;
                }                
                if (TaroSocketConfig.Debug) {
                    DateTime end = DateTime.Now;
                    log.Write(LogLvl.Med, "ReceiveMessage from {0} ended at time {1:T}.{2:000} ({3:N0}ms since start) -> {4} bytes",
                              state.Sock.RemoteEndPoint, end, end.Millisecond,
                              end.Subtract(state.receiveStarted).TotalMilliseconds, state.PacketData.Length);
                }

                // check message type here. if it is a ping message, treat it 
                // differently from a regular message
                if (state.MessageType == TaroSocketConfig.TARO_MSG_PING) {
                    if (TaroSocketConfig.Debug) {
                        log.Write(LogLvl.Med, "EndReceive got ping message of length {0} on socket {1} -> {2}",
                                  state.PacketData.Length, state.source, state.dest);
                    }
                    // there is no need to parse the ping message
                    state.Reset();
                    pingCounter++;
                } else if (state.MessageType == TaroSocketConfig.TARO_MSG_APP) {
                    
                    if (TaroSocketConfig.Debug) {
                        log.Write(LogLvl.Med, "EndReceive got message of length {0} on socket {1} -> {2}",
                                  state.PacketData.Length, state.source, state.dest);
                    }
                    
                    // Deserialize message into RouteMessage object
                    state.PacketData.Position = 0;
                    RouteMsg rm;
                    if (TaroSocketConfig.USE_COMPRESSION) 
                    {
#if COMPRESSION_AVAILABLE
                        MemoryStream ms = new MemoryStream();
                        CompressUtil.UnCompress(state.PacketData, ms);
                        ms.Position = 0;
                        lock( serializer ) 
                        {
                            if (serializer is XmlSerializer)
                                rm = (RouteMsg) ((XmlSerializer) serializer).Deserialize(ms);
                            else
                                rm = (RouteMsg) ((BinaryFormatter) serializer).Deserialize(ms);
                        }
                        ms.Close();
#else
                        Debug.Assert(false, "Message compression not supported in this release");
                        rm = null;
#endif
                    } 
                    else 
                    {
                        lock( serializer ) 
                        {
                            if (serializer is XmlSerializer)
                                rm = (RouteMsg) ((XmlSerializer) serializer).Deserialize(state.PacketData);
                            else
                                rm = (RouteMsg) ((BinaryFormatter) serializer).Deserialize(state.PacketData);
                        }
                    }
                    state.Reset();

                    localEvent.numMessagesReceivedFromWire++;
                    msgCounter++;
                    if (callback != null) {
                        callback(msgCounter);
                    }
                    
                    // inform the localEvent queue that an incoming message has arrived
                    localEvent.RaiseEvent(0, node.GetNodeIdAddressBind, rm);
                } else {
                    Console.WriteLine("TaroSocket ReceiveCallback: unknown message type!");
                }
            }

            // call BeginReceive to listen for new messages on this socket
            result = state.Sock.BeginReceive(state.Buffer, 0, state.Buflen, SocketFlags.None, recvCallback,
                                             state);
        }

        // This is a thread which receives messages from the server socket.
        private void ServerMain() 
        {
            try {
                while( true ) {
                    Socket newSocket = serverSocket.Accept();
                    if (newSocket == null) {
                        Console.WriteLine("Winsock error: " + 
                                          Convert.ToString(System.Runtime.InteropServices.Marshal.GetLastWin32Error()));
                    } else if (!newSocket.Connected ) {
                        Console.WriteLine("Socket not connected: " +
                                          Convert.ToString(System.Runtime.InteropServices.Marshal.GetLastWin32Error()));
                    } else {
                        newSocket.SetSocketOption(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, 1);
                        SocketState sockState = new SocketState(newSocket);
                        sockState.source = newSocket.RemoteEndPoint.ToString();
                        sockState.dest = localEP.ToString();
                        if (TaroSocketConfig.Debug) {
                            Console.WriteLine("Accepted connection from {0} -> {1}", sockState.source,
                                              sockState.dest);
                        }
                        if ( recvCallback == null )
                            recvCallback = new AsyncCallback (this.ReceiveCallback);
                        IAsyncResult result = newSocket.BeginReceive(sockState.Buffer, 0, sockState.Buflen, 
                                                                     SocketFlags.None, recvCallback,
                                                                     sockState);
                    }
                }
            }
            catch (SocketException e) {
                Console.WriteLine("ServerMain socket exception: {0}", e.ToString());                
            }
            catch (Exception e) {
                Console.WriteLine("ServerMain error: {0}", e.ToString());
            }
        }
    }
}
