
using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Xml.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
//using ZlibDotNetWrapper;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Taro {

    public class TaroSocketConfig 
    {
        // there are two kinds of messages: pings and application messages.
        public const int TARO_MSG_UNKNOWN = 0;
        public const int TARO_MSG_PING = 1;
        public const int TARO_MSG_APP = 2;

        // if true, print lots of detailed debugging messages
        public static bool Debug = false;
        public static bool USE_COMPRESSION = false;
        public static bool DUMP_XML = false;
        public static bool USE_MODELNET = false;
    }
    
    // the AddressPair class is used as the key for the Hashtable that contains the cache
    // entries for the socket cache
    public class AddressPair 
    {
        IPEndPoint srcEp;
        IPEndPoint destEp;
        
        // constructor
        public AddressPair(IPEndPoint source, IPEndPoint dest)
        {
            srcEp = source;
            destEp = dest;
        }
        
        public IPEndPoint DestEp 
        {
            get { return destEp; }
        }
        
        public IPEndPoint SrcEp
        {
            get { return srcEp; }
        }
        
        // Overridden equals (used for Hashtables)
        public override bool Equals(Object obj) {
            if (!(obj is AddressPair))
                return false;
            return Equals((AddressPair)obj);
        }
        
        public bool Equals(AddressPair ap)
        {
            if (this.srcEp.Equals(ap.srcEp) && this.destEp.Equals(ap.destEp)) {
                return true;
            }
            return false;
        }
        
        public override int GetHashCode()
        {
            int hashVal = srcEp.Address.GetHashCode();
            hashVal += srcEp.Port.GetHashCode();
            hashVal += destEp.Address.GetHashCode();
            hashVal += destEp.Port.GetHashCode();
            
            return hashVal;
        }
    }
        
    // a connection cache so that we don't always need to create a new TCP connection
    // each time we have a message to send to a neighbor
    public class SocketCache 
    {
        Hashtable sc;    // currently open sockets are stored here

        private TaroLocalEvent localEvent;

        // useful cache stats for debugging
        int numLookups;  
        int numHits;

        private int cacheSize = 200;  // try to keep the cache at this size - not a hard limit

        // constructor
        public SocketCache(TaroLocalEvent localEvent)
        {
            sc = new Hashtable();
            numLookups = 0;
            numHits = 0;
            this.localEvent = localEvent;
        }
        
        // a socket cache entry
        public class CacheEntry 
        {
            Socket s;
            bool failed = false;
            AddressPair ap;
            long age;
            bool inUse;
            private TaroLocalEvent localEvent;

            public CacheEntry(AddressPair ap, TaroLocalEvent localEvent)
            {
                s = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                age = localEvent.GetTime();
                inUse = true;
                this.ap = ap;
                this.localEvent = localEvent;
            }
            
            public Socket Socket
            {
                get { return s; }
            }
            
            public AddressPair Ap
            {
                get { return ap; }
            }

            public bool Failed
            {
                get { return failed; }
                set { failed = value; }
            }
            
            public void Done()
            {
                age = localEvent.GetTime();
                inUse = false;
            }

            public long Age 
            {
                get { return age; }
            }
            
            public bool InUse
            {
                get { return inUse; }
                set { 
                    // no one should be setting inUse = false without calling Done()
                    Debug.Assert(value == true);
                    inUse = value;
                }
            }
        }

        public AddressPair ConvertToAddressPair(string source, string dest)
        {
            int colon;
            AddressPair retval = null;
            
            // convert the source and destination endpoint information from string
            // format (e.g. 10.0.0.1:2000) to IPEndPoint objects
            colon = source.IndexOf(':');
            Debug.Assert(colon != -1);
            string srcIpStr = source.Substring(0, colon);
            int srcPort = Int32.Parse(source.Substring(colon + 1));
            colon = dest.IndexOf(':');
            Debug.Assert(colon != -1);
            string destIpStr = dest.Substring(0, colon);
            int destPort = Int32.Parse(dest.Substring(colon + 1));
            IPAddress destIpAddr = IPAddress.Parse(destIpStr);
            IPAddress srcIpAddr = IPAddress.Parse(srcIpStr);
            IPEndPoint srcEp = new IPEndPoint(srcIpAddr,0);
            IPEndPoint destEp = new IPEndPoint(destIpAddr,destPort);
            retval = new AddressPair(srcEp, destEp);

            return retval;
        }
        
        // the CheckSocket method is used to verify that it is OK to dequeue a message and initiate a
        // send worker thread to send the message.  If the socket exists and it is in use, return false.
        // Otherwise, return true.
        public bool CheckSocket(string source, string dest) 
        {
            // create an AddressPair object, which is used as the key
            // for lookups in the hashtable
            AddressPair apKey = ConvertToAddressPair(source, dest);

            CacheEntry sce = (CacheEntry) sc[apKey];
            if (sce == null) {
                return true;
            }
            
            return (!(sce.InUse));
        }
        
        // the GetSocket method is used to initiate a connection between the source and destination.
        // it works by looking up the source and destination in the cache, and creating a new tcp
        // connection if none already exists.
        public SocketCache.CacheEntry GetSocket(string source, string dest)
        {
            // create an AddressPair object, which is used as the key
            // for lookups in the hashtable
            AddressPair apKey = ConvertToAddressPair(source, dest);

            numLookups++;

            bool stillFull = false;
            // call EvictCacheEntry repeatedly until it returns false
            while (stillFull = EvictCacheEntry()) {};

            CacheEntry sce = (CacheEntry) sc[apKey];
            if (sce == null) {
                // create a new TCP socket
                sce = new CacheEntry(apKey, localEvent);
                sc.Add(apKey,sce);
            } else {
                // on a cache hit, mark the socket as in use
                sce.InUse = true;
                numHits++;
            }

            return sce;
        }

        // bind and connect the socket
        public bool ConnectSocket(CacheEntry sce)
        {
            DateTime connectStartedAt = DateTime.Now;
            TaroSocketSendMessage.log.Write(LogLvl.Med, "ConnectSocket {0} to {1} begun at time {2:T}.{3:000}",
                sce.Ap.SrcEp, sce.Ap.DestEp, connectStartedAt, connectStartedAt.Millisecond);
            try {
                if (sce.Socket.LocalEndPoint != null) {
                    Console.WriteLine("Error:  about to call bind on a socket that is already bound");
                    if (sce.Socket.Connected) {
                        Console.WriteLine("Furthermore, that socket is also connected. sce.Failed = {0}",sce.Failed);
                    } else {
                        Console.WriteLine("Fortunately, that socket is not connected. sce.Failed = {0}",sce.Failed);
                    }
                }
                sce.Socket.Bind(sce.Ap.SrcEp);
            }
            catch(SocketException e) {
                Console.WriteLine("Sender generated a SocketException - Bind(source ep = {0}) in ConnectSocket: {1}",
                                  sce.Ap.SrcEp, e.Message);
                sce.Failed = true;
                sce.Done();
                return false;
            }
            catch (Exception e) {
                Console.WriteLine("Error: {0}", e.ToString());
                sce.Failed = true;
                sce.Done();
                return false;
            }

            try {
                sce.Socket.SetSocketOption(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, 1);
                sce.Socket.Connect(sce.Ap.DestEp);
                if (!sce.Socket.Connected) 
                {
                    Console.WriteLine("Unable to connect to host: {0}", sce.Ap.DestEp);
                    sce.Failed = true;
                    sce.Done();
                    return false;
                }
                if (TaroSocketConfig.Debug) {
                    Console.WriteLine("Socket connection established: {0} -> {1}", sce.Ap.SrcEp, 
                                      sce.Ap.DestEp);
                }
            }
            catch (SocketException e) 
            {
                Console.WriteLine("Sender ({0} in ConnectSocket, socket destination = {1}) generated a SocketException: {2}", 
                                  e.TargetSite, sce.Ap.DestEp, e.Message);
                sce.Failed = true;
                sce.Done();
                return false;
            }
            catch (Exception e) 
            {
                Console.WriteLine("Error: {0}", e.ToString());
                sce.Failed = true;
                sce.Done();
                return false;
            }
            DateTime connectFinishedAt = DateTime.Now;
            TaroSocketSendMessage.log.Write(LogLvl.Med, "ConnectSocket {0} to {1} ended at time {2:T}.{3:000} ({4:N0}ms since started)",
                sce.Ap.SrcEp, sce.Ap.DestEp,
                connectFinishedAt, connectFinishedAt.Millisecond, connectFinishedAt.Subtract(connectStartedAt).TotalMilliseconds);
            return true;
        }

        // EvictCacheEntry is called each time GetSocket creates a new TCP connection.
        // Since GetSocket is the only caller of EvictCacheEntry, and GetSocket is only called
        // by the mainSenderThread, then we know that no entries will be added to the sc HashTable
        // while this routine is running.  Worker threads may modify the value of InUse during this
        // routine, but that shouldn't be a problem.

        // This routine attempts to keep the size of the cache near the cacheSize field
        // returns true when an entry is evicted from the cache. otherwise, returns false.
        public bool EvictCacheEntry()
        {
            CacheEntry evictCandidate = null;
            
            if (sc.Count >= cacheSize) {
                long oldestEntry = 0;
                // try to evict an entry
                foreach (CacheEntry e in sc.Values) {
                    // if we find a failed connection, then there is no need to keep looking
                    // for the LRU connection
                    if (e.Failed == true) {
                        evictCandidate = e;
                        break;
                    }
                    if (e.InUse == false) {
                        if (evictCandidate == null) {
                            oldestEntry = e.Age;
                            evictCandidate = e;
                        }
                        if (e.Age < oldestEntry) {
                            oldestEntry = e.Age;
                            evictCandidate = e;
                        }
                    }
                }
                if (evictCandidate == null) {
                    return false;
                } else {
                    if (TaroSocketConfig.Debug) {
                        if (evictCandidate.Failed) {
                            Console.WriteLine("Evicting failed socket to {0}", evictCandidate.Ap.DestEp);
                        } else {
                            Console.WriteLine("Evicting LRU socket: dest = {0}", evictCandidate.Ap.DestEp);
                        }
                    }
                    try {
                        // perform the cache eviction
                        sc.Remove(evictCandidate.Ap);
                        if (evictCandidate.Socket.Connected) 
                            evictCandidate.Socket.Shutdown(SocketShutdown.Both);
                        evictCandidate.Socket.Close();
                    }
                    catch (SocketException e) {
                        Console.WriteLine("EvictCacheEntry socket cleanup generated a SocketException: {0}",
                                          e.ToString());
                    }
                    catch (Exception e) {
                        Console.WriteLine("EvictCacheEntry error: {0}", e.ToString());
                    }
                    return true;
                }
            } else {
                // if the cache isn't full, then we only look to evict
                // failed entries
                foreach (CacheEntry e in sc.Values) {
                    if (e.Failed == true) {
                        evictCandidate = e;
                        break;
                    }
                }
                if (evictCandidate == null) {
                    return false;
                } else {
                    if (TaroSocketConfig.Debug) {
                        Console.WriteLine("Evicting failed socket to {0}", evictCandidate.Ap.DestEp);
                    }
                    try {
                        sc.Remove(evictCandidate.Ap);
                        if (evictCandidate.Socket.Connected)
                            evictCandidate.Socket.Shutdown(SocketShutdown.Both);
                        evictCandidate.Socket.Close();
                    }
                    catch (SocketException e) {
                        Console.WriteLine("EvictCacheEntry socket cleanup generated a SocketException: {0}",
                                          e.ToString());
                    }
                    catch (Exception e) {
                        Console.WriteLine("EvictCacheEntry error: {0}", e.ToString());
                    }
                    return true;
                }
            }
        }
    }

    // An OutgoingMessage is the data type stored in the SocketSendQueue
    public class OutgoingMessage
    {
        byte[] data;
        string source;
        string dest;
        SocketCache.CacheEntry sce;
        bool sendFailed;
        AutoResetEvent messageDone;
        bool async;

        public DateTime enqueuedAt;

        // the following information is only used by the user callback
        ReplyCallback userCb;
        NodeIdAddressBind userDest;
        MessageRouting.RouteMsg rmObj;

        // the following information is only used by the ping callback
        TaroPingCallback pingCb;
        NodePhysicalAddress pingDest;

        public byte[] Data 
        {
            get { return data; }
            set { data = value; }
        }

        public string Source
        {
            get { return source; }
            set { source = value; }
        }
        
        public string Dest
        {
            get { return dest; }
            set { dest = value; }
        }

        public SocketCache.CacheEntry Sce
        {
            get { return sce; }
            set { sce = value; }
        }
        
        public bool SendFailed
        {
            get { return sendFailed; }
            set { sendFailed = value; }
        }
        
        public AutoResetEvent MessageDone 
        {
            get { return messageDone; }
        }
        
        public bool Async
        {
            get { return async; }
        }
        
        public ReplyCallback UserCb 
        {
            get { return userCb; }
        }

        public NodeIdAddressBind UserDest
        {
            get { return userDest; }
        }
        
        public MessageRouting.RouteMsg RmObj
        {
            get { return rmObj; }
        }

        public TaroPingCallback PingCb 
        {
            get { return pingCb; }
        }
        
        public NodePhysicalAddress PingDest
        {
            get { return pingDest; }
        }

        public OutgoingMessage(byte [] data, string source, string dest)
        {
            this.data = data;
            this.source = source;
            this.dest = dest;
            this.sendFailed = false;
            this.sce = null;
            messageDone = new AutoResetEvent(false);
            this.async = false;

            this.userCb = null;
            this.userDest = null;
            this.rmObj = null;
            
            this.pingCb = null;
            this.pingDest = null;
        }

        public OutgoingMessage(byte [] data, string source, string dest, ReplyCallback userCb, 
                               NodeIdAddressBind origDest, MessageRouting.RouteMsg rmObj)
        {
            this.data = data;
            this.source = source;
            this.dest = dest;
            this.sendFailed = false;
            this.sce = null;
            messageDone = new AutoResetEvent(false);
            this.async = true;
           
            this.userCb = userCb;
            this.userDest = origDest;
            this.rmObj = rmObj;
            
            this.pingCb = null;
            this.pingDest = null;
        }

        public OutgoingMessage(byte [] data, string source, string dest, TaroPingCallback pingCb, 
                               NodePhysicalAddress pingDest) 
        {
            this.data = data;
            this.source = source;
            this.dest = dest;
            this.sendFailed = false;
            this.sce = null;
            messageDone = new AutoResetEvent(false);
            this.async = true;

            this.userCb = null;
            this.userDest = null;
            this.rmObj = null;

            this.pingCb = pingCb;
            this.pingDest = pingDest;
        }

    }
    
    // a synchronized queue of outgoing messages
    public class SocketSendQueue 
    {
        ArrayList outputq;
        SocketCache sc;

        public SocketSendQueue(SocketCache sc)
        {
            lock(this) {
                outputq = new ArrayList();
                this.sc = sc;
            }
        }

        public void Enqueue(OutgoingMessage m) 
        {
            lock(this) {
                outputq.Add(m);
                // wake-up anyone waiting on an empty queue
                Monitor.Pulse(this);
            }
        }

        // when a sending thread is finished using an entry from the socket cache
        // it calls Wakeup so that NextMessage can rescan the outputq
        public void Wakeup()
        {
            lock(this) {
                Monitor.Pulse(this);
            }
        }
        
        // Find and return the first message on the outputq where the socket is not already
        // in use.
        public OutgoingMessage NextMessage()
        {
            lock (this) {
                OutgoingMessage m;
                
                while (true) {
                    if (outputq.Count > 0) {
                        // scan through the queue, looking for messages than are ready to send
                        for (int i = 0; i < outputq.Count; i++) {
                            m = (OutgoingMessage) outputq[i];
                            if (sc.CheckSocket(m.Source,m.Dest)) {

                                // note that it is OK to call GetSocket immediately after
                                // CheckSocket succeeds.  there is only one thread that converts
                                // SocketCache entries from free to InUse, and we're it.

                                m.Sce = sc.GetSocket(m.Source, m.Dest);
                                outputq.RemoveAt(i);
                                return m;
                            }
                        }
                    } 
                    // if we don't find anything to send, just wait for a signal
                    Monitor.Wait(this);
                }
            }
        }
    }
        
    // the SocketSendThread class is needed for the send worker threads.  a separate class is
    // needed because there is no easy way to pass arguments to the ThreadStart method
    public class SocketSendThread {
        
        private OutgoingMessage m;
        private AutoResetEvent threadFinished;
        private SocketCache sc;
        private SocketSendQueue sq;
        private TaroLocalEvent localEvent;
        private byte[] recvBytes;
        private byte[] eom;

        public SocketSendThread(OutgoingMessage m, SocketCache sc, SocketSendQueue sq,
                                TaroLocalEvent localEvent, AutoResetEvent threadFinished)
        {
            this.m = m;
            this.threadFinished = threadFinished;
            this.sc = sc;
            this.sq = sq;
            this.localEvent = localEvent;
            this.recvBytes = new byte[4];
            this.eom = new byte[4];
        }
        
        public void SendMessage()
        {
           try {
                Debug.Assert(m.Sce.InUse == true);
                bool success = true;
                // the socket returned by the socket cache might be a new socket,
                // which still needs to be connected
                if (!m.Sce.Socket.Connected) {
                    success = sc.ConnectSocket(m.Sce);
                    if (!success) 
                    {
                        m.SendFailed = true;
                    }
                }
                if (success) {
                    // Send the message
                    DateTime start = DateTime.Now;
                    TaroSocketSendMessage.log.Write(LogLvl.Med,
                        "SendMessage {0} bytes from {1} to {2} start at time {3:T}.{4:000} ({5:N0}ms since enqueue)",
                        m.Data.Length, m.Sce.Socket.LocalEndPoint, m.Sce.Socket.RemoteEndPoint, start, start.Millisecond,
                        start.Subtract(m.enqueuedAt).TotalMilliseconds);

                    m.Sce.Socket.Send(m.Data, m.Data.Length, SocketFlags.None);
                    
                    // Wait until sender has received the acknowledgement message
                    int readBytes;
                    do {
                        readBytes = m.Sce.Socket.Receive(recvBytes, 0, 4, SocketFlags.None);
                    } while( readBytes==0 || recvBytes[readBytes-1]!=0 );

                    DateTime end = DateTime.Now;
                    TaroSocketSendMessage.log.Write(LogLvl.Med,
                        "SendMessage {0} bytes from {1} to {2} ended at time {3:T}.{4:000} ({5:N0}ms since enqueue, {6:N0}ms since start)",
                        m.Data.Length, m.Sce.Socket.LocalEndPoint, m.Sce.Socket.RemoteEndPoint, end, end.Millisecond,
                        end.Subtract(m.enqueuedAt).TotalMilliseconds, end.Subtract(start).TotalMilliseconds);

                    // Increment counter of number of actual messages sent
                    localEvent.numMessagesSentToWire++;
                }
            }
            catch (SocketException e)
            {
                m.SendFailed = true;
                m.Sce.Failed = true;
                Console.Write("Socket {0}->{1} in SendMessage generated a SocketException:",
                              m.Sce.Ap.SrcEp, m.Sce.Ap.DestEp);
                Console.WriteLine(" TargetSite={0} Message={1}",
                                  e.TargetSite, e.Message);
            }
            catch (Exception e) 
            {
                m.SendFailed = true;
                m.Sce.Failed = true;
                Console.WriteLine("Error: {0}", e.ToString());
            }
            finally {
                m.Sce.Done();

                if (m.Async) {
                    // for async messages, we invoke the callback, if needed
                    if (m.UserCb != null) {
                        ReplyCallback cb = m.UserCb;
                        cb(m.UserDest,!(m.SendFailed),m.RmObj);
                    } else if (m.PingCb != null) {
                        TaroPingCallback cb = m.PingCb;
                        cb(m.PingDest,!(m.SendFailed));
                    }
                } else {
                    // for synchronous messages, we inform any thread who is waiting
                    // on this message that the send is complete
                    m.MessageDone.Set();
                }

                // call wakeup on the socket send queue to indicate that we've
                // finished using one of the sockets.  there may be other outgoing
                // messages waiting to use that socket.
                sq.Wakeup();

                // inform the mainSendThread that we're done
                threadFinished.Set();
            }
        }
    }

    public class TaroSocketSendMessage : TaroSendMessage {

        private int             msgCounter;
        private int             pingCounter;
        private CounterCallback callback;
        private TaroLocalEvent  localEvent;
        
        private SocketCache     sc;
        private Thread          sendMainThread;
        private SocketSendQueue sq;

        private const int MAX_THREADS = 50;
        private int numActiveThreads = 0;
        Thread[] workerThread = new Thread[MAX_THREADS];
        AutoResetEvent[] threadFinishedEvent = new AutoResetEvent[MAX_THREADS];

        public const string CATEGORY = "TaroSocketSend";
        public static Log log = new Log(CATEGORY);

        // Default constructor
        public TaroSocketSendMessage(TaroLocalEvent localEvent, object serializer, CounterCallback callback)
            : base( serializer)
        {
            this.localEvent = localEvent;
            this.callback = callback;
            this.msgCounter = 0;
            this.pingCounter = 0;
            this.sc = new SocketCache(localEvent);
            this.sq = new SocketSendQueue(this.sc);


            for (int i=0; i < MAX_THREADS; i++) {
                threadFinishedEvent[i] = new AutoResetEvent(false);
            }
            
            sendMainThread = new Thread(new ThreadStart(SendMain));
            sendMainThread.Start();
        }

        private void SendMain()
        {
            while (true) {
                if (numActiveThreads < MAX_THREADS) {
                    // block until a new message is available
                    OutgoingMessage m = sq.NextMessage();

                    int idx = -1;
                    
                    // find an empty thread slot
                    for (int i = 0; i < MAX_THREADS; i++) {
                        if (workerThread[i] == null) {
                            idx = i;
                            break;
                        }
                    }
                    
                    // at this point, we know its ok to create another thread since
                    // this is the only creation point for sender threads
                    SocketSendThread threadObj = new SocketSendThread(m, sc, sq, localEvent,
                                                                      threadFinishedEvent[idx]);
                    workerThread[idx] = new Thread(new ThreadStart(threadObj.SendMessage));
                    workerThread[idx].Start();
                    numActiveThreads++;
                }
                
                if (numActiveThreads == MAX_THREADS) {
                    int idx;
                    DateTime waitStartedAt = DateTime.Now;
                    log.Write(LogLvl.Med, "SendMain MAX_THREADS limit {0} hit at time {1:T}.{2:000}", MAX_THREADS,
                        waitStartedAt, waitStartedAt.Millisecond);

                    // block, waiting for any thread to signal completion   
                    idx = WaitHandle.WaitAny(threadFinishedEvent);
                    DateTime waitFinishedAt = DateTime.Now;
                    log.Write(LogLvl.Med, "SendMain MAX_THREADS wait finished at time {0:T}.{1:000} ({2:N0}ms since started)",
                        waitFinishedAt, waitFinishedAt.Millisecond, waitFinishedAt.Subtract(waitStartedAt).TotalMilliseconds);

                    workerThread[idx] = null;
                    numActiveThreads--;
                    // check for any remaining threads that have signalled completion 
                    // in this case, we avoid blocking with 0 timeout
                    while ((idx = WaitHandle.WaitAny(threadFinishedEvent,0,false)) != WaitHandle.WaitTimeout) {
                        workerThread[idx] = null;
                        numActiveThreads--;
                    }
                } else if (numActiveThreads < MAX_THREADS) {
                    int idx = 0;
                    // check for any threads that signal completion (avoid blocking)
                    while ((idx = WaitHandle.WaitAny(threadFinishedEvent,0,false)) != WaitHandle.WaitTimeout) {
                        workerThread[idx] = null;
                        numActiveThreads--;
                    }
                }
            }
        }
        
        private bool DoSend(NodeIdAddressBind source, NodeIdAddressBind destination,
                            MessageRouting.RouteMsg obj, bool async, ReplyCallback userCb)
        {
            // perform address mapping for modelnet: convert destination IP addresses of the 
            // form 10.0.x.y to 10.128.x.y
            string destinationAddr = destination.GetAddress.Address;
            if (TaroSocketConfig.USE_MODELNET && destinationAddr.StartsWith("10.0")) {
                string delimStr = ".";
                char [] delimiter = delimStr.ToCharArray();
                string [] split = null;
                
                split = destinationAddr.Split(delimiter, 5);
                destinationAddr="10.128." + split[2] + "." + split[3];
            }

            RouteMsg rm = (RouteMsg) obj;

            // Serialize the MessageRouting.RouteMsg 
            MemoryStream ms = new MemoryStream();
            lock (serializer) {
                if (serializer is XmlSerializer)
                    ((XmlSerializer) serializer).Serialize(ms, rm);
                else
                    ((BinaryFormatter) serializer).Serialize(ms, rm);
            }

            byte[] messageBytes;
            if (TaroSocketConfig.USE_COMPRESSION) 
            {
#if COMPRESSION_AVAILABLE
                ms.Position = 0;
                MemoryStream ms2 = new MemoryStream();
                CompressUtil.Compress(ms, ms2, CompressUtil.CompresionLevel.DefaultCompression);
                ms.Close(); 
                ms2.Close();
                messageBytes = ms2.ToArray();
#else
                Debug.Assert(false, "Message compression not supported in this release");
                messageBytes = null;
#endif
            } 
            else 
            {
                ms.Close();
                messageBytes = ms.ToArray();
                if (TaroSocketConfig.DUMP_XML && serializer is XmlSerializer) DumpXml(messageBytes);
            }

            // create an array of bytes with the appropriate format:
            //  [ <msg-type> <msg-length> <message> ]
            ms = new MemoryStream();
            BinaryWriter w = new BinaryWriter(ms);
            w.Write(TaroSocketConfig.TARO_MSG_APP);
            w.Write(messageBytes.Length);
            w.Write(messageBytes);
            w.Flush();
            w.Close();
            byte[] sendBytes = ms.ToArray();
            
            if (TaroSocketConfig.Debug) {
                Console.WriteLine("Debug: Sending message ({0} bytes) to {1}", 
                                  sendBytes.Length, destinationAddr);
            }
            OutgoingMessage m = null;
            if (async) {
                m = new OutgoingMessage(sendBytes, source.GetAddress.Address,
                                        destinationAddr, userCb, destination, obj);
            } else {
                m = new OutgoingMessage(sendBytes, source.GetAddress.Address,
                                        destinationAddr);
            }

            if (TaroSocketConfig.Debug) {
                m.enqueuedAt = DateTime.Now;
                log.Write(LogLvl.Med, "SendMessage {0} bytes from {1} to {2} enque at time {3:T}.{4:000}",
                          m.Data.Length, source.GetAddress.Address, destinationAddr, m.enqueuedAt, 
                          m.enqueuedAt.Millisecond);
            }
            
            sq.Enqueue(m);

            // increment the counter of attempted message sends, and
            // invoke the callback once for each attempted send
            msgCounter++;
            if (callback != null) {
                callback(msgCounter);
            }

            if (!async) {
                // wait for a signal that the message has been acknowledged by
                // the receiver
                m.MessageDone.WaitOne();
                if (userCb != null) {
                    userCb(destination, !(m.SendFailed), obj);
                }
                
                // for synchronous calls, the return value is meaningful
                return (!(m.SendFailed));
            }
            
            // async calls return true unless the failure happens early on...
            return true;
        }
        
        private bool DoPing(NodePhysicalAddress source, NodePhysicalAddress destination,
                            TaroPing ping, bool async, TaroPingCallback pingCb) 
        {
            // perform address mapping for modelnet: convert destination IP addresses of the 
            // form 10.0.x.y to 10.128.x.y
            string destinationAddr = destination.Address;
            if (TaroSocketConfig.USE_MODELNET && destinationAddr.StartsWith("10.0")) {
                string delimStr = ".";
                char [] delimiter = delimStr.ToCharArray();
                string [] split = null;
                
                split = destinationAddr.Split(delimiter, 5);
                destinationAddr="10.128." + split[2] + "." + split[3];
            }

            // Serialize the ping
            MemoryStream ms = new MemoryStream();
            lock (serializer) {
                if (serializer is XmlSerializer)
                    ((XmlSerializer) serializer).Serialize(ms, ping);
                else
                    ((BinaryFormatter) serializer).Serialize(ms, ping);
            }

            byte[] messageBytes = ms.ToArray();
            if (TaroSocketConfig.USE_COMPRESSION) 
            {
#if COMPRESSION_AVAILABLE
                ms.Position = 0;
                MemoryStream ms2 = new MemoryStream();
                CompressUtil.Compress(ms, ms2, CompressUtil.CompresionLevel.DefaultCompression);
                ms.Close(); 
                ms2.Close();
                messageBytes = ms2.ToArray();
#else
                Debug.Assert(false, "Message compression not supported in this release");
#endif
            } 
            else 
            {
                ms.Close();
                messageBytes = ms.ToArray();
                if (TaroSocketConfig.DUMP_XML && serializer is XmlSerializer) DumpXml(messageBytes);
            }

            ms = new MemoryStream();
            BinaryWriter w = new BinaryWriter(ms);
            w.Write(TaroSocketConfig.TARO_MSG_PING);
            w.Write(messageBytes.Length);
            w.Write(messageBytes);
            w.Flush();
            w.Close();
            byte[] sendBytes = ms.ToArray();

            if (TaroSocketConfig.Debug) {
                Console.WriteLine("Debug: Sending ping ({0} bytes) to {1}", 
                                  sendBytes.Length, destinationAddr);
            }
            OutgoingMessage m = null;
            if (async) {
                m = new OutgoingMessage(sendBytes, source.Address,
                                        destinationAddr, pingCb, destination);
            } else {
                m = new OutgoingMessage(sendBytes, source.Address,
                                        destinationAddr);
            }
            
            sq.Enqueue(m);

            // increment the counter of attempted ping sends
            pingCounter++;

            if (!async) {
                // wait for a signal that the ping message has been acknowledged by
                // the receiver
                m.MessageDone.WaitOne();
                if (pingCb != null) {
                    pingCb(destination, !(m.SendFailed));
                }
                
                // for synchronous calls, the return value is meaningful
                return (!(m.SendFailed));
            }
            
            // async calls return true unless the failure happens early on...
            return true;
        }
        
        // Send a message on a TCP socket from the source to the destination
        public override bool Send(NodeIdAddressBind source, NodeIdAddressBind destination,
                                  MessageRouting.RouteMsg obj)
        {
            // call DoSend, inform it that this is a synchronous call
            bool success = DoSend(source, destination, obj, false, null);

            return success;
        }

        // Asynchronous version of Send
        public override void BeginSend(NodeIdAddressBind source, NodeIdAddressBind destination,
                                       MessageRouting.RouteMsg obj, ReplyCallback cb)
        {
#if SYNCHRONOUS
            // NOTUSED - just for debugging purposes, make all calls synchronous
            bool success = DoSend(source, destination, obj, false, null);
            cb(destination, success, obj);
            return;
#else
            // this is the correct asynchronous version
            bool success = DoSend(source, destination, obj, true, cb);
            return;
#endif
        }

        // Send a ping message
        public override bool Ping(NodePhysicalAddress source, NodePhysicalAddress dest) 
        {
            TaroPing ping = new TaroPing();
            ping.from = source.Address;
            bool success = DoPing(source, dest, ping, false, null);
            return success;
        }
        
        // Send an asynchronous ping message
        public override void BeginPing( NodePhysicalAddress source, NodePhysicalAddress dest, 
                                        TaroPingCallback cb) 
        {
#if SYNCHRONOUS
            // NOTUSED - just for debugging purposes, make all calls synchronous
            TaroPing ping = new TaroPing();
            ping.from = source.Address;
            bool success = DoPing(source, dest, ping, false, null);
            cb(dest,success);
            return;
#else
            // this is the correct asynchronous version
            TaroPing ping = new TaroPing();
            ping.from = source.Address;
            DoPing(source, dest, ping, true, cb);
            return;
#endif
        }

        private void DumpXml(byte[] messageBytes)
        {
            Console.WriteLine("BeginXml");
            foreach (byte b in messageBytes) 
            {
                Console.Write(Convert.ToChar(b));
            }
            Console.WriteLine("EndXml");
        }
    }
}
