using System;
using System.Collections;
using System.Diagnostics;
using MessageRouting;

namespace Microsoft.Research.Taro
{
    /// <summary>
    /// The TaroTopology class implements the MessageRouting.Topology class. It provides an
    /// implementation of the Distance() function which caches the result of the ReallyCheckDistance()
    /// function.
    /// 
    /// TODO: The internal cache does not properly eject stale entries.
    /// </summary>
    public class TaroTopology : MessageRouting.Topology {

        // Nested classes
        internal class CacheEntry {
            public DateTime     expireAt;
            public int          distance;
        };

        // Class members
        protected TaroSendMessage   sendMessage;
        protected Node              localNode;
        protected Hashtable         cache;
        Ping                        ping;

        // Constants
        protected const int         CACHE_LIFETIME_SEC = 2 * 60;    // Cache distances for 2 minutes

        // Constructor
        public TaroTopology( TaroSendMessage sm, Node localNode, bool useIcmp ) {
            this.sendMessage = sm;
            this.localNode = localNode;
            this.cache = new Hashtable();

            if( useIcmp ) {
                ping = new Ping();
            } else {
                ping = null;
            }
        }

        // Measure the distance between a and b, relative to the topology of the network and an
        // appropriate distance measure
        public int Ping(NodePhysicalAddress a, NodePhysicalAddress b, bool disableCache) {
            
            // We can only determine the distance between us and another node
            Debug.Assert( a==localNode.GetPhysicalAddr );
        
            if( ! disableCache ) {
                CacheEntry c = GetCacheEntry(b);
                if( c!=null ) {
                    // Found a cached entry!
                    //Console.WriteLine("Distance to {0} is {1}ms (CACHED)", b.IPAddress, c.distance );
                    return c.distance;
                }
            }
            
            int dist = CheckDistanceSync(a,b);
            InsertCacheEntry(b,dist);

            //Console.WriteLine("Distance to {0} is {1}ms (NEW)", b.IPAddress, dist );
            return dist;
        }

        // Actually check the distance
        int CheckDistanceSync(NodePhysicalAddress a, NodePhysicalAddress b) {
            bool alive, forceOneDistance=false;
            int duration;
            int err;

            Debug.Assert( a==localNode.GetPhysicalAddr );

            // If the two nodes are on the same machine then force the distance to 1
            string hostname1 = ((TaroAddress) a).GetHostname();
            string hostname2 = ((TaroAddress) b).GetHostname();
            if( hostname1==hostname2 ) {
                forceOneDistance=true;
            }

            // Send the ping
            if( ping!=null ) {
                // Measure the RTT of an ICMP 'Ping'
                duration = (int) ping.SendPingHost(hostname2, out err);
                alive = (err==0);
            } else {
                // Measure the RTT of an explicit 'Ping' message
                DateTime start = DateTime.Now;
                alive = sendMessage.Ping( a, b );
                DateTime end = DateTime.Now;
                TimeSpan diff = end.Subtract( start );
                duration = (int) diff.TotalMilliseconds;
            }

            // Return the result
            if( alive ) {
                //Console.WriteLine( "RealDistance to {0} is {1}ms", b.IPAddress, duration );
                if( forceOneDistance || (duration==0) ) {
                    return 1;
                } else {
                    InsertCacheEntry(b, duration);
                    return duration;
                }
            } else {
                Console.WriteLine("** Node {0} is dead **", b.Address);
                return -1;
            }

        }

        // Asynchronously measure the distance between a and b.
        public void BeginPing(NodePhysicalAddress a, NodePhysicalAddress b, bool disableCache, PingCallback cb) {
            bool forceOneDistance=false;

            Debug.Assert( a==localNode.GetPhysicalAddr );
            Debug.Assert( cb!=null );

            // First try the cache, if it is allowed
            if( ! disableCache ) {
                CacheEntry entry = GetCacheEntry(b);
                if( entry!=null ) {
                    cb(b, entry.distance);
                    return;
                }
            }

            // If the two nodes are on the same machine then force the distance to 1
            string hostname1 = ((TaroAddress) a).GetHostname();
            string hostname2 = ((TaroAddress) b).GetHostname();
            if( hostname1==hostname2 ) {
                forceOneDistance=true;
            }

            // Start a Taro 'Ping'
            DateTime start = DateTime.Now;
            AsyncPingCont continuation = new AsyncPingCont(this, cb, start, forceOneDistance);
            TaroPingCallback internalCB = new TaroPingCallback(continuation.Callback);
            sendMessage.BeginPing(a, b, internalCB);
        }

        // Continuation for asynchronous pings
        class AsyncPingCont {
            TaroTopology    topo;
            PingCallback    cb;
            DateTime        start;
            bool            forceOneDistance;

            public AsyncPingCont(TaroTopology topo, PingCallback cb, DateTime start, bool forceOneDistance) {
                this.topo = topo;
                this.cb = cb;
                this.start = start;
                this.forceOneDistance = forceOneDistance;
            }

            // Called when the Ping has completed
            public void Callback(NodePhysicalAddress destination, bool success) {
                DateTime end = DateTime.Now;
                TimeSpan diff = end.Subtract(start);
                int duration = (int) diff.TotalMilliseconds;

                // Return the result
                if( success ) {
                    //Console.WriteLine( "RealDistance to {0} is {1}ms", b.IPAddress, duration );
                    if( forceOneDistance || (duration==0) ) {
                        duration = 1;
                    } else {
                        topo.InsertCacheEntry(destination, duration);
                    }
                } else {
                    //Console.WriteLine("** Node {0} is dead **", b.IPAddress);
                    duration = -1;
                }

                cb(destination, duration);
            }
        }

        // This must return another seed node
        public NodePhysicalAddress GetClosest(NodePhysicalAddress n) {
            throw new CheatingNotAllowedException();
        }

        // Try to find the distance in our cache
        internal CacheEntry GetCacheEntry(NodePhysicalAddress n) {
            lock( this ) {
                CacheEntry c = (CacheEntry) cache[n];
                if( c==null ) {
                    // No cached entry
                    return null;
                }
                if( c.expireAt.CompareTo(DateTime.Now) < 0 ) {
                    // Entry has expired
                    //Console.WriteLine("Cached distance to {0} has expired!", n.IPAddress);
                    return null;
                }
                return c;
            }
        }

        // Insert an entry into the cache
        internal void InsertCacheEntry(NodePhysicalAddress b, int dist) {
            lock( this ) {
                CacheEntry c = new CacheEntry();
                c.distance = dist;
                c.expireAt = DateTime.Now.AddSeconds(CACHE_LIFETIME_SEC);
                cache[b] = c;
            }
        }

    }

    public class CheatingNotAllowedException : System.Exception {
    }
}
