using System;
using System.Collections;
using System.Threading;
using System.Xml.Serialization;

using MessageRouting;

namespace Overlook 
{
//========================= API message base classes =========================//

    /// <summary>
    /// All Overlook client requests derive from DirApiRequestMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("DirApiRequestMsg",Namespace=SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)]
    abstract public class DirApiRequestMsg: OverlayRequestMsg 
    {
        private NodeId destId;

        // Constructor - the destination node id
        public DirApiRequestMsg(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest)
            : base(localNode, cb, 30 * 1000 /* delayMs */, false)
        {
            this.destId = dest;
            this.Source = source;
        }

        [XmlElementAttribute("Destination",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public NodeId Destination 
        {
            set { destId = value; }
            get { return destId; }
        }
    }

    /// <summary>
    /// All Overlook client replies derive from DirApiReplyMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("DirApiReplyMsg",Namespace=SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)]
    public class DirApiReplyMsg: OverlayReplyMsg 
    {
        private NodeIdAddressBind source;
        public object result;

        public DirApiReplyMsg(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
            : base(requestMsg)
        {
            this.source = source;
            this.result = result;
        }

        [XmlElementAttribute("Source",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public NodeIdAddressBind Source 
        {
            set { source = (NodeIdAddressBind) value; }
            get { return source; }
        }
    }

//========================= DirAPI class =========================//

    public abstract class DirAPI
    {

        private Node myServer;
        private SendMessage sendMsg;

        public DirAPI(Node myServer, SendMessage msg) 
        {
            this.myServer = myServer;
            this.sendMsg = msg;
        }

//========================= mkdir =========================//

        public class MkDirApiRequest : DirApiRequestMsg 
        {
            public string fullPathName;

            public MkDirApiRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, string fullPathName)
                : base(localNode, cb, source, dest)
            {
                this.fullPathName = fullPathName;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.MkDir(this);
            }
        }

        public class MkDirApiReply : DirApiReplyMsg
        {
            public MkDirApiReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        /// Create a directory.
        public void MkDir(string fullPathName, MkDirCallback callback) 
        {
            if (DirApplication.DebugOutput) Console.WriteLine("mkdir {0}", fullPathName);

            MkDirApiRequest request = new MkDirApiRequest(myServer, callback,
                myServer.GetNodeIdAddressBind, myServer.GetNodeIdAddressBind.GetNodeId, fullPathName);
            myServer.SendMessageDirect(request, myServer.GetNodeIdAddressBind);
        }

        public abstract class MkDirCallback: OverlayReplyCallback
        {
            public Object result;
            public override void Execute(Node localNode)
            {
                if (DirApplication.DebugOutput) Console.WriteLine("DirAPI.MkDirCallback");
                if( this.Reply!=null ) {
                    // Successfully received a reply
                    result = ((MkDirApiReply) this.Reply).result;
                } else {
                    // Timed out
                    result = new DirException("MkDir Timed Out");
                }
                Callback();
            }
            public abstract void Callback(); // Client code needs to provide this method
        }

//========================= rmdir =========================//

        public class RmDirApiRequest : DirApiRequestMsg 
        {
            public string fullPathName;

            public RmDirApiRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, string fullPathName)
                : base(localNode, cb, source, dest)
            {
                this.fullPathName = fullPathName;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.RmDir(this);
            }
        }

        public class RmDirApiReply : DirApiReplyMsg
        {
            public RmDirApiReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        /// Delete a directory.
        public void RmDir(string fullPathName, RmDirCallback callback) 
        {
            if (DirApplication.DebugOutput) Console.WriteLine("rmdir {0}", fullPathName);

            RmDirApiRequest request = new RmDirApiRequest(myServer, callback,
                myServer.GetNodeIdAddressBind, myServer.GetNodeIdAddressBind.GetNodeId, fullPathName);
            myServer.SendMessageDirect(request, myServer.GetNodeIdAddressBind);
        }

        public abstract class RmDirCallback: OverlayReplyCallback 
        {
            public Object result;
            public override void Execute(Node localNode) 
            {
                if (DirApplication.DebugOutput) Console.WriteLine("DirAPI.RmDirCallback");
                if( this.Reply!=null ) {
                    // Received a reply
                    result = ((RmDirApiReply) this.Reply).result;
                } else {
                    // Timed out
                    result = new DirException("RmDir Timed Out");
                }
                Callback();
            }
            public abstract void Callback(); // Client code needs to provide this method
        }

//========================= ListDir =========================//

        public class ListDirApiRequest : DirApiRequestMsg 
        {
            public string fullPathName;

            public ListDirApiRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, string fullPathName)
                : base(localNode, cb, source, dest)
            {
                this.fullPathName = fullPathName;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ListDir(this);
            }
        }

        public class ListDirApiReply : DirApiReplyMsg
        {
            public ListDirApiReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        /// List a directory.
        /// Returns an array of directory entry descriptor objects.
        /// Fails if the directory does not exist.
        public void ListDir(string fullPathName, ListDirCallback callback) 
        {
            if (DirApplication.DebugOutput) Console.WriteLine("listdir {0}", fullPathName);

            ListDirApiRequest request = new ListDirApiRequest(myServer, callback,
                myServer.GetNodeIdAddressBind, myServer.GetNodeIdAddressBind.GetNodeId, fullPathName);
            myServer.SendMessageDirect(request, myServer.GetNodeIdAddressBind);
        }

        public abstract class ListDirCallback: OverlayReplyCallback 
        {
            public Object result;
            public override void Execute(Node localNode) 
            {
                if (DirApplication.DebugOutput) Console.WriteLine("DirAPI.ListDirCallback");
                if( this.Reply!=null ) {
                    // Received a reply
                    result = ((ListDirApiReply) this.Reply).result;
                    if (result is DirEntryArray) {
                        DirEntry[] entries = (DirEntry[]) ((DirEntryArray) result).Entries;
                        result = entries;
                    }
                    // else return exception or other value as result
                } else {
                    // Timed out
                    result = new DirException("ListDir Timed Out");
                }
                Callback();
            }
            public abstract void Callback(); // Client code needs to provide this method
        }

//========================= GetNameValue =========================//

        public class GetNameValueApiRequest : DirApiRequestMsg 
        {
            public NameValueEntry nameValueEntry;

            public GetNameValueApiRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, NameValueEntry nameValueEntry)
                : base(localNode, cb, source, dest)
            {
                this.nameValueEntry = nameValueEntry;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.GetNameValue(this);
            }
        }

        public class GetNameValueApiReply : DirApiReplyMsg
        {
            public GetNameValueApiReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        /// Return a {name, value} entry from a directory.
        public void GetNameValue(string dirPathName, string name, GetNameValueCallback callback) 
        {
            if (DirApplication.DebugOutput) Console.WriteLine("getNameValue {0}, {1}", dirPathName, name);
            NameValueEntry e = new NameValueEntry(dirPathName, name);

            GetNameValueApiRequest request = new GetNameValueApiRequest(myServer, callback,
                myServer.GetNodeIdAddressBind, myServer.GetNodeIdAddressBind.GetNodeId, e);
            myServer.SendMessageDirect(request, myServer.GetNodeIdAddressBind);
        }

        public abstract class GetNameValueCallback: OverlayReplyCallback 
        {
            public Object result;
            public override void Execute(Node localNode) 
            {
                if (DirApplication.DebugOutput) Console.WriteLine("DirAPI.GetNameValueCallback");
                if( this.Reply!=null ) {
                    // Received a reply
                    result = ((GetNameValueApiReply) this.Reply).result;
                } else {
                    // Timed out
                    result = new DirException("GetNameValue Timed Out");
                }
                Callback();
            }
            public abstract void Callback(); // Client code needs to provide this method
        }

//========================= UpdateNameValue =========================//

        public class UpdateNameValueApiRequest : DirApiRequestMsg 
        {
            public NameValueEntry nameValueEntry;

            public UpdateNameValueApiRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, NameValueEntry nameValueEntry)
                : base(localNode, cb, source, dest)
            {
                this.nameValueEntry = nameValueEntry;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.UpdateNameValue(this);
            }
        }

        public class UpdateNameValueApiReply : DirApiReplyMsg
        {
            public UpdateNameValueApiReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        /// Add/update a {name, value} pair in a directory.
        public void UpdateNameValue(string dirPathName, string name, string val, UpdateNameValueCallback callback) 
        {
            if (DirApplication.DebugOutput) Console.WriteLine("updateNameValue {0}, {1}, {2}", dirPathName, name, val);
            NameValueEntry e = new NameValueEntry(dirPathName, name, DirEntryType.nameValuePair, val);

            UpdateNameValueApiRequest request = new UpdateNameValueApiRequest(myServer, callback,
                myServer.GetNodeIdAddressBind, myServer.GetNodeIdAddressBind.GetNodeId, e);
            myServer.SendMessageDirect(request, myServer.GetNodeIdAddressBind);
        }

        public abstract class UpdateNameValueCallback: OverlayReplyCallback 
        {
            public Object result;
            public override void Execute(Node localNode)
            {
                if (DirApplication.DebugOutput) Console.WriteLine("DirAPI.UpdateNameValueCallback");
                if( this.Reply!=null ) {
                    // Received a reply
                    result = ((UpdateNameValueApiReply) this.Reply).result;
                } else {
                    // Timed out
                    result = new DirException("UpdateNameValue Timed Out");
                }
                Callback();
            }
            public abstract void Callback(); // Client code needs to provide this method
        }

        public abstract void ReceivedEvent(DirectoryId dirId, Object e, NodeIdAddressBind from);
    }
}
