using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using SkipNet = Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Common;

using Microsoft.Research.Herald.Logger;

namespace Overlook {

    public class SerializationConfig
    {
        public const string OVERLOOK_NAMESPACE = "http://research.microsoft.com/sn/Herald/Overlook";
        public static Type [] serializationTypes = new Type[]
        {
            typeof(DirRequestMsg),
            typeof(DirReplyMsg),
            typeof(DirAsyncMsg),
            typeof(DirException),
            typeof(DirApplication.MkDirRequest),
            typeof(DirApplication.MkDirReply),
            typeof(DirApplication.RmDirRequest),
            typeof(DirApplication.RmDirReply),
            typeof(DirApplication.ListDirRequest),
            typeof(DirApplication.ListDirReply),
            typeof(DirApplication.GetNameValueRequest),
            typeof(DirApplication.GetNameValueReply),
            typeof(DirApplication.UpdateNameValueRequest),
            typeof(DirApplication.UpdateNameValueReply),
            typeof(DirApplication.CreateReplicaRequest),
            typeof(DirApplication.CreateReplicaReply),
            typeof(DirApplication.MovePrimaryRequest),
            typeof(DirApplication.MovePrimaryReply),
            typeof(DirApplication.UpdateReplicaRequest),
            typeof(DirApplication.UpdateReplicaReply),
            typeof(DirApplication.ResyncReplicaRequest),
            typeof(DirApplication.ResyncReplicaReply),
            typeof(DirApplication.HighLatencyHintMsg),
            typeof(DirApplication.ReplicaHeartBeatMsg),
            typeof(DirApplication.VerifyPrimaryMsg),
            typeof(DirApplication.RmReplicaMsg),
            typeof(ValueEntry),
            typeof(NameValueEntry),
            typeof(DirReplicaDescriptor),
            typeof(DirUpdateInfo),
            typeof(ResyncDescriptor)
        };
    }

//========================= Server message base classes =========================//

    /// <summary>
    /// All Overlook server request messages derive from DirRequestMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("DirRequestMsg",Namespace=SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)]
    abstract public class DirRequestMsg: OverlayRequestMsg
    {
        private NodeId dest;
        private NodeIdAddressBind last, prevLast;

        public DirRequestMsg() {} // For XML Serialization only

        // Constructor - the destination node id
        public DirRequestMsg(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest)
            : base(localNode, cb, 10 * 1000 /* TimeoutMs */, false)
        {
            this.dest = dest;
            this.Source = source;
        }

        // This is called whenever a message is routed through the overlay
        // to a new node
        public override bool CheckContinueRouting(MessageRouting.Node localNode)
        {
            DirRequestMsg msg = this;
            msg.AddNodeId(localNode.GetNodeIdAddressBind);
            // if (DirApplication.DebugOutput) Console.WriteLine("CheckContinueRouting through {0}", localNode.GetNodeIdAddressBind.GetAddress);

            // Replicas process all requests EXCEPT:
            // - updateNameValue, movePrimary, verifyPrimary, and rmDir requests.
            //   Those go to the primary replica node to be processed.
            // - resyncReplica requests originating from the same node.

            if (!(msg.Destination is DirectoryId)) return true; // Continue if not sent to a directory

            DirectoryId dirId = (DirectoryId) msg.Destination;
            DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
            if (! dirApp.directories.Contains(dirId)) return true; // Continue if we don't have a copy of this directory

            DirectoryDescriptor dir = (DirectoryDescriptor)dirApp.directories[dirId];
            if (dir.Orphaned) return true;  // Don't use orphaned directories

            if (msg is DirApplication.UpdateNameValueRequest) return true; // Goes to primary
            if (msg is DirApplication.MovePrimaryRequest) return true; // Goes to primary
            if (msg is DirApplication.RmDirRequest) return true; // Goes to primary
                    
            if (msg.Source.Equals(localNode.GetNodeIdAddressBind) &&  msg is DirApplication.ResyncReplicaRequest)
                return true; // Don't service ResyncReplica at ourself

            if (DirApplication.DebugOutput) Console.WriteLine("CheckContinueRouting returning false for {0} on node {1}",
                                                this, localNode.GetNodeIdAddressBind.GetAddress);
            this.ArrivedAtDestination(localNode);
            return false;
        }

        [XmlElementAttribute("Destination",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public NodeId Destination 
        {
            set { dest = value; }
            get { return dest; }
        }

        [XmlElementAttribute("Last",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public NodeIdAddressBind Last 
        {
            get {return last;}
            set {last = value;}
        }

        [XmlElementAttribute("PrevLast",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public NodeIdAddressBind PrevLast 
        {
            get {return prevLast;}
            set {prevLast = value;}
        }

        public NodeIdAddressBind LastNodeId() 
        {
            if (prevLast != null) return prevLast;
            else return null;
        }

        public void AddNodeId(NodeIdAddressBind next) 
        {
            prevLast = last;
            last = next;
        }
    }

    /// <summary>
    /// All Overlook server reply messages derive from DirReplyMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("DirReplyMsg",Namespace=SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)]
    public class DirReplyMsg: OverlayReplyMsg 
    {
        private NodeIdAddressBind source;
        public object result;

        public DirReplyMsg() {} // For XML Serialization only

        public DirReplyMsg(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
            : base(requestMsg)
        {
            this.source = source;
            this.result = result;
        }

        [XmlElementAttribute("Source",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public NodeIdAddressBind Source 
        {
            set { source = value; }
            get { return source; }
        }
    }

    /// <summary>
    /// All Overlook server asynchronous messages derive from DirAsyncMsg.
    /// </summary>
    [Serializable,
    XmlRootAttribute("DirAsyncMsg",Namespace=SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)]
    abstract public class DirAsyncMsg: OverlayMsg
    {
        private NodeIdAddressBind source;
        private NodeId dest;

        public DirAsyncMsg() {} // For XML Serialization only

        // Constructor - the destination node id
        public DirAsyncMsg(NodeIdAddressBind source, NodeId dest)
        {
            this.source = source;
            this.dest = dest;
        }
       
        [XmlElementAttribute("Source",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public NodeIdAddressBind Source 
        {
            get { return source; }
            set { source = value; }
        }

        [XmlElementAttribute("Destination",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public NodeId Destination 
        {
            set { dest = value; }
            get { return dest; }
        }

        // We count on the default CheckContinueRouting behavior that always returns true.
        // Needed in particular for VerifyPrimaryMsg, which must be routed to the primary.
    }

    /// <summary>
    /// Overlook methods can return instances of this exception.
    /// </summary>
    [Serializable,
    XmlRootAttribute("DirException",Namespace=SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)]
    public class DirException 
    {
        public DirException() 
        {
        }

        public DirException(string errMsg) 
        {
            this.errMsg = errMsg;
        }

        private string errMsg;

        [XmlElementAttribute("ErrMsg",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public string ErrMsg 
        {
            get {return this.errMsg;}
            set {this.errMsg = value;}
        }
    }
    
//========================= DirApplication class =========================//

    /// <summary>
    /// The main Overlook server class.
    /// </summary>
    public class DirApplication: Application {
        public const string CATEGORY = "Overlook";
        private Log log = new Log(CATEGORY);

        private static int defaultReplicationNumber = 1;
        private static Random rng = new Random(0);

        public Node myServer;
        public SendMessage sendMsg;
        public LocalEvent local;
        public DirAPI dirApi;
        static public bool useCLB;

        public Hashtable directories;

        private static int[] numOperations = new int[NumDirOperations];
        private static long[] operationTimes = new long[NumDirOperations];
        private static long[] operationTimes2 = new long[NumDirOperations];
        private static int [] operationMin = new int[NumDirOperations];
        private static int [] operationMax = new int[NumDirOperations];
        private static Histogram[] operationHisto = new Histogram[NumDirOperations];

        private static bool replicationEnabled = true;
        private static bool highLatencyReplicationEnabled = true;
        private static bool debugOutput = false;
        private static bool debugTimingOutput = false;
        private static bool debugMsgIdOutput = false;

        private static bool debugResyncReplica = false;  // True to enable code coverage of some failure code paths
        public static int resyncReplicaCounter = 0;       // Used to simulate occasional silent parent replicas

        private static bool debugHighLatencyReplication = false; // True to enable code coverage of latency-based replication
        public static int highLatencyReplicaCounter = 0;    // Used to simulate occasional high latency get operations

        private const long highLatencyReplicationThreshold = 4 * 1000;  // 4 seconds
        private const long highLatencyReplicationExpirationLimit = 10 * 1000; // 10 seconds
        private const long highLatencyReplicationMaxUsage = 20;

        public Hashtable cachedDirs;
        public static int numCachedEntryHits;

        private HeartbeatTimer heartbeatTimer;
        private MaintenanceTimer maintenanceTimer;

        public enum DirOperations {mkDir, rmDir, listDir, getNameValue, updateNameValue, createReplica, updateReplica, resyncReplica, movePrimary};
        private const int NumDirOperations = ((int)(DirOperations.movePrimary)) + 1;

        static DirApplication() {
            int numBuckets = 40;
            int[] bucketBounds = new int[numBuckets];
            for (int i = 0; i < numBuckets; i++) {
                if (((i + 1) * 100) <= 2000)
                    bucketBounds[i] = (i + 1) * 100;
                else
                    bucketBounds[i] = (i + 1) * 1000;
            }

            for (int i = 0; i < NumDirOperations; i++) {
                operationMin[i] = int.MaxValue;
                operationMax[i] = int.MinValue;
                operationHisto[i] = new Histogram(bucketBounds);
            }
        }

        public void DirectoryInit(Node myServer, SendMessage s, LocalEvent l, DirAPI dirApi, bool useCLB) {
            this.myServer = myServer;
            this.sendMsg = s;
            this.local = l;
            this.dirApi = dirApi;
            DirApplication.useCLB = useCLB;

            this.directories = new Hashtable();

            if (highLatencyReplicationEnabled) {
                this.cachedDirs = new Hashtable();
            }

            // Start timers
            long initialDelayMs = (long) (rng.NextDouble() * DirectoryDescriptor.heartBeatTimeoutInterval)+1;
            heartbeatTimer = new HeartbeatTimer( initialDelayMs, DirectoryDescriptor.heartBeatTimeoutInterval );
            myServer.SetTimer( heartbeatTimer );

            initialDelayMs = (long) (rng.NextDouble() * DirectoryDescriptor.maintIntervalLength)+1;
            maintenanceTimer = new MaintenanceTimer( initialDelayMs, DirectoryDescriptor.maintIntervalLength );
            myServer.SetTimer( maintenanceTimer );
        }

        /// <summary>
        /// Periodic timer to do heartbeats for each directory descriptor
        /// </summary>
        class HeartbeatTimer : Timer {
            public HeartbeatTimer( long initialDelayMs, long periodMs )
                : base( initialDelayMs, periodMs )
            {}

            public override void Execute( Node localNode ) {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                foreach( DirectoryDescriptor dir in dirApp.directories.Values ) {
                    dir.HeartbeatTimerFired();
                }
            }
        }

        /// <summary>
        /// Periodic timer to maintain counters for each directory descriptor
        /// </summary>
        class MaintenanceTimer : Timer {
            public MaintenanceTimer( long initialDelayMs, long periodMs )
                : base( initialDelayMs, periodMs )
            {}
            
            public override void Execute( Node localNode ) {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                foreach( DirectoryDescriptor dir in dirApp.directories.Values ) {
                    dir.MaintenanceTimerFired();
                }            
            }
        }

        public void ResetDirectoryNode() {
        }

        // Reset static parts of DirApplication completely.
        public static void ResetDirApplication() {
            ResetDirApplicationStats();
        }

        // Reset static statistics parts of DirApplication.
        public static void ResetDirApplicationStats() {
            for (int i = 0; i < NumDirOperations; i++) {
                numOperations[i] = 0;
                operationTimes[i] = 0;
                operationTimes2[i] = 0;
                operationMin[i] = int.MaxValue;
                operationMax[i] = int.MinValue;
                operationHisto[i].Clear();
            }
            numCachedEntryHits = 0;
        }

        public static bool ReplicationEnabled {
            get {return replicationEnabled;}
            set {replicationEnabled = value;}
        }

        public static int DefaultReplicationNumber {
            get {return defaultReplicationNumber;}
            set {defaultReplicationNumber = value;}
        }

        public static bool HighLatencyReplicationEnabled {
            get {return highLatencyReplicationEnabled;}
            set {highLatencyReplicationEnabled = value;}
        }

        public static bool DebugOutput 
        {
            get {return debugOutput;}
            set {debugOutput = value;}
        }

        public static bool DebugTimingOutput 
        {
            get {return debugTimingOutput;}
            set {debugTimingOutput = value;}
        }

        public static bool DebugMsgIdOutput 
        {
            get {return debugMsgIdOutput;}
            set {debugMsgIdOutput = value;}
        }

        public static bool DebugResyncReplica 
        {
            get {return debugResyncReplica;}
            set {debugResyncReplica = value;}
        }

        public static bool DebugHighLatencyReplication 
        {
            get {return debugHighLatencyReplication;}
            set {debugHighLatencyReplication = value;}
        }

        public static int[] NumOperations 
        {
            get {return numOperations;}
        }

        public static long[] OperationTimes {
            get {return operationTimes;}
        }

        public static long[] OperationTimesSquared {
            get {return operationTimes2;}
        }

        public static int[] OperationMin {
            get {return operationMin;}
        }

        public static int[] OperationMax {
            get {return operationMax;}
        }

        public static Histogram[] OperationHistogram {
            get {return operationHisto;}
        }

//========================= DirAppCallback =========================//

        public abstract class DirAppCallback : OverlayReplyCallback 
        {
            public OverlayRequestMsg requestMsg; // For use in constructing reply message
            public DirApplication app;
            public long opnStartTime, opnTime;

            public DirAppCallback(OverlayRequestMsg requestMsg, DirApplication app, long opnStartTime) 
            {
                this.requestMsg = requestMsg;
                this.app = app;
                this.opnStartTime = opnStartTime;
            }

            public DirAppCallback(DirApplication app) 
            {
                this.requestMsg = null; // No client request involved
                this.app = app;
                this.opnStartTime = app.local.GetTime();
            }

            public void UpdateStats(DirOperations opn) 
            {
                long opnEndTime = this.app.local.GetTime();
                opnTime = opnEndTime - this.opnStartTime;
                numOperations[(int)opn]++;
                operationTimes[(int)opn] += opnTime;
                operationTimes2[(int)opn] += opnTime * opnTime;
                if (operationMin[(int)opn] > (int) opnTime) operationMin[(int)opn] = (int) opnTime;
                if (operationMax[(int)opn] < (int) opnTime) operationMax[(int)opn] = (int) opnTime;
                operationHisto[(int)opn].Update((int)opnTime);
            }
        }

//========================= mkdir =========================//

        [Serializable, XmlRootAttribute("MkDirRequest", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class MkDirRequest : DirRequestMsg
        {
            public string fullPathName;

            public MkDirRequest() {} // For XML Serialization only

            public MkDirRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, string fullPathName)
                : base(localNode, cb, source, dest)
            {
                this.fullPathName = fullPathName;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerMkDir(this);
            }
        }

        [Serializable, XmlRootAttribute("MkDirReply", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class MkDirReply : DirReplyMsg
        {
            public MkDirReply() {} // For XML Serialization only

            public MkDirReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        // Client request for mkdir.
        public void MkDir(DirAPI.MkDirApiRequest msg) {
            string fullPathName = msg.fullPathName;
            DirectoryId dirId = new DirectoryId(fullPathName);
            if (debugOutput) Console.WriteLine("mkdir on node {0}: {1} ({2})", myServer.GetNodeIdAddressBind.GetAddress, fullPathName, dirId);

            if (!IsRootPathName(fullPathName)) {
                // Insert into parent directory.
                int indx = fullPathName.LastIndexOf("/");
                string parentDir;
                if (indx == 0)
                    parentDir = "/";
                else 
                {
                    parentDir = fullPathName.Substring(0, indx);
                    if (fullPathName.LastIndexOf("!/") == indx - 1)
                        parentDir += "/"; // CLB root
                }
                string relDirName = fullPathName.Substring(indx + 1);

                DirectoryId dirId1 = new DirectoryId(parentDir);
                NameValueEntry r = new NameValueEntry(parentDir, relDirName, DirEntryType.dirEntry, relDirName);

                MkDirParentCallback cb = new MkDirParentCallback(msg, this, fullPathName, this.local.GetTime(), msg.Source, msg.MsgId);
                UpdateNameValueRequest request = new UpdateNameValueRequest(myServer, cb, myServer.GetNodeIdAddressBind, dirId1, r);
                myServer.RouteMessage(request, useCLB ? (NodeId) dirId1.clbId : (NodeId) dirId1);
            }
            else {
                // Finish creating the directory.
                new MkDirParentCallback(msg, this, fullPathName, this.local.GetTime(), msg.Source, msg.MsgId).Execute(myServer);
            }
        }

        // This gets invoked when a reply is received to mkdir's request to
        // update the parent directory entry for a new directory.
        class MkDirParentCallback: DirAppCallback {
            private string fullPathName;
            private NodeIdAddressBind clientSrc;
            private string clientMsgId;

            public MkDirParentCallback(DirAPI.MkDirApiRequest msg, DirApplication app, string fullPathName, long opnStartTime, NodeIdAddressBind clientSrc, string clientMsgId)
                : base(msg, app, opnStartTime)
            {
                this.fullPathName = fullPathName;
                this.clientSrc = clientSrc;
                this.clientMsgId = clientMsgId;
            }

            public override void Execute(Node localNode) {
                DirReplyMsg replyMsg = (DirReplyMsg) Reply;
                DirectoryId dirId = new DirectoryId(fullPathName);
                if (debugOutput) {
                    Console.Write("MkDirParentCallback {0}, dirId: {1} on node {2}",
                        fullPathName, dirId.ToString(), localNode.GetNodeIdAddressBind.GetAddress);
                    if (debugMsgIdOutput) Console.Write(": msgId: {0}", clientMsgId);
                    Console.WriteLine();
                }

                if ((replyMsg != null) && (replyMsg.result is DirException)) {
                    // Got an error while trying to update the parent directory.
                    // Abort the operation and reply to the original (non-overlay) client.

                    DirAPI.MkDirApiReply reply = new DirAPI.MkDirApiReply(requestMsg, localNode.GetNodeIdAddressBind, replyMsg.result);
                    localNode.SendReply(reply, requestMsg.Source);
                    return;
                }

                MkDirCallback cb = new MkDirCallback(this.requestMsg, app, this.clientSrc, this.clientMsgId, app.local.GetTime());
                MkDirRequest request = new MkDirRequest(app.myServer, cb, app.myServer.GetNodeIdAddressBind, dirId,
                    fullPathName);
                if (debugOutput) {
                    Console.Write("MkDirParentCallback {0}, dirId: {1} on node {2}",
                        fullPathName, dirId.ToString(), app.myServer.GetNodeIdAddressBind.GetAddress);
                    if (debugMsgIdOutput) Console.Write(": msgId: {0}", request.MsgId);
                    Console.WriteLine();
                }
                app.myServer.RouteMessage(request, useCLB ? (NodeId) dirId.clbId : (NodeId) dirId);
            }
        }

        // Server request for mkdir
        void ServerMkDir(MkDirRequest msg) {
            DirectoryId dirId = (DirectoryId) msg.Destination;
            string fullPathName = msg.fullPathName;
            if (debugOutput) Console.WriteLine("Received mkdir request on node {0}: {1}", myServer.GetNodeIdAddressBind.GetAddress, fullPathName);
            log.Write(LogLvl.Med, "MkDir '{0}' (addr:{1})", dirId.Name, msg.Destination);

            object res;
            if (!directories.Contains(dirId)) {
                DirectoryDescriptor dd = 
                    new DirectoryDescriptor(this, fullPathName, dirId);
                directories[dirId] = dd;
                res = null;
                dd.CreateBackups(); // Create backup replicas at leaf set nodes
            }
            else {
                string errMsg = "Directory already exists";
                res = new DirException(errMsg);
            }

            MkDirReply replyMsg = new MkDirReply(msg, myServer.GetNodeIdAddressBind, res);
            myServer.SendReply(replyMsg, msg.Source);
        }
    
        // This gets invoked when a reply is received to a mkdir request.
        class MkDirCallback: DirAppCallback {
            private NodeIdAddressBind clientSrc;
            private string clientMsgId;

            public MkDirCallback(OverlayRequestMsg requestMsg, DirApplication app, NodeIdAddressBind clientSrc, string clientMsgId, long opnStartTime)
                : base(requestMsg, app, opnStartTime)
            {
                this.clientSrc = clientSrc;
                this.clientMsgId = clientMsgId;
            }

            public override void Execute(Node localNode) {
                if (debugOutput) {
                    Console.Write("MkDirCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);
                    if (debugMsgIdOutput) Console.Write(": msgId: {0}", clientMsgId);
                    Console.WriteLine();
                }
                this.UpdateStats(DirOperations.mkDir);

                // Reply to the original (non-overlay) client.

                DirAPI.MkDirApiReply replyMsg = new DirAPI.MkDirApiReply(requestMsg, localNode.GetNodeIdAddressBind,
                    (Reply != null) ? ((MkDirReply) Reply).result : null);
                localNode.SendReply(replyMsg, requestMsg.Source);
            }
        }

//========================= rmdir =========================//

        [Serializable, XmlRootAttribute("RmDirRequest", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class RmDirRequest : DirRequestMsg
        {
            public string fullPathName;

            public RmDirRequest() {} // For XML Serialization only

            public RmDirRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, string fullPathName)
                : base(localNode, cb, source, dest)
            {
                this.fullPathName = fullPathName;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerRmDir(this);
            }
        }

        [Serializable, XmlRootAttribute("RmDirReply", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class RmDirReply : DirReplyMsg
        {
            public RmDirReply() {} // For XML Serialization only

            public RmDirReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        public void RmDir(DirAPI.RmDirApiRequest msg) {
            string fullPathName = msg.fullPathName;
            DirectoryId dirId = new DirectoryId(fullPathName);
            if (debugOutput) Console.WriteLine("rmDir on node {0}: {1} ({2})", myServer.GetNodeIdAddressBind.GetAddress, fullPathName, dirId);

            if (!IsRootPathName(fullPathName)) {
                // Remove from parent directory.
                int indx = fullPathName.LastIndexOf("/");
                string parentDir;
                if (indx == 0)
                    parentDir = "/";
                else 
                {
                    parentDir = fullPathName.Substring(0, indx);
                    if (fullPathName.LastIndexOf("!/") == indx - 1)
                        parentDir += "/"; // CLB root
                }
                string relDirName = fullPathName.Substring(indx + 1);

                DirectoryId dirId1 = new DirectoryId(parentDir);
                NameValueEntry r = new NameValueEntry(parentDir, relDirName);

                RmDirParentCallback cb = new RmDirParentCallback(msg, this, fullPathName, this.local.GetTime(), msg.Source, msg.MsgId);
                UpdateNameValueRequest request = new UpdateNameValueRequest(myServer, cb, myServer.GetNodeIdAddressBind, dirId1, r);
                myServer.RouteMessage(request, useCLB ? (NodeId) dirId1.clbId : (NodeId) dirId1);
            }
            else {
                // Finish deleting the directory.
                new RmDirParentCallback(msg, this, fullPathName, this.local.GetTime(), msg.Source, msg.MsgId).Execute(myServer);
            }
        }

        // This gets invoked when a reply is received to rmdir's request to
        // remove the parent directory entry for a new directory.
        class RmDirParentCallback : DirAppCallback {
            private string fullPathName;
            private NodeIdAddressBind clientSrc;
            private string clientMsgId;

            public RmDirParentCallback(DirAPI.RmDirApiRequest msg, DirApplication app, string fullPathName, long opnStartTime, NodeIdAddressBind clientSrc, string clientMsgId)
                : base(msg, app, opnStartTime)
            {
                this.fullPathName = fullPathName;
                this.clientSrc = clientSrc;
                this.clientMsgId = clientMsgId;
            }

            public override void Execute(Node localNode) {
                DirReplyMsg replyMsg = (DirReplyMsg) Reply;
                DirectoryId dirId = new DirectoryId(fullPathName);
                if (debugOutput) {
                    Console.Write("RmDirParentCallback {0}, dirId: {1} on node {2}",
                        fullPathName, dirId.ToString(), localNode.GetNodeIdAddressBind.GetAddress);
                    if (debugMsgIdOutput) Console.Write(": msgId: {0}", clientMsgId);
                    Console.WriteLine();
                }

                RmDirCallback cb = new RmDirCallback(this.requestMsg, app, this.clientSrc, this.clientMsgId, app.local.GetTime());
                RmDirRequest request = new RmDirRequest(app.myServer, cb, app.myServer.GetNodeIdAddressBind, dirId,
                    fullPathName);
                app.myServer.RouteMessage(request, useCLB ? (NodeId) dirId.clbId : (NodeId) dirId);
            }
        }

        // Server request for rmdir
        void ServerRmDir(RmDirRequest msg) {
            DirectoryId dirId = (DirectoryId) msg.Destination;
            string fullPathName = msg.fullPathName;
            if (debugOutput) Console.WriteLine("Received rmdir request on node {0}: {1}", myServer.GetNodeIdAddressBind.GetAddress, fullPathName);

            object res;
            if (directories.Contains(dirId)) {
                DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];
                // No need to call ShouldBePrimary() because we're deleting this directory anyway
                dir.SendRmReplica();
                directories.Remove(dirId);
                res = null;
            }
            else {
                string errMsg = "Directory does not exist";
                res = new DirException(errMsg);
            }

            RmDirReply replyMsg = new RmDirReply(msg, myServer.GetNodeIdAddressBind, res);
            myServer.SendReply(replyMsg, msg.Source);
        }
    
        // This gets invoked when a reply is received to a rmdir request.
        class RmDirCallback : DirAppCallback {
            private NodeIdAddressBind clientSrc;
            private string clientMsgId;

            public RmDirCallback(OverlayRequestMsg requestMsg, DirApplication app, NodeIdAddressBind clientSrc, string clientMsgId, long opnStartTime)
                : base(requestMsg, app, opnStartTime)
            {
                this.clientSrc = clientSrc;
                this.clientMsgId = clientMsgId;
            }

            public override void Execute(Node localNode) {
                if (debugOutput) {
                    Console.Write("RmDirCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);
                    if (debugMsgIdOutput) Console.Write(": msgId: {0}", clientMsgId);
                    Console.WriteLine();
                }
                this.UpdateStats(DirOperations.rmDir);

                // Reply to the original (non-overlay) client.

                DirAPI.RmDirApiReply replyMsg = new DirAPI.RmDirApiReply(requestMsg, localNode.GetNodeIdAddressBind,
                    (Reply != null) ? ((RmDirReply) Reply).result : null);
                localNode.SendReply(replyMsg, requestMsg.Source);
            }
        }

//========================= ListDir =========================//

        [Serializable, XmlRootAttribute("ListDirRequest", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class ListDirRequest : DirRequestMsg
        {
            public string fullPathName;

            public ListDirRequest() {} // For XML Serialization only

            public ListDirRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, string fullPathName)
                : base(localNode, cb, source, dest)
            {
                this.fullPathName = fullPathName;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerListDir(this);
            }
        }

        [Serializable, XmlRootAttribute("ListDirReply", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class ListDirReply : DirReplyMsg
        {
            public ListDirReply() {} // For XML Serialization only

            public ListDirReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        public void ListDir(DirAPI.ListDirApiRequest msg) {
            string fullPathName = msg.fullPathName;
            DirectoryId dirId = new DirectoryId(fullPathName);
            if (debugOutput) Console.WriteLine("listdir on node {0}: {1} ({2})", myServer.GetNodeIdAddressBind.GetAddress, fullPathName, dirId);

            ListDirCallback cb = new ListDirCallback(msg, this, msg.Source, msg.MsgId, local.GetTime());
            ListDirRequest request = new ListDirRequest(myServer, cb, myServer.GetNodeIdAddressBind, dirId, fullPathName);
            myServer.RouteMessage(request, useCLB ? (NodeId) dirId.clbId : (NodeId) dirId);

        }

        void ServerListDir(ListDirRequest msg) {
            DirectoryId dirId = (DirectoryId) msg.Destination;
            string fullPathName = (string) msg.fullPathName;
            if (debugOutput) Console.WriteLine("Received listdir request on node {0}: {1} ({2})", myServer.GetNodeIdAddressBind.GetAddress, fullPathName, dirId);

            object res;
            if (directories.Contains(dirId)) {
                DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];
                dir.UpdateCounts(msg, DirectoryDescriptor.DirectoryCountType.query);

                DirEntry[] entries = new DirEntry[dir.entries.Count];
                int i = 0;
                foreach (string name in dir.entries.Keys) {
                    ValueEntry e = (ValueEntry) dir.entries[name];
                    DirEntryType entryType;
                    entryType = e.EntryType;
                    entries[i++] = new DirEntry(name, entryType);
                }
                res = new DirEntryArray(entries);
            }
            else {
                string errMsg = "Directory does not exist";
                res = new DirException(errMsg);
            }

            ListDirReply replyMsg = new ListDirReply(msg, myServer.GetNodeIdAddressBind, res);
            myServer.SendReply(replyMsg, msg.Source);
        }
    
        // This gets invoked when a reply is received to a listDir request.
        class ListDirCallback : DirAppCallback {
            private NodeIdAddressBind clientSrc;
            private string clientMsgId;

            public ListDirCallback(OverlayRequestMsg requestMsg, DirApplication app, NodeIdAddressBind clientSrc, string clientMsgId, long opnStartTime)
                : base(requestMsg, app, opnStartTime) {
                this.clientSrc = clientSrc;
                this.clientMsgId = clientMsgId;
            }

            public override void Execute(Node localNode) {
                if (debugOutput) {
                    Console.Write("ListDirCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);
                    if (debugMsgIdOutput) Console.Write(": msgId: {0}", clientMsgId);
                    Console.WriteLine();
                }
                this.UpdateStats(DirOperations.listDir);

                // Reply to the original (non-overlay) client.

                DirAPI.ListDirApiReply replyMsg = new DirAPI.ListDirApiReply(requestMsg, localNode.GetNodeIdAddressBind,
                    (Reply != null) ? ((ListDirReply) Reply).result : null);
                localNode.SendReply(replyMsg, requestMsg.Source);
            }
        }

//========================= GetNameValue =========================//

        [Serializable, XmlRootAttribute("GetNameValueRequest", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class GetNameValueRequest : DirRequestMsg
        {
            public NameValueEntry nameValueEntry;

            public GetNameValueRequest() {} // For XML Serialization only

            public GetNameValueRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, NameValueEntry nameValueEntry)
                : base(localNode, cb, source, dest)
            {
                this.nameValueEntry = nameValueEntry;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerGetNameValue(this);
            }
        }

        [Serializable, XmlRootAttribute("GetNameValueReply", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class GetNameValueReply : DirReplyMsg
        {
            public GetNameValueReply() {} // For XML Serialization only

            public GetNameValueReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        public void GetNameValue(DirAPI.GetNameValueApiRequest msg) {
            NameValueEntry entry = msg.nameValueEntry;
            string dirPathName = entry.EntryDirectory;
            DirectoryId dirId = new DirectoryId(dirPathName);
            if (debugOutput) {
                Console.Write("getNameValue on node {0}: {1},{2}",
                    myServer.GetNodeIdAddressBind.GetAddress, dirPathName, entry.EntryName);
                if (debugTimingOutput) Console.Write(" at time {0}", local.GetTime());
                Console.WriteLine();
            }

            if (highLatencyReplicationEnabled) {
                Hashtable cdir = (Hashtable) cachedDirs[dirPathName];
                if (cdir != null) {
                    HighLatencyCachedDirEntryInfo cEntryInfo = (HighLatencyCachedDirEntryInfo) cdir[entry.EntryName];
                    if (cEntryInfo != null) {
                        if (local.GetTime() < cEntryInfo.expirationTime && cEntryInfo.usageCount++ < highLatencyReplicationMaxUsage)
                        {
                            if (debugOutput) Console.WriteLine("getNameValue hit in the local cache on {0} for {1}/{2}",
                                                 myServer.GetNodeIdAddressBind.GetAddress, dirPathName, entry.EntryName);
                            numOperations[(int)DirOperations.getNameValue]++;
                            if (operationMin[(int)DirOperations.getNameValue] > 0) operationMin[(int)DirOperations.getNameValue] = 0;
                            operationHisto[(int)DirOperations.getNameValue].Update((int)0);
                            numCachedEntryHits++;

                            // Reply to the original (non-overlay) client.

                            DirAPI.GetNameValueApiReply replyMsg = new DirAPI.GetNameValueApiReply(msg, myServer.GetNodeIdAddressBind, cEntryInfo.cEntry);
                            myServer.SendReply(replyMsg, msg.Source);
                            return;
                        }
                        else 
                        {
                            if (debugOutput) Console.WriteLine("getNameValue local cache expired on {0} for {1}/{2}",
                                                 myServer.GetNodeIdAddressBind.GetAddress, dirPathName, entry.EntryName);
                            cdir.Remove(entry.EntryName);
                        }
                    }
                }
            }

            GetNameValueCallback cb = new GetNameValueCallback(msg, this, msg.Source, msg.MsgId, local.GetTime());
            GetNameValueRequest request = new GetNameValueRequest(myServer, cb, myServer.GetNodeIdAddressBind, dirId, entry);
            myServer.RouteMessage(request, useCLB ? (NodeId) dirId.clbId : (NodeId) dirId);
        }

        void ServerGetNameValue(GetNameValueRequest msg) {
            DirectoryId dirId = (DirectoryId) msg.Destination;
            NameValueEntry entry = msg.nameValueEntry;
            if (debugOutput) Console.WriteLine("Received GetNameValue request on node {0}: {1} {2}", 
                                 myServer.GetNodeIdAddressBind.GetAddress, 
                                 entry.EntryDirectory, entry.EntryName);

            object res;
            if (directories.Contains(dirId)) {
                DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];
                dir.UpdateCounts(msg, DirectoryDescriptor.DirectoryCountType.query);

                ValueEntry e = (ValueEntry) dir.entries[entry.EntryName];
                if (e == null) {
                    string errMsg = entry.FullPathName() + " does not exist";
                    res = new DirException(errMsg);
                }
                else {
                    entry.EntryType = e.EntryType;
                    entry.EntryValue = e.EntryValue;
                    if (entry.EntryType == DirEntryType.dirEntry) {
                        string errMsg = entry.FullPathName() + " is a directory";
                        res = new DirException(errMsg);
                    }
                    else {
                        res = entry;
                    }
                }
            }
            else {
                string errMsg = dirId.Name + " does not exist.";
                res = new DirException(errMsg);
            }

            GetNameValueReply replyMsg = new GetNameValueReply(msg, myServer.GetNodeIdAddressBind, res);
            myServer.SendReply(replyMsg, msg.Source);
        }
            
        // This gets invoked when a reply is received to a getNameValue request.
        class GetNameValueCallback : DirAppCallback {
            private NodeIdAddressBind clientSrc;
            private string clientMsgId;

            public GetNameValueCallback(OverlayRequestMsg requestMsg, DirApplication app, NodeIdAddressBind clientSrc, string clientMsgId, long opnStartTime)
                : base(requestMsg, app, opnStartTime) {
                this.clientSrc = clientSrc;
                this.clientMsgId = clientMsgId;
            }

            public override void Execute(Node localNode) {
                if (Reply == null) 
                {
                    // Just return the null on to our caller
                }
                else if (((GetNameValueReply) Reply).result is DirException) {
                    DirException err = (DirException) (((GetNameValueReply) Reply).result);
                    if (debugOutput) Console.WriteLine("GetNameValueCallback: {0}", err.ErrMsg);
                }
                else {
                    NameValueEntry entry = (NameValueEntry) (((GetNameValueReply) Reply).result);
                    if (debugOutput) {
                        Console.Write("GetNameValueCallback on node {0}",
                            localNode.GetNodeIdAddressBind.GetAddress);
                        if (debugMsgIdOutput) Console.Write(": msgId: {0}", clientMsgId);
                        Console.WriteLine();
                        Console.WriteLine("  {0} {1} {2}", entry.EntryName, entry.EntryType, entry.EntryValue);
                    }
                    this.UpdateStats(DirOperations.getNameValue);

                    if (highLatencyReplicationEnabled) {
                        // Check to see if we should send a hint that replication
                        // to counteract high latencies is needed.
                        if (this.opnTime > highLatencyReplicationThreshold ||
                            (debugHighLatencyReplication && ++highLatencyReplicaCounter % 10 == 0) ) {
                            if (debugOutput) Console.WriteLine("GetNameValue return on {0}: {1} -- high latency of {2}", app.myServer.GetNodeIdAddressBind.GetAddress, entry.EntryDirectory, this.opnTime);
                            DirectoryId dirId = new DirectoryId(entry.EntryDirectory);
                            if (!app.directories.Contains(dirId)) {
                                Hashtable cdir = (Hashtable)app.cachedDirs[entry.EntryDirectory];
                                if (cdir == null) 
                                {
                                    cdir = new Hashtable();
                                    app.cachedDirs[entry.EntryDirectory] = cdir;
                                }

                                if (debugOutput) Console.WriteLine("Request creation of directory replica");
                                if (! DebugHighLatencyReplication || highLatencyReplicaCounter % 20 == 0) 
                                {
                                    // If debugging high latency replication skip this message sometimes so cache is actually used
                                    DirApplication.HighLatencyHintMsg msg =
                                        new DirApplication.HighLatencyHintMsg(app.myServer.GetNodeIdAddressBind, dirId, entry.EntryDirectory);
                                    app.myServer.SendMessageDirect(msg, ((GetNameValueReply) Reply).Source);
                                }


                                long expiration = (opnTime < highLatencyReplicationExpirationLimit) ? opnTime : highLatencyReplicationExpirationLimit;
                                if (expiration < highLatencyReplicationThreshold) expiration = highLatencyReplicationThreshold;
                                cdir[entry.EntryName] = new HighLatencyCachedDirEntryInfo(entry, app.local.GetTime() + expiration);
                            }
                        }
                    }
                }

                // Reply to the original (non-overlay) client.

                DirAPI.GetNameValueApiReply replyMsg = new DirAPI.GetNameValueApiReply(requestMsg, localNode.GetNodeIdAddressBind,
                    (Reply != null) ? ((GetNameValueReply) Reply).result : null);
                localNode.SendReply(replyMsg, requestMsg.Source);
            }
        }

//========================= UpdateNameValue =========================//

        [Serializable, XmlRootAttribute("UpdateNameValueRequest", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class UpdateNameValueRequest : DirRequestMsg
        {
            public NameValueEntry nameValueEntry;

            public UpdateNameValueRequest() {} // For XML Serialization only

            public UpdateNameValueRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, NameValueEntry nameValueEntry)
                : base(localNode, cb, source, dest)
            {
                this.nameValueEntry = nameValueEntry;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerUpdateNameValue(this);
            }
        }

        [Serializable, XmlRootAttribute("UpdateNameValueReply", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class UpdateNameValueReply : DirReplyMsg
        {
            public UpdateNameValueReply() {} // For XML Serialization only

            public UpdateNameValueReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        public void UpdateNameValue(DirAPI.UpdateNameValueApiRequest msg) {
            NameValueEntry entry = msg.nameValueEntry;
            string dirPathName = entry.EntryDirectory;
            DirectoryId dirId = new DirectoryId(dirPathName);

            UpdateNameValueCallback cb = new UpdateNameValueCallback(msg, this, msg.Source, msg.MsgId, local.GetTime());
            UpdateNameValueRequest request = new UpdateNameValueRequest(myServer, cb, myServer.GetNodeIdAddressBind, dirId, entry);
            if (debugOutput) {
                Console.Write("updateNameValue on node {0}: {1},{2},{3},{4} ({5})", myServer.GetNodeIdAddressBind.GetAddress,
                    dirPathName, entry.EntryName, entry.EntryType, entry.EntryValue, dirId);
                if (debugMsgIdOutput) Console.Write(" (msgId: {0})", request.MsgId);
                Console.WriteLine();
            }
            myServer.RouteMessage(request, useCLB ? (NodeId) dirId.clbId : (NodeId) dirId);
        }

        void ServerUpdateNameValue(UpdateNameValueRequest msg) {
            DirectoryId dirId = (DirectoryId) msg.Destination;
            NameValueEntry entry = msg.nameValueEntry;
            if (debugOutput) {
                Console.Write("Received UpdateNameValue request on node {0}: {1} {2} {3} {4}", 
                    myServer.GetNodeIdAddressBind.GetAddress, entry.EntryDirectory, entry.EntryName, entry.EntryType, entry.EntryValue);
                if (debugMsgIdOutput) Console.Write("  (msgId: {0})", msg.MsgId);
                Console.WriteLine();
            }

            object res;
            bool replyToCaller = true;
            if (directories.Contains(dirId)) {
                log.Write(LogLvl.Med, "Updating ('{0}', '{1}') to ({2} '{3}')", dirId.Name, entry.EntryName,
                    entry.EntryType, entry.EntryValue);
                DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];
                dir.ShouldBePrimary();  // Only primary replica should receive this message
                dir.UpdateCounts(msg, DirectoryDescriptor.DirectoryCountType.update);
                ValueEntry e = new ValueEntry(entry.EntryType, entry.EntryValue);
                if (e.EntryValue != null) dir.entries[entry.EntryName] = e;
                else dir.entries.Remove(entry.EntryName);
                res = null;

                replyToCaller = dir.UpdateOtherReplicas(entry, msg);

                // If 'replyToCaller' is false here, we will reply after updating the replicas
            }
            else {
                string errMsg = "Non-existent directory";
                res = new DirException(errMsg);
            }
            if (replyToCaller) {
                UpdateNameValueReply replyMsg = new UpdateNameValueReply(msg, myServer.GetNodeIdAddressBind, res);
                myServer.SendReply(replyMsg, msg.Source);
            }
        }

        // This gets invoked when a reply is received to a updateNameValue request.
        class UpdateNameValueCallback : DirAppCallback {
            private NodeIdAddressBind clientSrc;
            private string clientMsgId;

            public UpdateNameValueCallback(OverlayRequestMsg requestMsg, DirApplication app, NodeIdAddressBind clientSrc, string clientMsgId, long opnStartTime)
                : base(requestMsg, app, opnStartTime) {
                this.clientSrc = clientSrc;
                this.clientMsgId = clientMsgId;
            }

            public override void Execute(Node localNode) {
                if (debugOutput) {
                    Console.Write("UpdateNameValueCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);
                    if (debugMsgIdOutput) Console.Write(": msgId: {0}", clientMsgId);
                    Console.WriteLine();
                }
                this.UpdateStats(DirOperations.updateNameValue);

                // Reply to the original (non-overlay) client.

                DirAPI.UpdateNameValueApiReply replyMsg = new DirAPI.UpdateNameValueApiReply(requestMsg, localNode.GetNodeIdAddressBind,
                    (Reply != null) ? ((UpdateNameValueReply) Reply).result : null);
                localNode.SendReply(replyMsg, requestMsg.Source);
            }
        }

//========================= Server replica messages & methods  =========================//

//========================= CreateReplica messages & server  =========================//

        [Serializable, XmlRootAttribute("CreateReplicaRequest", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class CreateReplicaRequest : DirRequestMsg
        {
            public DirReplicaDescriptor drd;
            public bool fromPrimary;

            public CreateReplicaRequest() {} // For XML Serialization only

            public CreateReplicaRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest,
                DirReplicaDescriptor drd, bool fromPrimary)
                : base(localNode, cb, source, dest)
            {
                this.drd = drd;
                this.fromPrimary = fromPrimary;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerCreateReplica(this);
            }
        }

        [Serializable, XmlRootAttribute("CreateReplicaReply", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class CreateReplicaReply : DirReplyMsg
        {
            public CreateReplicaReply() {} // For XML Serialization only

            public CreateReplicaReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        void ServerCreateReplica(CreateReplicaRequest msg) {
            DirReplicaDescriptor entry = msg.drd;
            if (debugOutput) {
                Console.Write("Received CreateReplica request on node {0}: {1}", 
                    myServer.GetNodeIdAddressBind.GetAddress, entry.name);
                if (debugTimingOutput) Console.Write(" at time {0}", this.local.GetTime());
                Console.WriteLine();
            }
            log.Write(LogLvl.Med, "Creating replica of '{0}'", entry.name);

            if( ! directories.ContainsKey(entry.dirId) ) {
                DirectoryDescriptor dd = new DirectoryDescriptor(this, entry.name, entry.dirId,
                    entry.EntriesHashtable(), msg.Source, entry.lastSeqNumSeen, msg.fromPrimary);
                directories[entry.dirId] = dd;
            } else {
                // Already have a replica for this directory. Update it in-place.
                log.Write(LogLvl.Med, "Replica of '{0}' already exists. Updating it.", entry.name);
                DirectoryDescriptor oldDir = (DirectoryDescriptor) directories[entry.dirId];
                oldDir.Update(msg, entry);
            }

            // Check if our high-latency cache contains this directory.
            if (highLatencyReplicationEnabled && cachedDirs.Contains(entry.name)) {
                cachedDirs.Remove(entry.name);
            }

            CreateReplicaReply replyMsg = new CreateReplicaReply(msg, myServer.GetNodeIdAddressBind, null);
            myServer.SendReply(replyMsg, msg.Source);
        }

//========================= MovePrimary messages & server  =========================//

        [Serializable, XmlRootAttribute("MovePrimaryRequest", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class MovePrimaryRequest : DirRequestMsg
        {
            public DirReplicaDescriptor drd;

            public MovePrimaryRequest() {} // For XML Serialization only

            public MovePrimaryRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, DirReplicaDescriptor drd)
                : base(localNode, cb, source, dest)
            {
                this.drd = drd;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerMovePrimary(this);
            }
        }

        [Serializable, XmlRootAttribute("MovePrimaryReply", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class MovePrimaryReply : DirReplyMsg
        {
            public MovePrimaryReply() {} // For XML Serialization only

            public MovePrimaryReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        void ServerMovePrimary(MovePrimaryRequest msg)
        {
            DirReplicaDescriptor drd = msg.drd;
            
            DirectoryId dirId = (DirectoryId) msg.Destination;
            if (debugOutput) {
                Console.Write("Received MovePrimary request on node {0}: {1}", 
                    myServer.GetNodeIdAddressBind.GetAddress, drd.name);
                if (debugTimingOutput) Console.Write(" at time {0}", this.local.GetTime());
                Console.WriteLine();
            }

            object result = null;
            DirectoryDescriptor dd = null;

            if (directories.Contains(dirId))
            {
                dd = (DirectoryDescriptor)directories[dirId];
                if (dd.IsPrimaryReplica) 
                {
                    if (DirApplication.DebugOutput) Console.WriteLine("    Node already primary for directory {0}", drd.name);
                    result = myServer.GetNodeIdAddressBind;
                } 
                else 
                {
                    if (DirApplication.DebugOutput) Console.WriteLine("    Upgrading replica to primary for directory {0}", drd.name);
                    log.Write(LogLvl.Med, "Upgrading replica of '{0}' to primary", drd.name);
                    dd.BecomePrimary(drd.EntriesHashtable(), msg.Source, drd.lastSeqNumSeen);
                    dd.CreateBackups(); // Create backup replicas at leaf set nodes
                    result = myServer.GetNodeIdAddressBind;
                }
            } 
            else 
            {
                if (DirApplication.DebugOutput) Console.WriteLine("    Creating new primary for directory {0}", drd.name);
                log.Write(LogLvl.Med, "Creating new primary replica of '{0}'", drd.name);
                dd = new DirectoryDescriptor(this, drd.name, drd.dirId, drd.EntriesHashtable(), (NodeIdAddressBind)msg.Source, drd.lastSeqNumSeen, false);
                directories[drd.dirId] = dd;
                dd.BecomePrimary(drd.EntriesHashtable(), msg.Source, drd.lastSeqNumSeen);
                dd.CreateBackups(); // Create backup replicas at leaf set nodes
                result = myServer.GetNodeIdAddressBind;
            }

            // Check if our high-latency cache contains this directory.
            if (highLatencyReplicationEnabled && cachedDirs.Contains(drd.name)) 
            {
                cachedDirs.Remove(drd.name);
            }

            MovePrimaryReply replyMsg = new MovePrimaryReply(msg, myServer.GetNodeIdAddressBind, result);
            myServer.SendReply(replyMsg, msg.Source);
        }

//========================= UpdateReplica messages & server  =========================//

        [Serializable, XmlRootAttribute("UpdateReplicaRequest", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class UpdateReplicaRequest : DirRequestMsg
        {
            public object Payload;
            public bool fromPrimary;

            public UpdateReplicaRequest() {} // For XML Serialization only

            public UpdateReplicaRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest,
                object payload, bool fromPrimary)
                : base(localNode, cb, source, dest)
            {
                this.Payload = payload;
                this.fromPrimary = fromPrimary;
                TimeoutMs = 6 * 1000;   // Timeout reply in 6 seconds since don't want to delay main reply for background updates
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerUpdateReplica(this);
            }
        }

        [Serializable, XmlRootAttribute("UpdateReplicaReply", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class UpdateReplicaReply : DirReplyMsg
        {
            public UpdateReplicaReply() {} // For XML Serialization only

            public UpdateReplicaReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result)
                : base(requestMsg, source, result)
            {
            }
        }

        void ServerUpdateReplica(UpdateReplicaRequest msg) 
        {
            object res = null;
            bool replyToCaller = true;
            if (msg.Payload is DirUpdateInfo) {
                DirUpdateInfo dui = (DirUpdateInfo) msg.Payload;
                NameValueEntry entry = (NameValueEntry) dui.entry;
                if (debugOutput) {
                    Console.Write("Received UpdateReplica request on node {0}: {1}, {2}, {3}, {4}", 
                        myServer.GetNodeIdAddressBind.GetAddress, entry.EntryDirectory, entry.EntryName, entry.EntryType, entry.EntryValue);
                    if (debugTimingOutput) Console.Write(" at time {0}", this.local.GetTime());
                    Console.WriteLine();
                }
                log.Write(LogLvl.Med, "Update ('{0}', '{1}') to ({2} '{3}') from {4}",
                    entry.EntryDirectory, entry.EntryName, entry.EntryType, entry.EntryValue, msg.Source);

                DirectoryId dirId = new DirectoryId(entry.EntryDirectory);
                if (directories.Contains(dirId)) {
                    DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];
                    dir.ParentMsgSeen = true;
                    if (dir.LastSeqNumSeen == (dui.seqNum - 1)) {
                        ValueEntry e = new ValueEntry(entry.EntryType, entry.EntryValue);
                        dir.entries[entry.EntryName] = e;

                        dir.LastSeqNumSeen = dui.seqNum;
                        replyToCaller = dir.UpdateOtherReplicas(dui, msg);
                    }
                    else {
                        // Reply with a request for all updates starting 
                        // from dir.LastSeqNumSeen + 1.
                        res = dir.LastSeqNumSeen + 1;
                    }
                    dir.AdjacentToPrimary = msg.fromPrimary;
                }
                else {
                    if (DirApplication.DebugOutput) Console.WriteLine("ServerUpdateReplica: Directory does not exist: {0}", entry.EntryDirectory);
                    log.Write(LogLvl.Med, "Cannot update -- directory does not exist!");
                }
            }
            else {
                // Received a replica update in the form of the entire directory.
                DirReplicaDescriptor drd = (DirReplicaDescriptor) msg.Payload;
                DirectoryId dirId = new DirectoryId(drd.name);
                if (debugOutput) Console.WriteLine("Received FULL UpdateReplica request on node {0}: {1}", 
                                     myServer.GetNodeIdAddressBind.GetAddress, drd.name);
                log.Write(LogLvl.Med, "Full Update '{0}' from {1}", drd.name, msg.Source.GetAddress.ToString());
                if (directories.Contains(dirId)) {
                    DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];

                    // Replace existing directory info with new info.
                    dir.FullDirectoryUpdate(drd.EntriesHashtable(), (NodeIdAddressBind)msg.Source, drd.lastSeqNumSeen);
                    dir.ParentMsgSeen = true;
                    dir.AdjacentToPrimary = msg.fromPrimary;

                    replyToCaller = dir.UpdateOtherReplicas(drd, msg);
                }
                else {
                    if (DirApplication.DebugOutput) Console.WriteLine("ServerUpdateReplica: Directory does not exist: {0}", drd.name);
                    log.Write(LogLvl.Med, "Cannot update -- directory does not exist!");
                }
            }

            if (replyToCaller) {
                UpdateReplicaReply replyMsg = new UpdateReplicaReply(msg, myServer.GetNodeIdAddressBind, res);
                myServer.SendReply(replyMsg, msg.Source);
            }
        }

//========================= ResyncReplica messages & server  =========================//

        [Serializable, XmlRootAttribute("ResyncReplicaRequest", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class ResyncReplicaRequest : DirRequestMsg
        {
            public ResyncDescriptor rd;

            public ResyncReplicaRequest() {} // For XML Serialization only

            public ResyncReplicaRequest(Node localNode, OverlayReplyCallback cb, NodeIdAddressBind source, NodeId dest, ResyncDescriptor rd)
                : base(localNode, cb, source, dest)
            {
                this.rd = rd;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerResyncReplica(this);
            }
        }

        [Serializable, XmlRootAttribute("ResyncReplicaReply", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class ResyncReplicaReply : DirReplyMsg
        {
            public bool fromPrimary;

            public ResyncReplicaReply() {} // For XML Serialization only

            public ResyncReplicaReply(MessageRouting.OverlayRequestMsg requestMsg, NodeIdAddressBind source, object result, bool fromPrimary)
                : base(requestMsg, source, result)
            {
                this.fromPrimary = fromPrimary;
            }
        }

        void ServerResyncReplica(ResyncReplicaRequest msg) 
        {
            DirectoryId dirId = (DirectoryId) msg.Destination;
            ResyncDescriptor rd = msg.rd;
            if (debugOutput) Console.WriteLine("Received ResyncReplica request on node {0}: {1} {2}", 
                                 myServer.GetNodeIdAddressBind.GetAddress, rd.dirName, rd.lastSeqNumSeen);

            object res = null;
            bool fromPrimary = false;
            if (directories.Contains(dirId)) {
                DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];

                // For now, we simply send the entire directory since we aren't yet keeping a log of recent updates.
                DirReplicaDescriptor drd = dir.CreateDirReplicaDescriptor();
                res = drd;
                fromPrimary = dir.IsPrimaryReplica;

                if( ! msg.Source.Equals(myServer.GetNodeIdAddressBind) ) {
                    // Add orphaned replica to our replicaNodes list (unless it's already there).
                    dir.AddReplica((NodeIdAddressBind)msg.Source, false);
                }
            }
            else {
                if (DirApplication.DebugOutput) Console.WriteLine("ServerResyncReplica: Directory does not exist: {0}", rd.dirName);
            }

            ResyncReplicaReply replyMsg = new ResyncReplicaReply(msg, myServer.GetNodeIdAddressBind, res, fromPrimary);
            myServer.SendReply(replyMsg, msg.Source);
        }

//========================= HighLatencyHint message & server  =========================//

        [Serializable, XmlRootAttribute("HighLatencyHintMsg", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class HighLatencyHintMsg : DirAsyncMsg
        {
            public string dirName;

            public HighLatencyHintMsg() {} // For XML Serialization only

            public HighLatencyHintMsg(NodeIdAddressBind source, NodeId dest, string dirName)
                : base(source, dest)
            {
                this.dirName = dirName;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerHighLatencyHint(this);
            }
        }

        void ServerHighLatencyHint(HighLatencyHintMsg msg) 
        {
            DirectoryId dirId = (DirectoryId) msg.Destination;
            string dirName = msg.dirName;
            if (debugOutput) Console.WriteLine("Received HighLatencyHint on node {0}: {1}", 
                                 myServer.GetNodeIdAddressBind.GetAddress, dirName);

            if (directories.Contains(dirId)) {
                DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];
                dir.CreateReplica((NodeIdAddressBind)msg.Source, false);
            }
            else {
                if (DirApplication.DebugOutput) Console.WriteLine("ServerHighLatencyHint: Directory does not exist: {0}", dirName);
            }
        }

//========================= ReplicaHeartBeat message & server  =========================//

        [Serializable, XmlRootAttribute("ReplicaHeartBeatMsg", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class ReplicaHeartBeatMsg : DirAsyncMsg
        {
            public string dirName;
            public bool fromPrimary;

            public ReplicaHeartBeatMsg() {} // For XML Serialization only

            public ReplicaHeartBeatMsg(NodeIdAddressBind source, NodeId dest, string dirName, bool fromPrimary)
                : base(source, dest)
            {
                this.dirName = dirName;
                this.fromPrimary = fromPrimary;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerReplicaHeartBeat(this);
            }
        }

        void ServerReplicaHeartBeat(ReplicaHeartBeatMsg msg) {
            string dirName = msg.dirName;
            DirectoryId dirId = new DirectoryId(dirName);
            if (debugOutput) Console.WriteLine("Received ReplicaHeartBeat on node {0}: {1}", 
                                 myServer.GetNodeIdAddressBind.GetAddress, dirName);

            if (directories.Contains(dirId)) {
                DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];
                if( msg.Source.Equals(dir.Parent) ) {
                    dir.Orphaned = false;
                    dir.ParentMsgSeen = true;
                    dir.AdjacentToPrimary = msg.fromPrimary;
                } else {
                    // If we are receiving heartbeats from a node who is not our
                    // parent, ignore them.
                }
            }
            else {
                if (DirApplication.DebugOutput) Console.WriteLine("ServerReplicaHeartBeat: Directory does not exist: {0}", dirName);
            }
        }

//========================= VerifyPrimary message & server  =========================//

        [Serializable, XmlRootAttribute("VerifyPrimaryMsg", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class VerifyPrimaryMsg : DirAsyncMsg
        {
            public DirReplicaDescriptor drd;

            public VerifyPrimaryMsg() {} // For XML Serialization only

            public VerifyPrimaryMsg(NodeIdAddressBind source, NodeId dest, DirReplicaDescriptor drd)
                : base(source, dest)
            {
                this.drd = drd;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerVerifyPrimary(this);
            }
        }

        void ServerVerifyPrimary(VerifyPrimaryMsg msg) 
        {
            DirReplicaDescriptor drd = msg.drd;
            string dirName = drd.name;
            DirectoryId dirId = new DirectoryId(dirName);
            if (debugOutput) Console.WriteLine("Received VerifyPrimary on node {0}: {1}", 
                                 myServer.GetNodeIdAddressBind.GetAddress, dirName);

            DirectoryDescriptor dd = null;

            if (directories.Contains(dirId))
            {
                dd = (DirectoryDescriptor)directories[dirId];
                if (dd.IsPrimaryReplica) 
                {
                    if (DirApplication.DebugOutput) Console.WriteLine("    Node already primary for directory {0}", drd.name);
                } 
                else 
                {
                    if (DirApplication.DebugOutput) Console.WriteLine("    Upgrading replica to primary for directory {0}", drd.name);
                    dd.BecomePrimary(drd.EntriesHashtable(), msg.Source, drd.lastSeqNumSeen);
                    dd.CreateBackups(); // Create backup replicas at leaf set nodes
                }
            } 
            else 
            {
                if (DirApplication.DebugOutput) Console.WriteLine("    Creating new primary for directory {0}", drd.name);
                dd = new DirectoryDescriptor(this, drd.name, drd.dirId, drd.EntriesHashtable(), (NodeIdAddressBind)msg.Source, drd.lastSeqNumSeen, false);
                directories[drd.dirId] = dd;
                dd.BecomePrimary(drd.EntriesHashtable(), msg.Source, drd.lastSeqNumSeen);
                dd.CreateBackups(); // Create backup replicas at leaf set nodes
            }

            // Check if our high-latency cache contains this directory.
            if (highLatencyReplicationEnabled && cachedDirs.Contains(drd.name)) 
            {
                cachedDirs.Remove(drd.name);
            }
        }

//========================= RmReplica message & server  =========================//

        [Serializable, XmlRootAttribute("RmReplicaMsg", Namespace=SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        public class RmReplicaMsg : DirAsyncMsg
        {
            public string dirName;

            public RmReplicaMsg() {} // For XML Serialization only

            public RmReplicaMsg(NodeIdAddressBind source, NodeId dest, string dirName)
                : base(source, dest)
            {
                this.dirName = dirName;
            }

            public override void ArrivedAtDestination(MessageRouting.Node localNode)
            {
                DirApplication dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
                dirApp.ServerRmReplica(this);
            }
        }

        void ServerRmReplica(RmReplicaMsg msg) 
        {
            string dirName = msg.dirName;
            DirectoryId dirId = new DirectoryId(dirName);
            if (debugOutput) Console.WriteLine("Received RmReplica on node {0}: {1}", 
                                 myServer.GetNodeIdAddressBind.GetAddress, dirName);

            if (directories.Contains(dirId)) 
            {
                DirectoryDescriptor dir = (DirectoryDescriptor)directories[dirId];
                if (dir == null) return;

                // Only remove our replica if the node that told us to do so is our parent
                // (this might not always be true due to race conditions)
                if (msg.Source.Equals(dir.Parent)) {
                    log.Write(LogLvl.Med, "Removing replica of '{0}'", dirName);
                    dir.SendRmReplica();
                    directories.Remove(dirId);
                }
            }
        }

//========================= DirApplication housekeeping methods  =========================//

        public override void NewNeighbour(MessageRouting.NodeIdAddressBind id) 
        {
            // if (DebugOutput) Console.WriteLine("NewNeighbour called in node {0} adding node {1}", myServer.GetNodeIdAddressBind.GetAddress, id.GetAddress);
            foreach (object dirID in directories.Keys) 
            {
                DirectoryDescriptor dd = (DirectoryDescriptor) directories[dirID];
                if (dd.IsPrimaryReplica) 
                {
                    dd.MovePrimary(); // Send a message trying to move the primary copy in case it needs to move
                }
            }
        }

        public override void NeighbourRemoved(MessageRouting.NodeIdAddressBind id, bool failure) 
        {
            if (DebugOutput) Console.WriteLine("DirApplication.NeighbourRemoved called in node {0} for node {1}, failure={2}",
                                     myServer.GetNodeIdAddressBind.GetAddress, id.GetAddress, failure);
            log.Write(LogLvl.Hi, "Neighbour removed {0} {1}", id.GetAddress.ToString(), failure ? "(failed)" : "");
            foreach (object dirID in directories.Keys) 
            {
                DirectoryDescriptor dd = (DirectoryDescriptor) directories[dirID];
                dd.NeighbourRemoved(id, failure);
            }
        }

        static public bool IsRootPathName(string s) 
        {
            if (s.Equals("/")) return true;
            int index = s.LastIndexOf("!/");
            if (index < 0) return false;
            if (index == s.Length - 2) return true;
            return false;
        }
    }

//========================= ValueEntry  =========================//

    [Serializable,
    XmlRootAttribute("ValueEntry",Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)]
    public class ValueEntry {
        private DirEntryType type;
        private string val;

        public ValueEntry() {
        }

        public ValueEntry(DirEntryType type, string val) {
            this.type = type;
            this.val = val;
        }

        [XmlElementAttribute("ValueEntryType",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public DirEntryType EntryType {
            get {return this.type;}
            set {this.type = value;}
        }

        [XmlElementAttribute("ValueEntryValue",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public string EntryValue {
            get {return this.val;}
            set {this.val = value;}
        }
    }


//========================= DirectoryDescriptor  =========================//

    public class DirectoryDescriptor {
        private string name;
        private DirectoryId dirId;
        private DirApplication app;

        // Duration of maintenance intervals (in msecs).
        public const long maintIntervalLength = 4 * 1000; // 4 seconds

        // Duration of maintenance interval for sending heartbeats to replicas.
        private const long heartBeatMaintIntervalLength = 4 * 1000; // 4 seconds

        // Keep track of the time that the current replica heartbeat timeout interval started.
        public const long heartBeatTimeoutInterval = 20 * 1000; // 20 seconds

        // Max. number of operations per second allowed before we replicate.
        private const int maxOpnsPerSecBeforeReplicate = 10; // Was 50 but never happened -- max 26/s with merc, 2/s with GT topology
        private const int maxOpnsBeforeReplicate = maxOpnsPerSecBeforeReplicate * ((int)maintIntervalLength) / 1000;

        private int numQueries;
        private int numUpdates;

        // Are we the primary replica?
        private bool primaryReplica;

        // Are we a possible failure backup for the primary replica?
        public bool AdjacentToPrimary;

        // Incoming traffic statistics for last-hop overlay forwarding nodes
        private Hashtable potentialReplicaNodes;

        // List of replicas we have created.
        private Hashtable replicaNodes;
        // Parent node who created us.  Null for primary replica.
        private NodeIdAddressBind myParent;

        // Next sequence number to use when updating replicas.
        private int seqNum;
        // Last update sequence number seen from parent.
        private int lastSeqNumSeen;

        // Flag to signal that we've heard from our parent replica during the past 
        // maint. interval
        private bool parentMsgSeen = true;
        // Flag to signal that we've declared ourself to be an orphaned replica.
        // Set to false again once we hear from an ancestor replica in response to
        // a resyncReplica request.
        private bool orphaned = false;

        public Hashtable entries;

        private void CheckConsistency()
        {
            Debug.Assert(! replicaNodes.Contains(app.myServer.GetNodeIdAddressBind));
            if (replicaNodes.Contains(app.myServer.GetNodeIdAddressBind)) // Repair data structure if resumed from assertion
            {
                Console.WriteLine("XXX CheckConsistency: node {0} in own replica list for {1} -- removing!!! XXX",
                    app.myServer.GetPhysicalAddr, DirectoryName);
                replicaNodes.Remove(app.myServer.GetNodeIdAddressBind); // Fix data structure
            }

        }

        private void Init(DirApplication app, string name, DirectoryId dirId) {
            this.name = name;
            this.dirId = dirId;
            this.app = app;
            this.entries = new Hashtable();
            this.potentialReplicaNodes = new Hashtable();
            this.replicaNodes = new Hashtable();
            this.parentMsgSeen = true;
        }

        // Create a new directory.
        public DirectoryDescriptor(DirApplication app, string name, DirectoryId dirId) {
            Init(app, name, dirId);
            this.entries = new Hashtable();
            this.myParent = null;
            this.seqNum = 1;  // Start out one ahead of the last seq. num. seen.
            this.lastSeqNumSeen = 0;
            this.primaryReplica = true;
            this.AdjacentToPrimary = false;
        }

        // Create a directory replica.
        public DirectoryDescriptor(DirApplication app, string name, DirectoryId dirId,
            Hashtable entries, NodeIdAddressBind myParent, int lastSeqNumSeen, bool adjacentToPrimary) {
            Init(app, name, dirId);
            this.entries = entries;
            this.myParent = myParent;
            this.seqNum = 0;  // Should never be used.
            this.lastSeqNumSeen = lastSeqNumSeen;
            this.primaryReplica = false;
            this.AdjacentToPrimary = adjacentToPrimary;
        }

        public void BecomePrimary(Hashtable entries, NodeIdAddressBind oldPrimary, int lastSeqNumSeen) 
        {
            CheckConsistency();
            this.entries = entries;
            this.myParent = null;
            this.seqNum = lastSeqNumSeen;
            this.lastSeqNumSeen = lastSeqNumSeen;
            this.primaryReplica = true;
            this.AdjacentToPrimary = false;
            if (! app.myServer.GetNodeIdAddressBind.Equals(oldPrimary))
                this.replicaNodes[oldPrimary] = new DirReplicaInfo(oldPrimary, false); // Old primary becomes simply a replica
            CheckConsistency();
        }

        public void BecomePrimary() 
        {
            CheckConsistency();
            this.myParent = null;
            this.seqNum = lastSeqNumSeen;
            this.primaryReplica = true;
            this.AdjacentToPrimary = false;
        }

        public void BecomeReplica(NodeIdAddressBind myParent)
        {
            CheckConsistency();
            this.myParent = myParent;

            // If my new parent is already one of my children,
            // remove it from the list of children
            if( this.replicaNodes.ContainsKey(myParent) ) {
                this.replicaNodes.Remove(myParent);
            }

            // We're no longer the primary. We should remove our backup replica(s).
            ArrayList replicas = new ArrayList( replicaNodes.Values );
            foreach (DirReplicaInfo dri in replicas) {
                if( dri.isBackup ) {
                    RemoveBackupReplica( dri.replicaNode );
                }
            }

            this.lastSeqNumSeen = this.seqNum;
            this.primaryReplica = false;
            this.AdjacentToPrimary = true;
        }

        private void RemoveBackupReplica( NodeIdAddressBind backupReplica ) {
            Log.Write(DirApplication.CATEGORY, LogLvl.Med, "Removing backup replica of '{0}' on node {1}",
                this.name, backupReplica.GetAddress.ToString());
            replicaNodes.Remove( backupReplica );
            DirApplication.RmReplicaMsg msg =
                new DirApplication.RmReplicaMsg(app.myServer.GetNodeIdAddressBind, backupReplica.GetNodeId, this.name);
            app.myServer.SendMessageDirect(msg, backupReplica);
        }

        public bool IsPrimaryReplica 
        {
            get {return this.primaryReplica;}
        }

        public NodeIdAddressBind Parent {
            get { return myParent; }
            set { myParent = value; }
        }

        // Called when node receives a message that should only go to the primary directory copy.
        // If we're not the primary it means the primary died, and we should become the new primary.
        public void ShouldBePrimary()
        {
            CheckConsistency();
            if (IsPrimaryReplica) return;

            if (DirApplication.DebugOutput) Console.WriteLine("    Upgrading node {0} to primary copy of dir {1}",
                                                app.myServer.GetPhysicalAddr, DirectoryName);
            BecomePrimary();
            CreateBackups();
            CheckConsistency();
        }

        // Update a directory with full content from a parent.
        public void FullDirectoryUpdate(Hashtable entries, NodeIdAddressBind myParent, int lastSeqNumSeen) {
            this.entries = entries;
            this.myParent = myParent;
            this.lastSeqNumSeen = lastSeqNumSeen;
        }

        public enum DirectoryCountType {query, update};

        private bool SilentParentReplica() {
            if (!this.primaryReplica && (!this.parentMsgSeen ||
                (DirApplication.DebugResyncReplica && ++DirApplication.resyncReplicaCounter % 2 == 0)))
                return true;
            else 
                return false;
        }

        public void HeartbeatTimerFired() {
            CheckConsistency();

            // Deal with replica heartbeat.
            SendReplicaHeartBeat();

            // Check if we have't heard from our parent and should issue
            // a ReplicaResync request.
            if (SilentParentReplica()) {
                this.orphaned = true;

                if (AdjacentToPrimary) {
                    // Primary may have disappeared.  Initiate process of designating new primary.
                    VerifyPrimary();
                }

                ResyncReplica();
            }

            // Reset the parent-seen flag so that we can determine
            // whether our parent contacts us during the upcoming
            // maintenance interval.
            this.ParentMsgSeen = false;
        }

        public void MaintenanceTimerFired() {
            CheckConsistency();
            this.ResetCounts();
        }

        public void Update(DirApplication.CreateReplicaRequest msg, DirReplicaDescriptor entry) {
            // Update the entries if they have a better sequence number
            if( entry.lastSeqNumSeen>this.lastSeqNumSeen ) {
                this.entries = entry.EntriesHashtable();
                this.lastSeqNumSeen = entry.lastSeqNumSeen;
            }

            // Set our parent to be the source of this message, unless that node
            // is already a child (which can happen due to race conditions)
            if( ! replicaNodes.ContainsKey(msg.Source) ) {
                this.myParent = msg.Source;
                this.AdjacentToPrimary = msg.fromPrimary;
            }
        }

        public void UpdateCounts(DirRequestMsg msg, DirectoryCountType whichCount) {
            CheckConsistency();

            // Update the counters and check for overload.

            // Update the counts.
            if (whichCount == DirectoryCountType.query) {
                this.numQueries++;

                NodeIdAddressBind lastNode = msg.LastNodeId();
                if (lastNode == null)
                    return; // If we're servicing local requests skip replica creation logic
                if (this.potentialReplicaNodes.Contains(lastNode)) {
                    int count = (int)this.potentialReplicaNodes[lastNode];
                    this.potentialReplicaNodes[lastNode] = count + 1;
                }
                else {
                    this.potentialReplicaNodes[lastNode] = 1;
                }
            }
            else {
                this.numUpdates++;
            }

            // Check to see if we've reached an overload state.
            if (this.OverLoaded())
                this.CreateReplica();
        }

        public void ResetCounts() {
            this.potentialReplicaNodes.Clear();
            this.numQueries = 0;
            this.numUpdates = 0;
        }

        public bool OverLoaded() {
            if (!DirApplication.ReplicationEnabled) return false;
            return ((this.numQueries > 0) &&
                ((this.numQueries + this.numUpdates) >= maxOpnsBeforeReplicate));
        }

        public DirReplicaDescriptor CreateDirReplicaDescriptor() {
            return new DirReplicaDescriptor(this.name, this.dirId, this.entries, this.lastSeqNumSeen);
        }

//========================= Client replica methods  =========================//

//========================= CreateReplica client  =========================//

        public void CreateReplica() {
            CheckConsistency();
            if (DirApplication.DebugOutput) Console.WriteLine("CreateReplica on node {0}: {1}", this.app.myServer.GetNodeIdAddressBind.GetAddress, this.name);
            // Find out which route the most number of requests have come from.
            int maxCount = 0;
            NodeIdAddressBind maxNode = null;
            
            // Find node through which the most requests came
            foreach (NodeIdAddressBind n in this.potentialReplicaNodes.Keys) 
            {
                // Don't consider nodes at which we've already created a replica.
                if (replicaNodes.Contains(n)) continue;

                int count = (int) this.potentialReplicaNodes[n];
                if (count > maxCount) {
                    maxCount = count;
                    maxNode = n;
                }
            }
            if (maxNode == null) return;  // There are no potential candidates for replication.

            ResetCounts();  // So noone else mistakenly also creates a replica.
            CreateReplica(maxNode, false);
            CheckConsistency();
        }

        public void CreateReplica(NodeIdAddressBind replicaNode, bool isBackup) {
            if (DirApplication.DebugOutput) {
                Console.Write("CreateReplica on {0}: send request to {1}",
                    this.app.myServer.GetNodeIdAddressBind.GetAddress, replicaNode.GetAddress);
                if (DirApplication.DebugTimingOutput) Console.Write(" at time {0}", this.app.local.GetTime());
                Console.WriteLine();
            }

            DirReplicaDescriptor drd = CreateDirReplicaDescriptor();

            CreateReplicaCallback cb = new CreateReplicaCallback(app);
            DirApplication.CreateReplicaRequest request =
                new DirApplication.CreateReplicaRequest(app.myServer, cb, app.myServer.GetNodeIdAddressBind, replicaNode.GetNodeId, drd, IsPrimaryReplica);
            app.myServer.SendMessageDirect(request, replicaNode);

            bool oldIsBackup = false;
            if (replicaNodes.Contains(replicaNode)) // Preserve isBackup state across CreateReplica calls
                oldIsBackup = ((DirReplicaInfo)(replicaNodes[replicaNode])).isBackup;

            // Record that replicaNode is now an actual replica instead of a potential replica.
            replicaNodes[replicaNode] = new DirReplicaInfo(replicaNode, isBackup | oldIsBackup);
            CheckConsistency();
        }

        class CreateReplicaCallback: DirApplication.DirAppCallback {
            public CreateReplicaCallback(DirApplication dirApp)
                : base(dirApp)
            {
            }

            public override void Execute(Node localNode) {
                if (DirApplication.DebugOutput) Console.WriteLine("CreateReplicaCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);
                this.UpdateStats(DirApplication.DirOperations.createReplica);
            }
        }

//========================= CreateBackups method  =========================//

        //
        // Create replicas at leaf set nodes adjoining the primary copy of a directory for fail-over purposes.
        //
        public void CreateBackups() {
            CheckConsistency();

            // If replication is not enabled, do nothing
            if( (! DirApplication.ReplicationEnabled) || DirApplication.DefaultReplicationNumber<=0 ) 
                return;

            SkipNet.Node snNode = (SkipNet.Node) app.myServer;
            MessageRouting.NodeId destId = DirApplication.useCLB ? (NodeId) dirId.clbId : (NodeId) dirId;

            // Determine the set of neighbor nodes that will take over should the local node fail.
            // This list is sorted by closest-failover-neighbour-first.
            ArrayList list = snNode.GetFailoverNeighbours( destId );

            // If the CLB Table is not yet up-to-date, the list may be null, in which case we bail
            if (list == null) {
                if (DirApplication.DebugOutput)
                    Console.WriteLine("Null FailoverNeighbour list on node {0}", app.myServer.GetPhysicalAddr);
                return;
            }
            Debug.Assert( ! list.Contains( app.myServer.GetNodeIdAddressBind ) );

            // Get the array of current backup replicas
            ArrayList currentBackups = new ArrayList();
            foreach( DirReplicaInfo dri in replicaNodes.Values ) {
                if( dri.isBackup ) {
                    currentBackups.Add( dri.replicaNode );
                }
            }

            // Determine the array of new backup replicas
            ArrayList newBackups = list;
            int extraNodes = newBackups.Count - DirApplication.DefaultReplicationNumber;
            if( extraNodes>0 ) {
                newBackups.RemoveRange( DirApplication.DefaultReplicationNumber, extraNodes );
            }

            // If we couldn't determine the desired number of backup replicas, just complain
            if( extraNodes<0 ) {
                if (DirApplication.DebugOutput)
                    Console.WriteLine("Insufficient FailoverNeighbour list on node {0}: needed {1} but only got {2}",
                        app.myServer.GetPhysicalAddr, DirApplication.DefaultReplicationNumber, newBackups.Count);
            }

            // Determine the set of nodes which need a new backup replica
            foreach( NodeIdAddressBind n in newBackups ) {
                if( (! currentBackups.Contains(n)) || !((DirReplicaInfo) replicaNodes[n]).isBackup ) {
                    // Create a backup replica on node n
                    if (DirApplication.DebugOutput)
                        Console.WriteLine("FailoverNeighbour on node {0}: {1}",
                            app.myServer.GetPhysicalAddr, n.GetAddress);
                    Log.Write(DirApplication.CATEGORY, LogLvl.Med, "Sending failover replica of '{0}' to {1}",
                        DirectoryName, n.GetAddress.ToString());

                    CreateReplica(n, true);
                }
            }

            // As long as we determined enough backup replicas,
            // remove current backup replicas which we no longer need
            if( extraNodes>=0 ) {
                foreach( NodeIdAddressBind n in currentBackups ) {
                    if( ! newBackups.Contains(n) ) {
                        RemoveBackupReplica(n);
                    }
                }
            }
        }

//========================= MovePrimary client  =========================//

        //
        // Route information necessary to move a directory's primary copy to another node.
        // If we receive this notification ourselves there's nothing to do because the primary didn't move.
        // Otherwise we convert the old primary to into a replica, since the primary really moved.
        //
        public void MovePrimary() 
        {
            CheckConsistency();
            if (DirApplication.DebugOutput)
                {
                Console.Write("MovePrimary on {0} for dir {1}",
                    app.myServer.GetNodeIdAddressBind.GetAddress, DirectoryName);
                if (DirApplication.DebugTimingOutput) Console.Write(" at time {0}", app.local.GetTime());
                Console.WriteLine();
            }

            // Route to ID of directory
            DirReplicaDescriptor drd = CreateDirReplicaDescriptor();

            MovePrimaryCallback cb = new MovePrimaryCallback(app, this);
            DirApplication.MovePrimaryRequest request =
                new DirApplication.MovePrimaryRequest(app.myServer, cb, app.myServer.GetNodeIdAddressBind, this.dirId, drd);
            app.myServer.RouteMessage(request, DirApplication.useCLB ? (NodeId) dirId.clbId : (NodeId) dirId);
        }

        class MovePrimaryCallback: DirApplication.DirAppCallback
        {
            private DirectoryDescriptor dd;
            public MovePrimaryCallback(DirApplication dirApp, DirectoryDescriptor dd)
                : base(dirApp)
            {
                this.dd = dd;
            }

            public override void Execute(Node localNode) {
                if (DirApplication.DebugOutput) Console.WriteLine("MovePrimaryCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);
                this.UpdateStats(DirApplication.DirOperations.movePrimary);
                if (! dd.IsPrimaryReplica) return;  // No work to do if not still the primary copy

                if (Reply != null && ((DirApplication.MovePrimaryReply) Reply).result != null)
                {
                    NodeIdAddressBind primaryLoc = (NodeIdAddressBind) ((DirApplication.MovePrimaryReply) Reply).result;
                    if (this.app.myServer.GetNodeIdAddressBind.Equals(primaryLoc))
                    {
                        if (DirApplication.DebugOutput) Console.WriteLine("    Primary not moved for {0} on node {1}",
                                                            dd.DirectoryName, this.app.myServer.GetNodeIdAddressBind.GetAddress);
                        return;
                    }
                    if (DirApplication.DebugOutput) Console.WriteLine("    Converting old primary for {0} on node {1} to replica",
                                                        dd.DirectoryName, this.app.myServer.GetNodeIdAddressBind.GetAddress);
                    Log.Write(DirApplication.CATEGORY, LogLvl.Med, "Demoting local copy of '{0}' to replica", dd.DirectoryName);
                    dd.BecomeReplica(primaryLoc);
                }
                else 
                {
                    dd.CreateBackups(); // Still the primary.  But our leaf set may have changed.  Make sure backups exist.
                }
            }
        }

//========================= VerifyPrimary client  =========================//

        //
        // Route information necessary to recreate a directory's primary copy on another node.
        // Needed when the primary fails so a new primary can be designated.
        //
        public void VerifyPrimary() 
        {
            CheckConsistency();
            if (DirApplication.DebugOutput) {
                Console.Write("VerifyPrimary on {0} for dir {1}",
                    app.myServer.GetNodeIdAddressBind.GetAddress, DirectoryName);
                if (DirApplication.DebugTimingOutput) Console.Write(" at time {0}", app.local.GetTime());
                Console.WriteLine();
            }

            // Route to ID of directory
            DirReplicaDescriptor drd = CreateDirReplicaDescriptor();

            DirApplication.VerifyPrimaryMsg msg =
                new DirApplication.VerifyPrimaryMsg(app.myServer.GetNodeIdAddressBind, this.dirId, drd);
            app.myServer.RouteMessage(msg, DirApplication.useCLB ? (NodeId) dirId.clbId : (NodeId) dirId);
        }

//========================= Per-directory NeighbourRemoved handler  =========================//

        public void NeighbourRemoved(MessageRouting.NodeIdAddressBind id, bool failure)
        {
            CheckConsistency();
            if (AdjacentToPrimary)
                VerifyPrimary();

            if (replicaNodes.Contains(id)) 
            {
                DirReplicaInfo dri = (DirReplicaInfo) replicaNodes[id];
                if (IsPrimaryReplica && dri.isBackup) 
                {
                    CreateBackups();    // Create a new backup since our current backup may be gone
                }

                if (failure) replicaNodes.Remove(id);   // Don't try to communicate with failed nodes
            }
        }

//========================= UpdateReplica client code  =========================//

        // Primary replica's update of other replicas.
        // Returns a flag indicating whether the caller should immediately reply
        // to the client (return value is true) or whether a reply to the caller
        // will be generated automatically after all other replicas have been updated
        // (return value is false).
        public bool UpdateOtherReplicas(NameValueEntry entry, DirRequestMsg msgToReplyTo) 
        {
            CheckConsistency();
            if (replicaNodes.Count == 0) return true;

            DirUpdateInfo dui = new DirUpdateInfo(entry, seqNum);
            this.lastSeqNumSeen = seqNum++;
            UpdateOtherReplicas(dui, msgToReplyTo);

            return false;
        }

        public bool UpdateOtherReplicas(DirMsgRecord dui, DirRequestMsg msgToReplyTo) {
            CheckConsistency();
            if (replicaNodes.Count == 0) return true;
            if (DirApplication.DebugOutput) {
                Console.Write("UpdateOtherReplicas on node {0}: {1}",
                    app.myServer.GetNodeIdAddressBind.GetAddress, this.name);
                if (DirApplication.DebugTimingOutput) Console.Write(" at time {0}", this.app.local.GetTime());
                Console.WriteLine();
            }

            int numReplicaNodes = replicaNodes.Count;
            UpdateReplicaCallback cb = new UpdateReplicaCallback(app, this, numReplicaNodes, msgToReplyTo);

            foreach (DirReplicaInfo dri in replicaNodes.Values) 
            {
                if (DirApplication.DebugOutput) Console.WriteLine("    Sending UpdateReplica from {0} to {1}", app.myServer.GetNodeIdAddressBind.GetAddress, dri.replicaNode.GetAddress);
                Log.Write(DirApplication.CATEGORY, LogLvl.Med, "Sending update to {0}", dri.replicaNode.GetAddress.ToString());
                DirApplication.UpdateReplicaRequest request =
                    new DirApplication.UpdateReplicaRequest(app.myServer, cb, app.myServer.GetNodeIdAddressBind, dri.replicaNode.GetNodeId, dui, IsPrimaryReplica);
                app.myServer.SendMessageDirect(request, dri.replicaNode);
            }

            return false;
        }

        class UpdateReplicaCallback: DirApplication.DirAppCallback {
            private DirectoryDescriptor dd;
            private int numReplicaMsgs;
            private int numRepliesSeen;

            // If we're the primary replica then this is the client request
            // message to reply to when we're done updating all other replicas.
            // Null if not.
            private DirRequestMsg msgToReplyTo;

            public UpdateReplicaCallback(DirApplication dirApp, DirectoryDescriptor dd, int numReplicaMsgs, DirRequestMsg msgToReplyTo)
                : base(dirApp)
            {
                this.dd = dd;
                this.numReplicaMsgs = numReplicaMsgs;
                this.numRepliesSeen = 0;
                this.msgToReplyTo = msgToReplyTo;
            }

            public override void Execute(Node localNode) {
                if (DirApplication.DebugOutput) Console.WriteLine("UpdateReplicaCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);

                if (Reply != null && ((DirApplication.UpdateReplicaReply) Reply).result is int) {
                    // Request for resending missed updates.
                    // Send either the requested updates or the entire
                    // directory if we haven't kept all the necessary
                    // updates.
                    // *** For now, just send the directory. ***
                    dd.FullUpdateReplica(((DirApplication.UpdateReplicaReply) Reply).Source);
                }
                this.numRepliesSeen++;
                if (this.numRepliesSeen == this.numReplicaMsgs) {
                    if (this.dd.primaryReplica) this.UpdateStats(DirApplication.DirOperations.updateReplica);

                    // Reply to original requestor.
                    if (this.msgToReplyTo == null) 
                    {
                        if (DirApplication.DebugOutput) Console.WriteLine("UpdateReplicaCallback: Missing replyTo msg");
                        return;
                    }
                    if (DirApplication.DebugOutput) {
                        Console.Write("UpdateReplicaCallback on {0}: Sending reply message to {1}",
                            this.app.myServer.GetPhysicalAddr, msgToReplyTo.Source.GetAddress);
                        if (DirApplication.DebugMsgIdOutput) Console.Write(" (msgId: {0})", msgToReplyTo.MsgId);
                        Console.WriteLine();
                    }

                    DirReplyMsg reply = null;
                    if (msgToReplyTo is DirApplication.UpdateNameValueRequest)
                        reply = new DirApplication.UpdateNameValueReply(msgToReplyTo, app.myServer.GetNodeIdAddressBind, null);
                    else if (msgToReplyTo is DirApplication.UpdateReplicaRequest)
                        reply = new DirApplication.UpdateReplicaReply(msgToReplyTo, app.myServer.GetNodeIdAddressBind, null);
                    else
                        Debug.Assert(msgToReplyTo is DirApplication.UpdateNameValueRequest || msgToReplyTo is DirApplication.UpdateReplicaRequest);
                    app.myServer.SendReply(reply, msgToReplyTo.Source);
                }
            }
        }

//========================= ReplicaHeartBeat client  =========================//

        public void SendReplicaHeartBeat() 
        {
            CheckConsistency();
            foreach (DirReplicaInfo dri in replicaNodes.Values) 
            {
                if (DirApplication.DebugOutput) Console.WriteLine("    Sending ReplicaHeartBeat from {0} to node {1}", this.app.myServer.GetNodeIdAddressBind.GetAddress, dri.replicaNode.GetAddress);

                DirApplication.ReplicaHeartBeatMsg msg =
                    new DirApplication.ReplicaHeartBeatMsg(app.myServer.GetNodeIdAddressBind, dri.replicaNode.GetNodeId, this.name, IsPrimaryReplica);
                app.myServer.SendMessageDirect(msg, dri.replicaNode);
            }
        }


//========================= RmReplica client  =========================//

        public void SendRmReplica() 
        {
            CheckConsistency();
            foreach (DirReplicaInfo dri in replicaNodes.Values) 
            {
                if (DirApplication.DebugOutput) Console.WriteLine("    Sending RmReplica from {0} to node {1}", this.app.myServer.GetNodeIdAddressBind.GetAddress, dri.replicaNode.GetAddress);

                DirApplication.RmReplicaMsg msg =
                    new DirApplication.RmReplicaMsg(app.myServer.GetNodeIdAddressBind, dri.replicaNode.GetNodeId, this.name);
                app.myServer.SendMessageDirect(msg, dri.replicaNode);
            }
        }

//========================= UpdateReplica client code for full update  =========================//

        public void FullUpdateReplica(NodeIdAddressBind replicaNode) 
        {
            CheckConsistency();
            DirReplicaDescriptor drd = CreateDirReplicaDescriptor();

            FullUpdateReplicaCallback cb = new FullUpdateReplicaCallback(app);
            DirApplication.UpdateReplicaRequest request =
                new DirApplication.UpdateReplicaRequest(app.myServer, cb, app.myServer.GetNodeIdAddressBind, replicaNode.GetNodeId, drd, IsPrimaryReplica);
            app.myServer.SendMessageDirect(request, replicaNode);
        }

        class FullUpdateReplicaCallback: DirApplication.DirAppCallback {
            public FullUpdateReplicaCallback(DirApplication dirApp)
                : base (dirApp)
            {}

            public override void Execute(Node localNode) {
                if (DirApplication.DebugOutput)
                    Console.WriteLine("FullUpdateReplicaCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);
            }
        }

//========================= ResyncReplica client  =========================//

        public void ResyncReplica() 
        {
            CheckConsistency();
            if (DirApplication.DebugOutput) Console.WriteLine("ResyncReplica on node {0}: {1}", this.app.myServer.GetNodeIdAddressBind.GetAddress, this.name);
            ResyncDescriptor rd = new ResyncDescriptor(this.name, this.lastSeqNumSeen);

            ResyncReplicaCallback cb = new ResyncReplicaCallback(app);
            DirApplication.ResyncReplicaRequest request =
                new DirApplication.ResyncReplicaRequest(app.myServer, cb, app.myServer.GetNodeIdAddressBind, this.dirId, rd);
            app.myServer.RouteMessage(request, DirApplication.useCLB ? (NodeId) dirId.clbId : (NodeId) dirId);
        }

        class ResyncReplicaCallback: DirApplication.DirAppCallback {
            public ResyncReplicaCallback(DirApplication dirApp)
                : base (dirApp)
            {
            }

            public override void Execute(Node localNode) {
                if (DirApplication.DebugOutput) Console.WriteLine("ResyncReplicaCallback on node {0}", localNode.GetNodeIdAddressBind.GetAddress);
                this.UpdateStats(DirApplication.DirOperations.resyncReplica);

                if (Reply != null && ((DirApplication.ResyncReplicaReply) Reply).result != null) 
                {
                    if (((DirApplication.ResyncReplicaReply) Reply).result is DirReplicaDescriptor) 
                    {
                        DirReplicaDescriptor drd = (DirReplicaDescriptor) ((DirApplication.ResyncReplicaReply) Reply).result;
                        DirectoryId dirId = new DirectoryId(drd.name);
                        if (DirApplication.DebugOutput) Console.WriteLine("Received FULL ResyncReplica reply on node {0}: {1}", 
                                                            app.myServer.GetNodeIdAddressBind.GetAddress, drd.name);

                        // Replace existing directory info with new info.
                        DirectoryDescriptor dir = (DirectoryDescriptor)this.app.directories[dirId];
                        dir.FullDirectoryUpdate(drd.EntriesHashtable(), ((DirApplication.ResyncReplicaReply) Reply).Source, drd.lastSeqNumSeen);
                        dir.ParentMsgSeen = true;
                        dir.Orphaned = false;
                        dir.AdjacentToPrimary = ((DirApplication.ResyncReplicaReply) Reply).fromPrimary;
                    }
                    else 
                    {
                        if (DirApplication.DebugOutput) Console.WriteLine("ResyncReplicaCallback: unexpected return type");
                    }
                }
            }
        }

//========================= DirectoryDescriptor utility methods  =========================//

        public bool Orphaned {
            get {return this.orphaned;}
            set {this.orphaned = value;}
        }

        public void AddReplica(NodeIdAddressBind replicaNode, bool isBackup) {
            CheckConsistency();
            if (!replicaNodes.Contains(replicaNode)) {
                replicaNodes[replicaNode] = new DirReplicaInfo(replicaNode, isBackup);
            }
            CheckConsistency();
        }

        public string DirectoryName {
            get {return this.name;}
        }

        public bool ParentMsgSeen {
            get {return parentMsgSeen;}
            set {parentMsgSeen = value;}
        }

        public int LastSeqNumSeen {
            get {return lastSeqNumSeen;}
            set {lastSeqNumSeen = value;}
        }

        public int NumReplicas {
            get {return replicaNodes.Count;}
        }
    }


//========================= DirReplicaInfo  =========================//

    public class DirReplicaInfo {
        public NodeIdAddressBind replicaNode;
        public bool isBackup;

        public DirReplicaInfo(NodeIdAddressBind replicaNode, bool isBackup) {
            this.replicaNode = replicaNode;
            this.isBackup = isBackup;   // True if a fail-over backup replica on a leaf set node of a primary replica
        }
    }


//========================= NameValueEntry  =========================//

    [Serializable] public abstract class DirMsgRecord {
        public abstract int DescriptorSize {get;}
    }

    [Serializable,
    XmlRootAttribute("NameValueEntry",Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)]
    public class NameValueEntry : DirMsgRecord {

        private string dirPathName;
        private string name;
        DirEntryType type;

        // This is null in GetNameValue requests.
        // If this is null as part of an UpdateNameValue request then 'name' should be deleted.
        private string val;

        public NameValueEntry() {
        }

        public NameValueEntry(string dirPathName, string name, DirEntryType type, string val) {
            this.dirPathName = dirPathName;
            this.name = name;
            this.type = type;
            this.val = val;
        }

        public NameValueEntry(string dirPathName, string name) {
            this.dirPathName = dirPathName;
            this.name = name;
            this.val = null;
        }

        public override int DescriptorSize {
            get {
                int msgSize = dirPathName.Length + 1 + name.Length + 1 + 4;
                if (val != null) msgSize += val.Length + 1;
                return msgSize;
            }
        }

        [XmlElementAttribute("NameValueEntryDirectory",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public string EntryDirectory {
            get {return dirPathName;}
            set {dirPathName = value;}
        }

        [XmlElementAttribute("NameValueEntryName",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public string EntryName {
            get {return name;}
            set {name = value;}
        }

        [XmlElementAttribute("NameValueEntryType",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public DirEntryType EntryType {
            get {return type;}
            set {type = value;}
        }

        [XmlElementAttribute("NameValueEntryValue",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public string EntryValue {
            get {return val;}
            set {val = value;}
        }


        public string FullPathName() {
            string s = this.dirPathName;
            if (! DirApplication.IsRootPathName(s))
                s += "/";
            s += this.name;
            return s;
        }
    }

//========================= DirReplicaDescriptor  =========================//

    [Serializable,
    XmlRootAttribute("DirReplicaDescriptor",Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)] 
    public class DirReplicaDescriptor : DirMsgRecord {
        public string name;
        public DirectoryId dirId;
        public int lastSeqNumSeen;
        private NameValueEntry[] entries;

        public DirReplicaDescriptor() {
        }

        public DirReplicaDescriptor(string name, DirectoryId dirId, Hashtable entriesTbl, int lastSeqNumSeen) {
            this.name = name;
            this.dirId = dirId;
            this.lastSeqNumSeen = lastSeqNumSeen;
            this.entries = new NameValueEntry[entriesTbl.Count];
            int i = 0;
            foreach (string eName in entriesTbl.Keys) {
                ValueEntry e = (ValueEntry) entriesTbl[eName];
                this.entries[i++] = new NameValueEntry(null, eName, e.EntryType, e.EntryValue);
            }
        }

        [XmlArray( ElementName="DirReplicaDescriptorEntries",
             Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="DirReplicaDescriptorElem", Type=typeof(NameValueEntry),
             IsNullable=true, Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
        public NameValueEntry[] Entries {
            get {return this.entries;}
            set {this.entries = value;}
        }

        public override int DescriptorSize {
            get {
                int msgSize = this.name.Length + 16 + 4;
                for (int i = 0; i < this.entries.Length; i++) {
                    msgSize += this.entries[i].EntryName.Length + 4 + 1;
                    if (this.entries[i].EntryValue != null) msgSize += this.entries[i].EntryValue.Length + 1;
                }
                return msgSize;
            }
        }

        public Hashtable EntriesHashtable() {
            Hashtable tbl = new Hashtable();
            for (int i = 0; i < this.entries.Length; i++) {
                tbl[this.entries[i].EntryName] = new ValueEntry(this.entries[i].EntryType, this.entries[i].EntryValue);
            }
            return tbl;
        }
    }

//========================= DirUpdateInfo  =========================//

    [Serializable,
    XmlRootAttribute("DirUpdateInfo",Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)] 
    public class DirUpdateInfo : DirMsgRecord {
        public NameValueEntry entry;
        public int seqNum;

        public DirUpdateInfo() {
        }

        public DirUpdateInfo(NameValueEntry entry, int seqNum) {
            this.entry = entry;
            this.seqNum = seqNum;
        }

        public override int DescriptorSize {
            get {return entry.DescriptorSize + 4;}
        }
    }

//========================= ResyncDescriptor  =========================//

    [Serializable,
    XmlRootAttribute("ResyncDescriptor",Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)] 
    public class ResyncDescriptor : DirMsgRecord {
        public string dirName;
        public int lastSeqNumSeen;

        public ResyncDescriptor() {
        }

        public ResyncDescriptor(string dirName, int lastSeqNumSeen) {
            this.dirName = dirName;
            this.lastSeqNumSeen = lastSeqNumSeen;
        }

        public override int DescriptorSize {
            get {return this.dirName.Length + 4;}
        }
    }

    public class HighLatencyCachedDirEntryInfo 
    {
        public NameValueEntry cEntry;
        public long expirationTime;
        public long usageCount;

        public HighLatencyCachedDirEntryInfo(NameValueEntry cEntry, long expirationTime)
        {
            this.cEntry = cEntry;
            this.expirationTime = expirationTime;
            this.usageCount = 0;
        }
    }
}
