using System;
using System.Diagnostics;
using System.Security.Cryptography;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.SkipNet;

namespace Overlook {
	/// <summary>
	/// Summary description for DirId.
	/// </summary>
	[Serializable,
	XmlRootAttribute("DirectoryId",Namespace=SerializationConfig.OVERLOOK_NAMESPACE,IsNullable=true)]
	public class DirectoryId: MessageRouting.NodeId {
		private static SHA1 md = new SHA1Managed();
		private static int M = 2;

		public ulong[] myId = new ulong[M];  // *** Public for now, to simplify serialization. ***
		private string dirName;
        public CLBId clbId; // XXX SkipNet specific

		public DirectoryId() {
		}

		public DirectoryId(string dirName) {
			this.dirName = dirName;

			byte[] data = new Byte[dirName.Length];
			char[] chars = dirName.ToCharArray(); 
			for (int i = 0; i < chars.Length; i++)
				data[i] = (byte) chars[i];
			byte[] digest = md.ComputeHash(data);
			SetNodeId(digest);

            // Calcluate a CLBId for the directory name

            int firstSepIndex = dirName.IndexOf('!');
            if (firstSepIndex < 0) 
            {
                clbId = new CLBId("", dirName); // No separator -- CLB route over entire overlay
            } 
            else 
            {
                Debug.Assert( firstSepIndex<dirName.Length-1, "dirName must not terminate with !" );
                
                // Set the constraint based on the first half of the CLB string
                string prefix = dirName.Substring(0, firstSepIndex);

                // Set the numeric ID based on the second half of the CLB string
                string suffix = dirName.Substring(firstSepIndex+1, dirName.Length-firstSepIndex-1);

                clbId = new CLBId(prefix, suffix);
            }

		}
	
		// Set the node id using a byte array
		private void SetNodeId(byte[] n) {
			for (int i=0, j=0; i<n.Length && j < M; i++) {
				myId[j] |= n[i];
				if ((i % 8) == 7) 
					j++;
				else
					myId[j] <<= 8;
			} 
		}

		[XmlElementAttribute("DirName",
			 Namespace=Overlook.SerializationConfig.OVERLOOK_NAMESPACE)]
		public string Name {
			get {return this.dirName;}
			set {this.dirName = value;}
		}

		// overridden hashcode (used for Hashtables)
		public override int GetHashCode() {
			return ((int) (myId[0] & 0xffffffff));	// use 32 bits from myId as hashcode
		}

		// overridden equals (used for Hashtables)
		public override bool Equals(Object obj) {
			if (!(obj is DirectoryId))
				return false;
			return Equals((DirectoryId)obj);
		}

		// test equality
		public bool Equals(DirectoryId d) {
			for (int i=0; i<M; i++) 
				if (myId[i] != d.myId[i]) return false;	
			return true;
		}
	
		// Return true if this id is >= to compare
		public bool GreaterThanOrEqual(DirectoryId compare) {
			for (int i = M-1; i >= 0; i--) {
				if (myId[i] != compare.myId[i]) {
					return myId[i] > compare.myId[i];
				}
			}
			return true;
		}

		// Return true if this id is <= to compare
		public bool LessThanOrEqual(DirectoryId compare) {
			for (int i = M-1; i >= 0; i--) {
				if (myId[i] != compare.myId[i]) {
					return myId[i] < compare.myId[i];
				}
			}
		
			return true;
		}
    
		public override string ToString() {
			string str = "";
			for (int i=M-1; i>=0; i--) 
				str += string.Format("{0:X16}",myId[i]);
			return str;
		}

		// Is this topic ID between l and r?
		public /* XXX override */ bool IsBetween(MessageRouting.NodeId l, MessageRouting.NodeId r) {
			DirectoryId left = (DirectoryId) l;
			DirectoryId right = (DirectoryId) r;

			if (left.Equals(right)) {
				// l == r
				if (this.Equals(left)) 
					return true;		// left == this == right
				else				   
					return false;
			}
			else if (left.LessThanOrEqual(right)) {
				// l < r (no wrapping)
				if (left.LessThanOrEqual(this) && this.LessThanOrEqual(right)) 
					return true; // left <= this <= right
				else 
					return false;
			}
			else {
				// l > r (wrapping)
				if (left.LessThanOrEqual(this) || right.GreaterThanOrEqual(this)) 
					return true; // left <= this OR right >= this
				else 
					return false;
			}
		}
	}
}
