//
// OverlookTryout -- Exercise Overlook in either the simulator or on the cluster
//

using System;
using System.Collections;
using System.Diagnostics;
using Application = MessageRouting.Application;
using SendMessage = MessageRouting.SendMessage;
using LocalEvent = MessageRouting.LocalEvent;
using Topology = MessageRouting.Topology;
using NodePhysicalAddress = MessageRouting.NodePhysicalAddress;
using SkipNet=Microsoft.Research.Herald.SkipNet;
using Overlook;
using Microsoft.Research.Herald.Common;
using Microsoft.Research.Herald.Common.ControlFlow;
using Microsoft.Research.Herald.Logger;

namespace OverlookTryout
{
    public class App
    {
        // Constants
        const string CATEGORY = "OverlookTryout";

        // Default parameters
        static bool quitImmediately = false;
        public static bool use_clb = true;
        static bool assertOnError = false;

        // Debugging controls
        public static bool LogDetails = false;
        static Log log = new Log(CATEGORY);
        
        public static string appOrgName, name_prefix, name_prefix_dir;

        // KeepAliveNode is just an arbitrarily chosen node in the org
        // given by name prefix. We keep this node alive otherwise the entire
        // org will be destroyed and its data will be lost.
        static int keepAliveNode = -1;

         // To be initialized by OverlookTryout_sim or OverlookTryout_cluster
        public static ExecutionEnvironment exEnv;
        public static ApplicationEnvironment appEnv;

        static Random random;
 
        public static ControlFlowApplication node0cfApp;

//================================ Helper functions for application steps ================================

        public static int errorCount = 0;

        // Count this error and assert if -AssertOnError specified
        static void CountError()
        {
            errorCount++;
            if (assertOnError)
                Debug.Assert(false);
        }

        // Count this error but don't assert even if -AssertOnError specified -- allows known bugs to be worked around
        static void CountErrorNoAssert()
        {
            errorCount++;
        }

        static object myMkDirResult = null;
        class myMkDirCallback : DirAPI.MkDirCallback
        {
            public int pc;
            public myMkDirCallback(int pc) { this.pc = pc;}
            public override void Callback() 
            {
                callbackStep = pc;
                // Null result here is fine

                myMkDirResult = result;
                ApplicationStepCallback();
            }
        }

        static object myRmDirResult = null;
        class myRmDirCallback : DirAPI.RmDirCallback 
        {
            public int pc;
            public myRmDirCallback(int pc) { this.pc = pc;}
            public override void Callback() 
            {
                callbackStep = pc;
                myRmDirResult = result;
                ApplicationStepCallback();
            }
        }

        static object myListDirResult = null;
        class myListDirCallback : DirAPI.ListDirCallback 
        {
            public int pc;
            public myListDirCallback(int pc) { this.pc = pc;}
            public override void Callback() 
            {
                callbackStep = pc;
                myListDirResult = result;
                ApplicationStepCallback();
            }
        }

        static object myGetNameValueResult = null;
        class myGetNameValueCallback : DirAPI.GetNameValueCallback 
        {
            public int pc;
            public myGetNameValueCallback(int pc) { this.pc = pc;}
            public override void Callback() 
            {
                callbackStep = pc;
                myGetNameValueResult = result;
                ApplicationStepCallback();
            }
        }

        static object myUpdateNameValueResult = null;
        class myUpdateNameValueCallback : DirAPI.UpdateNameValueCallback 
        {
            public int pc;
            public myUpdateNameValueCallback(int pc) { this.pc = pc;}
            public override void Callback() 
            {
                callbackStep = pc;
                // Null result is good here
                myUpdateNameValueResult = result;
                ApplicationStepCallback();
            }
        }

        // Called by program steps that don't generate a callback
        static void NullCallback()
        {
            if (ControlFlowApplication.DebugOutput && exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                Console.WriteLine("NullCallback: {0}, thread {1}", programStep,
                    System.Threading.Thread.CurrentThread.Name);
            callbackStep = callbackShouldBeFrom = programStep;
            ApplicationStepCallback();
        }

        static string myMkDirPath;
        static void DoMkDir(string path)
        {
            myMkDirPath = path;
            dirApi.MkDir(path, new myMkDirCallback(callbackShouldBeFrom = programStep));
        }

        static void CheckMkDir()
        {
            if (myMkDirResult is DirException) 
            {
                Console.WriteLine("CheckMkDir error: MkDir({0}) returned exception {1}",
                    myMkDirPath, ((DirException)myMkDirResult).ErrMsg);
                CountError();
                return;
            }
            Console.WriteLine("MkDir({0}) succeeded", myMkDirPath);
        }

        static string myUpdateNameValuePath;
        static string myUpdateNameValueName;
        static string myUpdateNameValueValue;
        static void DoUpdateNameValue(string path, string name, string val)
        {
            myUpdateNameValuePath = path;
            myUpdateNameValueName = name;
            myUpdateNameValueValue = val;
            dirApi.UpdateNameValue(path, name, val, new myUpdateNameValueCallback(callbackShouldBeFrom = programStep));
        }

        static void CheckUpdateNameValue()
        {
            if (myUpdateNameValueResult is DirException) 
            {
                Console.WriteLine("CheckUpdateNameValue error: UpdateNameValue({0}, {1}, {2}) returned unexpected exception {3}",
                    myUpdateNameValuePath, myUpdateNameValueName, myUpdateNameValueValue, ((DirException)myUpdateNameValueResult).ErrMsg);
                CountError();
            }
            else
            {
                if (myUpdateNameValueResult != null) 
                {
                    Console.WriteLine("CheckUpdateNameValue error: result not null");
                    CountError();
                    return;
                }
                Console.WriteLine("UpdateNameValue({0}, {1}, {2}) succeeded",
                    myUpdateNameValuePath, myUpdateNameValueName, myUpdateNameValueValue);
            }
        }

        static string myListDirPath;
        static void DoListDir(string path)
        {
            myListDirPath = path;
            dirApi.ListDir(path, new myListDirCallback(callbackShouldBeFrom = programStep));
        }

        static void CheckListDir(int nEntries)
        {
            if (myListDirResult == null)
            {
                Console.WriteLine("CheckListDir error: result null");
                CountError();
                return;
            }
            if (myListDirResult is DirException) 
            {
                Console.WriteLine("CheckListDir error: ListDir({0}) returned exception {1}",
                    myListDirPath, ((DirException)myListDirResult).ErrMsg);
                CountError();
                return;
            }
            if (! (myListDirResult is DirEntry []))
            {
                Console.WriteLine("CheckListDir error: ListDir({0}) unexpected value {1}",
                    myListDirPath, myListDirResult);
                CountError();
                return;
            }
            if (((DirEntry [])myListDirResult).Length != nEntries)
            {
                Console.WriteLine("CheckListDir error: result.Length {0} != expected entries {1}",
                    ((DirEntry[])myListDirResult).Length, nEntries);
                if (! use_clb) CountError(); else CountErrorNoAssert(); // XXX Allow bvt to pass despite known bug
                return;
            }            
            Console.WriteLine("ListDir({0}) succeeded => {1} entries", myListDirPath, nEntries);
        }

        static string myGetNameValuePath;
        static string myGetNameValueName;
        static void DoGetNameValue(string path, string name)
        {
            myGetNameValuePath = path;
            myGetNameValueName = name;
            dirApi.GetNameValue(path, name, new myGetNameValueCallback(callbackShouldBeFrom = programStep));
        }

        static void CheckGetNameValue(string val)
        {
            if (myGetNameValueResult == null) 
            {
                Console.WriteLine("CheckGetNameValue error: result null");
                CountError();
                return;
            }
            if (myGetNameValueResult is DirException) 
            {
                if (val == null) 
                {    // Null passed in when exception expected
                    Console.WriteLine("GetNameValue({0}, {1}) succeeded (received expected exception {2})",
                        myGetNameValuePath, myGetNameValueName, ((DirException)myGetNameValueResult).ErrMsg);
                }
                else
                {
                    Console.WriteLine("CheckGetNameValue error: GetNameValue({0}, {1}) returned unexpected exception {2}",
                        myGetNameValuePath, myGetNameValueName, ((DirException)myGetNameValueResult).ErrMsg);
                    CountError();
                }
            } 
            else if (myGetNameValueResult is NameValueEntry)
            {
                if (val != ((NameValueEntry) myGetNameValueResult).EntryValue) 
                {
                    Console.WriteLine("CheckGetNameValue error: result {0} != expected result {1}",
                        ((NameValueEntry) myGetNameValueResult).EntryValue, val == null ? "null" : val);
                    CountError();
                    return;
                }
                Console.WriteLine("GetNameValue({0}, {1}) succeeded => {2}", myGetNameValuePath, myGetNameValueName, val);
            }
            else
            {
                Console.WriteLine("CheckGetNameValue error: result not DirException or NameValueEntry");
                CountError();
            }
        }

        static string myRmDirPath;
        static void DoRmDir(string path)
        {
            myRmDirPath = path;
            dirApi.RmDir(path, new myRmDirCallback(callbackShouldBeFrom = programStep));
        }

        static void CheckRmDir()
        {
            if (myRmDirResult is DirException) 
            {
                Console.WriteLine("CheckRmDir error: RmDir({0}) returned exception {1}",
                    myRmDirPath, ((DirException)myRmDirResult).ErrMsg);
                CountError();
                return;
            }
            Console.WriteLine("RmDir({0}) succeeded", myRmDirPath);
        }

//================================ Control flow primitives ================================

        // XXX The code below is currently simulator-specific.
        // XXX This needs to turn into communication, rather than direct calls.

        public static int executeOnNode;
        public static int programStep;
        public static int loopIndex;
        public static object replyToken;
        static int callbackShouldBeFrom;
        static int callbackStep;

        const int SECONDS_TO_WAIT_FOR_RENDEZVOUS = 5 * 60; // Five minutes 

        public static volatile int awaitingRendezvousWithNode = 0;
        static void RendezvousWithNode(int nodeNumber)
        {
             // TBD right now assumes will be called with sequential node numbers
            appEnv.envRendezvousWithNode();
            exEnv.Go();

            if (nodeNumber != 0) // Avoid rendezvous on master control flow node to avoid circularity
            {
                Debug.Assert(App.awaitingRendezvousWithNode == nodeNumber - 1);

                long rendezvousStartTime = DateTime.Now.Ticks;
                if (ControlFlowApplication.DebugOutput && exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                    Console.WriteLine("RendezvousWithNode {0} started at time {1:T}", nodeNumber, DateTime.Now);
                int attempt = 0;
                int sleepForMs = 500; // Start with 1/2 second
                long timeoutMs = 5000; // Start with five seconds
                
                while (App.awaitingRendezvousWithNode < nodeNumber) // Wait for upcall to signal RendezvousWithNode completed
                {
                    attempt++;
                    if (ControlFlowApplication.DebugOutput) 
                    {
                        Console.Write("RendezvousWithNode to node {0} nodeNumber {1} attempt {2} timeoutMs {3}",
                            exEnv.NodeIdAddrBind(nodeNumber).GetAddress, nodeNumber, attempt, timeoutMs);
                        if (exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                            Console.Write(" at time {0:T}", DateTime.Now);
                        Console.WriteLine();
                    }

                    lock (exEnv.OverlayLock) 
                    {
                        node0cfApp.ClientRendezvousWithNode(exEnv.NodeIdAddrBind(nodeNumber), nodeNumber, attempt, timeoutMs);
                    }
                    exEnv.Go();
                    if (exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment) 
                    {
                        System.Threading.Thread.Sleep(sleepForMs);

                        if (App.awaitingRendezvousWithNode < nodeNumber &&
                            (DateTime.Now.Ticks - rendezvousStartTime)/10000000 > SECONDS_TO_WAIT_FOR_RENDEZVOUS)
                        {
                            Console.WriteLine("======== RendezvousWithNode for nodeNumber {0}, address {1} ran out of time -- {2} seconds exceeded -- giving up on it! ========",
                                nodeNumber, exEnv.NodeIdAddrBind(nodeNumber).GetAddress, SECONDS_TO_WAIT_FOR_RENDEZVOUS);
                            App.awaitingRendezvousWithNode = nodeNumber;
                            break;
                        }
                    }
                    // Exponential backoff
                    timeoutMs *= 2;
                    if (timeoutMs > SECONDS_TO_WAIT_FOR_RENDEZVOUS * 1000) timeoutMs = SECONDS_TO_WAIT_FOR_RENDEZVOUS * 1000;
                    sleepForMs *= 2;
                    if (sleepForMs > 8000) sleepForMs = 8000; // Check every 8 seconds no matter what
                }
                if (ControlFlowApplication.DebugOutput && exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                    Console.WriteLine("RendezvousWithNode {0} finished at time {1:T} after attempt {2}: took {3} seconds",
                        nodeNumber, DateTime.Now, attempt, (DateTime.Now.Ticks - rendezvousStartTime)/10000000);
            }
        }

        public static bool pendingOnNodeDoStep = false;
        static void OnNodeDoStep(int nodeNumber, int programStep, int loopIndex)
        {
            Debug.Assert(pendingOnNodeDoStep == false);
            pendingOnNodeDoStep = true;

            lock (exEnv.OverlayLock) 
            {
                node0cfApp.ClientOnNodeDoStep(exEnv.NodeIdAddrBind(nodeNumber),
                    nodeNumber, programStep, loopIndex);
            }

            while (pendingOnNodeDoStep) // Wait for upcall to signal OnNodeDoStep completed
            {
                exEnv.Go();
                if (exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                    System.Threading.Thread.Sleep(5); // XXX Replace with something better
            }
        }

        static void FailNode(int nodeNumber)
        {
            lock (exEnv.OverlayLock) 
            {
                node0cfApp.ClientFailNode(exEnv.NodeIdAddrBind(nodeNumber));
            }
            exEnv.Go();
        }

        static void ExitNode(int nodeNumber)
        {
            lock (exEnv.OverlayLock) 
            {
                node0cfApp.ClientExitNode(exEnv.NodeIdAddrBind(nodeNumber));
            }
            exEnv.Go();
        }

//================================ Application control flow code ================================

        static int rendezvousWith = 0;

        // Rendezvous with 4 nodes each time we're called until there's no more to rendezvous with
        static void RendezvousWithNodes()
        {
            for (int i = 0; i < 4; i++)
            {
                if (rendezvousWith >= exEnv.NodeCount) return;
                RendezvousWithNode(rendezvousWith);
                rendezvousWith++;
            }
        }

        static int failedThisPeriod = 0;
        static int periodLen = 0;
        static int[] failNodeOrdering;

        public static void ApplicationControlFlow()
        {
            foreach (int pc in new int [] {0, 100, 200, 300, 400, 500, 600})
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, pc, 0);
            }
            foreach (int pc in new int [] {700})
            {
                RendezvousWithNodes();
                OnNodeDoStep(exEnv.NodeCount > 1 ? 1 : 0, pc, 0);
            }
            foreach (int pc in new int [] {800, 820})
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, pc, 0);
            }
            if (exEnv.NodeCount > 2) FailNode(2); // Fail primary copy of /foo/bar directory
            foreach (int pc in new int [] {830, 850, 900, 1000, 1100, 1200, 1300})
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, pc, 0);
            }
            for (int i = 0; i < 50; i++)
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, 1400, i);
            }
            if (exEnv.NodeCount > 94)
                FailNode(94); // Fail secondary copy of /foo/bar directory

            foreach (int pc in new int [] {1500, 1550, 1600, 1700, 1800, 1850})
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, pc, 0);
            }
            for (int i = 0; i < 100; i++)
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, 1900, i);
            }
            foreach (int pc in new int [] {2000, 2100, 2200, 2300, 2400, 2500, 2600})
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, pc, 0);
            }
            for (int node = 0;
                node < (exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment
                ? rendezvousWith /* # nodes created */ // Cluster: On all non-failed nodes do...
                : 1); // Simulator: On node 0 do...
                node++)
            {
                if (node != 2 && node != 94) // Don't do it on failed nodes
                    OnNodeDoStep(node, 2650, 0); // DirApplication.DebugHighLatencyReplication = true;
            }
            foreach (int pc in new int [] {2700, 2800})
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, pc, 0);
            }
            for (int i = 0; i < 100; i++)
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, 2900, i);
            }
            foreach (int pc in new int [] {3000})
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, pc, 0);
            }
            for (int node = 0;
                node < (exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment
                ? rendezvousWith /* # nodes created */ // Cluster: On all non-failed nodes do...
                : 1); // Simulator: On node 0 do...
                node++)
            {
                if (node != 2 && node != 94) // Don't do it on failed nodes
                    OnNodeDoStep(node, 3050, 0); // DirApplication.DebugHighLatencyReplication = false;
            }
            foreach (int pc in new int [] {3100})
            {
                RendezvousWithNodes();
                OnNodeDoStep(0, pc, 0);
            }

            // Prepare an array of all node's indices that we 
            // This is the order in which to fail them
            int len = (exEnv.NodeCount > 2) ? exEnv.NodeCount - 2 : 1;
            failNodeOrdering = new int[len];
            for (int i=0; i<len; i++) failNodeOrdering[i] = i+1;

            // We must not kill the keep-alive node
            if (exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment && keepAliveNode == -1)
                keepAliveNode = 0; // XXX Short-term hack for the cluster until we get orgs figured out there
            Debug.Assert(keepAliveNode!=-1);
            if (keepAliveNode >= len) keepAliveNode = 0; // XXX
            failNodeOrdering[keepAliveNode] = -1;
                        
            // Randomly permute the indices
            for (int i=0; i<len-1; i++) 
            {
                int r = random.Next(len-1-i)+i+1;
                int tmp = failNodeOrdering[i];
                failNodeOrdering[i] = failNodeOrdering[r];
                failNodeOrdering[r] = tmp;
            }

            for (int i = 0; i < 201; i++)
            {
                RendezvousWithNodes();

                if (i > 0 && i < (exEnv.NodeCount - 1)) 
                {
                    int indexToFail = failNodeOrdering[i-1];
                    if (indexToFail != -1)
                    {
                        Console.Write("Failing node {0}", indexToFail);
                        Console.Write(" with address {0}", exEnv.NodePhysAddr(indexToFail));
                        Console.WriteLine();
                        log.Write(LogLvl.Med, "Failing node {0} ({1}th failure)", indexToFail, i);
                        FailNode(indexToFail);
                    } 
                    else
                    {
                        // -1 indicates the keep-alive node. We don't kill it.
                        keepAliveNode = 0;
                    }
                    failedThisPeriod++;
                }
                
                if (failedThisPeriod >= periodLen)
                {
                    // Choose a new length for this period -- i.e., fail 10% of the remaining nodes
                    failedThisPeriod = 0;
                    periodLen = (int) ((((double)(exEnv.NodeCount-i)) * 0.10)+1);

                    // Before making updates, give replicas enough time to declare themselves stale
                    // and enough time for R-Table repair
                    long delayMs = 2*DirectoryDescriptor.heartBeatTimeoutInterval;
                    if (SkipNet.Parameters.neighbourPingPeriod > delayMs)
                    {
                        delayMs = SkipNet.Parameters.neighbourPingPeriod;
                    }
                    if (exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                        if (delayMs > 10 * 1000) // Cap at 10 seconds on the cluster so this eventually terminates
                            delayMs = 10 * 1000;
                    exEnv.GoFor((int) delayMs);

                    OnNodeDoStep(0, 4000, i);

                    RendezvousWithNodes();
                    OnNodeDoStep(0, 4100, i);

                    RendezvousWithNodes();
                }
                OnNodeDoStep(0, 4200, i);

                RendezvousWithNodes();
                OnNodeDoStep(0, 4300, i);
            }

            for (int node = 1; node < exEnv.NodeCount; node++)
            {
                if (ControlFlowApplication.DebugOutput && exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                    Console.WriteLine("Calling ExitNode({0}) at time {1:T}: ", node, DateTime.Now);
                ExitNode(node); // Will cause a lot of failed sends on the cluster since sending to many failed nodes
            }
        }

//================================ Code for application program steps ================================

        public static SkipNet.Node localNode;
        static DirApplication dirApp;
        static DirAPI dirApi;

        static string val1;
        static string val2;

        //
        // Run one step of the application code.  Each step involves sending a request and waiting for a reply.
        //
        public static void ApplicationStep()
        {
            if (! exEnv.AmNode(executeOnNode) && exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                return; // XXX Won't be needed soon
            Debug.Assert(exEnv.AmNode(executeOnNode));
            dirApp = (DirApplication) localNode.GetApp(typeof(DirApplication));
            dirApi = dirApp.dirApi;

            if (LogDetails)
                Log.logStatistics(localNode); // JD 8/22/03 // Destroys reproducability -- placed under -L 8/22/03 mbj
            Console.Write("programStep = {0}: ", programStep);
            if (ControlFlowApplication.DebugOutput && exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                Console.Write("thread = {0}: ", System.Threading.Thread.CurrentThread.Name);
            if (ControlFlowApplication.DebugOutput && exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                Console.Write("at time {0:T}: ", DateTime.Now);

            switch (programStep)
            {
                // Line numbers, anyone? :-)
                case 0:
                    Console.WriteLine();
                    DoMkDir("/"); break;
                case 100:
                    DoMkDir("/foo"); break;
                case 200:
                    DoMkDir("/foo/bar"); break;
                case 300:
                    DoUpdateNameValue("/foo/bar", "a", "13"); break;
                case 400:
                    DoUpdateNameValue("/foo/bar", "b", "42"); break;
                case 500:
                    DoUpdateNameValue("/foo/bar", "c", "17"); break;
                case 600:
                    DoListDir("/foo/bar"); break;
                case 700:
                    DoGetNameValue("/foo/bar", "a"); break;
                case 800:
                    DoListDir("/foo"); break;
                case 820:
                    DoListDir("/foo"); break;
                case 830:
                    DoGetNameValue("/foo/bar", "a"); break;
                case 850:
                    DoUpdateNameValue("/foo/bar", "c", "169"); break;
                case 900:
                    DoGetNameValue("/foo/bar", "c"); break;
                case 1000: 
                    DoUpdateNameValue("/foo/bar", "b", "43"); break;
                case 1100:
                    DoGetNameValue("/foo/bar", "b"); break;
                case 1200:
                    DoUpdateNameValue("/foo/bar", "a", "23"); break;
                case 1300:
                    DoGetNameValue("/foo/bar", "a"); break;
                case 1400:
                    Console.Write("loopIndex = {0}: ", loopIndex);

                    // for (int i = 0; i < 50; i++) 
                    // {
                    //     GetNameValue("/foo/bar", "a");
                    // }
                    DoGetNameValue("/foo/bar", "a"); break;
                case 1500:
                    Console.WriteLine();
                    DoGetNameValue("/foo/bar", "a"); break;
                case 1550:
                    DoRmDir("/foo/bar"); break;
                case 1600:
                    // Try to get value from directory replica even though primary has been deleted
                    DoGetNameValue("/foo/bar", "a"); break;
                case 1700: 
                    DoMkDir("/foo/bar"); break;
                case 1800:
                    DoUpdateNameValue("/foo/bar", "a", "03221960"); break;
                case 1850:
                    DoGetNameValue("/foo/bar", "a"); break;
                case 1900:
                    Console.Write("loopIndex = {0}: ", loopIndex);
                    
                    // for (int i = 0; i < 100; i++) 
                    // {
                    //     GetNameValue("/foo/bar", "a");
                    // }
                    DoGetNameValue("/foo/bar", "a"); break;
                case 2000:
                    Console.WriteLine();
                    DoMkDir(name_prefix_dir); break; // Either a CLB root or just a regular directory
                case 2100:
                    DoMkDir(name_prefix + "foo"); break;
                case 2200:
                    DoMkDir(name_prefix + "foo/bar"); break;
                case 2300:
                    DoUpdateNameValue(name_prefix + "foo/bar", "a", "13"); break;
                case 2400:
                    DoUpdateNameValue(name_prefix + "foo/bar", "b", "42"); break;
                case 2500:
                    DoUpdateNameValue(name_prefix + "foo/bar", "c", "17"); break;
                case 2600:
                    DoListDir(name_prefix + "foo/bar"); break;
                case 2650:
                    DirApplication.DebugHighLatencyReplication = true;
                    NullCallback(); break;
                case 2700:
                    DoGetNameValue(name_prefix + "foo/bar", "a"); break;
                case 2800:
                    Console.WriteLine();
                    NullCallback(); break;
                case 2900:
                    Console.Write("loopIndex = {0}: ", loopIndex);

                    // for (int i = 0; i < 100; i++) 
                    // {
                    //     GetNameValue(name_prefix + "foo/bar", "a");
                    // }
                    DoGetNameValue(name_prefix + "foo/bar", "a"); break;
                case 3000:
                    DoRmDir(name_prefix + "foo/bar"); break;
                case 3050:
                    DirApplication.DebugHighLatencyReplication = false;
                    NullCallback(); break;
                case 3100:
                    DoMkDir(name_prefix + "foo/bar"); break;
                case 4000:
                    // for (int i = 0; i < 100; i++) 
                    // {
                    //     if (i > 0 && i < (nodes - 1)) fail node[i]; // Keep first & last nodes
                    //     if (i % UpdatePeriod == 0) {
                    //         UpdateNameValue("/foo/bar", "a", "val1 " + (i % UpdatePeriod));
                    //         UpdateNameValue(name_prefix + "foo/bar", "a", "val2 " + (i % UpdatePeriod));
                    //     }
                    //     GetNameValue("/foo/bar", "a");
                    //     GetNameValue(name_prefix + "foo/bar", "a");
                    // }

                    Console.Write("loopIndex = {0}: ", loopIndex);
                    val1 = "val1_" + loopIndex;
                    DoUpdateNameValue("/foo/bar", "a", val1); break;
                case 4100:
                    val2 = "val2_" + loopIndex;
                    DoUpdateNameValue(name_prefix + "foo/bar", "a", val2); break;
                case 4200:
                    Console.Write("loopIndex = {0}: ", loopIndex);
                    DoGetNameValue("/foo/bar", "a"); break;
                case 4300:
                    DoGetNameValue(name_prefix + "foo/bar", "a"); break;

                default:
                    Console.WriteLine("Unexpected program counter value " + programStep);
                    ExitProg(1);
                    break;
            }
            exEnv.Go();
        }

        //
        // Take appropriate action upon receiving the reply from an application program step.
        //
        public static void ApplicationStepCallback()
        {
            if (exEnv.Type() == ExecutionEnvironmentType.SimulatorEnvironment)
                Debug.Assert(callbackStep == callbackShouldBeFrom); // True if callback has occurred
            else if (callbackStep != callbackShouldBeFrom)
                Console.WriteLine("*** Assertion would have failed: callbackStep {0} != callbackShouldBeFrom {1} ***",
                    callbackStep, callbackShouldBeFrom);
    
            Console.Write("callbackStep = {0}: ", callbackStep);
            if (ControlFlowApplication.DebugOutput && exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                Console.Write("thread = {0}: ", System.Threading.Thread.CurrentThread.Name);
            if (ControlFlowApplication.DebugOutput && exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                Console.Write("at time {0:T}: ", DateTime.Now);

            switch (callbackStep)
            {
                case 0:
                    CheckMkDir(); break;
                case 100:
                    CheckMkDir(); break;
                case 200:
                    CheckMkDir(); break;
                case 300:
                    CheckUpdateNameValue(); break;
                case 400:
                    CheckUpdateNameValue(); break;
                case 500:
                    CheckUpdateNameValue(); break;
                case 600:
                    CheckListDir(3); break;
                case 700:
                    CheckGetNameValue("13"); break;
                case 800:
                    CheckListDir(1); break;
                case 820:
                    CheckListDir(1); break;
                case 830:
                    CheckGetNameValue("13"); break;
                case 850:
                    CheckUpdateNameValue(); break;
                case 900:
                    CheckGetNameValue("169"); break;
                case 1000: 
                    CheckUpdateNameValue(); break;
                case 1100:
                    CheckGetNameValue("43"); break;
                case 1200:
                    CheckUpdateNameValue(); break;
                case 1300:
                    CheckGetNameValue("23"); break;
                case 1400:
                    CheckGetNameValue("23"); break;
                case 1500:
                    CheckGetNameValue("23"); break;
                case 1550:
                    CheckRmDir(); break;
                case 1600:
                    CheckGetNameValue(null); break;
                case 1700: 
                    CheckMkDir(); break;
                case 1800:
                    CheckUpdateNameValue(); break;
                case 1850:
                    CheckGetNameValue("03221960"); break;
                case 1900:
                    CheckGetNameValue("03221960"); break;
                case 2000:
                    CheckMkDir(); break;
                case 2100:
                    CheckMkDir(); break;
                case 2200:
                    CheckMkDir(); break;
                case 2300:
                    CheckUpdateNameValue(); break;
                case 2400:
                    CheckUpdateNameValue(); break;
                case 2500:
                    CheckUpdateNameValue(); break;
                case 2600:
                    CheckListDir(3); break;
                case 2650:
                    Console.WriteLine(); break;
                case 2700:
                    CheckGetNameValue("13"); break;
                case 2800:
                    Console.WriteLine(); break;
                case 2900:
                    CheckGetNameValue("13"); break;
                case 3000:
                    CheckRmDir(); break;
                case 3050:
                    Console.WriteLine(); break;
                case 3100:
                    CheckMkDir(); break;
                case 4000:
                    CheckUpdateNameValue(); break;
                case 4100:
                    CheckUpdateNameValue(); break;
                case 4200:
                    CheckGetNameValue(val1); break;
                case 4300:
                    CheckGetNameValue(val2); break;
            }

            if (exEnv.Type() == ExecutionEnvironmentType.SimulatorEnvironment) 
                Debug.Assert(replyToken != null);
            else if (replyToken == null)
                Console.WriteLine("*** Assertion would have failed: (replyToken != null) on thread {0} ***",
                    System.Threading.Thread.CurrentThread.Name);

            // Send control flow reply message now that statement has been completed
            SkipNet.Node node = exEnv.GetNode(executeOnNode);
            ControlFlowApplication cfApp = (ControlFlowApplication) node.GetApp(typeof(ControlFlowApplication));
            if (replyToken != null) cfApp.SendOnNodeDoStepReply(replyToken);
            replyToken = null;
        }

//================================ Handle program args and related state ================================

        /// <summary>
        /// Print the various command-line options to OverlookTryout.
        /// </summary>
        public static void Usage()
        {
            Console.WriteLine("OverlookTryout");

            appEnv.envPrintUsage();
            Console.WriteLine("  Common Parameters:");
            CommonParameters.PrintUsage();

            Console.WriteLine("  Application Parameters:");
            Console.WriteLine("    [-{0}clb] -- Use CLB routing for dir names.  Default is enabled.", "{e,d}");

            Console.WriteLine("  Debugging Parameters:");
            Console.WriteLine("    [-D] -- Turns on directory application debugging output");
            Console.WriteLine("    [-T] -- Turns on directory application timing debugging output");
            Console.WriteLine("    [-M] -- Turns on directory application MsgId debugging output");
            Console.WriteLine("    [-C] -- Turns on control flow application debugging output");
            Console.WriteLine("    [-A] -- Causes program to assert after each error detected");
            Console.WriteLine("    [-quit] -- Causes program to exit without waiting for input");
            ExitProg(0);
        }
       
        /// <summary>
        /// Print the parameters that are being used for this experiment
        /// </summary>
        public static void PrintExperimentParameters()
        {
            Console.WriteLine("Running OverlookTryout with parameters");

            appEnv.envPrintArgs();
            Console.WriteLine("  Common Parameters:");
            CommonParameters.PrintArgs();

            // SkipNet parameters
            Console.WriteLine("  Application Parameters:");
            Console.WriteLine("    -{0}clb", use_clb ? "e" : "d");

            Console.WriteLine();
            if (LogDetails)
                Console.WriteLine("    -LogDetails -- Turns on application detailed logging output");
            if (DirApplication.DebugOutput)
                Console.WriteLine("    -DebugOutput -- Turns on directory application debugging output");
            if (DirApplication.DebugTimingOutput)
                Console.WriteLine("    -TimingDebugOutput -- Turns on directory application timing debugging output");
            if (DirApplication.DebugMsgIdOutput)
                Console.WriteLine("    -MsgIdDebugOutput -- Turns on directory application MsgId debugging output");
            if (ControlFlowApplication.DebugOutput)
                Console.WriteLine("    -ControlFlowDebugOutput -- Turns on control flow application debugging output");
            if (assertOnError)
                Console.WriteLine("    -AssertOnError -- Causes the progam to assert after each error detected");
            if (quitImmediately)
                Console.WriteLine("    -quit -- Exit without waiting for input");
            Console.WriteLine();
        }

        /// <summary>
        /// Process the command-line options.
        /// </summary>
        public static void GetOptions(string[] args, int starting_at) 
        {
            for (int i = starting_at; i < args.Length; i++) 
            {
                if (args[i].Length > 0 && ((args[i][0] == '-') || (args[i][0] == '/'))) 
                {
                    switch(args[i][1]) 
                    {
                        case 'q':
                            quitImmediately = true;
                            break;
                        case 'd':   
                            if (args[i].Substring(2).Equals("clb")) 
                            {
                                // -dclb
                                use_clb = false;
                            } 
                            else 
                            {
                                Usage();
                            }
                            break;
                        case 'e':
                            if (args[i].Substring(2).Equals("clb")) 
                            {
                                // -eclb
                                use_clb = true;
                            } 
                            else
                            {
                                Usage();
                            }
                            break;
                        case 'L':
                            LogDetails = true;
                            appEnv.envLogDetails = true;
                            break;
                        case 'D':
                            DirApplication.DebugOutput = true;
                            Log.SetCategoryFilterLevel(SkipNet.CTable.CATEGORY, LogLvl.Hi); // Turn back on if -DebugOutput
                            break;
                        case 'T':
                            DirApplication.DebugTimingOutput = true;
                            break;
                        case 'M':
                            DirApplication.DebugMsgIdOutput = true;
                            break;
                        case 'C':
                            ControlFlowApplication.DebugOutput = true;
                            break;
                        case 'A':
                            assertOnError = true;
                            break;

                        default:
                            Usage();
                            break;
                    }
                }
                else
                    Usage();
            }

            // Fixup & Verify parameters
            appEnv.envParameterFixup();

            // No sense in enabling the clb table if we're not doing clb searches
            if( ! use_clb ) 
            {
                SkipNet.Parameters.cTableEnabled = false;
            }
        }

//======================== Functions to be called from environemnt-specific code ========================

        public static void SetParameterDefaults()
        {
            exEnv.NodeCount = 200;
            appEnv.envSetParameterDefaults();

            DirApplication.DefaultReplicationNumber = 3;

            SkipNet.Parameters.b = 8;
            SkipNet.Parameters.denseRTable = true;
            SkipNet.Parameters.proximityTableEnabled = true;
            SkipNet.Parameters.cTableEnabled = true;
            SkipNet.Parameters.clbTableEnabled = true;
            SkipNet.Parameters.pingEnabled = true;
            SkipNet.Parameters.clbTableNbrListSize = DirApplication.DefaultReplicationNumber;
            // SkipNet.Parameters.leafSetHalfSize = 4; // Default was 8

            //Log.SetCategoryFilterLevel(App.CATEGORY, LogLvl.Med);
            //Log.SetCategoryFilterLevel(DirApplication.CATEGORY, LogLvl.Med);
            Log.SetCategoryFilterLevel(SkipNet.LeafSetFixup.CATEGORY, LogLvl.Disabled);
            Log.SetCategoryFilterLevel(SkipNet.CTable.CATEGORY, LogLvl.Disabled); // Eliminate "C-Table Join Failed" messages caused by our many intentional node failures

            ControlFlowApplication.DebugOutput = false;
        }

        public static void SetRandomSeed(int seed)
        {
            random = new Random(seed);
        }

        public static void PostProcessParameters()
        {
            if (CommonParameters.seed == -1) 
            {
                CommonParameters.seed = 19600322;
            }
            if (! DirApplication.DebugOutput) 
            {
                Log.SetCategoryFilterLevel(DirApplication.CATEGORY, LogLvl.Disabled);
                Log.SetCategoryFilterLevel(SkipNet.ProximityTable.CATEGORY, LogLvl.Disabled);
            }
        }

        /// <summary>
        /// Each overlay has various configuration parameters. We set them here.
        /// </summary>
        public static void SetOverlayParameters(bool doRandom) 
        {
            SkipNet.Parameters.clbTableEnabled = use_clb;
            SkipNet.Parameters.cTableEnabled = (use_clb && SkipNet.Parameters.cTableEnabled);
            SkipNet.Parameters.pTableUpdatePeriod=20*60*1000;  // Rebuild every 20 minutes
            SkipNet.Parameters.neighbourPingPeriod = 60*1000; // Ping every minute
            if (doRandom)
                SkipNet.Parameters.SetSeed(CommonParameters.seed);
        }

        public static void SetClbOrgName(string orgName)
        {
            appOrgName = orgName; // Save for later "application" code use
            Console.WriteLine("appOrgName = {0}", appOrgName);

            if (use_clb) 
            {
                name_prefix = name_prefix_dir = appOrgName + "!/";
            }
            else
            {
                name_prefix_dir = "/non-CLB_prefix";
                name_prefix = name_prefix_dir + "/";
            }
        }

        public static void InitNode(SkipNet.Node node, int nodeIndex, SendMessage s, LocalEvent l)
        {
            // Initialize Overlook.
            DirApplication dirApp = new DirApplication();
            dirApp.DirectoryInit(node, s, l, new OverlookTryoutDirAPI(node, s), App.use_clb);
            node.RegisterApp(dirApp, typeof(DirApplication));

            // Initialize ControlFlow
            ControlFlowApplication cfApp = new ControlFlowApplication(node, s, l, new ApplicationControlFlowUpcalls());
            node.RegisterApp(cfApp, typeof(ControlFlowApplication));

            if (nodeIndex == 0) App.node0cfApp = cfApp; // Save for use by application master control flow code
        }

        public static void UseNode(SkipNet.Node node, int nodeIndex)
        {
            int nodes_created = nodeIndex + 1;

            string keepAliveOrgName = appOrgName + "/";
            string nodeName = node.NodeIdAddressBind.NodeId.Name;
            if( keepAliveNode==-1 && nodeName.StartsWith(keepAliveOrgName) ) {
                keepAliveNode = nodeIndex;
            }

            if (nodes_created <= 200)
                Console.WriteLine("Created node {0} with address {1}", nodes_created - 1, node.GetPhysicalAddr);
        }

        public static void PrintStats() 
        {
            Console.WriteLine();
            Console.WriteLine("Times for operations:");
            int[] numOpns = DirApplication.NumOperations;
            long[] opnTimes = DirApplication.OperationTimes;
            long[] opnTimes2 = DirApplication.OperationTimesSquared;
            Array opns = DirApplication.DirOperations.GetValues(typeof(DirApplication.DirOperations));
            foreach (DirApplication.DirOperations op in opns) 
            {
                Console.WriteLine("{0,15}: n = {1,2}, av = {2}", op, numOpns[(int)op], (float)opnTimes[(int)op] / (float)numOpns[(int)op]);
            }
            Console.WriteLine("numCachedEntryHits = {0}", DirApplication.numCachedEntryHits);

            appEnv.envPrintStats();
        }

        public static void ExitProg(int exitCode) 
        {
            if (!quitImmediately) 
            {
                Console.WriteLine();
                Console.WriteLine("Hit Enter to exit program.");
                Console.ReadLine();
            } else if (exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                System.Threading.Thread.Sleep(2000); // Let messages get sent
            Environment.Exit(exitCode);
        }
    }

//================================================================================================

    /// <summary>
    /// Upcalls from ControlFlowApplication messages
    /// </summary>
    public class ApplicationControlFlowUpcalls : ControlFlowUpcalls
    {
        public void RendezvousWithNodeRequestUpcall(SkipNet.Node node, int nodeNumber, int attempt)
        {
            // Nothing to do but send a reply, which happens automatically
            if (ControlFlowApplication.DebugOutput)
            {
                Console.Write("RendezvousWithNode received on node {0} nodeNumber {1} attempt {2}",
                    node.GetNodeIdAddressBind.GetAddress, nodeNumber, attempt);
                if (App.exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                    Console.Write(" at time {0:T}", DateTime.Now);
                Console.WriteLine();
            }
        }

        public void RendezvousWithNodeReplyUpcall(ControlFlowApplication cfApp, bool timedOut, int nodeNumber, int attempt)
        {
            if (ControlFlowApplication.DebugOutput && App.exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                Console.WriteLine("RendezvousWithNodeReplyUpcall called for nodeNumber {0} attempt {1} at time {2:T}, timedOut = {3}",
                    nodeNumber, attempt, DateTime.Now, timedOut);
            if (timedOut) return;

            if (App.awaitingRendezvousWithNode < nodeNumber)
                App.awaitingRendezvousWithNode = nodeNumber;
        }

        public void OnNodeDoStepRequestUpcall(SkipNet.Node node, object replyToken, int nodeNumber, int programStep, int loopIndex)
        {
            Debug.Assert(App.replyToken == null); // If not null, then concurrent statement execution occurring
            App.replyToken = replyToken;
            if (ControlFlowApplication.DebugOutput && App.exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                Console.WriteLine("OnNodeDoStepRequestUpcall: replyToken = {0}, thread = {1} at time {2:T}",
                    replyToken is ControlFlowApplication.OnNodeDoStepRequest ?
                    ((ControlFlowApplication.OnNodeDoStepRequest) replyToken).programStep :
                    replyToken == null ? "null" : replyToken,
                    System.Threading.Thread.CurrentThread.Name, DateTime.Now);
            App.executeOnNode = nodeNumber;
            App.localNode = node;
            App.programStep = programStep;
            App.loopIndex = loopIndex;
            App.ApplicationStep();
        }

        public void OnNodeDoStepReplyUpcall(ControlFlowApplication cfApp, bool timedOut)
        {
            if (ControlFlowApplication.DebugOutput && App.exEnv.Type() == ExecutionEnvironmentType.ClusterEnvironment)
                Console.WriteLine("OnNodeDoStepReplyUpcall at time {0:T}, timedOut = {1}", DateTime.Now, timedOut);
            Debug.Assert(App.pendingOnNodeDoStep == true);
            App.pendingOnNodeDoStep = false;
        }

        public void FailNodeUpcall(SkipNet.Node node)
        {
            node.Failed = true;
            Console.WriteLine("================================ Failed node {0} ================================",
                node.GetPhysicalAddr);
            App.appEnv.envDone();
        }

        public void ExitNodeUpcall(SkipNet.Node node)
        {
            if (ControlFlowApplication.DebugOutput)
                Console.WriteLine("Received ExitNode upcall on node {0}", node.GetPhysicalAddr);
            App.appEnv.envDone();
        }
    }

    /// <summary>
    /// Interface to be supplied by both OverlookTryout_sim and OverlookTryout_cluster.
    /// </summary>
    public interface ApplicationEnvironment
    {
        void envPrintUsage();                   // Print environment-specific usage information
        void envPrintArgs();                    // Print environment-specific argument information
        void envPrintStats();                   // Print environment-specific stats
        bool envLogDetails {get; set;}          // Turn on/off logging environment-specific details
        void envSetParameterDefaults();         // Set environment-specific parameter defaults
        void envParameterFixup();               // Reconcile environment-specific parameter settings
        bool envRendezvousWithNode();           // Called to sequentially confirm a node's creation
        void envDone();                         // Called to tell the node that it should exit
    }
}
