//
// OverlookTryout_sim -- Exercise Overlook in the simulator
//

using System;
using System.Collections;
using System.Diagnostics;
using Application = MessageRouting.Application;
using SendMessage = MessageRouting.SendMessage;
using LocalEvent = MessageRouting.LocalEvent;
using Topology = MessageRouting.Topology;
using NodePhysicalAddress = MessageRouting.NodePhysicalAddress;
using SkipNet=Microsoft.Research.Herald.SkipNet;
using Overlook;
using Microsoft.Research.Herald.Common;
using Microsoft.Research.Herald.Logger;

using Simulator;
using Microsoft.Research.Herald.SimLib;

namespace OverlookTryout
{
	class Test: CheckIgnore, SimNodesClient, ApplicationEnvironment
	{
        static long start;

        static int transitTransitBand = 250000;
        static int transitStubBand = 5625;
        static int stubStubBand = 5625;

        static SimInterface s
        {
            get { return SimNodes.s; } // Allow use of just "s" instead of "SimNodes.s"
        }

//================================ CheckIgnore functions ================================

        public bool IgnoreThis(object msg) 
        {
            return ((MessageRouting.RouteMsg) msg).IsBackground();
        }

//================================ SimNodesClient functions ================================

        public void UseOrgs(Organization[] sortedOrgsList)
        {
            App.SetClbOrgName(sortedOrgsList[(sortedOrgsList.Length > 1) ? 1 : 0].Name());
        }

        public void InitNode(SkipNet.Node node, int nodeIndex)
        {
            App.InitNode(node, nodeIndex, s, s);
        }

        public void UseNode(SkipNet.Node node, int nodeIndex)
        {
            // Begin periodic tasks
            node.ActivateNode(true);
            s.ProcessUntilIgnoreEventsOnly();

            App.UseNode(node, nodeIndex);
        }

        public void Terminate(int exitCode)
        {
            App.ExitProg(exitCode);
        }

//================================ ApplicationEnvironment functions ================================

        /// <summary>
        /// Print environment-specific program usage information
        /// </summary>
        public void envPrintUsage()
        {
            SimLibParameters.PrintUsage();
        }

        /// <summary>
        /// Print environment-specific argument information
        /// </summary>
        public void envPrintArgs()
        {
            SimLibParameters.PrintArgs();
        }

        /// <summary>
        /// Turn on/off logging environment-specific details
        /// </summary>
        public bool envLogDetails 
        {
            get {return Organizations.DebugOutput;}
            set {Organizations.DebugOutput = value;}
        }

        /// <summary>
        /// Set environment-specific parameter defaults
        /// </summary>
        public void envSetParameterDefaults()
        {
            SimLibParameters.numOrgs = 10;
        }

        /// <summary>
        /// Reconcile environment-specific parameter settings
        /// </summary>
        public void envParameterFixup()
        {
            if (SimLibParameters.nodes < SimLibParameters.numOrgs) 
            {
                Console.WriteLine("WARNING: Setting numOrgs from {0} to {1}.", SimLibParameters.numOrgs, SimLibParameters.nodes);
                SimLibParameters.numOrgs = SimLibParameters.nodes;
            }
        }

        /// <summary>
        /// Print environment-specific stats
        /// </summary>
        public void envPrintStats() 
        {
            if (SimParameters.ModelLinkBandwidth)
            {
                Console.WriteLine("");
                NetTopology top = s.GetNetTopology;
                long timeNow = s.GetTime();
                NetTopology.LinkDelayStats[] ld = top.LinkDelays(timeNow);
                int len = ld.GetLength(0);
                Console.WriteLine("Link delays:");
                int numLinksUsed = 0;
                int numLinksWithDelays = 0;
                for (int i = 0; i < len; i++) 
                {
                    if (ld[i].numMessagesSent > 0)
                    {
                        numLinksUsed++;
                        Console.WriteLine("{0,3}({1,4})->{2,3}({3,4}): lat: {4,2}, bnd: {5,2}, nSent={6}, nDelay={7},  av={8,4}, max={9}", 
                            ld[i].sourceNodeIndex, 
                            ld[i].sourceLinkType, 
                            ld[i].destNodeIndex, 
                            ld[i].destLinkType, 
                            ld[i].latency,
                            ld[i].bandwidth,
                            ld[i].numMessagesSent, 
                            ld[i].numDelayedMessages,
                            ((float)ld[i].sumDelayTimes)/((float)ld[i].numMessagesSent), 
                            ld[i].maxDelayTime);
                        if (ld[i].numDelayedMessages > 0)
                            numLinksWithDelays++;
                    }
                }
                Console.WriteLine("numLinksUsed: {0}, numLinksWithDelays: {1}", numLinksUsed, numLinksWithDelays);
            }
        }

        static bool done_creating_nodes = false;

        // Create a single node, returning true if done creating nodes
        public bool envRendezvousWithNode()
        {
            bool done = SimNodes.CreateNode();
            if (done && ! done_creating_nodes) 
            {
                done_creating_nodes = true;

                long simseconds = s.GetTime() / 1000;
                Console.WriteLine("Simulated time: {0}:{1:D2}:{2:D2}",
                    simseconds / (60 * 60), (simseconds / 60) % 60, simseconds % 60);
            
                long seconds = (DateTime.Now.Ticks - start) / 10000000;
                Console.WriteLine("Elapsed time: {0}:{1:D2}", seconds / 60, seconds % 60);

                // Some overlays need to optimize their routing tables to work efficiently
                RunOptimizations();
            }

            return done;
        }

        // Called to tell the node that it should exit
        public void envDone()
        {
            // Nothing to do in the simulator environment
        }

        //================================ Environment-specific support functions ================================

        /// <summary>
        /// RunOptimizations: Some overlay types (SkipNet, Chord) need to optimize their
        /// routing tables before they can operate efficiently. We perform these optimizations here.
        /// </summary>
        public static void RunOptimizations() 
        {
            /*
            Console.WriteLine( "Beginning optimizations." );

            // The SkipNet needs to build its P table
            ActivateNodes(s);
            long delay = (long) ((SkipNet.Parameters.maxRebuilds+2.1) * SkipNet.Parameters.rebuildProximityIntervalsPeriod);
            s.ProcessForTime(delay);

            DeactivateNodes(s);
            //Console.WriteLine( "Accidental arrivals: {0}", SkipNet.FindEntries.accidentalArrivals );

            Console.WriteLine( "Finished optimizations.");
            long seconds = (DateTime.Now.Ticks - start) / 10000000;
            Console.WriteLine("Elapsed time: {0}:{1:D2} ", seconds / 60, seconds % 60);
            */
        }

        // Begin the proximity-join phase
        static public void ActivateNodes(SimInterface s) 
        {
            for (int i = 0; i < s.GetNodeArray.GetLength(0); i += 1) 
            {
                SkipNet.Node node = ((SkipNet.Node)(s.GetNodeArray[i]));
                if (null == node) return;
                node.ActivateNode(false);
            }
        }

        // End the proximity-join phase
        static public void DeactivateNodes(SimInterface s) 
        {
            for (int i = 0; i < s.GetNodeArray.GetLength(0); i += 1) 
            {
                SkipNet.Node node = ((SkipNet.Node)(s.GetNodeArray[i]));
                if (null == node) return;
                node.DeactivateNode();
            }
        }

//================================ Simulator environment main program ================================

		/// <summary>
		/// Main program for simulator version of OverlookTryout
		/// </summary>
		[STAThread]
        public static int Main(string[] args)
        {
            start = DateTime.Now.Ticks;
            long seconds, simseconds;

            Test test = new Test();
            App.appEnv = test;
            App.exEnv = new SimExecutionEnvironment();

            App.SetParameterDefaults();

            // Process the command line options
            int resume_at = 0;
            resume_at = SimLibParameters.ParseArgs(args, resume_at);
            if (resume_at < 0) App.Usage();
            resume_at = CommonParameters.ParseArgs(args, resume_at);
            if (resume_at < 0) App.Usage();
            App.GetOptions(args, resume_at);

            App.PostProcessParameters();

            NetTopology.SetRandomSeed(CommonParameters.seed);
            ZipfSelect.SetRandomSeed(CommonParameters.seed);
            OverlookTryout.App.SetRandomSeed(CommonParameters.seed);

            SimLibParameters.SetParameters();
            CommonParameters.SetParameters();

            App.PrintExperimentParameters();

            App.SetOverlayParameters(true);

            seconds = (DateTime.Now.Ticks - start) / 10000000;
            Console.WriteLine("Elapsed time: {0}:{1:D2}", seconds / 60, seconds % 60);

            Organizations orgs = SimNodes.InitOrgs(test, test, SimLibParameters.numOrgs, transitTransitBand, transitStubBand, stubStubBand);
            Log.EnableTimeStamping(Test.s);
            App.ApplicationControlFlow();
        
            App.PrintStats();

            Console.WriteLine();

            simseconds = s.GetTime() / 1000;
            Console.WriteLine("Simulated time: {0}:{1:D2}:{2:D2}",
                simseconds / (60 * 60), (simseconds / 60) % 60, simseconds % 60);
            seconds = (DateTime.Now.Ticks - start) / 10000000;
            Console.WriteLine("Elapsed time: {0}:{1:D2} ({2} seconds)", seconds / 60, seconds % 60, seconds);

            Console.WriteLine();
            if (App.errorCount == 0)
                Console.WriteLine("All OverlookTryout tests passed.");
            else
                Console.WriteLine("{0} OverlookTryout tests failed.", App.errorCount);
            App.ExitProg(0);
            return 0;
        }
	}
}
