//
// RPCTest_Cluster -- Exercise RPCTest on the cluster
//

using System;
using System.Collections;
using System.Diagnostics;
using System.Threading;
using Application = MessageRouting.Application;
using SendMessage = MessageRouting.SendMessage;
using LocalEvent = MessageRouting.LocalEvent;
using Topology = MessageRouting.Topology;
using NodePhysicalAddress = MessageRouting.NodePhysicalAddress;
using SkipNet=Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Common;
using Microsoft.Research.Herald.Common.ControlFlow;
using Microsoft.Research.Herald.Logger;
using Microsoft.Research.Herald.TLog;

using Microsoft.Research.Taro;
using Microsoft.Research.Herald.ClusterLib;

namespace RPCTest {
    class Test: ApplicationEnvironment, HeraldUI {
        static long start;
        static ClusterNode clusterNode;
        static Thread overlayThread;
        static volatile bool done = false;

        //================================ ApplicationEnvironment functions ================================

        /// <summary>
        /// Print environment-specific program usage information
        /// </summary>
        public void envPrintUsage() {
            ClusterLibParameters.PrintUsage();
        }

        /// <summary>
        /// Print environment-specific argument information
        /// </summary>
        public void envPrintArgs() {
            ClusterLibParameters.PrintArgs();
        }

        static bool myLogDetails = false; // XXX Unused, but needed to satisfy interface

        /// <summary>
        /// Turn on/off logging environment-specific details
        /// </summary>
        public bool envLogDetails {
            get {return myLogDetails;}
            set {myLogDetails = value;}
        }

        /// <summary>
        /// Set environment-specific parameter defaults
        /// </summary>
        public void envSetParameterDefaults() {
            // Pings
            SkipNet.Parameters.neighbourPingPeriod = 60*1000;                   // 60s
            SkipNet.Parameters.aggressiveFixups = true;

            // P-Table
            SkipNet.Parameters.pTableUpdatePeriod = 1*60*1000;     // 1 minute
            SkipNet.Parameters.maxRebuilds = 0;                                 // Keep rebuilding P-Table

            Log.SetCategoryFilterLevel(SkipNet.JoinRequest.CATEGORY, LogLvl.Lo);
            Log.SetCategoryFilterLevel(TaroLocalEvent.CATEGORY, LogLvl.Med);
            // Log.SetCategoryFilterLevel(TaroSocketSendMessage.CATEGORY, LogLvl.Med);
            // Log.SetCategoryFilterLevel(TaroSocketReceiveMessage.CATEGORY, LogLvl.Med);
            // TaroSocketConfig.Debug = true;
        }

        /// <summary>
        /// Reconcile environment-specific parameter settings
        /// </summary>
        public void envParameterFixup() {
        }

        /// <summary>
        /// Print environment-specific stats
        /// </summary>
        public void envPrintStats() {
        }

        /// <summary>
        /// Reset environment-specific state, such as counters, etc.
        /// </summary>
        public void envReset() {
            // Nothing to do.
        }

        /// <summary>
        /// Return local time.
        /// </summary>
        public long envGetTime() {
            return clusterNode.localEvent.GetTime();
        }

        // Called to tell the node that it should exit
        public void envExit() {
            done = true;
        }

        //================================ HeraldUI functions ================================

        public void Error( string format, params object[] arg ) {
            Console.WriteLine( format, arg );
        }

        public void Message( string format, params object[] arg ) {
            Console.WriteLine( format, arg );
        }

        public void MessageNoNewLine( string format, params object[] arg ) {
            Console.Write( format, arg );
        }

        public void HeraldMessage( string topic, string from, string message ) {
            Console.WriteLine( "Message to topic {0} from {1}:", topic, from );
            Console.WriteLine( message );
        }

        //================================ Cluster environment main program ================================

        /// <summary>
        /// Main program for cluster version of RPCTest
        /// </summary>
        [STAThread]
        public static int Main(string[] args) {
            start = DateTime.Now.Ticks;

            Test test = new Test();
            App.appEnv = test;
            clusterNode = new ClusterNode(test);
            App.exEnv = new ClusterExecutionEnvironment(clusterNode);

            App.SetParameterDefaults();

            // Process the command line options
            int resume_at = 0;
            resume_at = ClusterLibParameters.ParseArgs(args, resume_at);
            if (resume_at < 0) App.Usage();
            resume_at = CommonParameters.ParseArgs(args, resume_at);
            if (resume_at < 0) App.Usage();
            App.GetOptions(args, resume_at);

            App.PostProcessParameters();

            App.SetRandomSeed(CommonParameters.seed);

            ClusterLibParameters.SetParameters();
            CommonParameters.SetParameters();

            App.PrintExperimentParameters();
            App.SetOverlayParameters(true);

            // Concatenate ControlFlow and RPCTest serialization types to add to SkipNet & built-in types for XML Serializer
            Type [] extraTypes, types1, types2, types3;
            types1 = SimpleControlFlow_SerializationConfig.serializationTypes;
            types2 = RPCTest.SerializationConfig.serializationTypes;
            types3 = TLog.SerializationConfig.serializationTypes;

            extraTypes = new Type [types1.Length + types2.Length + types3.Length];
            Array.Copy(types1, 0, extraTypes, 0, types1.Length);
            Array.Copy(types2, 0, extraTypes, types1.Length, types2.Length);
            Array.Copy(types3, 0, extraTypes, types1.Length+types2.Length, types3.Length);

            clusterNode.InitNode(extraTypes);
            App.InitNode(clusterNode.node, ClusterLibParameters.nodeNumber, clusterNode.sendMessage, clusterNode.localEvent);
            Log.EnableTimeStamping(clusterNode.localEvent);
            Log.EnableLoadMonitoring();
            TLog.Init(clusterNode.localEvent);

            App.UseNode(clusterNode.node, ClusterLibParameters.nodeNumber);

            // Run the Overlay event loop in its own thread
            Console.WriteLine("Finished Overlay Initialization.  Beginning event loop.");
            overlayThread = new Thread( new ThreadStart(OverlayMain) );
            overlayThread.Name = "SkipNet Thread";
            overlayThread.Start();

            const int POLL_PERIOD_IN_SECONDS = 3; // See if done every 3 seconds
            const int TIME_LIMIT_IN_MINUTES = 6 * 60; // Run for at most 6 hours

            if (ClusterLibParameters.nodeNumber == 0) {
                App.Application(); // Run the experiment control flow only on node 0
            }
            else {
                for (int i = 0; i < TIME_LIMIT_IN_MINUTES * (60 / POLL_PERIOD_IN_SECONDS); i++) {
                    if (done) break;
                    Thread.Sleep(POLL_PERIOD_IN_SECONDS * 1000);
                }

                if (! done)
                    Console.WriteLine("======== Application ran out of time -- {0} minutes exceeded -- exiting! ========",
                        TIME_LIMIT_IN_MINUTES);
            }

            App.PrintStats();

            Console.WriteLine("Time: " + ((DateTime.Now.Ticks - start)/10000000) + " seconds.");

            Console.WriteLine();
            Console.WriteLine("All RPCTest tests passed.");
            App.ExitProg(0);
            return 0;
        }

        // The main procedure of the Overlay thread
        private static void OverlayMain() {
            try {
                clusterNode.localEvent.EventLoop();
            } 
            catch( Exception e ) {
                Console.WriteLine(e.ToString());
                clusterNode.receiveMessage.ShutDown();
                if (e.Message.Equals("R-Table Join Failed!"))
                {
                    Console.WriteLine("Either the seed node address {0} is incorrect or else the seed node isn't running or reachable", clusterNode.overlaySeed);
                }
                Environment.Exit(-1);
            }
        }
    }
}
