//
// RPCTest_sim -- RPC test application, simulator version
//

using System;
using System.Collections;
using System.Diagnostics;
using Application = MessageRouting.Application;
using SendMessage = MessageRouting.SendMessage;
using LocalEvent = MessageRouting.LocalEvent;
using Topology = MessageRouting.Topology;
using NodePhysicalAddress = MessageRouting.NodePhysicalAddress;
using SkipNet=Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Common;
using Microsoft.Research.Herald.Logger;
using Microsoft.Research.Herald.TLog;

using Simulator;
using Microsoft.Research.Herald.SimLib;

namespace RPCTest {
    class Test: CheckIgnore, SimNodesClient, ApplicationEnvironment {
        static long start;

        static int transitTransitBand = 250000;
        static int transitStubBand = 5625;
        static int stubStubBand = 5625;

        static SimInterface s {
            get { return SimNodes.s; } // Allow use of just "s" instead of "SimNodes.s"
        }

        //================================ CheckIgnore functions ================================

        public bool IgnoreThis(object msg) {
            return ((MessageRouting.RouteMsg) msg).IsBackground();
        }

        //================================ SimNodesClient functions ================================

        public void UseOrgs(Organization[] sortedOrgsList) {
            // No special processing after creating the orgs
        }

        public void InitNode(SkipNet.Node node, int nodeIndex) {
            App.InitNode(node, nodeIndex, s, s);
        }

        public void UseNode(SkipNet.Node node, int nodeIndex) {
            App.UseNode(node, nodeIndex);
            s.ProcessUntilIgnoreEventsOnly();
        }

        public void Terminate(int exitCode) {
            App.ExitProg(exitCode);
        }

        //================================ ApplicationEnvironment functions ================================

        /// <summary>
        /// Print environment-specific program usage information
        /// </summary>
        public void envPrintUsage() {
            SimLibParameters.PrintUsage();
        }

        /// <summary>
        /// Print environment-specific argument information
        /// </summary>
        public void envPrintArgs() {
            SimLibParameters.PrintArgs();
        }

        /// <summary>
        /// Turn on/off logging environment-specific details
        /// </summary>
        public bool envLogDetails {
            get {return Organizations.DebugOutput;}
            set {Organizations.DebugOutput = value;}
        }

        /// <summary>
        /// Set environment-specific parameter defaults
        /// </summary>
        public void envSetParameterDefaults() {
            SimLibParameters.numOrgs = 10;
        }

        /// <summary>
        /// Reconcile environment-specific parameter settings
        /// </summary>
        public void envParameterFixup() {
            if (SimLibParameters.nodes < SimLibParameters.numOrgs) {
                Console.WriteLine("WARNING: Setting numOrgs from {0} to {1}.", SimLibParameters.numOrgs, SimLibParameters.nodes);
                SimLibParameters.numOrgs = SimLibParameters.nodes;
            }
        }

        /// <summary>
        /// Print environment-specific stats
        /// </summary>
        public void envPrintStats() {
            // Nothing to print.
        }

        /// <summary>
        /// Reset environment-specific state, such as counters, etc.
        /// </summary>
        public void envReset() {
            s.GetNetTopology.ResetStats();
        }

        /// <summary>
        /// Return local time.
        /// </summary>
        public long envGetTime() {
            return s.GetTime();
        }

        /// <summary>
        /// Exit simulated node
        /// </summary>
        public void envExit() {
            // Noop in the simulator.
        }

        //================================ Environment-specific support functions ================================

        //================================ Simulator environment main program ================================

        /// <summary>
        /// Main program for simulator version of RPCTest
        /// </summary>
        [STAThread]
        public static int Main(string[] args) {
            start = DateTime.Now.Ticks;

            Test test = new Test();
            App.appEnv = test;
            App.exEnv = new SimExecutionEnvironment();

            App.SetParameterDefaults();

            // Process the command line options
            int resume_at = 0;
            resume_at = SimLibParameters.ParseArgs(args, resume_at);
            if (resume_at < 0) App.Usage();
            resume_at = CommonParameters.ParseArgs(args, resume_at);
            if (resume_at < 0) App.Usage();
            App.GetOptions(args, resume_at);

            App.PostProcessParameters();

            NetTopology.SetRandomSeed(CommonParameters.seed);
            ZipfSelect.SetRandomSeed(CommonParameters.seed);
            App.SetRandomSeed(CommonParameters.seed);

            SimLibParameters.SetParameters();
            CommonParameters.SetParameters();

            App.PrintExperimentParameters();

            App.SetOverlayParameters(true);

            long seconds = (DateTime.Now.Ticks - start) / 10000000;
            Console.WriteLine("Elapsed time: {0}:{1:D2}", seconds / 60, seconds % 60);
            Organizations orgs = SimNodes.InitNodes(test, test, SimLibParameters.numOrgs, transitTransitBand, transitStubBand, stubStubBand);
            seconds = (DateTime.Now.Ticks - start) / 10000000;
            Console.WriteLine("Elapsed time: {0}:{1:D2}", seconds / 60, seconds % 60);

            Log.EnableTimeStamping(Test.s);
            TLog.Init(Test.s);
            TLog.InitGlobalLoggingNode(App.myServer, App.exEnv);

            App.Application();
        
            App.PrintStats();

            Console.WriteLine("Time: " + ((DateTime.Now.Ticks - start)/10000000) + " seconds.");

            Console.WriteLine();
            Console.WriteLine("All RPCTest tests passed.");
            App.ExitProg(0);
            return 0;
        }
    }
}
