//
// Module to manage organizations.
//

using System;
using System.Collections;
using System.Diagnostics;
using System.IO;

namespace Microsoft.Research.Herald.SimLib
{
    //
    // Describes a single organization
    //
    public class Organization: IComparable
    {
        static Random random = null; // Used to generate random member names.
        public const string NamesDirectory = @"C:\SkipNet\Names";
        const string HostNamesFile = NamesDirectory + @"\HostNames.txt";
        static ArrayList hostNames = null;
        static int hostNamesIndex = 0;

        static Hashtable usedHostNames = new Hashtable();

        Organizations orgs;  // Pointer to an object that keeps track of all organizations
        int numMembers;  // Number of members this org has.
        String name;  // Name of the org.
        ArrayList members = new ArrayList();  // List of node records for members of this organization.
        Simulator.ZipfSelect zs = null;

        // If >= 0 then this is the index of a core router node in the network topology.
        // Useful for figuring out which GT stub or Mercator AS this organization is situated in.
        // If == -1 then the organization is not clustered within a single GT stub or Mercator AS.
        int topCoreNode = -1;

        // First node added to this organization.
        int firstNode = -1;


        private Organization(){}  // Don't allow this to be used.

        public Organization(int numMembers, String name, Organizations orgs)
        {
            this.orgs = orgs;
            this.numMembers = numMembers;
            this.name = name;
        }

        public static void SetRandomSeed(int seed) 
        {
            random = new Random(seed);
        }


        public int CompareTo(object x) 
        {
            return ((Organization)x).numMembers - this.numMembers;
        }


        public int NumberOfMembers() 
        {
            return this.numMembers;
        }


        public String Name() 
        {
            return this.name;
        }


        public void AddMemberNodeIndex(int nodeIndex)
        {
            if (members.Count == 0) this.firstNode = nodeIndex;
            members.Add(nodeIndex);
            orgs.AddMemberNodeIndex(nodeIndex, this);
        }

        public int GetMemberNodeIndex(int nodeIndex)
        {
            return (int) members[nodeIndex];
        }

        /// <summary>
        /// Pick a random member of the organization and return its members index.
        /// </summary>
        /// <returns></returns>
        public int PickRandomMember() 
        {
            return PickRandomMember(false);
        }

        /// <summary>
        /// Pick a random member of the organization and return its members index.
        /// If 'zipfSelect' then Zipf-select from the members array, else pick uniformly.
        /// </summary>
        /// <param name="zipfSelect"></param>
        /// <returns></returns>
        public int PickRandomMember(bool zipfSelect) 
        {
            if (this.zs == null) this.zs = new Simulator.ZipfSelect(this.numMembers);
            return (int) members[this.zs.Pick(zipfSelect)];
        }


        /// <summary>
        /// Randomly pick an unused name.
        /// </summary>
        /// <returns></returns>
        public String NewMemberName()
        {
            string hostName;
            while (true) 
            {
                if (hostNames == null) 
                {
                    Debug.Assert(random != null);
                    // Use a randomly generated member name.
                    hostName = random.Next().ToString();
                }
                else 
                {
                    // Use a real host name.
                    hostName = (string) hostNames[hostNamesIndex++];
                    Debug.Assert(hostNamesIndex <= hostNames.Count);
                }
                if (!usedHostNames.Contains(hostName)) 
                {
                    usedHostNames[hostName] = "1";
                    break;
                }
            }
            return this.name + "/" + hostName;
        }

        public int TopologyCoreNode 
        {
            get {return this.topCoreNode;}
            set {this.topCoreNode = value;}
        }

        public int FirstMemberNode 
        {
            get {return this.firstNode;}
        }

        public static void CheckNamesFile(string filename)
        {
            if (! File.Exists(filename)) 
            {
                Console.WriteLine();
                Console.WriteLine("File {0} does not exist.", filename);
                Console.WriteLine("Did you remember to copy the release's Names directory to");
                Console.WriteLine("{0} or create your own version of this file?", NamesDirectory);
                Environment.Exit(1);
            }
        }

        public static void InitializeRealMachineNames()
        {
            hostNames = new ArrayList();

            // Read in the file of host names to use for picking new 
            // member names for the organization's machines.
            CheckNamesFile(HostNamesFile);
            StreamReader sr = new StreamReader(HostNamesFile);
            while (true)
            {
                string hostName = sr.ReadLine();
                if (hostName == null) break;
                hostNames.Add(hostName);
            }
        }

        public static void ReleastHostNamesData()
        {
            hostNames = null;
            usedHostNames = null;
        }
    }


    //
    // Describes a set of organizations that obey a distribution of member sizes
    //
    public class Organizations
    {
        public static bool DebugOutput = false;

        Random random = null; // Used to generate random member names.
        const string OrgNamesFile = Organization.NamesDirectory + @"\OrgNames.txt";
        static ArrayList orgNames = null;

        static double cNormal = 0.0;
        static double[] weights;

        public int numOrgs;  // Number of organizations to represent.
        int numTotalNodes;  // Total number of nodes that these organizations should contain.

        int numAllocedNodes = 0;  // Internal variable used to track how many nodes we've already allocated.

        ArrayList orgsList = new ArrayList();  // List of organizations we've already created.
        Hashtable allMembers = new Hashtable();  // Table of all members of all organizations.

        public Organizations(int numOrgs, int numTotalNodes, int randomSeed) 
        {
            Organization.SetRandomSeed(randomSeed);

            this.random = new Random(randomSeed);
            this.numOrgs = numOrgs;
            this.numTotalNodes = numTotalNodes;
        }


        public string RandomOrgName()
        {
            if (orgNames == null) return random.Next().ToString();
            else
            {
                int indx = random.Next(orgNames.Count);
                return (string) orgNames[indx];
            }
        }


        //
        // Return the 'indx'-th organization.
        //
        public Organization CreateOrg(int indx) 
        {
            if (cNormal == 0.0) 
            {
                weights = new double[this.numOrgs];
                for (int i = 0; i < this.numOrgs; i++) 
                {
                    weights[i] = Math.Pow((double) (i+1),-1.25) + 0.5;  // Adding in the 0.5 makes the slope of the Zipf much flatter.
                    cNormal += weights[i];
                }
                cNormal = ((double)this.numTotalNodes) / cNormal;
            }

            int numMembers = (int) (weights[indx] * cNormal);
            if (random.Next(2) == 1) numMembers++;  // Adjust for integer round-down.
            if (numMembers < 1) 
            {
                numMembers = 1;
            }

            int numOrgsToGo = this.numOrgs - indx;
            int numNodesStillLeft = this.numTotalNodes - this.numAllocedNodes;
            if (indx == (this.numOrgs - 1)) 
            {
                numMembers = numNodesStillLeft;
            } 
            else 
            {  // Make sure we don't run out of nodes before we run out of orgs.
                while ((numOrgsToGo - 1) >= (numNodesStillLeft - numMembers)) 
                {
                    numMembers /= 2;
                }
                if ( numMembers<1 ) numMembers=1;
            }
            this.numAllocedNodes += numMembers;
            string orgName = RandomOrgName();
            Organization org = new Organization(numMembers, orgName, this);
            orgsList.Add(org);
            Debug.Assert(numMembers > 0, "Created an org with 0 members");
            if (DebugOutput)
                Console.WriteLine("New org: {0} with {1} members;  indx: {2}, numAlloced: {3}", org.Name(), org.NumberOfMembers(), indx, this.numAllocedNodes);
            return org;
        }


        /// <summary>
        /// Returns an array of Organization records, sorted by size of the organization
        /// (from largest to smallest).
        /// </summary>
        /// <returns></returns>
        public Organization[] GetListOfAllOrgs() 
        {
            Organization[] allOrgs = new Organization[orgsList.Count];
            orgsList.CopyTo(allOrgs);
            Array.Sort(allOrgs);
            return allOrgs;
        }


        public void AddMemberNodeIndex(int nodeIndex, Organization org)
        {
            allMembers[nodeIndex] = org;
        }

        public Organization GetOrgByMemberNodeIndex(int nodeIndex)
        {
            return (Organization) allMembers[nodeIndex];
        }


        public static void InitializeRealOrgNames()
        {
            orgNames = new ArrayList();

            // Read in the file of host names to use for picking new 
            // member names for the organization's machines.
            Organization.CheckNamesFile(OrgNamesFile);
            StreamReader sr = new StreamReader(OrgNamesFile);
            while (true)
            {
                string orgName = sr.ReadLine();
                if (orgName == null) break;
                orgNames.Add(orgName);
            }
        }

        public static void ReleastOrgNamesData()
        {
            orgNames = null;
        }
    }
}
