//
// Create simulator organizations and nodes
//

using System;
using System.Collections;
using System.Diagnostics;
using System.IO;

using Simulator;
using SkipNet=Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Common;

namespace Microsoft.Research.Herald.SimLib
{
    public class SimNodes
    {
        public static SimInterface s = null;
        static int nodes_created = 0;

        private static Organization org;
        private static Organization orgToClusterWith;
        private static bool isolatedOrg;

        private static SimNodePhysicalAddress firstAddr;
        private static int orgNodeCount;
        private static NetTopology top;
        private static int[] coreNodesToPickFrom;
        private static ZipfSelect zs;

        // CreateNodes
        // Given an organization, determine how many nodes should be in the organization and create them
        // If 'isolatedOrg' is true, this org will create its own disjoint SkipNet, isolated from the rest
        // of the SkipNet.

        static void PrepareNewOrg()
        {
            // These were formerly parameters to the CreateOrgNodes function
            orgToClusterWith = null;
            org = sortedOrgsList[orgNum];
            isolatedOrg = false;

            firstAddr = null;
            top = s.GetNetTopology;

            orgNodeCount = org.NumberOfMembers();
            int coreNode = -1;
            if (orgToClusterWith != null) coreNode = orgToClusterWith.TopologyCoreNode;
            coreNodesToPickFrom = top.CoreNodesForOrg(org.Name(), orgNodeCount, SimLibParameters.clusterByOrg, coreNode);
            if (SimLibParameters.clusterByOrg) org.TopologyCoreNode = coreNodesToPickFrom[0];
            zs = new ZipfSelect(coreNodesToPickFrom.Length);
            if (!SimLibParameters.clusterOrgsByZipf)zs.SetDefaultToUniform();
        }

        private static int orgNode;

        private static void CreateOrgNode() 
        {
            if (nodes_created % 1000 == 0) Console.WriteLine("Creating node {0}", nodes_created);
            SkipNet.Node node = new SkipNet.Node(org.NewMemberName());

            SimNodePhysicalAddress addr = top.NewPhysicalNode(coreNodesToPickFrom, zs);

            MessageRouting.NodePhysicalAddress seedNeighbor;
            if (nodes_created == 0)
                seedNeighbor = null;
            else
            {
                do 
                {
                    seedNeighbor = top.GetRandom(addr);
                } while (seedNeighbor.Failed);
            }

            // If 'isolatedOrg' is true, we deliberately do not give the first node a seed
            if (isolatedOrg)
            {
                if (orgNode == 0) 
                {
                    firstAddr = addr;
                    seedNeighbor = null;
                } 
                else
                {
                    seedNeighbor = firstAddr;
                }
            }
            s.CreateNode(node, addr, seedNeighbor);
            client.InitNode(node, nodes_created);
            org.AddMemberNodeIndex(nodes_created);  // This has to go after OverlayInit so that the node's internal state is fully instantiated.

            // Complete node join
            s.ProcessUntilIgnoreEventsOnly();

            client.UseNode(node, nodes_created);

            nodes_created++;
        }

        public static Organizations InitOrgs(SimNodesClient simNodesClient, CheckIgnore checkIgnore, int nOrgs,
            int transitTransitBand, int transitStubBand, int stubStubBand) 
        {
            numOrgs = nOrgs;
            client = simNodesClient;

            Console.WriteLine("Initialising Network Topology");
            bool useMerc = false;

            // for gt-itm
            string netFile = null;
            string routesFile = null;
            string distancesFile = null;

            // for mercator
            string asFile = null;
            string asRoutesFile = null;
            string routerRoutesFile = null;
            string routerFile = null;
            string gatewayFile = null;

            if (SimLibParameters.topology == "600") 
            {
                netFile = NetTopology.GraphsDirectory + "\\ts600-0.sc";
                routesFile = NetTopology.GraphsDirectory + "\\ts600-0-routesA.out";
                distancesFile = NetTopology.GraphsDirectory + "\\ts600-0-distancesA.out";
            } 
            else if (SimLibParameters.topology == "5050") 
            {
                netFile = NetTopology.GraphsDirectory + "\\ts5050-0.sc";
                routesFile = NetTopology.GraphsDirectory + "\\ts5050-0-routes.out";
                distancesFile = null;
            }
            else if (SimLibParameters.topology == "100") 
            {
                netFile = NetTopology.GraphsDirectory + "\\ts100-0.sc";
                routesFile = NetTopology.GraphsDirectory + "\\ts100-0-routes.out";
                distancesFile = null;
            } 
            else if (SimLibParameters.topology == "merc") 
            {
                if (SimLibParameters.extended_simulator_model) 
                {
                    Console.WriteLine("Extended Simulator doesn't support mercator topology");
                    client.Terminate(-1);
                }
                asFile = NetTopology.GraphsDirectory + "\\merc-as.sc";
                asRoutesFile = NetTopology.GraphsDirectory + "\\merc-as-routes.out";
                routerRoutesFile = NetTopology.GraphsDirectory + "\\new-merc-router-routes.out";
                routerFile = NetTopology.GraphsDirectory + "\\new-merc-routers.sc";
                gatewayFile = NetTopology.GraphsDirectory + "\\new-merc-inter-AS.out";
                useMerc = true;
            }
            else 
            {
                Console.WriteLine("Unrecognized topology \"{0}\" -- use \"100\" or \"600\" or \"5050\" or \"merc\"", SimLibParameters.topology);
                client.Terminate(-1);
            }
            if (SimLibParameters.extended_simulator_model) 
            {
                SimParameters.ModelLinkBandwidth = true;
                s = new Sim3(netFile, routesFile, distancesFile, checkIgnore,
                    transitTransitBand, transitStubBand, stubStubBand);
            }
            else 
            {
                if (useMerc) 
                {
                    s = new Sim(asFile,asRoutesFile, routerRoutesFile, routerFile, gatewayFile, checkIgnore);
                } 
                else 
                {
                    s = new Sim(netFile, routesFile, distancesFile, checkIgnore);
                }
            }

            Console.WriteLine("Memory usage after creation of simulator: " + (System.GC.GetTotalMemory(true)/(1024*1024)) + "MB");

            Console.WriteLine("Starting (" + SimLibParameters.nodes + "," + s.GetNetTopology.MaxLogicalNodes() + ")");

            // Create application nodes.
            if (SimLibParameters.useRealMachineNamesForNodes) Organization.InitializeRealMachineNames();
            if (SimLibParameters.useRealOrgNames) Organizations.InitializeRealOrgNames();
            orgs = new Organizations(numOrgs, SimLibParameters.nodes, CommonParameters.seed);
            for (int i = 0; i < numOrgs; i++) 
            {
                orgs.CreateOrg(i);
            }
            sortedOrgsList = orgs.GetListOfAllOrgs();
            client.UseOrgs(sortedOrgsList);

            // Initialize loop variables for "while (! done)" loop
            done = false;
            orgNum = -1; // Must be initialized to -1 since incremented in first call to CreateNode
            orgNode = orgNodeCount = 0;

            return orgs;
        }

        private static int orgNum;
        private static bool done;

        public static bool CreateNode()
        {
            if (done) return true;

            if (orgNode < orgNodeCount) // If nodes left within current organization
            {
                CreateOrgNode();
                orgNode++;
                goto check_done;
            }

            orgNum++;

            while (orgNum < numOrgs) // If organizations remaining
            {
                PrepareNewOrg();
                orgNode = 0;
                    
                if (orgNode < orgNodeCount) // If nodes left within current organization
                {
                    CreateOrgNode();
                    orgNode++;
                    goto check_done;
                }
                orgNum++;
            }

        check_done:
            if (orgNum >= numOrgs || (orgNum + 1 == numOrgs && orgNode >= orgNodeCount))
            {
                done = true;
                Cleanup();
            }

            return done;
        }

        private static void Cleanup()
        {
            // Free up memory by getting rid of no-longer-needed data structures.
            sortedOrgsList = null;
            s.GetNetTopology.FreeNodeCreationDataStructures();
            if (SimLibParameters.useRealMachineNamesForNodes) Organization.ReleastHostNamesData();
            if (SimLibParameters.useRealOrgNames) Organizations.ReleastOrgNamesData();

            top = null;
            coreNodesToPickFrom = null;
            zs = null;

            Console.WriteLine();
            Console.WriteLine("Memory usage after CreateNodes(): " + (System.GC.GetTotalMemory(true)/(1024*1024)) + "MB");
        }

        public static Organizations InitNodes(SimNodesClient simNodesClient, CheckIgnore checkIgnore, int nOrgs,
            int transitTransitBand, int transitStubBand, int stubStubBand)
        {
            Organizations orgs = InitOrgs(simNodesClient, checkIgnore, nOrgs, transitTransitBand, transitStubBand, stubStubBand);
            
            while (CreateNode() == false) {}

            return orgs;
        }

        private static Organizations orgs;
        private static int numOrgs;
        private static Organization[] sortedOrgsList;
        private static SimNodesClient client;
    }

    public interface SimNodesClient 
    {
        void UseOrgs(Organization[] sortedOrgsList);
        void InitNode(SkipNet.Node node, int nodeIndex);
        void UseNode(SkipNet.Node node, int nodeIndex);

        void Terminate(int exitcode);
    }
}
