using System;
using System.Collections;
using System.Diagnostics;

using SplayTree = Microsoft.Research.Herald.SplayTree;

namespace Simulator {

    /// <summary>
    /// An entry in the queue.
    /// </summary>
    class QueueEntry : IComparable {
        public long    point;
        public Object  obj;

        public QueueEntry(long point, Object obj) {
            this.point = point;
            this.obj = obj;
        }

        public int CompareTo(Object o) {
            QueueEntry e = (QueueEntry) o;
            return point.CompareTo(e.point);
        }

        public override bool Equals(Object o) {
            QueueEntry e = (QueueEntry) o;
            return obj.Equals(e.obj);
        }
    }

    /// <summary>
    /// A queue of messages, implemented using a Splay Tree and a Hashtable
    /// </summary>
    public class EventQueue {
        SplayTree   tree = new SplayTree();
        Hashtable   table = new Hashtable();
        long        timeNow = 0;

        // Add an entry to the queue
        public void InsertObject(long point, Object obj) {
            QueueEntry entry = new QueueEntry(point, obj);

            // Insert an entry in the splay tree for this point in time
            ArrayList entries = (ArrayList) tree.Search(point);
            if( entries!=null ) {
                entries.Add(entry);
            } else {
                entries = new ArrayList();
                entries.Add(entry);
                tree.Insert(point, entries);
            }

            // Insert this entry into the hashtable
            Debug.Assert( ! table.ContainsKey(obj) );
            table[obj] = entry;
        }

        // Get the current time of the queue
        public long GetTime {
            get { return timeNow; }
        }

        // Get an event from the queue that will fire next
        public Object GetNextEvent() {
            // Get the smallest key in the tree
            IComparable min = tree.GetMinKey();

            // If the tree is empty, return null
            if( min==null ) {
                return null;
            }

            // Get the value in the tree with the smallest key
            ArrayList entries = (ArrayList) tree.Search(min);
            Debug.Assert( entries!=null && entries.Count>0 );

            // Remove an entry from the list
            QueueEntry entry = (QueueEntry) entries[ entries.Count-1 ];
            entries.RemoveAt( entries.Count-1 );

            // The list of entries is empty. Remove this list from the tree.
            if( entries.Count==0 ) {
                tree.Delete( min );
            }

            // Remove this entry from the hashtable
            Debug.Assert( table.ContainsKey(entry.obj) );
            table.Remove(entry.obj);
            Debug.Assert( ! table.ContainsKey(entry.obj) );

            timeNow = (long) min;
            return entry.obj;
        }

        // Remove this object from the event queue
        public object Remove( Object obj ) {
            Debug.Assert( table.ContainsKey(obj) );
            QueueEntry entry = (QueueEntry) table[obj];

            // Get the entries in the tree with the matching key
            ArrayList entries = (ArrayList) tree.Search(entry.point);
            Debug.Assert( entries!=null && entries.Count>0 );

            // Remove the object from the list of entries
            int countBefore = entries.Count;
            entries.Remove(entry);
            Debug.Assert( entries.Count == countBefore-1 );

            // The list of entries is empty. Remove this list from the tree.
            if( entries.Count==0 ) {
                tree.Delete(entry.point);
            }

            // Remove the object from the table
            table.Remove(obj);
            Debug.Assert( ! table.ContainsKey(obj) );
            return entry.obj;
        }

    }
}