#define PING_STATS // Support for gathering Distance statisitics
#define NO_TRANSIT_LANS // To prevent LAN nodes from being assigned to transit nodes

namespace Simulator
{
    using System;
    using MessageRouting;
    using System.Collections;
    using System.IO;
    using System.Diagnostics;
    using System.Runtime.Serialization.Formatters.Binary;

    /// <summary>
    /// Simulates a network topology. The topology has a core and a set of 
    /// LAN's where all nodes in a LAN are directly attached to a node in the core.
    /// </summary>
    public class NetTopology : Topology
    {

        public enum TopologyType
        {
        GT_IM,
        Mercator
        };
        
        
        public const string GraphsDirectory = @"C:\SkipNet\Topologies";  // Directory in which network topology graphs are located.

        // Seed used with the various random number generators in this file.
        public static int randomSeed = 0;

        private const int AvgLanNodes = 100;       // Average number of nodes in each LAN
        private const int LanLat = 1;              // 1ms LAN latency
        private const int LanBand = SimParameters.DefaultLanBandwidth;
            private const int TransitTransitBand = SimParameters.DefaultTransitTransitBandwidth;
            private const int TransitStubBand = SimParameters.DefaultTransitStubBandwidth;
            private const int StubStubBand = SimParameters.DefaultStubStubBandwidth;
            private const int MaxDist = Int32.MaxValue;  

        private int nl;   // Number of nodes in LAN's
        private int nc;   // Number of nodes in network core, i.e., excluding LAN nodes
        private int nn;   // Total number of nodes in the network: numLanNodes+numCoreNodes
        private int maxn; // Maximum number of nodes that can be allocated
        private int numLinks; // Total number of physical links
        private int numModelnetLinks = 0; // used to enforce modelnet specific numbering (the two  simulator links describing a physical bidirectional link must have consecutive numbers)
        public Hashtable modelnetLinkNumbers = new Hashtable();

        public TopologyType TopType;

#if PING_STATS
        // These are used to gather information about who pings who
        public static bool record = false;
        public static bool heavyrecord = false;
        public static ArrayList src = new ArrayList();
        public static ArrayList dest = new ArrayList();
        public static Hashtable all = null;
#endif

        // Array with nas entries with an adjacency list representation of the AS overlay
        // Each NetAS includes intra-AS routes.
        private NetAS[] ass; 
        private int nas;     // Number of autonomous systems

        // Matrix with nas x nas entries where route[i,j] is the index of AS j's link that one
        // should take to route towards AS i 
        private ushort[,] asRoutes;

        // Array of maxn entries with an adjacency list representation of the entire network.
        // The index of a node is equal to its index in nodes.
        //AM change from private to public
        public NetNode[] nodes;

        private int logicalBase; // Minimum index that can be assigned to a logical node
        public int toAssign;    // Index of next physical node to assign to a logical node.
        public static Random r; // Used to assign LAN nodes to core nodes

        // Define a separate random number generator from r so that we have reproducible assignments of LAN nodes to core nodes.
        public static Random r1;  // Used to pick a random existing logical node.

        // Define yet another random number generator for selecting core nodes for an organization.
        public static Random rCore;

        private OverlayId defaultOid;
        
        void CheckTopologyFile(string filename)
        {
            if (!File.Exists(filename)) 
            {
                Console.WriteLine();
                Console.WriteLine("File {0} does not exist.", filename);
                Console.WriteLine("Did you remember to copy the release's Topologies directory to {0}?", GraphsDirectory);
                Environment.Exit(1);
            }
        }

        public NetTopology(string netFile, string routesFile, string distancesFile):
            this(netFile, routesFile, null, -1, -1, -1) {}

        public NetTopology(string netFile, string routesFile, string distancesFile,
            int transitTransitBand, int transitStubBand, int stubStubBand)
            // effects: Reads network core topology from the file named by netFile, and the matrix with the
            // routes between nodes in the topology from the file named by routesFile.
            // transitTransitBand, transitStubBand, and stubStubBand specify bandwidths to use for the 
            // respective link types.  If -1 is passed in then the default value is used.
            // The use of this constructor is deprecated it is here for backwards compatibility 
            // with the old file formats and non-hierarchical routing
        {
            TopType = TopologyType.GT_IM;

            // Define an overlayId for the default overlay network.
            defaultOid.idGlobal = 0;
            defaultOid.idLocal = 0;

            // Read core network from file.
            CheckTopologyFile(netFile);
            BinaryReader nr = new BinaryReader(File.OpenRead(netFile));

            // FORMAT binary little endian: 
            // First ushort gives number of nodes. Note that this means that there cannot
            // be more than UInt16.MaxValue nodes in the core. If more nodes are needed 
            // in the core change ushorts to ints...
            nc = nr.ReadUInt16();
            nl = 0;
            nn = nc+nl;
            maxn = nc*(AvgLanNodes+1);
                if (maxn < 262144 /* # nodes up to 2^18 */ + 10100 /* max topology */ + 1)
                maxn = 262144 /* # nodes up to 2^18 */ + 10100 /* max topology */ + 1;
            nodes = new NetNode[maxn];
            numLinks = 0;
            
            // With this value of logicalBase only LAN nodes are assigned to logical nodes.
            // Switch to 0 to assign logical nodes to core network nodes.
            logicalBase = nc; 
            toAssign = logicalBase; 
            
            // ...This is followed by a record for each node...
            for (int i=0; i < nc; i++) 
            {
                // ...The node record has a name, which is a string encoded as a byte len
                // followed by a sequence of bytes of that length, and a ushort with the number of 
                // links leaving the node followed by that number of link records.
                // (The index of a node record in the file is equal to the index
                // used to identify the node.)
                string name = nr.ReadString(); 
                
                NetNode newNode = new NetNode(i, 0, i, name, null);
                nodes[i] = newNode;
                nodes[i].AssignAddress(new SimNodePhysicalAddress(i));  // CoreNodesForOrg needs this.
//              Console.WriteLine("nodes[{0}]: {1}", i, nodes[i].Name()); // xxx mmt

                int nls = nr.ReadUInt16();
                for (int j=0; j < nls; j++) 
                {
                    // Each link record contains a ushort with the destination node index, followed 
                    // by an int with nominal latency, and another int with nominal bandwidth.
                    int dest = nr.ReadUInt16();
                    int lat = nr.ReadInt32();
                    int band = nr.ReadInt32();
                    newNode.AddLink(new NetLink(dest, lat, band));
                    numLinks++;
                }
            }

            // Set bandwidths of links in the core.
            if (transitTransitBand == -1) transitTransitBand = TransitTransitBand;
            if (transitStubBand == -1) transitStubBand = TransitStubBand;
            if (stubStubBand == -1) stubStubBand = StubStubBand;
            for (int i = 0; i < nc; i++) 
            {
                foreach (NetLink l in nodes[i])
                {
                    string typeStr = "?,?"; // mmt xxx
                    if (nodes[i].Name().StartsWith("T")) 
                    {
                        if (nodes[l.Destination()].Name().StartsWith("T")) 
                        {
                            typeStr = "T,T"; // mmt xxx
                            l.SetBandwidth(transitTransitBand);
                        }
                        else 
                        {
                            typeStr = "T,S"; // mmt xxx
                            l.SetBandwidth(transitStubBand);
                        }
                    }
                    else 
                    {
                        if (nodes[l.Destination()].Name().StartsWith("T")) 
                        {
                            typeStr = "S,T"; // mmt xxx
                            l.SetBandwidth(transitStubBand);
                        }
                        else 
                        {
                            typeStr = "S,S"; // mmt xxx
                            l.SetBandwidth(stubStubBand);
                        }
                    }
//                  Console.WriteLine("{0} -{1}-> {2}", nodes[i].Name(), typeStr, nodes[l.Destination()].Name());
                }
            }

            // Read route matrix from file
            CheckTopologyFile(routesFile);
            BinaryReader rr = new BinaryReader(File.OpenRead(routesFile));
 
            if (nc != (int) rr.ReadUInt16()) 
            {
                throw new SystemException("Routes file does not match network file");
            }

            // Read route matrix and rewrite it to match the new format
            ushort [,] route = new ushort[nc,nc];
            for (int i=0; i < nc; i++) 
            {
                for (int j=0; j < nc; j++) 
                {
                    route[i,j] = (ushort)nodes[j].LinkIndex(rr.ReadUInt16());
                    int dummy = rr.ReadUInt16();
                }
            }

            nas = 1;
            ass = new NetAS[nas];
            ass[0] = new NetAS(nc, route, nodes);
            asRoutes = new ushort[nas,nas];
            asRoutes[0,0] = 0;
            
            // Sanity check: route between two random nodes; if the routes matrix 
            // does not match the graph this should generate a null dereference 
            // exception with high probability.
            Route(Math.Abs(r.Next()%nn),Math.Abs(r.Next()%nn),false);
        }


        Hashtable asNodeLists;
        
        public NetTopology(string asFile, string asRoutesFile, string routerRoutesFile,
            string routerFile, string gatewayFile, int transitTransitBand, int transitStubBand, int stubStubBand) 
        {
            TopType = TopologyType.Mercator;

            // Define an overlayId for the default overlay network.
            defaultOid.idGlobal = 0;
            defaultOid.idLocal = 0;

            // Read router topology
            // FORMAT binary little endian: 
            // First int gives number of nodes...
            CheckTopologyFile(routerFile);
            BinaryReader nr = new BinaryReader(File.OpenRead(routerFile));
            nc = nr.ReadInt32();
            nl = 0;
            nn = nc+nl;
            maxn = nc*(AvgLanNodes+1);
            nodes = new NetNode[maxn];
            numLinks = 0;

            asNodeLists = new Hashtable(3001);  // Assuming that there are probably around O(3000) ASes.
            
            // With this value of logicalBase only LAN nodes are assigned to logical nodes.
            // Switch to 0 to assign logical nodes to core network nodes.
            logicalBase = nc; 
            toAssign = logicalBase; 
            
            // ...This is followed by a record for each node...
            for (int i=0; i < nc; i++) 
            {
                // ...The node record has a ushort with the id of its AS, a ushort 
                // with the index of the node in that AS, and ushort with the number of 
                // links leaving the node followed by that number of link records.
                // (The index of a node record in the file is equal to the index
                // used to identify the node.)

                int myas = nr.ReadUInt16();
                int asindex = nr.ReadUInt16();

                string myasName = myas.ToString();
                string name = myasName + "/" + i.ToString();
                // XXXXXneed to set node names right. Could classify ASs into 
                // stubs and transits and put the S: or T: followed by the AS name
                // and the nodes index within the AS.

                NetNode newNode = new NetNode(i, myas, asindex, name, null);
//                if (asindex > 60000) // xxx mmt
//                    Console.WriteLine("{0} ({1}) in {2} has AS index {3}", name, i, myas, asindex); // xxx mmt
                nodes[i] = newNode;
                nodes[i].AssignAddress(new SimNodePhysicalAddress(i));  // CoreNodesForOrg needs this.
                ArrayList asNodeList = (ArrayList) asNodeLists[myasName];
                if (asNodeList == null)
                {
                asNodeList = new ArrayList();
                asNodeLists.Add(myasName, asNodeList);
                }
                asNodeList.Add(i);
//              Console.WriteLine("nodes[{0}]: {1}", i, nodes[i].Name()); // xxx mmt

                int nls = nr.ReadUInt16();
                for (int j=0; j < nls; j++) 
                {
                    // Each link record contains an int with the destination node index, followed 
                    // by an int with nominal latency, and another int with nominal bandwidth.
                    int dest = nr.ReadInt32();
            //      int lat = nr.ReadInt32(); // Not there yet
                    int lat = 1;
            //      int band = nr.ReadInt32(); // Not there yet
                    int band = 1;
                    NetLink link = new NetLink(dest, lat, band);
                    newNode.AddLink(link);
                    if (i < dest)
                      {
                        link.Index = numModelnetLinks;
                        numModelnetLinks+=2;
                      }
                    else
                      {
                        NetLink existing = nodes[dest]. LinkTo (i);
                        Debug.Assert(existing != null );
                        link.Index = existing.Index + 1;     
                      }

//                      link.Index = numLinks;
                    numLinks++;
                }
            }

            // Read AS network from files.
            CheckTopologyFile(asFile);
            BinaryReader ASTop = new BinaryReader(File.OpenRead(asFile));

            CheckTopologyFile(routerRoutesFile);
            BinaryReader intraASRoutes = new BinaryReader(File.OpenRead(routerRoutesFile));

            CheckTopologyFile(gatewayFile);
            BinaryReader gateways = new BinaryReader(File.OpenRead(gatewayFile));

            nas = ASTop.ReadUInt16();
            ass = new NetAS[nas];
            for (int i=0; i < nas; i++) 
            {
                ass[i] = new NetAS(i, ASTop, intraASRoutes, gateways, nodes);
            }

            // Read inter AS routes
            CheckTopologyFile(asRoutesFile);
            BinaryReader interASRoutes = new BinaryReader(File.OpenRead(asRoutesFile));
            if (nas != (int) interASRoutes.ReadUInt16()) 
            {
                throw new SystemException("AS routes file does not match AS network file");
            }

            asRoutes = new ushort[nas,nas];
            for (int i=0; i < nas; i++) 
            {
                for (int j=0; j < nas; j++) 
                {
                    asRoutes[i,j] = interASRoutes.ReadUInt16();
                }
            }

            
            // Set bandwidths of links in the core.
            if (transitTransitBand == -1) transitTransitBand = TransitTransitBand;
            if (transitStubBand == -1) transitStubBand = TransitStubBand;
            if (stubStubBand == -1) stubStubBand = StubStubBand;
            for (int i = 0; i < nc; i++) 
            {
                foreach (NetLink l in nodes[i])
                {
                    string typeStr = "?,?"; // mmt xxx
                    int srcAS = nodes[i].getAS;
                    int dstAS = nodes[l.Destination()].getAS;
                    if (srcAS != dstAS) {
                        // Inter-AS link.
                        typeStr = "T,T";
                        l.SetBandwidth(ChooseInterASBandwidth());
                        l.SetLatency(ChooseInterASLatency());
                    }
                    else {
                        // Intra-AS link.
                        typeStr = "S,S"; // mmt xxx
                        l.SetBandwidth(stubStubBand);
                        l.SetLatency(r.Next(SimParameters.MinStubStubLatency, SimParameters.MaxStubStubLatency + 1));
                    }
                    /*
                    if (nodes[i].Name().StartsWith("T")) 
                    {
                        if (nodes[l.Destination()].Name().StartsWith("T")) 
                        {
                            typeStr = "T,T"; // mmt xxx
                            l.SetBandwidth(transitTransitBand);
                        }
                        else 
                        {
                            typeStr = "T,S"; // mmt xxx
                            l.SetBandwidth(transitStubBand);
                        }
                    }
                    else 
                    {
                        if (nodes[l.Destination()].Name().StartsWith("T")) 
                        {
                            typeStr = "S,T"; // mmt xxx
                            l.SetBandwidth(transitStubBand);
                        }
                        else 
                        {
                            typeStr = "S,S"; // mmt xxx
                            l.SetBandwidth(stubStubBand);
                        }
                    }
                    */
                    // Console.WriteLine("{0} ({1}) -{2}-> {3}  lat: {4}, bw: {5}", nodes[i].Name(), i, typeStr, nodes[l.Destination()].Name(), l.Latency(), l.Bandwidth());
                }
            }

            // Sanity check: route between two random nodes; if the routes matrix 
            // does not match the graph this should generate a null dereference 
            // exception with high probability.
        
            Route(Math.Abs(r.Next()%nn),Math.Abs(r.Next()%nn),false);
        }

        /// <summary>
        /// Choose an inter-AS link bandwidth.
        /// Algorithm:
        /// 2 classes of links: 3/4 OC3 links and 1/4 T3 links.
        /// </summary>
        /// <returns></returns>
        int ChooseInterASBandwidth() {
            int linkClass = r.Next(4);
            if (linkClass == 0) {
                return SimParameters.T3Bandwidth;
            }
            else {
                return SimParameters.OC3Bandwidth;
            }
        }

        /// <summary>
        /// Choose an inter-AS link latency..
        /// Algorithm: 
        /// 2 classes of links: 3/4 OC3 links and 1/4 T3 links
        /// Draw uniformly from [20,160] msecs for OC3 and from [200,400] for T3 links.
        /// </summary>
        /// <returns></returns>
        int ChooseInterASLatency() {
            double linkClass = r.NextDouble();
            if (linkClass > SimParameters.FractionOfOC3) {
                return r.Next(SimParameters.MinT3Latency, SimParameters.MaxT3Latency + 1);
            }
            else {
                return r.Next(SimParameters.MinOC3Latency, SimParameters.MaxOC3Latency + 1);
            }
        }

        public  void  ExportToModelnet (string filenameModel, string filenameRoutes, bool exportLat, bool exportBW )
        {
          Hashtable used = new Hashtable ();
          
          ExportRoutes (filenameRoutes, used);
          ExportModel(filenameModel, used, exportLat, exportBW);
        }

        public  void  ExportRoutes (string filename, Hashtable used  ) 
        {
          Console.Write("exporting modelnet  routes to {0} ...", filename);
          StreamWriter writer = new StreamWriter (filename, false );

            writer.WriteLine("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>");
            writer.WriteLine("<allpairs>");
            for (int source = logicalBase; source < toAssign ;source++ )
            {
                for (int destination = logicalBase; destination < toAssign ;destination++ )
                {
                  if (source == destination)
                    {
                      continue;
                    }

                    writer.Write("<path int_vndst=\"{1}\" int_vnsrc=\"{0}\" hops=\"",  source - logicalBase, destination - logicalBase);
                    ArrayList path = new ArrayList ();
                    RoutePath(source, destination, 0, 0, path );
                    foreach (int entry in   path)
                    {
                      used[entry] = 1;
                      writer.Write("{0} ", entry);
                    }
                    writer.WriteLine("\" />");
                }
            }
            writer.WriteLine("</allpairs>");
            writer.Close ();
            Console.WriteLine("done");
        }

        public  void  ExportModel (string filename, Hashtable used, bool exportLat, bool exportBW  ) 
        {
          Console.Write("exporting modelnet model to {0} ...", filename);
          StreamWriter writer = new StreamWriter (filename, false );

            writer.WriteLine("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>");
            writer.WriteLine("<model>");
            writer.WriteLine("  <hops>");
            for (int i =0; i < toAssign /*nc*/; i++) 
            {
                foreach (NetLink l in nodes[i])
                {
                    if (used[l.Index] == null ) {
                        continue;
                    }

                    int latency = 1;
                    if (exportLat) latency = l.Latency();
                    int bw = 10000;
                    if (exportBW) bw = l.Bandwidth();
                    writer.WriteLine("    <hop dbl_len=\"{0}\" dbl_kbps=\"{1}\" int_src=\"{2}\" int_dst=\"{3}\" int_delayms=\"{4}\" int_len=\"{5}\" int_qlen=\"{6}\" int_emul=\"0\" specs=\"transit-transit\" int_idx=\"{7}\" dbl_plr=\"{8}\" />", 
                        latency, bw, nodes[i].Index, l.Destination(),
                        latency, latency, 20, l.Index, 0);
                }
            }
            writer.WriteLine("  </hops>");
            writer.WriteLine("</model>");
            writer.Close ();
            Console.WriteLine("done");
        }

        public static void SetRandomSeed(int seed) {
        randomSeed = seed;
        r = new Random(randomSeed + 1234);
        r1 = new Random(randomSeed + 5678);
        rCore = new Random(randomSeed + 1342);
        }


        // Used to implement failure of gateways between ASes.
        bool FailGateways = false;
        int ASToFail = -1;

        /// <summary>
        /// Fail all transit nodes in the GT-IM model 
        /// or all gateway nodes in the Mercator model.
        /// </summary>
        public void FailAllGateways() {
        if (TopType == TopologyType.Mercator) {
            FailGateways = true;
        }
        else {
            // Fail all the transit nodes.
            for (int i = 0; i < nc; i++) {
            if (nodes[i].Name().StartsWith("T")) {
                nodes[i].Fail();
            }
            }
        }
        }

        /// <summary>
        /// Fail the gateways to the specified ASes.
        /// Only usable with the Mercator network topology.
        /// </summary>
        /// <param name="coreNodeIndex">Index of a core router node in the AS to fail.</param>
        public void FailGatewaysToAS(int coreNodeIndex) {
        if (TopType == TopologyType.Mercator) {
            FailGateways = true;
            ASToFail = nodes[coreNodeIndex].getAS;
        }
        else {
            throw new Exception("FailGatewaysToAS not implemented for GT-IM network topology");
        }
        }

        /// <summary>
        /// Unfail the gateways to the specified ASes.
        /// Only usable with the Mercator network topology.
        /// </summary>
        public void UnfailGatewaysToAS() {
        if (TopType == TopologyType.Mercator) {
            FailGateways = false;
        }
        else {
            throw new Exception("UnfailGatewaysToAS not implemented for GT-IM network topology");
        }
        }

        /// <summary>
        /// Fail the specified number of router core nodes.  Pick nodes randomly.
        /// If 'transitNodesOnly' then only fail transit nodes in the GT-IM model
        /// or gateway nodes in the Mercator model.
        /// </summary>
        /// <param name="node_core_failures"></param>
        /// <param name="transitNodesOnly"></param>
        public void FailCoreNodes(int node_core_failures, bool transitNodesOnly)
        {
        Random random = new Random(randomSeed + 137);
        int[] gNodes = null;
        if (TopType == TopologyType.Mercator && transitNodesOnly)
        {
            Hashtable gNodesTable = new Hashtable();
            for (int i = 0; i < ass.Length; i++)
            {
            int[] asGNodes = ass[i].GetGatewayNodes();
//          Console.WriteLine("gateway nodes for AS {0} (n={1}):", i, asGNodes.Length); // xxx mmt
            for (int j = 0; j < asGNodes.Length; j++)
            {
                int indx = asGNodes[j];
                object o = gNodesTable[indx];
                int v;
                if (o == null) v = 1;
                else v = ((int)o) + 1;
                gNodesTable[indx] = v;
//              Console.WriteLine("  {0} -> {1}", nodes[indx].Name(), v); // xxx mmt
            }
            }
            gNodes = new int[gNodesTable.Count];
            int n = 0;
            foreach (int indx in gNodesTable.Keys)
            {
            gNodes[n++] = indx;
            }
            Console.WriteLine("Number of Mercator gateway nodes: {0}", gNodes.Length);
        }
        for (int i = 0; i < node_core_failures; i++) 
        {
            if (transitNodesOnly && TopType == TopologyType.Mercator)
            {
            int indx = random.Next(gNodes.Length);
            bool quitFlag = true;  // true after loop => all gNodes are faulty
            for (int j = 0; j < gNodes.Length; j++)
            {
                int gIndx = gNodes[indx];
                if (!nodes[gIndx].IsFaulty())
                {
                nodes[gIndx].Fail();
                quitFlag = false;
                break;
                }
                indx++;
                if (indx == gNodes.Length) indx = 0;
            }
            if (quitFlag) 
                return;
            }
            else
            {
            int indx = random.Next(nc);
            int origIndx = indx;
            while (true)
            {
                bool breakOk = true;
                if (transitNodesOnly &&
                (TopType == TopologyType.GT_IM) &&
                !nodes[indx].Name().StartsWith("T"))
                {
                breakOk = false;
                }
                if (breakOk && !nodes[indx].IsFaulty()) break;
                indx++;
                if (indx == nc) indx = 0;

                if (indx == origIndx) 
                return;  // No more candidate nodes.  We're done.
            }
            nodes[indx].Fail();
            }
        }
        }


        public bool SameAS(int nIndx1, int nIndx2) {
        if (TopType == TopologyType.Mercator)
            return nodes[nIndx1].getAS == nodes[nIndx2].getAS;
        else {
            string n1 = nodes[nIndx1].Name();
            n1 = n1.Substring(0, n1.IndexOf("/"));
            string n2 = nodes[nIndx2].Name();
            n2 = n2.Substring(0, n2.IndexOf("/"));
            return n1.Equals(n2);
        }
        }
        

        // This removes the distance matrix - which decreases memory usage but increases computation time
        public void FreeAllSpareMemory() { }

        public void FreeNodeCreationDataStructures()
        {
        this.asNodeLists = null;
        }


        // Dividing line between large and small organizations.
        // Large orgs get spread across a larger set of core nodes:
        // - GT-IM => Spread across and entire S<digit>.<digit>/ prefix instead of an S<digit>.<digit>/<digit> prefix.
        // - Mercator => Spread across a large AS instead of a small AS.
        private const int LargeOrgSize = 1000;
        private const int maxMembersPerLAN = 100;

        /// <summary>
        /// Object used to sort routing core node entries.
        /// </summary>
        private class CoreNodeEntry: IComparable
        {
        private static int rootIndex;
        private static NetTopology top;

        public int nodeIndex;
        public int distance;

        public CoreNodeEntry(int nodeIndex)
        {
            this.nodeIndex = nodeIndex;
            distance = top.Ping(top.nodes[rootIndex].Address(), top.nodes[nodeIndex].Address(), false);
//          distance = top.Distance(new SimNodePhysicalAddress(rootIndex), new SimNodePhysicalAddress(nodeIndex));
        }

        public static void SetRoot(int rootIndexArg, NetTopology topArg)
        {
            rootIndex = rootIndexArg;
            top = topArg;
        }

        public int CompareTo(object x)
        {
            return distance - ((CoreNodeEntry) x).distance;
        }
        }

        /// <summary>
        /// Pick a set of core nodes to which to attach the LAN nodes for the specified
        /// organization.  Return an array with the set of indices into the 'nodes' array 
        /// for these nodes as function return value.
        /// If 'clusterByOrg' is false then simply return an array of all core router nodes.
        /// If 'pickSameRegionAs' is -1 then we ignore the org name and just pick 
        /// a random GT stub or Mercator AS and return an array
        /// of all core node indices belonging to that stub/AS, sorted in distance order from
        /// a randomly chosen "root" core node in the stub/AS.
        /// If 'pickSameRegionAs' >= 0 then we pick the same GT stub or Mercator AS as
        /// node[pickSameRegionAs] belongs to.
        /// </summary>
        /// <param name="org">Name or organization to choose core nodes for.</param>
        /// <param name="numMembers">Number of members of organization.</param>
        /// <param name="clusterByOrg">True => pick only core nodes from a single GT stub or Mercator AS.</param>
        /// <param name="pickSameRegionAs">If >= 0 then specifies a node[] index.  If clusterByOrg then pick the same GT stub or Mercator AS as node[pickSameRegionAs].</param>
        /// <returns></returns>
        public int[] CoreNodesForOrg(string org, int numMembers, bool clusterByOrg, int pickSameRegionAs)
        {
        int[] cn;

        if (!clusterByOrg)
        {
            // Return all core nodes, in the order they are in the nodes array 
            // (i.e. not sorted by distance to anything).
            cn = new int[nc];
            for (int i = 0; i < nc; i++) cn[i] = i;
            return cn;
        }

        int cindex;
        int startIndx;
        int endIndx;
        CoreNodeEntry[] cne;
        if (TopType == TopologyType.GT_IM)
        {

        retry:

            if (pickSameRegionAs >= 0) {
            cindex = pickSameRegionAs;
            }
            else {
            cindex = Math.Abs(rCore.Next()%nc);

            // Make sure we pick a stub and not a transit node.
            while (nodes[cindex].Name()[0] == 'T') {
                cindex = Math.Abs(rCore.Next()%nc);
            }
            }
            startIndx = cindex;
            endIndx = cindex;

            string stub = nodes[cindex].Name();
            int sepIndx;
            if (numMembers > LargeOrgSize)
            {
            // Spread org across all nodes whose name is a prefix like "S:2.3/".
            sepIndx = stub.IndexOf("/");
            }
            else
            {
            // Spread org across all nodes whose name is a prefix like "S:2.3/5.".
            sepIndx = stub.IndexOf("/");
            sepIndx = stub.IndexOf(".", sepIndx);
            }
            stub = stub.Substring(0, sepIndx);

            // Find the beginning of the set of core nodes belonging to this stub.
            while (nodes[startIndx].Name().StartsWith(stub))
            {
            startIndx--;
            if (startIndx == -1) break;
            }
            startIndx++;

            // Find the end of the set of core nodes belonging to this stub.
            while(nodes[endIndx].Name().StartsWith(stub))
            {
            endIndx++;
            if (endIndx == nc) break;
            }

            int cnSize = endIndx - startIndx;
            if ((pickSameRegionAs == -1) && (cnSize * maxMembersPerLAN < numMembers)) {
            goto retry;
            }
            cne = new CoreNodeEntry[cnSize];
            CoreNodeEntry.SetRoot(cindex, this);
            int j = 0;
            for (int i = startIndx; i < endIndx; i++)
            {
            cne[j++] = new CoreNodeEntry(i);
            }
        }
        else
        {
            while (true)
            {
            if (pickSameRegionAs >= 0) {
                cindex = pickSameRegionAs;
            }
            else {
                cindex = Math.Abs(rCore.Next()%nc);
            }
            startIndx = cindex;
            endIndx = cindex;

            int sepIndx = nodes[cindex].Name().IndexOf("/");
            string aS = nodes[cindex].Name().Substring(0, sepIndx);

            ArrayList asNodeList = (ArrayList) this.asNodeLists[aS];
            int cnSize = asNodeList.Count;

            if ((pickSameRegionAs == -1) && 
                ((cnSize * maxMembersPerLAN) < numMembers)) {
                continue; // This AS is too small for this organization.
            }

            cne = new CoreNodeEntry[cnSize];
            CoreNodeEntry.SetRoot(cindex, this);
            int j = 0;
            Debug.Assert(asNodeList != null);
            IEnumerator ie = asNodeList.GetEnumerator();
            while (ie.MoveNext())
            {
                int indx = (int) ie.Current;
                cne[j++] = new CoreNodeEntry(indx);
            }

            break;
            }
        }
        Array.Sort(cne);

        cn = new int[cne.Length];
        for (int i = 0; i < cne.Length; i++) cn[i] = cne[i].nodeIndex;
//      Console.WriteLine("Sorted array for org {0} with {1} members", org, numMembers); // xxx
//      for (int i = 0; i < cne.Length; i++) Console.WriteLine("{0}: {1}, {2} {3}", i, cne[i].nodeIndex, cne[i].distance, nodes[cne[i].nodeIndex].Name()); // xxx
        return cn;
        }

        public SimNodePhysicalAddress NewPhysicalNode(int[] coreNodesToPickFrom, ZipfSelect zs) 
            // effects: Assigns a new physical node and returns the address
            // of the pysical node.
        {
            if (toAssign < maxn) 
            {
                SimNodePhysicalAddress addr = new SimNodePhysicalAddress(toAssign);
                if (nodes[toAssign] == null) 
                {
                    // Allocate a new LAN node and assign it to a 
                    // core node randomly with uniform probability
                    int cindex = zs.Pick();
                    
#if NO_TRANSIT_LANS
                    // To prevent logical nodes in the transits
                    while (nodes[coreNodesToPickFrom[cindex]].Name()[0] == 'T') 
                    {
                        cindex = zs.Pick();
                    }
#endif
                    
                        cindex = coreNodesToPickFrom[cindex];
//                  Console.WriteLine("NewPhysicalNode: {0}  {1}", cindex, nodes[cindex].Name()); // xxx
                    NetLink cl = new NetLink(toAssign, LanLat, LanBand);
                    nodes[cindex].AddLanLink(cl);
                    cl.Index = numModelnetLinks;
                    numLinks++;
                    numModelnetLinks++;

                    NetNode newNode = new NetNode(toAssign, nodes[cindex].getAS, 
                                                    nodes[cindex].getASIndex, nodes[cindex].Name()+"LAN", cl);
                    nodes[toAssign] = newNode;
                
                    cl =new NetLink(cindex, LanLat, LanBand);
                    cl.Index = numModelnetLinks;
                    newNode.AddLink(cl);
                    numLinks++;
                    numModelnetLinks++;
                    nn++;
                }
                nodes[toAssign].AssignAddress(addr);
                toAssign++;
                return addr;
            }
            return null;
        }

        // Create a "minimal" logical node that will be associated with a router node.
        // We need this in order to associate a NodePhysicalAddress with a router node.
        public SimNodePhysicalAddress NewRouterNode(int nodeIndx)
        {
            SimNodePhysicalAddress addr = new SimNodePhysicalAddress(nodeIndx);
            return addr;
        }

        public int MaxLogicalNodes() 
            // effects: returns the maximum number of logical nodes that may be assigned in this topology
        {
            return maxn - logicalBase;
        }

        public int NumLogicalNodes()
            // effects: returns the number of logical nodes that have been assigned to physical nodes
        {
            return toAssign - logicalBase;
        }

        public NetNode GetNode(int i)
            // requires: i >= logicalBase() && i < logicalBase()+NumLogicalNodes()
            // effects: a generator for the NetNodes in this that are assigned to logical nodes
        {
            Debug.Assert(i >= LogicalBase() && i < LogicalBase()+NumLogicalNodes());
            return nodes[i];            
        }
 
        public int LogicalBase()
        {
            return logicalBase;
        }

        public int NumLinks()
            // effects: Returns the number of physical links in this
        {
            return numLinks;
        }

        // measure the distance between a and b, relative to the topology of the network and an
        // appropriate distance measure
        public void BeginPing(NodePhysicalAddress a, NodePhysicalAddress b, bool disableCache, PingCallback cb) {
            int distance;
            distance = Ping(a, b, disableCache);
            cb(b, distance);
        }


        public int Ping(NodePhysicalAddress a, NodePhysicalAddress b, bool alive) 
        {
            if (a.Failed || b.Failed)
                return -1;
            
            int ia = ((SimNodePhysicalAddress)a).GetNum;
            int ib = ((SimNodePhysicalAddress)b).GetNum;
            
#if PING_STATS
            // This was added by Ant for getting information about the number of pings
            // performed......
            if (heavyrecord & !alive)
            {
                Hashtable p = (Hashtable) all[ia];
                if (p == null)
                {
                    p = new Hashtable();
                    all.Add(ia,p);
                }
                if (!p.Contains(ib))
                {
                    p.Add(ib,true);
                    if (record)
                    {
                        src.Add(ia);
                        dest.Add(ib);
                    }
                }
            }
            else if (record & !alive)
            {
                src.Add(ia);
                dest.Add(ib);
            }
            // This is the end of the bit added by Ant
#endif

                if (ia == ib) return 0;  // src == dest

            return Route(ia, ib, false);
        }

        public int Distance(NodePhysicalAddress a, NodePhysicalAddress b, int msgSize) 
        {
        return Route(((SimNodePhysicalAddress)a).GetNum,((SimNodePhysicalAddress)b).GetNum, msgSize, 0);
        }

        public int Delay(NodePhysicalAddress a, NodePhysicalAddress b)
        {
            if (a.Failed || b.Failed)
                return -1;
            return Route(((SimNodePhysicalAddress)a).GetNum,((SimNodePhysicalAddress)b).GetNum, true);
        }

        public int Delay(NodePhysicalAddress a, NodePhysicalAddress b, int msgSize, long timeNow)
        {
            if (a.Failed || b.Failed)
                return -1;
            return Route(((SimNodePhysicalAddress)a).GetNum,((SimNodePhysicalAddress)b).GetNum, msgSize, timeNow);
        }

        //XXXXX We may no longer need this
        public int OneHopDelay(NodePhysicalAddress a, NodePhysicalAddress b, int msgSize, long timeNow, out NodePhysicalAddress nextA)
        {
            nextA = null;
            //      if ((((SimNodePhysicalAddress)a).Node != null && ((SimNodePhysicalAddress)a).Node.Failed) || ((SimNodePhysicalAddress)b).Node.Failed)
            if (a.Failed || b.Failed)
                return -1;
            return OneHopRoute(((SimNodePhysicalAddress)a).GetNum,((SimNodePhysicalAddress)b).GetNum, msgSize, timeNow, out nextA);
        }

        public int Route(int a, int b, bool stats)
        {
            int msgSize = stats ? RouteMsg.DefaultMsgSize : 0;
            return Route(a, b, msgSize, 0);
        }
    
        public int Route(int a, int b, int msgSize, long timeNow) 
            // requires: 0 <= a,b < nn
            // effects: Compute the delay in the route between a, b and update link statistics
                // If timeNow is 0 then don't update link statistics and 
                // queueing information (in the case where we're modeling bandwidth -- otherwise this is meaningless).
        {   
            Debug.Assert(a >= 0 && a < nn && b >= 0 && b < nn);

            if (a == b) return 0;  // src == dest

            int delay = 0;
            int s = a;
            int d = b;
            NetNode ns = nodes[s];
            NetNode nd = nodes[d];
//          string nsPrefix = ns.Name(); // xxx mmt
//          nsPrefix = nsPrefix.Substring(0, nsPrefix.IndexOf("/"));
//          string ndPrefix = nd.Name();
//          ndPrefix = ndPrefix.Substring(0, ndPrefix.IndexOf("/"));
//          if (!nsPrefix.Equals(ndPrefix))
//          {
//          Console.WriteLine("<<< Route {0} -> {1}", ns.Name(), nd.Name()); // xxx mmt
//          ndPrefix = nd.Name();  // noop
//          }

            // Find a node that has an entry in the routing tables
            while (!ns.InCore() || ns.DegreeCore <= 1) 
            {
                NetLink l = ns.FirstLink();
                delay += l.MessageDelay(msgSize, timeNow);

                s = l.Destination();
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }

                ns = nodes[s];
            }

            while (!nd.InCore() || nd.DegreeCore <= 1) 
            {
                int newd = nd.FirstLink().Destination();
                NetLink l = (nd.InCore()) ? nodes[newd].LinkTo(d) : nd.LinkFromCore();
                //XXXXX TODO: set link from core in degree 1 core routers to avoid 
                // doing the search above repeatedly

                d = newd;
                nd = nodes[d];              
                if (nd.IsFaulty()) 
                {
                    return -1;
                }

                delay += l.MessageDelay(msgSize, timeNow);
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }
            }

            // Find route between s and d:
            // -> Start by finding route between their ASs
            int sas = ns.getAS;
            int das = nd.getAS;
            if (FailGateways) {
            if (sas != das) {
                if (ASToFail == -1 || ASToFail == sas || ASToFail == das)
                return -1;
            }
            }
            while (sas != das) 
            {
                int lindex = asRoutes[das,sas];
                int news;
                    int rtDelay = ass[sas].routeThrough(s, lindex, out news, msgSize, timeNow); 
                    if (rtDelay == -1) return -1;
                delay += rtDelay;
                sas = ass[sas].neighbor(lindex);
    
                NetNode nns = nodes[news];
                if (nns.IsFaulty()) 
                {
                    return -1;
                }
    
                NetLink l = nns.LinkToAS(sas, nodes);
                rtDelay = l.MessageDelay(msgSize, timeNow);
                    delay += rtDelay;
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }
                s = l.Destination();
            }

            // Since sas == das, route to destination within sas
            int delayRoute = ass[sas].route(s, d, msgSize, timeNow);
                if (delayRoute == -1) return -1;
                delay += delayRoute;

            return delay;
        }
    
        public int RoutePath(int a, int b, int msgSize, long timeNow,ArrayList path) 
            // requires: 0 <= a,b < nn
            // effects: Compute the delay in the route between a, b and update link statistics
            // If timeNow is 0 then don't update link statistics and 
            // queueing information (in the case where we're modeling bandwidth -- otherwise this is meaningless).
        {   
            Debug.Assert(a >= 0 && a < nn && b >= 0 && b < nn);
            if (a == b) return 0;  // src == dest

            int delay = 0;
            int s = a;
            int d = b;
            NetNode ns = nodes[s];
            NetNode nd = nodes[d];
            //          string nsPrefix = ns.Name(); // xxx mmt
            //          nsPrefix = nsPrefix.Substring(0, nsPrefix.IndexOf("/"));
            //          string ndPrefix = nd.Name();
            //          ndPrefix = ndPrefix.Substring(0, ndPrefix.IndexOf("/"));
            //          if (!nsPrefix.Equals(ndPrefix))
            //          {
            //          Console.WriteLine("<<< Route {0} -> {1}", ns.Name(), nd.Name()); // xxx mmt
            //          ndPrefix = nd.Name();  // noop
            //          }

            // Find a node that has an entry in the routing tables
            while (!ns.InCore() || ns.DegreeCore <= 1) 
            {
                NetLink l = ns.FirstLink();
                delay += l.MessageDelay(msgSize, timeNow);
                path.Add(l.Index);

                s = l.Destination();
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }

                ns = nodes[s];
            }

            while (!nd.InCore() || nd.DegreeCore <= 1) 
            {
                int newd = nd.FirstLink().Destination();
                NetLink l = (nd.InCore()) ? nodes[newd].LinkTo(d) : nd.LinkFromCore();
                //XXXXX TODO: set link from core in degree 1 core routers to avoid 
                // doing the search above repeatedly

                d = newd;
                nd = nodes[d];              
                if (nd.IsFaulty()) 
                {
                    return -1;
                }

                delay += l.MessageDelay(msgSize, timeNow);
                path.Add(l.Index);
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }
            }

            // Find route between s and d:
            // -> Start by finding route between their ASs
            int sas = ns.getAS;
            int das = nd.getAS;
            if (FailGateways) 
            {
                if (sas != das) 
                {
                    if (ASToFail == -1 || ASToFail == sas || ASToFail == das)
                        return -1;
                }
            }

            while (sas != das) 
            {
                int lindex = asRoutes[das,sas];
                int news;
                int rtDelay = ass[sas].routeThroughPath(s, lindex, out news, msgSize, timeNow, path); 
                if (rtDelay == -1) return -1;
                delay += rtDelay;
                sas = ass[sas].neighbor(lindex);
    
                NetNode nns = nodes[news];
                if (nns.IsFaulty()) 
                {
                    return -1;
                }
    
                NetLink l = nns.LinkToAS(sas, nodes);
                rtDelay = l.MessageDelay(msgSize, timeNow);
                delay += rtDelay;
                path.Add(l.Index);

                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }
                s = l.Destination();
            }

            // Since sas == das, route to destination within sas
            int delayRoute = ass[sas].routePath(s, d, msgSize, timeNow, path);
            if (delayRoute == -1) return -1;
            delay += delayRoute;

            return delay;
        }
    


        public int OneHopRoute(int a, int b, int msgSize, long timeNow, out NodePhysicalAddress nextA) 
            // requires: 0 <= a,b < nn
            // effects: Compute the first-hop delay in the route between a, b and update link statistics
            // if stats is true. Returns -1 if there is a faulty node along the route.
        {   
            Debug.Assert(false, "Not implemented yet");
            Debug.Assert(a >= 0 && a < nn && b >= 0 && b < nn);
            int delay = 0;
            nextA = null;  // In case things go wrong.

            if (a == b) return 0;  // src == dest
/*
            int s = a;
            int d = b;
            NetNode ns = nodes[s];
            NetNode nd = nodes[d];

            // Find a node that has an entry in the routing tables
            while (!ns.InCore() || ns.DegreeCore <= 1) 
            {
                NetLink l = ns.FirstLink();
                delay += l.MessageDelay(msgSize, timeNow);

                s = l.Destination();
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }

                nextA = GetRouterAddress(s);
                return delay;
            }

            while (!nd.InCore() || nd.DegreeCore <= 1) 
            {
                int newd = nd.FirstLink().Destination();
                NetLink l = (nd.InCore()) ? nodes[newd].LinkTo(d) : nd.LinkFromCore();
                
                d = newd;
                nd = nodes[d];
                
                if (nd.IsFaulty()) 
                {
                    return -1;
                }

                if (d == s) 
                {
                    delay += l.MessageDelay(msgSize, timeNow);
                    if (msgSize > 0) 
                    {
                        l.NewPacket(msgSize);
                    }

                    nextA = GetRouterAddress(d);
                    return delay;
                }
            }

            // Find route between s and d:
            // -> Start by finding route between their ASs
            int sas = ns.getAS;
            int das = nd.getAS;
            while (sas != das) 
            {
                int lindex = asRoutes[das,sas];
                int news;
                delay += ass[sas].routeThrough(s, lindex, out news, msgSize, timeNow); 
                sas = ass[sas].neighbor(lindex);
    
                NetNode nns = nodes[news];
                if (nns.IsFaulty()) 
                {
                    return -1;
                }
    
                NetLink l = nns.LinkToAS(sas, nodes);
                delay += l.MessageDelay(msgSize, timeNow);
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }
                s = l.Destination();
            }

            // Since sas == das, route to destination within sas
            delay += ass[sas].route(s, d, msgSize, timeNow);

            return delay;
        }   
            if (a == b) return 0;  // src == dest

            NetNode na = nodes[a];
            NetNode nb = nodes[b];
            
            NetLink l = null;

            int s = a;
            int d = b;
            if (!na.InCore()) 
            {
                l = na.FirstLink();
                delay = l.MessageDelay(msgSize, timeNow);

                s = l.Destination();
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }

                nextA = GetRouterAddress(s);
                return delay;
            }

            if (!nb.InCore()) 
            {
                d = nb.FirstLink().Destination();

                if (s == d)
                {
                    if (nodes[d].IsFaulty()) 
                    {
                        return -1;
                    }

                    l = nb.LinkFromCore();
                    delay = l.MessageDelay(msgSize, timeNow);
                    if (msgSize > 0) 
                    {
                        l.NewPacket(msgSize);
                    }

                    nextA = GetRouterAddress(b);
                    return delay;
                }
            }

            // Find route between s and d
            int nextHop = d;
            RouteEntry re = route[s,d];
            NetNode np = null;
            while (s != d) 
            {
                nextHop = d;
                re = route[s,d];
                int p = re.pred;
                np = nodes[p];
                if (np.IsFaulty()) 
                {
                    return -1;
                }
                d = p;
            }
            l = np.LinkTo(d, re.lindex);
            delay = l.MessageDelay(msgSize, timeNow);
            if (msgSize > 0) 
            {
                l.NewPacket(msgSize);
            }
            nextA = GetRouterAddress(nextHop);
            */
            return delay;
        }

        private NodePhysicalAddress GetRouterAddress(int d)
        {
        NodePhysicalAddress addr = nodes[d].Address();
        if (addr == null)
        {
            nodes[d].AssignAddress(NewRouterNode(d));
            addr = nodes[d].Address();
        }
        return addr;
        }

        public NodePhysicalAddress GetRandom(NodePhysicalAddress n)
        {
        return GetRandom(n, defaultOid);
        }

        public NodePhysicalAddress GetRandom(NodePhysicalAddress n, OverlayId oid)
        {
        int ni = ((SimNodePhysicalAddress)n).GetNum;
        int nln = NumLogicalNodes();

        if (nln == 0) return null;

        NodePhysicalAddress la = null;
        while (true)
        {
            int choice = logicalBase + r1.Next(toAssign - logicalBase);
            if (ni == choice) continue;
            la = nodes[choice].Address();
            if (SimParameters.SupportMultipleOverlayNetworks)
            {
            if (la.GetLogicalNode(oid) != null)
            {
                break;
            }
            }
            else
            {
            break;
            }
        }
        return la;
        }
    
        public NodePhysicalAddress GetClosest(NodePhysicalAddress n)
        {
        return GetClosest(n, defaultOid);
        }


        public NodePhysicalAddress GetClosest(NodePhysicalAddress n, OverlayId oid) 
            // effects: Return the physical address of the logical node (other than n) 
            // that is closest to n
        {
            int minN = 0;
            int minD = MaxDist;
            int ni = ((SimNodePhysicalAddress)n).GetNum;
            int nln = NumLogicalNodes();

            if (nln == 0) 
                return null;

            if (logicalBase == nc && nln*10 > nc) 
            {
                // There is a good probability there is a logical node in ni's LAN;
                // if so that node is one of the closest.
                int ci = nodes[ni].FirstLink().Destination();

                while (true) 
                {
                    int li = nodes[ci].RandomLanLink(ni);
                    if (li >= 0 && li != ni) 
                    {
                        SimNodePhysicalAddress la = nodes[li].Address();
                        if ((la != null) && (!la.Failed)) // Added by ANT
                        {
                            if (SimParameters.SupportMultipleOverlayNetworks)
                            {
                                if (la.GetLogicalNode(oid) != null)
                                {
                                    return la;
                                }
                            }
                            else
                            {
                                return la;
                            }
                        }
                    }

                    if (nodes[ci].DegreeCore == 1)
                        ci = nodes[ci].FirstLink().Destination();
                    else
                        break;
                }
            }

            for (int i=logicalBase; i < toAssign; i++ ) 
            {
                if (ni != i)
                {
                    int tmp = Route(ni, i, false);
                    if (tmp < minD) 
                    {
                        if ((nodes[i].Address().GetLogicalNode(oid) != null) && (!nodes[i].Address().Failed)) // added by ANT
                        {
                            minD = tmp;
                            minN = i;
                        }
                    }
                }
            }
    
            return nodes[minN].Address();
        }
    
    
        public SimNodePhysicalAddress[] GetSetNthClosest(NodePhysicalAddress node, int num)
            // effects: Return the physical address of the num logical nodes that are closest to node
            // (including node).
        {
            if (SimParameters.SupportMultipleOverlayNetworks) throw new Exception("Can't use GetSetNthClosest methodif multiple overlay networks are allowed");
            
            ArrayList minN = new ArrayList(num);
            ArrayList minD = new ArrayList(num);
        
            for (int i = 0; i < num; i++)
            {
                minD.Add(MaxDist);
                minN.Add(-1);               
            }

            int ni = ((SimNodePhysicalAddress)node).GetNum;
        
            if (NumLogicalNodes() == 0) return null;
        
            for (int i=logicalBase; i < toAssign; i++ ) 
            {
                int tmp = Route(ni, i, false);
                for (int j = 0; j < num; j++)
                {
                    if (tmp < (int) minD[j]) 
                    {
                        minN.Insert(j,i);
                        minD.Insert(j,tmp);
                        minN.RemoveAt(minN.Count-1);
                        minD.RemoveAt(minD.Count-1);
                        break;
                    }
                }
            }
    
            //return SimNodePhysicalAddress.GetNode((int) minN[num]);
            SimNodePhysicalAddress[] result = new SimNodePhysicalAddress[minN.Count];
            for (int i = 0; i < minN.Count; i++)
                result[i] = nodes[(int) minN[i]].Address();
            return result;
        }   

        // This function is used by PrintStats
        private void Update(long[] data, long pcount)
        {
            if (pcount > data[2])
                data[2] = pcount;
            if (pcount < data[3])
                data[3] = pcount;
            data[1] += pcount;
            data[0]++;
        }

        // This function is used by PrintStats
        private void PrintData(long[] data, String title)
        {
            Console.WriteLine(title);
            Console.WriteLine("   min :" + data[3]);
            Console.WriteLine("   max :" + data[2]);
            Console.WriteLine("   average =" + (double)data[1]/(double)data[0]);
        }

        private void PrintData2(long[] data, String title)
        {
            Console.Write(title+","+data[3]+","+data[2]+","+(double)data[1]/(double)data[0] +",");
        }

        public void PrintStats(bool brief) 
        {
            // Number, Sum, Max, Min
            long[] Lan = {0,0,0,Int32.MaxValue};
            long[] InStub = {0,0,0,Int32.MaxValue};
            long[] InTransit = {0,0,0,Int32.MaxValue};
            long[] InterTransit = {0,0,0,Int32.MaxValue};
            long[] StubTransit = {0,0,0,Int32.MaxValue};

            for (int i=0; i < nn; i++) 
            {
                NetNode n = nodes[i];
                foreach (NetLink l in n) 
                {
                    long pcount = l.GetPacketCount;

                    if (!n.InCore() || !nodes[l.Destination()].InCore())  // This is a LAN link
                    { // Now check there is a node on the address - otherwise we skew the results
                        if ((!n.InCore() && n.Address() != null) || (!nodes[l.Destination()].InCore() && (nodes[l.Destination()].Address() != null)))
                            Update(Lan, pcount);
                    }
                    else if (n.Name().StartsWith("T") && nodes[l.Destination()].Name().StartsWith("T")) // Transit -> Transit
                    { 
                        if (n.Name().StartsWith(nodes[l.Destination()].Name().Substring(0,nodes[l.Destination()].Name().IndexOf("."))))
                            Update(InTransit, pcount); // This is within the same transit
                        else
                            Update(InterTransit, pcount);
                    }
                    else if ((n.Name().StartsWith("S") && nodes[l.Destination()].Name().StartsWith("T")) || 
                        (n.Name().StartsWith("T") && nodes[l.Destination()].Name().StartsWith("S"))) // Stub Transit or vice-versa
                        Update(StubTransit, pcount); 
                    else if (n.Name().StartsWith("S") && nodes[l.Destination()].Name().StartsWith("S")) // Within a stub
                        Update(InStub, pcount);
                    else 
                        Console.WriteLine("!!!!!!!!! Not caught: " + n.Name() + " -> " + nodes[l.Destination()].Name());
                } 
            }
            if (brief)
            {
                PrintData2(Lan, "LAN");
                PrintData2(InStub, "InStub");
                PrintData2(StubTransit, "StubTransit");
                PrintData2(InTransit, "InTransit");
                PrintData2(InterTransit, "BetweenTransits");
                Console.WriteLine("");
            }
            else
            {
                Console.WriteLine("Link stress:");
                PrintData(Lan, "LAN");
                PrintData(InStub, "Within Stub:");
                PrintData(StubTransit, "Stub to Transit:");
                PrintData(InTransit, "Within Transit:");
                PrintData(InterTransit, "Between Transits:");
            }
        }


        public void DumpStats(bool includeLan) 
        {
            long max = Int32.MinValue;
            for (int i=0; i < nn; i++) 
                foreach (NetLink l in nodes[i]) 
                    if (l.GetPacketCount > max)
                        max = l.GetPacketCount;

            int[] counts = new int[max+1];

            for (int i=0; i < nn; i++) 
            {
                NetNode n = nodes[i];
                foreach (NetLink l in n) 
                {
                    long pcount = l.GetPacketCount;

                    if (includeLan)
                        counts[pcount]++;
                    else
                    {
                        if (!(!n.InCore() || !nodes[l.Destination()].InCore()))  // This is a LAN link
                            counts[pcount]++;
                    }
                } 
            }
            for (int i = 0; i < counts.Length; i++)
                Console.WriteLine(i + "," + counts[i]);
        }

        public void DumpStatsCompact(bool includeLan) 
        {
        long max = Int32.MinValue;
        for (int i=0; i < nn; i++) 
            foreach (NetLink l in nodes[i]) 
            if (l.GetPacketCount > max)
                max = l.GetPacketCount;

        int[] counts = new int[max+1];

        for (int i=0; i < nn; i++) 
        {
            NetNode n = nodes[i];
            foreach (NetLink l in n) 
            {
            long pcount = l.GetPacketCount;

            if (includeLan)
                counts[pcount]++;
            else
            {
                if (!(!n.InCore() || !nodes[l.Destination()].InCore()))  // This is a LAN link
                counts[pcount]++;
            }
            } 
        }
        for (int i = 0; i < counts.Length; i++)
            if (counts[i] > 0)
            Console.WriteLine(i + "," + counts[i]);
        }
        
        public int[] LinkStresses() 
            // effects: Returns a new array with an entry for the link stresses in each link.
            // The links are sorted in a canonical order and the array has size NumLinks()
        {
            int [] ls = new int[NumLinks()];
            int j = 0;
            for (int i=0; i < nn; i++) 
            { 
                foreach (NetLink l in nodes[i]) 
                {
                    ls[j] = (int)l.GetPacketCount;
                    j++;
                }
            }
            return ls;
        }

        public void DumpStats2(bool includeLan) 
        {
            //Include cumulative results + average link stress, do not take 0 into account
            long max = Int32.MinValue;
            for (int i=0; i < nn; i++) 
                foreach (NetLink l in nodes[i]) 
                    if (l.GetPacketCount > max)
                        max = l.GetPacketCount;

            int[] counts = new int[max+1];

            for (int i=0; i < nn; i++) 
            {
                NetNode n = nodes[i];
                foreach (NetLink l in n) 
                {
                    long pcount = l.GetPacketCount;

                    if (includeLan)
                        counts[pcount]++;
                    else
                    {
                        if (!(!n.InCore() || !nodes[l.Destination()].InCore()))  // This is a LAN link
                            counts[pcount]++;
                    }
                } 
            }
            int cumul=0;
            int total=0;
            for (int i = 0; i < counts.Length; i++)
            {
                if (counts[i]!=0) 
                {
                    cumul+= counts[i];
                    total+=i*counts[i];
                    Console.WriteLine(i + "," + counts[i] +"," + cumul);
                }
            }
            Console.WriteLine("AverageLink stress" + total/cumul);
        }

        public enum LinkType {tran, stub, lan};

        public struct LinkDelayStats 
        {
        public int sourceNodeIndex;
        public LinkType sourceLinkType;
        public int destNodeIndex;
        public LinkType destLinkType;
        public int latency;
        public int bandwidth;
        public int numMessagesSent;
        public int numDelayedMessages;
        public double sumDelayTimes;
        public double sumDelayTimesSquared;
        public int maxDelayTime;
        }

        public LinkDelayStats[] LinkDelays(long timeNow) 
        {
        int numLinksUsed = 0;
        int numLinks = NumLinks();
        for (int i = 0; i < nn; i++) {
            foreach (NetLink l in nodes[i]) {
            if (l.NumMessagesSent > 0) numLinksUsed++;
            }
        }
        LinkDelayStats[] ld = new LinkDelayStats[numLinksUsed];
        int j = 0;
        for (int i = 0; i < nn; i++) 
        {
            foreach (NetLink l in nodes[i]) 
            {
            if (l.NumMessagesSent > 0) 
            {
                ld[j].sourceNodeIndex = i;
                string n = nodes[i].Name();
                if (n.Equals("LAN")) ld[j].sourceLinkType = LinkType.lan;
                else if (n.StartsWith("T")) ld[j].sourceLinkType = LinkType.tran;
                else ld[j].sourceLinkType = LinkType.stub;
                ld[j].destNodeIndex = l.Destination();
                n = nodes[l.Destination()].Name();
                if (n.Equals("LAN")) ld[j].destLinkType = LinkType.lan;
                else if (n.StartsWith("T")) ld[j].destLinkType = LinkType.tran;
                else ld[j].destLinkType = LinkType.stub;
                ld[j].latency = l.Latency();
                ld[j].bandwidth = l.Bandwidth();
                ld[j].numMessagesSent = l.NumMessagesSent;
                ld[j].numDelayedMessages = l.NumDelayedMessages;
                ld[j].sumDelayTimes = l.SumDelayTimes;
                ld[j].sumDelayTimesSquared = l.SumDelayTimesSquared;
                ld[j].maxDelayTime = l.MaxDelayTime;
                j++;
            }
            }
        }
        return ld;
        }
                    
        public void ResetStats()
        {
            for (int i=0; i < nn; i++) 
            {
                nodes[i].ResetStats();
            }
        }

        private static bool debugOutput = false;

        public static void DebugOutput(bool flag) 
        {
            debugOutput = flag;
        }

        public String GetLocationDescription(NodePhysicalAddress a)
        {
            String result;
            NetNode p = nodes[nodes[((SimNodePhysicalAddress)a).GetNum].FirstLink().Destination()];
            result = p.Name() + " LINKS: ";
            foreach (NetLink n in p)
            {
                if (!nodes[n.Destination()].Name().StartsWith("LAN"))
                    result += nodes[n.Destination()].Name() +",";
            }
            return result;
        }

        public class LinkStressEnumerator
        {
            private int curNode;
            private int curLink;
            private NetTopology top;
        
            public LinkStressEnumerator(NetTopology t)
            {
                curNode = 0;
                curLink = -1;
                top = t;
            }

            public bool MoveNext()
            {
                ++curLink; 
                if (curLink < top.nodes[curNode].Degree) 
                {
                    return true;
                } 
        
                curNode++;
                if (curNode < top.nn) 
                {
                    curLink = 0;
                    return true;
                }

                return false;
            }

            public void Reset()
            {
                curLink = -1;
                curNode = 0;
            }

            public long Current
            {
                get
                {
                    return top.nodes[curNode].LinkToIndex(curLink).GetPacketCount;
                }
            }
        }

    }

    
    public class NetNode : IComparable {
        private int index;            // index of node in NetTopology nodes
        private bool faulty;          // true iff node is faulty
        private NetLink linkFromCore; // null if node is in core; otherwise link from core node to this

        private int nodeAS;  // Index of this node's AS
        private int ASIndex; // Index of this node in AS's tables

        // If not null this node has been assigned to a logical node with
        // the address below
        public SimNodePhysicalAddress logicalNodeAddress;
 
        // links that start at this node ordered by increasing node index
        private ArrayList links;

        // The links after (and including) firstLan are nodes in the LAN of this.
        // if this is a core node. Otherwise, firstLan = -1
        private int firstLan; 
 
        // Name that indicates nodes position in the topology. It is topology specific.
        private string name;

        // Keep track of current time on this node and how much to "push" it forward to account for
        // the overhead of sending messages.
        private long timeNow = 0L;
        private int numMsgsSent = 0;

        //AM Add XXXXXXX where is this used ?
        public NetNode() {
        }

        // Create a router netnode XXXXXXX where is this used ?
        public NetNode(int i, SimNodePhysicalAddress logicalNodeAddress) {
            index = i;
            faulty = false;
            linkFromCore = null;
            this.logicalNodeAddress = logicalNodeAddress;
            firstLan = -1;
            name = "Router";
            links = new ArrayList(1);
        }

        public int CompareTo(Object obj) {
            NetNode nn = (NetNode) obj;
            return String.Compare(this.name,2,nn.name,2,10);
        }
            
        public NetNode(int i, int nas, int asi, string n, NetLink c) { 
            // effects: Creates a new netnode with index i. If c is null the node is in the core
            // otherwise c must be the link from the core node to this.
            index = i;
            faulty = false;

            nodeAS = nas;
            ASIndex = asi;
            
            linkFromCore = c;
            // Debug.Assert(c.Destination() == index);

            logicalNodeAddress = null;
            firstLan = -1;
            
            name = n;
            
            links = new ArrayList();
        }

        public string Name() {
            return name;
        }

        public int Index {
            get { return index; }
        }

        public int getAS {
            get { return nodeAS; }
        }

        public int getASIndex {
            get { return ASIndex; }
        }

        public int Degree {
            get { return links.Count; }
        }

        public int DegreeCore {
            // Number of links to core nodes
            get { 
                if (firstLan < 0) {
                    return links.Count;
                } 
                else {
                    return firstLan;
                }
            }
        }

        public void AssignAddress(SimNodePhysicalAddress addr) {
            // effects: Assigns a logical node to this with address addr
            logicalNodeAddress = addr;
        }

        public SimNodePhysicalAddress Address() { 
            // effects: Returns the address associated with this
            // or null if there is no logical node associated with this
            return logicalNodeAddress;
        }

        public bool InCore() { 
            // effects: Return true iff this node is in the network core
            return linkFromCore == null;
        }

        public NetLink LinkFromCore() {
            // effects: If InCore() returns null; otherwise returns the link from this' parent
            // in the core to this.
            return linkFromCore;
        }

        public void Fail() { 
            // effects: marks this node as faulty; this prevents it from routing messages
            faulty = true;
        }

        public bool IsFaulty() {
            // effects: returns true iff this node is faulty
            return faulty;
        }

        public void Repair() {
            // effects: repairs node such that it is no longer faulty
            faulty = false;
        }

        public void AddLink(NetLink l) { 
            // effects: Adds a new link to the end of the array of links of this
            links.Add(l);
        }

        public void AddLanLink(NetLink l) { 
            // effects: Same as AddLink but link is marked as a LAN link
            if (firstLan < 0)
                firstLan = links.Count;
            links.Add(l);
        }

        
        public int LinkIndex(int dindex) {
            // effects: Returns the index of the link from this to the node with index dindex
            // or -1 if there is no such link.
            int sz = links.Count;
            for (int i=0; i < sz; i++) {
                NetLink l = (NetLink)links[i];
                if (l.Destination() == dindex)
                    return i;
            }
            return sz;
        }

        public NetLink LinkTo(int dindex) {
            // effects: Returns the link from this to the node with index dindex
            // or null if there is no such link. This is slow when there is a 
            // large number of links and it should be used sparingly.
            int sz = links.Count;
            for (int i=0; i < sz; i++) {
                NetLink l = (NetLink)links[i];
                if (l.Destination() == dindex)
                    return l;
            }
            return null;    
        }

        public NetLink LinkToIndex(int lindex) {
            // effects: Returns the link from this with index "lindex".
            NetLink l = (NetLink)links[lindex];
            return l;       
        }

        public NetLink LinkToAS(int das, NetNode[] nodes) {
            int sz = links.Count;
            for (int i=0; i < sz; i++) {
                NetLink l = (NetLink)links[i];
                if (nodes[l.Destination()].getAS == das)
                    return l;
            }
            return null;    
        }

        public NetLink FirstLink() { 
            // effects: Returns the first link in this' array of links.
            // If !InCore() this is the link to this' parent node in the core.
            return (NetLink) links[0];
        }

        public int FirstLanLink() {
            // effects: returns the index of the first node in this' LAN or -1
            // if this is not a core node with LAN links.
            if (firstLan > 0) 
                return ((NetLink)links[firstLan]).Destination();
            return -1;
        }

        public int RandomLanLink(int src) {
            // effects: returns the index of a random node in this' LAN different from
            // src or -1 if this is not a core node with LAN links.
            if (firstLan > 0) {
                int numLan = links.Count-firstLan;
                int rand = Math.Abs(NetTopology.r.Next());
                int ret = ((NetLink)links[firstLan+rand%numLan]).Destination();
                if (ret == src && numLan > 1) {
                    ret = ((NetLink)links[firstLan+(rand+1)%numLan]).Destination();
                }
                return ret;
            }
            return -1;
        }


        public void ResetStats() {
            // effects: Resets the statistics for all the links out of this.
            int sz = links.Count;
            for (int i=0; i < sz; i++) {
                ((NetLink)links[i]).ResetStats();
            }
        }

        public IEnumerator GetEnumerator() { 
            // effects: returns an enumerator for the entire list of links
            return links.GetEnumerator();
        }

        public long CurrentNodeTime {
            get {return this.timeNow;}
            set {this.timeNow = value;}
        }

        public int NumMsgsSentAtCurrentTime {
            get {return this.numMsgsSent;}
            set {this.numMsgsSent = value;}
        }
    }


    public class NetLinkBandwidthFields
    {
        // Total number of messages that have been sent.
        public int numMessagesSent;
        // Number of messages that couldn't be sent immediately.
        public int numDelayedMessages;

        // Time till bandwidth is available again.
        // *** in usecs ***
        public long timeWhenLinkAvail;

        // Statistics for time delay when messages are to be sent.
        // *** all kept in usecs ***
        public double sumDelayTimes;  // Sum of delay times when a message is delayed.
        public double sumDelayTimes2;  // Square.
        public long maxDelayTime;
        public long bCount;   // Number of bytes sent over link
        public int cBandwidth; // Current bandwidth
    }


    public class NetLink 
    {
        private int dest;     // index of destination node

        // Current latency.
        private int cLatency;
        private int index;
     
        // Statistics:
        private long pCount;   // Number of packets sent over link
        
        // Fields relating to bandwidth modeling.
        private NetLinkBandwidthFields bw;

        public NetLink(int d, int l, int b)
            // effects: Creates a new network link with destination index d, 
            // latency l, and bandwidth b 
        {
            dest = d;
            cLatency = l;
            if (SimParameters.ModelLinkBandwidth) 
            {
                bw = new NetLinkBandwidthFields();
                bw.cBandwidth = b;
            }
            else bw = null;

            ResetStats();
        }

        public long GetPacketCount
            // effects: returns the number of packets sent over link
        {
            get { return pCount; }
        }
        
        public int Destination() 
            // effects: returns the index of the destination of the link
        {
            return dest;
        }

        public void NewPacket(int msgSize)
            // effects: updates the statistics to reflect that a packet was sent on the link
        {
            pCount++;
            if (bw != null)
                bw.bCount += msgSize;
        }

        public void ResetStats() 
        {
            pCount = 0L;
            if (SimParameters.ModelLinkBandwidth)
            {
                bw.timeWhenLinkAvail = 0L;
                bw.numMessagesSent = 0;
                bw.numDelayedMessages = 0;
                bw.sumDelayTimes = 0.0;
                bw.sumDelayTimes2 = 0.0;
                bw.maxDelayTime = 0L;
                bw.bCount = 0L;
            }
        }

        private static bool debugOutput = false;

        public static void DebugOutput(bool flag) 
        {
            debugOutput = flag;
        }

        public int Latency() 
            // effect: returns the current link latency
        {
            return cLatency;
        }

        public void SetLatency(int lat) {
            cLatency = lat;
        }

        public int Bandwidth() 
            // effect: returns the current link bandwidth
        {
            if (bw != null)
                return bw.cBandwidth;
            else 
                return 1;
        }

        public void SetBandwidth(int bnd) 
        {
            if (bw != null)
                bw.cBandwidth = bnd;
        }

        // Return the delay that would be incurred on this link by a message of the specified size.
        // Update link stats and queueing delays unless timeNow is 0 (or we're not modeling bandwidth).
        public int MessageDelay(int msgSize, long timeNow) 
        {
        if (SimParameters.ModelLinkBandwidth)
        {
            if (timeNow == 0) 
            {
            return cLatency + msgSize / bw.cBandwidth;
            }

            bw.numMessagesSent++;
            timeNow = timeNow * 1000;  // convert to usecs
            int msgSendInterval = msgSize * 1000 / bw.cBandwidth;  // in usecs
            if (bw.timeWhenLinkAvail <= timeNow) 
            {
            bw.timeWhenLinkAvail = timeNow + msgSendInterval;
            }
            else 
            {
            bw.numDelayedMessages++;
            long qDelay = bw.timeWhenLinkAvail - timeNow;
            bw.sumDelayTimes += (double) qDelay;
            bw.sumDelayTimes2 += (double)qDelay * (double)qDelay;
            if (!(qDelay >= 0 && bw.sumDelayTimes >= 0.0 && bw.sumDelayTimes2 >= 0.0)) 
            {
                Console.WriteLine("Link delays negative! {0}, {1}, {2}", qDelay, bw.sumDelayTimes, bw.sumDelayTimes2);
                int a = 0;
                int b = 1/a;
            }
            if (qDelay > bw.maxDelayTime) bw.maxDelayTime = qDelay;
            bw.timeWhenLinkAvail += msgSendInterval;
            }
            int delay = ((int)((bw.timeWhenLinkAvail - timeNow) / 1000)) + cLatency;
            if (debugOutput) 
            Console.WriteLine("dst: {0}, msgSz: {1}, lat: {2}, bnd: {3}, dTime: {4}, dSend: {5}, delay: {6}", dest, msgSize, cLatency, bw.cBandwidth, bw.timeWhenLinkAvail - timeNow, msgSendInterval, delay);
            return delay;
        }
        else
        {
            return cLatency;
        }
        }

        public long TimeWhenLinkAvail 
        {
        get {return this.bw.timeWhenLinkAvail / 1000;}
        }

        public int NumMessagesSent 
        {
        get {return this.bw.numMessagesSent;}
        }

        public int NumDelayedMessages 
        {
        get {return this.bw.numDelayedMessages;}
        }

        public double SumDelayTimes 
        {
        get {return this.bw.sumDelayTimes / 1000.0;}
        }

        public double SumDelayTimesSquared 
        {
        get {return this.bw.sumDelayTimes2 / 1000000.0;}
        }

        public int MaxDelayTime 
        {
        get {return (int) (this.bw.maxDelayTime / 1000);}
        }
        public int  Index 
        {
        set {index = value;}
        get {return index;}
        }

    }



    public class NetAS
    {
        private int index;  // AS id which is the index in NetTopology ass
        private int size;   // Number of routers in this AS
        private int tsize;  // Number of routers in this AS with degree > 1

        private int nlinks; // Number of links out of this AS
        private ushort[] links; // Links out of this AS

        // Matrix with tsize x tszie entries where route[i,j] is the index of router j's link that one
        // should take to route towards router i 
        private ushort[,] routes; // Intra AS routes

        // Matrix with tsize x nlinks where closestGateway[i,j] returns the AS-index of the router
        // that is closest to i and has a link to the AS at the tip of the AS-link with index j
        private ushort[,] closestGateway; 

        private NetNode[] nodes;  // Array of NetTopology nodes

        public NetAS(int ind, BinaryReader ASTop, BinaryReader intraASRoutes, BinaryReader gateways, NetNode[] ns)
            // effects: Creates a new AS object with id "i" using the file readers passed as arguments
            // These readers must be positioned just before the entries of this AS. The first one must
            // contain the AS topology, the second one the routes inside ASs, and the third one the table
            // with the closest gateways.
        {
            index = ind;

            // Read links out of this AS
            nlinks = ASTop.ReadUInt16();
            links = new ushort[nlinks];
            for (int j=0; j < nlinks; j++) 
            {
                links[j] = ASTop.ReadUInt16();
            }

            // Read intra AS routing table
            tsize = intraASRoutes.ReadUInt16();
            routes = new ushort[tsize,tsize];
            for (int i=0; i < tsize; i++) 
            {
                for (int j=0; j < tsize; j++) 
                {
                    routes[i,j] = intraASRoutes.ReadUInt16();
                }
            }
        
            // Read table with closest gateways for each neighboring AS
            int nls = gateways.ReadUInt16();
            if (nls != nlinks) 
            {
                throw new SystemException("Intra AS routes file does not match gateways file");
            }

            closestGateway = new ushort[tsize,nlinks];
            for (int i=0; i < tsize; i++) 
            {
                for (int j=0; j < nlinks; j++) 
                {
                    closestGateway[i,j] = gateways.ReadUInt16();
                }
            }

            nodes = ns;

        }

        public NetAS(int sz, ushort[,] _routes, NetNode[] ns)
            // effects: The use of this constructor is deprecated. It is used to provide backwards
            // compatibility with non-hierarchical routing.
        {
            index = 0;
            nlinks = 0;
            links = null;
            tsize = size = sz;
            routes = _routes;
            closestGateway = null;
            nodes = ns;
        }

        public int[] GetGatewayNodes()
        {
        int asNodeIndex = 0;
        int[] gNodes;
            while (true) {
                int gNode = 0;  // Used to record if a node is a gateway node.
                if (this.nodes[asNodeIndex] == null) {  // Run past the end of the nodes?
                    gNodes = new int[1];
                    gNodes[0] = gNode;
                    return gNodes;
                }
                if (this.nodes[asNodeIndex].getAS == this.index) {
                    if (this.nodes[asNodeIndex].getASIndex != 65535) break;
                    else gNode = asNodeIndex;
                }
                asNodeIndex++;
            }
            NetNode ns = this.nodes[asNodeIndex];
            gNodes = new int[this.nlinks];
            int nextGNode = 0;
            for (int i = 0; i < this.nlinks; i++) {
                ushort asGatewayIndex = this.closestGateway[ns.getASIndex, i];
                int gatewayNodeIndex;
                routeInternal(ns, asGatewayIndex, out gatewayNodeIndex, 0, 0);
                gNodes[nextGNode++] = gatewayNodeIndex;
            }
            return gNodes;
        }

        public int routeThrough(int s, int lindex, out int news, int msgSize, long timeNow) 
        {
            NetNode ns = nodes[s];
            ushort d = closestGateway[ns.getASIndex,lindex];

            return routeInternal(ns, d, out news, msgSize, timeNow);
        }

        public int route (int s, int d, int msgSize, long timeNow) 
        {
            int news = 0;
            NetNode ns = nodes[s];
            int ad = nodes[d].getASIndex;
            int delay = routeInternal(ns, ad, out news, msgSize, timeNow); 
            Debug.Assert(news == d);

            return delay;
        }

        private int routeInternal(NetNode ns, int d, out int news, int msgSize, long timeNow) 
        {
            int delay = 0;
            int s = ns.getASIndex;
            news = ns.Index;

            while (s != d) 
            {
                int lindex = routes[d,s];
                NetLink l = ns.LinkToIndex(lindex);
                ns = nodes[l.Destination()];
                s = ns.getASIndex;

                if (ns.IsFaulty()) 
                {
                    return -1;
                }

                delay += l.MessageDelay(msgSize, timeNow);
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }
            }

            news = ns.Index;
            return delay;
        }

        public int routeThroughPath(int s, int lindex, out int news, int msgSize, long timeNow, ArrayList path) 
        {
            NetNode ns = nodes[s];
            ushort d = closestGateway[ns.getASIndex,lindex];

            return routeInternalPath(ns, d, out news, msgSize, timeNow, path);
        }

        public int routePath (int s, int d, int msgSize, long timeNow, ArrayList path) 
        {
            int news = 0;
            NetNode ns = nodes[s];
            int ad = nodes[d].getASIndex;
            int delay = routeInternalPath(ns, ad, out news, msgSize, timeNow, path); 
            Debug.Assert(news == d);

            return delay;
        }

        private int routeInternalPath(NetNode ns, int d, out int news, int msgSize, long timeNow, ArrayList path) 
        {
            int delay = 0;
            int s = ns.getASIndex;
            news = ns.Index;

            while (s != d) 
            {
                int lindex = routes[d,s];
                NetLink l = ns.LinkToIndex(lindex);
                ns = nodes[l.Destination()];
                s = ns.getASIndex;

                if (ns.IsFaulty()) 
                {
                    return -1;
                }

                delay += l.MessageDelay(msgSize, timeNow);
                if (msgSize > 0) 
                {
                    l.NewPacket(msgSize);
                }
                path.Add(l.Index);
            }

            news = ns.Index;
            return delay;
        }

        public int neighbor(int lindex) 
        {
            Debug.Assert(lindex >= 0 && lindex < links.Length);
            return links[lindex];
        }

    }


    public class ZipfSelect
    {
    static Random r = null;
    int numDataElements;
    double[] sums;
    bool defaultSelectZipf = true;

    public ZipfSelect(int numDataElements)
    {
        this.numDataElements = numDataElements;
        sums = new double[numDataElements];
        double sum = 0.0;
        for (int i = 0; i < sums.Length; i++) {
        sum += 1.0 / ((double)(i+2));
        }
        for (int i = 0; i < sums.Length; i++) {
        if (i == 0)
            sums[i] = 1.0 / ((double)(i+2)) / sum;
        else
            sums[i] = sums[i-1] + 1.0 / ((double)(i+2)) / sum;
        }
    }

    public static void SetRandomSeed(int seed) {
        r = new Random(seed + 13);
    }

    public void SetDefaultToUniform() {
        this.defaultSelectZipf = false;
    }

    public int Pick() {
        return Pick(this.defaultSelectZipf);
    }

    public int Pick(bool zipfSelect)
    {
        int indx;
        if (zipfSelect)
        {
        double u = r.NextDouble();
        indx = 0;
        while (u > sums[indx]) indx++;
        }
        else
        {
        indx = r.Next(this.numDataElements);
        }
        return indx;
    }
    }
}
    

    
