using System;
using MessageRouting;
using System.Collections;

using Microsoft.Research.Herald.Logger;

namespace Simulator
{

    public interface SimInterface: SendMessage, LocalEvent, TransportLayer
    {
        void ResetSimulator();

        int[] GetNodeMessageReceiveStats();

        void ResetStats();

        Node[] GetNodeArray { get; }

        Node[] GetOverlayNodeArray(OverlayId oid);

        NetTopology GetNetTopology { get; }

        long GetTime();

        long DistanceStat { get; }

        void ProcessAllEvents();

        void ProcessUntilIgnoreEventsOnly();

        void ProcessForTime(long ms);
    }


    /// <summary>
    /// Summary description for Simulator.
    /// </summary>
    public class Sim : SimInterface
    {
        private EventQueue eventq;
        private NetTopology top;
        private Node[] nodes;
        private int nodePtr = 0;
        private long eventsIn = 0;
        private long localEvents = 0;
        private CheckIgnore local;
        //public bool D_FLAG = false;

        public static int[] nodeNumMessagesSentToWire;
        public static int[] nodeNumMessagesReceivedFromWire;

        public long GetNumMessagesSentToWire(Node node) 
        {
            int portNum = node.GetNodeIdAddressBind.GetAddress.Port;
            return nodeNumMessagesSentToWire[portNum];
        }
        public long GetNumMessagesReceivedFromWire(Node node)
        {
            int portNum = node.GetNodeIdAddressBind.GetAddress.Port;
            return nodeNumMessagesReceivedFromWire[portNum];
        }

        public static int[] nodeNumMsgsReceived;

        // this could be replaced by the argument passed in on the command line
        // another alternative is to grow the node array dynamically as 
        // successive calls to CreateNode() are made
        // this constant seems fine for now
        const int MAX_NODES_TO_SIMULATE = 128000; // JD 8-22-03

        public Sim(string netFile, string routesFile, string distancesFile, CheckIgnore ignore)
        {
            eventq = new EventQueue();
            top = new NetTopology(netFile, routesFile, distancesFile);
            
            int totalNumberOfNodes = top.LogicalBase() + MAX_NODES_TO_SIMULATE;
            nodes = new Node[totalNumberOfNodes];
            this.local = ignore;
            
            if (SimParameters.TrackMessageReceipts) 
            {
                nodeNumMsgsReceived = new int[totalNumberOfNodes];
                nodeNumMessagesSentToWire = new int[totalNumberOfNodes];
                nodeNumMessagesReceivedFromWire = new int[totalNumberOfNodes];
                ResetStats();
            }
        }

        public Sim(string asFile, string asRoutesFile, string routerRoutesFile,
            string routerFile, string gatewayFile, CheckIgnore ignore)
        {
            eventq = new EventQueue();
            top = new NetTopology(asFile, asRoutesFile, routerRoutesFile, routerFile, gatewayFile, -1, -1, -1);

            int totalNumberOfNodes = top.LogicalBase() + MAX_NODES_TO_SIMULATE;
            nodes = new Node[totalNumberOfNodes];
            this.local = ignore;

            if (SimParameters.TrackMessageReceipts) 
            {
                nodeNumMsgsReceived = new int[totalNumberOfNodes];
                nodeNumMessagesSentToWire = new int[totalNumberOfNodes];
                nodeNumMessagesReceivedFromWire = new int[totalNumberOfNodes];
                ResetStats();
            }
        }

        public void ResetSimulator()
        {
            //          eventq = new EventQueue();
            //          timeNow = 0L;
        }

        // Return an array of numbers of messages received by each node.
        // The index into the array is the same as the one to use with GetNodeArray.
        public int[] GetNodeMessageReceiveStats()
        {
            return nodeNumMsgsReceived;
        }

        public void ResetStats() 
        {
            if (nodeNumMsgsReceived == null) return;
            
            for (int i = 0; i < nodeNumMsgsReceived.Length; i++) 
            {
                nodeNumMsgsReceived[i] = 0;
                nodeNumMessagesSentToWire[i] = 0;
                nodeNumMessagesReceivedFromWire[i] = 0;
            }
        }

        public Node[] GetNodeArray 
        { 
            get 
            { 
                if (SimParameters.SupportMultipleOverlayNetworks) throw new Exception("Can't use GetNodeArray if multiple overlay networks are allowed");
                return nodes; 
            } 
        }

        public Node[] GetOverlayNodeArray(OverlayId oid)
        {
            Node[] singleNetworkNodes = new Node[top.MaxLogicalNodes()];
            int j = 0;
            for (int i = 0; i < nodePtr; i++)
            {
                if (nodes[i].OverlayId.idGlobal == oid.idGlobal &&
                    nodes[i].OverlayId.idLocal == oid.idLocal)
                { 
                    singleNetworkNodes[j++] = nodes[i];
                }
            }
            return singleNetworkNodes;
        }
        
        public NetTopology GetNetTopology { get { return top; } }

        public Topology GetTopology()
        {
            return top;
        }
        
        public long GetTime() { return eventq.GetTime; }

        long diststat = 0;

        public long DistanceStat
        {
            get { return diststat; }
            set { diststat = 0; }
        }

        //public long[] msgs = new long[100];

            private const int NumMicroSecsForMsgSend = 200;  // Number of microseconds needed to send a message on a machine.

        public bool Send(NodeIdAddressBind source, NodeIdAddressBind dest, MessageRouting.RouteMsg obj)
        {
            if (!source.GetAddress.Failed && !dest.GetAddress.Failed)
            {
                // Drop messages if enabled.
                if (SimParameters.DropMessages && source != dest) {
                    double dice = NetTopology.r.NextDouble();
                    if (dice < SimParameters.DropFraction) {
                        // Drop this message.
                        return false;
                    }
                }

                            NetNode srcNode = this.top.nodes[source.GetAddress.Port];
                            if (srcNode.CurrentNodeTime == eventq.GetTime) {
                                srcNode.NumMsgsSentAtCurrentTime++;
                            }
                            else {
                                srcNode.CurrentNodeTime = eventq.GetTime;
                                srcNode.NumMsgsSentAtCurrentTime = 1;
                            }

                int msgSentD = (srcNode.NumMsgsSentAtCurrentTime * NumMicroSecsForMsgSend + 1000) / 1000;

                int srcIndex = source.GetAddress.Port;
                int destIndex = dest.GetAddress.Port;
                if (srcIndex != destIndex) 
                {
                    nodeNumMessagesSentToWire[srcIndex]++;;
                } 
                

                // Note: during initialization dest.GetNodeId may be null, so we check for that special-case.
                int d = top.Delay(source.GetAddress,dest.GetAddress);
                if (d < 0)
                    return false; // The node no longer exists
                if (srcIndex != destIndex) 
                {
                    nodeNumMessagesReceivedFromWire[destIndex]++;
                } 
                
                d += msgSentD;  // Adjust for time required to send messages on the source node at the current time.
                diststat += d;
                //msgs[((RouteMsg) obj).MessageType]++;
                eventsIn++;
                if (local.IgnoreThis(obj))
                    localEvents++;
                eventq.InsertObject(eventq.GetTime+d,new Event(dest,obj));
                //if (D_FLAG)
                //Console.WriteLine(((SimNodePhysicalAddress)source.GetAddress).ToString() +" -> "+ ((SimNodePhysicalAddress)dest.GetAddress).ToString() +" " + ((RouteMsg)obj).StringMessageType); // mmt xxx
                return true;
            }
            return false;
        }

        // Send a message from the source node to the destination node asynchronously
        public void BeginSend(MessageRouting.NodeIdAddressBind source, 
            MessageRouting.NodeIdAddressBind destination, MessageRouting.RouteMsg obj, ReplyCallback cb)
        {
            bool success = Send(source, destination, obj);
            cb(destination, success, (MessageRouting.RouteMsg) obj);
        }

        public void RaiseEvent(long ms, NodeIdAddressBind me, MessageRouting.RouteMsg obj)
        {
            if (!me.GetAddress.Failed)
            {
                eventsIn++;
                if (local.IgnoreThis(obj))
                    localEvents++;
                eventq.InsertObject(eventq.GetTime+ms,new Event(me,obj));
                //if (E_FLAG)
                //Console.WriteLine(((SimNodePhysicalAddress)me.GetAddress).ToString() +" -> "+ ((SimNodePhysicalAddress)me.GetAddress).ToString()+" " + ((RouteMsg)obj).StringMessageType); // mmt xxx
            }
        }

        public void RemoveEvent(MessageRouting.RouteMsg message) {
            Event dummy = new Event(null, message);
            Event n = (Event) eventq.Remove(dummy);
            eventsIn--;
            if (local.IgnoreThis(n.GetMessage))
                localEvents--;
        }

        public Node CreateNode(Node node, NodePhysicalAddress addr, Object nodeInitData)
        {
            //          if (nodePtr == 0)
            //          {
            //              nodes[nodePtr] = node;
            //              SimNodePhysicalAddress addr = top.NewLogicalNode(nodes[nodePtr]);
            //              nodes[nodePtr++].InitNode(this, this, top, addr, null, param);  // first node
            //              return nodes[0];
            //          }
            //          else
            //          {
            //              nodes[nodePtr] = node;
            //              SimNodePhysicalAddress addr = top.NewLogicalNode(nodes[nodePtr]);           
            //              NodePhysicalAddress seedNeighbors = top.GetClosest(addr);
            //              nodes[nodePtr++].InitNode(this, this, top, addr, nodeInitData, param);
            //              return nodes[nodePtr-1];
            //          }
            addr.AttachLogicalNode(node);
            nodes[nodePtr] = node;
            nodes[nodePtr].InitNode(this, this, top, addr, (NodePhysicalAddress) nodeInitData);
            return nodes[nodePtr++];
        }

        public void ProcessAllEvents()
        {
            for(;;)
            {
                Event n = (Event) eventq.GetNextEvent();
                if( n==null ) break;
                ProcessEvent(n);
            }
        }

        public void ProcessUntilIgnoreEventsOnly()
        {
            for(;;)
            {
                if(eventsIn - localEvents <= 0) break;
                Event n = (Event) eventq.GetNextEvent();
                if( n==null ) break;
                ProcessEvent(n);
            }
        }

        public void ProcessForTime(long ms)
        {
            long endTime = eventq.GetTime + ms;
            //  Console.WriteLine("NOW: " + eventq.GetTime + " END: " + endTime);
            for(;;)
            {
                Event n = (Event) eventq.GetNextEvent();
                if( n==null ) break;
                if( eventq.GetTime>endTime ) {
                    // Oops: This event fires in the future
                    eventq.InsertObject(eventq.GetTime, n);
                    break;
                }
                ProcessEvent(n);
            }
        }

        private void ProcessEvent(Event n)
        {
            eventsIn--;
            if (local.IgnoreThis(n.GetMessage))
                localEvents--;
            else if (SimParameters.TrackMessageReceipts)
                nodeNumMsgsReceived[((SimNodePhysicalAddress)(n.GetDestination.GetAddress)).Port] += 1;
            if (!n.GetDestination.GetAddress.Failed) 
            {
                RouteMsg msg = (RouteMsg) n.GetMessage;
                //              Console.WriteLine("OUT: " + timeNow);
                //if (Pastry.Parameters.DEBUG_PASTRY)
                //    Console.WriteLine("LOW_LEVEL RAISE: " + n.GetDestination.GetNodeId + " " + ((RouteMsg) n.GetMessage).MessageType + " " + eventq.GetTime);

                Log.SetIdentity( n.GetDestination.GetAddress.ToString() );

                if (SimParameters.SupportMultipleOverlayNetworks) {
                    OverlayId msgOid = msg.Overlay;
                    Node node = n.GetDestination.GetAddress.GetLogicalNode(msgOid);
                    Log.SetIdentity( node.GetNodeIdAddressBind.GetAddress.ToString() );
                    node.RaiseEventUpcall(msg);
                } else {
                    Node node = ((SimNodePhysicalAddress)(n.GetDestination.GetAddress)).Node;
                    Log.SetIdentity( node.GetNodeIdAddressBind.GetAddress.ToString() + " (" + node.GetNodeIdAddressBind.GetNodeId.ToString() + ")" );
                    node.RaiseEventUpcall(msg);
                }

                Log.SetIdentity( null );
            }
        }
    }

    
    /// <summary>
    /// An Event that will be inserted into the event queue.
    /// </summary>
    class Event {
        private Object actualMsg;
        private NodeIdAddressBind destination;

        public NodeIdAddressBind GetDestination { get { return destination; } }

        public Object GetMessage { get { return actualMsg; }}

        public Event(NodeIdAddressBind dest, Object obj) {
            actualMsg = obj;
            destination = dest;
        }

        public override bool Equals(Object o) {
            Event e = (Event) o;
            return actualMsg.Equals( e.actualMsg );
        }

        public override int GetHashCode() {
            int hc = actualMsg.GetHashCode();
            return hc;
        }
    }
}
