using System;
using System.Collections;
using System.Diagnostics;

namespace Microsoft.Research.Herald {

    /// <summary>
    /// Splay tree implementation.
    ///
    /// Any element that supports the IComparable interface can be stored in the tree.
    /// 
    /// Multiple elements with equal keys cannot directly be stored in the tree. If you need to store
    /// multiple equal elements, you will need to layer this functionality on top of the Splay tree
    /// (i.e., insert elements into the Splay tree that actually contain arrays of your data elements).
    /// 
    /// This code is derived from Daniel Sleator's public domain implementation of top-down splaying,
    /// available at ftp://ftp.cs.cmu.edu/user/sleator/splaying. This algorithm is described in:
    /// 
    ///    "Self-adjusting Binary Search Trees" Sleator and Tarjan,
    ///    JACM Volume 32, No 3, July 1985, pp 652-686.
    /// 
    /// ****************************************************************************
    /// WARNING: Consult with LCA before incorporating this code into products.
    /// 
    /// Important Notice Concerning Non-Microsoft Components:  
    /// This code, data, and/or other components do NOT belong to Microsoft and are
    /// only licensed to Microsoft for limited research use. Please do not use or
    /// share them with others outside of your research team without talking first
    /// to your legal contact at MSR Legal.  This code, data, and/or other components
    /// must be purged as soon as the work they are related to is completed. If you
    /// have any discussions about a potential technology transfer within Microsoft
    /// that either involves use of this code/data/other components or use of output
    /// derived from them, please contact MSR Legal first.
    /// </summary>
    public class SplayTree {

        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        
        // The root of the tree
        TreeNode    root = null;

        // State used during Splay operation
        TreeNode N = new TreeNode();

        // Special values for min and max
        static TreeNode Min = new TreeNode();
        static TreeNode Max = new TreeNode();

        
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************

        /// <summary>
        /// Insert the given value with the given key into the Splay tree.
        /// If an element with an equal key already exists, its value will be replaced with the new value.
        /// </summary>
        public void Insert(IComparable key, Object dataValue) {
            TreeNode newNode = new TreeNode(key, dataValue);
            if( root==null ) {
                root = newNode;
                return;
            }

            Splay(newNode);

            // The root node is now the closest node to newNode

            if( newNode.CompareTo(root)<0 ) {
                // Key is not in the tree. Insert new node at the root.
                newNode.Left = root.Left;
                newNode.Right = root;
                root.Left = null;
                root = newNode;
            } else if ( newNode.CompareTo(root)>0 ) {
                // Key is not in the tree. Insert new node at the root.
                newNode.Right = root.Right;
                newNode.Left = root;
                root.Right = null;
                root = newNode;
            } else { 
                // Key equals the closest node. Replace the dataValue.
                root.Value = dataValue;
            }
        }


        /// <summary>
        /// Search for a value in the tree that equals the given key.
        /// </summary>
        /// <param name="element">The key to search for</param>
        /// <returns>The value that was found, or null</returns>
        public Object Search(IComparable key) {
            TreeNode node = SearchInternal(key);
            if( node==null ) {
                return null;
            } else {
                return node.Value;
            }
        }


        /// <summary>
        /// Delete the value from the Splay tree that equals the given key.
        /// </summary>
        /// <returns>Returns true if successful, or false if not found</returns>
        public bool Delete(IComparable key) {
            TreeNode    node = SearchInternal(key);
            bool        result;
            
            // Node is not in the tree
            if( node==null ) {
                return false;
            }
            
            Debug.Assert( node==root );
            Debug.Assert( key.CompareTo(node.Key)==0 );
            
            // Remove the root node from the tree
            if( root.Left == null ) {
                root = root.Right;
            } else if( root.Right == null ) {
                root = root.Left;
            } else {
                TreeNode right = root.Right;
                root = root.Left;
                Splay(right);
                root.Right = right;
            }

            return true;
        }

        /// <summary>
        /// Get the minimum key in the tree.
        /// </summary>
        /// <returns>The minimum key, if one exists, or null otherwise</returns>
        public IComparable GetMinKey() {

            // Handle empty tree
            if( root==null ) {
                return null;
            }
            
            // Splay the minimum node to the root
            Splay(Min);
            return root.Key;
        }


        //**********************************************************************************************
        // Private Methods
        //**********************************************************************************************

        /// <summary>
        /// Search for the closest node to element, and return it.
        /// </summary>
        TreeNode SearchInternal(IComparable key) {
            TreeNode        right;
            IComparable     result;
            
            // Tree is empty
            if( root==null ) {
                return null;
            }

            // Splay the closest element to the top
            // TODO: Removing this memory allocation might improve search performance
            TreeNode dummy = new TreeNode(key, null);
            Splay(dummy);

            if( dummy.CompareTo(root)==0 ) { 
                // Found the element.
                return root;
            }

            // It wasn't there
            return null;
        }


        /// <summary>
        /// Bring the node closest to splayNode up to the root of the tree.
        /// </summary>
        /// <param name="splayNode">The node to splay to the root of the tree</param>
        /// <param name="forceMin">If true, ignore the splayNode parameter and just splay the minimum</param>
        /// <param name="forceMax">If true, ignore the splayNode parameter and just splay the maximum</param>
        void Splay(TreeNode splayNode) {
            TreeNode l, r, y, t;

            Debug.Assert( root!=null );
            t = root;

            N.Left = N.Right = null;
            l = r = N;

            for (;;) {

                if( SplayCompare(splayNode, t)<0 ) {
                    // Search left subtree
                    
                    if( t.Left == null ) {
                        // Hit the bottom of the tree
                        break;
                    }
                    
                    if( SplayCompare(splayNode, t.Left)<0 ) {

                        // Single Right Rotation:
                        //      t              y
                        //     / \            / \
                        //    y   C   -->    A   t
                        //   / \                / \
                        //  A   B              B   C

                        y = t.Left;
                        t.Left = y.Right;
                        y.Right = t;

                        // Current node is now y
                        t = y;

                        if( t.Left == null ) {
                            // Hit the bottom of the tree
                            break;
                        }
                    }

                    // Link right
                    r.Left = t;
                    r = t;
                    t = t.Left;

                } else if( SplayCompare(splayNode,t)>0 ) {
                    // Search right subtree

                    if( t.Right == null ) {
                        // Hit the bottom of the tree
                        break;
                    }

                    if( SplayCompare(splayNode,t.Right)>0 ) {

                        // Single Left Rotation:
                        //      t               y
                        //     / \             / \
                        //    A   y    -->    t   C
                        //       / \         / \
                        //      B   C       A   B

                        y = t.Right;
                        t.Right = y.Left;
                        y.Left = t;

                        // Current node is now y
                        t = y;

                        if( t.Right == null ) {
                            // Hit the bottom of the tree
                            break;
                        }
                    }

                    // Link left
                    l.Right = t;
                    l = t;
                    t = t.Right;
                } 
                else {
                    break;
                }
            }

            // Assemble
            l.Right = t.Left;
            r.Left = t.Right;
            t.Left = N.Right;
            t.Right = N.Left;
            root = t;
        }

        
        /// <summary>
        /// Compare splayNode with compNode, while correctly handling Min and Max.
        /// This is used by the Splay() method.
        /// </summary>
        /// <param name="splayNode">The node we are currently splaying</param>
        /// <param name="compNode">The node to compare splayNode with</param>
        /// <returns>-1 if splayNode is less, 0 if they are equal, 1 if splayNode is greater</returns>
        int SplayCompare( TreeNode splayNode, TreeNode compNode ) {
            if( splayNode==Min ) return -1;
            if( splayNode==Max ) return 1;
            return splayNode.CompareTo(compNode);
        }


        //**********************************************************************************************
        // Nested Classes
        //**********************************************************************************************
        
        /// <summary>
        /// A node in the Splay tree.
        /// </summary>
        class TreeNode : IComparable {
            TreeNode        left, right;
            IComparable     key;
            Object          dataValue;
 
            public TreeNode Left {
                get { return left; }
                set { left = value; }
            }	

            public TreeNode Right {
                get { return right; }
                set { right = value; }
            }	

            public IComparable Key {
                get { return key; }
                set { key = value; }
            }

            public Object Value {
                get { return dataValue; }
                set { dataValue = value; }
            }

            // Construct an empty tree node
            public TreeNode() {
                this.left = this.right = null;
            }

            // Construct a tree node with the given key and value
            public TreeNode(IComparable key, Object dataValue) {
                Debug.Assert(key!=null);
                this.key = key;
                this.dataValue = dataValue;
                this.left = null;
                this.right = null;
            }

            // Compare ordering of two tree nodes
            public int CompareTo( Object o ) {
                TreeNode n = (TreeNode) o;
                Debug.Assert( key!=null && n.key!=null );
                return key.CompareTo(n.key);
            }

            // Determine equality of two tree nodes
            public override bool Equals( Object o ) {
                return (CompareTo(o) == 0);
            }
        }

    }

}