using System;
using MessageRouting;
using System.Collections;
using System.Threading;
using System.Diagnostics;

namespace Simulator
{
    /// <summary>
    /// Simulator based on Geogia-Tech network topology.
    /// This simulator also models queuing delays at nodes.
    /// </summary>
    public class Sim3 : SimInterface
    {
        private EventQueue eventq;
        private NetTopology top;
        private Node[] nodes;
        private int nodePtr = 0;
        private long eventsIn = 0;
        private long localEvents = 0;
        private CheckIgnore local;
        private Thread simThread;
        //public bool D_FLAG = false;

        private long[] nodeTimes;
        private const long UpcallServiceTime = 1;  // Time to process an event upcall.

        public static int[] nodeNumOpns;
        public static long[] nodeSumOpns;
        public static int[] nodeMaxOpns; 

        public static int[] nodeNumMsgsReceived;

        public Sim3(string netFile, string routesFile, string distancesFile, CheckIgnore ignore) :
            this(netFile, routesFile, distancesFile, ignore, -1, -1, -1) 
        {
        }

        public Sim3(string netFile, string routesFile, string distancesFile, CheckIgnore ignore,
            int transitTransitBand, int transitStubBand, int stubStubBand)
        {
            eventq = new EventQueue();
            top = new NetTopology(netFile, routesFile, distancesFile, transitTransitBand, transitStubBand, stubStubBand);
            nodes = new Node[top.MaxLogicalNodes()];
            this.local = ignore;

            nodeTimes = new long[top.MaxLogicalNodes()];
            nodeNumOpns = new int[top.MaxLogicalNodes()];
            nodeSumOpns = new long[top.MaxLogicalNodes()];
            nodeMaxOpns = new int[top.MaxLogicalNodes()];
#if SeparateSimThread
        SimulatorThread st = new SimulatorThread(this);
        simThread = new Thread(new ThreadStart(st.RunSim));
        simThread.Start();
#endif
            if (SimParameters.TrackMessageReceipts) 
            {
                nodeNumMsgsReceived = new int[top.MaxLogicalNodes()];
                ResetStats();
            }
        }

        public void ResetSimulator()
        {
            //          eventq = new EventQueue();
            //          timeNow = 0L;
        }

        // Return an array of numbers of messages received by each node.
        // The index into the array is the same as the one to use with GetNodeArray.
        public int[] GetNodeMessageReceiveStats()
        {
            return nodeNumMsgsReceived;
        }
        public long GetNumMessagesSentToWire(Node node) 
        {
            Debug.Assert(false);
            return 0;
        }
        public long GetNumMessagesReceivedFromWire(Node node)
        {
            Debug.Assert(false);
            return 0;
        }

        public void ResetStats() 
        {
            if (nodeNumMsgsReceived == null) return;

            for (int i = 0; i < nodeNumMsgsReceived.Length; i++) 
            {
                nodeNumMsgsReceived[i] = 0;
            }
        }

        public Node[] GetNodeArray 
        { 
            get 
            {
                if (SimParameters.SupportMultipleOverlayNetworks) throw new Exception("Can't use GetNodeArray if multiple overlay networks are allowed");
                return nodes; 
            } 
        }

        public Node[] GetOverlayNodeArray(OverlayId oid)
        {
            Node[] singleNetworkNodes = new Node[top.MaxLogicalNodes()];
            int j = 0;
            for (int i = 0; i < nodePtr; i++)
            {
                if (nodes[i].OverlayId.idGlobal == oid.idGlobal &&
                    nodes[i].OverlayId.idLocal == oid.idLocal) 
                {
                    singleNetworkNodes[j++] = nodes[i];
                }
            }
            return singleNetworkNodes;
        }
        
        public NetTopology GetNetTopology { get { return top; } 
        }

        public Topology GetTopology()
        {
            return top;
        }
        
        public long GetTime() 
        { 
            long t = eventq.GetTime;
            return t; 
        }

        long diststat = 0;

        public long DistanceStat
        {
            get { return diststat; }
            set { diststat = 0; }
        }

        // Find out what time it is on a specific node.
        public long GetNodeTime(NodeIdAddressBind n) 
        {
            return nodeTimes[((SimNodePhysicalAddress)(n.GetAddress)).Port];
        }

        // Let the node indicate what time it has.
        public void SetNodeTime(NodeIdAddressBind n, long nTimeNow) 
        {
            nodeTimes[((SimNodePhysicalAddress)(n.GetAddress)).Port] = nTimeNow;
        }

        //public long[] msgs = new long[100];

        public bool Send(NodeIdAddressBind source, NodeIdAddressBind dest, MessageRouting.RouteMsg obj)
        {
            RouteMsg msg = (RouteMsg) obj;
            if (!source.GetAddress.Failed)
            {
                // Note: nodeTimes is guaranteed to be bigger than eventq.GetTime because
                // of how we manage time recording before upcalls. 
                long tNode = nodeTimes[((SimNodePhysicalAddress)(source.GetAddress)).Port];
                long tNow = eventq.GetTime;
                if (tNode > tNow) tNow = tNode;
                int d = 0;
                // Compute a non-zero delay if the source and dest aren't the same.
                // Note: during initialization dest.GetNodeId may be null, so we check for that special-case.
                NodePhysicalAddress nextDestAddr = dest.GetAddress;
                if (SimParameters.ModelIntermediateLinkQueuing)
                {
                    //              Console.WriteLine("Send({0},{1},{2},{3}):  ", source.GetAddress,dest.GetAddress, msg.MessageSize, tNow); // xxx
                    d = top.OneHopDelay(source.GetAddress,dest.GetAddress, msg.MessageSize, tNow, out nextDestAddr);
                }
                else
                {
                    d = top.Delay(source.GetAddress,dest.GetAddress, msg.MessageSize, tNow);
                    //              Console.WriteLine("Send({0},{1},{2},{3}):  d: {4}", source.GetAddress,dest.GetAddress, msg.MessageSize, tNow, d); // xxx mmt
                }
                if (d < 0)
                    return false; // The node no longer exists
                diststat += d;
                //msgs[msg.MessageType]++;
                eventsIn++;
                if (local.IgnoreThis(msg))
                    localEvents++;
                //          long t = eventq.GetTime+d;
                long t = tNow + d;
                Event e;
                if (SimParameters.ModelIntermediateLinkQueuing) e = new Event3(source.GetAddress, dest, nextDestAddr, msg, tNow);
                else e = new Event(dest, msg);
                eventq.InsertObject(t, e);
                //if (D_FLAG)
                //Console.WriteLine(((SimNodePhysicalAddress)source.GetAddress).ToString() +" -> "+ ((SimNodePhysicalAddress)dest.GetAddress).ToString() +" " + ((RouteMsg)obj).StringMessageType); // mmt xxx
                return true;
            }
            return false;
        }

        // Send a message from the source node to the destination node asynchronously
        public void BeginSend(MessageRouting.NodeIdAddressBind source, 
            MessageRouting.NodeIdAddressBind destination, MessageRouting.RouteMsg obj, ReplyCallback cb) {
            bool success = Send(source, destination, obj);
            cb(destination, success, (MessageRouting.RouteMsg) obj);
        }

        public void RaiseEvent(long ms, NodeIdAddressBind me, MessageRouting.RouteMsg obj)
        {
            if (!me.GetAddress.Failed)
            {
                eventsIn++;
                if (local.IgnoreThis(obj))
                    localEvents++;
                long tNow = eventq.GetTime;
                long t= tNow + ms;
                Event e;
                if (SimParameters.ModelIntermediateLinkQueuing) e = new Event3(me.GetAddress, me, me.GetAddress, obj, tNow);
                else e = new Event(me, obj);
                eventq.InsertObject(t, e);
                //if (E_FLAG)
                //Console.WriteLine(((SimNodePhysicalAddress)me.GetAddress).ToString() +" -> "+ ((SimNodePhysicalAddress)me.GetAddress).ToString()+" " + msg.StringMessageType); // mmt xxx
            }
        }

        public void RemoveEvent(MessageRouting.RouteMsg message) {
            Event dummy = new Event(null, message);
            eventq.Remove(dummy);
        }

        public Node CreateNode(Node node, NodePhysicalAddress addr, Object nodeInitData)
        {
            addr.AttachLogicalNode(node);
            nodes[nodePtr] = node;
            nodes[nodePtr].InitNode(this, this, top, addr, (NodePhysicalAddress) nodeInitData);
            return nodes[nodePtr++];
        }

        public void ProcessAllEvents() {
            for(;;) {
                Event n = (Event) eventq.GetNextEvent();
                if( n==null ) break;
                ProcessEvent(n);
            }
        }

        public void ProcessUntilIgnoreEventsOnly() {
            for(;;) {
                if(eventsIn - localEvents <= 0) break;
                Event n = (Event) eventq.GetNextEvent();
                if( n==null ) break;
                ProcessEvent(n);
            }
        }

        public void ProcessForTime(long ms) {
            long endTime = eventq.GetTime + ms;
            //  Console.WriteLine("NOW: " + eventq.GetTime + " END: " + endTime);
            for(;;) {
                Event n = (Event) eventq.GetNextEvent();
                if( n==null ) break;
                if( eventq.GetTime>endTime ) {
                    // Oops: This event fires in the future
                    eventq.InsertObject(eventq.GetTime, n);
                    break;
                }
                ProcessEvent(n);
            }
        }

        private void ProcessEvent(Event n)
        {
            eventsIn--;
            if (local.IgnoreThis(n.GetMessage))
                localEvents--;
            else if (SimParameters.TrackMessageReceipts)
                nodeNumMsgsReceived[((SimNodePhysicalAddress)(n.GetDestination.GetAddress)).Port] += 1;

            if (!n.GetDestination.GetAddress.Failed) 
            {
                long tNow = eventq.GetTime;
                if (SimParameters.ModelIntermediateLinkQueuing)
                {
                    NodeIdAddressBind dest = n.GetDestination;
                    Event3 n3 = (Event3)n;
                    if (n3.GetNextDestAddr != dest.GetAddress)
                    {
                        // This is a router event: a message has arrived at an intermediate IP router node
                        // and needs to be forwarded to the next node on its path.
                        RouteMsg msg = (RouteMsg) n.GetMessage;
                        NodePhysicalAddress nextDestAddr;
                        int d = top.OneHopDelay(n3.GetNextDestAddr, dest.GetAddress, msg.MessageSize, tNow, out nextDestAddr);
                        //              Console.WriteLine("Forward({0},{1}): d = {2},  nextDest = {3}", n3.GetNextDestAddr, dest.GetAddress, d, nextDestAddr); // xxx
                        if (d > 0)
                        {
                            eventsIn++;
                            long t = tNow + d;
                            Event3 e = new Event3(n3.SrcAddr, dest, nextDestAddr, msg, n3.GetStartTime);
                            eventq.InsertObject(t, e);
                        }
                        return;
                    }
                    else if (n3.SrcAddr != dest.GetAddress)
                    {
                        //              Console.WriteLine("{0} -> {1}  tStart: {2},  d: {3}", n3.SrcAddr, dest.GetAddress, n3.GetStartTime, tNow - n3.GetStartTime); // xxx mmt
                    }
                }

                long nTime = nodeTimes[((SimNodePhysicalAddress)(n.GetDestination.GetAddress)).Port];
                if (nTime > tNow) 
                {
                    int indx = ((SimNodePhysicalAddress)(n.GetDestination.GetAddress)).Port;
                    long del = nTime - tNow;
                    nodeNumOpns[indx]++;
                    nodeSumOpns[indx] += del;
                    if (nodeMaxOpns[indx] < del) nodeMaxOpns[indx] = (int)del;

                    tNow = eventq.GetTime;
                    nodeTimes[((SimNodePhysicalAddress)(n.GetDestination.GetAddress)).Port] = nTime + UpcallServiceTime;
                }
                else 
                {
                    int indx = ((SimNodePhysicalAddress)(n.GetDestination.GetAddress)).Port;
                    nodeNumOpns[indx]++;

                    nodeTimes[((SimNodePhysicalAddress)(n.GetDestination.GetAddress)).Port] = tNow + UpcallServiceTime;
                }
                RouteMsg routeMsg = (RouteMsg) n.GetMessage;
                if (SimParameters.SupportMultipleOverlayNetworks)
                {
                    OverlayId msgOid = routeMsg.Overlay;
                    Node nd = n.GetDestination.GetAddress.GetLogicalNode(msgOid); // xxx
                    //          Console.WriteLine("Deliver event to {0} on overlay net ({1},{2})", nd.GetNodeIdAddressBind.GetAddress, msgOid.idGlobal, msgOid.idLocal); // xxx
                    n.GetDestination.GetAddress.GetLogicalNode(msgOid).RaiseEventUpcall(routeMsg);
                }
                else
                    ((SimNodePhysicalAddress)(n.GetDestination.GetAddress)).Node.RaiseEventUpcall(routeMsg);
            }
        }
    }
    
    class Event3: Event
    {
        private NodePhysicalAddress srcAddr;
        private NodePhysicalAddress nextDestAddr;
        private long startTime;

        public NodePhysicalAddress GetNextDestAddr { get { return nextDestAddr; } }

        public long GetStartTime { get {return startTime;}}

        public NodePhysicalAddress SrcAddr { get { return srcAddr; } }

        public Event3(NodePhysicalAddress srcAddr, NodeIdAddressBind dest, NodePhysicalAddress nextDestAddr, Object obj, long tNow): base(dest, obj)
        {
            this.nextDestAddr = nextDestAddr;
            this.startTime = tNow;
            this.srcAddr = srcAddr;
        }
    }

}
