using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// This class is used to identify the destination of CLB messages.
    /// </summary>
[Serializable, XmlRootAttribute("CLBId",
    Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class CLBId : MessageRouting.NodeId {

        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        private const char          SEPCHAR = '!';
        

        //**********************************************************************************************
        // Public Members (for serialization convenience)
        //**********************************************************************************************
        public string       prefixConstraint;
        public NumericId    randomId;
        public string       wholeString;
        
        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        /// <summary>
        /// Default constructor for serialization.
        /// </summary>
        public CLBId() {}

        /// <summary>
        /// Generate a CLB identifier with the given prefix constraint. Hash the string identifier
        /// to obtain the random identifier.
        /// </summary>
        public CLBId( string prefixConstraint, string identifier ) {
            this.prefixConstraint = prefixConstraint;
            this.randomId = new NumericId();
            randomId.GenerateFromString( identifier );

            this.wholeString = prefixConstraint + "!" + identifier;
        }
        
        /// <summary>
        /// Generate a CLB identifier with the given prefix constraint and the given
        /// random identifier.
        /// </summary>
        public CLBId( string prefixConstraint, NumericId randomId ) {
            this.prefixConstraint = prefixConstraint;
            this.randomId = randomId;
        }
        
        /// <summary>
        /// Generate a CLB identifier with no prefix constraint and the given
        /// random identifier.
        /// </summary>
        public CLBId( NumericId randomId ) {
            this.prefixConstraint = "";
            this.randomId = randomId;
        }
        
        /// <summary>
        /// Generate a CLB identifier from a single string in the "CLB-Domain!CLB-Suffix" format.
        /// </summary>
        /// <param name="source">Address of the node sending the message</param>
        /// <param name="background">Is this a background message? (simulator only)</param>
        /// <param name="clbString">A string of the form "constraint!identifier" that identifies the destination.
        /// The constraint may be the empty string but the identifier may not.</param>
        public CLBId( string clbString ) {
            // Check the CLB string
            int firstSepIndex = clbString.IndexOf(SEPCHAR);
            int lastSepIndex = clbString.LastIndexOf(SEPCHAR);
            Debug.Assert( firstSepIndex>=0 && firstSepIndex==lastSepIndex, "clbString must contain exactly one separator" );
            Debug.Assert( firstSepIndex<clbString.Length-1, "clbString must not terminate with separator" );
                
            // Set the constraint based on the first half of the CLB string
            this.prefixConstraint = clbString.Substring(0, firstSepIndex);

            // Set the numeric ID based on the second half of the CLB string
            string s = clbString.Substring(firstSepIndex+1, clbString.Length-firstSepIndex-1);
            this.randomId = new NumericId();
            this.randomId.GenerateFromString(s);

            this.wholeString = clbString;
        }
    
    
        //**********************************************************************************************
        // Properties
        //**********************************************************************************************
        
        public string PrefixConstraint {
            get { return prefixConstraint; }
        }
        
        public NumericId NumericId {
            get { return randomId; }
        }
        
        
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************

        public override string ToString() {
            Debug.Assert( this.wholeString!=null );
            return this.wholeString;
        }

        // Test two CLBIds for equality
        public bool Equals( CLBId c ) {
            return ( this.prefixConstraint==c.prefixConstraint
                && this.randomId.Equals(c.randomId) );
        }
        
        // Overridden equals (used for Hashtables)
        public override bool Equals(Object obj) {
            if (!(obj is CLBId))
                return false;
            return Equals((CLBId)obj);
        }

        // Overridden hashcode (used for Hashtables)
        public override int GetHashCode() {
            int hash1 = prefixConstraint.GetHashCode();
            int hash2 = randomId.GetHashCode();
            return hash1+hash2;
        }
    }
    
    
    /// <summary>
    /// An abstract SkipNet message that is routed using Constrained Load Balancing.
    /// Note that the routing path traversed by this message is loop-free, in the sense
    /// that CheckContinueRouting() is never called at a node more than once.
    /// Unfortunately, the logic needed to achieve this is a bit of a nightmare.
    /// Should be rewritten for clarity some day.
    /// </summary>
    [Serializable, XmlRootAttribute("RouteByCLBMsg",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class RouteByCLBMsg : SkipNet.RouteMsg {

        //**********************************************************************************************
        // Constants
        //**********************************************************************************************

        [XmlType(Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, TypeName="CLBPhase")]
        public enum Phase   { First,        // Haven't yet switched direction at this level
                              Second,       // Have switched direction at this level
                              Best          // Have found the destination node
                            }

        public const string CATEGORY = "CLB";
        
        
        //**********************************************************************************************
        // Public Members (public for serialization convenience)
        //**********************************************************************************************
        
        // The destination
        private CLBId               clbId;
        public string               constraint;
        public NumericId            destID;

        // The routing state
        public int                  currentH;
        public NodeIdAddressBind    start, best;
        public NumericId            bestDiff;
        public Direction            dir;
        public Phase                phase;
        public bool                 committingMode;
        public bool                 failed;
        ArrayList                   pendingCommits; 

        // Hack added for SOSP experiments
        public bool                 forceRTable;

        // Statistics:
        // When the message arrives at its destination, cTableHops/(cTableHops+rTableHops-1) gives the
        // percentage of hops that were taken with the C-Table.
        public int                 cTableHops;
        public int                 rTableHops;
        
        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        public RouteByCLBMsg() : base() {
            Init();
        }
        
        /// <summary>
        /// Constructor for a message that is routed by Constrained Load Balancing.
        /// </summary>
        /// <param name="source">Address of the node sending the message</param>
        /// <param name="background">Is this a background message? (simulator only)</param>
        /// <param name="clbId">This CLB ID defines the destination of the message</param>
        public RouteByCLBMsg( NodeIdAddressBind source, CLBId clbId, OverlayMsg payload )
            : base(source, false, payload)
        {
            this.clbId = clbId;
            this.constraint = clbId.PrefixConstraint;
            this.destID = clbId.NumericId;

            Init();
        }

        /// <summary>
        /// Initialize the routing state.
        /// </summary>
        private void Init() {
            currentH = -1;
            start = null;
            best = null;
            dir = Direction.Right;
            bestDiff = null;
            phase = Phase.First;
            cTableHops = 0;
            rTableHops = 0;
            forceRTable = false;
            committingMode = false;
            failed = false;
            pendingCommits = new ArrayList();
        }
        

        //**********************************************************************************************
        // Properties
        //**********************************************************************************************
        
        [XmlElementAttribute("Destination",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public override MessageRouting.NodeId Destination {
            get { return clbId; }
            set { clbId = (CLBId) value; }
        }

        [XmlArray( ElementName="PendingCommits",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="PendingNode", Type=typeof(NodeIdAddressBind),
             IsNullable=true, Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public ArrayList PendingCommits {
            get { return pendingCommits; }
            set { pendingCommits = value; }
        }


        //**********************************************************************************************
        // Overriding Members
        //**********************************************************************************************

        /// Called to find the next hop address towards the destination.
        public override NodeIdAddressBind GetNextHop( Node localNode ) {
            NodeIdAddressBind   localAddrBind = localNode.NodeIdAddressBind;
            NodeIdAddressBind   nextHop=null;
            bool                continueRouting;
            
            // If routing has failed, drop the message
            if( failed ) {
                return null;
            }
            
            // If the local node is not even within the prefix constraint, route
            // towards the destination prefix
            if( ! IsNodeWithinConstraint(localAddrBind) ) {
                return RouteTowardsPrefix(localNode);
            }

            // Special case: Routing directly to this node's numeric ID
            // (this is used during CLB Join)
            if( destID.Equals(localAddrBind.NumericId) ) {
                return localAddrBind;
            }

            // The local numericId matches 'matchingHeight' digits with the destination numericId
            int matchingHeight = destID.LongestCommonPrefixLength( localAddrBind.NumericId );

            // If the matchingHeight has increased, reset the state for this level
            if( matchingHeight > this.currentH ) {
                int prevH = this.currentH;
                this.currentH = matchingHeight;
                this.start = localAddrBind;
                this.best = null;               // 'best' will be set below
                this.phase = Phase.First;
                this.dir = Direction.Right;

                Debug.Assert( committingMode==false );

                if( prevH<0 ) {
                    // We've just finished routing by name ID. There should be no pending commits,
                    // and we should just begin the numeric ID routing phase below.
                    Debug.Assert( pendingCommits.Count==0 );
                } else {
                    // We can now commit at the nodes on the pending list.
                    Debug.Assert( pendingCommits.Count>0 );

                    if( pendingCommits.Count==1 ) {
                        // Special case: The local node is the only one on the pending list.
                        // We can just commit right away and clear the pending list.
                        // No need to switch to committing mode.
                        Debug.Assert( localAddrBind.Equals(pendingCommits[0]) );
                        pendingCommits.Clear();
                        Log.Write(CATEGORY, LogLvl.Med, "Committing at {0}", localAddrBind);
                        continueRouting = payload.CheckContinueRouting(localNode);
                        if( ! continueRouting ) return null;
                    } else {
                        // Switch to committing mode and retrace the path.
                        committingMode = true;
                        nextHop = (NodeIdAddressBind) pendingCommits[0];
                        return nextHop;
                    }
                }
            }

            // 'Committing mode' means either:
            //  - we are retracing our steps along the pending path (after loops have been removed from it)
            //  - we are routing towards the 'best' node
            if( committingMode ) {
                if( phase!=Phase.Best ) {
                    pendingCommits.RemoveAt(0);
                    if( pendingCommits.Count==0 ) {
                        committingMode = false;
                    } else {
                        nextHop = (NodeIdAddressBind) pendingCommits[0];
                        // TODO: Verify that nextHop is still a neighbour in some table
                    }
                } else {
                    // In the best phase, we just route using the R-Table.
                    // A bit of a kludge, this.
                    nextHop = RouteUsingRTable(localNode);
                    this.rTableHops++;
                }
            }

            if( ! committingMode ) {
                // Check if this node is 'better' (i.e. a closer match to the destination)
                // than any node seen before at this level
                NumericId diff = destID.Difference( localAddrBind.NumericId );
                if( best==null || diff.CompareTo(bestDiff)<0 ) {
                    this.best = localAddrBind;
                    this.bestDiff = diff;
                }

                // The local node is within the destination prefix -- try to use the CTable
                // to route towards the destination random ID.
                if( Parameters.cTableEnabled && (! this.forceRTable) ) {
                    nextHop = localNode.CTable.GetNextHop(destID);

                    // As long as the entry from the CLB Table is not the local node, and matches
                    // the destination prefix, we can use it as the next hop
                    if(    ( ! nextHop.Equals( localAddrBind ) )
                        && ( IsNodeWithinConstraint(nextHop) ) )
                    {
                        this.cTableHops++;
                        return nextHop;
                    }
                }
        
                // The CLB Table wasn't much help. We must route using the R Table.
                nextHop = RouteUsingRTable(localNode);
                this.rTableHops++;
            }

            return nextHop;
        }
        
        /// <summary>
        /// After each time SkipNet tries to send this message to another node,
        /// this callback notifies the message whether the send was successful or not.
        /// </summary>
        public override void SendCallback( Node localNode, NodeIdAddressBind dest, bool success ) {
            if( ! success ) {
                if( currentH<0 ) {
                    // We're still routing by name ID and haven't yet reached the CLB domain
                } else {
                    // We're routing by numeric ID. Now we need to pay careful attention to failures.

                    // If we can't send to the start node (i.e. the first node we encountered in this ring),
                    // that's disasterous
                    Debug.Assert( start!=null );
                    if( dest.Equals(start) ) {
                        failed = true;
                    }

                    // When in committing mode, if we fail to retrace the pending-path, that's disasterous
                    if( committingMode ) {
                        failed = true;
                    }
                }
            }
        }

        /// <summary>
        /// This is called when SkipNet suggests we should call CheckContinueRouting() on the payload.
        /// This CLB message must decide for itself whether or not to call CheckContinueRouting(),
        /// in order to avoid routing loops.
        /// </summary>
        public override bool CallCCR( Node localNode ) {
            if(    currentH<0           // Still routing by name ID
                || committingMode       // Routing by numeric ID, in committing mode
            ) {
                // Actually call CheckContinueRouting().
                Log.Write(CATEGORY, LogLvl.Med, "Committing at {0}", localNode.NodeIdAddressBind);
                bool continueRouting = localNode.MessageCCRHook( payload );
                return continueRouting;
            } else {
                // We're routing by numeric ID, but we're in pending mode.
                // In pending mode, we just add the node to our pending list.
                // CheckContinueRouting() is not called until we later commit.
                Log.Write(CATEGORY, LogLvl.Med, "Pending at {0}", localNode.NodeIdAddressBind);
                pendingCommits.Add( localNode.NodeIdAddressBind );
                return true;
            }
        }

        
        //**********************************************************************************************
        // Private Methods
        //**********************************************************************************************

        // Determines if a node's name ID is within the constraint defined by the given prefix string
        private bool IsNodeWithinConstraint( NodeIdAddressBind node ) {
            return node.NodeId.IsWithinPrefix( constraint );
        }

        // Route towards nodes with the given name ID prefix
        private NodeIdAddressBind RouteTowardsPrefix( Node localNode ) {
            NodeId prefixID = new NodeId(constraint);
            return localNode.RouteByName(prefixID, RouteByNameOption.PrefixMatch, (! forceRTable));
        }
        
        // Route using the R-Table
        private NodeIdAddressBind RouteUsingRTable( Node localNode ) {
            NodeIdAddressBind   localAddrBind = localNode.NodeIdAddressBind;
            LeafSet             ls = localNode.RoutingTable.Neighbours[ currentH ];
            
            GetNextHop:

            // Always clear the pending-commit path at the start node
            if( localAddrBind.Equals(start) ) {
                // This is what ensures that the pending path contains no loops
                Log.Write(CATEGORY, LogLvl.Med, "Clearing pending nodes");
                pendingCommits.Clear();
            }

            // Special case: If we're routing towards the previously-found best node,
            // just use route-by-name within this ring
            if( phase==Phase.Best ) {
                committingMode = true;

                NodeIdAddressBind nextHop = ls.RouteByName(best.NodeId, RouteByNameOption.Normal);
                if( ! nextHop.Equals(localAddrBind) ) {
                    Log.Write(CATEGORY, LogLvl.Med, "Hopping to best node");
                }
                if( nextHop.Equals(localAddrBind) && ! localAddrBind.Equals(best) ) {
                    // Can't reach the previously determined 'best' node. Just drop the message.
                    return null;
                } else {
                    return nextHop;
                }
            }

            // Pseudocode for the algorithm to determine the next hop:
            //  - Define S to be the set of all neighbours at the current level and direction
            //  - Define T to be the subset of S that stay within the constaint and do not wrap
            //  - If T is empty
            //    - Either reverse direction (if in First Phase) or halt (if in Second Phase)
            //  - Let U be the subset of nodes in T that have a better numeric ID prefix match than currentH
            //  - If U is non-empty
            //    - Let u be the node in U with best distance
            //    - Commit any pending non-committed hops
            //    - Make a 'committing hop' to node u
            //  - Else
            //    - Let t be the node in T with best distance
            //    - If t's distance is better than the current best distance
            //      - Make a 'non-committing hop' to node t
            //    - Else
            //      - If T = S (i.e., all neigbours are within the constraint)
            //        - Make a 'non-committing hop' to the farthest node in S
            //      - Else
            //        - Either reverse direction (if in First Phase) or halt (if in Second Phase)

            // Find set S: the neighbours of the local node at the current level and direction
            ArrayList S;
            if( dir==Direction.Left ) {
                S = ls.LeftHalf.GetAllNbrs();
            } else {
                S = ls.RightHalf.GetAllNbrs();
            }
            
            // Find set T: the neighbours in set S that stay within the constaint and don't wrap
            ArrayList T = new ArrayList( S.Count );
            for( int i=0; i<S.Count; i++ ) {
                NodeIdAddressBind s = (NodeIdAddressBind) S[i];
                if( (! IsNodeWithinConstraint(s)) || Wrapped(localAddrBind.NodeId, s.NodeId) ) {
                    break;
                }
                T.Add(s);
            }

            // Check if no neighbours stay within the constraint and don't wrap
            if( T.Count==0 ) {
                goto SwitchOrHalt;
            }
            
            // Search set T for: 
            //  - u, the neighbour in set T that has the best numeric prefix match
            //    (but this prefix match _must_ be better than the current height)
            //  - t, the neighbour in set T that has the best distance
            //    (but this distance _must_ be better than the current distance)
            NodeIdAddressBind u=null, t=null;
            int uHeight = this.currentH;
            NumericId tDist = this.bestDiff;
            for( int i=0; i<T.Count; i++ ) {
                NodeIdAddressBind n = (NodeIdAddressBind) T[i];
                int h = destID.LongestCommonPrefixLength(n.NumericId);
                NumericId d = destID.Difference(n.NumericId);
                if( h>uHeight ) {
                    u = n;
                    uHeight = h;
                }
                if( d.CompareTo(tDist)<0 ) {
                    t = n;
                    tDist = d;
                }
            }

            // If u exists, just jump to u
            if( u!=null ) {
                return u;
            }

            // No neighbour in set T has a better numeric prefix match. 

            // If t exists, just jump to t
            if( t!=null ) {
                return t;
            }
            
            // There is no reason to prefer any of the neighbours.

            // If all neighbours stay within the constraint, then just go to
            // the farthest neighbour
            if( T.Count==S.Count ) {
                return (NodeIdAddressBind) S[ S.Count-1 ];
            }
            
            // Invoked when we either have to switch direction or halt (i.e. go to the best node)
            SwitchOrHalt:
            if( phase==Phase.First ) {
                this.phase = Phase.Second;
                this.dir = Direction.Left;      // Switch direction
            } else {
                phase = Phase.Best;
            }

            if( localAddrBind.Equals(start) ) {
                goto GetNextHop;                // Still at the start node. Just start again.
            } else {
                return start;                   // Hop to the start node
            }
        }

        // Determine if the step from 'local' to 'next' in the current direction
        // wrapped around
        private bool Wrapped( NodeId local, NodeId next ) {
            if( dir==Direction.Right ) {
                return ( next.CompareTo(local) <= 0 );
            } else {
                return ( next.CompareTo(local) >= 0 );
            }
        }
    }
    

}
