using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// This class stores a table of all CLB Neighbours.
    /// </summary>
    public class CLBTable {
    
        /// <summary>
        /// This class defines one row in the CLBTable
        /// </summary>
        public class CLBTableRow {
            public CLBSet               clbSet;
            public string               clbDomain;

            public CLBTableRow( Node localNode, string clbDomain ) {
                this.clbDomain = clbDomain;
                this.clbSet = new CLBSet(localNode, localNode.NodeIdAddressBind, Parameters.clbTableNbrListSize);
            }
        }
        
        
        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        Node                localNode;
        NodeIdAddressBind   localAddrBind;
        CLBTableRow[]       rows;
        
        
        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        public const string CATEGORY = "CLBTable";


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        public CLBTable( Node localNode ) {
            this.localNode = localNode;
            this.localAddrBind = localNode.NodeIdAddressBind;

            // Get the list of all CLB domains that this node belongs to
            string[] clbDomains = localAddrBind.NodeId.GetAllCLBDomains();
            
            // Build the table for all these CLB domains
            if( clbDomains!=null ) {
                rows = new CLBTableRow[ clbDomains.Length+1 ];
                rows[0] = new CLBTableRow(localNode, "");
                for( int i=0; i<clbDomains.Length; i++ ) {
                    rows[i+1] = new CLBTableRow(localNode, clbDomains[i]);
                }
            } else {
                rows = new CLBTableRow[1];
                rows[0] = new CLBTableRow(localNode, "");
            }
        }
        

        //**********************************************************************************************
        // Public properties
        //**********************************************************************************************

        public CLBSet this[string clbDomain] {
            get {
                foreach( CLBTableRow row in rows ) {
                    if( clbDomain.Equals(row.clbDomain) ) {
                        return row.clbSet;
                    }
                }
                return null;
            }
            set {
                foreach( CLBTableRow row in rows ) {
                    if( clbDomain.Equals(row.clbDomain) ) {
                        row.clbSet = value;
                    }
                }
                throw new ApplicationException("Error: Bad clb domain");
            }
        }


        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************

        /// <summary>
        /// Send a delayed message to ourself instructing us to ping our neighbours
        /// </summary>
        internal void ScheduleUpdate(double delayFactor) {
            long initialDelayMs = (long) (delayFactor * Parameters.clbTableUpdatePeriod);
            Timer pingTimer = new CLBTableUpdateTimer(initialDelayMs, Parameters.clbTableUpdatePeriod);
            localNode.SetTimer(pingTimer);
        }

        /// <summary>
        /// Send out messages to update our 
        /// </summary>
        internal void UpdateNbrsNow() {
            // Time to update the list of neighbours in the CLB Domains
            Log.Write(CATEGORY, LogLvl.Lo, "Updating CLB Domain Neighbours now");

            for( int i=0; i<rows.Length; i++ ) {
                GatherCLBNbrs msg = new GatherCLBNbrs(localNode, rows[i].clbDomain);
                RouteMsg routeMsg = new RoutePayloadControlMsg(localAddrBind, msg);
                localNode.RouteMessageInternal(routeMsg);
            }
        }

        /// <summary>
        /// A node has failed and thus we are to remove it from the routing table, if it exists.
        /// 
        /// We don't need to worry about calling RebuildNeighbours() here because the table
        /// that it rebuilds does not have entries from the CLB Table.
        /// </summary>
        /// <param name="n"></param>
        internal void NodeFailed( NodeIdAddressBind n ) {
            foreach( CLBTableRow row in rows ) {
                row.clbSet.NodeFailed( n );
            }
        }

        
        //**********************************************************************************************
        // Private methods and nested classes
        //**********************************************************************************************
        
        /// <summary>
        /// A periodic timer used to update the CLB neighbours.
        /// </summary>
        class CLBTableUpdateTimer : Timer {
            public CLBTableUpdateTimer(long initialDelayMs, long periodMs)
                : base(initialDelayMs, periodMs) {}
            public override void Execute( MessageRouting.Node node ) {
                Node localNode = (Node) node;

                if( ! localNode.clbRebuildEnabled ) {
                    this.Periodic = false;
                    return;
                }

                localNode.CLBTable.UpdateNbrsNow();
            }
        }
    }

        
    /// <summary>
    /// Gather the CLB Neighbours 
    /// </summary>
    [Serializable, XmlRootAttribute("GatherCLBNbrs", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class GatherCLBNbrs : OverlayMsg, RouteControl {

        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        
        public const string         CATEGORY = "GatherCLBNbrs";


        //**********************************************************************************************
        // Private members
        //**********************************************************************************************
        public  NodeIdAddressBind   source;             // The source of this fixup message
        public  string              clbDomain;          // The CLB domain we are searching

        public  int                 curHeight;          // Current height
        public  Direction           dir;                // Current direction

        public  CLBSet              newCLBSet;          // The CLBSet that we are building up
        public  bool                needMinusOneNbr;    // If true, we can stop whenever we see a neighbour
                                                        // at this level s.t. hisNumId[h]=myNumId[h]-1
        public  bool                needPlusOneNbr;     // If true, we can stop whenever we see a neighbour
                                                        // at this level s.t. hisNumId[h]=myNumId[h]+1
        public  bool                complete;           // If true, the message is complete

        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor needed for serialization
        public GatherCLBNbrs() {}
        
        /// <summary>
        /// This message finds the node at level 'height' that should be our neighbour
        /// at level height+1.
        /// </summary>
        /// <param name="localAddrBind">The source node</param>
        /// <param name="h">The height at which to examine nodes</param>
        public GatherCLBNbrs( Node localNode, string clbDomain ) {
            this.source = localNode.NodeIdAddressBind;
            this.clbDomain = clbDomain;
            this.curHeight = Parameters.maxHeight-1;
            this.dir = Direction.Right;
            this.newCLBSet = new CLBSet(null, localNode.NodeIdAddressBind, SkipNet.Parameters.clbTableNbrListSize);
        }
        

        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************

        /// <summary>
        /// Determine the next hop for this message
        /// </summary>
        public NodeIdAddressBind GetNextHop( Node localNode ) {
            NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;
            RoutingTable rt = localNode.RoutingTable;

            Debug.Assert( IsNodeWithinCLBDomain(localAddrBind) );

            Log.Write(CATEGORY, LogLvl.Lo, "Received GatherCLBNbrs. Src={0}, Domain={1}", source, clbDomain); 

            // If the source has failed, drop the message
            if( source==null ) {
                return null;
            }

            Start:;

            // If we're at the source and either stepping to the right or complete, then check to see if we're done
            if( localAddrBind.Equals(source) && (dir==Direction.Right || complete) ) {
                if( newCLBSet.IsComplete() || complete ) {
                    // The message is complete
                    return localAddrBind;
                }
                if( newCLBSet.IsLeftComplete() ) {
                    while( curHeight>=0 && source.NumericId.GetDigit(curHeight)==SkipNet.Parameters.b-1 ) {
                        // Can skip this level: there are no greater nodes than the ones we've already seen
                        curHeight--;
                    }
                    needPlusOneNbr = true;      // We can stop when we find a node with numID[h]=srcNumID[h]+1
                }
                if( newCLBSet.IsRightComplete() ) {
                    while( curHeight>=0 && source.NumericId.GetDigit(curHeight)==0 ) {
                        // Can skip this level: there are no lesser nodes than the ones we've already seen
                        curHeight--;
                    }
                    needMinusOneNbr = true;     // We can stop when we find a node with numID[h]=srcNumID[h]-1
                }
                if( curHeight<0 ) {
                    // If we've examined all levels then we're done
                    return localAddrBind;
                }
            }

            // Continue searching the current ring

            // Inserting each node into the CLB Set
            if( ! localAddrBind.Equals(source) ) {
                newCLBSet.AddNode(localAddrBind);
            }

            // If we need a "minus-one" neighbour and this node fits the bill, we're done
            if(    needMinusOneNbr
                && localAddrBind.NumericId.GetDigit(curHeight)==source.NumericId.GetDigit(curHeight)-1 )
            {
                // TODO: Should probably clear newCLBSet.left here
                complete = true;
                return source;
            }

            // If we need a "plus-one" neighbour and this node fits the bill, we're done
            if(    needPlusOneNbr
                && localAddrBind.NumericId.GetDigit(curHeight)==source.NumericId.GetDigit(curHeight)+1 )
            {
                // TODO: Should probably clear newCLBSet.right here
                complete = true;
                return source;
            }

            // Find set S: the neighbours of the local node at the current level and direction
            LeafSet ls = rt.Neighbours[ curHeight ];
            ArrayList S = ls.GetHalf(dir).GetAllNbrs();

            // If the first node of S is still within the CLB domain and doesn't wrap,
            // hop to that node
            if( S.Count>0 ) {
                NodeIdAddressBind s = (NodeIdAddressBind) S[0];
                if( IsNodeWithinCLBDomain(s) && ! Wrapped(localAddrBind.NodeId, s.NodeId) ) {
                    return s;
                }
            }

            // Next phase:
            // - If currently going right, go left and send back to the source node
            // - If currently going left, drop down a level, go right, and send msg back to the source
            if( dir==Direction.Left ) curHeight--;
            dir = Node.OppositeDir(dir);

            // Go back to the source node. If we're already at the source node,
            // just go to the start of this function again
            if( localAddrBind.Equals(source) ) {
                goto Start;
            } else {
                return source;
            }
        }
        
        /// <summary>
        /// The message has arrived at its final destination.
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node node ) {
            Node localNode = (Node) node;
            NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;

            Log.Write(CATEGORY, LogLvl.Med, "GatherCLBNbrs Complete. Domain={0}", clbDomain);
            
            CLBSet myCLBSet = localNode.CLBTable[ clbDomain ];
            myCLBSet.UpdateNbrs( this.newCLBSet );
        }
        
        /// <summary>
        /// After each time SkipNet tries to send this message to another node,
        /// this callback notifies the message whether the send was successful or not.
        /// </summary>
        public void SendCallback( Node localNode, NodeIdAddressBind dest, bool success ) {
            if( (! success) && source.Equals(dest) ) {
                source = null;
            }
        }

        /// <summary>
        /// An indication to the simulator whether this is a 'background' message or not.
        /// Background messages are ignored by "ProcessUntilIgnoreEventsOnly()".
        /// </summary>
        public bool IsBackground() {
            return true;
        }


        //**********************************************************************************************
        // Private methods
        //**********************************************************************************************

        /// <summary>
        /// Determines if the given node is still within the CLB domain we're searching
        /// </summary>
        private bool IsNodeWithinCLBDomain( NodeIdAddressBind node ) {
            return node.NodeId.IsWithinPrefix( clbDomain );
        }

        /// <summary>
        /// Determine if the step from 'local' to 'next' in the current direction
        /// wrapped around.
        /// </summary>
        private bool Wrapped( NodeId local, NodeId next ) {
            if( dir==Direction.Right ) {
                return ( next.CompareTo(local) <= 0 );
            } else {
                return ( next.CompareTo(local) >= 0 );
            }
        }
    }

 
    /// <summary>
    /// Stores a set of pointers for routing in a CLB domain
    /// </summary>
    [Serializable, XmlRootAttribute("CLBSet",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class CLBSet : RoutingSet {

        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor needed for serialization
        public CLBSet() {}
        
        /// <summary>
        /// Leaf Set Constructor
        /// </summary>
        /// <param name="localAddrBind">Address of the local node</param>
        /// <param name="localNode">Local node object</param>
        /// <param name="maxSize">The maximum size of _half_ of the clb set</param>
        public CLBSet(Node localNode, NodeIdAddressBind localAddrBind, int maxSize)
            : base(localNode, localAddrBind, maxSize, true)
        {
            left = new CLBSetHalf( localNode, localAddrBind, Direction.Left, maxSize, true );
            right = new CLBSetHalf( localNode, localAddrBind, Direction.Right, maxSize, true );
        }


        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************
        
        /// <summary>
        /// Try to add a single node to the routing set
        /// </summary>
        /// <param name="newNode">The node to add</param>
        /// <returns>True if the node was added</returns>
        public override bool AddNode( NodeIdAddressBind newNode ) {
            int cmp = newNode.NumericId.CompareTo( OwnerNode.NumericId );
            if( cmp<0 ) {
                return left.AddNode(newNode);
            } else {
                return right.AddNode(newNode);
            }
        }

        /// <summary>
        /// Try to add an array of nodes to our leaf set.
        /// </summary>
        /// <param name="newNodes">An array of NodeIdAddressBind objects</param>
        /// <returns>True if any node was added</returns>
        public override bool AddNodes( ArrayList newNodes ) {
            throw new ApplicationException("Not implemented");
        }
        
        /// <summary>
        /// Returns true if the left side of the CLBSet has been fully constructed.
        /// </summary>
        public bool IsLeftComplete() {
            CLBSetHalf l = (CLBSetHalf) left;
            return (l.NbrCount==l.MaxSize);
        }
        
        /// <summary>
        /// Returns true if the right side of the CLBSet has been fully constructed.
        /// </summary>
        public bool IsRightComplete() {
            CLBSetHalf r = (CLBSetHalf) right;
            return (r.NbrCount==r.MaxSize);
        }

        /// <summary>
        /// Returns true if the CLBSet has been fully constructed.
        /// </summary>
        public bool IsComplete() {
            return IsLeftComplete() && IsRightComplete();
        }

        /// <summary>
        /// Update this CLB set with the given CLB set
        /// </summary>
        public void UpdateNbrs( CLBSet newCLBSet ) {
            CLBSetHalf l = (CLBSetHalf) left;
            CLBSetHalf r = (CLBSetHalf) right;
            l.UpdateNbrs( (CLBSetHalf) newCLBSet.left );
            r.UpdateNbrs( (CLBSetHalf) newCLBSet.right );
        }

        public ArrayList GetNbrsForDest( CLBId destId ) {
            ArrayList leftNbrs = left.GetAllNbrs();
            ArrayList rightNbrs = right.GetAllNbrs();
            leftNbrs.AddRange( rightNbrs );
            NbrComparer comparer = new NbrComparer( destId.NumericId );
            leftNbrs.Sort( comparer );

            int extraSize = leftNbrs.Count-Parameters.clbTableNbrListSize;
            if( extraSize>0 ) {
                leftNbrs.RemoveRange(Parameters.clbTableNbrListSize, extraSize);
            }
            return leftNbrs;
        }

        class NbrComparer : IComparer {
            NumericId dest;
            public NbrComparer( NumericId dest ) {
                this.dest = dest;
            }
            public int Compare(object px, object py) {
                NumericId x = ((NodeIdAddressBind) px).NumericId;
                NumericId y = ((NodeIdAddressBind) py).NumericId;
                
                // First, we prefer longer prefix matches
                int xmatch = x.LongestCommonPrefixLength(dest);
                int ymatch = y.LongestCommonPrefixLength(dest);
                if( xmatch>ymatch ) {
                    return -1;
                } else if( ymatch>xmatch ) {
                    return 1;
                }

                // Next, we consider absolute distance
                NumericId xdiff = x.Difference(dest);
                NumericId ydiff = y.Difference(dest);
                int result = xdiff.CompareTo(ydiff);
                return result;
            }
        }
    }
    
    
    /// <summary>
    /// Implements one half of a leaf set (either left or right)
    /// </summary>
    [Serializable, XmlRootAttribute("CLBSetHalf",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class CLBSetHalf : RoutingSetHalf {
        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        public CLBSetHalf( Node localNode, NodeIdAddressBind localAddrBind, Direction dir, int maxSize,
            bool authoritative ) : base(localNode, localAddrBind, dir, maxSize, authoritative)
        {}
        

        //**********************************************************************************************
        // Public Properties
        //**********************************************************************************************
        
        public int NbrCount {
            get { return this.nbrs.Count; }
        }


        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************
        
        public void UpdateNbrs( CLBSetHalf newHalf ) {
            this.nbrs = newHalf.nbrs;
        }

        
        //**********************************************************************************************
        // Private Methods
        //**********************************************************************************************

        /// <summary>
        /// Determine if node B lies strictly between nodes A and C, taking into
        /// account the direction of this leaf-set half.
        /// </summary>
        protected override bool LiesBetween( NodeIdAddressBind A, NodeIdAddressBind B, NodeIdAddressBind C ) {
            return NumericId.LiesBetweenDir(this.dir, A.NumericId, B.NumericId, C.NumericId);
        }

        /// <summary>
        /// Determine if going from node A to node B in direction 'dir' causes a wrap.
        /// </summary>
        private bool Wrap( Direction dir, NumericId A, NumericId B ) {
            if( Direction.Left==dir ) {
                return( A.CompareTo(B)<0 );
            } else {
                return( A.CompareTo(B)>0 );
            }
        }
    }

}