using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// This class stores a routing table that is designed for CLB Routing operations.
    /// </summary>
    public class CTable {
    
        /// <summary>
        /// This class defines one entry in the CTable
        /// </summary>
        public class CTableEntry {
            public NodeIdAddressBind    node;
            public int                  dist;
            public int                  prefixLen;

            public CTableEntry( NodeIdAddressBind node, int dist, int prefixLen ) {
                this.node = node;
                this.dist = dist;
                this.prefixLen = prefixLen;
            }
        }
        
        
        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        Node                localNode;
        Topology            topo;
        NodeIdAddressBind   localAddrBind;
        int                 b;
        int                 height;
        CTableEntry[,]      entries;
        
        
        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        public const string CATEGORY = "CTable";


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        public CTable( Node localNode ) {
            this.localNode = localNode;
            this.topo = localNode.Topology;
            this.localAddrBind = localNode.NodeIdAddressBind;
            this.height = Parameters.maxHeight;
            this.b = Parameters.b;
            this.entries = new CTableEntry[ height, b ];
        }
    
    
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************
    
        /// <summary>
        /// Returns an array containing the NodeIdAddressBind object for all
        /// entries in the table.
        /// </summary>
        /// <returns></returns>
        public ArrayList GetAllEntries() {
            ArrayList result = new ArrayList();

            for( int i=0; i<height; i++ ) {
                for( int j=0; j<b; j++ ) {
                    CTableEntry entry = entries[i,j];
                    if( entry!=null ) {
                        result.Add( entry.node );
                    }
                }
            }

            return result;
        }

        /// <summary>
        /// Returns an array containing PTableEntry objects for all entries in the table.
        /// This is used so that the P-Table can populate itself from the C-Table.
        /// </summary>
        public ArrayList GetArrayOfPTableEntries() {
            ArrayList result = new ArrayList();

            for( int i=0; i<height; i++ ) {
                for( int j=0; j<b; j++ ) {
                    CTableEntry entry = entries[i,j];
                    if( entry!=null ) {
                        result.Add( new PTableEntry(entry.node, entry.dist) );
                    }
                }
            }

            return result;            
        }

        /// <summary>
        /// Try to fill this table with new entries. Entries are only taken if they are better than
        /// an existing entry.
        /// 
        /// NOTE: This function calls "RebuildNeigbours()"
        /// </summary>
        /// <param name="newEntries">An ArrayList of NodeIdAddressBind objects</param>
        /// <param name="localNode">The local node</param>
        /// <param name="notifyWhenDone">If true, notify the neighbours when complete</param>
        public void Merge( Node localNode, ArrayList newEntries, bool notifyWhenDone ) {
            // Insert the new entries into a hash table to remove duplicates
            Hashtable table = new Hashtable( newEntries.Count );
            foreach( Object o in newEntries ) {
                if( ! table.ContainsKey(o) ) {
                    table.Add( o, null );
                }
            }

            ArrayList uniqueEntries = new ArrayList(table.Keys);
            MergeCont cont = new MergeCont(localNode, uniqueEntries, notifyWhenDone);
            cont.LoopTop();
        }

        
        /// <summary>
        /// Continuation used to merge new entries into the table.
        /// 
        /// NOTE: This continuation calls "RebuildNeigbours()"
        /// </summary>
        class MergeCont : Continuation {
            Node                localNode;
            ArrayList           newEntries;
            bool                notifyWhenDone;

            NodeIdAddressBind   newNode;
            int                 i, row, col, distance;

            public MergeCont( Node localNode, ArrayList newEntries, bool notifyWhenDone ) {
                this.localNode = localNode;
                this.newEntries = newEntries;

                this.i = 0;
                this.notifyWhenDone = notifyWhenDone;
            }

            // The top of the loop in which we check the new entries
            public void LoopTop() {
                NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;
                Topology topo = localNode.Topology;

                Top:
                if( i<newEntries.Count ) {
                    newNode = (NodeIdAddressBind) newEntries[i];
    
                    // Don't bother checking ourself!
                    if( newNode.Equals(localAddrBind) ) {
                        i++; goto Top;
                    }

                    // Determine the row/column the new node should live in
                    row = localAddrBind.NumericId.LongestCommonPrefixLength( newNode.NumericId );
                    col = newNode.NumericId.GetDigit( row );

                    // Measure the distance from the local node to the new node
                    PingCallback cb = new PingCallback(this.Callback);
                    topo.BeginPing(localAddrBind.GetAddress, newNode.GetAddress, false, cb);
                } else {
                    LoopDone();
                }
            }

            // Note: Not in a thread-safe state. Schedule the "Execute" function to be called.
            void Callback(NodePhysicalAddress destination, int distance) {
                this.distance = distance;
                localNode.ScheduleContinuation(this);
            }

            // Called from the event-loop when the ping has completed.
            // This is the latter half of the loop.
            public override void Execute( MessageRouting.Node node ) {
                NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;

                if( distance>=0 ) {
                    localNode.CTable.TryToAddNode(newNode, distance, row, col);
                }

                i++;
                LoopTop();
            }

            // This is called when the loop is complete
            void LoopDone() {
                if( this.notifyWhenDone ) {
                    // Notify all neighbours in the CLB Table
                    localNode.CTable.NotifyNeighbours( localNode );
                }

                localNode.RebuildNeighbours(null);
            }
        }

        /// <summary>
        /// Try to add the new node to the CLB Table
        /// 
        /// NOTE: The caller must call "RebuildNeigbours()" afterwards
        /// </summary>
        /// <param name="newNode">The node to add</param>
        /// <param name="distance">The distance to this node</param>
        internal void TryToAddNode(NodeIdAddressBind newNode, int distance, int row, int col) {
            int prefixLen = -1;
            if( Parameters.preferLongerPrefix ) {
                prefixLen = NodeId.NumCommonCLBDomains(localAddrBind.NodeId, newNode.NodeId);
            }
            
            // Replace the current entry if the new node is better
            if( IsBetterThan(newNode, distance, prefixLen, entries[row,col]) ) {
                entries[row,col] = new CTableEntry(newNode, distance, prefixLen);
            }
        }
        
        /// <summary>
        /// Notify neighbours that they are in our routing table. This gives them an
        /// opportunity to add us to their routing table.
        /// </summary>
        /// <param name="localNode"></param>
        public void NotifyNeighbours( Node localNode ) {
            // Get the list of all our neighbours
            ArrayList nbrs = GetAllEntries();

            // Get a list of all our neighbours plus ourself
            ArrayList nbrsAndMe = new ArrayList(nbrs);
            nbrsAndMe.Add( localAddrBind );

            // Send a notification message to each neighbour
            foreach( NodeIdAddressBind nbr in nbrs ) {
                // Note: We don't particularly care if this message fails to arrive, so
                // we don't use a callback here.
                UpdateCTableMsg msg = new UpdateCTableMsg(localAddrBind, nbrsAndMe);
                localNode.SendMessageDirect(msg, nbr);
            }
        }
    
        /// <summary>
        /// Determine the next hop in routing towards the destination.
        /// </summary>
        /// <param name="destID">The destination ID</param>
        /// <returns>The address of the next hop</returns>
        public NodeIdAddressBind GetNextHop( NumericId destID ) {

            // Determine the row/column to examine
            int row = destID.LongestCommonPrefixLength( localAddrBind.NumericId );
            int col = destID.GetDigit( row );
        
            // If the entry is empty, return the local node address
            if( entries[row,col]==null ) {
                return localAddrBind;
            } else {
                return entries[row,col].node;
            }
        }

        /// <summary>
        /// A node has failed and thus we are to remove it from the routing table, if it exists.
        /// 
        /// NOTE: The caller must call 'RebuildNeighbours()' if the node was removed, otherwise
        /// the table of all neighbours will be stale.
        /// </summary>
        /// <param name="nodeAddr">The failed node</param>
        /// <returns>True if the node was removed</returns>
        public bool NodeFailed( NodeIdAddressBind failedNode ) {
            NumericId failedRid = failedNode.NumericId;
        
            // Determine the row/column to examine
            int row = failedRid.LongestCommonPrefixLength( localAddrBind.NumericId );
            int col = failedRid.GetDigit( row );
        
            // If the entry in that row/column matches the failed entry, remove it
            if( entries[row,col]!=null && failedNode.Equals(entries[row,col].node) ) {
                entries[row,col] = null;
                return true;
            }

            return false;
        }

        /// <summary>
        /// Insert each neighbour into the given Hashtable 
        /// </summary>
        /// <param name="ht">Table to insert neighbours into</param>
        public void GetAllNeighbours(Hashtable ht) {
            for( int i=0; i<height; i++ )  {
                for( int j=0; j<b; j++ ) {
                    if( ( entries[i,j]!=null ) && (! ht.ContainsKey(entries[i,j].node)) ) {
                        ht.Add(entries[i,j].node, null);
                    }
                }
            }
        }
        
        /// <summary>
        /// Start a periodic timer instructing us to ping our neighbours
        /// </summary>
        internal void SchedulePing(double delayFactor) {
            long initialDelayMs = (long) (delayFactor * Parameters.neighbourPingPeriod);
            Timer pingTimer = new CTablePingTimer(initialDelayMs, Parameters.neighbourPingPeriod);
            localNode.SetTimer(pingTimer);
        }

        /// <summary>
        /// Send out a message to rebuild the C-Table
        /// </summary>
        internal void Rebuild() {
            NodePhysicalAddress seedAddr = localNode.seed;
            NodeIdAddressBind rebuildFromNode=null;
            CTableEntry closestEntry = GetClosest();
            
            // We prefer to use the closest entry in the C-Table as the seed.
            if( closestEntry!=null ) {
                rebuildFromNode = closestEntry.node;
            } else {
                // If no entry exists we use the original seed.
                if( seedAddr!=null ) {
                    rebuildFromNode = new NodeIdAddressBind(seedAddr, null, null);
                } else {
                    // If no original seed, use the closest leaf-set node
                    LeafSet ls = localNode.RoutingTable.LeafSet;
                    ArrayList leaves = ls.GetAllNbrs();
                    if( leaves.Count>0 ) {
                        rebuildFromNode = ls.GetClosest(Direction.Right);
                    }
                }
            }

            if( rebuildFromNode!=null ) {
                CLBJoinRequest clbJoin = new CLBJoinRequest(localNode, Node.TIMEOUT_DELAY_MS);
                localNode.SendMessageDirect(clbJoin, rebuildFromNode);
            }
        }
        

        //**********************************************************************************************
        // Private Methods
        //**********************************************************************************************

        /// <summary>
        /// Determine if the new node would be better than the old entry
        /// </summary>
        /// <param name="newNode">The new node</param>
        /// <param name="dist">The distance from the new node to the local node</param>
        /// <param name="oldEntry">The old entry</param>
        private bool IsBetterThan( NodeIdAddressBind newNode, int dist, int prefixLen, CTableEntry oldEntry ) {

            // A new node is better than no node at all
            if( oldEntry==null )
                return true;
        
            if( Parameters.preferLongerPrefix ) {

                // If the new node has a longer prefix match, it is better
                if( prefixLen>oldEntry.prefixLen ) {
                    return true;
                }

                // If the new node has a shorter prefix match, it is worse
                if( prefixLen<oldEntry.prefixLen ) {
                    return false;
                }
        
            }
            
            // If the new node has shorter distance, it is better
            if( dist<oldEntry.dist ) {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Get the CTableEntry with the least distance to the local node
        /// </summary>
        private CTableEntry GetClosest() {
            CTableEntry best=null;

            for( int i=0; i<height; i++ ) {
                for( int j=0; j<b; j++ ) {
                    CTableEntry entry = entries[i,j];
                    if( entry!=null ) {
                        if( best==null || entry.dist<best.dist ) {
                            best = entry;
                        }
                    }
                }
            }

            return best;
        }

    }


    /// <summary>
    /// A periodic timer used to ping our neighbours.
    /// </summary>
    class CTablePingTimer : Timer {
        public CTablePingTimer(long initialDelayMs, long periodMs)
            : base(initialDelayMs, periodMs) {}

        public override void Execute( MessageRouting.Node node ) {
            Node localNode = (Node) node;
            localNode.CTable.Rebuild();
        }
    }


    /// <summary>
    /// A request message used to build the CTable.
    /// </summary>
    [Serializable, XmlRootAttribute("CLBJoinRequest",
                       Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class CLBJoinRequest : OverlayRequestMsg {

        //**********************************************************************************************
        // Constants
        //**********************************************************************************************

        [XmlType(Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, TypeName="CLBJoinPhase")]
        public enum Phase   {
            Seed,         // Send from joining node to seed
            Routing       // Route By Name
        }


        //**********************************************************************************************
        // Public Members (for serialization convenience)
        //**********************************************************************************************        
        public NodeIdAddressBind    joiningNode;
        public ArrayList            entries;
        public Phase                phase;
    
    
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************        
    
        /// <summary>
        /// Default constructor for serialization
        /// </summary>
        public CLBJoinRequest() {}
        
        /// <summary>
        /// Constructor for a CLBJoin message
        /// </summary>
        /// <param name="source">The joining node</param>
        /// <param name="dest">The CLBId of the joining node</param>
        public CLBJoinRequest( MessageRouting.Node localNode, long delayMs )
            : base( localNode, new CLBJoinCallback(), delayMs, true )
        {
            this.joiningNode = (NodeIdAddressBind) localNode.GetNodeIdAddressBind;
            this.entries = new ArrayList();
            this.phase = Phase.Seed;
        }
    
    
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************        
    
        /// <summary>
        /// Called on each hop of the CLB join message.
        /// </summary>
        /// <returns>Always returns true</returns>
        public override bool CheckContinueRouting( MessageRouting.Node node ) {
            Node localNode = (Node) node;
            CTable cTable = localNode.CTable;

            Log.Write(CTable.CATEGORY, LogLvl.Med, "CLB Join message from {0} passing through", joiningNode);

            // Add all entries from our local CLB Table to the CLB Join message
            entries.AddRange( cTable.GetAllEntries() );
            entries.Add( localNode.NodeIdAddressBind );

            return true;
        }
        
        /// <summary>
        /// Called when the Join message arrives at its destination (the closest node to the joining node).
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node node ) {
            Node localNode = (Node) node;

            Log.Write(CTable.CATEGORY, LogLvl.Med, "CLB Join message from {0} arrived", joiningNode);

            switch( phase ) {
                case Phase.Seed:
                    // The message has arrived at the seed node
                    RouteByCLBMsg snMsg = new RouteByCLBMsg(localNode.NodeIdAddressBind, 
                        new CLBId( joiningNode.NumericId ), this);
                    this.phase = Phase.Routing;
                    localNode.ResumeRouting(snMsg);
                    break;

                case Phase.Routing:
                    CLBJoinReply reply = new CLBJoinReply(this, entries);
                    node.SendReply(reply, this.joiningNode);
                    break;
            }
        }
    }
        
        
    /// <summary>
    /// A CLB Join reply message that contains candidate CTable entries.
    /// </summary>
    [Serializable, XmlRootAttribute("CLBJoinReply",
                       Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class CLBJoinReply : OverlayReplyMsg {
        public ArrayList       entries;

        /// <summary>
        /// Default constructor for serialization
        /// </summary>
        public CLBJoinReply() {}

        /// <summary>
        /// Constructor for CLBJoinReply.
        /// </summary>
        /// <param name="requestMsg">The original request</param>
        /// <param name="entries">The CLB table entries that were found</param>
        public CLBJoinReply( OverlayRequestMsg requestMsg, ArrayList entries )
            : base(requestMsg)
        {
            this.entries = entries;
        }
    }


    /// <summary>
    /// Callback that is executed when join completes or times out.
    /// </summary>
    public class CLBJoinCallback : OverlayReplyCallback {
        public override void Execute( MessageRouting.Node node ) {
            Node localNode = (Node) node;
            NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;
            CTable cTable = localNode.CTable;

            // Check if join request timed out
            CLBJoinReply joinReply = (CLBJoinReply) this.reply;
            if( joinReply!=null ) {
                // Add the new entries found during the CLB Join to our local CLB Table
                cTable.Merge(localNode, joinReply.entries, true);
            } else {
                // TODO: Do something better to handle this failure?
                Log.Write(CTable.CATEGORY, LogLvl.Hi, "C-Table Join Failed!");
            }            
        }
    }
    
    
    /// <summary>
    /// This message is used to notify another node that we have added it to our C-Table.
    /// </summary>
    [Serializable, XmlRootAttribute("UpdateCTableMsg",
                       Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class UpdateCTableMsg : OverlayMsg {
        
        //**********************************************************************************************
        // Serialized Members
        //**********************************************************************************************
        public NodeIdAddressBind   source;
        ArrayList           entries;        // Serialized with a property below
        
        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor for serialization
        public UpdateCTableMsg() : base() {}

        public UpdateCTableMsg( NodeIdAddressBind source, ArrayList entries ) {
            this.source = source;
            this.entries = entries;
        }

        
        //**********************************************************************************************
        // Public properties
        //**********************************************************************************************

        [XmlArray( ElementName="Entries",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="NodeIdAddressBind", Type=typeof(NodeIdAddressBind),
             IsNullable=true, Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public ArrayList Entries {
            get { return entries; }
            set { entries = value; }
        }
        

        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************

        /// <summary>
        /// This method is called when another node has added us to its CLB Table
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node node ) {
            Node localNode = (Node) node;

            Log.Write(CTable.CATEGORY, LogLvl.Lo, "Received UpdateCTable msg from {0}", source);
            CTable localTable = localNode.CTable;

            localTable.Merge(localNode, this.entries, false);
        }
    }

}
