//*******************************************************************************************************
//
// Implements the messages used to repair SkipNet rings
//
//*******************************************************************************************************

using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// Examine the nodes at level h to find the correct neighbour at level h+1.
    /// </summary>
    [Serializable, XmlRootAttribute("FixupHigher",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class FixupHigher : OverlayMsg, RouteControl {

        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        public const string     CATEGORY = "FixupHigher";


        //**********************************************************************************************
        // Private members
        //**********************************************************************************************
        public  NodeIdAddressBind   source;             // The source of this fixup message
        public  int                 height;             // The level at which the source started the fixup
        public  bool                firstHop;           // Is this the first hop by the message?
        public  bool                complete;           // Is the message complete?
        public  Direction           dir;
        public  bool                background;         // Indication to the simulator if this is a background message
        private ArrayList           nbrs;               // The leaf set of the correct neighbour at level h+1.

        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor needed for serialization
        public FixupHigher() : base() {}
        
        /// <summary>
        /// This message finds the node at level 'height' that should be our neighbour
        /// at level height+1.
        /// </summary>
        /// <param name="localAddrBind">The source node</param>
        /// <param name="h">The height at which to examine nodes</param>
        public FixupHigher( NodeIdAddressBind localAddrBind, int height, Direction direction,
            bool background )
        {
            this.source = localAddrBind;
            this.height = height;
            this.firstHop = true;
            this.complete = false;
            this.nbrs = null;
            this.dir = direction;
            this.background = background;
        }
        
        
        //**********************************************************************************************
        // Public properties
        //**********************************************************************************************

        [XmlArray( ElementName="Neighbours",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="NodeIdAddressBind", Type=typeof(NodeIdAddressBind),
             IsNullable=true, Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public ArrayList Neighbours {
            get { return nbrs; }
            set { nbrs = value; }
        }
        

        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************

        /// <summary>
        /// Determine the next hop for this message
        /// </summary>
        public NodeIdAddressBind GetNextHop( Node localNode ) {
            NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;
            RoutingTable rt = localNode.RoutingTable;

            Log.Write(CATEGORY, LogLvl.Lo, "Received FixupHigher. Src={0}, Ht={1}, Dir={2}",
                source, height, dir); 

            // If this is the first hop, treat it specially
            if( firstHop ) {
                firstHop = false;
                NodeIdAddressBind nbr = rt.GetNeighbour(dir, height);
                Debug.Assert( nbr!=null );

                if( nbr.Equals(localAddrBind) ) {
                    // We have no neighbours at this level. The fixup was vacuously successful.
                    return null;
                } else {
                    Log.Write(CATEGORY, LogLvl.Med, "Sending FixupHigher to {0}", nbr);
                    return nbr;
                }
            }

            // If the source has failed, drop the message
            if( source==null ) {
                return null;
            }

            // If the message is complete, route back to the source.
            // If we have already arrived at the source, don't route any further.
            if( this.complete || localAddrBind.Equals(source) ) {
                return source;
            }
            
            // Determine how many digits the local node shares with the source node (i.e. the level of
            // the highest ring containing both nodes)
            int commonH = source.NumericId.LongestCommonPrefixLength(localAddrBind.NumericId);
            Debug.Assert( commonH>=height && commonH<Parameters.maxHeight );
            
            // If this node shares more than 'height' digits with the source,
            // then our mission is complete. Go back to the source.
            if( commonH>height ) {
                this.complete = true;
                this.nbrs = rt.Neighbours[height+1].GetAllNbrs();
                this.nbrs.Add(localAddrBind);
                return source;
            }
            
            // Get our next neighbour at this level
            NodeIdAddressBind nextHop = localNode.RoutingTable.GetNeighbour(dir, height);

            // If there is no next neighbour, or the source should be between this node and
            // the next neighbour, then this message is complete.
            if(    nextHop.Equals(localAddrBind)
                || NodeId.LiesBetweenDir(dir, localAddrBind.NodeId, source.NodeId, nextHop.NodeId) )
            {
                this.complete = true;
                Log.Write(CATEGORY, LogLvl.Med, "No more nbrs. Returning to source {0}", this.source);
                return source;      
            } else {
                Log.Write(CATEGORY, LogLvl.Med, "Sending to node {0}", nextHop);
                return nextHop;
            }
        }

        /// <summary>
        /// The message has arrived at its final destination.
        /// </summary>
        /// <param name="localNode"></param>
        public override void ArrivedAtDestination( MessageRouting.Node node ) {
            Node localNode = (Node) node;

            Log.Write(CATEGORY, LogLvl.Med, "FixupHigher Complete. Ht={0}, Dir={1}", height, this.dir); 

            if( this.nbrs==null ) {
                Log.Write(CATEGORY, LogLvl.Med, "FixupHigher was unsuccessful", height);
                // TODO: Initiate CLB Table construction here
            } else {
                localNode.RoutingTable.RepairLevel(height+1, this.nbrs, this.background);
            }
        }
        
        /// <summary>
        /// After each time SkipNet tries to send this message to another node,
        /// this callback notifies the message whether the send was successful or not.
        /// </summary>
        public void SendCallback( Node localNode, NodeIdAddressBind dest, bool success ) {
            if( (! success) && source.Equals(dest) ) {
                source = null;
            }
        }
        
        /// <summary>
        /// An indication to the simulator whether this is a 'background' message or not.
        /// Background messages are ignored by "ProcessUntilIgnoreEventsOnly()".
        /// </summary>
        public bool IsBackground() {
            return background;
        }
    }

}
