//*******************************************************************************************************
//
// Implements the message used to Join a SkipNet
//
//*******************************************************************************************************

using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// Contains the state of and performs logic for a Join operation.
    /// </summary>
    [Serializable, XmlRootAttribute("JoinRequest",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class JoinRequest : OverlayRequestMsg {

        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        
        [XmlType(Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, TypeName="JoinPhase")]
        public enum Phase   {
            Seed,         // Send from joining node to seed
            Routing       // Route By Name
        }

        public const string    CATEGORY = "SkipNetJoin";


        //**********************************************************************************************
        // Public Members (public for serialization convenience)
        //**********************************************************************************************

        public NodeIdAddressBind    joiningNode;    // The node joining the SkipNet
        public NodeIdAddressBind    destination;    // The node closest to the joining node in the SkipNet
        public Phase                phase;          


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        // Default constructor
        public JoinRequest() {}

        public JoinRequest( MessageRouting.Node localNode, long delayMs )
            : base( localNode, new JoinCallback(), delayMs, true )
        {
            this.joiningNode = (NodeIdAddressBind) localNode.GetNodeIdAddressBind;
            this.phase = Phase.Seed;
        }


        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************
        
        /// <summary>
        /// Called each time a phase completes.
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node node ) {
            Node localNode = (Node) node;
            NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;

            Log.Write(CATEGORY, LogLvl.Med, "Received Join message from NodeId {0} at address {1}",
                joiningNode, joiningNode.GetAddress.ToString());

            switch( phase ) {
                case Phase.Seed:
                    // The message has arrived at the seed node
                    RouteByNameMsg snMsg = new RouteByNameMsg(localAddrBind, false,
                        joiningNode.NodeId, RouteByNameOption.Normal, this);
                    this.phase = Phase.Routing;

                    // Note: ResumeRouting may result in a recursive call to this function
                    localNode.ResumeRouting(snMsg);
                    break;

                case Phase.Routing:
                    // The message has arrived at the destination node (the node nearest to
                    // the joining node)
                    LeafSet ls = localNode.RoutingTable.LeafSet;
                    ArrayList allLeaves = ls.GetAllNbrs();
                    allLeaves.Add(localAddrBind);

                    // Send a JoinReply message back to the joining node
                    JoinReply reply = new JoinReply(this, allLeaves);
                    localNode.SendReply(reply, joiningNode);
                    break;
            }
        }
    }


    /// <summary>
    /// A Join reply message that contains the new leaf set for the joining node.
    /// </summary>
    [Serializable, XmlRootAttribute("JoinReply",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class JoinReply : OverlayReplyMsg {
        private ArrayList       nbrs;

        //**********************************************************************************************
        // Properties
        //**********************************************************************************************

        [XmlArray( ElementName="Neighbours",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="NodeIdAddressBind", Type=typeof(NodeIdAddressBind),
             IsNullable=true, Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public ArrayList Neighbours {
            get { return nbrs; }
            set { nbrs = value; }
        }
        

        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        /// <summary>
        /// Default constructor for serialization
        /// </summary>
        public JoinReply() {}

        /// <summary>
        /// Constructor for JoinReply.
        /// </summary>
        /// <param name="requestMsg">The original request</param>
        /// <param name="entries">The neighbours of the joining node</param>
        public JoinReply( OverlayRequestMsg requestMsg, ArrayList nbrs )
            : base(requestMsg)
        {
            this.nbrs = nbrs;
        }
    }


    /// <summary>
    /// Callback that is executed when join completes or times out.
    /// </summary>
    public class JoinCallback : OverlayReplyCallback {
        public override void Execute( MessageRouting.Node node ) {
            Node localNode = (Node) node;

            // Check if join request timed out
            JoinReply joinReply = (JoinReply) this.reply;
            if( joinReply==null ) {
                throw new ApplicationException("R-Table Join Failed!");
            }
            
            // The completed join message has arrived back at the joining node.
            // Repair the routing table using the neighbours that we have
            // found using the Join message.
            localNode.RoutingTable.RepairLevel(0, joinReply.Neighbours, false);
        }
    }

}
