using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// Stores a set of pointers to other SkipNet nodes.
    /// </summary>
    [Serializable, XmlRootAttribute("LeafSet",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class LeafSet : RoutingSet {

        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor needed for serialization
        public LeafSet() {}
        
        /// <summary>
        /// Leaf Set Constructor
        /// </summary>
        /// <param name="localAddrBind">Address of the local node</param>
        /// <param name="localNode">Local node object</param>
        /// <param name="maxSize">The maximum size of _half_ of the leaf set</param>
        /// <param name="bottomLevel">True if this Leaf Set is the bottom level of the R-Table</param>
        public LeafSet(Node localNode, int maxSize, bool authoritative)
            : base(localNode, localNode.NodeIdAddressBind, maxSize, authoritative)
        {
            left = new LeafSetHalf( localAddrBind, localNode, Direction.Left, maxSize, authoritative );
            right = new LeafSetHalf( localAddrBind, localNode, Direction.Right, maxSize, authoritative );
        }


        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************

        /// <summary>
        /// Determine if the leaf set contains the destination node id.
        /// If the option is PrefixMatch, returns true if the local node or any node
        /// in the leaf set matches the prefix.
        /// </summary>
        /// <param name="destID">The destination NodeId</param>
        public bool Contains( NodeId destID, RouteByNameOption option ) {
            NodeIdAddressBind nextHop = TryRouteOneHop(destID, option);
            if( nextHop!=null ) {
                return true;
            } else {
                return false;
            }
        }
        
        /// <summary>
        /// Try to route to the destination in one hop.
        /// If the option is PrefixMatch, return any node (including local node) that matches the prefix
        /// If the option is DisallowEqual, return the closest node to the destination but does not equal it.
        /// If unable to route in one hop, returns null.
        /// </summary>
        public NodeIdAddressBind TryRouteOneHop(NodeId destID, RouteByNameOption option) {
            return Route(destID, option, true);
        }
        
        /// <summary>
        /// Determine the next hop in routing towards the destination.
        /// If the option is PrefixMatch, return any node (including local node) that matches the prefix
        /// If the option is DisallowEqual, return the closest node to the destination but does not equal it.
        /// If the local node is the destination, returns the localAddrBind.
        /// </summary>
        public NodeIdAddressBind RouteByName( NodeId destID, RouteByNameOption option ) {
            return Route(destID, option, false);
        }

        public ArrayList GetFailoverNeighbours( NodeId destID ) {
            // Get a list of all unique neighbours
            ArrayList allNbrs = left.GetAllNbrs();
            ArrayList rightNbrs = right.GetAllNbrs();
            foreach( NodeIdAddressBind n in rightNbrs ) {
                if( ! allNbrs.Contains(n) ) {
                    allNbrs.Add(n);
                }
            }

            // Sort the list of neighbours in failover-order relative to this destination
            LeafSetComparer comparer = new LeafSetComparer( destID );
            allNbrs.Sort( comparer );
            
            // Trim any extra nodes from the list
            int extraSize = allNbrs.Count-Parameters.clbTableNbrListSize;
            if( extraSize>0 ) {
                allNbrs.RemoveRange(Parameters.clbTableNbrListSize, extraSize);
            }
            
            return allNbrs;
        }

        /// <summary>
        /// Send out ping(s) to fix up the leaf set at the given height in the given direction.
        /// If smart pings are enabled, we send a single LeafSetFixup message.
        /// If not, we send we send individual simple ping messages.
        /// </summary>
        public void SendPing( Direction dir, int height, bool background ) {
            if( Parameters.smartPingEnabled ) {
                SendLeafSetFixup(dir, height, background);
            } else {
                LeafSetHalf half = (LeafSetHalf) ( dir==Direction.Left ? left : right );
                half.SendSimplePings(height, background);
            }
        }

        
        //**********************************************************************************************
        // Private methods
        //**********************************************************************************************

        /// <summary>
        /// Internal routing mechanism. Determines the next hop in routing towards the destination.
        /// If the option is PrefixMatch, any node that has a prefix match with the destination is ok.
        /// If the option is DisallowEqual, the destination must be strictly between the localID and the destID.
        /// If one hop is true, this will return null if it cannot route to the destination in one hop.
        /// If one hop is false, this will simply return the best node in routing towards the destination.
        /// </summary>
        /// <returns>
        /// If the local node is the final destination, returns the local address.
        /// Otherwise, returns the address of the next hop.
        /// </returns>
        private NodeIdAddressBind Route(NodeId destID, RouteByNameOption option, bool oneHop) {
            NodeIdAddressBind   nextHop;
            LeafSetHalf         half = null;

            // Compare the destination with the local node id
            int cmp = destID.CompareTo(localAddrBind.NodeId);

            ////////////////////////////////////////////////////////////////////////////////////
            // Matches with local node

            // PrefixMatch
            if( option==RouteByNameOption.PrefixMatch && localAddrBind.NodeId.IsWithinPrefix(destID) ) {
                return localAddrBind;
            }

            // DisallowEqual
            if( option==RouteByNameOption.DisallowEqual ) {
                Debug.Assert( cmp!=0 );
            }
            
            // Normal / PrefixMatch option
            if(    option==RouteByNameOption.Normal
                || option==RouteByNameOption.PrefixMatch
                || option==RouteByNameOption.DisallowEqual )
            {
                if( cmp<0 ) {
                    half = (LeafSetHalf) left;
                } else if( cmp==0 ) {
                    return localAddrBind;
                } else {
                    half = (LeafSetHalf) right;
                }
            }

            ////////////////////////////////////////////////////////////////////////////////////
            
            Debug.Assert( half!=null );

            // Either route in one hop or route, as appropriate
            nextHop = half.Route(destID, option, oneHop);
            
            return nextHop;
        }

        /// <summary>
        /// Send out a smart ping (LeafSetFixup) to fix up the leaf set at the given height in
        /// the given direction.
        /// </summary>
        /// <param name="background">[Simulator] If true, the ping should be treated as
        /// a background message. If false, the ping is a foreground message</param>
        private void SendLeafSetFixup( Direction dir, int height, bool background ) {
            // First determine the set of nodes that we plan to visit during the fixup
            LeafSetHalf half = (LeafSetHalf) ( dir==Direction.Left ? left : right );
            ArrayList nbrs = half.GetAllNbrs();

            // Next, make upcalls to any applications that registered
            // for ping notifications, and build up a list of piggyback-objects
            // that we should send along to our neighbours.
            ArrayList allPiggybacks = new ArrayList();
            foreach( NodeIdAddressBind n in nbrs ) {
                ArrayList arr = localNode.PingNotification(n);
                if( arr!=null && arr.Count>0 ) {
                    allPiggybacks.AddRange(arr);
                }
            }

            // Finally, send off the LeafSetFixup message
            LeafSetFixup fixup = new LeafSetFixup(localAddrBind, dir, height, allPiggybacks, background);
            RouteMsg routeMsg = new RoutePayloadControlMsg(localAddrBind, fixup);
            localNode.RouteMessageInternal(routeMsg);
        }
    }
    
    
    /// <summary>
    /// A comparator used to sort leaf set nodes in order of closeness to the given destination
    /// </summary>
    class LeafSetComparer : IComparer {
        NodeId dest;
        public LeafSetComparer( NodeId dest ) { this.dest = dest; }

        public int Compare( object px, object py ) {
            NodeId x = ((NodeIdAddressBind) px).NodeId;
            NodeId y = ((NodeIdAddressBind) py).NodeId;

            int xcmp = x.CompareTo(dest); if( xcmp<0 ) xcmp=-1; if( xcmp>0 ) xcmp=1;
            int ycmp = y.CompareTo(dest); if( ycmp<0 ) ycmp=-1; if( ycmp>0 ) ycmp=1;

            // Handle case of x equal to dest
            if( xcmp==0 ) {
                // If y equal to dest it's a tie, otherwise x wins
                return ( ycmp==0 ) ? 0 : -1;
            }

            // Handle case of both lesser or both greater than dest
            if( xcmp==ycmp ) {
                if( xcmp==-1 ) return - x.CompareTo(y);
                if( xcmp==1 )  return   x.CompareTo(y);
            }

            // Handle case of one lesser or one greater
            Debug.Assert( (xcmp==-1 && ycmp==1) || (xcmp==1 && ycmp==-1) );

            // Without loss of generality, make x lesser and y greater
            // (but keep track of whether we made the switch)
            int result, switchedXfactor=1;
            if( xcmp==1 ) { NodeId t=x; x=y; y=t; switchedXfactor=-1; }
                
            // Use the final, complicated rules to choose the destination:
            // no wrapping, longest prefix match, etc.
            int xPrefixMatch = dest.LongestCommonPrefixLength(x);
            int yPrefixMatch = dest.LongestCommonPrefixLength(y);

            // The destination is the node with the longer prefix match with the destination.
            if( xPrefixMatch>yPrefixMatch ) {
                result = -1;
            } else if( yPrefixMatch>xPrefixMatch ) {
                result = 1;
            } else {
                // If the prefix length are equal, go to the right.
                result = 1;
            }

            return result * switchedXfactor;
        }
    }
    
    
    /// <summary>
    /// Implements one half of a leaf set (either left or right)
    /// </summary>
    [Serializable, XmlRootAttribute("LeafSetHalf",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class LeafSetHalf : RoutingSetHalf {
        
        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        
        int         numPingsSent;
        int         numRepliesRecvd;
        bool        anyUpdates;
        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        public LeafSetHalf() {}

        public LeafSetHalf( NodeIdAddressBind localAddrBind, Node localNode, Direction dir, int maxSize,
            bool authoritative ) : base(localNode, localAddrBind, dir, maxSize, authoritative )
        {
            numPingsSent = numRepliesRecvd = 0;
            anyUpdates = false;
        }
        

        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************
        
        /// <summary>
        /// Internal routing mechanism. Determines the next hop in routing towards the destination.
        /// If the option is PrefixMatch, any node that has a prefix match with the destination is ok.
        /// If the option is DisallowEqual, the destination must be strictly between the localID and the destID.
        /// If bottomLevel is true, this will return null if it cannot route to the destination in one hop.
        /// If bottomLevel is true, this will use the complicated rules for determining the next hop in the
        /// case of a destination name that lies between two nodes.
        /// 
        /// NOTE: This should only be called by LeafSet.
        /// </summary>
        /// <returns>
        /// If the local node is the final destination, returns the local address.
        /// Otherwise, returns the address of the next hop.
        /// </returns>
        public NodeIdAddressBind Route(NodeId destID, RouteByNameOption option, bool bottomLevel) {
            // TODO: What about the special case when nbrs.Count==0?

            // Find the best next hop in this leafset half.
            NodeIdAddressBind prev = localAddrBind;
            for( int i=0; i<nbrs.Count; i++ ) {
                NodeIdAddressBind leaf = (NodeIdAddressBind) nbrs[i];
                int cmp = leaf.NodeId.CompareTo(destID);

                // PrefixMatch option: Any node with a prefix match is OK
                if(   option==RouteByNameOption.PrefixMatch
                    && leaf.NodeId.IsWithinPrefix(destID) ) {
                    return leaf;
                }
                
                // DisallowEqual option: If this node is equal then we want the previous node
                if( option==RouteByNameOption.DisallowEqual && cmp==0 ) {
                    return prev;
                }

                // Check if the leaf matches the destination exactly.
                if( option==RouteByNameOption.Normal && cmp==0 ) {
                    return leaf;
                }
                
                // Check if the destination lies between the previous node and the i'th leaf
                if(   (   option==RouteByNameOption.Normal
                       || option==RouteByNameOption.PrefixMatch
                       || option==RouteByNameOption.DisallowEqual )
                   && (NodeId.LiesBetweenDir(dir, prev.NodeId, destID, leaf.NodeId) )
                ) {

                    if( bottomLevel ) {
                        // Use the final, complicated rules to choose the destination:
                        // no wrapping, longest prefix match, etc.
                        int prevPrefixMatch = destID.LongestCommonPrefixLength(prev.NodeId);
                        int leafPrefixMatch = destID.LongestCommonPrefixLength(leaf.NodeId);

                        // If going from the previous node to the leaf causes a wrap, the previous
                        // node is the destination.
                        if( Wrap(dir,prev.NodeId,leaf.NodeId) ) {
                            return prev;
                        }

                        // The destination is the node with the longer prefix match with the destination.
                        // If the prefix length are equal, go to the right.
                        if( prevPrefixMatch>leafPrefixMatch ) {
                            return prev;
                        } else if( leafPrefixMatch>prevPrefixMatch ) {
                            return leaf;
                        } else {
                            if( Direction.Left==dir ) {
                                return prev;
                            } else {
                                return leaf;
                            }
                        }
                    } else {
                        // This is not the bottom level of the routing table. We can
                        // use the simple rules to choose the next hop.
                        return prev;
                    }
                }

                prev = leaf;
            }
            
            if( bottomLevel ) {
                // No appropriate leaf found. Unable to route in one hop.
                return null;
            } else {
                // Use the farthest hop.
                return prev;
            }
        }
        
        /// <summary>
        /// Send 'simple pings' to all neighbours at the given height, in the given direction.
        /// </summary>
        /// <param name="background">[Simulator] If true, the ping should be treated as
        /// a background message. If false, the ping is a foreground message</param>
        public void SendSimplePings( int height, bool background ) {
            // First determine the set of nodes that need to be pinged
            ArrayList nbrs = GetAllNbrs();

            // Pings must timeout before the next batch of pings that we send
            Debug.Assert( LeafSetSimplePing.TIMEOUT < Parameters.neighbourPingPeriod );

            // Send off a ping message to each node that is to be pinged
            foreach( NodeIdAddressBind nbr in nbrs ) {
                // First find out if any apps want to piggyback objects in this ping
                ArrayList piggybackObjs = localNode.PingNotification(nbr);

                // Send out a ping message. This is a request-reply message, but we
                // don't do any retries
                LeafSetSimplePingCallback cb = new LeafSetSimplePingCallback(nbr);
                LeafSetSimplePing ping = new LeafSetSimplePing(localNode, height, dir,
                    nbr, piggybackObjs, cb, background);
                ping.MaxRetries = 0;
                localNode.SendBackgroundMessageDirect(ping, nbr);
            }
            
            numPingsSent = nbrs.Count;
            numRepliesRecvd = 0;
            anyUpdates = false;
        }

        
        //**********************************************************************************************
        // Private Methods
        //**********************************************************************************************

        /// <summary>
        /// Determine if node B lies strictly between nodes A and C, taking into
        /// account the direction of this leaf-set half.
        /// </summary>
        protected override bool LiesBetween( NodeIdAddressBind A, NodeIdAddressBind B, NodeIdAddressBind C ) {
            return NodeId.LiesBetweenDir( dir, A.NodeId, B.NodeId, C.NodeId );
        }

        /// <summary>
        /// Determine if going from node A to node B in direction 'dir' causes a wrap.
        /// </summary>
        private bool Wrap( Direction dir, NodeId A, NodeId B ) {
            if( Direction.Left==dir ) {
                return( A.CompareTo(B)<0 );
            } else {
                return( A.CompareTo(B)>0 );
            }
        }
    }


    /// <summary>
    /// This message is used as a 'simple ping' between nodes to keep the R-Table rings consistent
    /// </summary>
    [Serializable, XmlRootAttribute("LeafSetSimplePing", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class LeafSetSimplePing : OverlayRequestMsg {
        
        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        public const string         CATEGORY = "SimplePing";
        public static int           TIMEOUT = 45*1000;
        public static int NumSimplePingsSent = 0;
        public static int NumSimplePingsRcvd = 0;


        //**********************************************************************************************
        // Serialized Members
        //**********************************************************************************************

        public NodeIdAddressBind    source;         // The node sending the ping
        public int                  height;         // The height of the ring that this ping operates on
        public Direction            dir;            // The direction in that ring
        public NodeIdAddressBind    dest;           // The destination node
        public PiggybackWrapper[]   piggybacks;     // Objects that are piggybacked on this message
        public bool                 background;     // If true, this message is considered a 'background'
                                                    // message by the simulator

        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor for serialization
        public LeafSetSimplePing() : base() {}

        public LeafSetSimplePing( Node localNode, int height, Direction dir, NodeIdAddressBind dest,
            ArrayList piggybacks, LeafSetSimplePingCallback cb, bool background )
            : base( localNode, cb, TIMEOUT, true )
        {
            Debug.Assert( ! Parameters.smartPingEnabled );

            this.source = localNode.NodeIdAddressBind;
            this.height = height;
            this.dir = dir;
            this.dest = dest;
            this.background = background;

            if( piggybacks!=null && piggybacks.Count>0 ) {
                this.piggybacks = (PiggybackWrapper[]) piggybacks.ToArray( typeof(PiggybackWrapper) );
            }
            NumSimplePingsSent++;
        }


        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************

        /// <summary>
        /// This method is called when another node has added us to its CLB Table
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node node ) {
            Node localNode = (Node) node;
            NumSimplePingsRcvd++;

            Log.Write(CATEGORY, LogLvl.Lo, "Received LeafSetSimplePing msg from {0}", source.GetAddress.ToString());

            // Verify that this node should indeed be in this ring
            int commonH = source.NumericId.LongestCommonPrefixLength( localNode.NodeIdAddressBind.NumericId );
            Debug.Assert( commonH>=height && commonH<Parameters.maxHeight );

            // Add the source into our leafset at this height
            LeafSet ls = localNode.RoutingTable.Neighbours[height];
            ls.AddNode(source);

            // Send back a reply to the source
            OverlayReplyMsg reply = new LeafSetSimplePingReply(this);
            localNode.SendBackgroundReply(reply, source);

            HandlePiggybacks( localNode );
        }


        //**********************************************************************************************
        // Private methods
        //**********************************************************************************************

        /// <summary>
        /// Process any objects that have been piggybacked onto this ping message.
        /// </summary>
        private void HandlePiggybacks( Node localNode ) {
            if( piggybacks==null ) {
                return;
            }

            Debug.Assert( piggybacks.Length>0 );
            foreach( PiggybackWrapper w in piggybacks ) {
                Debug.Assert( w.dest.Equals(localNode.NodeIdAddressBind) );
                localNode.InvokePingReceivedUpcall(source, w.key, w.piggybackObj);
            }
        }
    }


    /// <summary>
    /// Reply to a LeafSetSimplePing. Doesn't need to contain anything.
    /// </summary>
    [Serializable, XmlRootAttribute("LeafSetSimplePingReply", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class LeafSetSimplePingReply : OverlayReplyMsg 
    {
        public LeafSetSimplePingReply() : base() {} // For XML Serializer
        public LeafSetSimplePingReply( LeafSetSimplePing ping ) : base(ping) {}
    }


    /// <summary>
    /// Callback that is invoked when a LeafSetSimplePingReply is received, or the reply times out
    /// </summary>
    public class LeafSetSimplePingCallback : OverlayReplyCallback {
        Node                    localNode;
        NodeIdAddressBind       dest;
        bool                    appLevelPing;

        public LeafSetSimplePingCallback( NodeIdAddressBind dest ) {
            this.dest = dest;
            this.appLevelPing = true;
        }
        
        /// <summary>
        /// This method is invoked in two ways:
        ///  1) When we receive a reply for a LeafSetSimplePing, or the reply times out
        ///  2) If the low-level ping fails (via TopoPingCallback)
        /// </summary>
        /// <param name="node"></param>
        public override void Execute( MessageRouting.Node node ) {
            localNode = (Node) node;

            Debug.Assert( dest!=null );

            // Initially, this callback is used for the app-level ping
            if( appLevelPing ) {
                if( Reply==null ) {
                    // This section is invoked if the app-level SimplePing failed
                    localNode.Topology.BeginPing(localNode.NodeIdAddressBind.GetAddress,
                        dest.GetAddress, true, new PingCallback(TopoPingCallback));
                }
            } else {
                // This section is invoked if Topology.BeginPing() failed
                localNode.NodeFailed(dest);
            }
        }

        /// <summary>
        /// This is the callback for Topology.BeginPing()
        /// </summary>
        public void TopoPingCallback(NodePhysicalAddress destAddr, int dist) {
            if( dist<0 ) {
                appLevelPing = false;
                localNode.ScheduleContinuation( this );
            }
        }
    }


    /// <summary>
    /// Fix up a leaf set in the given direction at the given height.
    /// </summary>
    [Serializable, XmlRootAttribute("LeafSetFixup", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class LeafSetFixup : OverlayMsg, RouteControl {
        
        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        public const string         CATEGORY = "LeafSetFixup";
        public static int NumLeafSetFixupSent = 0;
        public static int NumLeafSetFixupRcvd = 0;


        //**********************************************************************************************
        // Data members
        //**********************************************************************************************
        public  NodeIdAddressBind   source;             // The source of this RingFixup message
        public  bool                firstHop;           // Is this the first hop of the message?
        public  bool                complete;           // Is the fixup complete?
        public  int                 desiredNbrs;        // The desired number of neighbours per level
        public  Direction           dir;                // The direction in which the source is fixing neighbours
        public  int                 height;             // The level at which the source started the fixup
        public  bool                background;         // Indication to the simulator if this is a background message
        public PiggybackWrapper[]          piggybacks;  // Objects that are piggybacked on this message

        public ArrayList            liveNbrs;           // The neighbours we have found so far

        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor needed for serialization
        public LeafSetFixup() {}
        
        /// <summary>
        /// Fixup the leaf set in the given direction starting at the given height.
        /// </summary>
        /// <param name="localAddrBind">The source node</param>
        /// <param name="dir">The direction</param>
        /// <param name="height">The height to start at</param>
        /// <param name="piggybacks">An array of PiggybackWrapper objects to be delivered by this ping</param>
        /// <param name="background">Indicates if this message should be treated as a 'background'
        /// message by the simulator</param>
        public LeafSetFixup( NodeIdAddressBind localAddrBind, Direction dir, int height, 
            ArrayList piggybacks, bool background )
        {
            Debug.Assert( Parameters.smartPingEnabled );

            this.source = localAddrBind;
            this.firstHop = true;
            this.complete = false;
            this.dir = dir;
            this.height = height;
            if( piggybacks!=null && piggybacks.Count>0 ) {
                this.piggybacks = (PiggybackWrapper[]) piggybacks.ToArray( typeof(PiggybackWrapper) );
            }
            this.background = background;
            
            int nbrSetSize = Parameters.denseRTable ? Parameters.b-1 : 1;
            if( height==0 ) {
                nbrSetSize = Math.Max( nbrSetSize, Parameters.leafSetHalfSize );
            }
            Debug.Assert( nbrSetSize>0 );       // Each level must have non-zero maximum size
            this.desiredNbrs = nbrSetSize;       // 2*n+1 also could be a good choice here
            liveNbrs = new ArrayList(desiredNbrs);

            NumLeafSetFixupSent++;
        }


        //**********************************************************************************************
        // Route Control methods
        //**********************************************************************************************

        /// <summary>
        /// Determine the next hop for this message
        /// </summary>
        public NodeIdAddressBind GetNextHop( Node localNode ) {
            NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;

            Log.Write(CATEGORY, LogLvl.Lo, "Received LeafSetFixup. Src={0}, Nbrs={1}",
                (source == null ? null : source.GetAddress.ToString()), liveNbrs.Count); 

            // If this is the first hop, treat it specially
            if( firstHop ) {
                firstHop = false;
                NodeIdAddressBind nbr = WhereNext( localNode );
                Debug.Assert( nbr!=null );

                if( nbr.Equals(localAddrBind) ) {
                    // We have no neighbours to traverse. The fixup was vacuously successful.
                    return null;
                } else {
                    Log.Write(CATEGORY, LogLvl.Med, "Sending LeafSetFixup to {0}", nbr.GetAddress.ToString());
                    return nbr;
                }
            }

            // If the source has failed, drop the message
            if( source==null ) {
                return null;
            }

            HandlePiggybacks( localNode );

            // If the message is complete, route back to the source.
            // If we have already arrived at the source, don't route any further.
            if( this.complete || localAddrBind.Equals(source) ) {
                return source;
            }

            VerifyCurrentNode( localNode );

            // Add this node to the set of neighbours in the message
            if( ! liveNbrs.Contains(localAddrBind) ) {
                liveNbrs.Add(localAddrBind);
            }

            // Update the local node's neighbours in this ring, using the nodes encountered so far
            UpdateLocalNbrs( localNode );
            
            // If we have accumulated enough neighbours, return to the source
            if( liveNbrs.Count>=desiredNbrs ) {
                this.complete = true;
                Log.Write(CATEGORY, LogLvl.Med, "Complete. Returning to source {0}", source.GetAddress.ToString());
                return source;
            }
       
            // It is possible source may have neighbours in the opposite direction of this message
            // that were not encountered by this message. Look for them now.
            ArrayList myNbrs = GetOppositeNbrs( localNode );
            foreach( NodeIdAddressBind nbr in myNbrs ) {
                if( LiesBetween(source, nbr, localAddrBind) && !liveNbrs.Contains(nbr) ) {
                    // Found a node that was not encountered yet but should have been.
                    // Try to route there next.
                    Log.Write(CATEGORY, LogLvl.Hi, "Node {0} was not encountered but should have been", nbr.GetAddress.ToString());
                    return nbr;
                }
            }
                        
            // Get our next neighbour at this level
            NodeIdAddressBind nextHop = WhereNext( localNode );

            // If there is no next neighbour, or the source should be between this node and
            // the next neighbour, then this message is complete.
            if( nextHop.Equals(localAddrBind) || LiesBetween(localAddrBind, source, nextHop) ) {
                this.complete = true;
                Log.Write(CATEGORY, LogLvl.Med, "No more nbrs. Returning to source {0}", source.GetAddress.ToString());
                return source;
            } else {
                Log.Write(CATEGORY, LogLvl.Med, "Sending RingFixup to {0}", nextHop.GetAddress.ToString());
                return nextHop;
            }
        }
                    
        /// <summary>
        /// After each time SkipNet tries to send this message to another node,
        /// this callback notifies the message whether the send was successful or not.
        /// </summary>
        public void SendCallback( Node localNode, NodeIdAddressBind dest, bool success ) {
            if( ! success ) {
                Log.Write(CATEGORY, LogLvl.Med, "Send failed to node {0}", dest.GetAddress.ToString());

                if( source.Equals(dest) ) {
                    source = null;
                }

                if( liveNbrs.Contains(dest) ) {
                    liveNbrs.Remove(dest);
                }
            }
        }

        /// <summary>
        /// An indication to the simulator whether this is a 'background' message or not.
        /// Background messages are ignored by "ProcessUntilIgnoreEventsOnly()".
        /// </summary>
        public bool IsBackground() {
            return background;
        }


        //**********************************************************************************************
        // OverlayMsg functions
        //**********************************************************************************************

        /// <summary>
        /// The message has arrived at its final destination.
        /// We now examine the set of live neighbours that were discovered,
        /// add them to our routing table, and remove any dead neighbours.
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node node ) {
            NumLeafSetFixupRcvd++;
            Node localNode = (Node) node;
            NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;

            Log.Write(CATEGORY, LogLvl.Med, "LeafSetFixup complete!");
            
            ArrayList myNbrs = GetAllMyNbrs( localNode );
            LeafSetFixupCont cont = new LeafSetFixupCont(localNode, myNbrs, liveNbrs, this);
            cont.LoopTop();
        }

        /// <summary>
        /// Continuation used when a RingFixup is complete.
        /// </summary>

        /// <summary>
        /// Continuation used when a RingFixup is complete.
        /// </summary>
        class LeafSetFixupCont : Continuation {
            Node                localNode;
            ArrayList           myNbrs, liveNbrs;
            LeafSetFixup        fixupMsg;

            bool                anyChanges;
            NodeIdAddressBind   nbr;
            int                 i, distance;

            public LeafSetFixupCont( Node localNode, ArrayList myNbrs, ArrayList liveNbrs,
                LeafSetFixup fixupMsg )
            {
                this.localNode = localNode;
                this.myNbrs = myNbrs;
                this.liveNbrs = liveNbrs;
                this.fixupMsg = fixupMsg;

                this.anyChanges = false;
                this.i = 0;
            }

            // First, we loop over the set: (Our Neighbours) \ (Live Neigbours in Message)
            // and explicitly check if those nodes are alive. If not, remove them.

            // The top of the loop in which we check if our neighbours are dead
            public void LoopTop() {
                NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;
                Topology topo = localNode.Topology;

                Top:;

                if( i<myNbrs.Count ) {
                    nbr = (NodeIdAddressBind) myNbrs[i];
                    if( liveNbrs.Contains(nbr) ) {
                        i++;
                        goto Top;
                    }
                    Log.Write(LeafSetFixup.CATEGORY, LogLvl.Hi, "Node {0} is missing from LiveNbrs", nbr.GetAddress.ToString());

                    PingCallback cb = new PingCallback(this.Callback);
                    topo.BeginPing(localAddrBind.GetAddress, nbr.GetAddress, true, cb);
                } else {
                    LoopDone();
                }
            }

            // Note: Not in a thread-safe state. Schedule the "Execute" function to be called.
            void Callback(NodePhysicalAddress destination, int distance) {
                this.distance = distance;
                localNode.ScheduleContinuation(this);
            }

            // Called from the event-loop when the ping has completed.
            // This is the latter half of the loop.
            public override void Execute( MessageRouting.Node node ) {
                Node localNode = (Node) node;

                if( distance==-1 ) {
                    Log.Write(LeafSetFixup.CATEGORY, LogLvl.Hi, "Ping failed! Node {0} is dead!", nbr.GetAddress.ToString());
                    anyChanges |= localNode.NodeFailed(nbr);
                } else {
                    Log.Write(LeafSetFixup.CATEGORY, LogLvl.Hi, "Ping succeeded. Node {0} is alive.", nbr.GetAddress.ToString());
                }
                i++;
                LoopTop();
            }
            
            // Called when the loop checking our neighbours is complete
            void LoopDone() {
                // Update our neighbours at this level (adding live nodes only)
                fixupMsg.UpdateLocalNbrsFinal( localNode, anyChanges );
            }
        }
        
        //**********************************************************************************************
        // Private methods
        //**********************************************************************************************
        
        /// <summary>
        /// Determine which node to visit next
        /// </summary>
        private NodeIdAddressBind WhereNext( Node localNode ) {
            RoutingTable rt = localNode.RoutingTable;
            return rt.GetNeighbour(dir, height);
        }

        /// <summary>
        /// Process any objects that have been piggybacked onto this ping message.
        /// </summary>
        private void HandlePiggybacks( Node localNode ) {
            if( piggybacks==null ) {
                return;
            }

            Debug.Assert( piggybacks.Length>0 );
            foreach( PiggybackWrapper w in piggybacks ) {
                if( w.dest.Equals(localNode.NodeIdAddressBind) ) {
                    localNode.InvokePingReceivedUpcall(source, w.key, w.piggybackObj);
                }
            }
        }

        /// <summary>
        /// Sanity check to verify that this node should be in this leaf set.
        /// </summary>
        private void VerifyCurrentNode( Node localNode ) {
            int commonH = source.NumericId.LongestCommonPrefixLength(localNode.NodeIdAddressBind.NumericId);
            Debug.Assert( commonH>=height && commonH<Parameters.maxHeight,
                "Numeric ID collision. Either two nodes have the same name or Parameters.maxHeight is too small");
        }

        /// <summary>
        /// Update the local node's neighbours in this ring, using the nodes encountered so far
        /// </summary>
        private void UpdateLocalNbrs( Node localNode ) {
            RoutingTable rt = localNode.RoutingTable;

            // Generate a new array containing the source too
            ArrayList nbrsAndSource = new ArrayList( liveNbrs );
            nbrsAndSource.Add( source );

            // Add the live nodes (plus source) to the leaf set at this level
            rt.AddNodes(this.height, nbrsAndSource);

            // In aggressive mode, add the live nodes to the leaf set at lower levels too
            if( Parameters.aggressiveFixups ) {
                for( int h=this.height-1; h>=0; h-- ) {
                    rt.AddNodes(h, liveNbrs);
                }
            }
        }

        private ArrayList GetOppositeNbrs( Node localNode ) {
            RoutingTable rt = localNode.RoutingTable;
            LeafSet ls = rt.Neighbours[height];
            Direction opposite = Node.OppositeDir(dir);
            ArrayList myNbrs = ls.GetHalf(opposite).GetAllNbrs();
            return myNbrs;
        }

        private bool LiesBetween( NodeIdAddressBind A, NodeIdAddressBind B, NodeIdAddressBind C ) {
            return NodeId.LiesBetweenDir(dir, A.NodeId, B.NodeId, C.NodeId);
        }

        private ArrayList GetAllMyNbrs( Node localNode ) {
            LeafSet ls = localNode.RoutingTable.Neighbours[ this.height ];
            ArrayList myNbrs = ls.GetHalf(dir).GetAllNbrs();
            return myNbrs;
        }

        private void UpdateLocalNbrsFinal( Node localNode, bool anyChanges ) {
            NodeIdAddressBind   localAddrBind = localNode.NodeIdAddressBind;
            RoutingTable        rt = localNode.RoutingTable;
        
            // Update our neighbours at this level (adding live nodes only)
            anyChanges |= rt.AddNodes(height, liveNbrs);
        
            // While we're at it, update our neighbours at lower levels too
            if( Parameters.aggressiveFixups ) {
                for( int h=height-1; h>=0; h-- ) {
                    rt.AddNodes(h, liveNbrs);
                }
            }

            // If any changes were made at this level, we also fixup the next higher level
            if( anyChanges ) {
                FixupHigher fixupHigher = new FixupHigher(localAddrBind, height, dir, background);
                RoutePayloadControlMsg routeMsg = new RoutePayloadControlMsg(localAddrBind, fixupHigher);
                localNode.RouteMessageInternal(routeMsg);
            }
        }
 
    }


    /// <summary>
    /// Wrapper used to insert piggybacked objects into a LeafSetFixup.
    /// </summary>
    [Serializable, XmlRootAttribute("PiggybackWrapper", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class PiggybackWrapper {
        public NodeIdAddressBind dest;
        //public Object key;
        //public Object piggybackObj;
        public int key;
        public byte[] piggybackObj;


        [XmlArray( ElementName="PiggybackObjs",
             Namespace=SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="PiggybackObjsElem", Type=typeof(byte),
             IsNullable=false, Namespace=SerializationConfig.SKIPNET_NAMESPACE)]
        public byte[] PiggybackObj 
        {
            get { return piggybackObj; }
            set { piggybackObj = value; }
        }
 
        public PiggybackWrapper() {}

        public PiggybackWrapper( NodeIdAddressBind dest, Object key, Object piggybackObj ) {
            this.dest = dest;
            this.key = (int) key;
            this.piggybackObj = (byte[]) piggybackObj;
        }
    }

}
