using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// Implements a SkipNet Node.
    /// </summary>
    public class Node : MessageRouting.Node {

        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************

        // These three objects define the message-delivery layer beneath us
        LocalEvent                      localEvent;
        SendMessage                     send;
        Topology                        topo;

        // Information about the local node
        string                          nodeName = null;
        NodePhysicalAddress             localAddr;
        NodeIdAddressBind               localAddrBind;
        
        // Routing Tables
        RoutingTable                    routingTable;
        ProximityTable                  proximityTable;
        CTable                          cTable;
        CLBTable                        clbTable;
        
        internal bool                   proxRebuildEnabled=true;
        internal bool                   clbRebuildEnabled=true;
        internal bool                   pingEnabled=true;
        internal NodePhysicalAddress    seed;

        // Misc.
        Log                             log;
        Hashtable                       pingSendUpcalls;
        Hashtable                       pingReceiveUpcalls;


        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        public const string             CATEGORY="SkipNetNode";
        internal const long             TIMEOUT_DELAY_MS = 2 * 60 * 1000; // Try allowing joins to take up to 2 minutes
        const int                       MAXHOPS=256;
        
        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
    
        // Standard Constructor
        public Node(LocalEvent local, SendMessage output, Topology top, NodePhysicalAddress myAddr,
            NodePhysicalAddress seed, string identifier) : base()
        {
            this.nodeName = identifier;
            InitNode(local, output, top, myAddr, seed);
        }
    
        // Constructor to be used when InitNode() is directly called later.
        public Node(string identifier) : base() {
            this.nodeName = identifier;
        }
    
        public Node() : base() {}
        
        
        //**********************************************************************************************
        // Properties
        //**********************************************************************************************

        // Inherited from MessageRouting
        public override MessageRouting.NodeIdAddressBind GetNodeIdAddressBind { 
            get { return localAddrBind; }
        } 

        // Inherited from MessageRouting
        public override NodePhysicalAddress GetPhysicalAddr {
            get { return localAddr; }
        }

        public NodeIdAddressBind NodeIdAddressBind { 
            get { return localAddrBind; }
        } 

        public RoutingTable RoutingTable {
            get { return routingTable; }
            set { routingTable=value; }
        }

        internal ProximityTable ProximityTable {
            get { return proximityTable; }
        }

        internal CTable CTable {
            get { return cTable; }
        }

        internal CLBTable CLBTable {
            get { return clbTable; }
        }

        public Topology Topology {
            get { return topo; }
        }

   
        //**********************************************************************************************
        // Initialization code
        //**********************************************************************************************
        
        /// <summary>
        /// Initialize a node and either join an existing net or start a new net.
        /// NOTE: After calling InitNode, you must call ActivateNode to activate it.
        /// </summary>
        /// <param name="local">Used for queueing messages to ourself</param>
        /// <param name="output">Used for sending messages to other nodes</param>
        /// <param name="top">Used to measure network proximity</param>
        /// <param name="myAddr">Our local physical address</param>
        /// <param name="seedNodeAddr">Our seed node for joining the net (optional)</param>
        public override void InitNode(LocalEvent local, SendMessage output, Topology topo,
            NodePhysicalAddress myAddr, NodePhysicalAddress seedNodeAddr)
        {
            log = new Log(CATEGORY);

            // Store the MessageRouting objects that were passed as parameters
            this.localEvent = local;
            this.send = output;
            this.topo = topo;

            // Initialize the local node's identifiers
            this.localAddr = myAddr;
            if( this.localAddrBind==null ) {
                NumericId numericId = new NumericId();
                numericId.GenerateFromString(this.nodeName);
                this.localAddrBind = new NodeIdAddressBind(myAddr, new NodeId(this.nodeName), numericId);

                log.Write(LogLvl.Med, "New Node. NameID: {0}, NumericID: {1}", this.nodeName, numericId.ToString());
            }

            // Initialize the various tables
            this.routingTable = new RoutingTable(this);
            if( Parameters.proximityTableEnabled ) {
                this.proximityTable = new ProximityTable(this);
            }
            if( Parameters.cTableEnabled ) {
                this.cTable = new CTable(this);
            }
            if( Parameters.clbTableEnabled ) {
                this.clbTable = new CLBTable(this);
            }
            this.pingSendUpcalls = new Hashtable();
            this.pingReceiveUpcalls = new Hashtable();
            
            // Join / start a SkipNet
            this.seed = seedNodeAddr;
            if( seedNodeAddr!=null ) {
                // Join an existing SkipNet
                StartJoin(seedNodeAddr);
            }
        }

        /// <summary>
        /// This is essentially the second half of InitNode(). It is separated out so that
        /// the simulator can 'Activate' a node after all nodes have joined the SkipNet.
        /// Schedules periodic tasks, etc.
        /// </summary>
        /// <param name="immediate">If true, periodic tasks start immediately</param>
        public void ActivateNode( bool immediate ) {
            double delayFactor;

            // If 'immediate' is true, start periodic tasks immediately
            if( immediate ) {
                delayFactor = 0;
            } else {
                delayFactor = NumericId.RandomDouble();
            }

            // Periodically ping our neighbours, if enabled
            if( Parameters.pingEnabled ) {
                pingEnabled = true;
                routingTable.SchedulePing( delayFactor );
            }
            
            // Periodically rebuild the P-Table, if enabled
            if( Parameters.proximityTableEnabled ) {
                SchedulePTableRebuild( delayFactor );
            }

            // Periodically rebuild the C-Table, if enabled
            if( Parameters.cTableEnabled ) {
                cTable.SchedulePing( delayFactor );
            }

            // Periodically rebuild the CLB table, if enabled
            if( Parameters.clbTableEnabled ) {
                clbRebuildEnabled = true;
                clbTable.ScheduleUpdate( delayFactor );
            }
        }
                
        /// <summary>
        /// Deactivate the node: stop P-Table rebuild, CLB Table rebuild, pings, etc.
        /// </summary>
        public void DeactivateNode() {
            proxRebuildEnabled = false;
            clbRebuildEnabled = false;
            pingEnabled = false;
        }


        //**********************************************************************************************
        // Join code
        //**********************************************************************************************
        
        /// <summary>
        /// Start the process of joining an existing net.
        /// </summary>
        public void StartJoin(NodePhysicalAddress seed) {
            NodeIdAddressBind seedAddrBind = new NodeIdAddressBind(seed, null, null);
            
            // Send a Join request to our seed node
            JoinRequest join = new JoinRequest(this, TIMEOUT_DELAY_MS);
            SendMessageDirect(join, seedAddrBind);
        
            // If the CLB Table is enabled, send a CLB Join request to build it
            if( Parameters.cTableEnabled ) {
                cTable.Rebuild();
            }        
        }


        //**********************************************************************************************
        // Message Routing code
        //**********************************************************************************************
        
        /// <summary>
        /// This function is the core of the SkipNet event-loop.
        /// It is called every time a new message is received from the message queue.
        /// This message may have come from another node, or it may be a local event, 
        /// such as a continuation, or a timer.
        /// </summary>
        /// <param name="msg">The message object</param>
        protected override void RaiseEvent(MessageRouting.RouteMsg msg) {
            SkipNet.RouteMsg snMsg = (SkipNet.RouteMsg) msg;
            
            /*
            OverlayMsg payload = snMsg.Payload;
            string payloadType;
            if( payload==null ) {
                payloadType = "(null)";
            } else {
                payloadType = payload.GetType().ToString();     // Note: This call is expensive
            }
            log.Write(LogLvl.Lo, "Received {0} message, source={1}", payloadType, localAddrBind);*/
            
            ResumeRouting(snMsg);
        }

        /// <summary>
        /// Internal method used to route a new RouteMsg.
        /// </summary>
        /// <param name="snMsg">The message object</param>
        public override void RouteMessageInternal( MessageRouting.RouteMsg routeMsg ) {
            RouteMsg snMsg = (RouteMsg) routeMsg;
            Debug.Assert( snMsg.Payload!=null );

            // "Downcall" to the MessageRouting Node object
            MessageSourceHook( snMsg.Payload, snMsg );

            ResumeRouting( snMsg );
        }

        /// <summary>
        /// Called to continue routing a message. This function could be called when a message
        /// is plucked off the event queue, or it could be called indirectly (via RouteMessageInternal)
        /// when the local node or application sends a new message.
        /// </summary>
        /// <param name="msg">The message object</param>
        internal void ResumeRouting(RouteMsg snMsg) {
            bool                continueRouting;
            
            // For debugging / measurement purposes, keep a hop count and a
            // path of all nodes visited by this message
            if( ++snMsg.Hops>=MAXHOPS ) {
                log.Write(LogLvl.Hi, "***** SERIOUS ERROR *****");
                log.Write(LogLvl.Hi, "Too many hops! Probable routing loop or buggy message.");
                log.Write(LogLvl.Hi, "Msg type={0}, Payload type={1}", snMsg.GetType(), snMsg.Payload.GetType());
                Debugger.Break();
                return;
            }
            if( Parameters.recordMessagePath ) {
                snMsg.AddToPath( localAddrBind );
            }

            // Ask the message if we should keep routing
            continueRouting = snMsg.CallCCR(this);
            if( ! continueRouting ) {
                return;
            }
            
            // Try to send the message
            SendMessageCont continuation = new SendMessageCont(snMsg);
            continuation.Execute(this);
        }

        /// <summary>
        /// Continuation used to send messages (and retry sends as necessary).
        /// Also contains a callback that fires when the asynchronous send completes.
        /// </summary>
        private class SendMessageCont : Continuation {
            const int           MAXRETRIES=64;

            RouteMsg            msg;
            int                 retryCount;
        
            /// <summary>
            /// This stores the parameters needed for the continuation.
            /// </summary>
            public SendMessageCont(RouteMsg msg) {
                this.msg = msg;
                this.retryCount = 0;
            }

            /// <summary>
            /// This is the body of the continuation.
            /// </summary>
            public override void Execute(MessageRouting.Node node) {
                Node localNode = (Node) node;
                NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;
                NodeIdAddressBind nextHop;

                // Check that we haven't retried too many times
                if( retryCount>MAXRETRIES ) {
                    Debug.Assert(false, "Too many retries! This message is buggy.");
                    return;
                }
                
                // Find the next hop for this message
                nextHop = msg.GetNextHop(localNode);
            
                if( nextHop==null ) {
                    // If the next hop is null, this means to drop the message on the floor
                    return;
                } else if( nextHop.Equals(localAddrBind) ) {
                    // Message has arrived at its destination

                    // Print statistics for C-Table routing efficiency
                    if( msg is RouteByCLBMsg ) {
                        RouteByCLBMsg clbMsg = (RouteByCLBMsg) msg;
                        --clbMsg.rTableHops;
                        double sum = clbMsg.rTableHops + clbMsg.cTableHops;
                        double pct = (sum>0) ? (((double) clbMsg.cTableHops) / sum) : 1;
                        Log.Write(Node.CATEGORY, LogLvl.Lo, "CTable: {0}  RTable: {1}  Pct: {2:0.00}",
                            clbMsg.cTableHops, clbMsg.rTableHops, pct);
                    }

                    localNode.MessageArrivedHook( msg.Payload );
                } else {
                    // Send on to next hop
                    ReplyDirectCallback cb = new ReplyDirectCallback(this.Callback);
                    localNode.SendMessageAsync(nextHop, msg, cb);
                }
            }

            /// <summary>
            /// This is called from the event loop when the asynchronous send completes.
            /// </summary>
            private void Callback(MessageRouting.Node node, MessageRouting.NodeIdAddressBind dest,
                bool success, MessageRouting.RouteMsg sentMsg)
            {
                Node                localNode = (Node) node;
                NodeIdAddressBind   destination = (NodeIdAddressBind) dest;

                msg.SendCallback(localNode, destination, success);

                if( ! success ) {
                    // Failed to send the message. Try routing again.
                    retryCount++;
                    Execute(localNode);
                }
            }
        }

        /// <summary>
        /// Send the given message asynchronously to the given node. Invoke the callback when complete.
        /// </summary>
        /// <param name="nextHop">The node to send the message to</param>
        /// <param name="msg">The message to send</param>
        /// <param name="cb">Delegate that is called when send is complete. The delegate is
        /// called by the main event-loop so thread-safety is not a concern.
        /// This parameter may be null, in which case no callback is called.</param>
        internal void SendMessageAsync( NodeIdAddressBind nextHop, RouteMsg msg,
            ReplyDirectCallback userCallback )
        {
            AsyncSendCont sendCont = new AsyncSendCont(this, userCallback, msg.IsBackground());
            ReplyCallback cb = new ReplyCallback(sendCont.Callback);
            send.BeginSend(localAddrBind, nextHop, msg, cb);
        }

        /// <summary>
        /// Continuation that is called when the asynchronous send completes.
        /// </summary>
        private class AsyncSendCont : Continuation {
            // Parameters
            Node                    localNode;
            ReplyDirectCallback     cb;         // May be null.
            NodeIdAddressBind       dest;
            bool                    success;
            RouteMsg                sentMsg;
            bool                    background = false;

            // Constructor: save parameters
            public AsyncSendCont( Node localNode, ReplyDirectCallback cb ) {
                this.localNode = localNode;
                this.cb = cb;
            }
   
            // Constructor: save parameters
            public AsyncSendCont( Node localNode, ReplyDirectCallback cb, bool background ) {
                this.localNode = localNode;
                this.cb = cb;
                this.background = background;
            }
         
            // Called when the asynchronous send completes.
            // We can only do thread-safe operations at this point.
            internal void Callback( MessageRouting.NodeIdAddressBind dest, bool success,
                MessageRouting.RouteMsg sentMsg )
            {
                // Save the parameters and schedule this continuation to be called
                this.dest = (NodeIdAddressBind) dest;
                this.success = success;
                this.sentMsg = (RouteMsg) sentMsg;
                localNode.ScheduleContinuation(this);
            }

            // Called by the event-loop when the asynchronous send completes.
            // (Non-thread-safe operations may be used now.)
            public override void Execute( MessageRouting.Node node ) {
                Node localNode = (Node) node;

                // If the send failed, remove the destination node from our routing tables.
                if( ! success ) {
                    localNode.NodeFailed( dest );
                }

                // If the user specified a callback, call it.
                if( cb!=null ) {
                    cb(localNode, dest, success, sentMsg);
                }
            }

            // Determine if this is a "background" message or not.
            // Asynchronous Send Continuations are logically part of the message-delivery
            // process so they are not "background" messages.
            public override bool IsBackground() {
                return this.background;
            }
        }
        
        /// <summary>
        /// Schedule a message to be delivered to the local node after the given delay.
        /// This function is thread-safe.
        /// </summary>
        internal void ScheduleLocalMessage( long delayMs, LocalMsg msg ) {
            msg.Source = localAddrBind;
            localEvent.RaiseEvent(delayMs, this.localAddrBind, msg);
        }
        
        /// <summary>
        /// Given a destination name ID, determine the next hop to be taken
        /// in order to route towards that name ID.
        /// This is a bit of a hack, so that SV can disable the P-Table.
        /// </summary>
        /// <param name="destID">The destination name ID</param>
        /// <param name="allowEqual">Are we allowed to arrive at the node with name ID equal to destID?</param>
        /// <returns>The address of the next hop</returns>
        internal NodeIdAddressBind RouteByName( NodeId destID, RouteByNameOption option,
            bool pTableEnabled )
        {
            NodeIdAddressBind nextHop = localAddrBind;
            LeafSet leafSet = routingTable.Neighbours[0];

            // First check if the destination is in the leaf set.
            nextHop = leafSet.TryRouteOneHop(destID, option);
            if( nextHop!=null ) {
                // If so, route there directly.
                return nextHop;
            }
            Debug.Assert( ! destID.Equals(localAddrBind.NodeId) );
            
            // If the P-Table is enabled, try to use it to route.
            if( Parameters.proximityTableEnabled && pTableEnabled ) {
                nextHop = proximityTable.RouteByName(destID, option);
            }
            
            // If we had no luck with the P-Table, try the R-Table.
            if( nextHop==null || nextHop.Equals(localAddrBind) ) {
                nextHop = routingTable.RouteByName(destID, option);
            }
            
            return nextHop;
        }

        // RouteByName, by default allowing the P-Table to be used.
        internal NodeIdAddressBind RouteByName( NodeId destID, RouteByNameOption option ) {
            return RouteByName( destID, option, true );
        }

        
        //**********************************************************************************************
        // Message Routing APIs
        //**********************************************************************************************

        /// <summary>
        /// Route an overlay message using the overlay routing system.
        /// </summary>
        /// <param name="msg">The message to send</param>
        /// <param name="dest">A string representation of the CLB destination in "foo!bar" syntax</param>
        /// <param name="secure">Specifies if this message should be routed securely</param>
        public override void RouteMessage(OverlayMsg msg, string dest, bool secure) {
            CLBId destId = new CLBId(dest);
            RouteMessage(msg, destId, secure);
        }

        /// <summary>
        /// Route an overlay message using the overlay routing system.
        /// </summary>
        /// <param name="msg">The message to send</param>
        /// <param name="dest">A node ID representing the destination. This may be a
        /// CLBId or a SkipNet.NodeId</param>
        /// <param name="secure">Specifies if this message should be routed securely</param>
        public override void RouteMessage(OverlayMsg msg, MessageRouting.NodeId dest, bool secure) {
            // Create a RouteByCLBMsg to send the OverlayMsg through the overlay
            CLBId destId;
            if( dest is CLBId ) {
                destId = (CLBId) dest;
            } else {
                destId = new CLBId( dest.ToString() + "!nodeid" );
            }
            RouteByCLBMsg snMsg = new RouteByCLBMsg(localAddrBind, destId, msg);
            snMsg.forceRTable = secure;
            RouteMessageInternal( snMsg );
        }

        /// <summary>
        /// Send an overlay message directly to the given destination node.
        /// </summary>
        /// <param name="msg">The message to send</param>
        /// <param name="dest">The destination node</param>
        public override void SendMessageDirect(OverlayMsg msg, MessageRouting.NodeIdAddressBind dest) {
            // Create a RouteDirectMsg to send the OverlayMsg through the overlay
            RouteDirectMsg snMsg = new RouteDirectMsg(localAddrBind, (NodeIdAddressBind) dest, msg);
            RouteMessageInternal( snMsg );
        }

        /// <summary>
        /// Send an overlay message directly to the given destination node.
        /// Send as a background message.
        /// </summary>
        /// <param name="msg">The message to send</param>
        /// <param name="dest">The destination node</param>
        public override void SendBackgroundMessageDirect(OverlayMsg msg, MessageRouting.NodeIdAddressBind dest) {
            // Create a RouteDirectMsg to send the OverlayMsg through the overlay
            RouteDirectMsg snMsg = new RouteDirectMsg(localAddrBind, (NodeIdAddressBind) dest, true, msg);
            RouteMessageInternal( snMsg );
        }

        /// <summary>
        /// Schedule a continuation to execute.
        /// </summary>
        public override void ScheduleContinuation(Continuation cont) {
            LocalMsg snMsg = new LocalMsg(cont);
            ScheduleLocalMessage(0, snMsg);
        }

        /// <summary>
        /// Set a timer to execute.
        /// </summary>
        public override void SetTimer(Timer t) {
            LocalMsg snMsg = new LocalMsg(t);
            
            // If the timer is to run immediately, don't bother scheduling it
            if( t.NextDelayMs<=0 ) {
                ResumeRouting(snMsg);
            } else {
                ScheduleLocalMessage(t.NextDelayMs, snMsg);
            }
        }

        /// <summary>
        /// Cancel a pending timer.
        /// </summary>
        public override void CancelTimer(Timer t) {
            Debug.Assert( t!=null );
            LocalMsg snMsg = new LocalMsg(t);
            localEvent.RemoveEvent(snMsg);
        }

            
        //**********************************************************************************************
        // Proximity Code
        //**********************************************************************************************
        
        /// <summary>
        /// Start a periodic timer instructing us to start rebuilding the P-Table
        /// </summary>
        internal void SchedulePTableRebuild( double delayFactor ) {
            long initialDelayMs = (long) (delayFactor * Parameters.pTableUpdatePeriod);

            // Don't rebuild if disabled or if rebuild count has been exceeded
            if( ! proxRebuildEnabled
                || (SkipNet.Parameters.maxRebuilds>0 && proximityTable.InitCount>=SkipNet.Parameters.maxRebuilds)
            ) {
                return;
            }
            
            Timer rebuildTimer = new PTableRebuildTimer(initialDelayMs, Parameters.pTableUpdatePeriod);
            SetTimer(rebuildTimer);
        }
    

        //**********************************************************************************************
        // Neighbours
        //**********************************************************************************************

        /// <summary>
        /// This function is used to determine the addresses of the nodes in this
        /// node's leaf set.
        /// </summary>
        /// <returns>An array of NodeIdAddressBind objects</returns>
        public ArrayList GetLeafSet() {
            ArrayList leaves;
            
            leaves = routingTable.LeafSet.GetAllNbrs();

            return leaves;
        }

        /// <summary>
        /// Get the CLB neighbours for the given destination id.
        /// </summary>
        /// <param name="destId"></param>
        /// <returns>
        /// </returns>
        public ArrayList GetFailoverNeighbours( MessageRouting.NodeId destID ) {
            ArrayList nbrs;

            if( destID is CLBId ) {
                CLBId clbID = (CLBId) destID;

                // Get the CLB set for this destination's CLB domain
                string clbDomain = clbID.PrefixConstraint;

                // CLB domains often don't end with the necessary delimeter.
                // If this is the case, add the delimiter.
                if( clbDomain!="" && clbDomain[clbDomain.Length-1]!=NodeId.DELIMITER ) {
                    clbDomain += NodeId.DELIMITER;
                }

                CLBSet clbSet = clbTable[ clbDomain ];
                if( clbSet==null ) {
                    // If the user gave us a bogus clb domain, return null
                    return null;
                }

                nbrs = clbSet.GetNbrsForDest( clbID );
            } else {
                LeafSet ls = routingTable.Neighbours[0];
                NodeId snDestID = new NodeId( destID.ToString() );
                nbrs = ls.GetFailoverNeighbours( snDestID );
            }

            return nbrs;
        }

        /// <summary>
        /// This function is called when we fail to send a message to one of our neighbours.
        /// We notify all routing tables that this node has failed.
        /// 
        /// NOTE: This function calls "RebuildNeigbours()" if necessary.
        /// </summary>
        /// <param name="failedNode">The node that has failed</param>
        /// <returns>True if the node was removed</returns>
        public bool NodeFailed( NodeIdAddressBind failedNode ) {
            bool removed=false;

            // If the failedNode is not a complete NodeIdAddressBind (as would be
            // the case with the seed node), ignore this failure.
            if( failedNode.NodeId==null || failedNode.NumericId==null ) {
                return false;
            }

            if( routingTable!=null ) {
                removed |= routingTable.NodeFailed( failedNode );
            }
            if( proximityTable!=null ) {
                removed |= proximityTable.NodeFailed( failedNode );
            }
            if( cTable!=null ) {
                removed |= cTable.NodeFailed( failedNode );
            }
            if( clbTable!=null ) {
                clbTable.NodeFailed( failedNode );
            }

            // If the node was actually removed, update the table of all neighbours.
            if( removed ) {
                RebuildNeighbours(failedNode);
            }

            return removed;
        }

        /// <summary>
        /// This method is called when a new node is added to our _leaf set_.
        /// TODO: Do we need something like this for CLB?
        /// </summary>
        /// <param name="newNode">The new node</param>
        internal void NewNeighbour( NodeIdAddressBind newNode ) {
            Debug.Assert( appTable!=null );
            foreach( Application a in appTable.Values ) {
                a.NewNeighbour( newNode );
            }
        }

        internal void NeighbourRemoved(NodeIdAddressBind failedNode, bool failed)
        {
            // Inform higher-level applications about node failure
            Debug.Assert(appTable != null);
            foreach (Application a in appTable.Values) 
            {
                a.NeighbourRemoved(failedNode, failed);
            }
        }

        /// <summary>
        /// Populate the 'allNeighbours' table with all nodes in the R-Table, P-Table and C-Table.
        /// If failedNode != null then the rebuild was caused by failure of the specified node,
        /// allowing NeighbourRemoved() to pass the value true as the node failed parameter.
        /// </summary>
        internal void RebuildNeighbours(NodeIdAddressBind failedNode) {
            Hashtable oldAllNeighbours = allNeighbours;
            allNeighbours = new Hashtable();
            RoutingTable.GetAllNeighbours(allNeighbours);
            if( Parameters.proximityTableEnabled ) {
                ProximityTable.GetAllNeighbours(allNeighbours);
            }
            if( Parameters.cTableEnabled ) {
                CTable.GetAllNeighbours(allNeighbours);
            }

            if (oldAllNeighbours == null) oldAllNeighbours = new Hashtable();

            // Compare old and new tables, calling NeighbourRemoved() and NewNeighbour() when there are differences.
            foreach (object oldKey in oldAllNeighbours.Keys) 
            {
                if (! allNeighbours.Contains(oldKey)) 
                {
                    NodeIdAddressBind oldNeighbour = (NodeIdAddressBind) oldKey;
                    NeighbourRemoved(oldNeighbour, oldNeighbour.Equals(failedNode));
                }
            }
            foreach (object newKey in allNeighbours.Keys) 
            {
                if (! oldAllNeighbours.Contains(newKey)) 
                {
                    NodeIdAddressBind newNeighbour = (NodeIdAddressBind) newKey;
                    NewNeighbour(newNeighbour);
                }
            }
        }

        
        //**********************************************************************************************
        // Misc.
        //**********************************************************************************************

        // Create a node ID based on a given string
        public override MessageRouting.NodeId GenerateNodeId(string s) {
            if (s.IndexOf("!") >= 0)
                return new CLBId(s);
            else
                return new NodeId(s);
        }
        
        /// <summary>
        /// Convert left -> right or right -> left
        /// </summary>
        internal static Direction OppositeDir( Direction dir ) {
            if( dir==Direction.Left ) {
                return Direction.Right;
            } else {
                return Direction.Left;
            }
        }


        //**********************************************************************************************
        // Ping Piggybacking
        //**********************************************************************************************

        /// <summary>
        /// Register a delegate that will be called when SkipNet is about to
        /// ping a neighbour.
        /// </summary>
        /// <param name="pingSendUpcall">Upcall that is invoked when a ping is about to be sent</param>
        /// <param name="pingReceiveUpcall">Upcall that is invoked when a piggybacked ping is received</param>
        /// <param name="key">The key that identifies the application</param>
        public void RegisterPingNotificationUpcall( PingSendUpcall pingSendUpcall,
            PingReceiveUpcall pingReceiveUpcall, Object key )
        {
            Debug.Assert( pingSendUpcall!=null );
            Debug.Assert( pingReceiveUpcall!=null );
            Debug.Assert( key!=null );

            pingSendUpcalls.Add( key, pingSendUpcall );
            pingReceiveUpcalls.Add( key, pingReceiveUpcall );
        }

        /// <summary>
        /// Called internally by the SkipNet ping code when we are about to ping node 'n'.
        /// Calls all registered ping-notification-upcalls and builds a list of 'piggyback'
        /// objects that we will deliver to node 'n' along with our ping.
        /// </summary>
        internal ArrayList PingNotification( NodeIdAddressBind n ) {
            ArrayList piggybacks = new ArrayList();
            foreach( Object key in pingSendUpcalls.Keys ) {
                // Get the upcall associated with this key and invoke it
                PingSendUpcall upcall = (PingSendUpcall) pingSendUpcalls[ key ];
                Object piggybackObj = upcall( this, n );

                // If we are to piggyback an object in this ping, add this
                // to our array of piggybacked objects
                if( piggybackObj!=null ) {
                    piggybacks.Add( new PiggybackWrapper(n, key, piggybackObj) );
                }
            }
            return piggybacks;
        }

        /// <summary>
        /// Called internally when we receive a piggybacked ping object that was
        /// destined for this node. We determine which handler 
        /// </summary>
        /// <param name="key"></param>
        /// <param name="piggybackObj"></param>
        internal void InvokePingReceivedUpcall( NodeIdAddressBind source, Object key, Object piggybackObj ) {
            if( ! pingReceiveUpcalls.ContainsKey(key) ) {
                log.Write(LogLvl.Hi, "Received a piggybacked-ping for key {0} but no upcall registered!", key);
                return;
            }

            PingReceiveUpcall upcall = (PingReceiveUpcall) pingReceiveUpcalls[ key ];
            upcall( this, source, piggybackObj );
        }


        //**********************************************************************************************
        // Console Dump
        //**********************************************************************************************
        
        public void DumpLeafSet() {
            ArrayList left=null, right=null;
            
            SkipNet.LeafSet leafSet = routingTable.Neighbours[0];
            if( leafSet==null ) {
                Console.WriteLine( "Leaf Set Disabled!" );
                return;
            }

            left = new ArrayList( leafSet.LeftHalf.Neighbours );
            right = new ArrayList( leafSet.RightHalf.Neighbours );

            Console.WriteLine("--- Left Leaf Set ---");
            foreach( NodeIdAddressBind n in left ) {
                Console.WriteLine( n.GetNodeId.ToString() + " (" + n.GetAddress.ToString() + ")" );
            }

            Console.WriteLine("--- Right Leaf Set ---");
            foreach( NodeIdAddressBind n in right ) {
                Console.WriteLine( n.GetNodeId.ToString() + " (" + n.GetAddress.ToString() + ")" );
            }

        }

        // Dump all rows of the routing table until an empty row is discovered
        public void DumpRoutingTable() {
            bool empty;
            int row=0;
            Console.WriteLine("---- Routing Table ----");
            do {
                empty = DumpRoutingTable(row++);
            } while( ! empty );
        }

        // Dumps one row of the routing table. If the row was empty returns true. If the
        // row was non-empty, returns false.
        public bool DumpRoutingTable(int row) {

            SkipNet.LeafSet leafSet = routingTable.Neighbours[row];

            ArrayList left=null, right=null;
            left = new ArrayList( leafSet.LeftHalf.Neighbours );
            right = new ArrayList( leafSet.RightHalf.Neighbours );

            // If this row is empty, we're done
            if( left.Count==0 && right.Count==0 ) {
                return true;
            }

            Console.WriteLine("--- Level {0} Left Neighbours ---", row);
            foreach( NodeIdAddressBind n in left ) {
                Console.WriteLine( n.GetNodeId.ToString() + " (" + n.GetAddress.ToString() + ")" );
            }

            Console.WriteLine("--- Level {0} Right Neighbours ---", row);
            foreach( NodeIdAddressBind n in right ) {
                Console.WriteLine( n.GetNodeId.ToString() + " (" + n.GetAddress.ToString() + ")" );
            }

            return false;
        }

        // Dump the SkipNet P-Table
        public void DumpPTable() {
            Console.WriteLine("----- P-Table -----");

            ProximityTable pt = this.proximityTable;
            if( pt==null ) {
                Console.WriteLine("--- P-Table Disabled ---");
                return;
            }
            
            pt.Dump();
        }

        // DumpNeighbours: Dump a set of neighbours
        internal void DumpNeighbours( ArrayList nbrs ) {
            if( nbrs==null ) {
                Console.WriteLine("(null)");
            } else {
                for( int i=0; i<nbrs.Count; i++ ) {
                    if( nbrs[i]!=null ) {
                        Console.WriteLine( "{0}: {1}", i, nbrs[i].ToString() );
                    }
                }
            }
        }
        
        
        //**********************************************************************************************
        // Simulation-specific code
        //**********************************************************************************************

        private bool failed = false;
        public override bool Failed {
            get { return failed; } 
            set { failed = value; localAddr.Failed = value; }
        }

        public void BuildPerfectPTable( MessageRouting.Node[] allNodes ) {
            if( proximityTable!=null ) {
                proximityTable.BuildPerfect( allNodes );
            }
        }
    }


    /// <summary>
    /// Delegate for an upcall that is invoked by SkipNet when it is about to ping a node.
    /// An application should use Node.RegisterPingNotificationUpcall if it is interested
    /// in receiving these upcalls.
    /// </summary>
    /// <param name="dest">The node that we are about to ping</param>
    public delegate Object PingSendUpcall( Node localNode, NodeIdAddressBind n );


    /// <summary>
    /// Delegate for an upcall that is invoked by SkipNet when it receives a piggybacked
    /// object that is to be delivered on this node.
    /// </summary>
    /// <param name="piggybackObj">The piggybacked object that was received</param>
    public delegate void PingReceiveUpcall( Node localNode, NodeIdAddressBind source, Object piggybackObj );


    /// <summary>
    /// Used in various places to indicate direction within a ring.
    /// </summary>
    public enum Direction {
        Left, Right
    }


    /// <summary>
    /// Options used in routing by name
    /// </summary>
    public enum RouteByNameOption {
        Normal,             // Simply find the node that is closest to the given name.
        
        PrefixMatch,        // Any node for which the given name is a prefix is acceptable.
                            // If no node matches this prefix, the closest name to the prefix is used.
                            // This option is used by CLB searches.
        
        DisallowEqual,      // Find the node that is closest (in whatever the current direction is),
                            // but not equal to, the node with the given name. Since use of this option
                            // yields different results depending on the direction of routing, this
                            // should only be used for special operations.
    }

}