using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// Implements a Skip Net NodeId (i.e. NameId) identifier.
    /// </summary>
    [Serializable,
        XmlRootAttribute("NodeId",Namespace=SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class NodeId : MessageRouting.NodeId, IComparable {

        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        public const char   DELIMITER = '/';
        static char[]       DELIMITER_ARR = new char[]{ DELIMITER };


        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        string              identifier;
        

        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        // Default constructor (for serialization)
        public NodeId() {}

        public NodeId( string id ) {
            this.identifier = id;
        }
        
        
        //**********************************************************************************************
        // Public Properties
        //**********************************************************************************************
        
        [XmlElementAttribute("Name",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public string Name {
            get { return identifier; }
            set { identifier = value; }
        }
            
        
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************
        
        // Compare ordering of two NodeIds
        public int CompareTo(object comp) {
            NodeId compNode = (NodeId) comp;

            // String.CompareOrdinal is fast -- it ignores locales. This is preferable in the simulator.
            return String.CompareOrdinal(this.identifier,compNode.identifier);

            // Case-insensitive string comparison
            //return String.Compare( this.identifier, compNode.identifier, true );
        }

        // Determine if B is located between A and C, allowing the possibility that A==B
        public static bool LiesBetweenIncl( NodeId A, NodeId B, NodeId C ) {
            // There are 3 cases:
            // Case 1:  A --- B --- C
            // Case 2:  C --- A --- B
            // Case 3:  B --- C --- A
            // Case 4:  A==B --- C   or  C --- A==B  or   A==B==C   
            if(    
                   (A.CompareTo(B)<0 && B.CompareTo(C)<0)   // Case 1
                || (C.CompareTo(A)<0 && A.CompareTo(B)<0)   // Case 2
                || (B.CompareTo(C)<0 && C.CompareTo(A)<0)   // Case 3
                || (A.Equals(B)) )                          // Case 4
            {
                return true;
            } else {
                return false;
            }
        }

        // Determine if B is located between A and C, allowing for the possibility that A==C
        public static bool LiesBetween( NodeId A, NodeId B, NodeId C ) {
            // There are 3 cases:
            // Case 1:  A --- B --- C
            // Case 2:  C --- A --- B
            // Case 3:  B --- C --- A
            // Case 4:  C==A --- B
            if(    (A.CompareTo(B)<0 && B.CompareTo(C)<0)   // Case 1
                || (C.CompareTo(A)<0 && A.CompareTo(B)<0)   // Case 2
                || (B.CompareTo(C)<0 && C.CompareTo(A)<0)   // Case 3
                || (C.Equals(A)) )                          // Case 4
            {
                return true;
            } else {
                return false;
            }
        }

        // Determine if B is located between A and C, when going in direction given by 'dir'
        public static bool LiesBetweenDir( Direction dir, NodeId A, NodeId B, NodeId C ) {
            if( dir==Direction.Left ) {
                return LiesBetween(C,B,A);
            } else {
                return LiesBetween(A,B,C);
            }
        }

        // Determines if this name ID has the given prefix
        public bool IsWithinPrefix( string prefix ) {
            return identifier.StartsWith( prefix );
        }
        
        // Determines if this name ID has the given prefix
        public bool IsWithinPrefix( NodeId prefix ) {
            return IsWithinPrefix( prefix.identifier );
        }
        
        // Compute the length of the longest common prefix of two name IDs, ignoring delimiters
        public int LongestCommonPrefixLength( NodeId n ) {
            int len=0;
            while( true ) {
                if(len>=this.identifier.Length || len>=n.identifier.Length) return len;
                if(this.identifier[len]!=n.identifier[len]) return len;
                len++;
            }
        }
            
        /// <summary>
        /// Compute the number of the common CLB Domains of ids A and B.
        /// "CLB Domains" are substrings starting at the beginning of the string and ending
        /// with a delimiter character (or the end of the string).
        /// For example, NumCommonCLBDomains("com/msft/windows", "com/msft/office")=2.
        /// </summary>
        public static int NumCommonCLBDomains( NodeId A, NodeId B ) {
            int i;

            string[] partsA = A.identifier.Split( DELIMITER_ARR );
            string[] partsB = B.identifier.Split( DELIMITER_ARR );

            for( i=0; i<partsA.Length && i<partsB.Length; i++ ) {
                if( ! partsA[i].Equals(partsB[i]) ) {
                    return i;
                }
            }

            return i;
        }

        /// <summary>
        /// Get an array containing all CLB Domains of this NodeId.
        /// Examples:
        ///              Input                                        Output
        /// "nickhar"                            null
        /// "com/nickhar"                        {"com/"}
        /// "com/microsoft/research/nickhar"     {"com/", "com/microsoft/", "com/microsoft/research/"}
        /// </summary>
        public string[] GetAllCLBDomains() {
            string[] parts = identifier.Split( DELIMITER_ARR );

            // Node doesn't belong to any CLB domains
            if( parts.Length<2 ) {
                return null;
            }

            // Build list of CLB domains
            string[] result = new string[ parts.Length-1 ];
            result[0] = parts[0] + DELIMITER;

            for( int i=1; i<parts.Length-1; i++ ) {
                result[i] = result[i-1] + parts[i] + DELIMITER;
            }

            return result;
        }

        // Return the string form of a node id
        public override string ToString() {
            return identifier;
        }

        // Test two NodeIds for equality
        public bool Equals(NodeId n) {
            return this.identifier.Equals(n.identifier);
        }

        // Overridden equals (used for Hashtables)
        public override bool Equals(Object obj) {
            if (!(obj is NodeId))
                return false;
            return Equals((NodeId)obj);
        }

        // Overridden hashcode (used for Hashtables)
        public override int GetHashCode() {
            return identifier.GetHashCode();
        }
    }

}
