using System;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// Implements a NodeIdAddressBind object, which binds a NodeId to a NodePhysicalAddress.
    /// </summary>
    [Serializable, XmlRootAttribute("NodeIdAddressBind",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class NodeIdAddressBind : MessageRouting.NodeIdAddressBind, IComparable {

        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        private NodeId              id;     // Id
        private NodePhysicalAddress addr;   // Physical address
        private NumericId           rid;


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        // Default constructor (for serialization)
        public NodeIdAddressBind() {
        }

        public NodeIdAddressBind(NodePhysicalAddress addr, NodeId nodeid, NumericId randomid) {
            this.id = nodeid;
            this.addr = addr;
            this.rid = randomid;
        }


        //**********************************************************************************************
        // Properties
        //**********************************************************************************************

        // Use this for internal access within Pastry without having to do runtime
        // conversions between MessageRouting.NodeId and Pastry.NodeId.
        [XmlIgnore]
        public NodeId NodeId {
            get { return id; }
            set { id = value; }
        }

        [XmlElementAttribute("NumericId",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public NumericId NumericId {
            get { return rid; }
            set { rid=value; }
        }

        [XmlElementAttribute("GetNodeId",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public override MessageRouting.NodeId GetNodeId {
            get { return id; }
            set { id = (NodeId) value; }
        }

        [XmlElementAttribute("GetAddress",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public override NodePhysicalAddress GetAddress {
            get { return addr; }
            set { addr=value; }
        }


        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************

        // Overridden equals (used for Hashtables)
        public override bool Equals(Object o) {
            if (!(o is NodeIdAddressBind)) 
                return false;

            NodeIdAddressBind n = (NodeIdAddressBind) o;
            bool idEqual=false;

            // Determine if the node ids are equal, being careful to handle null node ids
            if( this.id==null || n.id==null ) {
                idEqual = true;
            } else {
                idEqual = NodeId.Equals(n.NodeId);
            }

            return ( idEqual && GetAddress.Equals(n.GetAddress) );
        }

        // Overridden hashcode (used for Hashtables)
        public override int GetHashCode() {
            if( id==null ) {
                // Possible during join -- we do not know the id of the seed node
                return 0;
            } else {
                return id.GetHashCode();
            }
        }

        // Overriden convert to string
        public override string ToString() {
            if( id!=null )
                return id.ToString();
            else
                return "(null)";
        }
        
        // Compare ordering of two NodeIdAddressBinds
        public int CompareTo(object comp) {
            NodeIdAddressBind compAddr = (NodeIdAddressBind) comp;
            return this.NodeId.CompareTo( compAddr.NodeId );
        }
    }
}
