using System;
using System.Diagnostics;
using System.Security.Cryptography;
using System.Text;
using System.Xml.Serialization;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// A NumericId is just a simple string of random numbers.
    /// </summary>
    [Serializable, XmlRootAttribute("NumericId",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class NumericId : ICloneable {

        //**********************************************************************************************
        // Static members
        //**********************************************************************************************
        static Random               rng;
        
        
        //**********************************************************************************************
        // Private members
        //**********************************************************************************************
        static SHA1CryptoServiceProvider    md;         // SHA1 Hash object. Used for generating random numbers
        static ulong                        seed;       // Seed for random numbers
        static ulong                        counter;    // Counter for random numbers
        static System.Text.UnicodeEncoding  enc;        // Text encoder for hashing strings

        int[]           r;          // The array of random coin-flips
        int             b;          // The number of faces on the random coin
        int             height;     // The number of random flips that we make, minus 


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        public NumericId() {
            b = Parameters.b;
            Debug.Assert( b>1 && b<=16, "SkipNet Base Parameter out of Range" );
        }

        // Static constructor
        static NumericId() {
            rng = new Random();
            md = new SHA1CryptoServiceProvider();
            counter = 0;
            enc = new System.Text.UnicodeEncoding();
        }
        
        
        //**********************************************************************************************
        // Public Properties
        //**********************************************************************************************
        
        [XmlElementAttribute("RandomBits",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public string RandomBits {
            get { return ToString(); }
            set { ReadFromString(value); }
        }

        [XmlIgnore]
        public int Height {
            get { return height; }
            set { height = value; }
        }
        
        
        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************

        public static void SetSeed( int seed ) {
            rng = new Random( seed );
            NumericId.seed = (ulong) seed;
        }
        
        internal static double RandomDouble() {
            lock( rng ) {
                return rng.NextDouble();
            }
        }

        /* Alternate implementation of Generate() that uses a simple random number generator
        public void Generate() {
            height = Parameters.maxHeight-1;
            r = new int[height];
            for( int i=0; i<height; i++ ) {
                // TODO: Synchronize here
                r[i] = rng.Next(b);
            }
        }
        */
    
        // Generate a random ID using SHA1 with default height and b values
        public void Generate() {
            byte[] input = new Byte[16];
            int i, j;

            lock( md ) {
                height = Parameters.maxHeight;
                r = new int[height];

                // Generate initial input from counter & seed
                // Note: counter is static so we are synchronizing for it too
                ulong n = counter++;
                for( i=0; i<8; i++ ) {
                    input[i] = (byte) (n & 0xFF);
                    n >>= 8;
                }
                n = NumericId.seed;
                for( i=0; i<8; i++ ) {
                    input[8+i] = (byte) (n & 0xFF);
                    n >>= 8;
                }
                
                // Generate 'height' random digits
                for( i=0; i<height; i+=16 ) {
                    input = md.ComputeHash(input);
                    for( j=0; j<16; j++ ) {
                        r[i+j] = input[j] % b;
                        if( i+j>=height-1 ) {
                            return;
                        }
                    }
                }
            }
        }
        
        // Generate a random ID using a SHA1 hash of a given string
        public void GenerateFromString( string str ) {
            byte[] input = new Byte[str.Length*2];
            int i, j;
            
            lock( md ) {
                height = Parameters.maxHeight;
                r = new int[height];
            
                // Initial input is given
                // Note: enc is also static so we are synchronizing for it too.
                enc.GetBytes( str, 0, str.Length, input, 0 );
            
                // Generate 'height' random digits
                for( i=0; i<height; i+=16 ) {
                    input = md.ComputeHash(input);
                    for( j=0; j<16; j++ ) {
                        r[i+j] = input[j] % b;
                        if( i+j>=height-1 ) {
                            return;
                        }
                    }
                }
            }
        }

        // Create a new NumericId that is a copy of this NumericId
        public object Clone() {
            NumericId result = new NumericId();

            result.r = (int[]) this.r.Clone();
            result.b = this.b;
            result.height = this.height;

            return result;
        }

        // Compute the length of the longest common prefix of two random IDs
        public int LongestCommonPrefixLength( NumericId ridB ) {
            Debug.Assert(this.b==ridB.b);
            int len=0;
            while( true ) {
                if(len>=this.height || len>=ridB.height) return len;
                if(this.r[len]!=ridB.r[len]) return len;
                len++;
            }
        }

        // Get the n'th digit
        public int GetDigit( int n ) {
            Debug.Assert(n<height);
            return this.r[n];
        }

        // Set the n'th digit
        public void SetDigit( int n, int digit ) {
            Debug.Assert(n<height);
            this.r[n] = digit;
        }

        // Convert a NumericId to a string
        public override string ToString() {
            StringBuilder builder = new StringBuilder(Parameters.maxHeight+2);
            for( int i=0; i<height; i++ ) {
                builder.AppendFormat("{0:X}", r[i] );
            }
            return builder.ToString();
        }

        // Read in a NumericId from a string
        public void ReadFromString(string s) {
            r = new int[ s.Length ];
            height = s.Length;

            if( Parameters.b>10 ) {
                throw new ApplicationException( "NumericIds don't support b>=10!" );
            }

            int zero = (int) Char.GetNumericValue('0');
            for( int i=0; i<s.Length; i++ ) {
                // Old code:
                //string c = s.Substring(i,1);
                //r[i] = Int32.Parse(c, System.Globalization.NumberStyles.HexNumber);

                r[i] = ((int) Char.GetNumericValue(s[i])) - zero;
            }
        }

        // For putting NumericIds into a hash table
        public override int GetHashCode() {
            int hash=0;
            for( int i=0; i<r.Length; i++ ) {
                hash = 314159265*hash + r[i];
            }
            return hash;
        }
        
        // Test two NumericIds for equality
        public bool Equals( NumericId ridB ) {
            // Base and height must be equal
            if( (this.b!=ridB.b) || (this.height!=ridB.height) ) {
                return false;
            }
            
            // Check that all digits are the same
            for( int i=0; i<height; i++ ) {
                if( this.r[i]!=ridB.r[i] ) {
                    return false;
                }
            }

            return true;
        }
        
        // Compare ordering of two NumericIds
        public int CompareTo(object comp) {
            NumericId ridB = (NumericId) comp;

            Debug.Assert( this.height>0 && ridB.height>0 );

            // Base and height must be equal
            if( (this.b!=ridB.b) || (this.height!=ridB.height) ) {
                throw new Exception( "Cannot compare NumericIds with different base / height" );
            }
            
            // Check all digits
            for( int i=0; i<height; i++ ) {
                if( this.r[i]<ridB.r[i] ) {
                    return -1;
                } else if( this.r[i]>ridB.r[i] ) {
                    return 1;
                }
            }
            
            // They are equal
            return 0;
        }
        
        // Determine if B is located between A and C, allowing for the possibility that A==C
        public static bool LiesBetween( NumericId A, NumericId B, NumericId C ) {
            // There are 3 cases:
            // Case 1:  A --- B --- C
            // Case 2:  C --- A --- B
            // Case 3:  B --- C --- A
            // Case 4:  C==A --- B
            if(    (A.CompareTo(B)<0 && B.CompareTo(C)<0)   // Case 1
                || (C.CompareTo(A)<0 && A.CompareTo(B)<0)   // Case 2
                || (B.CompareTo(C)<0 && C.CompareTo(A)<0)   // Case 3
                || (C.Equals(A)) ) {                        // Case 4
                return true;
            } else {
                return false;
            }
        }

        // Determine if B is located between A and C, when going in direction given by 'dir'
        public static bool LiesBetweenDir( Direction dir, NumericId A, NumericId B, NumericId C ) {
            if( dir==Direction.Left ) {
                return LiesBetween(C,B,A);
            } else {
                return LiesBetween(A,B,C);
            }
        }
        
        // Compute the absolute difference of this and ridParam
        public NumericId Difference( NumericId ridParam ) {
            /* Compute absolute difference */
            int cmp = this.CompareTo( ridParam );
            
            // Subtract the larger from the smaller
            if( cmp<0 ) {
                return Subtract( ridParam, this );
            } else {
                return Subtract( this, ridParam );
            }

            /* Subtract ridParam from this (not absolute difference)
            return Subtract( this, ridParam );*/
        }
        
        
        //**********************************************************************************************
        // Private methods
        //**********************************************************************************************
        
        // Subtract Y from X
        private static NumericId Subtract( NumericId X, NumericId Y ) {
            int i, x, y, b=X.b, borrow=0, height=X.height;

            // Create a NumericId to store the difference
            NumericId diff = new NumericId();
            diff.height = height;
            diff.r = new int[ height ];
            
            for( i=height-1; i>=0; i-- ) {
                x = X.r[i] - borrow;
                if( x<0 ) { x+=b; borrow=1; }
                else { borrow=0; }
                
                y = Y.r[i];
                
                if( x<y ) {
                    Debug.Assert( borrow==0 );
                    x+=b; borrow=1;
                }
                Debug.Assert( x>=y );
                
                diff.r[i] = x-y;
            }

            // If X>=Y then borrow must be 0. If X<Y then borrow must be 1.
            int cmp = X.CompareTo(Y);
            Debug.Assert( (borrow==0 && cmp>=0) || (borrow==1 && cmp<0) );
            
            return diff;
        }

    }
}
