using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {


    /// <summary>
    /// The ProximityTable is a type of routing table specialized for incorporating
    /// network proximity when routing by name ID.
    /// </summary>
    [Serializable, XmlRootAttribute("ProximityTable",
                       Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class ProximityTable {
        
        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        
        public const string     CATEGORY = "PTable";

        
        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        
        // The node to whom the table belongs
        Node                    localNode;

        // The id and address of the node to whom the table belongs
        NodeIdAddressBind       localAddrBind;

        // The two halves of the table
        ProximityTableHalf      left, right;

        // How many times has the P-Table been rebuilt?
        int                     proxTableRebuilds=0;

        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor for serialization
        public ProximityTable() {}

        public ProximityTable( Node localNode ) {
            this.localNode = localNode;
            this.localAddrBind = localNode.NodeIdAddressBind;
            this.left = new ProximityTableHalf( Direction.Left, localAddrBind, localNode.Topology );
            this.right = new ProximityTableHalf( Direction.Right, localAddrBind, localNode.Topology );
        }
        
        
        //**********************************************************************************************
        // Public properties
        //**********************************************************************************************

        // Have both halves of the proximity table been successfully initialized?
        public bool Initialized {
            get { return left.Initialized && right.Initialized; }
        }

        // How many times have both halves of the proximity table been successfully initialized?
        public int InitCount {
            get { return Math.Min(left.InitCount, right.InitCount); }
        }


        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************

        /// <summary>
        /// Completely rebuild the P-Table. Rebuild the list of neighbours, rebuild the intervals,
        /// and send out a new FindEntries message.
        /// </summary>
        public void Rebuild() {
            Log.Write(CATEGORY, LogLvl.Med, "Rebuilding P-Table at node {0} count={1}",
                localAddrBind.GetAddress.ToString(), ++proxTableRebuilds );

            if( EnsureCurrentIntervals() ) {
                // Even if we already had an array of intervals, we update our neighbours
                // so that the P-Table will be rebuilt with the newest intervals.
                UpdateNeighbours();
            }

            // Rebuild the left and right halves of the table.
            left.RebuildEntries( localNode, this );
            right.RebuildEntries( localNode, this );
        }

        /// <summary>
        /// Build a perfect P-Table, given the list of all overlay nodes.
        /// This is only used by the simulator to compare the P-Table constructed by the usual
        /// algorithm against optimal P-Tables.
        /// 
        /// NOTE: This function calls "RebuildNeigbours()"
        /// </summary>
        /// <param name="allNodes">Array of all overlay nodes</param>
        public void BuildPerfect(MessageRouting.Node[] allNodes) {
            Log.Write(CATEGORY, LogLvl.Med, "Building perfect P-Table at node {0}", localAddrBind.GetAddress.ToString());

            EnsureCurrentIntervals();

            ArrayList allAddrs = new ArrayList();
            for( int i=0; i<allNodes.GetLength(0); i++ ) {
                if( allNodes[i]==null ) continue;
                NodeIdAddressBind n = (NodeIdAddressBind) allNodes[i].GetNodeIdAddressBind;
                allAddrs.Add( new PTableEntry(n, 0) );
            }
            allAddrs.Sort();

            // Fill the left side
            int numProcessed = left.AddToCurrentEntries( allAddrs );
            left.InitCount++;
            
            // Remove all processed elements and reverse the ordering for
            // merging with the right half.
            allAddrs.RemoveRange(0, numProcessed);
            allAddrs.Reverse();
            
            // Fill the right side
            right.AddToCurrentEntries( allAddrs );
            right.InitCount++;

            // Rebuild the list of all of this node's neighbours
            localNode.RebuildNeighbours(null);
        }

        /// <summary>
        /// Try to add the given list of entries as current entries in this P-Table.
        /// 
        /// NOTE: This function calls "RebuildNeigbours()" if necessary
        /// </summary>
        /// <param name="newEntries">An array of PTableEntry objects</param>
        public void AddToCurrentEntries( ArrayList newEntries ) {
            bool added=false;

            EnsureCurrentIntervals();
            
            newEntries.Sort();
            
            // Fill the left side
            int numProcessed = left.AddToCurrentEntries( newEntries );
            if( numProcessed>0 ) {
                added = true;
            }
            
            // Remove all processed elements and reverse the ordering for
            // merging with the right half.
            newEntries.RemoveRange(0, numProcessed);
            newEntries.Reverse();
            
            // Fill the right side
            numProcessed = right.AddToCurrentEntries( newEntries );
            if( numProcessed>0 ) {
                added = true;
            }

            // Rebuild the list of all of this node's neighbours
            if( added ) {
                localNode.RebuildNeighbours(null);
            }
        }

        /// <summary>
        /// Try to add the single given entry as a current entry in this P-Table.
        /// 
        /// NOTE: This function calls "RebuildNeigbours()" if necessary
        /// </summary>
        /// <param name="newEntries"></param>
        public void AddEntry( NodeIdAddressBind n ) {
            // Build an array of all entries in the source table
            ArrayList entryArray;
            entryArray = new ArrayList(1);
            entryArray.Add( new PTableEntry(n,0) );

            AddToCurrentEntries(entryArray);
        }

        /// <summary>
        /// Update the current intervals with the new intervals.
        /// Generate a new array of "current entries" by choosing from the existing
        /// "current entries" and the array of "new entries".
        /// 
        /// NOTE: This function calls "RebuildNeigbours()"
        /// </summary>
        /// <param name="dir">Which side of the P-Table to update</param>
        /// <param name="newIntervals">The array of new intervals</param>
        /// <param name="newEntries">The array of new entries</param>
        public void ChooseNewEntries( Direction dir, ArrayList newIntervals, ArrayList newEntries ) {
            if( dir==Direction.Left ) {
                left.ChooseNewEntries( newIntervals, newEntries );
            } else {
                right.ChooseNewEntries( newIntervals, newEntries );
            }

            // Rebuild the list of all of this node's neighbours
            localNode.RebuildNeighbours(null);
        }

        /// <summary>
        /// Send a message to all P-Table entries in given direction
        /// to let them know that they are in our proximity table.
        /// </summary>
        /// <param name="dir">Indicates which half of the table we're interested in</param>
        /// <param name="localNode">The local node</param>
        public void NotifyEntries( Direction dir, Node localNode ) {
            if( dir==Direction.Left ) {
                left.NotifyEntries( localNode );
            } else {
                right.NotifyEntries( localNode );
            }
        }

        /// <summary>
        /// Get all entries from this table, sorted by name id.
        /// </summary>
        public ArrayList GetAllEntries() {
            // Build an array of all entries in the source table
            ArrayList allEntries = left.GetEntries();
            allEntries.Add( new PTableEntry(localAddrBind, 0) );
            
            ArrayList rightEntries = right.GetEntries();
            rightEntries.Reverse();
            allEntries.AddRange( rightEntries );
            
            // ASSERT: allSourceEntries is sorted here

            return allEntries;
        }
        
        /// <summary>
        /// Determine the next hop in routing towards the destination.
        /// The destination may not be equal to the local node.
        /// The localNode may not be within the prefix if the PrefixMatch option is on.
        /// </summary>
        public NodeIdAddressBind RouteByName( NodeId destID, RouteByNameOption option ) {
            NodeIdAddressBind nextHop=null;

            // Compare the local node ID with the destination
            NodeId localID = localAddrBind.NodeId;
            int cmp = destID.CompareTo(localID);
            
            // Sanity check.
            Debug.Assert( cmp!=0 );
            if( option==RouteByNameOption.PrefixMatch ) {
                Debug.Assert( ! localAddrBind.NodeId.IsWithinPrefix(destID) );
            }

            // Route either to the left or the right
            if( cmp<0 ) {
                // Destination ID is to the left
                nextHop = left.RouteByName(destID, option);
                
                // The next hop must always be to our left (or null if table not initialized)
                Debug.Assert( nextHop==null || nextHop.NodeId.CompareTo(localID)<=0 );
            } else {
                // Destination ID is to the right
                nextHop = right.RouteByName(destID, option);
                
                // The next hop must always be to our right (or null if table not initialized)
                Debug.Assert( nextHop==null || nextHop.NodeId.CompareTo(localID)>=0 );
            }
            
            return nextHop;
        }
        
        /// <summary>
        /// A node has failed and thus we are to remove it from the table, if it exists.
        /// 
        /// NOTE: The caller must call 'RebuildNeighbours()' if the node was removed, otherwise
        /// the table of all neighbours will be stale.
        /// </summary>
        /// <param name="nodeAddr">The failed node</param>
        /// <returns>True if the node was removed</returns>
        public bool NodeFailed( NodeIdAddressBind nodeAddr ) {
            bool removed;
            removed  = left.NodeFailed( nodeAddr );
            removed |= right.NodeFailed( nodeAddr );
            return removed;
        }
        
        /// <summary>
        /// Insert each neighbour into the given Hashtable 
        /// </summary>
        /// <param name="ht">Table to insert neighbours into</param>
        public void GetAllNeighbours(Hashtable ht) {
            left.GetAllNeighbours(ht);
            right.GetAllNeighbours(ht);
        }

        /// <summary>
        /// Dump the P-Table intervals and entries
        /// </summary>
        public void Dump() {
            Console.WriteLine("--- Left Intervals ---");
            localNode.DumpNeighbours(left.currentIntervals);

            Console.WriteLine("--- Right Intervals ---");
            localNode.DumpNeighbours(right.currentIntervals);

            Console.WriteLine("--- Left Entries ---");
            foreach( PTableEntry entry in left.GetEntries() ) {
                Console.WriteLine( "{0} ({1})", entry.node, entry.dist );
            }

            Console.WriteLine("--- Right Entries ---");
            foreach( PTableEntry entry in right.GetEntries() ) {
                Console.WriteLine( "{0} ({1})", entry.node, entry.dist );
            }
        }
        
        
        //**********************************************************************************************
        // Private methods
        //**********************************************************************************************

        /// <summary>
        /// Update our "neighbour lists": lists of all R-Table neighbours to the
        /// left and right of the local node
        /// </summary>
        private void UpdateNeighbours() {
            ArrayList leftNbrs, rightNbrs, leaves;
            leftNbrs = new ArrayList();
            rightNbrs = new ArrayList();

            RoutingTable rt = localNode.RoutingTable;

            for( int h=1; h<Parameters.maxHeight; h++ ) {
                // Construct list of left neighbours. Any order is OK.
                leaves = rt.Neighbours[h].LeftHalf.GetAllNbrs();
                foreach( NodeIdAddressBind n in leaves ) {
                    if( n.CompareTo(localAddrBind)<0 ) {
                        leftNbrs.Add(n.NodeId);
                    }
                }

                // Construct list of right neighbours. Any order is OK.
                leaves = rt.Neighbours[h].RightHalf.GetAllNbrs();
                foreach( NodeIdAddressBind n in leaves ) {
                    if( n.CompareTo(localAddrBind)>0 ) {
                        rightNbrs.Add(n.NodeId);
                    }
                }
            }

            // Update the lists of neigbours in the two half-tables
            left.UpdateNeighbours( leftNbrs );
            right.UpdateNeighbours( rightNbrs );
        }

        /// <summary>
        /// Ensure that we have built an array of current intervals for this P-Table.
        /// 
        /// NOTE: This function calls "RebuildNeigbours()" if necessary
        /// </summary>
        /// <returns>True if there already was an array of current intervals,
        /// false if there was no existing array</returns>
        bool EnsureCurrentIntervals() {
            if( left.currentIntervals!=null ) {
                Debug.Assert( right.currentIntervals!=null );
                return true;
            }

            Debug.Assert( right.currentIntervals==null );
            Debug.Assert( left.currentEntries==null );
            Debug.Assert( right.currentEntries==null );

            // No current entries! Update our neighbours and our current intervals.
            UpdateNeighbours();

            left.EnsureCurrentIntervals();
            right.EnsureCurrentIntervals();

            // Now, if we have a C-Table, use the C-Table to populate the P-Table.
            CTable cTable = localNode.CTable;
            if( cTable!=null ) {
                ArrayList cTableEntries = cTable.GetArrayOfPTableEntries();
                cTableEntries.Sort();
        
                // Fill the left side
                int numProcessed = left.AddToCurrentEntries( cTableEntries );
        
                // Remove all processed elements and reverse the ordering for
                // merging with the right half.
                cTableEntries.RemoveRange(0, numProcessed);
                cTableEntries.Reverse();
        
                // Fill the right side
                right.AddToCurrentEntries( cTableEntries );
            }

            // Rebuild the list of all of this node's neighbours
            localNode.RebuildNeighbours(null);

            return false;
        }

    }

    
    /// <summary>
    /// One half of the proximity table.
    /// </summary>
    [Serializable, XmlRootAttribute("ProximityTableHalf",
                       Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class ProximityTableHalf {

        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        
        // The id and address of the local node
        NodeIdAddressBind       localAddrBind;
        
        // Which direction is this half of the proximity table?
        Direction               dir;

        // The neighbours at a particular height (note: not the P-Table entries!).
        // Stored as an ArrayList of NodeId objects.
        ArrayList               neighbours;
        
        // The current intervals and current entries. 'currentIntervals' is an array of NodeId objects.
        // 'currentEntries' contains PTableEntry objects and always corresponds to currentIntervals.
        internal ArrayList      currentIntervals;
        internal PTableEntry[]  currentEntries;
        
        // How many times has this half of the table been initialized?
        int                     initCount;

        // The topology is used to measure distance in the network 
        Topology                topo;
        
        
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************
        
        // Default constructor needed for serialization
        public ProximityTableHalf() {}

        public ProximityTableHalf( Direction dir, NodeIdAddressBind localAddrBind,
            MessageRouting.Topology topo ) : this()
        {
            this.dir = dir;
            this.topo = topo;
            this.localAddrBind = localAddrBind;
            this.neighbours = new ArrayList();
            this.initCount = 0;
        }
        
        
        //**********************************************************************************************
        // Properties
        //**********************************************************************************************
        
        // Has this half of the proximity table been successfully initialized?
        public bool Initialized {
            get { return initCount > 0; }
        }

        // How many times has this half of the proximity table been successfully initialized?
        public int InitCount {
            get { return initCount; }
            set { initCount = value; }
        }

        
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************
        
        /// <summary>
        /// Update the cached information about our neighbours with the
        /// given list of neighbours.
        /// </summary>
        /// <param name="nbrs">An array of NodeId objects</param>
        public void UpdateNeighbours( ArrayList nbrs ) {
            this.neighbours = nbrs;

            // Sort the list of intervals from farthest to closest
            neighbours.Sort();
            if( dir==Direction.Right ) {
                neighbours.Reverse();
            }
        }

        /// <summary>
        /// Ensure that we have built the arrays of current intervals and entries.
        /// </summary>
        public void EnsureCurrentIntervals() {
            if( currentIntervals==null ) {
                currentIntervals = GetIntervals();

                // Count how many current entries we need for these intervals
                int numEntries = currentIntervals.Count;
                if( numEntries<0 ) numEntries=0;
            
                // Create a new array of current entries
                currentEntries = new PTableEntry[ numEntries ];
            }
        }
        
        /// <summary>
        /// Build an array of new intervals and send out a FindEntries message to
        /// populate those intervals.
        /// </summary>
        /// <param name="seed">
        /// Note: the 'seed' param may be null, in which case it will try to use
        /// existing table entries as a seed.
        /// </param>
        public void RebuildEntries( Node localNode, ProximityTable proximityTable ) {
            NodePhysicalAddress seedAddr = localNode.seed;
            NodeIdAddressBind rebuildFromNode=null;
            PTableEntry closestEntry = GetClosest();
            
            // We prefer to use the closest entry in this half of the table as the seed.
            if( closestEntry!=null ) {
                rebuildFromNode = closestEntry.node;
            } else {
                // If no entry exists we use the original seed.
                if( seedAddr!=null ) {
                    rebuildFromNode = new NodeIdAddressBind(seedAddr, null, null);
                } else {
                    // If no original seed, use the closest leaf-set node
                    LeafSet ls = localNode.RoutingTable.LeafSet;
                    ArrayList leaves = ls.GetAllNbrs();
                    if( leaves.Count>0 ) {
                        rebuildFromNode = ls.GetClosest(Direction.Right);
                    }
                }
            }

            ArrayList newIntervals = GetIntervals();
            int intervalCount = newIntervals.Count-1;
            
            // Send a FindEntries message to find possible entries for this table
            if( intervalCount>0 ) {
                if( rebuildFromNode!=null ) {
                    SendFindEntries( localNode, dir, rebuildFromNode, newIntervals );
                }
            } else {
                // If there are zero intervals then we consider this half of the
                // table to be already initialized
                initCount++;
            }
        }

        /// <summary>
        /// Try to update the entries in this half of the proximity table with the
        /// given list of new entries.
        /// </summary>
        /// <param name="newEntries">
        /// An array of PTableEntry objects. The caller must sort this array appropriately
        /// for this half of the proximity table.
        /// </param>
        /// <returns>Returns the number elements in 'newEntries' that were processed</returns>
        public int AddToCurrentEntries( ArrayList newEntries ) {
            return ChooseEntriesInternal(currentIntervals, currentEntries, newEntries);
        }

        /// <summary>
        /// Update the current intervals with the new intervals.
        /// Generate a new array of "current entries" using the existing "current entries"
        /// and the array of "new entries".
        /// </summary>
        /// <param name="newIntervals">The array of new intervals</param>
        /// <param name="newEntries">The array of new entries</param>
        public void ChooseNewEntries( ArrayList newIntervals, ArrayList newEntries ) {
            PTableEntry[] entries = new PTableEntry[ newIntervals.Count ];
            ChooseEntriesInternal(newIntervals, entries, newEntries);
            currentIntervals = newIntervals;
            currentEntries = entries;

            initCount++;
        }
        
        /// <summary>
        /// Return all current entries in this half of the proximity table. Never returns null.
        /// For the left proximity table half, the array is sorted in increasing order.
        /// For the right proximity table half, the array is sorted in decreasing order.
        /// </summary>
        /// <returns>An array of PTableEntry objects. This array can be modified.</returns>
        public ArrayList GetEntries() {
            ArrayList allCurrentEntries = new ArrayList();

            if( currentEntries!=null ) {
                foreach( PTableEntry entry in currentEntries ) {
                    if( entry!=null ) {
                        allCurrentEntries.Add( entry );
                    }
                }
            }
            
            return allCurrentEntries;
        }
        
        /// <summary>
        /// Determine the next hop in routing towards the destination.
        /// This should probably only be called by ProximityTable.RouteByName.
        /// </summary>
        /// <returns>
        /// May return null if this P-Table Half has not been initialized.
        /// </returns>
        public NodeIdAddressBind RouteByName( NodeId destID, RouteByNameOption option ) {
            NodeIdAddressBind prev = localAddrBind;
            
            if( ! Initialized ) {
                // Not initialized -- we should fall back to the R-Table
                return null;
            }
            
            // Examine all of the current entries and choose the best hop.
            for( int i=currentEntries.Length-1; i>=0; i-- ) {
                PTableEntry entry = (PTableEntry) currentEntries[i];
                if( entry==null ) continue;
                NodeIdAddressBind n = entry.node;
                
                // Special case: Handle prefix matches
                if(    option==RouteByNameOption.PrefixMatch
                    && n.NodeId.IsWithinPrefix(destID) ) {
                    return n;
                }

                // If n is too far, return the previous entry
                if( IsFarther(destID, n.NodeId) ) {
                    return prev;
                }
                
                // If n is the destination, we either return it or the previous node
                // depending on whether the DisallowEqual option is turned on.
                if( destID.Equals(n.NodeId) ) {
                    if( option==RouteByNameOption.DisallowEqual ) {
                        return prev;
                    } else {
                        return n;
                    }
                }
                
                prev = n;
            }
            
            // None of the entries went past or equalled the destination. Return
            // the last entry we considered.
            return prev;
        }
        
        /// <summary>
        /// A node has failed and thus we are to remove it from the list of
        /// current entries, if it exists.
        /// 
        /// NOTE: The caller must call 'RebuildNeighbours()' if the node was removed, otherwise
        /// the table of all neighbours will be stale.
        /// </summary>
        /// <param name="nodeAddr">The failed node</param>
        /// <returns>True if the node was actually removed</returns>
        public bool NodeFailed( NodeIdAddressBind nodeAddr ) {
            bool removed = false;

            // No current entries
            if( currentEntries==null ) {
                return false;
            }
            
            // Examine all current entries
            for( int i=0; i<currentEntries.Length; i++ ) {
                // This entry has already gone AWOL or was unfilled
                if( currentEntries[i]==null ) {
                    continue;
                }
                
                // If this entry is the failed node, remove it
                if( nodeAddr.Equals(currentEntries[i].node) ) {
                    currentEntries[i] = null;
                    removed = true;
                }
            }

            return removed;
        }
    
        /// <summary>
        /// Send a message to all entries in this half of the table
        /// to let them know that they are in our proximity table.
        /// </summary>
        /// <param name="localNode">The local node</param>
        public void NotifyEntries( Node localNode ) {
            for( int i=0; i<currentEntries.Length; i++ ) {
                // This entry has already gone AWOL or was unfilled
                if( currentEntries[i]==null ) {
                    continue;
                }

                Log.Write(ProximityTable.CATEGORY, LogLvl.Lo,
                    "Notifying Node {0} that he is in our P-Table", currentEntries[i].node.GetAddress.ToString());

                // Stuff our proximity table in an UpdatePTable message
                // and send it off to our P-Table entry.
                // Note: It doesn't really matter if the message fails, so we don't
                // specify a callback.
                OverlayMsg msg = new UpdatePTableMsg(localAddrBind);
                localNode.SendMessageDirect(msg, currentEntries[i].node);
            }
        }

        /// <summary>
        /// Insert each neighbour into the given Hashtable 
        /// </summary>
        /// <param name="ht">Table to insert neighbours into</param>
        public void GetAllNeighbours(Hashtable ht) {
            if( currentEntries==null ) return;

            for( int i=0; i<currentEntries.Length; i++ ) {
                if( (currentEntries[i]!=null) && 
                    (! ht.ContainsKey(currentEntries[i].node)) )
                {
                    ht.Add(currentEntries[i].node, null);
                }
            }
        }
        

        //**********************************************************************************************
        // Private Methods
        //**********************************************************************************************
        
        /// <summary>
        /// Given our list of neighbours, create a list of intervals.
        /// Note: The list of neighbours must previously have been updated using UpdateNeighbours().
        /// </summary>
        /// <returns>An array of NodeIds that are the intervals</returns>
        private ArrayList GetIntervals() {
            ArrayList intervals = new ArrayList();

            // Assert: List of neighbours is sorted

            // Copy all neighbours, removing duplicates, into the newIntervals array
            for( int i=0; i<neighbours.Count; i++ ) {
                if( i==0 || ! neighbours[i].Equals(neighbours[i-1]) ) {
                    intervals.Add( neighbours[i] );
                }
            }
            
            return intervals;
        }

        /// <summary>
        /// Send a FindEntries message to the seed node.
        /// A FindEntriesCallback will be called when the reply is received (or it times out).
        /// </summary>
        private void SendFindEntries( Node localNode, Direction dir, NodeIdAddressBind seed,
            ArrayList newIntervals ) {
            FindEntriesCallback cb = new FindEntriesCallback();
            FindEntriesRequest fe = new FindEntriesRequest(localNode, cb,
                Node.TIMEOUT_DELAY_MS, dir, newIntervals);
            localNode.SendMessageDirect(fe, seed);
        }

        /// <summary>
        /// Given a list of intervals, a list of entries, and a list of new entries
        /// try to replace the entries with any new entries that may be preferable.
        /// </summary>
        /// <param name="intervals">An array of node ids that delineate the entries</param>
        /// <param name="entries">The array of currently selected entries</param>
        /// <param name="newEntries">An array of new entries to be inserted into 'entries'</param>
        /// <returns>Returns the number elements in 'newEntries' that were processed</returns>
        private int ChooseEntriesInternal( ArrayList intervals, PTableEntry[] entries, ArrayList newEntries ) {
            int i, j;

            if( intervals==null || intervals.Count<2 ) {
                // If there are not enough intervals, just do nothing.
                return 0;
            }
            
            Debug.Assert( entries.Length==intervals.Count );
            
            // Merge 'newEntries' into 'entries'
            for( i=0, j=0; i<newEntries.Count && j<intervals.Count; ) {
                PTableEntry newEntry = (PTableEntry) newEntries[i];
                NodeId n1 = newEntry.node.NodeId;
                NodeId n2 = (NodeId) ((j==0) ? null : intervals[j-1]);
                NodeId n3 = (NodeId) intervals[j];
                
                // Does n1 lie inside the interval n2-n3?
                // Note: If the direction is left then n2<n3 and the interval is [n2,n3)
                // If the direction is right then n3<n2 and the interval is [n3,n2)
                if( MergeCompareLess(n1,n2) ) {
                    // n1 is too small. Try next n1.
                    i++;
                } else if( ! MergeCompareLess(n1,n3) ) {
                    // n1 is too big. Try next [n2,n3).
                    j++;
                } else {
                    // n1 lies in [n2,n3). Try replacing the current entry.
                    int currentDist=Int32.MaxValue, newDist;

                    // Only check the distances if the entries are not equal
                    if( entries[j]==null || ! entries[j].node.Equals(newEntry.node) ) {
                    
                        // Get the distance of the current entry
                        if( entries[j]!=null ) {
                            currentDist = entries[j].dist;
                            Debug.Assert( currentDist!=0 );
                        }
                        
                        // Get the distance of the new entry
                        // TODO: Must use topo.BeginPing instead
                        newDist = topo.Ping( localAddrBind.GetAddress, newEntry.node.GetAddress, true );
                        if( newDist==-1 ) newDist=Int32.MaxValue;
                        
                        // If the new entry is better than the current entry, use the new entry instead
                        if( newDist<currentDist ) {
                            Log.Write(ProximityTable.CATEGORY, LogLvl.Lo,
                                "AddToCurrentEntries: Adding node {0}!", newEntry.node.GetAddress.ToString());
                            entries[j] = new PTableEntry( newEntry.node, newDist );
                        }

                    }
                    
                    i++;
                }
            }

            return i;
        }

        /// <summary>
        /// Determine if n1 is 'less than' n2, for the purposes of merging. The comparison
        /// and strictness depends on direction.
        /// </summary>
        private bool MergeCompareLess( NodeId n1, NodeId n2 ) {
            if( n2==null ) return false;
            if( dir==Direction.Left ) {
                return (n1.CompareTo(n2) < 0);
            } else {
                return (n1.CompareTo(n2) >= 0);
            }
        }
        
        /// <summary>
        /// Returns true if b is farther than a. The comparison depends on direction.
        /// </summary>
        private bool IsFarther( NodeId a, NodeId b ) {
            if( dir==Direction.Left ) {
                return b.CompareTo(a)<0;
            } else {
                return b.CompareTo(a)>0;
            }
        }

        /// <summary>
        /// Search through the current entries and return the closest one.
        /// </summary>
        /// <returns>Returns a PTableEntry object or null</returns>
        private PTableEntry GetClosest() {            
            // Search for the closest entry
            if( currentEntries==null ) return null;
        
            PTableEntry bestEntry=null;
            foreach( PTableEntry entry in currentEntries ) {
                if( entry==null ) continue;
                Debug.Assert( entry.dist!=0 );
                if( bestEntry==null || entry.dist<bestEntry.dist ) {
                    bestEntry = entry;
                }
            }
            
            return bestEntry;
        }
    }

    
    /// <summary>
    /// This object is routed during a proximity join to gather entries to
    /// fill the source node's proximity table.
    /// </summary>
    [Serializable, XmlRootAttribute("FindEntriesRequest",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class FindEntriesRequest : OverlayRequestMsg, RouteControl {
        
        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        
        public enum Phase   {
            Seed,         // Send from joining node to seed
            Routing       // Route By Name
        }
        
        
        //**********************************************************************************************
        // Members (public for serialization convenience)
        //**********************************************************************************************
        
        public NodeIdAddressBind    source;
        public Direction            dir;
        public Phase                phase;
        public static int           accidentalArrivals=0;
        
        // The name ID intervals that we're trying to fill
        ArrayList                   newIntervals;

        // An array of lists of possible entries (which correspond to the newIntervals).
        // The inner lists contain PTableEntry objects.
        SortedList[]                possibleEntries;
        
        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor for serialization
        public FindEntriesRequest() : base() {}

        public FindEntriesRequest( Node localNode, OverlayReplyCallback cb, long delayMs,
            Direction dir, ArrayList newIntervals ) : base(localNode, cb, delayMs, true)
        {
            this.source = localNode.NodeIdAddressBind;
            this.dir = dir;
            this.phase = Phase.Seed;

            // Set up the array of intervals and the lists of possible entries
            this.newIntervals = newIntervals;

            // Count how many possible entries we need for these intervals
            // (used to use newIntervals.Count-1 here but now we have an extra
            // entry that lies outside the farthest interval)
            int numEntries = newIntervals.Count;
            
            // Create a new array of possible entries
            possibleEntries = new SortedList[ numEntries ];
            for( int i=0; i<numEntries; i++ ) {
                possibleEntries[i] = new SortedList(new PTableEntryComparer(dir),
                    Parameters.maxPossibleEntries);
            }
        }

        
        //**********************************************************************************************
        // Public Properties
        //**********************************************************************************************
        
        [XmlArray(ElementName = "NewIntervals",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="NodeId", Type=typeof(NodeId),
             IsNullable=true, Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public ArrayList NewIntervals {
            get { return newIntervals; }
            set {
                Debug.Assert( value!=null );
                this.newIntervals = value;
            }
        }

        // Serialize the Possible Entries. Note that SortedList objects cannot be serialized
        // so we must convert them to arrays here. Lame but sadly necessary.
        [XmlArray( ElementName="PossibleEntries",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="PTableEntryArray", Type=typeof(PTableEntry[]),
             IsNullable=true, Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public PTableEntry[][] PossibleEntries {
            get {
                if( possibleEntries==null ) return null;
                PTableEntry[][] result = new PTableEntry[ possibleEntries.Length ][];
                for( int i=0; i<possibleEntries.Length; i++ ) {
                    if( possibleEntries[i]==null ) continue;
                    result[i] = new PTableEntry[ possibleEntries[i].Count ];
                    ICollection keys = possibleEntries[i].Keys;
                    keys.CopyTo( result[i], 0 );
                }
                return result;
            }
            set {
                if( value==null ) possibleEntries = null;
                possibleEntries = new SortedList[ value.Length ];
                for( int i=0; i<value.Length; i++ ) {
                    possibleEntries[i] = new SortedList(new PTableEntryComparer(dir), Parameters.maxPossibleEntries);
                    if( value[i]==null ) continue;
                    for( int j=0; j<value[i].Length; j++ ) {
                        possibleEntries[i].Add( value[i][j], null );
                    }
                }
            }
        }


        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************
        
        /// <summary>
        /// This method is called during the 'routing' phase to gather P-Table entries.
        /// </summary>
        public NodeIdAddressBind GetNextHop( Node node ) {
            Node                localNode = (Node) node;
            NodeIdAddressBind   nextHop=null;
            NodeId              localID = (NodeId) localNode.GetNodeIdAddressBind.GetNodeId;
            
            Debug.Assert( phase==Phase.Routing );

            Log.Write(ProximityTable.CATEGORY, LogLvl.Lo,
                "Processing FindEntries from {0} ({1})", source, dir );

            // If the source has failed, drop the message
            if( source==null ) {
                return null;
            }
            
            // Fill the possible entries in this message with entries
            // from the local node's proximity table.
            AddPossibleEntriesFromTable( localNode.ProximityTable );

            // While we're at it, fill the local node's proximity table with entries
            // from this message too
            if( Parameters.aggressivelyFindEntries ) {
                ArrayList allPossibleEntries = new ArrayList( possibleEntries.Length*Parameters.maxPossibleEntries );
                foreach( SortedList sl in possibleEntries ) {
                    Debug.Assert( sl!=null );
                    allPossibleEntries.AddRange( sl.Keys );
                }

                localNode.ProximityTable.AddToCurrentEntries(allPossibleEntries);
            }
            
            // Choose the next unfilled interval to fill
            PTableInterval nextInterval = GetNextInterval();
            if( nextInterval==null ) {
                // Mission complete -- Cause ArrivedAtDestination() to be called, which will reply to source
                Log.Write(ProximityTable.CATEGORY, LogLvl.Lo,
                    "!!! FindEntries Mission Complete {0} ({1}) !!!", source, dir );
                return localNode.NodeIdAddressBind;
            }
            
            // Route towards the interval
            if( localID.CompareTo(nextInterval.lower)<0 ) {
                // Local node is to the left of the interval.
                // Route towards the interval's upper bound (but not equal to it).
                nextHop = localNode.RouteByName(nextInterval.upper, RouteByNameOption.DisallowEqual);

                if( NodeId.LiesBetween(localID,nextInterval.upper,nextHop.NodeId) ) {
                    // Failure: We will never manage to find entries for this interval
                    // Just drop the message
                    return null;
                }
            } else if( localID.CompareTo(nextInterval.upper)>=0 ) {
                // Local node is to the right of the interval
                // Route toward the interval's lower bound (or equal to it).
                nextHop = localNode.RouteByName(nextInterval.lower, RouteByNameOption.Normal);

                if( NodeId.LiesBetween(nextHop.NodeId,nextInterval.lower,localID) ) {
                    // Failure: We will never manage to find entries for this interval
                    // Just drop the message
                    return null;
                }
            } else {
                // This should not happen -- if the local node is in the interval then
                // the interval should not be empty
                Debug.Assert( false, "Error: localID is in the interval!" );
            }
            
            return nextHop;
        }
        
        /// <summary>
        /// This function is called when the FindEntriesRequest message arrives at
        /// the seed node (Phase.Seed) or when it has filled all entries (Phase.Routing).
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node node ) {
            Node localNode = (Node) node;

            if( phase==Phase.Seed ) {
                // If this seed node is on the wrong side of the source, reply with no entries
                if( IsFarther(localNode.NodeIdAddressBind.NodeId,source.NodeId) ) {
                    FindEntriesReply reply = new FindEntriesReply(this, dir, null, null);
                    localNode.SendReply(reply, source);
                    return;
                }

                phase = Phase.Routing;
                RouteMsg routeMsg = new RoutePayloadControlMsg(localNode.NodeIdAddressBind, this);
                localNode.ResumeRouting(routeMsg);
            } else {
                Debug.Assert( phase==Phase.Routing );
                FindEntriesReply reply = new FindEntriesReply(this, dir, newIntervals, possibleEntries);
                localNode.SendReply(reply, source);
            }
        }

        /// <summary>
        /// Notifies this message that a node has failed.
        /// </summary>
        public void SendCallback( Node localNode, NodeIdAddressBind dest, bool success ) {
            // If the send succeeded, we don't care
            if( success ) {
                return;
            }

            // If the source fails, that's not good
            if( source.Equals(dest) ) {
                source = null;
            }

            // Remove the failed node from the possible entries in this message
            // (otherwise they might be reintroduced into the current node's routing
            // tables in GetNextHop()).
            if( Parameters.aggressivelyFindEntries ) {
                foreach( SortedList sl in possibleEntries ) {
                    for( int i=0; i<sl.Count; i++ ) {
                        PTableEntry entry = (PTableEntry) sl.GetKey(i);
                        if( entry.node.Equals(dest) ) {
                            sl.RemoveAt(i);
                            break;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// An indication to the simulator whether this is a 'background' message or not.
        /// Background messages are ignored by "ProcessUntilIgnoreEventsOnly()".
        /// </summary>
        public bool IsBackground() {
            return true;
        }


        //**********************************************************************************************
        // Private methods
        //**********************************************************************************************        
                
        /// <summary>
        /// Fill the possible entries of this message with the current entries from the source table.
        /// </summary>
        private void AddPossibleEntriesFromTable( ProximityTable sourceTable ) {
            ArrayList allSourceEntries = sourceTable.GetAllEntries();
            if( dir==Direction.Right ) allSourceEntries.Reverse();
            
            // ASSERT: allSourceEntries is sorted in ascending order if dir==Left and descending order if dir==Right

            AddPossibleEntries( allSourceEntries );
        }

        /// <summary>
        /// Fill the possible entries in this half of the proximity table with the
        /// given list of new entries. This must be called from ProximityTable.AddPossibleEntriesFromTable.
        /// </summary>
        /// <param name="newEntries">
        /// An array of PTableEntry objects. This array will be modified.
        /// </param>
        private void AddPossibleEntries( ArrayList newEntries ) {
            int i, j;
            
            if( newIntervals==null || newIntervals.Count<2 ) {
                Debug.Assert(false, "Should never happen");
                return;
            }
            
            Debug.Assert( possibleEntries.Length==newIntervals.Count );
            
            // Merge newEntries into possibleEntries
            for( i=0, j=0; i<newEntries.Count && j<newIntervals.Count; ) {
                PTableEntry newEntry = (PTableEntry) newEntries[i];
                NodeId n1 = newEntry.node.NodeId;
                NodeId n2 = (NodeId) ((j==0) ? null : newIntervals[j-1]);
                NodeId n3 = (NodeId) newIntervals[j];
                
                // Does n1 lie inside the interval n2-n3?
                // Note: If the direction is left then n2<n3 and the interval is [n2,n3)
                // If the direction is right then n3<n2 and the interval is [n3,n2)
                if( MergeCompareLess(n1,n2) ) {
                    // n1 is too small. Try next n1.
                    i++;
                } else if( ! MergeCompareLess(n1,n3) ) {
                    // n1 is too big. Try next [n2,n3).
                    j++;
                } else {
                    // n1 lies in [n2,n3). Add to our possible entries.
                    AddPossibleEntry( j, newEntry.node );
                    i++;
                }
            }
        }

        /// <summary>
        /// Try to add a new node to the possible entry list
        /// </summary>
        private void AddPossibleEntry( int i, NodeIdAddressBind node ) {
            SortedList arr = possibleEntries[i];
            PTableEntry entry = new PTableEntry(node, 0);
            bool containsEntry = arr.Contains(entry);
            if( !containsEntry && arr.Count<Parameters.maxPossibleEntries ) {
                Log.Write(ProximityTable.CATEGORY, LogLvl.Lo,
                    "Add possible entry {0} to {1}'s {2} join msg", node, source, dir);
                arr.Add( entry, null );
            }
        }

        /// <summary>
        /// Determine if n1 is 'less than' n2, for the purposes of merging. The comparison
        /// and strictness depends on direction.
        /// </summary>
        private bool MergeCompareLess( NodeId n1, NodeId n2 ) {
            if( n2==null ) return false;
            if( dir==Direction.Left ) {
                return (n1.CompareTo(n2) < 0);
            } else {
                return (n1.CompareTo(n2) >= 0);
            }
        }
                
        /// <summary>
        /// Returns true if b is farther than a. The comparison depends on direction.
        /// </summary>
        private bool IsFarther( NodeId a, NodeId b ) {
            if( dir==Direction.Left ) {
                return b.CompareTo(a)<0;
            } else {
                return b.CompareTo(a)>0;
            }
        }
                
        /// <summary>
        /// Finds the largest new interval without any possible entries.
        /// </summary>
        /// <returns>Returns the interval or null if all have possible entries.</returns>
        public PTableInterval GetNextInterval() {
            PTableInterval result;
            
            Debug.Assert( possibleEntries.Length==newIntervals.Count );
            
            // Find the lowest numbered interval without any possible entries
            for( int i=1; i<possibleEntries.Length; i++ ) {
                if( possibleEntries[i]!=null && possibleEntries[i].Count>0 ) {
                    // We have found some possible entries for the i'th interval
                    continue;
                }
                
                // No entries found for the i'th interval. Return an Interval object
                // to the caller.
                NodeId n1 = (NodeId) newIntervals[i-1];
                NodeId n2 = (NodeId) newIntervals[i];
                
                Debug.Assert( !n1.Equals(n2) );
                if( n1.CompareTo(n2)<0 ) {
                    result = new PTableInterval(n1,n2);
                } else {
                    result = new PTableInterval(n2,n1);
                }
                return result;
            }
            
            return null;
        }
    }
 
    
    /// <summary>
    /// Reply to a FindEntriesRequest message.
    /// </summary>
    [Serializable, XmlRootAttribute("FindEntriesReply",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class FindEntriesReply : OverlayReplyMsg {

        //**********************************************************************************************
        // Members (public for serialization convenience)
        //**********************************************************************************************
        public Direction        dir;

        // The name ID intervals that we're trying to fill
        ArrayList               newIntervals;

        // An array of possible entries for the new P-Table; elements are PTableEntry objects.
        ArrayList               possibleEntries;


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        public FindEntriesReply() {}
        public FindEntriesReply(FindEntriesRequest request, Direction dir, ArrayList newIntervals,
            SortedList[] possibleEntriesParam) : base(request)
        {
            this.dir = dir;

            if( newIntervals==null ) {
                // Empty reply
                Debug.Assert( possibleEntriesParam==null );
            } else {
                this.newIntervals = newIntervals;
                this.possibleEntries = new ArrayList();
                for( int i=0; i<possibleEntriesParam.Length; i++ ) {
                    if( possibleEntriesParam[i]==null ) continue;
                    this.possibleEntries.AddRange( possibleEntriesParam[i].Keys );
                }
            }
        }
        

        //**********************************************************************************************
        // Public Properties
        //**********************************************************************************************
        
        [XmlArray(ElementName = "NewIntervals",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="NodeId", Type=typeof(NodeId),
             IsNullable=true, Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public ArrayList NewIntervals {
            get { return newIntervals; }
            set {
                Debug.Assert( value!=null );
                this.newIntervals = value;
            }
        }

        [XmlArray(ElementName = "PossibleEntries",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="PTableEntry", Type=typeof(PTableEntry),
             IsNullable=true, Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public ArrayList PossibleEntries {
            get { return possibleEntries; }
            set {
                Debug.Assert( value!=null );
                this.possibleEntries = value;
            }
        }

    }


    /// <summary>
    /// Callback that is executed when join completes or times out.
    /// </summary>
    public class FindEntriesCallback : OverlayReplyCallback {
        public override void Execute( MessageRouting.Node node ) {
            Node localNode = (Node) node;

            // Check if join request timed out
            FindEntriesReply feReply = (FindEntriesReply) this.reply;
            if( feReply==null ) {
                // TODO: Do something better to handle this failure?
                Log.Write(ProximityTable.CATEGORY, LogLvl.Hi, "P-Table Join Failed!");
                return;
            }
            
            // If this is not an empty reply, update the local table
            if( feReply.NewIntervals!=null ) {
                Log.Write(ProximityTable.CATEGORY, LogLvl.Lo,
                    "***** FindEntries {0} is Complete()! *****", feReply.dir);

                localNode.ProximityTable.ChooseNewEntries(feReply.dir,
                    feReply.NewIntervals, feReply.PossibleEntries);
                localNode.ProximityTable.NotifyEntries(feReply.dir, localNode);
            } else {
                Log.Write(ProximityTable.CATEGORY, LogLvl.Lo,
                    "***** FindEntries {0} failed *****", feReply.dir);
            }
        }
    }
    

    /// <summary>
    /// This message is used to notify another node that we have added it to our P-Table.
    /// </summary>
    [Serializable, XmlRootAttribute("UpdatePTableMsg",
                       Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class UpdatePTableMsg : OverlayMsg 
    {

        //**********************************************************************************************
        // Public Members (public only for serialization)
        //**********************************************************************************************
        public NodeIdAddressBind       source;

        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor for serialization
        public UpdatePTableMsg() {}

        public UpdatePTableMsg( NodeIdAddressBind source ) {
            this.source = source;
        }
        

        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************

        /// <summary>
        /// This method is called when another node has added us to its Proximity Table
        /// </summary>
        public override void ArrivedAtDestination( MessageRouting.Node node ) {
            Node localNode = (Node) node;
            ProximityTable localTable = localNode.ProximityTable;
            
            Log.Write(ProximityTable.CATEGORY, LogLvl.Lo, "Received UpdatePTable msg from {0}", source);
            
            localTable.AddEntry( source );

            // TODO: Could be more aggressive here
        }
    }


    /// <summary>
    /// A timer that periodically fires to rebuild the P-Table.
    /// </summary>
    public class PTableRebuildTimer : Timer {
        public PTableRebuildTimer(long initialDelayMs, long periodMs)
            : base(initialDelayMs, periodMs) {}

        /// <summary>
        /// This method is called when the timer fires.
        /// </summary>
        public override void Execute( MessageRouting.Node node ) {
            Node localNode = (Node) node;
            NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;

            ProximityTable pt = localNode.ProximityTable;
            
            // Don't rebuild if disabled or if rebuild count has been exceeded
            if(    (! localNode.proxRebuildEnabled)
                || (SkipNet.Parameters.maxRebuilds>0 && pt.InitCount>=SkipNet.Parameters.maxRebuilds)
            ) {
                // Prevent this timer from firing again
                this.Periodic = false;
                return;
            }

            pt.Rebuild();
        }
    }


    /// <summary>
    /// This structure describes a P Table "Interval", which is a range of identifiers
    /// in which we intend to store a neighbour pointer.
    /// </summary>
    public class PTableInterval {
        public NodeId       lower, upper;
        
        // Default constructor for serialization
        public PTableInterval() {}

        public PTableInterval( NodeId lower, NodeId upper ) {
            Debug.Assert( lower.CompareTo(upper)<0 );
            this.lower = lower;
            this.upper = upper;
        }
    }
    

    /// <summary>
    /// This structure describes a single neighbour entry in the P Table, which consists
    /// of the node id/address pair, and the cached distance to that node.
    /// </summary>
    [Serializable, XmlRootAttribute("PTableEntry",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class PTableEntry : IComparable {
        // Public -> automatically serialized
        public NodeIdAddressBind    node;
        public int                  dist;

        // Default constructor for serialization
        public PTableEntry() {}

        // Constructor
        public PTableEntry( NodeIdAddressBind node, int dist ) {
            this.node = node;
            this.dist = dist;
        }

        // Compare ordering of two PTableEntry objects
        public int CompareTo(object comp) {
            PTableEntry compEntry = (PTableEntry) comp;
            return node.NodeId.CompareTo(compEntry.node.NodeId);
        }

        // Overridden equals
        public override bool Equals(Object o) {
            if (!(o is PTableEntry)) 
                return false;
            PTableEntry ent = (PTableEntry) o;
            return this.node.Equals( ent.node );
        }
        
        // Overridden hashcode (used for Hashtables)
        public override int GetHashCode() {
            return node.GetHashCode();
        }
    }
    
    
    /// <summary>
    /// Compare two PTableEntry objects in forward or reverse order.
    /// </summary>
    public class PTableEntryComparer : IComparer {
        int factor;

        public PTableEntryComparer( Direction dir ) {
            if( dir==Direction.Left ) {
                factor = 1;
            } else {
                factor = -1;
            }
        }

        public int Compare( object A, object B ) {
            PTableEntry a = (PTableEntry) A;
            PTableEntry b = (PTableEntry) B;
            return factor * a.CompareTo( b );
        }
    }
}
