using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {
    
    /// <summary>
    /// A RoutingTable contains the neighbours of a node, stored as an array of LeafSets at various
    /// heights. The neighbours are constructed by the join logic, which is not implemented here.
    /// This class also contains logic to route messages by name ID.
    /// </summary>
    public class RoutingTable {

        //**********************************************************************************************
        // Constants
        //**********************************************************************************************
        public const string     CATEGORY = "RTable";


        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************
        Node                    localNode;
        NodeIdAddressBind       localAddrBind;
        LeafSet[]               nbrs;
        int                     height;
    
    
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        public RoutingTable( Node localNode ) {
            this.localNode = localNode;
            this.localAddrBind = localNode.NodeIdAddressBind;
            this.height = Parameters.maxHeight;

            nbrs = new LeafSet[ height ];
            for( int h=0; h<height; h++ ) {
                // The lowest level of the R-Table is called "the Leaf Set" (but in fact we use
                // leaf sets to store the neighbours at every level). The bottom level leaf set's size
                // is given by Parameters.leafSetHalfSize. For higher levels, a dense R-Table has size
                // b-1 and a sparse R-Table has size 1.
                int maxSize = Parameters.denseRTable ? Parameters.b-1 : 1;
                if( h==0 ) {
                    maxSize = Math.Max( maxSize, Parameters.leafSetHalfSize );
                }

                nbrs[h] = new SkipNet.LeafSet(localNode, maxSize, (h==0) );
            }
        }
    
    
        //**********************************************************************************************
        // Properties
        //**********************************************************************************************

        public int Height {
            get { return height; }
            set { height = value; }
        }
    
        public LeafSet[] Neighbours {
            get { return nbrs; }
            set {
                Debug.Assert( value!=null );
                Debug.Assert( value.Length==height );
                nbrs = value;
            }
        }

        public LeafSet LeafSet {
            get { return nbrs[0]; }
        }
    
    
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************

        /// <summary>
        /// Get the neighbour at height h (0 <= h < height) in the given direction
        /// The neighbour is the node with closest NodeId in the given direction among the nodes
        /// that match h digits with the local node.
        /// </summary>
        /// <param name="h">The height</param>
        /// <returns> Returns the NodeIdAddressBind for that neighbour. If no such neighbour exists
        /// (i.e. the ring at height h contains just this node) this function returns the local node address.
        /// </returns>
        public NodeIdAddressBind GetNeighbour(Direction dir, int h) {
            Debug.Assert( 0<=h && h<height );
            NodeIdAddressBind nbr = nbrs[h].GetClosest(dir);
            if( nbr==null ) {
                nbr = localAddrBind;
            }
            return nbr;
        }

        /// <summary>
        /// Try to add this node as a neighbour at height h
        /// The given node must match at least h numeric id digits with the local node
        /// 
        /// NOTE: This function calls "RebuildNeigbours()" if necessary.
        /// </summary>
        /// <param name="h">The height</param>
        /// <param name="n">The NodeIdAddressBind for that neighbour.</param>
        public bool AddNode(int h, NodeIdAddressBind n) {
            Debug.Assert( 0<=h && h<height );
            bool added = nbrs[h].AddNode(n);
            return added;
        }

        /// <summary>
        /// Try to add the nodes in the given array as neighbours at height h
        /// The given nodes must all match at least h numeric id digits with the local node
        /// 
        /// NOTE: This function calls "RebuildNeigbours()" if necessary.
        /// </summary>
        /// <param name="h">The height</param>
        /// <param name="n">An array of NodeIdAddressBind objects</param>
        public bool AddNodes(int h, ArrayList newNodes) {
            Debug.Assert( 0<=h && h<height );
            bool added = nbrs[h].AddNodes(newNodes);
            return added;
        }
        
        /// <summary>
        /// Determine the next hop in routing towards the destination.
        /// </summary>
        /// <param name="dest">The destination ID</param>
        /// <param name="option">Routing options</param>
        /// <returns>The address of the next hop</returns>
        public NodeIdAddressBind RouteByName( NodeId destID, RouteByNameOption option ) {
            NodeIdAddressBind nextHop;
            
            // Determine if we should route to the left or to the right
            int cmp = destID.CompareTo(localAddrBind.NodeId);
            if( cmp<0 ) {
                // Destination ID is to the left
                nextHop = RouteByName(destID, Direction.Left, option);
                
                // The next hop must always be to our left
                Debug.Assert( nextHop.CompareTo(localAddrBind)<=0 );
            } else if( cmp>0 ) {
                // Destination ID is to the right
                nextHop = RouteByName(destID, Direction.Right, option);
                
                // The next hop must always be to our right
                Debug.Assert( nextHop.CompareTo(localAddrBind)>=0 );
            } else {
                // Destination ID exactly matches local ID
                // Use the level-0 leaf set to figure out what to do.
                return nbrs[0].RouteByName(destID, option);
            }
            
            return nextHop;
        }
                
        /// <summary>
        /// A node has failed and thus we are to remove it from the routing table, if it exists.
        /// 
        /// NOTE: The caller must call 'RebuildNeighbours()' if the node was removed, otherwise
        /// the table of all neighbours will be stale.
        /// </summary>
        /// <param name="nodeAddr">The failed node</param>
        /// <returns>True if the node was removed</returns>
        public bool NodeFailed( NodeIdAddressBind nodeAddr ) {
            bool removed=false;
            for( int h=0; h<height; h++ ) {
                removed |= nbrs[h].NodeFailed(nodeAddr);
            }
            return removed;
        }

        /// <summary>
        /// Insert each neighbour into the given Hashtable 
        /// </summary>
        /// <param name="ht">Table to insert neighbours into</param>
        public void GetAllNeighbours(Hashtable ht) {
            for( int h=0; h<height; h++ ) {
                nbrs[h].GetAllNeighbours(ht);
            }
        }

        /// <summary>
        /// Start a periodic timer instructing us to ping our neighbours
        /// </summary>
        internal void SchedulePing(double delayFactor) {
            long initialDelayMs = (long) (delayFactor * Parameters.neighbourPingPeriod);
            Timer pingTimer = new RTablePingTimer(initialDelayMs, Parameters.neighbourPingPeriod);
            localNode.SetTimer(pingTimer);
        }

        public static int NumNeighborRebuilds = 0;

        /// <summary>
        /// Take the set of nodes 'nbrs' and try to insert them at level 'height'.
        /// Recursively fix higher levels if necessary.
        /// </summary>
        internal void RepairLevel( int height, ArrayList newNbrs, bool background ) {
            RoutePayloadControlMsg  routeMsg;
            FixupHigher             fixupHigher;
            bool                    addedLeft=false, addedRight=false;

            Log.Write(RoutingTable.CATEGORY, LogLvl.Lo, "Repairing level {0}", height);

            // Try to insert each of the nodes, one-by-one
            foreach( NodeIdAddressBind n in newNbrs ) {
                LeafSet ls = (LeafSet) this.nbrs[height];
                addedLeft  |= ls.LeftHalf.AddNode(n);
                addedRight |= ls.RightHalf.AddNode(n);
            }

            // If the neighbour sets have changed, rebuild the list of all neighbours
            if( addedLeft || addedRight ) {
                NumNeighborRebuilds++;
                localNode.RebuildNeighbours(null);
            }
            
            if( addedLeft ) {
                LeafSet ls = (LeafSet) this.nbrs[height];

                Log.Write(RoutingTable.CATEGORY, LogLvl.Lo, "Fixing level {0} to left", height);

                // We added some new nodes to the routing table. We should send
                // out a ping so that they will point to us as well.
                ls.SendPing(Direction.Left, height, background);

                // Send FixupHigher messages to try to fix pointers at higher levels too
                fixupHigher = new FixupHigher(localAddrBind, height, Direction.Left, background);
                routeMsg = new RoutePayloadControlMsg(localAddrBind, fixupHigher);
                localNode.RouteMessageInternal(routeMsg);
            }

            if( addedRight ) {
                LeafSet ls = (LeafSet) this.nbrs[height];

                Log.Write(RoutingTable.CATEGORY, LogLvl.Lo, "Fixing level {0} to right", height);

                // We added some new nodes to the routing table. We should send
                // out a ping so that they will point to us as well.
                ls.SendPing(Direction.Right, height, background);
      
                // Send FixupHigher messages to try to fix pointers at higher levels too
                fixupHigher = new FixupHigher(localAddrBind, height, Direction.Right, background);
                routeMsg = new RoutePayloadControlMsg(localAddrBind, fixupHigher);
                localNode.RouteMessageInternal(routeMsg);
            }
        }

        /// <summary>
        /// Dump the routing table to the console.
        /// </summary>
        public void Dump() {
            for( int h=0; h<height; h++ ) {
                ArrayList leaves = nbrs[h].GetAllNbrs();
                if( leaves.Count>0 ) {
                    Console.Write( "Level {0}: ", h );
                    leaves = nbrs[h].LeftHalf.GetAllNbrs();
                    leaves.Reverse();
                    foreach( NodeIdAddressBind n in leaves ) {
                        Console.Write( "{0} ", n );
                    }

                    Console.Write(" / ");

                    leaves = nbrs[h].RightHalf.GetAllNbrs();
                    foreach( NodeIdAddressBind n in leaves ) {
                        Console.Write( "{0} ", n );
                    }
                    Console.WriteLine("");
                } else {
                    Console.WriteLine("");
                    break;
                }
            }
        }
    

        //**********************************************************************************************
        // Private Methods
        //**********************************************************************************************

        /// <summary>
        /// Determine the next hop in routing towards the destination.
        /// 
        /// NOTE: This should only be called by RTable.RouteByName.
        /// </summary>
        /// <param name="dest">The destination ID</param>
        /// <param name="dir">The direction in which to route</param>
        /// <param name="option">Routing options</param>
        /// <returns>The address of the next hop</returns>
        private NodeIdAddressBind RouteByName( NodeId destID, Direction dir, RouteByNameOption option ) {
            NodeIdAddressBind nextHop, bestHop=localAddrBind;

            // We could not arrive at the destination in one hop. Allow multiple hops,
            // and higher-level pointers.
            for( int h=0; h<height; h++ ) {
                nextHop = nbrs[h].RouteByName(destID, option);

                // PrefixMatch option: any matching node is acceptable
                if(    option==RouteByNameOption.PrefixMatch
                    && nextHop.NodeId.IsWithinPrefix(destID) ) {
                    return nextHop;
                }

                if(    bestHop.Equals(localAddrBind)
                    || (   NodeId.LiesBetweenDir(dir,localAddrBind.NodeId,bestHop.NodeId,nextHop.NodeId)
                        && ! nextHop.Equals(localAddrBind) )
                ) {
                    bestHop = nextHop;
                }
            }
            
            return bestHop;
        }
    }

    
    /// <summary>
    /// A periodic timer used to ping our neighbours.
    /// </summary>
    class RTablePingTimer : Timer {
        public RTablePingTimer(long initialDelayMs, long periodMs)
            : base(initialDelayMs, periodMs) {}

        public override void Execute( MessageRouting.Node node ) {
            Node localNode = (Node) node;
            NodeIdAddressBind localAddrBind = localNode.NodeIdAddressBind;

            // If pings have been disabled, abort this timer
            if( ! localNode.pingEnabled ) {
                this.Periodic = false;
                return;
            }
                
            // TODO: Probably don't want to ping them all at once
            Log.Write(RoutingTable.CATEGORY, LogLvl.Med, "*** Pinging neighbours now... ***");
            for( int h=0; h<Parameters.maxHeight; h++ ) {
                LeafSet ls = localNode.RoutingTable.Neighbours[h];

                // Send pings to left and right
                ls.SendPing(Direction.Left, h, true);
                ls.SendPing(Direction.Right, h, true);
                
                // If we're doing simple pings, we also need to send FixupHigher messages
                if( ! Parameters.smartPingEnabled ) {
                    FixupHigher fixupHigher;
                    RoutePayloadControlMsg routeMsg;

                    fixupHigher = new FixupHigher(localAddrBind, h, Direction.Left, true);
                    routeMsg = new RoutePayloadControlMsg(localAddrBind, fixupHigher);
                    localNode.RouteMessageInternal(routeMsg);

                    fixupHigher = new FixupHigher(localAddrBind, h, Direction.Right, true);
                    routeMsg = new RoutePayloadControlMsg(localAddrBind, fixupHigher);
                    localNode.RouteMessageInternal(routeMsg);
                }
            }
        }
    }

}
