using System;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;

namespace Microsoft.Research.Herald.SkipNet {


    public class SerializationConfig {
        public const string SKIPNET_NAMESPACE = "http://research.microsoft.com/sn/Herald/SkipNet";
    }


    /// <summary>
    /// Functions that are used to control routing.
    /// </summary>
    internal interface RouteControl {
        /// <summary>
        /// Used to determine the next hop for a RouteMsg.
        /// If the returned next hop is the local node's address, the message is
        /// considered to have arrived at its destination.
        /// Must be overriden in derived classes.
        /// </summary>
        NodeIdAddressBind GetNextHop( Node localNode );

        /// <summary>
        /// After each time SkipNet tries to send this message to another node,
        /// this callback notifies the message whether the send was successful or not.
        /// </summary>
        void SendCallback( Node localNode, NodeIdAddressBind dest, bool success );

        /// <summary>
        /// An indication to the simulator whether this is a 'background' message or not.
        /// Background messages are ignored by "ProcessUntilIgnoreEventsOnly()".
        /// </summary>
        bool IsBackground();
    }


    /// <summary>
    /// An abstract message that can be routed by SkipNet
    /// </summary>
    [Serializable,
        XmlRootAttribute("RouteMsg", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public abstract class RouteMsg : MessageRouting.RouteMsg, RouteControl {

        //**********************************************************************************************
        // Protected Members
        //**********************************************************************************************
        protected NodeIdAddressBind     source = null;
        protected OverlayMsg            payload;
        protected bool                  background = false;
        
        
        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************

        // Serialized Members
        int                         hops;
        string                      path;

        // Non-serialized Members
        int                         msgId;

        // Static Members
        static int                  msgCounter=0;


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        // Default constructor
        public RouteMsg() {
            ResetMessage(null, false, null);
        }

        // Set the contents of a message, using Skip Net objects as parameters
        public RouteMsg(NodeIdAddressBind source, bool background) {
            ResetMessage(source, background, null);
        }

        // Set the contents of a message, using Skip Net objects as parameters
        public RouteMsg(NodeIdAddressBind source, bool background, OverlayMsg payload) {
            ResetMessage(source, background, payload);
        }
        

        //**********************************************************************************************
        // Properties
        //**********************************************************************************************

        [XmlElementAttribute("Source",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public override MessageRouting.NodeIdAddressBind Source {
            set { source = (NodeIdAddressBind) value; }
            get { return source; }
        }

        [XmlElementAttribute("Payload",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public OverlayMsg Payload {
            get { return payload; }
            set { payload = value; }
        }
        
        [XmlElementAttribute("Hops",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public int Hops {
            get { return hops; }
            set { hops = value; }
        }

        [XmlElementAttribute("Path",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public string Path {
            get { return path; }
            set { path = value; }
        }
        
        
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************
        
        // Set the contents of a message, using Skip Net objects as parameters
        public void ResetMessage(NodeIdAddressBind source, bool background, OverlayMsg payload) {
            this.background = background;
            this.source = source;
            this.payload = payload;
            this.hops = 0;
            this.path = "";         
            this.msgId = msgCounter++;
        }

        // Add a node to the message path
        public void AddToPath( NodeIdAddressBind nextHop ) {
            path = path + nextHop + ", ";
        }

        /// <summary>
        /// An indication to the simulator whether this is a 'background' message or not.
        /// Background messages are ignored by "ProcessUntilIgnoreEventsOnly()".
        /// </summary>
        public override bool IsBackground() {
            return this.background;
        }

        /// <summary>
        /// Determine if this RouteMsg equals the given object.
        /// Equality is determined by whether they contain the same payload object.
        /// </summary>
        /// <returns>True if equal, false otherwise</returns>
        public override bool Equals( object o ) {
            if( ! (o is RouteMsg) ) {
                return false;
            }
            RouteMsg msg = (RouteMsg) o;
            return ( this.payload == msg.payload );
        }

        /// <summary>
        /// Returns a hash code for use with Hashtables.
        /// </summary>
        public override int GetHashCode() {
            return this.payload.GetHashCode();
        }


        //**********************************************************************************************
        // Abstract RouteControl Methods
        //**********************************************************************************************

        public abstract NodeIdAddressBind GetNextHop( Node localNode );

        public abstract void SendCallback( Node localNode, NodeIdAddressBind dest, bool success );

        /// <summary>
        /// This is called when SkipNet suggests we should call "CheckContinueRouting" on the payload.
        /// </summary>
        public virtual bool CallCCR( Node localNode ) {
            return localNode.MessageCCRHook( payload );
        }
    }
    

    /// <summary>
    /// This class is a message from a SkipNet node to itself. It is used
    /// to implement continuations and timers.
    /// </summary>
    public class LocalMsg : RouteMsg {

        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        public LocalMsg( Continuation cont ) : base() {
            this.payload = cont;
            this.background = true;
        }
        
        //**********************************************************************************************
        // Overriding Methods
        //**********************************************************************************************

        public override MessageRouting.NodeId Destination {
            get { return null; }
            set {}
        }

        public override NodeIdAddressBind GetNextHop( Node localNode ) {
            return localNode.NodeIdAddressBind;
        }

        public override void SendCallback( Node localNode, NodeIdAddressBind dest, bool success ) {
            if( ! success ) {
                throw new InvalidOperationException("Local node should never fail!");
            }
        }

        public override bool IsBackground() {
            Continuation cont = (Continuation) payload;
            return cont.IsBackground();
        }
    }

    
    /// <summary>
    /// A message from a SkipNet node directly to another node.
    /// </summary>
    [Serializable,
        XmlRootAttribute(ElementName="RouteDirectMsg",Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class RouteDirectMsg : RouteMsg {

        //**********************************************************************************************
        // Private Members
        //**********************************************************************************************

        // Serialized members
        NodeIdAddressBind       dest;


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        /// <summary>
        /// Default constructor
        /// </summary>
        public RouteDirectMsg() {}

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="source">The source node</param>
        /// <param name="dest">The destination node</param>
        /// <param name="payload">The overlay message to deliver</param>
        /// <param name="cb">A callback that is invoked when the send completes. May be null.</param>
        public RouteDirectMsg(NodeIdAddressBind source, NodeIdAddressBind dest, OverlayMsg payload)
            : base(source, false, payload)
        {
            this.dest = dest;
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="source">The source node</param>
        /// <param name="dest">The destination node</param>
        /// <param name="payload">The overlay message to deliver</param>
        /// <param name="cb">A callback that is invoked when the send completes. May be null.</param>
        /// <param name="background">Indicate if this should be treated as a background message.</param>
        public RouteDirectMsg(NodeIdAddressBind source, NodeIdAddressBind dest, bool background, OverlayMsg payload)
            : base(source, background, payload) {
            this.dest = dest;
        }


        //**********************************************************************************************
        // Properties
        //**********************************************************************************************
        
        [XmlElementAttribute("Destination",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public override MessageRouting.NodeId Destination {
            get { return dest.NodeId; }
            set {  }
        }
        
        [XmlElementAttribute("DestAddrBind",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public NodeIdAddressBind DestAddrBind {
            get { return dest; }
            set { dest = (NodeIdAddressBind) value; }
        }


        //**********************************************************************************************
        // Overriding Methods
        //**********************************************************************************************

        public override NodeIdAddressBind GetNextHop( Node localNode ) {
            return dest;
        }

        /// <summary>
        /// Callback that is invoked when the send to the destination node completes,
        /// either successfully or unsuccessfully.
        /// </summary>
        public override void SendCallback( Node localNode, NodeIdAddressBind dest, bool success ) {
            Debug.Assert( source.Equals(localNode.NodeIdAddressBind) );
            Debug.Assert( dest.Equals(this.dest) );

            if( ! success ) {
                this.dest = null;
            }
        }
    }
    

    /// <summary>
    /// An abstract SkipNet message that is routed by Name ID.
    /// </summary>
    [Serializable,
        XmlRootAttribute("RouteByNameMsg", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public class RouteByNameMsg : SkipNet.RouteMsg {

        //**********************************************************************************************
        // Public Members (public for serialization convenience)
        //**********************************************************************************************
        public NodeId               destID;
        public RouteByNameOption    option;


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        public RouteByNameMsg() : base() {}

        public RouteByNameMsg(NodeIdAddressBind source, bool background, NodeId destID,
            RouteByNameOption option, OverlayMsg payload) : base(source, background, payload)
        {
            this.destID = destID;
            this.option = option;
        }
        

        //**********************************************************************************************
        // Properties
        //**********************************************************************************************
        
        [XmlElementAttribute("Destination",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public override MessageRouting.NodeId Destination {
            get { return destID; }
            set { destID = (NodeId) value; }
        }


        //**********************************************************************************************
        // Overriding Members
        //**********************************************************************************************

        /// Called to find the next hop address towards a destination name ID. If allowEqual is
        /// false then the destination must be the closest node to the destination ID that is
        /// not beyond it.
        public override NodeIdAddressBind GetNextHop( Node localNode ) {
            return localNode.RouteByName( this.destID, this.option );
        }

        /// <summary>
        /// After each time SkipNet tries to send this message to another node,
        /// this callback notifies the message whether the send was successful or not.
        /// </summary>
        public override void SendCallback( Node localNode, NodeIdAddressBind dest, bool success ) {
            // This message doesn't store any NodeIdAddressBind objects so we
            // don't care whether the send succeeded or not.
        }
    }


    /// <summary>
    /// A RouteMsg where the payload can do the routing control.
    /// The Payload must implement the RoutingControl interface.
    /// </summary>
    [Serializable,
        XmlRootAttribute(ElementName="RoutePayloadControlMsg",Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
    public class RoutePayloadControlMsg : SkipNet.RouteMsg {
        
        public RoutePayloadControlMsg() : base() {}

        public RoutePayloadControlMsg(NodeIdAddressBind source, OverlayMsg payload)
            : base(source, false, payload)
        {
            Debug.Assert( payload is RouteControl );
        }

        [XmlElementAttribute("Destination",
             Namespace=MessageRouting.SerializationConfig.MESSAGEROUTING_NAMESPACE)]
        public override MessageRouting.NodeId Destination {
            get { return null; }
            set { }
        }

        /// <summary>
        /// An indication to the simulator whether this is a 'background' message or not.
        /// Background messages are ignored by "ProcessUntilIgnoreEventsOnly()".
        /// </summary>
        public override bool IsBackground() {
            Debug.Assert( payload is RouteControl );
            RouteControl rcPayload = (RouteControl) payload;
            return rcPayload.IsBackground();
        }

        /// <summary>
        /// Called to get the next hop for this message.
        /// </summary>
        public override NodeIdAddressBind GetNextHop( Node localNode ) {
            RouteControl control = (RouteControl) payload;
            return control.GetNextHop(localNode);
        }

        /// <summary>
        /// After each time SkipNet tries to send this message to another node,
        /// this callback notifies the message whether the send was successful or not.
        /// </summary>
        public override void SendCallback( Node localNode, NodeIdAddressBind dest, bool success ) {
            RouteControl control = (RouteControl) payload;
            control.SendCallback( localNode, dest, success );
        }        
    }

}
