using System;
using System.Collections;
using System.Diagnostics;
using System.Xml.Serialization;

using MessageRouting;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.SkipNet {

    /// <summary>
    /// Stores a set of pointers to other SkipNet nodes.
    /// </summary>
    [Serializable, XmlRootAttribute("RoutingSet",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public abstract class RoutingSet {

        //**********************************************************************************************
        // Protected members
        //**********************************************************************************************
        protected NodeIdAddressBind   localAddrBind;
        protected Node                localNode;            // Not serialized
        protected RoutingSetHalf      left, right;


        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************
        
        // Default constructor needed for serialization
        public RoutingSet() {}
        
        /// <summary>
        /// Routing Set Constructor
        /// Subclasses must create the left and right RoutingSetHalf objects.
        /// </summary>
        /// <param name="localAddrBind">Address of the local node</param>
        /// <param name="localNode">Local node object. May be null.</param>
        /// <param name="maxSize">The maximum size of _half_ of the leaf set</param>
        /// <param name="bottomLevel">True if this set is the authoritative set for its kind of routing</param>
        public RoutingSet(Node localNode, NodeIdAddressBind localAddrBind, int maxSize, bool authoritative) {
            this.localNode = localNode;
            this.localAddrBind = localAddrBind;
        }


        //**********************************************************************************************
        // Public Properties
        //**********************************************************************************************
        
        [XmlElementAttribute("OwnerNode",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public NodeIdAddressBind OwnerNode {
            get { return localAddrBind; }
            set { localAddrBind = value; }
        }

        [XmlElementAttribute("LeftHalf",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public RoutingSetHalf LeftHalf {
            get { return left; }
            set { left = value; }
        }
        
        [XmlElementAttribute("RightHalf",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public RoutingSetHalf RightHalf {
            get { return right; }
            set { right = value; }
        }

        
        //**********************************************************************************************
        // Public methods
        //**********************************************************************************************
        
        /// <summary>
        /// Try to add a single node to the routing set
        /// 
        /// NOTE: This function calls "RebuildNeigbours()" if necessary.
        /// </summary>
        /// <param name="newNode">The node to add</param>
        /// <returns>True if the node was added</returns>
        public virtual bool AddNode( NodeIdAddressBind newNode ) {
            bool added=false;

            added |= left.AddNode( newNode );
            added |= right.AddNode( newNode );

            if( added && localNode!=null ) {
                localNode.RebuildNeighbours(null);
            }

            return added;
        }

        /// <summary>
        /// Try to add an array of nodes to our leaf set.
        /// 
        /// NOTE: This function calls "RebuildNeigbours()" if necessary.
        /// </summary>
        /// <param name="newNodes">An array of NodeIdAddressBind objects</param>
        /// <returns>True if any node was added</returns>
        public virtual bool AddNodes( ArrayList newNodes ) {
            bool added=false;

            foreach( NodeIdAddressBind newNode in newNodes ) {
                added |= left.AddNode( newNode );
                added |= right.AddNode( newNode );
            }

            if( added && localNode!=null ) {
                localNode.RebuildNeighbours(null);
            }

            return added;
        }

        /// <summary>
        /// Get the specified half of the leaf set.
        /// </summary>
        public RoutingSetHalf GetHalf( Direction dir ) {
            if( dir==Direction.Left ) {
                return left;
            } else {
                return right;
            }
        }

        /// <summary>
        /// Get the closest node in the given direction.
        /// </summary>
        /// <param name="dir"></param>
        /// <returns>The NodeIdAddressBind or null</returns>
        public NodeIdAddressBind GetClosest( Direction dir ) {
            if( dir==Direction.Left ) {
                return left.GetClosest();
            } else {
                return right.GetClosest();
            }
        }
        
        /// <summary>
        /// Get an array containing all NodeIdAddressBinds in this routing set.
        /// This array may be modified. It is ordered: (farthest to left, ..., farthest to right).
        /// </summary>
        public ArrayList GetAllNbrs() {
            // Get the neighbours in the two halves of the routing set.
            ArrayList result = left.GetAllNbrs();
            result.Reverse();
            result.AddRange( right.GetAllNbrs() );
            
            return result;
        }
        
        /// <summary>
        /// A node has failed and thus we are to remove it from the routing set, if it exists.
        /// 
        /// NOTE: The caller must call 'RebuildNeighbours()' if the node was removed, otherwise
        /// the table of all neighbours will be stale.
        /// </summary>
        /// <returns>True if the node was actually removed</returns>
        public bool NodeFailed( NodeIdAddressBind nodeAddr ) {
            bool removed;
            removed  = left.RemoveNode( nodeAddr );
            removed |= right.RemoveNode( nodeAddr );
            return removed;
        }

        /// <summary>
        /// Insert each neighbour into the given Hashtable 
        /// </summary>
        /// <param name="ht">Table to insert neighbours into</param>
        public void GetAllNeighbours(Hashtable ht) {
            left.GetAllNeighbours(ht);
            right.GetAllNeighbours(ht);
        }
    }
    
    
    /// <summary>
    /// Implements one half of a routing set (either left or right)
    /// </summary>
    [Serializable, XmlRootAttribute("RoutingSetHalf",
        Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE,IsNullable=true)]
    public abstract class RoutingSetHalf {
        
        //**********************************************************************************************
        // Protected Members
        //**********************************************************************************************
        protected Node                localNode;      // The local node object (not serialized)
        protected NodeIdAddressBind   localAddrBind;  // The id and address of the owning node
        protected Direction           dir;            // Is this a left or right set?
        protected int                 maxSize;        // The maximum size of the set. 0 means no maximum.
        protected ArrayList           nbrs;           // The array of neighbours in this set.
        protected bool                authoritative;  // Is this routing-set half authoritative?

        
        //**********************************************************************************************
        // Constructors
        //**********************************************************************************************

        // Default constructor needed for serialization
        public RoutingSetHalf() {}
        
        // localNode may be null to dissociate this set from a node
        public RoutingSetHalf( Node localNode, NodeIdAddressBind localAddrBind, Direction dir,
            int maxSize, bool authoritative )
        {
            this.localAddrBind = localAddrBind;
            this.localNode = localNode;
            this.dir = dir;
            this.maxSize = maxSize;
            if( maxSize==0 ) {
                this.nbrs = new ArrayList();    // Unlimited size
            } else {
                this.nbrs = new ArrayList( maxSize+1 );
            }
            this.authoritative = authoritative;
        }
        
        
        //**********************************************************************************************
        // Public Properties
        //**********************************************************************************************
        
        [XmlElementAttribute("OwnerNode", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public NodeIdAddressBind OwnerNode {
            get { return localAddrBind; }
            set { localAddrBind = value; }
        }

        [XmlElementAttribute("Direction", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public Direction Direction {
            get { return dir; }
            set { dir = value; }
        }

        [XmlElementAttribute("MaxSize", Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE)]
        public int MaxSize {
            get { return maxSize; }
            set { maxSize = value; }
        }

        /// <summary>
        /// Get all neighbours in this routing-set half. The array should not be modified.
        /// </summary>
        [XmlArray( ElementName="Neighbours",
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        [XmlArrayItem( ElementName="Nbrs", Type=typeof(NodeIdAddressBind),
             Namespace=SkipNet.SerializationConfig.SKIPNET_NAMESPACE, IsNullable=true)]
        public ArrayList Neighbours {
            get { return nbrs; }
            set {
                Debug.Assert( value!=null );
                nbrs = value;
            }
        }

        
        //**********************************************************************************************
        // Public Methods
        //**********************************************************************************************
        
        /// <summary>
        /// Add information from the given routing set to our own.
        /// 
        /// NOTE: The caller must call 'RebuildNeighbours()' if a node was added, otherwise
        /// the table of all neighbours will be stale.
        /// </summary>
        /// <param name="nbr">The appropriate half of the neighbour's routing set</param>
        /// <returns>True if any node was actually added</returns>
        public bool Add( RoutingSetHalf nbr ) {
            bool added = false;
            foreach( Object o in nbr.nbrs ) {
                NodeIdAddressBind n = (NodeIdAddressBind) o;
                added |= AddNode( n );
            }
            return added;
        }
        
        /// <summary>
        /// Try to add a single node to this routing set.
        /// 
        /// NOTE: The caller must call 'RebuildNeighbours()' if a node was added, otherwise
        /// the table of all neighbours will be stale.
        /// </summary>
        /// <param name="newNode">The node to add</param>
        /// <param name="authoritative">True if this routing set is authoritative</param>
        public bool AddNode( NodeIdAddressBind newNode ) {
            int i;
            bool added=false;

            Debug.Assert( nbrs.Count<=maxSize || maxSize==0 );

            // Never insert the local node into its own routing set
            if( newNode.Equals(localAddrBind) ) {
                return false;
            }

            // Find this newNode's position in our routing set
            NodeIdAddressBind prev = localAddrBind;
            for( i=0; i<nbrs.Count; i++ ) {
                NodeIdAddressBind existingNode = (NodeIdAddressBind) nbrs[i];

                // If the new node already exists in the routing set, don't insert it
                if( existingNode.Equals(newNode) ) {
                    return false;
                }
                
                if( LiesBetween(prev, newNode, existingNode) ) {
                    // The new node should be inserted before the i'th node
                    break;
                }
                prev = existingNode;
            }
            
            // Check if this new node should be inserted
            if( i<maxSize || maxSize==0 ) {
                nbrs.Insert( i, newNode );

                // If the routing set has overflowed, remove the farthest node
                if( nbrs.Count>maxSize && maxSize!=0 ) {
                    nbrs.RemoveAt( nbrs.Count-1 );
                }

                added = true;
            }
            
            // OPTIONAL: If the new node is now the closest node to us,
            // we can proactively update our routing table.
            if( i==0 ) {
                //NeighbourPing nbrPing = new NeighbourPing(newNode, dir, 0);
                //nbrPing.Process(localNode);
            }
            
            return added;
        }
        
        /// <summary>
        /// Get an array containing all nodes in this routing set.
        /// The nodes are ordered (closest, ..., farthest).
        /// This array may be modified.
        /// </summary>
        public ArrayList GetAllNbrs() {
            return (ArrayList) nbrs.Clone();
        }
        
        /// <summary>
        /// A node has failed and should be removed from this half of the routing set.
        /// 
        /// NOTE: The caller must call 'RebuildNeighbours()' if the node was removed, otherwise
        /// the table of all neighbours will be stale.
        /// </summary>
        /// <param name="node">A node, which may or may not exist in the routing set</param>
        /// <returns>True if the node was actually removed</returns>
        public bool RemoveNode( NodeIdAddressBind node ) {
            for( int i=0; i<nbrs.Count; i++ ) {
                NodeIdAddressBind existingNode = (NodeIdAddressBind) nbrs[i];
                if( existingNode.Equals(node) ) {
                    nbrs.RemoveAt(i);
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Insert each neighbour into the given Hashtable 
        /// </summary>
        /// <param name="ht">Table to insert neighbours into</param>
        public void GetAllNeighbours(Hashtable ht) {
            for( int i=0; i<nbrs.Count; i++ ) {
                NodeIdAddressBind existingNode = (NodeIdAddressBind) nbrs[i];
                if( ! ht.ContainsKey(existingNode) ) 
                    ht.Add(existingNode, null);
            }
        }

        /// <summary>
        /// Returns the node that is closest to the local node in this routing-set half.
        /// </summary>
        public NodeIdAddressBind GetClosest() {
            if( nbrs.Count==0 ) {
                return null;
            } else {
                return (NodeIdAddressBind) nbrs[0];
            }
        }
        
        /// <summary>
        /// Returns the farthest node from the local node in this routing-set half.
        /// </summary>
        public NodeIdAddressBind GetFarthest() {
            if( nbrs.Count==0 ) {
                return null;
            } else {
                return (NodeIdAddressBind) nbrs[ nbrs.Count-1 ];
            }
        }
        
        
        //**********************************************************************************************
        // Private Methods
        //**********************************************************************************************

        /// <summary>
        /// Determine if node B lies strictly between nodes A and C, taking into
        /// account the direction of this routing-set half.
        /// </summary>
        protected abstract bool LiesBetween( NodeIdAddressBind A, NodeIdAddressBind B, NodeIdAddressBind C );
    }

}