using System;
using System.Collections;
using System.Diagnostics;

using MessageRouting;
using Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald {

    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    /// <summary>
    /// Queue that holds all messages passed between nodes.
    /// Hacked together from TaroLocalEvent.
    /// </summary>
    public class TestQueue : MessageRouting.LocalEvent {

        private TestMessageQueue    mq;

        public TestQueue() {
            mq = new TestMessageQueue();
        }

        // Add a message to the queue
        public void RaiseEvent(long ms, MessageRouting.NodeIdAddressBind me, MessageRouting.RouteMsg message) {
            SkipNet.RouteMsg    rm = (SkipNet.RouteMsg) message;
            TestAddress         meAddr = (TestAddress) me.GetAddress;
            SkipNet.Node        receivingNode = meAddr.Node;

            //Console.WriteLine("Raising local event: delay={0}ms, type={1}", ms, rm.MessageType );
            QueueEntry p = new QueueEntry();
            p.Node = receivingNode;
            p.Message = rm;
            p.FireAt = mq.CurrentTime + ms;
            p.Background = rm.IsBackground();

            //Console.WriteLine("Event will fire at: {0}", p.FireAt);
            mq.Add( p );
        }

        public void RemoveEvent(MessageRouting.RouteMsg message) {
            mq.Remove(message);
        }

        // Process messages until all foreground messages have been processed
        public void ProcessAllForegroundMsgs() {
            while( mq.NumForegroundMsgs>0 ) {
                // Remove the next message from the queue
                QueueEntry p = mq.Dequeue();

                // Process this message on the appropriate node
                Log.SetIdentity( p.Node.GetNodeIdAddressBind.GetAddress.ToString() );
                p.Node.RaiseEventUpcall( p.Message );
                Log.SetIdentity( null );
            }
        }

        // Process messages for the given duration
        public void ProcessForTime( long duration ) {
            long endTime = mq.CurrentTime + duration;
            for(;;) {
                QueueEntry p = mq.Dequeue();
                if( p==null ) {
                    return;
                } else if( mq.CurrentTime<=endTime ) {
                    Log.SetIdentity( p.Node.GetNodeIdAddressBind.ToString() );
                    p.Node.RaiseEventUpcall( p.Message );
                    Log.SetIdentity( null );
                } else {
                    mq.Add(p);
                    return;
                }
            }
        }

        public long GetTime() {
            return mq.CurrentTime;
        }

        public void Clear() {
            mq.Clear();
        }
        public long GetNumMessagesSentToWire(MessageRouting.Node node) { return 0;}
        public long GetNumMessagesReceivedFromWire(MessageRouting.Node node) { return 0;}
    }


    /// <summary>
    /// An Entry in the TestMessageQueue
    /// </summary>
    public class QueueEntry {
        public SkipNet.Node         Node;
        public SkipNet.RouteMsg		Message;
        public long			        FireAt;
        public bool                 Background;
    }
	

    /// <summary>
    /// A queue of messages
    /// </summary>
    public class TestMessageQueue 
    {
        SortedList  list = new SortedList();
        Hashtable   table = new Hashtable();
        int         numMsgs = 0;
        int         numForegroundMsgs = 0;
        long        now = 0;    // Current time

        public long CurrentTime 
        {
            get { return now; }
        }

        public int NumForegroundMsgs 
        {
            get { return numForegroundMsgs; }
        }

        // Add the message to the queue
        public void Add( QueueEntry p ) 
        {

            // Add message to the list sorted by delivery time
            ArrayList listEntry = (ArrayList) list[p.FireAt];
            if( listEntry!=null ) 
            {
                listEntry.Add( p );
            } 
            else 
            {
                listEntry = new ArrayList();
                listEntry.Add( p );
                list.Add( p.FireAt, listEntry );
            }

            // Add the message to the table accessed by key
            Debug.Assert( ! table.ContainsKey(p.Message) );
            table[ p.Message ] = p;

            // Increment message counters
            numMsgs++;
            if(!p.Background) numForegroundMsgs++;
        }

        // Get the next event on the queue to be ready.
        // Returns null if no event on the queue is ready.
        public QueueEntry Dequeue() 
        {
            QueueEntry p=null;
            ArrayList listEntry;

            // No messages are to be processed
            if( list.Count==0 ) 
            {
                return null;
            }

            // Extract the next Message to process
            listEntry = (ArrayList) list.GetByIndex(0);
            p = (QueueEntry) listEntry[ listEntry.Count-1 ];

            // We got a message. Advance the current time.
            now = p.FireAt;
            
            // Remove the entry from the list
            listEntry.RemoveAt( listEntry.Count-1 );
            if( listEntry.Count==0 ) 
            {
                list.RemoveAt(0);
            }
            Debug.Assert( table.ContainsKey(p.Message) );
            table.Remove(p.Message);

            // Decrement message counters
            numMsgs--;
            if(!p.Background) numForegroundMsgs--;

            return p;
        }

        /// <summary>
        /// Remove an event from the event queue.
        /// </summary>
        /// <param name="Message">A Message acting as a "comparison key" that is equal to the
        /// message to be removed.</param>
        public QueueEntry Remove(MessageRouting.RouteMsg msg) 
        {
            // Retrieve the event matching the given key
            Debug.Assert( table.ContainsKey(msg) );
            QueueEntry p = (QueueEntry) table[msg];
            Debug.Assert( msg.Equals(p.Message) );
            
            // Remove the event
            ArrayList listEntry = (ArrayList) list[p.FireAt];
            Debug.Assert( listEntry!=null );
            Debug.Assert( listEntry.Contains(p) );
            listEntry.Remove(p);
            if( listEntry.Count==0 ) 
            {
                list.Remove(p.FireAt);
            }
            table.Remove(msg);

            // Decrement message counters
            numMsgs--;
            if(!p.Background) numForegroundMsgs--;

            // Return the event
            return p;
        }

        /// <summary>
        /// Clear all events from the queue
        /// </summary>
        public void Clear() 
        {
            list.Clear();
            table.Clear();
            numMsgs = 0;
            numForegroundMsgs = 0;
            now = 0;
        }
    }
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// For the simulator, we use a Node object as its physical address.
    /// </summary>
    public class TestAddress : MessageRouting.NodePhysicalAddress {
        SkipNet.Node     node;

        public TestAddress(SkipNet.Node node) {
            this.node = node;
        }

        public SkipNet.Node Node {
            get { return node; }
        }

        public override string Address {
            get { return node.NodeIdAddressBind.ToString(); }
            set { Debug.Assert(false, "Not implemented"); }
        }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////////

    public class TestTopology : MessageRouting.Topology {

        public TestTopology() {}

        // Measure the distance between a and b, relative to the topology of the network and an
        // appropriate distance measure
        public int Ping(NodePhysicalAddress a, NodePhysicalAddress b, bool disableCache) {
            // Check if node a or b has failed
            SkipNet.Node aNode = (SkipNet.Node) ((TestAddress) a).Node;
            SkipNet.Node bNode = (SkipNet.Node) ((TestAddress) b).Node;
            if( aNode.Failed || bNode.Failed ) {
                return -1;
            }

            return 1;
        }

        // Asynchronous call to ping node b
        public void BeginPing(NodePhysicalAddress a, NodePhysicalAddress b, bool disableCache, PingCallback cb) {
            int distance;
            distance = Ping(a, b, disableCache);
            cb(b, distance);
        }


        // This must return another seed node
        public NodePhysicalAddress GetClosest(NodePhysicalAddress n) {
            Debug.Assert( false, "Not Yet Implemented" );
            return null;
        }
    }
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////

    public class TestSend : MessageRouting.SendMessage {
        TestQueue       tq;
        TestTopology    tt;
        bool            dropReply;

        public TestSend( TestQueue tq, TestTopology tt ) {
            this.tq = tq;
            this.tt = tt;
        }

        public bool DropReply {
            set { dropReply = value; }
        }

        // Send a message from the source node to the destination node
        public bool Send(MessageRouting.NodeIdAddressBind source,
            MessageRouting.NodeIdAddressBind destination, MessageRouting.RouteMsg msg)
        {
            // If we're told to drop OverlayReplyMsg objects, do so
            SkipNet.RouteMsg snMsg = (SkipNet.RouteMsg) msg;
            if( (snMsg.Payload is OverlayReplyMsg) && dropReply ) {
                return true;
            }

            // Measure the delay between the source and the destination
            int delay = tt.Ping(source.GetAddress, destination.GetAddress, false);
            if( delay<0 ) {
                // If the delay is negative, one of the nodes has failed
                return false;
            }

            // Enqueue the message at the destination node, after the given delay
            tq.RaiseEvent(delay, destination, msg);

            return true;
        }

        // Send a message from the source node to the destination node asynchronously
        public void BeginSend(MessageRouting.NodeIdAddressBind source, 
            MessageRouting.NodeIdAddressBind destination, MessageRouting.RouteMsg msg, ReplyCallback cb)
        {
            bool success = Send(source, destination, msg);
            cb(destination, success, msg);
        }

    }

}