using System;
using System.Collections;
using System.Diagnostics;

using Microsoft.Research.Herald.SkipNet;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald {

    /// <summary>
    /// Application to test SkipNet.
    /// </summary>
    class SkipNetTest {

        internal const string   CATEGORY = "SkipNetTest";

        // Start-up code
        public static int Main(string[] args) {
            Console.WriteLine("***** Starting SkipNet Test *****");

            SkipNetTest test = new SkipNetTest();
            return test.Run(args);
        }

        // Private members
        TestQueue       tq;
        TestTopology    tt;
        TestSend        ts;
        Random          rng;
        int             numNodes;
        int             skipnetBase;

        // Main test code
        public int Run(string[] args) {
            //Log.SetCategoryFilterLevel(CATEGORY, LogLvl.Lo);
            //Log.SetCategoryFilterLevel(SkipNet.Node.CATEGORY, LogLvl.Med);
            //Log.SetCategoryFilterLevel(SkipNet.JoinRequest.CATEGORY, LogLvl.Lo);
            Log.SetCategoryFilterLevel(SkipNet.LeafSetFixup.CATEGORY, LogLvl.Hi);
            //Log.SetCategoryFilterLevel(SkipNet.FixupHigher.CATEGORY, LogLvl.Lo);

            tq = new TestQueue();
            tt = new TestTopology();
            ts = new TestSend(tq, tt);
            rng = new Random(0);
            SkipNet.Parameters.SetSeed(0);

            SkipNet.Parameters.b = skipnetBase = 8;
            SkipNet.Parameters.pingEnabled = false;
            SkipNet.Parameters.proximityTableEnabled = false;
            SkipNet.Parameters.cTableEnabled = false;
            SkipNet.Parameters.clbTableEnabled = false;
            SkipNet.Parameters.maxHeight = 32;

            NumericIdTest();

            RTableTest();
            tq.Clear();

            NeighbourTest();
            tq.Clear();

            PiggybackTest();
            tq.Clear();

            RequestReplyTest();
            tq.Clear();

            FailoverTest();
            tq.Clear();

            Console.WriteLine("All SkipNet tests passed");
            /*Console.WriteLine("[... Press Enter ...]");
            Console.ReadLine();*/
            return 0;
        }


        ////////////////////////////////////////////////////////////////////////////////////////////////////////


        void NumericIdTest() {
            NumericId A = new NumericId(), B = new NumericId(), C;
            A.ReadFromString("0000");
            B.ReadFromString("7777");

            C = A.Difference(B);
            
            /*
            // Non-absolute-value test
            if( ! C.ToString().Equals("0001") ) {
                Log.Write(CATEGORY, LogLvl.Hi, "NumericId subtraction failed");
                Debug.Assert( false );
            }
            */

        
            // Absolute-value test
            if( ! C.ToString().Equals("7777") ) {
                Log.Write(CATEGORY, LogLvl.Hi, "NumericId subtraction failed");
                Debug.Assert( false );
            }
        }


        ////////////////////////////////////////////////////////////////////////////////////////////////////////


        // Test that the R-Table is being built properly
        void RTableTest() {
            numNodes = 512;
            Node[] allNodes = new Node[ numNodes ];

            Log.Write(CATEGORY, LogLvl.Hi, "***** Testing Ring Construction *****");

            TestAddress firstNodeAddr = null;
            for( int i=0; i<numNodes; i++ ) {
                // Generate the name of the i'th node
                string nameId;
                nameId = String.Format("node-{0}", rng.Next()%1000000000);
                //nameId = String.Format("node-{0:0000}", i);

                // Create the i'th node
                Node node = new Node(nameId);
                TestAddress addr = new TestAddress(node);
                node.InitNode(tq, ts, tt, addr, firstNodeAddr);
                node.ActivateNode( false );
                allNodes[i] = node;
                Log.Write(CATEGORY, LogLvl.Med, "Creating Node {0}, {1}", nameId, node.NodeIdAddressBind.NumericId);

                // Maintain the address of the first node
                if( i==0 ) firstNodeAddr=addr;

                // Process messages so that this node can join
                tq.ProcessAllForegroundMsgs();
            }
            
            // Generate the array of rings at level i
            for( int h=0;; h++ ) {
                Log.Write(CATEGORY, LogLvl.Hi, "--- Checking rings at level {0} ---", h);
                ArrayList[] rings = GenerateRings(allNodes, h);
                if( rings==null ) break;
                CheckRings(allNodes, h, rings);
            }

            Log.Write(CATEGORY, LogLvl.Hi, "R-Table check complete!");
        }

        // Generate the correct array of rings at level height.
        // If all the rings contain only 1 node, return null.
        ArrayList[] GenerateRings(Node[] allNodes, int height) {
            int	numRings = (int) Math.Pow(skipnetBase, height);
            int	i;
            bool nonTrivialRingExists=false;

            // Create the array of rings
            ArrayList[] rings = new ArrayList[ numRings ];

            // Insert the nodes into their rings
            foreach( Node n in allNodes ) {
                int ringNum=0;
                for( i=0; i<height; i++ ) {
                    ringNum = (ringNum*skipnetBase) + n.NodeIdAddressBind.NumericId.GetDigit(i);
                }

                // Insert the node into its ring, creating the ring array if necessary
                if( rings[ringNum]==null )
                    rings[ringNum] = new ArrayList();
                rings[ringNum].Add(n);
            }

            // Sort the rings
            NodeComparer comparer = new NodeComparer();
            for( i=0; i<numRings; i++ ) {
                if( rings[i]==null ) continue;
                rings[i].Sort( comparer );
                if( rings[i].Count>1 ) nonTrivialRingExists = true;
            }

            if( nonTrivialRingExists ) {
                return rings;
            } else {
                return null;
            }
        }

        // Check that the nodes' actual pointers match the rings
        void CheckRings(Node[] allNodes, int height, ArrayList[] rings) {
            int maxNbrs;
            
            // Compute the maxmimum number of neighbours each node should have at this level
            if( height==0 ) {
                if( SkipNet.Parameters.denseRTable ) {
                    maxNbrs = ( SkipNet.Parameters.leafSetHalfSize > skipnetBase-1 ) ?
                        SkipNet.Parameters.leafSetHalfSize : skipnetBase-1;
                } else {
                    maxNbrs = SkipNet.Parameters.leafSetHalfSize;
                }
            } else {
                maxNbrs = ( SkipNet.Parameters.denseRTable ) ? skipnetBase-1 : 1;
            }

            // Check each ring
            for( int whichRing=0; whichRing<rings.Length; whichRing++ ) {
                ArrayList ring = rings[whichRing];

                if( ring==null || ring.Count<=1 ) {
                    continue;
                } else {
                    Log.Write(CATEGORY, LogLvl.Med, "Checking ring {0}", whichRing);
                }

                // Compute the actual number of neighbours each node should have in this ring
                int numNbrs = (ring.Count-1<maxNbrs) ? ring.Count-1 : maxNbrs;

                for( int i=0; i<ring.Count; i++ ) {
                    Node n = (Node) ring[i];
                    LeafSet ls = n.RoutingTable.Neighbours[height];
                    ArrayList left = ls.LeftHalf.GetAllNbrs();
                    ArrayList right = ls.RightHalf.GetAllNbrs();
                    Log.Write(CATEGORY, LogLvl.Lo, "Node {0}, {1}", n.NodeIdAddressBind.NodeId, n.NodeIdAddressBind.NumericId);

                    // Verify that this node has the correct number of neighbours
                    Debug.Assert( left.Count==numNbrs );
                    Debug.Assert( right.Count==numNbrs );

                    // Examine all the neighbours of this node and check them
                    for( int j=1; j<=numNbrs; j++ ) {
                        // Determine who the j'th neighbour should be
                        int l = i-j;    if(l<0)             l+=ring.Count;
                        int r = i+j;    if(r>=ring.Count)   r-=ring.Count;
                        NodeIdAddressBind expectedL = ((Node) ring[l]).NodeIdAddressBind;
                        NodeIdAddressBind expectedR = ((Node) ring[r]).NodeIdAddressBind;
                        
                        // Determine who the j'th neighbour really is
                        NodeIdAddressBind actualL = (NodeIdAddressBind) left[j-1];
                        NodeIdAddressBind actualR = (NodeIdAddressBind) right[j-1];

                        if( ! actualL.Equals(expectedL) ) {
                            Log.Write(CATEGORY, LogLvl.Hi, "Left pointer should be {0} but is {1}",
                                expectedL, actualL);
                            Debug.Assert( false );
                        }

                        if( ! actualR.Equals(expectedR) ) {
                            Log.Write(CATEGORY, LogLvl.Hi, "Right pointer should be {0} but is {1}",
                                expectedR, actualR);
                            Debug.Assert( false );
                        }

                        l=r=0;      // Dummy statement for putting a breakpoint on
                    }
                }
            }
        }

        /// <summary>
        /// Comparer class for comparing nodes
        /// </summary>
        class NodeComparer : IComparer {
            public int Compare( object x, object y ) {
                Node nx = (Node) x, ny = (Node) y;
                NodeId ix = nx.NodeIdAddressBind.NodeId, iy = ny.NodeIdAddressBind.NodeId;
                return ix.CompareTo(iy);
            }
        }


        ////////////////////////////////////////////////////////////////////////////////////////////////////////


        // Test that the neighbour table is being built properly
        void NeighbourTest() {
            numNodes = 512;
            Node[] allNodes = new Node[ numNodes ];

            Log.Write(CATEGORY, LogLvl.Hi, "***** Testing Neighbour Table *****");

            TestAddress firstNodeAddr = null;
            int i;
            for( i=0; i<numNodes; i++ ) {
                // Generate the name of the i'th node
                string nameId;
                nameId = String.Format("node-{0}", rng.Next()%1000000000);
                //nameId = String.Format("node-{0:0000}", i);

                // Create the i'th node
                Node node = new Node(nameId);
                TestAddress addr = new TestAddress(node);
                node.InitNode(tq, ts, tt, addr, firstNodeAddr);
                node.ActivateNode( false );
                allNodes[i] = node;
                Log.Write(CATEGORY, LogLvl.Med, "Creating Node {0}, {1}", nameId, node.NodeIdAddressBind.NumericId);

                // Maintain the address of the first node
                if( i==0 ) firstNodeAddr=addr;

                // Process messages so that this node can join
                tq.ProcessAllForegroundMsgs();
            }

            // Check that the R-Table and neighbour table are in 1-1 correspondence
            for( i=0; i<numNodes; i++ ) {
                // Build an array containing all R-Table entries
                Hashtable rTableNbrs = new Hashtable();
                SkipNet.LeafSet[] rtable = allNodes[i].RoutingTable.Neighbours;
                for( int h=0; h<rtable.Length; h++ ) {
                    ArrayList levelHNbrs = rtable[h].GetAllNbrs();
                    foreach( NodeIdAddressBind n in levelHNbrs ) {
                        if( ! rTableNbrs.ContainsKey(n) )
                            rTableNbrs.Add(n, null);
                    }
                }

                // Check that all of the R-Table neighbours are in the neighbour table
                foreach( NodeIdAddressBind n in rTableNbrs.Keys ) {
                    Debug.Assert( allNodes[i].IsNeighbour(n) );
                }

                // Check that all entries in the neighbour table are in the R-Table
                ArrayList nbrTable = allNodes[i].GetAllNeighbours();
                foreach( NodeIdAddressBind n in nbrTable ) {
                    Debug.Assert( rTableNbrs.ContainsKey(n) );
                }
            }

            Log.Write(CATEGORY, LogLvl.Hi, "Neighbour table check complete!");
        }


        ////////////////////////////////////////////////////////////////////////////////////////////////////////


        // Test that failover works properly
        void FailoverTest() {
            int i, numOrgs;

            Log.Write(CATEGORY, LogLvl.Hi, "***** Testing Failover *****");
            //Log.SetCategoryFilterLevel(SkipNet.CLBTable.CATEGORY, LogLvl.Lo);
            //Log.SetCategoryFilterLevel(SkipNet.GatherCLBNbrs.CATEGORY, LogLvl.Lo);
            Log.SetCategoryFilterLevel(SkipNet.LeafSetFixup.CATEGORY, LogLvl.Disabled);
            
            SkipNet.Parameters.clbTableEnabled = true;
            SkipNet.Parameters.clbTableUpdatePeriod = 60*1000;
            SkipNet.Parameters.pingEnabled = true;
            SkipNet.Parameters.neighbourPingPeriod = 60*1000;

            // Create organizations
            numOrgs = 5;
            string[] orgNames = new String[ numOrgs ];
            for( i=0; i<numOrgs; i++ ) {
                orgNames[i] = String.Format("{0}", rng.Next()%1000000000);
            }

            // Create the nodes
            numNodes = 512;
            Node[] allNodes = new Node[ numNodes ];
            
            TestAddress firstNodeAddr = null;
            for( i=0; i<numNodes; i++ ) {
                // Choose an org for this node
                int orgNum = rng.Next() % numOrgs;

                // Generate the name of the i'th node
                int nodeNum = rng.Next() % 1000000000;
                string nameId;
                nameId = String.Format("org{0}/node{1}", orgNames[orgNum], nodeNum);
                //nameId = String.Format("node-{0:0000}", i);

                // Create the i'th node
                Node node = new Node(nameId);
                TestAddress addr = new TestAddress(node);
                node.InitNode(tq, ts, tt, addr, firstNodeAddr);
                allNodes[i] = node;
                Log.Write(CATEGORY, LogLvl.Med, "Creating Node {0}, {1}", nameId, node.NodeIdAddressBind.NumericId);

                // Maintain the address of the first node
                if( i==0 ) firstNodeAddr=addr;

                // Process messages so that this node can join
                tq.ProcessAllForegroundMsgs();
            }
            
            // Generate the CLB Table on all nodes
            for( i=0; i<numNodes; i++ ) {
                Log.SetIdentity( allNodes[i].NodeIdAddressBind.ToString() );
                allNodes[i].ActivateNode( false );
            }
            
            // Run a series of failover tests
            for( int testNum=0; testNum<5; testNum++ ) {
                tq.ProcessForTime( SkipNet.Parameters.neighbourPingPeriod + SkipNet.Parameters.clbTableUpdatePeriod );

                // Choose a start node
                Log.Write(CATEGORY, LogLvl.Hi, "--- CLB Failover Test {0} ---", testNum+1);
                Node startNode = allNodes[ rng.Next() % numNodes ];

                // Choose a destination id
                MessageRouting.NodeId destId;
                string destStr;
                int orgNum = rng.Next() % (numOrgs+2);
                if( orgNum==numOrgs+1 ) {
                    // Leafset test
                    destStr = String.Format("org{0}/dest", rng.Next());
                    destId = new SkipNet.NodeId(destStr);
                } else if( orgNum==numOrgs ) {
                    // CLB over all nodes test
                    destStr = String.Format("!clb-test-{0}", testNum);
                    destId = new CLBId(destStr);
                } else {
                    // CLB within a domain test
                    destStr = String.Format("org{0}!clb-test-{1}", orgNames[orgNum], testNum);
                    destId = new CLBId(destStr);
                }                

                // Route the first message to that destination id
                CLBTestMsg testMsg = new CLBTestMsg();
                startNode.RouteMessage(testMsg, destId);
                tq.ProcessAllForegroundMsgs();

                // Get the identity of the original destination
                Node originalDest = clbTestDest;
                Debug.Assert( originalDest!=null );

                // Get the failover neighbours from this node
                ArrayList failoverNbrs = originalDest.GetFailoverNeighbours( destId );
                Debug.Assert( failoverNbrs!=null && failoverNbrs.Count>=SkipNet.Parameters.clbTableNbrListSize );

                // Fail the original destination node
                originalDest.Failed = true;

                // Route again to the same destination
                testMsg = new CLBTestMsg();
                startNode.RouteMessage(testMsg, destId);
                tq.ProcessAllForegroundMsgs();

                // Should arrive at the first failover neighbour
                Node failoverDest = clbTestDest;
                NodeIdAddressBind firstFailoverNbr = (NodeIdAddressBind) failoverNbrs[0];
                if( ! failoverDest.NodeIdAddressBind.Equals(firstFailoverNbr) ) {
                    Log.Write(CATEGORY, LogLvl.Hi, "Failover Test Error!");
                    Debug.Assert( false );
                }

                Log.Write(CATEGORY, LogLvl.Med, "Success. Repairing Rings.");
            }

            // Disable the options that we enabled
            SkipNet.Parameters.clbTableEnabled = false;
            SkipNet.Parameters.pingEnabled = false;
        }

        public static SkipNet.Node clbTestDest;

        public class CLBTestMsg : MessageRouting.OverlayMsg {
            public override void ArrivedAtDestination( MessageRouting.Node localNode ) {
                Node node = (SkipNet.Node) localNode;
                SkipNetTest.clbTestDest = node;
            }
        }


        ////////////////////////////////////////////////////////////////////////////////////////////////////////


        // Test the Request-Reply mechanism
        public int CallbackCounter=0;
        void RequestReplyTest() {
            // Create a SkipNet to work with.
            numNodes = 512;
            Node[] allNodes = new Node[ numNodes ];

            Log.Write(CATEGORY, LogLvl.Hi, "***** Testing Request-Reply Mechanism *****");

            TestAddress firstNodeAddr = null;
            for( int i=0; i<numNodes; i++ ) {
                // Generate the name of the i'th node
                string nameId;
                nameId = String.Format("node-{0}", rng.Next()%1000000000);
                //nameId = String.Format("node-{0:0000}", i);

                // Create the i'th node
                Node node = new Node(nameId);
                TestAddress addr = new TestAddress(node);
                node.InitNode(tq, ts, tt, addr, firstNodeAddr);
                node.ActivateNode( false );
                allNodes[i] = node;
                Log.Write(CATEGORY, LogLvl.Med, "Creating Node {0}, {1}", nameId, node.NodeIdAddressBind.NumericId);

                // Maintain the address of the first node
                if( i==0 ) firstNodeAddr=addr;

                // Process messages so that this node can join
                tq.ProcessAllForegroundMsgs();
            }

            // Run for 6 minutes, in order to garbage collect request-reply state
            tq.ProcessForTime( 6*60*1000 );

            // Choose random source and destination nodes
            Node source=allNodes[rng.Next()%numNodes];
            Node dest=allNodes[rng.Next()%numNodes];


            // Send a message from source to dest. Reply immediately.
            TestRequest request = new TestRequest(source,
                new TestRequestCB(this, true),
                30*1000, source.NodeIdAddressBind);
            source.RouteMessage(request, dest.NodeIdAddressBind.NodeId);
            tq.ProcessAllForegroundMsgs();
            Debug.Assert( CallbackCounter==1 );     // Ensure that the callback is called with a reply
            tq.ProcessForTime(360*1000);
            Debug.Assert( CallbackCounter==1 );


            // Send a message from source to dest but drop reply on first attempt.
            request = new TestRequest(source,
                new TestRequestCB(this, true),
                30*1000, source.NodeIdAddressBind);
            ts.DropReply = true;
            source.RouteMessage(request, dest.NodeIdAddressBind.NodeId);
            tq.ProcessForTime(29*1000);
            Debug.Assert( CallbackCounter==1 );     // Reply is dropped so we should not have received it yet
            ts.DropReply = false;
            tq.ProcessForTime(5*1000);
            Debug.Assert( CallbackCounter==2 );     // Automatic retry is sent and so now we should have received reply


            // Send a message from source to dest but never reply
            request = new TestRequest(source,
                new TestRequestCB(this, false),
                30*1000, source.NodeIdAddressBind);
            ts.DropReply = true;
            source.RouteMessage(request, dest.NodeIdAddressBind.NodeId);
            tq.ProcessForTime(4*29*1000);
            Debug.Assert( CallbackCounter==2 );     // Replies are all dropped so we should not receive them
            tq.ProcessForTime(5*1000);
            Debug.Assert( CallbackCounter==3 );     // Request times out
            ts.DropReply = false;

        }


        ////////////////////////////////////////////////////////////////////////////////////////////////////////

        Node pingSource;
        NodeIdAddressBind pingDest;
        Object pingObj;
        bool gotPing;

        void PiggybackTest() {
            // Create a SkipNet to work with.
            numNodes = 512;
            Node[] allNodes = new Node[ numNodes ];

            Log.Write(CATEGORY, LogLvl.Hi, "***** Testing Piggybacked Pings *****");
            SkipNet.Parameters.pingEnabled = true;

            TestAddress firstNodeAddr = null;
            PingSendUpcall sendUpcall = new PingSendUpcall( TestPingSendUpcall );
            PingReceiveUpcall receiveUpcall = new PingReceiveUpcall( TestPingReceiveUpcall );
            for( int i=0; i<numNodes; i++ ) {
                // Generate the name of the i'th node
                string nameId;
                nameId = String.Format("node-{0}", rng.Next()%1000000000);
                //nameId = String.Format("node-{0:0000}", i);

                // Create the i'th node
                Node node = new Node(nameId);
                TestAddress addr = new TestAddress(node);
                node.InitNode(tq, ts, tt, addr, firstNodeAddr);
                node.ActivateNode( false );
                // node.RegisterPingNotificationUpcall(sendUpcall, receiveUpcall, "SkipNetTest");
                node.RegisterPingNotificationUpcall(sendUpcall, receiveUpcall, 14); //JD 9/2/03 -- see previous hack for why this hack
                allNodes[i] = node;
                Log.Write(CATEGORY, LogLvl.Med, "Creating Node {0}, {1}", nameId, node.NodeIdAddressBind.NumericId);

                // Maintain the address of the first node
                if( i==0 ) firstNodeAddr=addr;

                // Process messages so that this node can join
                tq.ProcessAllForegroundMsgs();
            }

            for( int j=0; j<10; j++ ) {
                Log.Write(CATEGORY, LogLvl.Hi, "--- Piggyback test {0} ---", j+1);

                pingSource = allNodes[ rng.Next(numNodes) ];
                ArrayList allNbrs = pingSource.GetAllNeighbours();
                pingDest = (NodeIdAddressBind) allNbrs[ rng.Next(allNbrs.Count) ];
                //pingObj    = "Piggyback Me!"; // JD 9/2/03 HACK HACK
                byte[] foo = new byte[4];
                foo[0] = 0x1;
                foo[1] = 0x2;
                foo[2] = 0x3;
                foo[3] = 0x4;
                pingObj = foo;
                gotPing = false;
                
                tq.ProcessForTime( SkipNet.Parameters.neighbourPingPeriod );
                Debug.Assert( gotPing );
                Log.Write(CATEGORY, LogLvl.Med, "Node {0} received ping from {1}",
                    pingDest, pingSource.NodeIdAddressBind);
            }

            SkipNet.Parameters.pingEnabled = false;
        }

        Object TestPingSendUpcall( Node localNode, NodeIdAddressBind n ) {
            if( pingSource==null || localNode!=pingSource ) {
                return null;
            }
            if( ! n.Equals(pingDest) ) {
                return null;
            }
            return pingObj;
        }

        void TestPingReceiveUpcall( Node localNode, NodeIdAddressBind source, Object piggybackObj ) {
            Debug.Assert( localNode.NodeIdAddressBind==pingDest );
            Debug.Assert( source.Equals(pingSource.NodeIdAddressBind) );
            Debug.Assert( piggybackObj==pingObj );
            gotPing = true;
        }
    }

    
    /// <summary>
    /// Request Message
    /// </summary>
    class TestRequest : MessageRouting.OverlayRequestMsg {
        NodeIdAddressBind source;
        const int NUM_RETRIES = 3;

        public TestRequest(Node localNode, MessageRouting.OverlayReplyCallback cb, long delayMs,
            NodeIdAddressBind source) : base(localNode, cb, delayMs, false)
        {
            this.source = source;
            base.MaxRetries = NUM_RETRIES;
        }

        public override void ArrivedAtDestination( MessageRouting.Node localNode ) {
            TestReply reply = new TestReply(this);
            localNode.SendReply(reply, source);
        }
    }


    /// <summary>
    /// Reply message
    /// </summary>
    class TestReply : MessageRouting.OverlayReplyMsg {
        public TestReply( MessageRouting.OverlayRequestMsg requestMsg ) : base(requestMsg) {}
    }


    /// <summary>
    /// Request callback
    /// </summary>
    class TestRequestCB : MessageRouting.OverlayReplyCallback {
        SkipNetTest test;
        bool shouldHaveReply;

        public TestRequestCB( SkipNetTest test, bool shouldHaveReply ) {
            this.test = test;
            this.shouldHaveReply = shouldHaveReply;
        }

        public override void Execute( MessageRouting.Node localNode ) {
            if( shouldHaveReply ) {
                Debug.Assert( this.reply!=null );
            } else {
                Debug.Assert( this.reply==null );
            }
            test.CallbackCounter++;
        }
    }

}