using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Xml.Serialization;
using MessageRouting;
using Microsoft.Research.Herald.Common;
using ControlFlow = Microsoft.Research.Herald.Common.ControlFlow;
using Microsoft.Research.Herald.Logger;

namespace Microsoft.Research.Herald.TLog
{
	/// <summary>
	/// "Timed" logging package.
	/// Generates log messages that contain time stamps.
	/// Two kinds of timed logging are available:
	/// - Local log messages employ the host-local clock to generate time stamps.
	/// - Global log messages send log messages to a central node, where they are
	///   logged using the central node's clock and written to a designated log file.
	/// </summary>
	public class TLog
	{
            // *** Static TLog package ***

            static LocalEvent localEvent;

            // Identity of the central node to which global log messages will be sent.
            // This variable is set during an initial clock synchronization phase that the
            // central node must initiate with all participating log nodes.
            public static NodeIdAddressBind globalNode = null;
            // Table of latencies to log nodes.
            // Used to adjust clock times to account for log message latency.
            public static Hashtable nodeMsgLatencies;

            static Hashtable enabledLevelTable;
            static LogLvl defaultLevel = LogLvl.Hi;


            /// <summary>
            /// Static initialization.  
            /// Must be called once before this package is used.
            /// </summary>
            public static void Init(LocalEvent localEventParam) {
                localEvent = localEventParam;
                enabledLevelTable = new Hashtable();
                defaultLevel = LogLvl.Hi;
                nodeMsgLatencies = new Hashtable();
            }

            /// <summary>
            /// Called by central logging node (assumed to be node 0 in 'exEnv'.
            /// Contracts all nodes known to 'exEnv' and measures RTT to each one.
            /// This information will be used to adjust the clock time that is
            /// listed for a log message received from a given node.
            /// </summary>
            public static void InitGlobalLoggingNode(SkipNet.Node node, ExecutionEnvironment exEnv) {
                // Figure out how much time to adjust log message times by
                NodeIdAddressBind[] logNodes = new NodeIdAddressBind[exEnv.NodeCount];
                for (int i = 0; i < logNodes.Length; i++) {
                    logNodes[i] = exEnv.NodeIdAddrBind(i);
                }
                Node overlayNode = exEnv.GetNode(0);

                RTTSyncReplyCallback cb = new RTTSyncReplyCallback(overlayNode, exEnv, logNodes, localEvent, 0);
                RTTSyncMsg msg = new RTTSyncMsg(overlayNode, cb);
                lock (exEnv.OverlayLock) {
                    overlayNode.SendMessageDirect(msg, logNodes[0]);
                }
                while (!ClockSyncDone) {
                    exEnv.GoFor(1000);
                }
            }


            /// <summary>
            /// Write a message to the log that is output at the given logging level.
            /// The 'category' for this message is as given.
            /// </summary>
            /// <param name="category">The category for this message</param>
            /// <param name="level">The severity level of this message</param>
            /// <param name="formatString">The string format of this message</param>
            /// <param name="args">The parameters for the string formatting</param>
            public static void Write( string category, LogLvl level, string formatString, params object[] args ) {
                Debug.Assert( level!=LogLvl.Disabled );

                string s = WriteStr(category, level, formatString, args);
                if (s == null) {
                    return;
                }
                else {
                    Console.WriteLine("{0} {1}", GetTime(0), s);
                }
            }

            public static void WriteGlobal( Node overlayNode, string category, LogLvl level, string formatString, params object[] args ) {
                Debug.Assert( level!=LogLvl.Disabled );

                string s = WriteStr(category, level, formatString, args);
                if (s == null) {
                    return;
                }
                else {
                    GlobalLogMsg msg = new GlobalLogMsg(overlayNode, s);
                    overlayNode.SendMessageDirect(msg, globalNode);
                }
            }

            /// <summary>
            /// Specify file to which global logging messages should be written.
            /// </summary>
            /// <param name="filename">The name of the logging file</param>
            public static void SetGlobalOutputFile( string filename ) {}

            /// <summary>
            /// Set the default filter level. Messages at this level or above
            /// will be printed by default, unless overridden with SetCategoryFilterLevel().
            /// </summary>
            public static void SetDefaultFilterLevel( LogLvl level ) {
                defaultLevel = level;
            }

            /// <summary>
            /// Enable output for the given category, at the given level and above.
            /// </summary>
            public static void ClearCategoryFilters() {
                enabledLevelTable = new Hashtable();
            }

            /// <summary>
            /// Enable output for the given category, at the given level and above.
            /// </summary>
            public static void SetCategoryFilterLevel( string category, LogLvl level ) {
                enabledLevelTable[category] = level;
            }

            /// <summary>
            /// Determine if a message with the given category and level should be filtered.
            /// </summary>
            /// <returns>True if it should be filtered, False if it should be output</returns>
            private static bool ShouldFilter( string msgCategory, LogLvl msgLevel ) {
                int iMsgLevel = (int) msgLevel;
                int iFilterLevel;

                if( enabledLevelTable.ContainsKey(msgCategory) ) {
                    LogLvl filterLevel = (LogLvl) enabledLevelTable[msgCategory];
                    iFilterLevel = (int) filterLevel;
                } else {
                    iFilterLevel = (int) defaultLevel;
                }

                return (iMsgLevel > iFilterLevel);
            }

            
            // *** Instance-based (i.e. non-static) package ***

            private string     category; 
            
            public TLog(string category) {
                this.category = category;
            }

            /// <summary>
            /// Write a message to the log that is output at the given logging level.
            /// The 'category' for this message is determined by the instance.
            /// </summary>
            /// <param name="level">The severity level of this message</param>
            /// <param name="formatString">The string format of this message</param>
            /// <param name="args">The parameters for the string formatting</param>
            public void Write( LogLvl level, string formatString, params object[] args ) {
                Write( category, level, formatString, args );
            }


            // *** Internal routines ***
 
            /// <summary>
            /// Produce a message string.
            /// The 'category' for this message is as given.
            /// If the logging level is not high enough then return null.
            /// </summary>
            /// <param name="category">The category for this message</param>
            /// <param name="level">The severity level of this message</param>
            /// <param name="formatString">The string format of this message</param>
            /// <param name="args">The parameters for the string formatting</param>
            static string WriteStr( string category, LogLvl level, string formatString, params object[] args ) {
                Debug.Assert( level!=LogLvl.Disabled );

                // Check if this message should be filtered out
                if( ShouldFilter(category, level) ) {
                    return null;
                }
            
                // Output the message
                string s = String.Format(formatString, args);
                string identity = Log.GetIdentity();
                if( identity == null ) {
                    return String.Format("{0}: {1}", category, s);
                } else {
                    return String.Format("{0} Node {1}: {2}", category, identity, s);
                }
            }

            public static string GetTime(long timeAdjustment) {
                long t = localEvent.GetTime() - timeAdjustment;
                DateTime dateTime = new DateTime(10000*t);
                // 10000 converts the msecs retruned by GetTime() to the 100-nanosec unit expected by
                // the DateTime constructor
                return (t.ToString() + " (" + dateTime.ToLongTimeString() + ") ");
            }

            public static bool ClockSyncDone = false;


            /// <summary>
            /// Needed for XML serialization of messages.
            /// </summary>
            public class SerializationConfig {
                public const string LOG_NAMESPACE = "http://research.microsoft.com/sn/Herald/LOG";
 
                public static Type [] serializationTypes = new Type[] {
                                                                          typeof(RTTSyncMsg),
                                                                          typeof(RTTSyncReplyMsg),
                                                                          typeof(GlobalLogMsg)
                                                                      };
            }

            /// <summary>
            /// Message used to measure round-trip time to a node.
            /// </summary>
            [Serializable,
                XmlRootAttribute("RTTSyncMsg",Namespace=SerializationConfig.LOG_NAMESPACE,IsNullable=true)]
                public class RTTSyncMsg : OverlayRequestMsg {
        
                public NodeIdAddressBind Requestor;


                public RTTSyncMsg() {}

                /// <summary>
                /// Constructor.
                /// </summary>
                /// <param name="localNode"></param>
                public RTTSyncMsg( MessageRouting.Node localNode, OverlayReplyCallback cb): base(localNode, cb, ControlFlow.Parameters.RPCTimeout, false) {
                    this.Requestor = localNode.GetNodeIdAddressBind;
                }
 

                /// <summary>
                /// Simply reply to indicate we are alive.
                /// </summary>
                /// <param name="overlayNode"></param>
                public override void ArrivedAtDestination( MessageRouting.Node overlayNode ) {

                    //            Log.Write("TLog", LogLvl.Hi, "RTTSyncMsg {0}", overlayNode.GetNodeIdAddressBind.GetNodeId.ToString());

                    TLog.globalNode = this.Requestor;

                    OverlayReplyMsg replyMsg = new RTTSyncReplyMsg(overlayNode.GetNodeIdAddressBind, this);
                    overlayNode.SendReply(replyMsg, this.Requestor);
                }
            }

            /// <summary>
            /// Reply to an RTTSyncMsg.
            /// </summary>
            [Serializable,
                XmlRootAttribute("RTTSyncReplyMsg",Namespace=SerializationConfig.LOG_NAMESPACE,IsNullable=true)]
                public class RTTSyncReplyMsg : OverlayReplyMsg {

                public NodeIdAddressBind Replier;  // Node sending the reply message.

        
                public RTTSyncReplyMsg() {
                }

                public RTTSyncReplyMsg( NodeIdAddressBind replier, OverlayRequestMsg requestMsg ): base(requestMsg) {
                    this.Replier = replier;
                }
            }

            public class RTTSyncReplyCallback: OverlayReplyCallback {

                Node overlayNode;
                ExecutionEnvironment exEnv;
                NodeIdAddressBind[] logNodes;
                LocalEvent localEvent;
                int logNodesIndex;
                long startTime;
                long endTime;

                public RTTSyncReplyCallback(Node overlayNode, ExecutionEnvironment exEnv, NodeIdAddressBind[] logNodes, LocalEvent localEvent, int logNodesIndex) {
                    this.overlayNode = overlayNode;
                    this.exEnv = exEnv;
                    this.logNodes = logNodes;
                    this.localEvent = localEvent;
                    this.startTime = localEvent.GetTime();
                    this.logNodesIndex = logNodesIndex;
                }

                public override void Execute( MessageRouting.Node overlayNode ) {
                    // Record msg latency to a node.
                    this.endTime = this.localEvent.GetTime();
                    long rtt = this.endTime - this.startTime;
                    TLog.nodeMsgLatencies[this.logNodes[this.logNodesIndex].GetAddress.ToString()] = rtt / 2;

                    Log.Write("TLog", LogLvl.Hi, "RTTSyncReplyCallback on node {0} addr {1}: RTT: {2} at time {3:T}",
                        logNodesIndex, this.logNodes[this.logNodesIndex].GetAddress.ToString(), rtt, DateTime.Now);

                    // Initiate an RTT message interaction on the next node in logNodes.
                    this.logNodesIndex++;
                    if (this.logNodesIndex < this.logNodes.Length) {
                        RTTSyncReplyCallback cb = new RTTSyncReplyCallback(this.overlayNode, this.exEnv, this.logNodes, this.localEvent, this.logNodesIndex);
                        RTTSyncMsg msg = new RTTSyncMsg(this.overlayNode, cb);
                        lock (this.exEnv.OverlayLock) {
                            overlayNode.SendMessageDirect(msg, this.logNodes[this.logNodesIndex]);
                        }
                    }
                    else {
                        TLog.ClockSyncDone = true;
                    }
                }
            }


            /// <summary>
            /// This message is used to send a log message to the central logging node.
            /// </summary>
            [Serializable,
                XmlRootAttribute("GlobalLogMsg",Namespace=SerializationConfig.LOG_NAMESPACE,IsNullable=true)]
                public class GlobalLogMsg : OverlayMsg {

                public NodeIdAddressBind Sender;
                public string logMsg;

                public GlobalLogMsg() {}

                public GlobalLogMsg(Node overlayNode, string logMsg) {
                    this.Sender = overlayNode.GetNodeIdAddressBind;
                    this.logMsg = logMsg;
                }

                public override void ArrivedAtDestination( MessageRouting.Node overlayNode ) {
                    long lat = (long) TLog.nodeMsgLatencies[this.Sender.GetAddress.ToString()];
                    //Console.WriteLine("{0} {1}", TLog.GetTime(lat), this.logMsg);
                    Log.Write("tlog", LogLvl.Hi, "{0} {1}", TLog.GetTime(lat), this.logMsg);
                }
            }
        }
}
